unit ACRCrypto;

interface

uses SysUtils,Classes,

{$I ACRVer.inc}

// Accuracer units

 {$IFDEF DEBUG_LOG}
     ACRDebug,
 {$ENDIF}
     ACRExcept,
     ACRConst,
     ACRTypes
{$IFDEF ENCRYPTION_ON}
     ,ACRRng,
     ACRCipher,
     ACRCipher1,
     ACRHash,
     ACRDecUtil
{$ENDIF}
     ;

  procedure ACREncryptBuffer(const CryptoInfo: TACRCryptoInfo;
                                     Buffer: PChar; BufferSize: Integer);
  procedure ACRDecryptBuffer(const CryptoInfo: TACRCryptoInfo;
                                     Buffer: PChar; BufferSize: Integer);
  function ACRCountCRC(Value: Cardinal; Buffer: PChar; BufferSize: Integer): Cardinal;
  function ACRCountCRC16(Value: Word; Buffer: PChar; BufferSize: Integer): Word;
  procedure ACRGenerateRandomBuffer(Buffer: PChar; BufferSize: Integer);
  function ACRCreateCryptoHeader(const CryptoInfo: TACRCryptoInfo): TACRCryptoHeader;
  function ACRIsKeyValid(const CryptoHeader: TACRCryptoHeader; const CryptoInfo: TACRCryptoInfo): Boolean;

implementation


//------------------------------------------------------------------------------
// encrypt buffer
//------------------------------------------------------------------------------
procedure ACREncryptBuffer(const CryptoInfo: TACRCryptoInfo;
                                     Buffer: PChar; BufferSize: Integer);
{$IFDEF ENCRYPTION_ON}
var cr: TCipher;
{$ENDIF}
begin
 if (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
  Exit;
{$IFDEF ENCRYPTION_ON}
 case CryptoInfo.CryptoAlgorithm of
  ACR_Cipher_Rijndael_128:
   begin
     cr := TCipher_Rijndael.Create(CryptoInfo.Password,nil);
     cr.HashClass := THash_RipeMD128;
   end;
  ACR_Cipher_Rijndael_256:
     cr := TCipher_Rijndael.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Blowfish:
     cr := TCipher_Blowfish.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Twofish_128:
   begin
     cr := TCipher_Twofish.Create(CryptoInfo.Password,nil);
     cr.HashClass := THash_RipeMD128;
   end;
  ACR_Cipher_Twofish_256:
     cr := TCipher_Twofish.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Square:
     cr := TCipher_Square.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Single_8:
     cr := TCipher_1DES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Double_8:
     cr := TCipher_2DES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Double_16:
     cr := TCipher_2DDES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Triple_8:
     cr := TCipher_3DES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Triple_16:
     cr := TCipher_3DDES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Triple_24:
     cr := TCipher_3TDES.Create(CryptoInfo.Password,nil)
 else
  raise EACRException.Create(10699,ErrorLUnknownCryptoAlgorithm,
    [CryptoInfo.CryptoAlgorithm]);
 end;
 try
  case CryptoInfo.CryptoMode of
   ACR_Cipher_Mode_CTS: cr.Mode := cmCTS;
   ACR_Cipher_Mode_CBC: cr.Mode := cmCBC;
   ACR_Cipher_Mode_CFB: cr.Mode := cmCFB;
   ACR_Cipher_Mode_OFB: cr.Mode := cmOFB;
  else
   raise EACRException.Create(10701,ErrorLUnknownCryptoMode,
    [CryptoInfo.CryptoMode]);
  end; 
  cr.Initialized := False; 
  if (CryptoInfo.Password = '') then
   begin
    if (CryptoInfo.UseInitVector) then 
     cr.Init(CryptoInfo.KeyInfo.Key,CryptoInfo.KeyInfo.KeySize,@CryptoInfo.InitVector) 
    else
     cr.Init(CryptoInfo.KeyInfo.Key,CryptoInfo.KeyInfo.KeySize,cr.Vector);
   end 
  else
   begin 
    if (CryptoInfo.UseInitVector) then
     cr.InitKey(CryptoInfo.Password,@CryptoInfo.InitVector) 
    else
     cr.InitKey(CryptoInfo.Password,cr.Vector); 
   end;
  cr.EncodeBuffer(Buffer^,Buffer^,BufferSize); 
 finally 
  cr.Free;
 end; 
{$ELSE} 
 raise EACRException.Create(10708,ErrorLEncryptionIsSwitchedOff); 
{$ENDIF} 
end; // ACREncryptBuffer 
	
	
//------------------------------------------------------------------------------
// decrypt buffer
//------------------------------------------------------------------------------
procedure ACRDecryptBuffer(const CryptoInfo: TACRCryptoInfo; 
                                     Buffer: PChar; BufferSize: Integer);
{$IFDEF ENCRYPTION_ON} 
var cr: TCipher;
{$ENDIF} 
begin 
 if (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
  Exit; 
{$IFDEF ENCRYPTION_ON}
 case CryptoInfo.CryptoAlgorithm of 
  ACR_Cipher_Rijndael_128: 
   begin
     cr := TCipher_Rijndael.Create(CryptoInfo.Password,nil);
     cr.HashClass := THash_RipeMD128; 
   end; 
  ACR_Cipher_Rijndael_256: 
     cr := TCipher_Rijndael.Create(CryptoInfo.Password,nil); 
  ACR_Cipher_Blowfish: 
     cr := TCipher_Blowfish.Create(CryptoInfo.Password,nil); 
  ACR_Cipher_Twofish_128: 
   begin
     cr := TCipher_Twofish.Create(CryptoInfo.Password,nil);
     cr.HashClass := THash_RipeMD128; 
   end;
  ACR_Cipher_Twofish_256:
     cr := TCipher_Twofish.Create(CryptoInfo.Password,nil); 
  ACR_Cipher_Square:
     cr := TCipher_Square.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Single_8: 
     cr := TCipher_1DES.Create(CryptoInfo.Password,nil); 
  ACR_Cipher_Des_Double_8: 
     cr := TCipher_2DES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Double_16:
     cr := TCipher_2DDES.Create(CryptoInfo.Password,nil); 
  ACR_Cipher_Des_Triple_8:
     cr := TCipher_3DES.Create(CryptoInfo.Password,nil);
  ACR_Cipher_Des_Triple_16: 
     cr := TCipher_3DDES.Create(CryptoInfo.Password,nil); 
  ACR_Cipher_Des_Triple_24:
     cr := TCipher_3TDES.Create(CryptoInfo.Password,nil)
 else
  raise EACRException.Create(10700,ErrorLUnknownCryptoAlgorithm, 
    [CryptoInfo.CryptoAlgorithm]); 
 end; 
 try 
  case CryptoInfo.CryptoMode of
   ACR_Cipher_Mode_CTS: cr.Mode := cmCTS; 
   ACR_Cipher_Mode_CBC: cr.Mode := cmCBC; 
   ACR_Cipher_Mode_CFB: cr.Mode := cmCFB; 
   ACR_Cipher_Mode_OFB: cr.Mode := cmOFB;
  else
   raise EACRException.Create(10702,ErrorLUnknownCryptoMode,
    [CryptoInfo.CryptoMode]); 
  end;
  cr.Initialized := False;
  if (CryptoInfo.Password = '') then
   begin 
    if (CryptoInfo.UseInitVector) then 
     cr.Init(CryptoInfo.KeyInfo.Key,CryptoInfo.KeyInfo.KeySize,@CryptoInfo.InitVector)
    else
     cr.Init(CryptoInfo.KeyInfo.Key,CryptoInfo.KeyInfo.KeySize,cr.Vector);
   end
  else
   begin
    if (CryptoInfo.UseInitVector) then 
     cr.InitKey(CryptoInfo.Password,@CryptoInfo.InitVector)
    else
     cr.InitKey(CryptoInfo.Password,cr.Vector); 
   end;
  cr.DecodeBuffer(Buffer^,Buffer^,BufferSize);
 finally 
  cr.Free; 
 end; 
{$ELSE} 
 raise EACRException.Create(10709,ErrorLEncryptionIsSwitchedOff);
{$ENDIF} 
end; // ACRDecryptBuffer 
	
	
//------------------------------------------------------------------------------
// calculate CRC32
//------------------------------------------------------------------------------ 
function ACRCountCRC(Value: Cardinal; Buffer: PChar; BufferSize: Integer): Cardinal; 
begin
{$IFDEF ENCRYPTION_ON}
 Result := CRC32(Value,Buffer,BufferSize); 
{$ELSE} 
 raise EACRException.Create(10710,ErrorLEncryptionIsSwitchedOff); 
{$ENDIF}
end; // ACRCountCRC
	
	
//------------------------------------------------------------------------------
// calculate CRC16 
//------------------------------------------------------------------------------
function ACRCountCRC16(Value: Word; Buffer: PChar; BufferSize: Integer): Word; 
begin
{$IFDEF ENCRYPTION_ON}
 Result := CRC16(Value,Buffer,BufferSize);
{$ELSE}
 raise EACRException.Create(10710,ErrorLEncryptionIsSwitchedOff);
{$ENDIF}
end; // ACRCountCRC16


//------------------------------------------------------------------------------
// generate random buffer
//------------------------------------------------------------------------------
procedure ACRGenerateRandomBuffer(Buffer: PChar; BufferSize: Integer);
{$IFDEF ENCRYPTION_ON}
var rng:  TRandom;
    size: Integer;

 function GenRndString(Len: Integer): String;
 var i: Integer;
 begin
  Result := '';
  for i := 1 to Len do
   Result := Result + Chr(Random(MaxInt) mod 255+1);
 end;

{$ENDIF}
begin
{$IFDEF ENCRYPTION_ON}
 Size := 100;
 rng := TRandom_LFSR.Create(GenRndString(Size),2032,False,nil);
 try
   rng.Seed('',-1);
   rng.Buffer(Buffer^,BufferSize);
 finally
   rng.Free;
 end;
{$ELSE}
 raise EACRException.Create(10711,ErrorLEncryptionIsSwitchedOff);
{$ENDIF}
end; // ACRGenerateRandomBuffer


//------------------------------------------------------------------------------
// create crypto header
//------------------------------------------------------------------------------
function ACRCreateCryptoHeader(const CryptoInfo: TACRCryptoInfo): TACRCryptoHeader;
begin
 Result.CryptoAlgorithm := CryptoInfo.CryptoAlgorithm;
 if (Result.CryptoAlgorithm <> ACR_Cipher_None) then
  begin
{$IFDEF ENCRYPTION_ON}
   ACRGenerateRandomBuffer(@Result.ControlBlock,Sizeof(Result.ControlBlock));
   Result.ControlBlockCRC := ACRCountCRC(0,@Result.ControlBlock,Sizeof(Result.ControlBlock));
   Result.CryptoMode := CryptoInfo.CryptoMode;
   ACREncryptBuffer(CryptoInfo,@Result.ControlBlock,Sizeof(Result.ControlBlock));
   // ask password or key if crypto info was not set
   if (CryptoInfo.Password = '') then
     Result.CryptoAskPassword := 0
   else
     Result.CryptoAskPassword := 1;
{$ELSE}
 raise EACRException.Create(10713,ErrorLEncryptionIsSwitchedOff);
{$ENDIF}
  end;
end; // ACRCreateCryptoHeader


//------------------------------------------------------------------------------
// Return true if CryptoHeader can be decrypted by KeyInfo
//------------------------------------------------------------------------------
function ACRIsKeyValid(const CryptoHeader: TACRCryptoHeader; const CryptoInfo: TACRCryptoInfo): Boolean;
var ch: TACRCryptoHeader;
begin
 if (CryptoInfo.CryptoAlgorithm = ACR_Cipher_None) then
  Result := True
 else
  begin
{$IFDEF ENCRYPTION_ON}
   ch := CryptoHeader;
   ACRDecryptBuffer(CryptoInfo,@ch.ControlBlock,Sizeof(ch.ControlBlock));
   Result := (ch.ControlBlockCRC = ACRCountCRC(0,@ch.ControlBlock,Sizeof(ch.ControlBlock)));
{$ELSE}
 raise EACRException.Create(10714,ErrorLEncryptionIsSwitchedOff);
{$ENDIF}
  end;
end; // ACRIsKeyValid

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRCrypto> initialized');
{$ENDIF}

end.
