unit ACRNetwork;

interface

{$I ACRVer.inc}

uses
  Classes, SysUtils,
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}
{$IFDEF LINUX}
  Libc,
{$ENDIF}

// Accuracer units

{$IFDEF DEBUG_LOG}
  ACRDebug,
{$ENDIF}
  ACRConst,
  ACRExcept,
  ACRMemory;

const
  MAX_PORT = 2147483646;

  PF_INET         =  2;
  SOCK_DGRAM      =  2;
  SOCKET_ERROR    = -1;
  AF_INET         =  2;
  IPPROTO_UDP     = 17;             { user datagram protocol }
  INADDR_NONE     = -1;

  FD_SETSIZE      =  1; // default value = 64;

type

  TSocket = Integer;

  SunB = packed record
    s_b1, s_b2, s_b3, s_b4: Byte;
  end;

  SunW = packed record
    s_w1, s_w2: Word;
  end;

  PInAddr = ^TInAddr;
  TACRin_addr = record
    case integer of
      0: (S_un_b: SunB);
      1: (S_un_w: SunW);
      2: (S_addr: LongInt);
  end;
  TInAddr = TACRin_addr;

  TACRsockaddr_in = record
    case Integer of
      0: (sin_family: Word;
          sin_port:   Word;
          sin_addr:   TInAddr;
          sin_zero:   array[0..7] of Char);
      1: (sa_family:  Word;
          sa_data:    array[0..13] of Char)
  end;
  TSockAddr = TACRsockaddr_in;

  PHostEnt = ^THostEnt;
  TACRhostent = record
    h_name: PChar;
    h_aliases: ^PChar;
    h_addrtype: Smallint;
    h_length: Smallint;
    case Byte of
      0: (h_addr_list: ^PChar);
      1: (h_addr: ^PChar)
  end;
  THostEnt = TACRhostent;

  PFDSet = ^TFDSet;
  TFDSet = record
    fd_count: Integer;
    fd_array: array[0..FD_SETSIZE-1] of TSocket;
  end;

  PTimeVal = ^TTimeVal;
  TACRtimeval = record
    tv_sec: Longint;
    tv_usec: Longint;
  end;
  TTimeVal = TACRtimeval;

const
  WSADESCRIPTION_LEN     =   256;
  WSASYS_STATUS_LEN      =   128;
type
  TWSAData = record
    wVersion: Word;
    wHighVersion: Word;
    szDescription: array[0..WSADESCRIPTION_LEN] of Char;
    szSystemStatus: array[0..WSASYS_STATUS_LEN] of Char;
    iMaxSockets: Word;
    iMaxUdpDg: Word;
    lpVendorInfo: PChar;
  end;

{$IFDEF LINUX}

(*$HPPEMIT '#include <sys/socket.h>'*)

{$ENDIF}


// Forward Declarations

  TACRListenerThread = class;

////////////////////////////////////////////////////////////////////////////////
//
// TACRapiNetwork
//
////////////////////////////////////////////////////////////////////////////////

  TACRDataReceivedNotifyEvent = procedure(
                             Buffer:    PChar;
                             Count:     Integer;
                             FromHost:  String;
                             FromPort:  Integer
                                          ) of object;

  TACRDisconnectNotifyEvent = procedure(
                             FromHost:  String;
                             FromPort:  Integer
                                          ) of object;

  TACRapiNetwork = class (TObject)
   private
    FOnDataReceived:      TACRDataReceivedNotifyEvent;
    FOnDisconnect:        TACRDisconnectNotifyEvent;
    FSocket:              Integer;
    FPacketSize:          Integer;
    FLocalHost:           String;
    FLocalPort:           Integer;
    FRemoteHost:          String;
    FRemotePort:          Integer;
    FListener:            TACRListenerThread;
    FActive:              Boolean;
    FDisconnected:        Boolean;
    procedure SetActive(Value: Boolean);
    procedure SetRemoteHost(Host: String);
    function GetRemoteHost: String;
    procedure SetRemotePort(Port: Integer);
    function GetRemotePort: Integer;
    procedure SetLocalHost(Host: String);
    function GetLocalHost: String;
    procedure SetLocalPort(Port: Integer);
    function GetLocalPort: Integer;
    procedure StartListening;
    procedure StopListening;
    procedure Open;
    procedure Close;
   public
//   protected
    property OnDataReceived: TACRDataReceivedNotifyEvent
                                    read FOnDataReceived write FOnDataReceived;
    property OnDisconnect: TACRDisconnectNotifyEvent
                                    read FOnDisconnect write FOnDisconnect;
   public
    constructor Create;
    destructor Destroy; override;
    procedure SendBuffer(
                            Buffer:     PChar;
                            Count:      Integer
                         );
   public
    property Active: Boolean read FActive write SetActive;
    property RemoteHost: String read GetRemoteHost write SetRemoteHost;
    property RemotePort: Integer read GetRemotePort write SetRemotePort;
    property LocalHost: String read GetLocalHost write SetLocalHost;
    property LocalPort: Integer read GetLocalPort write SetLocalPort;
  end; // TACRapiNetwork


  TACRListenerThread = class(TThread)
  private
    FapiNetwork:  TACRapiNetwork;
  protected
    procedure Execute; override;
  public
    constructor Create(apiNetwork: TACRapiNetwork);
    destructor Destroy; override;
  end;// TACRListenerThread


  TACROnDisconnectThread = class(TThread)
  private
    FapiNetwork:    TACRapiNetwork;
  protected
    procedure Execute; override;
  public
    constructor Create(
                        apiNetwork:    TACRapiNetwork
                      );
    destructor Destroy; override;
  end;// TACROnDisconnectThread



// Procedures

function SocketError: Integer;
function LookupHostAddr(const hn: string): string;

{$IFDEF MSWINDOWS}
const
  winsocket = 'wsock32.dll';
  kernel32  = 'kernel32.dll';

function socket(af, Struct, protocol: Integer): TSocket; stdcall;
function bind(s: TSocket; var addr: TSockAddr; namelen: Integer): Integer; stdcall;
function sendto(s: TSocket; var Buf; len, flags: Integer;
                  var addrto: TSockAddr; tolen: Integer): Integer; stdcall;
function recvfrom(s: TSocket; var Buf; len, flags: Integer;
                  var from: TSockAddr; var fromlen: Integer): Integer; stdcall;
function select(nfds: Integer; readfds, writefds, exceptfds: PFDSet;
                  timeout: PTimeVal): Longint; stdcall;
function inet_addr(cp: PChar): Longint; stdcall; {PInAddr;}  { TInAddr }
function gethostbyname(name: PChar): PHostEnt; stdcall;
function htons(hostshort: Word): Word; stdcall;
function ntohs(netshort: Word): Word; stdcall;
function closesocket(s: TSocket): Integer; stdcall;
function WSAStartup(wVersionRequired: Word; var WSData: TWSAData): Integer; stdcall;
function WSACleanup: Integer; stdcall;
function WSAGetLastError: Integer; stdcall;
function TerminateThread(hThread: THandle; dwExitCode: Longword): Boolean; stdcall;
{$ENDIF}

var
  FNetworkThreads:     TThreadList;

implementation


////////////////////////////////////////////////////////////////////////////////
//
// TACRapiNetwork
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRapiNetwork.Create;
begin
  inherited Create;
  FActive := False;
  FDisconnected := False;
  FPacketSize := ACRDefaultPacketSize;
  FLocalHost := '';
  FLocalPort := ACRDefaultClientPort;
  FRemoteHost := ACRDefaultHost;
  FRemotePort := ACRDefaultServerPort;
  FListener := nil;
  StartListening;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRapiNetwork.Create>  Socket #'+IntToStr(Integer(FSocket)));
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRapiNetwork.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRapiNetwork.Destroy>  Socket #'+IntToStr(Integer(FSocket)));
{$ENDIF}
  StopListening;
  FDisconnected := True;
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRapiNetwork.Destroy - FINISHED');
{$ENDIF}
end;// Destoy


//------------------------------------------------------------------------------
// SetRemoteHost
//------------------------------------------------------------------------------
procedure TACRapiNetwork.SetRemoteHost(Host: String);
begin
  FRemoteHost := Host;
end;


//------------------------------------------------------------------------------
// GetRemoteHost
//------------------------------------------------------------------------------
function TACRapiNetwork.GetRemoteHost: String;
begin
  Result := FRemoteHost;
end;


//------------------------------------------------------------------------------
// SetRemotePort
//------------------------------------------------------------------------------
procedure TACRapiNetwork.SetRemotePort(Port: Integer);
begin
  FRemotePort := Port;
end;


//------------------------------------------------------------------------------
// GetRemotePort
//------------------------------------------------------------------------------
function TACRapiNetwork.GetRemotePort: Integer;
begin
  Result := FRemotePort;
end;


//------------------------------------------------------------------------------
// GetLocalHost
//------------------------------------------------------------------------------
function TACRapiNetwork.GetLocalHost: String;
begin
  Result := FLocalHost;
end;


//------------------------------------------------------------------------------
// GetLocalPort
//------------------------------------------------------------------------------
function TACRapiNetwork.GetLocalPort: Integer;
begin
  Result := FLocalPort;
end;


//------------------------------------------------------------------------------
// SetLocalPort
//------------------------------------------------------------------------------
procedure TACRapiNetwork.SetLocalPort(Port: Integer);
begin
  if FLocalPort = Port then Exit;
  if FListener <> nil then
    StopListening; // you must close socket to change Local parameter
  FLocalPort := Port;
  StartListening;
end;


//------------------------------------------------------------------------------
// SetLocalHost
//------------------------------------------------------------------------------
procedure TACRapiNetwork.SetLocalHost(Host: String);
begin
  if FLocalHost = Host then Exit;
  if FListener <> nil then
    StopListening; // you must close socket to change Local parameter
  FLocalHost := Host;
  StartListening;
end;


//------------------------------------------------------------------------------
// SetActive
//------------------------------------------------------------------------------
procedure TACRapiNetwork.SetActive(Value: Boolean);
begin
  if Value = FActive then Exit; // you must close socket to change Local parameters
  if Value then
    Open
  else
    Close;
  FActive := Value;
end;// SetActive


//------------------------------------------------------------------------------
// StartListening
//------------------------------------------------------------------------------
procedure TACRapiNetwork.StartListening;
begin
  if FListener <> nil then
    raise EACRException.Create(40028, ErrorRNetworkListenerStarted, [SocketError]);
  if Active = False then
    Open;
  FListener := TACRListenerThread.Create(self);
end; // StartListening


//------------------------------------------------------------------------------
// StopListening
//------------------------------------------------------------------------------
procedure TACRapiNetwork.StopListening;
begin 
  if FListener = nil then 
    raise EACRException.Create(40029, ErrorRNetworkListenerNotStarted, [SocketError]);
  FListener.Free;
  FListener := nil; 
  Close; 
  FActive := False;
end; // StopListening
 
	
//------------------------------------------------------------------------------ 
// Open
//------------------------------------------------------------------------------ 
procedure TACRapiNetwork.Open;
var 
  RetCode:    Integer;
  Addr:       TSockAddr; 
  Bound:      Boolean; 
  i:          Integer;
begin 
  // open socket 
  FSocket := socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP); 
  if FSocket = SOCKET_ERROR then 
    raise EACRException.Create(40032, ErrorRCannotOpenSocket, ['socket', SocketError]); 
  // set parameters
  Addr.sin_family := AF_INET;
  Addr.sin_addr.s_addr := inet_addr(pchar(LookupHostAddr(FLocalHost)));
  // binding
  Bound := False;
  for i:=FLocalPort to MAX_PORT do 
   begin
    Addr.sin_port := htons(FLocalPort); 
    if bind(FSocket, Addr, sizeof(Addr)) = SOCKET_ERROR then
      RetCode := SocketError 
    else 
     begin
      Bound := True; 
      break; // bound
     end; 
    if RetCode = 10048 then // port is already in use 
      inc(FLocalPort)       // search for another port
    else
      raise EACRException.Create(40032, ErrorRCannotOpenSocket, ['bind', RetCode]); 
   end; // binding 
  if not Bound then 
    raise EACRException.Create(40032, ErrorRCannotOpenSocket, ['bind: cannot bound', RetCode]); 
end; // Open 
 
 
//------------------------------------------------------------------------------
// Close
//------------------------------------------------------------------------------ 
procedure TACRapiNetwork.Close;
var
  RetCode:    Integer; 
begin
{$IFDEF MSWINDOWS}
  RetCode := closesocket(FSocket); 
{$ENDIF} 
{$IFDEF LINUX} 
  RetCode := Libc.__close(FSocket);
{$ENDIF}
  if RetCode = SOCKET_ERROR then 
    raise EACRException.Create(40033, ErrorRCannotCloseSocket, [SocketError]);
end; // Close
 
 
//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRapiNetwork.SendBuffer( 
                          Buffer: PChar; 
                          Count:  Integer); 
var 
  Flags:  Integer;
  Addr: TSockAddr; 
begin 
  // set parameters 
  Flags := 0;
  Addr.sin_family := AF_INET;
  Addr.sin_addr.s_addr := inet_addr(pchar(LookupHostAddr(FRemoteHost)));
  Addr.sin_port := htons(FRemotePort); 
  FillChar(Addr.sin_zero, SizeOf(Addr.sin_zero), 0);
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
 try
{$ENDIF} 
  // send packet 
  if FDisconnected then
    Exit
  else
    if sendto(FSocket, Buffer^, Count, Flags, Addr, sizeof(Addr)) = SOCKET_ERROR then
      raise EACRException.Create(40030, ErrorRNetSend, [SocketError]);
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
 except 
  on E: Exception do
    begin
aaWriteToLog('**************************************************************'); 
aaWriteToLog('ACRNetwork> TACRapiNetwork.SendBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************'); 
    end; 
 end; 
{$ENDIF} 
end; // SendBuffer
 
 
// TACRapiNetwork
	
	
	
//////////////////////////////////////////////////////////////////////////////// 
// 
// TACRListenerThread
//
//////////////////////////////////////////////////////////////////////////////// 
 
 
//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRListenerThread.Create(apiNetwork: TACRapiNetwork);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
var 
  NetworkThreads:     TList;
{$ENDIF} 
begin
  inherited Create(False);
  Priority := tpNormal; //tpHigher; //tpHighest;
  FapiNetwork := apiNetwork;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRListenerThread.Create> Thread #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  FNetworkThreads.Add(Self);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> Listener Thread Added');
  NetworkThreads := FNetworkThreads.LockList;
  try
aaWriteToLog('TACRNetwork> Network Threads Count='+IntToStr(NetworkThreads.Count));
  finally
   FNetworkThreads.UnlockList;
  end;
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRListenerThread.Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
var
  NetworkThreads:     TList;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRListenerThread.Destroy> Thread #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
{$IFDEF MSWINDOWS}
//  TerminateThread(Handle, 0);
{$ENDIF}
{$IFDEF LINUX}
{$ENDIF}
  inherited Destroy;
  FNetworkThreads.Remove(Self);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> Listener Thread Removed');
  NetworkThreads := FNetworkThreads.LockList;
  try
aaWriteToLog('TACRNetwork> Network Threads Count='+IntToStr(NetworkThreads.Count));
  finally
   FNetworkThreads.UnlockList;
  end;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRListenerThread.Destroy> Thread #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' TERMINATED');
{$ENDIF}
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRListenerThread.Execute;
var
  Buffer:     PChar;
  BufferSize: Integer;
  Count:      Integer;
  FromHost:   String;
  FromPort:   Integer;
  Flags:      Integer;
  Addr:       TSockAddr;
  AddrLen:    Integer;
  Time:       TTimeVal;
  pTime:      PTimeVal;
  State:      TFDSet;
  pState:     PFDSet;
  ErrorCode:  Integer;
label
  Start;
begin
 try
  // set parameters
  Flags := 0;
  AddrLen := sizeof(Addr);
  FillChar(Addr, SizeOf(Addr), 0);
  Addr.sin_family := AF_INET;
Start:
  BufferSize := FapiNetwork.FPacketSize;
  // receive data
  Buffer := MemoryManager.GetMem(BufferSize);
  try
   // wait for incoming packet
   repeat
    if Terminated then
      Exit;
    State.fd_count := 1;
    State.fd_array[0] := FapiNetwork.FSocket;
    pState := @State;
    Time.tv_sec := 0;
    Time.tv_usec := 1;
    pTime := @Time;
    Count := select(0, pState, nil, nil, pTime);
    if Count = SOCKET_ERROR then
     begin
      ErrorCode := SocketError;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRNetwork> TACRListenerThread.Execute - "select" returned SOCKET_ERROR='+IntToStr(ErrorCode));
{$ENDIF}
      if (Assigned(FapiNetwork.FOnDisconnect))
      and (ErrorCode=10054)
      then
       begin
        TACROnDisconnectThread.Create(FapiNetwork);
        Exit;
       end
      else
        raise EACRException.Create(40031, ErrorRNetReceive, ['select', ErrorCode]);
     end;
   until (Count > 0) and (State.fd_array[0] = FapiNetwork.FSocket);
   // receive packet
{$IFDEF MSWINDOWS}
   Count := recvfrom(FapiNetwork.FSocket, Buffer^, BufferSize, Flags, Addr, AddrLen);
{$ENDIF}
{$IFDEF LINUX}
   Count := Libc.recvfrom(FapiNetwork.FSocket, Buffer^, BufferSize, Flags, @Addr, @AddrLen);
{$ENDIF}
   if Count = SOCKET_ERROR then
    begin
     ErrorCode := SocketError;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRNetwork> TACRListenerThread.Execute - "recvfrom" returned SOCKET_ERROR='+IntToStr(ErrorCode));
{$ENDIF}
      if (Assigned(FapiNetwork.FOnDisconnect))
      and (ErrorCode=10054)
      then
       begin
        TACROnDisconnectThread.Create(FapiNetwork);
        Exit;
       end
      else
        raise EACRException.Create(40031, ErrorRNetReceive, ['recvfrom', ErrorCode]);
    end;
   // get parameters
   FromHost := IntToStr(Addr.sin_addr.S_un_b.s_b1)+'.'+
               IntToStr(Addr.sin_addr.S_un_b.s_b2)+'.'+
               IntToStr(Addr.sin_addr.S_un_b.s_b3)+'.'+
               IntToStr(Addr.sin_addr.S_un_b.s_b4);
   FromPort := Integer (ntohs(Addr.sin_port));
   // send event
   if Assigned(FapiNetwork.FOnDataReceived) then
     FapiNetwork.FOnDataReceived(Buffer, Count, FromHost, FromPort);
  finally
   MemoryManager.FreeAndNilMem(Buffer);
  end;
  goto Start;
 except
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRNetwork> TACRListenerThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
    end;
{$ENDIF}
 end;
end; // Execute

// TACRListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACROnDisconnectThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACROnDisconnectThread.Create(
                        apiNetwork:    TACRapiNetwork
                                          );
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
var
  NetworkThreads:     TList;
{$ENDIF}
begin
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
  FapiNetwork := apiNetwork;
  FNetworkThreads.Add(Self);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> OnDisconnect Thread Added');
  NetworkThreads := FNetworkThreads.LockList;
  try
aaWriteToLog('TACRNetwork> Network Threads Count='+IntToStr(NetworkThreads.Count));
  finally
   FNetworkThreads.UnlockList;
  end;
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACROnDisconnectThread.Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
var
  NetworkThreads:     TList;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('MsgNetwork> TMsgOnDisconnectThread.Destroy - START');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACROnDisconnectThread.Destroy> Thread #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  inherited Destroy;
  FNetworkThreads.Remove(Self);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> OnDisconnect Thread Removed');
  NetworkThreads := FNetworkThreads.LockList;
  try
aaWriteToLog('TACRNetwork> Network Threads Count='+IntToStr(NetworkThreads.Count));
  finally
   FNetworkThreads.UnlockList;
  end;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACROnDisconnectThread.Destroy> Thread #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' TERMINATED');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('MsgNetwork> TMsgOnDisconnectThread.Destroy - FINISH');
{$ENDIF}
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACROnDisconnectThread.Execute;
begin
  FapiNetwork.FDisconnected := True;
  FapiNetwork.FOnDisconnect(FapiNetwork.RemoteHost, FapiNetwork.RemotePort);
end; // Execute

// TACROnDisconnectThread



////////////////////////////////////////////////////////////////////////////////
//
// Procedures
//
////////////////////////////////////////////////////////////////////////////////

function SocketError: Integer;
begin
 {$IFDEF MSWINDOWS}
  Result := WSAGetLastError;
 {$ENDIF}
 {$IFDEF LINUX}
  Result := errno;
 {$ENDIF}

{$IFNDEF MSWINDOWS}
 {$IFNDEF LINUX}
  Result := 0;
 {$ENDIF}
{$ENDIF}
end;


function LookupHostAddr(const hn: string): string;
var
  h: PHostEnt;
begin
  Result := '';
  if hn <> '' then
  begin
    if hn[1] in ['0'..'9'] then
    begin
      if inet_addr(pchar(hn)) <> INADDR_NONE then
        Result := hn;
    end
    else
    begin
      h := gethostbyname(pchar(hn));
      if h <> nil then
        with h^ do
        Result := format('%d.%d.%d.%d', [ord(h_addr^[0]), ord(h_addr^[1]),
      		  ord(h_addr^[2]), ord(h_addr^[3])]);
    end;
  end
  else Result := '0.0.0.0';
end;

{$IFDEF MSWINDOWS}
function socket;            external    winsocket name 'socket';
function bind;              external    winsocket name 'bind';
function sendto;            external    winsocket name 'sendto';
function select;            external    winsocket name 'select';
function recvfrom;          external    winsocket name 'recvfrom';
function inet_addr;         external    winsocket name 'inet_addr';
function gethostbyname;     external    winsocket name 'gethostbyname';
function htons;             external    winsocket name 'htons';
function ntohs;             external    winsocket name 'ntohs';
function closesocket;       external    winsocket name 'closesocket';
function WSAStartup;        external    winsocket name 'WSAStartup';
function WSACleanup;        external    winsocket name 'WSACleanup';
function WSAGetLastError;   external    winsocket name 'WSAGetLastError';
function TerminateThread;   external    kernel32  name 'TerminateThread';
{$ENDIF}

var
{$IFDEF MSWINDOWS}
  WSAData:            TWSAData;
{$ENDIF}
  NetworkThreads:     TList;
  i:                  Integer;
  Err:                Boolean;

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRNetwork> initialized');
{$ENDIF}
  FNetworkThreads := TThreadList.Create;

{$IFDEF MSWINDOWS}
  if WSAStartup($0101, WSAData) = SOCKET_ERROR then
    raise EACRException.Create(40034, ErrorRWSAStartup, [SocketError]);
{$ENDIF}

finalization

// Terminate all network threads...
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> Terminate all network threads...');
{$ENDIF}
  NetworkThreads := FNetworkThreads.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> Hang Count='+IntToStr(NetworkThreads.Count));
{$ENDIF}
  try
   for i:= NetworkThreads.Count-1 downto 0 do
    begin
     Err := TerminateThread(TThread(NetworkThreads[i]).Handle, 0);
     if not Err then
      begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> TerminateThread failed. '+ErrorRCannotKillThread+IntToStr(Err));
{$ENDIF}
      end;
     if TThread(NetworkThreads[i]).Handle <> 0 then
       Err := CloseHandle(TThread(NetworkThreads[i]).Handle);
     if not Err then
      begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> TerminateThread failed. '+ErrorRCannotKillThread+IntToStr(Err));
{$ENDIF}
      end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> Count='+IntToStr(NetworkThreads.Count));
{$ENDIF}
    end;
  finally
   FNetworkThreads.UnlockList;
  end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  sleep(1000);
  NetworkThreads := FNetworkThreads.LockList;
  try
aaWriteToLog('TACRNetwork> Rest Count='+IntToStr(NetworkThreads.Count));
  finally
   FNetworkThreads.UnlockList;
  end;
{$ENDIF}

{$IFDEF MSWINDOWS}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> WSACleanup...');
{$ENDIF}
  if WSACleanup = SOCKET_ERROR then
    raise EACRException.Create(40035, ErrorRWSACleanup, [SocketError]);
{$ENDIF}

  FNetworkThreads.Free;

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRNetwork> FINISH');
{$ENDIF}

end.

