unit ACRTempEngine;

{$I ACRVer.inc}

interface

uses SysUtils, Classes,

// Accuracer units

     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRCriticalSection,
     ACRExcept,
     ACRBase,
     ACRPage,
     ACRBaseEngine,
     ACRMemory,
     ACRCompression,
     ACRTypes,
     ACRConverts,
     ACRConst;

type
   TACRTemporaryTableData = class;


////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryRecordManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRTemporaryRecordManager = class (TACRBaseRecordManager)
   private
    FRecordsPerPage:        Integer;
    FPageSize:              Integer;
    FCachedRecordCount:     Integer;
    FCache:                 PChar;
    FAllocatedRecordCount:  Integer;
    FTempPage:              PChar;
    FTempPageRecordCount:   Integer;
    FTempPageFile:          TACRFileStream;
    FMaxCachedRecordCount:  Integer;
    FAllocRecordsBy:        Integer;

    procedure SaveTempPage;
    function ReadRecord(
                        var RecordBuffer: TACRRecordBuffer;
                        RecordID: TACRRecordNo
                       ): Boolean;
    // return result for attempt of getting record relatively to first position
    // and set RecordID to new record ID
    function GetRecordFromFirstPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID:  TACRRecordNo
                                       ): TACRGetRecordResult;
    // return result for attempt of getting record relatively to last position
    // and set RecordID to new record ID
    function GetRecordFromLastPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID:  TACRRecordNo
                                      ): TACRGetRecordResult;
    // return result for attempt of getting record relatively any position
    // and set RecordID to new record ID
    function GetRecordFromAnyPosition(
            GetRecordMode: TACRGetRecordMode;
            var RecordID:  TACRRecordNo
                                      ): TACRGetRecordResult;
   public
    constructor Create(
                        RecordBufferSize:     Integer;
                        RecordsPerPage:       Integer;
                        AllocRecordsBy:       Integer
                      );
    destructor Destroy; override;

    procedure Empty(SessionID: TACRSessionID = INVALID_SESSION_ID); override;

    // return true if record exists
    function IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID = INVALID_SESSION_ID): Boolean; override;
    // add record and return its number
    function AddRecord(
                       RecordBuffer:   TACRRecordBuffer;
                        var RecordID:  TACRRecordID;
                        SessionID:     TACRSessionID = INVALID_SESSION_ID
                       ): Boolean; override;
    procedure GetRecordBuffer(var NavigationInfo: TACRNavigationInfo); override;
    // return record no
    function GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo; override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
  end; // TACRTemporaryRecordManager


////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryDatabaseData
//
////////////////////////////////////////////////////////////////////////////////


  TACRTemporaryDatabaseData = class (TACRDatabaseData)
   public
    constructor Create;
    destructor Destroy; override;
    // create table data
    function CreateTableData(Cursor: TACRCursor): TACRTableData; override;
  end; // TACRTemporaryDatabaseData


////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryTableData
//
////////////////////////////////////////////////////////////////////////////////


  TACRTemporaryTableData = class (TACRTableData)
   private
    FBLOBFile:          TACRFileStream;
    FRecordsPerPage:    Integer;
    FAllocRecordsBy:    Integer;
    FPageManager:       TACRPageManager;

   protected
    function GetPageManager: TACRPageManager; override;
    procedure CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo = INVALID_PAGE_NO); override;
    procedure CreateBLOBFile;
    procedure DeleteBLOBFile;
    // return filter bitmap rec no by record id
    function GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo; override;
    // return filter bitmap rec no by record id
    function GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID; override;
   public
    constructor Create(
                        aDatabaseData:  TACRDatabaseData;
                        RecordsPerPage: Integer = 10;
                        AllocRecordsBy: Integer = 1000
                      );
    destructor Destroy; override;

    procedure CreateTable(
                          Cursor: TACRCursor;
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); override;
    procedure DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean = False); override;
    procedure EmptyTable(Cursor: TACRCursor); override;
    procedure OpenTable(Cursor: TACRCursor); override;
    procedure CloseTable(Cursor: TACRCursor); override;
    procedure AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor); override;

    // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
    function CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer; override;

    function InsertRecord(var Cursor: TACRCursor): Boolean; override;
    function DeleteRecord(Cursor: TACRCursor): Boolean; override;
    function UpdateRecord(Cursor: TACRCursor): Boolean; override;
    procedure DeleteVisibleRecords(Cursor: TACRCursor); override;

    function InternalGetRecordCount: TACRRecordNo; override;
    // move cursor to specified position and set current record id in cursor
    procedure InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo); override;
    // get current record position from cursor
    function InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo; override;

    //---------------------------------------------------------------------------
    // BLOB methods
    //---------------------------------------------------------------------------
    function InternalCreateBlobStream(
              Cursor:   TACRCursor;
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ): TACRStream; override;

    procedure WriteBLOBFieldToRecordBuffer(
              Cursor:     TACRCursor;
              FieldNo:    Integer;
              BLOBStream: TACRStream
              ); override;

    procedure ClearBLOBFieldInRecordBuffer(
              var RecordBuffer: TACRRecordBuffer;
              FieldNo:    Integer
              ); override;

    procedure LoadTableFromStream(
                        Stream:               TStream
                       ); override;
    procedure SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      ); override;
  end; // TACRTemporaryTableData


implementation


uses

// Accuracer units

  ACRLocalEngine,
  ACRMemEngine;


////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryRecordManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// save temporary page
//------------------------------------------------------------------------------
procedure TACRTemporaryRecordManager.SaveTempPage;
var WriteBytes,WriteSize:   Integer;
    OldPos,NewPos:          Int64;
begin
 if (FTempPageRecordCount > 0) then
  begin
   if (FTempPageRecordCount > FRecordsPerPage) then
    raise EACRException.Create(10126,ErrorLInvalidRecordCount,
      [FTempPageRecordCount,FRecordsPerPage]);
   OldPos := FTempPageFile.Position;
   NewPos := Int64(FRecordCount - FMaxCachedRecordCount - FTempPageRecordCount) *
             Int64(FRecordBufferSize);
   FTempPageFile.Position := NewPos;
   if (FTempPageFile.Position <> NewPos) then
    raise EACRException.Create(10128,ErrorLCannotSetPosition,
      [NewPos,OldPos,FTempPageFile.Size]);
   OldPos := FTempPageFile.Position;
   WriteSize := FTempPageRecordCount * FRecordBufferSize;
   WriteBytes := FTempPageFile.Write(FTempPage^,WriteSize);
   if (WriteBytes <> WriteSize) then
    raise EACRException.Create(10125,ErrorLWriteToStream,
      [OldPos,FTempPageFile.Size,WriteSize,WriteBytes]);
   FTempPageRecordCount := 0;
  end;
end; // SaveTempPage


//------------------------------------------------------------------------------
// read record
//------------------------------------------------------------------------------
function TACRTemporaryRecordManager.ReadRecord(
                        var RecordBuffer: TACRRecordBuffer;
                        RecordID: TACRRecordNo
                       ): Boolean;
var ReadBytes:          Integer;
    OldPos,NewPos:      Int64;
begin
 if (RecordID < FMaxCachedRecordCount) then
  begin
   try
     Move((FCache + RecordID * FRecordBufferSize)^,RecordBuffer^,
      FRecordBufferSize);
   except
     raise EACRException.Create(10127,ErrorLRetreivingRecord);
   end;
   Result := True;
  end
 else
  begin
   SaveTempPage;
   OldPos := FTempPageFile.Position;
   NewPos := Int64(RecordID - FMaxCachedRecordCount) *
             Int64(FRecordBufferSize);
   FTempPageFile.Position := NewPos;
   if (FTempPageFile.Position <> NewPos) then
    raise EACRException.Create(10129,ErrorLCannotSetPosition,
      [NewPos,OldPos,FTempPageFile.Size]);
   OldPos := FTempPageFile.Position;
   ReadBytes := FTempPageFile.Read(RecordBuffer^,FRecordBufferSize);
   if (ReadBytes <> FRecordBufferSize) then
    raise EACRException.Create(10130,ErrorLCannotReadFromStream,
      [OldPos,FTempPageFile.Size,FRecordBufferSize,ReadBytes]);
   Result := True;
  end;
end; // ReadRecord


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively to first position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRTemporaryRecordManager.GetRecordFromFirstPosition(
        GetRecordMode: TACRGetRecordMode;
        var RecordID:  TACRRecordNo
                                   ): TACRGetRecordResult;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    Result := grrBOF;
   end;
  grmCurrent:
   begin
    Result := grrError;
   end;
  grmNext:
   begin
    RecordID := 0;
    Result := grrOK;
   end;
 end; // GetRecordMode
end; // GetRecordFromFirstPosition


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively to last position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRTemporaryRecordManager.GetRecordFromLastPosition(
        GetRecordMode: TACRGetRecordMode;
        var RecordID:  TACRRecordNo
                                  ): TACRGetRecordResult;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    RecordID := FRecordCount-1;
    Result := grrOK;
   end;
  grmCurrent:
   begin
    Result := grrError;
   end;
  grmNext:
   begin
    Result := grrEOF;
   end;
 end; // GetRecordMode
end; // GetRecordFromLastPosition


//------------------------------------------------------------------------------
// return result for attempt of getting record relatively any position
// and set RecordID to new record ID
//------------------------------------------------------------------------------
function TACRTemporaryRecordManager.GetRecordFromAnyPosition(
        GetRecordMode: TACRGetRecordMode;
        var RecordID:  TACRRecordNo
                                  ): TACRGetRecordResult;
begin
 Result := grrError;
 case GetRecordMode of
  grmPrior:
   begin
    Result := grrBOF;
    if (RecordID > 0) then
     begin
      Dec(RecordID);
      Result := grrOK;
     end; // RecordID > 0
   end;
  grmCurrent:
   begin
    if ((RecordID >= FRecordCount) or (RecordID < 0)) then
     // current record does not exist
     Result := grrError
    else
     Result := grrOK;
   end;
  grmNext:
   begin
    Result := grrEOF;
    if (RecordID < FRecordCount-1) then
     begin
      Inc(RecordID);
      Result := grrOK;
     end;
   end;
 end; // GetRecordMode
end; // GetRecordFromAnyPosition


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRTemporaryRecordManager.Create(
                        RecordBufferSize:     Integer;
                        RecordsPerPage:       Integer;
                        AllocRecordsBy:       Integer
                      );
begin
 if (RecordBufferSize = 0) then
  raise EACRException.Create(10124,ErrorLInvalidRecordSize);
 InitializeCriticalSection(FCSect);
 FRecordCount := 0;
 FRecordBufferSize := RecordBufferSize;
 FRecordsPerPage := RecordsPerPage;
 if (FRecordsPerPage = 0) then
  FRecordsPerPage := 1;
 FMaxCachedRecordCount := ((MemoryManager.GetFreeMemorySize div 10) div
                          FRecordBufferSize);
// FMaxCachedRecordCount := 1;
 FPageSize := FRecordsPerPage * FRecordBufferSize;
 FTempPage := MemoryManager.GetMem(FPageSize);
 FTempPageRecordCount := 0;
 FCachedRecordCount := 0;
 FCache := nil;
 FAllocRecordsBy := AllocRecordsBy;
 FTempPageFile := nil;
end; // Create;


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRTemporaryRecordManager.Destroy;
begin
 Empty;
 MemoryManager.FreeAndNilMem(FTempPage);
 DeleteCriticalSection(FCSect);
 inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// Empty
//------------------------------------------------------------------------------
procedure TACRTemporaryRecordManager.Empty(SessionID: TACRSessionID);
var
    FileName: String;
begin
 FRecordCount := 0;
 FTempPageRecordCount := 0;
 if (FCache <> nil) then
  MemoryManager.FreeAndNilMem(FCache);
 FCachedRecordCount := 0;
 FAllocatedRecordCount := 0;
 if (FTempPageFile <> nil) then
  begin
   FileName := FTempPageFile.FileName;
   FTempPageFile.Free;
   SysUtils.DeleteFile(FileName);
   FTempPageFile := nil;
  end;
end; // Empty


//------------------------------------------------------------------------------
// return true if record exists
//------------------------------------------------------------------------------
function TACRTemporaryRecordManager.IsRecordExists(var RecordID: TACRRecordID; SessionID: TACRSessionID): Boolean;
var RecordNo: TACRRecordNo;
begin
 RecordNo := 0;
 Move(RecordID,RecordNo,sizeof(RecordID));
 Result := ((FRecordCount > 0) and (RecordNo < FRecordCount) and (RecordNo > 0));
end; // IsRecordExists


//------------------------------------------------------------------------------
// add record and return its number
//------------------------------------------------------------------------------
function TACRTemporaryRecordManager.AddRecord(
                                              RecordBuffer: TACRRecordBuffer;
                                              var RecordID:  TACRRecordID;
                                              SessionID:     TACRSessionID
                                              ): Boolean;
var SaveToTemp: Boolean;
begin
 try
   SaveToTemp := False;
   if (FCachedRecordCount >= FAllocatedRecordCount) then
    begin
     // need extending cache or writing to temporary file
     if (FAllocatedRecordCount < FMaxCachedRecordCount) then
      begin
       // create or extend cache
       if (FAllocatedRecordCount + FAllocRecordsBy <= FMaxCachedRecordCount) then
        Inc(FAllocatedRecordCount,FAllocRecordsBy)
       else
        FAllocatedRecordCount := FMaxCachedRecordCount;
       if (FCache = nil) then
        FCache := MemoryManager.GetMem(FAllocatedRecordCount * FRecordBufferSize)
       else
        MemoryManager.ReallocMem(FCache,(FAllocatedRecordCount * FRecordBufferSize));
       Inc(FCachedRecordCount);
      end
     else
      begin
       SaveToTemp := True;
       // need writing to temporary file
       if (FTempPageFile = nil) then
        FTempPageFile := TACRFileStream.Create(GetTempFileName,fmCreate);
       SaveTempPage;
       Move(RecordBuffer^,(FTempPage + FTempPageRecordCount * FRecordBufferSize)^,
         FRecordBufferSize);
       Inc(FTempPageRecordCount);
      end;
    end
   else
    begin
     // no need in extending
     Inc(FCachedRecordCount);
    end;
   if (not SaveToTemp) then
    Move(RecordBuffer^,(FCache + FRecordCount * FRecordBufferSize)^,
         FRecordBufferSize);
   Move(FRecordCount,RecordID,sizeof(RecordID));
   Inc(FRecordCount);
   Result := True;
 except
   Result := False;
 end;
end; // AddRecord


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
procedure TACRTemporaryRecordManager.GetRecordBuffer(var NavigationInfo: TACRNavigationInfo);
var RecordNo: TACRRecordNo;
begin
 RecordNo := 0;
 Move(NavigationInfo.RecordID,RecordNo,sizeof(NavigationInfo.RecordID));
 if (FRecordCount = 0) then
  begin
   NavigationInfo.GetRecordResult := grrEOF;
   Exit;
  end;
 // get record relatively to the first position
 if (NavigationInfo.FirstPosition) then
  begin
   NavigationInfo.GetRecordResult := GetRecordFromFirstPosition(
                                        NavigationInfo.GetRecordMode,
                                        RecordNo
                                                               );
   if (NavigationInfo.GetRecordResult = grrOK) then
    NavigationInfo.FirstPosition := False;
  end
 else
 // get record relatively to the last position
 if (NavigationInfo.LastPosition) then
  begin
   NavigationInfo.GetRecordResult := GetRecordFromLastPosition(
                                        NavigationInfo.GetRecordMode,
                                        RecordNo
                                                               );
   if (NavigationInfo.GetRecordResult = grrOK) then
    NavigationInfo.LastPosition := False;
  end
 else
  NavigationInfo.GetRecordResult := GetRecordFromAnyPosition(
                                        NavigationInfo.GetRecordMode,
                                        RecordNo
                                                               );
 if (NavigationInfo.GetRecordResult = grrOK) then
  begin
   NavigationInfo.GetRecordResult := grrError;
   if (ReadRecord(NavigationInfo.RecordBuffer,RecordNo)) then
    NavigationInfo.GetRecordResult := grrOK; 
  end; // record retrieved successfully 
 Move(RecordNo,NavigationInfo.RecordID,sizeof(NavigationInfo.RecordID));
end; // GetRecordBuffer
 
 
//------------------------------------------------------------------------------
// return record no
//------------------------------------------------------------------------------ 
function TACRTemporaryRecordManager.GetApproximateRecNo(RecordID: TACRRecordID; SessionID: TACRSessionID): TACRRecordNo;
begin 
 Result := 0;
 Move(RecordID,Result,Sizeof(RecordID)); 
 if (Result >= FRecordCount) then
  Result := -1 
 else
  begin 
   Inc(Result); 
  end;
end; // GetApproximateRecNo 
 
 
//------------------------------------------------------------------------------ 
// LoadFromStream 
//------------------------------------------------------------------------------
procedure TACRTemporaryRecordManager.LoadFromStream(Stream: TStream);
var Size: Int64;
begin
  LoadDataFromStream(FRecordBufferSize,SizeOf(FRecordBufferSize),Stream,11217);
  LoadDataFromStream(FRecordCount,SizeOf(FRecordCount),Stream,11218); 
  LoadDataFromStream(FRecordsPerPage,SizeOf(FRecordsPerPage),Stream,11219);
  LoadDataFromStream(FPageSize,SizeOf(FPageSize),Stream,11220); 
  LoadDataFromStream(FCachedRecordCount,SizeOf(FCachedRecordCount),Stream,11221);
  LoadDataFromStream(FAllocatedRecordCount,SizeOf(FAllocatedRecordCount),Stream,11222); 
  LoadDataFromStream(FMaxCachedRecordCount,SizeOf(FMaxCachedRecordCount),Stream,11223); 
  LoadDataFromStream(FAllocRecordsBy,SizeOf(FAllocRecordsBy),Stream,11224);
  if (FAllocatedRecordCount = 0) then 
   FCache := nil
  else 
   begin 
    if (FCache = nil) then
      FCache := MemoryManager.GetMem(FAllocatedRecordCount * FRecordBufferSize)
    else 
      MemoryManager.ReallocMem(FCache,(FAllocatedRecordCount * FRecordBufferSize)); 
   end; // recordcount > 0 
  if (FTempPage = nil) then 
    FTempPage := MemoryManager.GetMem(FPageSize) 
  else 
    MemoryManager.ReallocMem(FTempPage,(FPageSize)); 
  // load cache
  if (FAllocatedRecordCount > 0) then
    LoadDataFromStream(FCache^,FCachedRecordCount * FRecordBufferSize,Stream,11225); 
  // load temp file
  LoadDataFromStream(Size,SizeOf(Size),Stream,11227);
  if (Size > 0) then 
   begin
    if (FTempPageFile = nil) then
      FTempPageFile := TACRFileStream.Create(GetTempFileName,fmCreate) 
    else 
      FTempPageFile.Size := 0; 
    FTempPageFile.LoadFromStreamWithPosition(Stream,Stream.Position,Size);
   end
  else 
   FTempPageFile := nil; 	
end; // LoadFromStream
 
 
//------------------------------------------------------------------------------
// SaveToStream
//------------------------------------------------------------------------------
procedure TACRTemporaryRecordManager.SaveToStream(Stream: TStream); 
var Size: Int64; 
begin 
  SaveDataToStream(FRecordBufferSize,SizeOf(FRecordBufferSize),Stream,11215); 
  SaveDataToStream(FRecordCount,SizeOf(FRecordCount),Stream,11216);
  SaveDataToStream(FRecordsPerPage,SizeOf(FRecordsPerPage),Stream,11208); 
  SaveDataToStream(FPageSize,SizeOf(FPageSize),Stream,11209); 
  SaveDataToStream(FCachedRecordCount,SizeOf(FCachedRecordCount),Stream,11210); 
  SaveDataToStream(FAllocatedRecordCount,SizeOf(FAllocatedRecordCount),Stream,11211);
  SaveDataToStream(FMaxCachedRecordCount,SizeOf(FMaxCachedRecordCount),Stream,11212);
  SaveDataToStream(FAllocRecordsBy,SizeOf(FAllocRecordsBy),Stream,11213);
  // save cache 
  if (FAllocatedRecordCount > 0) then
    SaveDataToStream(FCache^,FCachedRecordCount * FRecordBufferSize,Stream,11214);
  // save temp file
  Size := 0; 
  if (FRecordCount > FMaxCachedRecordCount) then 
   Size := FTempPageFile.Size;
  SaveDataToStream(Size,SizeOf(Size),Stream,11226);
  if (Size > 0) then
   FTempPageFile.SaveToStream(Stream);
end; // SaveToStream
	
 
////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryDatabaseData 
//
////////////////////////////////////////////////////////////////////////////////
 
 
//------------------------------------------------------------------------------ 
// create 
//------------------------------------------------------------------------------
constructor TACRTemporaryDatabaseData.Create; 
begin 
  inherited Create;
  DatabaseName := ACRTemporaryDatabaseName;
end;// Create
	
 
//------------------------------------------------------------------------------ 
// destroy
//------------------------------------------------------------------------------
destructor TACRTemporaryDatabaseData.Destroy; 
begin 
{ 
 while (FTableDataList.Count > 0) do
  begin
    if (FTableDataList.Items[0] = nil) then
     raise EACRException.Create(10379, ErrorLNilPointer);
    if (TACRTableData(FTableDataList.Items[0]).CursorList.Count > 0) then
     TACRTableData(FTableDataList.Items[0]).DeleteTable; 
  end;
} 
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// create table data
//------------------------------------------------------------------------------
function TACRTemporaryDatabaseData.CreateTableData(Cursor: TACRCursor): TACRTableData;
begin
  Result := TACRTemporaryTableData.Create(Self);
  Result.TableName := Cursor.TableName;
  FTableDataList.Add(Result);
end;// CreateTableData


////////////////////////////////////////////////////////////////////////////////
//
// TACRTemporaryTableData
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// GetPageManager
//------------------------------------------------------------------------------
function TACRTemporaryTableData.GetPageManager: TACRPageManager;
begin
  Result := FPageManager;
end;// GetPageManager


//------------------------------------------------------------------------------
// create record manager
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.CreateRecordManager(TablePFSMapFileStartPageNo: TACRPageNo);
begin
 if (FRecordManager <> nil) then
  FRecordManager.Free;
 FRecordManager := TACRTemporaryRecordManager.Create(GetRecordBufferSize,
  FRecordsPerPage,FAllocRecordsBy);
end; // CreateRecordManager


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.CreateBLOBFile;
begin
 if (FBLOBFieldsPresent) then
  begin
   DeleteBLOBFile;
   FBLOBFile := TACRFileStream.Create(GetTempFileName,fmCreate);
  end;
end;


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.DeleteBLOBFile;
var FileName: String;
begin
 if (FBLOBFile <> nil) then
  begin
   FileName := FBLOBFile.FileName;
   FBLOBFile.Free;
   FBLOBFile := nil;
   SysUtils.DeleteFile(FileName);
  end;
end;


//------------------------------------------------------------------------------
// return filter bitmap rec no by record id
//------------------------------------------------------------------------------
function TACRTemporaryTableData.GetBitmapRecNoByRecordID(RecordID: TACRRecordID): TACRRecordNo;
begin
  Result := 0;
  Move(RecordID,Result,sizeof(RecordID));
end; // GetBitmapRecNoByRecordID


//------------------------------------------------------------------------------
// return filter bitmap rec no by record id
//------------------------------------------------------------------------------
function TACRTemporaryTableData.GetRecordIDByBitmapRecNo(RecordNo: TACRRecordNo): TACRRecordID;
begin
  Move(RecordNo,Result,sizeof(Result));
end; // GetRecordIDByBitmapRecNo


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRTemporaryTableData.Create(
                    aDatabaseData:  TACRDatabaseData;
                    RecordsPerPage: Integer;
                    AllocRecordsBy: Integer
                  );
begin
 inherited Create(aDatabaseData);
 FPageManager := TACRTemporaryPageManager.Create;
 FRecordsPerPage := RecordsPerPage;
 FAllocRecordsBy := AllocRecordsBy;
 FBLOBFile := nil;
 FInMemory := False;
 FTemporary := True;
end; // Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRTemporaryTableData.Destroy;
begin
  EmptyTable(nil);
  DeleteAllIndexes(nil);
  inherited Destroy;
  FPageManager.Free;
end; // Destroy


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.CreateTable(
                          Cursor: TACRCursor;
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                                         );
begin
 if (FieldDefs.Count <= 0) then
  raise EACRException.Create(10132,ErrorLNoFields);
 TableName := Cursor.TableName;
 CheckFieldDefinitions(FieldDefs);
 CheckIndexDefinitions(IndexDefs);
 CheckConstraintDefinitions(ConstraintDefs);
 CreateFieldManager(FieldDefs);
 CreateIndexManager(IndexDefs);
 CreateConstraintManager(ConstraintDefs);
 if (FConstraintManager.ConstraintDefs.ForeignKeysExists) then
  raise EACRException.Create(11426,ErrorLForeignKeysAreNotSupportedInTempTable);
 CreateRecordManager;
 InitCursor(Cursor);
 CreateBLOBFile;
end;// CreateTable


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.DeleteTable(Session: TACRBaseSession; Cascade: Boolean; DesignMode: Boolean);
begin
 inherited DeleteTable(Session,Cascade,DesignMode);
 if (FIndexManager.IndexDefs.Count > 0) then
  begin
   if (Session <> nil) then
    FSessionID := Session.SessionID
   else
    FSessionID := SYSTEM_SESSION_ID;
   FIndexManager.DropAllIndexes(FSessionID);
   FIndexManager.PageManager.ApplyChanges(FSessionID);
  end;
 if ((not DesignMode) or (FCursorList.Count = 0)) then
   Free;
end; // DeleteTable


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.EmptyTable(Cursor: TACRCursor);
begin
 DeleteBLOBFile;
 inherited EmptyTable(Cursor);
 if (FRecordManager <> nil) then
  begin
   FRecordManager.Empty;
  end;
 FPageManager.ApplyChanges(SYSTEM_SESSION_ID);
 if (Cursor <> nil) then
  FPageManager.ApplyChanges(Cursor.Session.SessionID);
end; // EmptyTable


//------------------------------------------------------------------------------
// open table
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.OpenTable(Cursor: TACRCursor);
begin
 TableName := Cursor.TableName;
 inherited OpenTable(Cursor);
end;// OpenTable


//------------------------------------------------------------------------------
// close table
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.CloseTable(Cursor: TACRCursor);
begin
 inherited CloseTable(Cursor);
 if ((FCursorList.Count <= 0) and (not Cursor.CreateTableStarted)) then
  begin
   DeleteTable(Cursor.Session,True);
  end;
end;// CloseTable


//------------------------------------------------------------------------------
// add index
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.AddIndex(IndexDef: TACRIndexDef; Cursor: TACRCursor);
begin
 Lock;
 try
   inherited AddIndex(IndexDef,Cursor);
   FPageManager.ApplyChanges(Cursor.Session.SessionID);
 finally
  Unlock;
 end;
end; // AddIndex


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRTemporaryTableData.CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
var id1,id2: TACRRecordNo;
begin
 id1 := 0;
 id2 := 0;
 Move(RecordID1,id1,sizeof(RecordID1));
 Move(RecordID2,id2,sizeof(RecordID2));
 if (id1 = id2) then
  Result := 0
 else
 if (id1 > id2) then
  Result := 1
 else
  Result := -1;
end; // CompareRecordID


//------------------------------------------------------------------------------
// insert record
//------------------------------------------------------------------------------
function TACRTemporaryTableData.InsertRecord(var Cursor: TACRCursor): Boolean;
var
    RecordID:       TACRRecordID;
    Pos:            Pointer;
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10134,ErrorLNilPointer);
  if (Cursor.CurrentRecordBuffer = nil) then
   raise EACRException.Create(10135,ErrorLNilPointer);
  // add record to first empty space
  Pos := Cursor.SavePosition;
  try
    try
      Result := FRecordManager.AddRecord(Cursor.CurrentRecordBuffer,RecordID);
      Move(RecordID,Cursor.CurrentRecordID,sizeof(Cursor.CurrentRecordID));
    except
      Result := False;
    end;
    if (Result) then
     begin
       FIndexManager.InsertRecord(Cursor);
       if (FIndexManager.IndexDefs.Count > 0) then
         FPageManager.ApplyChanges(Cursor.Session.SessionID);
       Cursor.FirstPosition := False;
       Cursor.LastPosition := False;
       if (TACRRecordBitmap(Cursor.RecordBitmap).Active) then
        UpdateRecordBitmapAfterInsertRecord(Cursor,Pos);
     end;
  finally
    Cursor.FreePosition(Pos);
  end;
end; // InsertRecord;


//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
function TACRTemporaryTableData.DeleteRecord(Cursor: TACRCursor): Boolean;
begin
 Result := False;
end; // DeleteRecord


//------------------------------------------------------------------------------
// update record
//------------------------------------------------------------------------------
function TACRTemporaryTableData.UpdateRecord(Cursor: TACRCursor): Boolean;
begin
 Result := False;
end; // UpdateRecord


//------------------------------------------------------------------------------
// delete all visible records;
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.DeleteVisibleRecords(Cursor: TACRCursor);
begin
  raise EACRException.Create(11301,ErrorLOperationIsNotSupported);
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRTemporaryTableData.InternalGetRecordCount: TACRRecordNo;
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10133,ErrorLNilPointer);
  Result := FRecordManager.GetRecordCount;
end; // GetRecordCount


//------------------------------------------------------------------------------
// move cursor to specified position and set current record id in cursor
//------------------------------------------------------------------------------
procedure TACRTemporaryTableData.InternalSetRecNo(Cursor: TACRCursor; RecNo: TACRRecordNo);
var RecordID: TACRRecordNo;
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10136,ErrorLNilPointer);
  if (RecNo > 1) then
   RecordID := RecNo-1
  else
   RecordID := 0;
  Move(RecordID,Cursor.CurrentRecordID,sizeof(Cursor.CurrentRecordID));
end; // SetRecNo


//------------------------------------------------------------------------------
// get current record position from cursor
//------------------------------------------------------------------------------
function TACRTemporaryTableData.InternalGetRecNo(Cursor: TACRCursor): TACRRecordNo;
begin
  if (FRecordManager = nil) then
   raise EACRException.Create(10137,ErrorLNilPointer);
  Result := 0;
  Move(Cursor.CurrentRecordID,Result,sizeof(Cursor.CurrentRecordID));
  Inc(Result);
end; // GetRecNo


//---------------------------------------------------------------------------
// create BLOB stream
//---------------------------------------------------------------------------
function TACRTemporaryTableData.InternalCreateBlobStream(
          Cursor:   TACRCursor;
          ToInsert: Boolean;
          FieldNo:  Integer;
          OpenMode: TACRBLOBOpenMode
          ): TACRStream;
var TempStream:             TACRTemporaryStream;
    CompressedStream:       TACRCompressedBLOBStream;
    Offset:                 Integer;
    DiskOffset,OldPos:      Int64;
    ReadBytes,ReadSize:     Integer;
    BLOBDescriptor:         TACRBLOBDescriptor;
    PartialBLOBDescriptor:  TACRPartialTemporaryBLOBDescriptor;
begin
 if (FBLOBFile = nil) then
  raise EACRException.Create(10140,ErrorLNilPointer);
 BLOBDescriptor.CompressionAlgorithm :=
     Byte(FieldManager.FieldDefs[FieldNo].BLOBCompressionAlgorithm);
 BLOBDescriptor.CompressionMode := FieldManager.FieldDefs[FieldNo].BLOBCompressionMode;
 BLOBDescriptor.BlockSize := FieldManager.FieldDefs[FieldNo].BLOBBlockSize;
 BLOBDescriptor.StartPosition := 0;
 TempStream := TACRTemporaryStream.Create;
 // create new compressed stream
 if ((ToInsert) or (OpenMode = bomWrite) or
     (CheckNullFlag(FieldNo,Cursor.CurrentRecordBuffer))) then
  begin
   // empty stream
   BLOBDescriptor.NumBlocks := 0;
   BLOBDescriptor.UncompressedSize := 0;
   CompressedStream := TACRCompressedBLOBStream.Create(TempStream,
    BLOBDescriptor,True);
   Result := TACRLocalBLOBStream.Create(CompressedStream,Cursor,OpenMode,FieldNo);
  end // empty stream
 else
  begin
   // copy value from TableData
   Offset := FieldManager.FieldDefs[FieldNo].MemoryOffset;
   // offset to blob field data
   Move(PChar(Cursor.CurrentRecordBuffer + Offset)^,DiskOffset,Sizeof(DiskOffset));
   OldPos := FBLOBFile.Position;

   // set position in BLOB file
   FBLOBFile.Position := DiskOffset;
   if (FBLOBFile.Position <> DiskOffset) then
     raise EACRException.Create(10141,ErrorLCannotSetPosition,
       [DiskOffset,OldPos,FBLOBFile.Size]);

   // read partial BLOB descriptor
   OldPos := FBLOBFile.Position;
   ReadSize := sizeof(TACRPartialTemporaryBLOBDescriptor);
   ReadBytes := FBLOBFile.Read(PartialBLOBDescriptor,ReadSize);
   if (ReadBytes <> ReadSize) then
    raise EACRException.Create(10142,ErrorLCannotReadFromStream,
      [OldPos,FBLOBFile.Size,ReadSize,ReadBytes]);

   // read compressed blob data
   TempStream.LoadFromStreamWithPosition(FBLOBFile,FBLOBFile.Position,
      PartialBLOBDescriptor.CompressedSize);

   // copy partial blob descriptor
   BLOBDescriptor.NumBlocks := PartialBLOBDescriptor.NumBlocks;
   BLOBDescriptor.UncompressedSize := PartialBLOBDescriptor.UncompressedSize;
   CompressedStream := TACRCompressedBLOBStream.Create(TempStream,
    BLOBDescriptor,False);
   Result := TACRLocalBLOBStream.Create(CompressedStream,Cursor,OpenMode,FieldNo);
  end; // copy value from TableData
end; // InternalCreateBlobStream


//---------------------------------------------------------------------------
// write BLOB field to record buffer
//---------------------------------------------------------------------------
procedure TACRTemporaryTableData.WriteBLOBFieldToRecordBuffer(
          Cursor:     TACRCursor;
          FieldNo:    Integer;
          BLOBStream: TACRStream
          );
var
    WriteSize,WriteBytes:     Integer;
    PartialBLOBDescriptor:    TACRPartialTemporaryBLOBDescriptor;
    DiskOffset:               Int64;
begin
 if (FBLOBFile = nil) then
  raise EACRException.Create(10144,ErrorLNilPointer);
 if (BLOBStream.Modified) then
  begin
   if (BLOBStream.Size = 0) then
    begin
     // empty stream
     SetNullFlag(True,FieldNo,Cursor.CurrentRecordBuffer);
    end
   else
    begin
     PartialBLOBDescriptor.NumBlocks := TACRCompressedBLOBStream(
        TACRLocalBLOBStream(BLOBStream).TemporaryStream).BLOBDescriptor.NumBlocks;
     PartialBLOBDescriptor.UncompressedSize := TACRCompressedBLOBStream(
        TACRLocalBLOBStream(BLOBStream).TemporaryStream).BLOBDescriptor.UncompressedSize;
     PartialBLOBDescriptor.CompressedSize := TACRCompressedBLOBStream(
        TACRLocalBLOBStream(BLOBStream).TemporaryStream).CompressedSize;
     SetStreamPosition(FBLOBFile,FBLOBFile.Size,10422);
     DiskOffset := FBLOBFile.Position;
     // save blob descriptor
     WriteSize := sizeof(PartialBLOBDescriptor);
     WriteBytes := FBLOBFile.Write(PartialBLOBDescriptor,WriteSize);
     if (WriteBytes <> WriteSize) then
      raise EACRException.Create(10145,ErrorLCannotWriteToStream,
        [DiskOffset,FBLOBFile.Size,WriteSize,WriteBytes]);
     // save compressed data to blob file
     TACRStream(TACRCompressedBLOBStream(TACRLocalBLOBStream(BLOBStream).TemporaryStream).
        CompressedStream).SaveToStream(FBLOBFile);
     // store offset of blob descriptor in blob file to record buffer
     Move(DiskOffset,PChar(Cursor.CurrentRecordBuffer +
      FieldManager.FieldDefs[FieldNo].MemoryOffset)^,sizeof(DiskOffset));
     SetNullFlag(False,FieldNo,Cursor.CurrentRecordBuffer);
    end; // not empty stream
  end; // Modified
end; // WriteBLOBFieldToRecordBuffer

                             
//---------------------------------------------------------------------------
// clear BLOB field in record buffer
//---------------------------------------------------------------------------
procedure TACRTemporaryTableData.ClearBLOBFieldInRecordBuffer(
          var RecordBuffer: TACRRecordBuffer;
          FieldNo:    Integer
          );
begin
; // do nothing by default
end; // ClearBLOBFieldInRecordBuffer


//---------------------------------------------------------------------------
// load from stream
//---------------------------------------------------------------------------
procedure TACRTemporaryTableData.LoadTableFromStream(Stream: TStream);
var Size: Int64;
    s:    String;
    i:    Integer;
begin
  repeat
   s := GetTemporaryName(ACRTemporaryTableName);
  until (not FDatabaseData.TableExists(s));
  FTableName := s;
  // load fields
  if (FFieldManager <> nil) then
   FFieldManager.Free;
  FFieldManager := TACRBaseFieldManager.Create(Self,FSequenceManager);
  FFieldManager.LoadFromStream(Stream);
  FFieldManager.FieldDefs.RecalcFieldOffsets;
  FBLOBFieldsPresent := False;
  for i := 0 to FFieldManager.FieldDefs.Count - 1 do
   if (IsBLOBFieldType(FFieldManager.FieldDefs[i].BaseFieldType)) then
    begin
     FBLOBFieldsPresent := True;
     if (FFieldManager.FieldDefs[i].BLOBBlockSize = 0) then
      raise EACRException.Create(11333,ErrorLZeroBlockSizeIsNotAllowedForField,[FFieldManager.FieldDefs[i].Name]);
    end;

  if (FConstraintManager <> nil) then
    FConstraintManager.Free;
  // Create ConstraintManager
  FConstraintManager := TACRBaseConstraintManager.Create(Self);

  // load indexes
  if (FIndexManager <> nil) then
   FIndexManager.Free;
  FIndexManager := TACRBaseIndexManager.Create(Self);
  FIndexManager.LoadFromStream(Stream);

  CreateRecordManager;
  FRecordManager.LoadFromStream(Stream);

  LoadDataFromStream(Size,SizeOf(Size),Stream,11195);
  if (Size > 0) then
   begin
    CreateBLOBFile;
    FBLOBFile.LoadFromStreamWithPosition(Stream,Stream.Position,Size);
   end;
end; // LoadTableFromStream


//---------------------------------------------------------------------------
// save to stream
//---------------------------------------------------------------------------
procedure TACRTemporaryTableData.SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
var Size: Int64;
begin
  FFieldManager.SaveToStream(Stream);
  FIndexManager.SaveToStream(Stream);
  FRecordManager.SaveToStream(Stream);
  if (FBLOBFile = nil) then
    Size := 0
  else
    Size := FBLOBFile.Size;
  SaveDataToStream(Size,SizeOf(Size),Stream,11194);
  if (Size > 0) then
    FBLOBFile.SaveToStream(Stream);
  Stream.Position := Stream.Size;  
end; // SaveTableToStream


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRTempEngine> initialization finished');
{$ENDIF}

end.
