unit ACRBase;

interface

uses SysUtils, Classes, Db,

// Accuracer units
{$I ACRVer.inc}

     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRMemory,
     ACRCompression,
     ACRLexer,
     ACRTypes,
     ACRConverts,
     ACRTypesRoutines,
     ACRVariant,
     ACRExcept,
     ACRConst;

type
 // Events
 TACRFilterRecord = Pointer;


type

  TACRBaseSession = class;
  TACRCursor = class;
  TACRIndexDef = class;
  TACRFieldDef = class;

  TACRFieldDefs = class;
  TACRIndexDefs = class;

  TACRCursorPos = record
   FirstPosition: Boolean;
   LastPosition:  Boolean;
   RecordID:      TACRRecordID;
  end;
  PACRCursorPos = ^TACRCursorPos;



{$IFDEF SERVER_VERSION}
////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseServer
//
////////////////////////////////////////////////////////////////////////////////
{
  TACRBaseServer = class (TComponent) // to avid circular modules reference
   protected
    FLocalPort:               String;
    FServerID:                Integer;
   published
    property LocalPort: String read FLocalPort write FLocalPort;
    property ServerID: Integer read FServerID write FServerID;
  end;// TACRBaseServer
}  
{$ENDIF}

////////////////////////////////////////////////////////////////////////////////
//
// TACRMetaObjectDef
//
////////////////////////////////////////////////////////////////////////////////


  TACRMetaObjectDef = class (TObject)
   private
    FName:       TACRObjectName;
    FObjectID:   TACRObjectID;
    FSchemaName: String;
    FSchemaID:   TACRObjectID;
   public
    constructor Create;
    procedure Assign(Source: TACRMetaObjectDef); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
   public
    property Name: TACRObjectName read FName write FName;
    property ObjectID: TACRObjectID read FObjectID write FObjectID;
    property SchemaName: String read FSchemaName write FSchemaName;
    property SchemaID: TACRObjectID read FSchemaID write FSchemaID;
  end; // TACRMetaObjectDef


////////////////////////////////////////////////////////////////////////////////
//
// TACRSequenceDef
//
////////////////////////////////////////////////////////////////////////////////


  TACRSequenceDef = class (TACRMetaObjectDef)
   private
    FDataType:  TACRBaseFieldType;
    FMinValue:  TACRSequenceValue;
    FMaxValue:  TACRSequenceValue;
    FLastValue: TACRSequenceValue;
    FIncrement: TACRSequenceValue;
    FCycled:    Boolean;
   public
    // constructor
    constructor Create;
    // Assign data from an another sequence
    procedure   Assign(Source: TACRMetaObjectDef); override;
    // GetNextVal
    function GetNextVal: TACRSequenceValue; virtual;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
   public
     property DataType:  TACRBaseFieldType read FDataType  write FDataType;
     property MinValue:  TACRSequenceValue read FMinValue  write FMinValue;
     property MaxValue:  TACRSequenceValue read FMaxValue  write FMaxValue;
     property LastValue: TACRSequenceValue read FLastValue write FLastValue;
     property Increment: TACRSequenceValue read FIncrement write FIncrement;
     property Cycled:    Boolean           read FCycled    write FCycled;
  end; // TACRSequenceDef




////////////////////////////////////////////////////////////////////////////////
//
// TACRFieldDef
//
////////////////////////////////////////////////////////////////////////////////


  TACRFieldDef = class (TACRMetaObjectDef)
   private
    FBaseFieldType:             TACRBaseFieldType;
    FAdvancedFieldType:         TACRAdvancedFieldType;
    FFieldSize:                 Integer;
    FDiskDataSize:              Integer; // size of binary data in disk page or table file
    FMemoryDataSize:            Integer; // size of binary data in record buffer
    FDiskOffset:                Integer; // offset to binary data in disk page or table file
    FMemoryOffset:              Integer; // offset to binary data in record buffer
    FFieldNoReference:          Integer; // index of this field in TableData.FieldManager.FieldDefs

    // Default Value
    FSequenceDefObjectId:       TACRObjectID;
    FDefaultValue:              TACRVariant;

    // Blob data
    FBLOBCompressionAlgorithm:  TACRCompressionAlgorithm;
    FBLOBCompressionMode:       TACRCompressionMode;
    FBLOBBlockSize:             Integer;

    // Autoinc settings
    FAutoincIncrement:    Int64;
    FAutoincInitialValue: Int64;
    FAutoincMinValue:     Int64;
    FAutoincMaxValue:     Int64;
    FAutoincCycled:       ByteBool;

    FEngineVersion:       Double;

   private
    procedure RecalcInternalSizes;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TACRMetaObjectDef); override;
    procedure SetFieldDefDataType(
                                  AdvancedFieldType: TACRAdvancedFieldType;
                                  FieldSize:        Integer = 0
                                 ); overload;
    procedure SetFieldDefDataType(
                                  BaseFieldType: TACRBaseFieldType;
                                  FieldSize:        Integer = 0
                                 ); overload;

    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;

   public
    property BaseFieldType: TACRBaseFieldType read FBaseFieldType write FBaseFieldType;
    property AdvancedFieldType: TACRAdvancedFieldType read FAdvancedFieldType write FAdvancedFieldType;
    property FieldSize: Integer read FFieldSize write FFieldSize;
    property DiskOffset: Integer read FDiskOffset write FDiskOffset;
    property MemoryOffset: Integer read FMemoryOffset write FMemoryOffset;
    property DiskDataSize: Integer read FDiskDataSize write FDiskDataSize;
    property MemoryDataSize: Integer read FMemoryDataSize write FMemoryDataSize;
    property FieldNoReference: Integer read FFieldNoReference write FFieldNoReference;

    //property DefaultValueType: TACRDefaultValueType read FDefaultValueType write FDefaultValueType;
    property SequenceDefObjectId: TACRObjectID read FSequenceDefObjectId write FSequenceDefObjectId;
    property DefaultValue: TACRVariant read FDefaultValue write FDefaultValue;

    property AutoincIncrement: Int64    read FAutoincIncrement  write   FAutoincIncrement;
    property AutoincInitialValue: Int64 read FAutoincInitialValue write FAutoincInitialValue;
    property AutoincMinValue:  Int64    read FAutoincMinValue   write   FAutoincMinValue;
    property AutoincMaxValue:  Int64    read FAutoincMaxValue   write   FAutoincMaxValue;
    property AutoincCycled:    ByteBool read FAutoincCycled     write   FAutoincCycled;

    property BLOBCompressionAlgorithm: TACRCompressionAlgorithm read FBLOBCompressionAlgorithm write FBLOBCompressionAlgorithm;
    property BLOBCompressionMode: TACRCompressionMode read FBLOBCompressionMode write FBLOBCompressionMode;
    property BLOBBlockSize: Integer read FBLOBBlockSize write FBLOBBlockSize;

    property EngineVersion: Double read FEngineVersion write FEngineVersion;
  end; // TACRFieldDef


////////////////////////////////////////////////////////////////////////////////
//
// TACRIndexDef
//
////////////////////////////////////////////////////////////////////////////////


  TACRIndexType = (itBTree,itAnotherOne);

  TACRIndexColumn = class (TObject)
   private
    FFieldName:        String;
    FDescending:       ByteBool;
    FCaseInsensitive:  ByteBool;
   public
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
   public
    property FieldName: String read FFieldName write FFieldName;
    property Descending: ByteBool read FDescending write FDescending;
    property CaseInsensitive: ByteBool read FCaseInsensitive write FCaseInsensitive;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRIndexDef
//
////////////////////////////////////////////////////////////////////////////////

  TACRIndexDef = class (TACRMetaObjectDef)
   private
    FIndexType:           TACRIndexType; // BTree or other
    FIndexColumns:        array of TACRIndexColumn;
    FUnique:              ByteBool;
    FPrimary:             ByteBool;
    FRootPageNo:          TACRPageNo;
    FTemporary:           ByteBool;

    function GetIndexColumn(Index: Integer): TACRIndexColumn;
    function GetColumnCount: Integer;
    procedure SetColumnCount(Value: Integer);
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TACRMetaObjectDef); override;
    procedure AssignByNames(FieldNames, AscDescList, CaseSensitivityList: TStringList);
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    // return -1 if field was not found, otherwise return column index
    function FindField(FieldName: String): Integer;
   public
    property IndexType: TACRIndexType read FIndexType write FIndexType;
    property Columns[Index: Integer]: TACRIndexColumn read GetIndexColumn;
    property ColumnCount: Integer read GetColumnCount write SetColumnCount;
    property Unique: ByteBool read FUnique write FUnique;
    property Primary: ByteBool read FPrimary write FPrimary;
    property RootPageNo: TACRPageNo read FRootPageNo write FRootPageNo;
    property Temporary: ByteBool read FTemporary write FTemporary;
  end; // TACRIndexDef


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDef
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintType = (ctPK, ctFK, ctUnique, ctNotNull, ctCheck, ctFKAction);
  TACRConstraintDef = class (TACRMetaObjectDef)
   private
    FConstraintType:  TACRConstraintType;
   public
    procedure Assign(Source: TACRMetaObjectDef); override;
   public
    property ConstraintType: TACRConstraintType read FConstraintType write FConstraintType;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefNotNull
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintDefNotNull = class (TACRConstraintDef)
   private
    FTableSchemaName:          TACRObjectName;  // Shema
    FTableSchemaObjectID:      TACRObjectID;
    FTableName:                TACRObjectName;  // Table
    FTableObjectID:            TACRObjectID;
    FColumnName:               TACRObjectName;  // Column
    FColumnObjectID:           TACRObjectID;
   public
    constructor Create;
    procedure Assign(Source: TACRMetaObjectDef); override;
    procedure SetNames(
                        ColumnName: String;
                        TableName: String = '';
                        SchemaName: String = ''
                       );
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
   public

// never used
    property TableSchemaName: TACRObjectName read FTableSchemaName write FTableSchemaName;
    property TableSchemaObjectID: TACRObjectID read FTableSchemaObjectID write FTableSchemaObjectID;
    property TableName: TACRObjectName read FTableName write FTableName;
    property TableObjectID: TACRObjectID read FTableObjectID write FTableObjectID;
// never used

    property ColumnName: TACRObjectName read FColumnName write FColumnName;
    property ColumnObjectID: TACRObjectID read FColumnObjectID write FColumnObjectID;
  end;// TACRConstraintDefNotNull


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefCheck
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintDefCheck = class (TACRConstraintDef)
   private
    FTableSchemaName:     TACRObjectName;  // Shema
    FTableSchemaObjectID: TACRObjectID;
    FTableName:           TACRObjectName;  // Table
    FTableObjectID:       TACRObjectID;
    FColumnName:          TACRObjectName;  // Column
    FColumnObjectID:      TACRObjectID;
    FMinValue:            TACRVariant;
    FMaxValue:            TACRVariant;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TACRMetaObjectDef); override;
    procedure SetNames(
                        ColumnName: String;
                        TableName: String = '';
                        SchemaName: String = ''
                       );
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
   public
    property MinValue: TACRVariant read FMinValue;
    property MaxValue: TACRVariant read FMaxValue;

// never used
    property TableSchemaName: TACRObjectName read FTableSchemaName write FTableSchemaName;
    property TableSchemaObjectID: TACRObjectID read FTableSchemaObjectID write FTableSchemaObjectID;
    property TableName: TACRObjectName read FTableName write FTableName;
    property TableObjectID: TACRObjectID read FTableObjectID write FTableObjectID;
// never used

    property ColumnName: TACRObjectName read FColumnName write FColumnName;
    property ColumnObjectID: TACRObjectID read FColumnObjectID write FColumnObjectID;
  end;//TACRConstraintDefCheck


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefUnique
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintColumn = record
    ColumnName:          TACRObjectName;  // Column
    ColumnObjectID:      TACRObjectID;
  end;


  TACRConstraintDefUnique = class (TACRConstraintDef)
   private
    FTableSchemaName:          TACRObjectName;  // Shema
    FTableSchemaObjectID:      TACRObjectID;
    FTableName:                TACRObjectName;  // Table
    FTableObjectID:            TACRObjectID;
    FIndexName:                TACRObjectName;  // Index ID
    FIndexObjectID:            TACRObjectID;
   public
    Columns: array of TACRConstraintColumn; // Columns
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TACRMetaObjectDef); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
   public

// never used
    property TableSchemaName: TACRObjectName read FTableSchemaName write FTableSchemaName;
    property TableSchemaObjectID: TACRObjectID read FTableSchemaObjectID write FTableSchemaObjectID;
    property TableName: TACRObjectName read FTableName write FTableName;
    property TableObjectID: TACRObjectID read FTableObjectID write FTableObjectID;
// never used


    property IndexName: TACRObjectName read FIndexName write FIndexName;
    property IndexObjectID: TACRObjectID read FIndexObjectID write FIndexObjectID;
  end;//TACRConstraintDefUnique


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefPrimary
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintDefPrimary = class (TACRConstraintDefUnique)
   public
    constructor Create;
  end;//TACRConstraintDefPrimary


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefForeignKeyAction
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintForeignKeyMatchType = (cfkmtDefault,cfkmtFull,cfkmtPartial);
  TACRConstraintForeignKeyAction = (cfkaDefault, cfkaCascade,cfkaSetNull,cfkaSetDefault,cfkaNoAction);

  // FKAction is an action of master table on update or/and on delete for FK in detail table
  TACRConstraintDefForeignKeyAction = class (TACRConstraintDef)
   private
    FReferencedTableName:         TACRObjectName;  // Referenced Table
    FReferencedTableObjectID:     TACRObjectID;
    FReferencedFKName:            TACRObjectName;  // Referenced FK
    FReferencedFKObjectID:        TACRObjectID;
    FDeleteAction:                TACRConstraintForeignKeyAction;
    FUpdateAction:                TACRConstraintForeignKeyAction;
    FMatchType:                   TACRConstraintForeignKeyMatchType;
   public
    Columns: array of TACRConstraintColumn; // Columns in detail table for Set Null, SetDefault, Cascade
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TACRMetaObjectDef); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
   public
    property ReferencedTableName: TACRObjectName read FReferencedTableName write FReferencedTableName;
    property ReferencedTableObjectID: TACRObjectID read FReferencedTableObjectID write FReferencedTableObjectID;
    property ReferencedFKName: TACRObjectName read FReferencedFKName write FReferencedFKName;
    property ReferencedFKObjectID: TACRObjectID read FReferencedFKObjectID write FReferencedFKObjectID;
    property DeleteAction: TACRConstraintForeignKeyAction read FDeleteAction write FDeleteAction;
    property UpdateAction: TACRConstraintForeignKeyAction read FUpdateAction write FUpdateAction;
    property MatchType: TACRConstraintForeignKeyMatchType read FMatchType write FMatchType;
  end;//TACRConstraintDefForeignKeyAction


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefForeignKey
//
////////////////////////////////////////////////////////////////////////////////


  TACRConstraintDefForeignKey = class (TACRConstraintDefForeignKeyAction)
   public
    constructor Create;
  end;//TACRConstraintDefForeignKey


////////////////////////////////////////////////////////////////////////////////
//
// Meta Objects Defs
//
////////////////////////////////////////////////////////////////////////////////

 TACRMetaObjectDefs = class(TObject)
  protected
   FDefsList:   TACRSortedStringPtrArray;
   FLoadedItemCount:  Integer;
  private
   function GetCount: Integer; virtual;
   function GetDef(Index: Integer): TACRMetaObjectDef;
   procedure SetDef(Index: Integer; Value: TACRMetaObjectDef);

   procedure Add(MetaObjectDef: TACRMetaObjectDef); virtual;
   function InternalAddCreated: TACRMetaObjectDef; virtual;
  public
   procedure LoadFromStream(Stream: TStream); virtual;
   procedure SaveToStream(Stream: TStream); virtual;

   constructor Create;
   destructor Destroy; override;
   procedure Assign(Source: TACRMetaObjectDefs); virtual;

   procedure Delete(Index: Integer); virtual;
   procedure Insert(Index: Integer; MetaObjectDef: TACRMetaObjectDef); virtual;
   procedure Move(CurIndex, NewIndex: Integer); virtual;
   procedure Clear; virtual;

   function GetDefNumberByName(Name: String): Integer; 
   function GetDefByName(Name: String): TACRMetaObjectDef;
   function GetDefNumberByObjectId(id: TACRObjectID): Integer;
   function GetDefByObjectId(id: TACRObjectID): TACRMetaObjectDef;
  public
   property Count: Integer read GetCount;
   property Items[Index: Integer]: TACRMetaObjectDef read GetDef write SetDef; default;
 end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRIndexDefs
//
////////////////////////////////////////////////////////////////////////////////


 TACRIndexDefs = class(TACRMetaObjectDefs)
  private
   function GetIndexDef(Index: Integer): TACRIndexDef; virtual;
   procedure SetIndexDef(Index: Integer; Value: TACRIndexDef); virtual;
   function InternalAddCreated: TACRMetaObjectDef; override;
  public
   // AscDesc and CaseSensitivity lists should contain constants ACR_ASC, ACR_DESC, ACR_NO_CASE, ACR_CASE
   function IsIndexExists(FieldNames, AscDescList, CaseSensitivityList: TStringList): Boolean;
   function FindIndex(FieldNames, AscDescList, CaseSensitivityList: TStringList): TACRObjectID;
   function FindPrimaryIndex: TACRIndexDef;
   function AddCreated: TACRIndexDef;
   function GetIndexDefByName(Name: String): TACRIndexDef;
   procedure LoadFromStream(Stream: TStream); override;
   procedure SaveToStream(Stream: TStream); override;
  public
   property Items[Index: Integer]: TACRIndexDef read GetIndexDef write SetIndexDef; default;
 end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRFieldDefs
//
////////////////////////////////////////////////////////////////////////////////


 TACRFieldDefs = class(TACRMetaObjectDefs)
  private
   FVarcharOrBLOBFieldsExists: Boolean;
   FAutoIncFieldsExists:       Boolean;
   FDefaultValuesExists:       Boolean;
   FEngineVersion:             Double;

   function GetDef(Index: Integer): TACRFieldDef;
   procedure SetDef(Index: Integer; Value: TACRFieldDef);
   function InternalAddCreated: TACRMetaObjectDef; override;
  public
   function AddCreated: TACRFieldDef;
   function GetFieldDefByName(Name: String): TACRFieldDef;

   procedure RecalcFieldOffsets; 
   function GetMemoryRecordBufferSize: Integer;

   procedure LoadFromStream(Stream: TStream); override;
   procedure SaveToStream(Stream: TStream); override;
  public
   property Items[Index: Integer]: TACRFieldDef read GetDef write SetDef; default;
   property VarcharOrBLOBFieldsExists: Boolean read FVarcharOrBLOBFieldsExists;
   property AutoIncFieldsExists: Boolean read FAutoIncFieldsExists;
   property DefaultValuesExists: Boolean read FDefaultValuesExists;
   property EngineVersion: Double read FEngineVersion write FEngineVersion;
 end;//TACRFieldDefs


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefs
//
////////////////////////////////////////////////////////////////////////////////

 TACRConstraintDefs =  class(TACRMetaObjectDefs)
  private
   FForeignKeysExists: Boolean;
   FForeignKeysActionsExists: Boolean;
  private
   function GetDef(Index: Integer): TACRConstraintDef;
   procedure SetDef(Index: Integer; Value: TACRConstraintDef);
   function GetForeignKeysActionsUpdateExists: Boolean;
   function GetForeignKeysActionsDeleteExists: Boolean;
  public
   constructor Create;
   procedure Assign(Source: TACRMetaObjectDefs); override;
   procedure Delete(Index: Integer); override;
   // Create TACRConstraintDefNotNull and add it into list
   function AddNotNull: TACRConstraintDefNotNull;
   // Create TACRConstraintDefCheck and add it into list
   function AddCheck: TACRConstraintDefCheck;
   function AddPK: TACRConstraintDefPrimary;
   function AddUnique: TACRConstraintDefUnique;
   function AddFK: TACRConstraintDefForeignKey;
   function AddFKAction: TACRConstraintDefForeignKeyAction;

   procedure LoadFromStream(Stream: TStream); override;
   procedure SaveToStream(Stream: TStream); override;
  public
   property Items[Index: Integer]: TACRConstraintDef read GetDef write SetDef; default;
   property ForeignKeysExists: Boolean read FForeignKeysExists write FForeignKeysExists;
   property ForeignKeysActionsExists: Boolean read FForeignKeysActionsExists  write FForeignKeysActionsExists;
   property ForeignKeysActionsUpdateExists: Boolean read GetForeignKeysActionsUpdateExists;
   property ForeignKeysActionsDeleteExists: Boolean read GetForeignKeysActionsDeleteExists;
 end;//TACRConstraintDefs



////////////////////////////////////////////////////////////////////////////////
//
// TACRSequenceDefs
//
////////////////////////////////////////////////////////////////////////////////

 TACRSequenceDefs = class(TACRMetaObjectDefs)
  private
   function GetDef(Index: Integer): TACRSequenceDef;
   procedure SetDef(Index: Integer; Value: TACRSequenceDef);
   function InternalAddCreated: TACRMetaObjectDef; override;
  public
   function AddCreated: TACRSequenceDef; 
   function GetSequenceDefByName(Name: String): TACRSequenceDef;
  public
   property Items[Index: Integer]: TACRSequenceDef read GetDef write SetDef; default;
 end;

 TACRRestructureInfo = record
    FRestructureBLOBCompression:   TACRCompression;
    FRestructureFieldDefs:         TACRFieldDefs;
    FRestructureIndexDefs:         TACRIndexDefs;
    FRestructureConstraintDefs:    TACRConstraintDefs;
 end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRCursor
//
////////////////////////////////////////////////////////////////////////////////


  TACRCursor = class (TObject)
   public
    FMemoryTableAllocBy:            Integer;
    FSettingProjection:             Boolean;
    FTableName:                     String;
    FIndexName:                     String;
    FIndexID:                       TACRObjectID;
    FReadOnly:                      Boolean;
    FExclusive:                     Boolean;
    FInMemory:                      Boolean;
    FTemporary:                     Boolean;
    FSession:                       TACRBaseSession;
    FIsOpen:                        Boolean;
    FPhysicalOrder:                 Boolean;
    FCurrentRecordPositionInIndex:  TACRIndexPosition;
    // current record buffer
    FCurrentRecordBuffer:           TACRRecordBuffer;
    // distinct
    FDistinctRecordBuffer:          TACRRecordBuffer;
    // buffer with original record, stored on InternalEdit by TACRDataset
    FEditRecordBuffer:              TACRRecordBuffer;
    FConstraintDefs:                TACRConstraintDefs;
    FBLOBCompression:               TACRCompression;
    FFieldDefs:                     TACRFieldDefs;
    FVisibleFieldDefs:              TACRFieldDefs; // visible fields (projection)
    FIndexDefs:                     TACRIndexDefs;
    FBLOBStreams:                   TList;
    FRecordBitmap:                  Pointer;
    FIsClientCursor:                Boolean;
    FCreateTableStarted:            Boolean;
    FRandomOrder:                   Boolean;
   protected
    // BLOBS will be stored as 6 bytes reference:
    // DiskEngine: 4 bytes PAGEID, 2 bytes ObjectID
    // TemporaryEngine: 4 bytes pointer to BLOBDescriptor record, 2 bytes not used
    // MemoryEngine: 4 bytes pointer to BLOBCompressedStream, 2 bytes not used

    // Record Buffer sizes and offsets:
    //              + FieldValuesOffset
    //                                   + BookmarOffset
    //                                              + CalculatedFieldsOffset
    // +------------+--------------------+----------+-----------------+
    // | Null Flags | Field Values,      | Bookmark | Calculated and  |
    // |            | References To BLOB | Bookmark | Lookup Fields   |
    // +------------+--------------------+----------+-----------------+
    //                                              + RecordSize
    //                                                                + RecordBufferSize

    // Key buffer:
    //                                                   + KeyBufferSize
    //                                   + KeyOffset
    // +------------+--------------------+---------------+
    // | Null Flags | Field Values,      | TACRKeyBuffer |
    // |            | References To BLOB |               |
    // +------------+--------------------+---------------+

    FErrorCode:                     TACRErrorCode;
    FErrorMessage:                  String;
    FIsDesignMode:                  Boolean;
    FRecordBufferSize:              Integer;
    FRecordSize:                    Integer;
    FKeyBufferSize:                 Integer;
    FKeyOffset:                     Integer;
    FKeyFieldCount:                 Integer;
    FFieldValuesOffset:             Integer;
    FCalculatedFieldsOffset:        Integer;
    FBookmarkOffset:                Integer;
    FFilterExpression:              Pointer;
    FSQLFilterExpression:           Pointer;
    FKeyBuffer:                     TACRRecordBuffer;
    FRangeStartBuffer:              TACRRecordBuffer;
    FRangeEndBuffer:                TACRRecordBuffer;
    FRangeStartExclusive:           Boolean;
    FRangeEndExclusive:             Boolean;
    FRangeStartKeyFieldCount:       Integer;
    FRangeEndKeyFieldCount:         Integer;
    FDistinctFieldCount:            Integer;
    FRepair:                        Boolean;
   protected
    procedure SetIndexName(Value: String);
   public
    property RecordBufferSize: Integer read FRecordBufferSize write FRecordBufferSize;
    property RecordSize: Integer read FRecordSize write FRecordSize;
    property FieldValuesOffset: Integer read FFieldValuesOffset write FFieldValuesOffset;
    property CalculatedFieldsOffset: Integer read FCalculatedFieldsOffset write FCalculatedFieldsOffset;
    property BookmarkOffset: Integer read FBookmarkOffset write FBookmarkOffset;
    property KeyOffset: Integer read FKeyOffset write FKeyOffset;
    property KeyBufferSize: Integer read FKeyBufferSize write FKeyBufferSize;
    property Repair: Boolean read FRepair write FRepair;
   public

    // table operations
    procedure CreateTable(
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); virtual; abstract;
    procedure DeleteTable(Cascade: Boolean = False); virtual; abstract;
    procedure EmptyTable; virtual; abstract;
    procedure RenameTable(NewTableName: String); virtual; abstract;
    procedure RenameField(FieldName, NewFieldName: String); virtual;
    function RepairTable(var Log: String; NewSession: Pointer = nil): Boolean; virtual; abstract;
    procedure AddForeignKey(ConstraintDef: TACRConstraintDefForeignKey); virtual; abstract;
    procedure DeleteConstraint(Name: String; Cascade: Boolean; FKPartialDelete: Boolean); virtual; abstract;
    procedure LoadTableFromStream(
                        Stream:               TStream
                       ); virtual; abstract;
    procedure SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      ); virtual; abstract;
    function ExportTableToSQL(
                              ExportStructure:      Boolean = True;
                              AddDropTableCommand:  Boolean = True;
                              ExportIndexes:        Boolean = True;
                              AddDropIndexCommand:  Boolean = True;
                              ExportData:           Boolean = True;
                              ExportBLOBFields:     Boolean = True;
                              UseBracketsForNames:  Boolean = True
                            ): String; virtual; abstract;
    procedure InternalInitFieldDefs; virtual; abstract;
    procedure OpenTableByFieldDefs(
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                       ); virtual; abstract;
    procedure UpdateTableDefinitions; virtual;
    procedure CloseTable; virtual; abstract;
    // initialize record buffer
    procedure InternalInitRecord(RecordBuffer: TACRRecordBuffer; InsertMode: Boolean); virtual; abstract;

    // index operations
    function GetIndexDefs: TACRIndexDefs; virtual;
    procedure AddIndex(IndexDef: TACRIndexDef); virtual; abstract;
    procedure DeleteIndex(Name: String); virtual; abstract;
    procedure DeleteAllIndexes; virtual; abstract;
    // return index name of the index or '' if not found
    function FindIndex(FieldNamesList,
                AscDescList, CaseSensitivityList: TStringList): String; virtual; abstract;
    // return index name
    function FindOrCreateIndex(FieldNamesList,
                AscDescList, CaseSensitivityList: TStringList): String; virtual; abstract;

    // check field value and if not null move data from RecordBuffer to Buffer
    function GetFieldData(
                          FieldNo:      Integer; // field no
                          Buffer:       Pointer; // buffer
                          RecordBuffer: TACRRecordBuffer // record buffer
                         ): Boolean;
    // set field data from Buffer to RecordBuffer
    procedure SetFieldData(
                            FieldNo:       Integer;
                            Buffer:        Pointer;
                            RecordBuffer:  TACRRecordBuffer // record buffer
                          );
    procedure GetFieldValue(Value: TACRVariant; FieldNo: Integer; DirectAccess: Boolean; CopyFlag: Boolean = True);
    procedure SetFieldValue(
                            Value:        TACRVariant;
                            FieldNo:      Integer;
                            DirectAccess: Boolean;
                            RecordBuffer: TACRRecordBuffer = nil
                           );
    // allocate record buffer and set null flags
    function AllocateRecordBuffer: TACRRecordBuffer;
    // free record buffer
    procedure FreeRecordBuffer(var Buffer: TACRRecordBuffer);
    // allocate record buffer and set null flags
    function AllocateKeyRecordBuffer: TACRRecordBuffer;
    // initialize record buffer
    procedure InternalInitKeyBuffer(RecordBuffer: TACRRecordBuffer);


    //---------------------------------------------------------------------------
    // navigation & bookmark methods
    //---------------------------------------------------------------------------

    // return true if current record exists
    function IsRecordExists: Boolean; virtual; abstract;
    function GetRecordCount: TACRRecordNo; virtual; abstract;
    // get record
    function GetRecordBuffer(
              GetRecordMode:  TACRGetRecordMode
              ): TACRGetRecordResult; virtual; abstract;
    // go to record
    procedure SetRecNo(Value: Int64); virtual; abstract;
    // return current record number
    function GetRecNo: Int64; virtual; abstract;
    // go to first record
    procedure InternalFirst;
    // go to last record
    procedure InternalLast; 
    function SavePosition: Pointer;
    procedure RestorePosition(Pos: Pointer);
    procedure FreePosition(var Pos: Pointer);

    //---------------------------------------------------------------------------
    // insert, edit, post, delete methods
    //---------------------------------------------------------------------------

    // insert record
    procedure InternalInsert; virtual; abstract;
    // edit record
    procedure InternalEdit; virtual; abstract;
    // cancels updates
    procedure InternalCancel(ToInsert: Boolean); virtual; abstract;
    // update record
    procedure InternalPost(ToInsert: Boolean); virtual; abstract;
    // delete record
    procedure InternalDelete; virtual; abstract;
    procedure DeleteVisibleRecords; virtual; abstract;
    procedure UpdateVisibleRecords(FieldNames:   TStringList;
                                   values:       array of TACRVariant;
                                   SkipFKCheck:  Boolean = False
                                   ); virtual; abstract;

    //---------------------------------------------------------------------------
    // search & filter methods
    //---------------------------------------------------------------------------

    // disable record bitmap
    procedure DisableRecordBitmap; virtual; abstract;
    // set distinct
    procedure ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList); virtual; abstract;
    // apply projection
    procedure ApplyProjection(FieldNamesList, AliasList: TStringList); virtual; abstract;
    procedure ActivateFilters(
                              FilterText:      String;
                              CaseInsensitive: Boolean;
                              PartialKey:      Boolean
                            ); virtual; abstract;
    procedure DeactivateFilters; virtual; abstract;
    function Locate(
                    const KeyFields: String;
                    const KeyValues: Variant;
                    CaseInsensitive: Boolean;
                    PartialKey:      Boolean
              ): Boolean; virtual; abstract;
    function FindKey(SearchCondition: TACRSearchCondition): Boolean; virtual; abstract;
    function IsIndexApplied: Boolean;
    function IsDistinctApplied: Boolean;
    function IsFilterApplied: Boolean;
    function IsRangeApplied: Boolean;
    function IsViewRestricted: Boolean;
    procedure ResetRange; virtual; abstract;
    procedure ApplyRange(
                          StartBuffer, EndBuffer: TACRRecordBuffer;
                          StartKeyFieldCount:     Integer;
                          EndKeyFieldCount:       Integer;
                          StartExclusive:         Boolean;
                          EndExclusive:           Boolean
                        ); virtual; abstract;
    // set SQL Filter
    procedure SetSQLFilter(FilterExpr: TObject); virtual; abstract;

    //---------------------------------------------------------------------------
    // BLOB methods
    //---------------------------------------------------------------------------

    function InternalCreateBlobStream(
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ):TACRStream; virtual; abstract;

    procedure InternalCloseBLOB(FieldNo: Integer); virtual; abstract;

    // clear blob streams
    procedure ClearBLOBStreams(WriteOnly: Boolean = False); virtual; abstract;

    function LastAutoincValue(FieldNo: Integer): Int64; virtual; abstract;
    procedure SetLastAutoincValue(Value: Int64; FieldNo: Integer); virtual; abstract;

    procedure LockTableForRead; virtual; abstract;
    procedure UnlockTableForRead; virtual; abstract;

    procedure LockTableS(ErrorCode: Integer); virtual; abstract;
    procedure UnlockTableS(ErrorCode: Integer); virtual; abstract;
    procedure LockTableIRW(ErrorCode: Integer); virtual; abstract;
    procedure UnlockTableIRW(ErrorCode: Integer); virtual; abstract;
    procedure LockTableRW(ErrorCode: Integer); virtual; abstract;
    procedure UnlockTableRW(ErrorCode: Integer); virtual; abstract;

    function TryToLockTableS: Boolean; virtual; abstract;
    function TryToUnlockTableS: Boolean; virtual; abstract;
    function TryToLockTableIRW: Boolean; virtual; abstract;
    function TryToUnlockTableIRW: Boolean; virtual; abstract;
    function TryToLockTableRW: Boolean; virtual; abstract;
    function TryToUnlockTableRW: Boolean; virtual; abstract;
    // Destroy should call DatabaseData.FreeTableData
   public
    // for bookmarks
    CurrentRecordID:               TACRRecordID;
    // position for navigation
    FirstPosition:                 Boolean;
    LastPosition:                  Boolean;
    FilterRecord:                  TACRFilterRecord;
    Dataset:                       Pointer;

    property RandomOrder: Boolean read FRandomOrder write FRandomOrder;
    property RecordBitmap: Pointer read FRecordBitmap write FRecordBitmap;
    property DistinctFieldCount: Integer read FDistinctFieldCount write FDistinctFieldCount;
    property FilterExpression: Pointer read FFilterExpression write FFilterExpression;
    property SQLFilterExpression: Pointer read FSQLFilterExpression write FSQLFilterExpression;
    property BLOBStreams: TList read FBLOBStreams;
    property ErrorCode: TACRErrorCode read FErrorCode write FErrorCode;
    property ErrorMessage: String read FErrorMessage write FErrorMessage;
    property IsDesignMode: Boolean read FIsDesignMode write FIsDesignMode;
    property TableName: String read FTableName write FTableName;
    property IndexName: String read FIndexName write SetIndexName;
    property IndexID: TACRObjectID read FIndexID write FIndexID;
    property ReadOnly: Boolean read FReadOnly write FReadOnly;
    property Exclusive: Boolean read FExclusive write FExclusive;
    property InMemory: Boolean read FInMemory write FInMemory;
    property MemoryTableAllocBy: Integer read FMemoryTableAllocBy write FMemoryTableAllocBy;
    property Temporary: Boolean read FTemporary write FTemporary;
    property Session: TACRBaseSession read FSession write FSession;
    property IsOpen: Boolean read FIsOpen write FIsOpen;
    property RecordCount: Int64 read GetRecordCount;
    property FieldDefs: TACRFieldDefs read FFieldDefs;
    property VisibleFieldDefs: TACRFieldDefs read FVisibleFieldDefs;
    property IndexDefs: TACRIndexDefs read GetIndexDefs;
    property ConstraintDefs: TACRConstraintDefs read FConstraintDefs;
    // set it before call CreateTable
    property BLOBCompression: TACRCompression read FBLOBCompression
             write FBLOBCompression;
    property KeyFieldCount: Integer read FKeyFieldCount write FKeyFieldCount;
    property KeyBuffer: TACRRecordBuffer read FKeyBuffer write FKeyBuffer;
    property RangeStartBuffer: TACRRecordBuffer read FRangeStartBuffer
             write FRangeStartBuffer;
    property RangeEndBuffer:   TACRRecordBuffer read FRangeEndBuffer
             write FRangeEndBuffer;
    property RangeStartExclusive: Boolean read FRangeStartExclusive
             write FRangeStartExclusive;
    property RangeEndExclusive: Boolean read FRangeEndExclusive
             write FRangeEndExclusive;
    property RangeStartKeyFieldCount: Integer read FRangeStartKeyFieldCount
             write FRangeStartKeyFieldCount;
    property RangeEndKeyFieldCount: Integer read FRangeEndKeyFieldCount
             write FRangeEndKeyFieldCount;
    property CurrentRecordBuffer: TACRRecordBuffer read FCurrentRecordBuffer
              write FCurrentRecordBuffer;
    property DistinctRecordBuffer: TACRRecordBuffer read FDistinctRecordBuffer
              write FDistinctRecordBuffer;
    property EditRecordBuffer: TACRRecordBuffer read FEditRecordBuffer
              write FEditRecordBuffer;
    property PhysicalOrder: Boolean read FPhysicalOrder write FPhysicalOrder;
    property CurrentRecordPositionInIndex: TACRIndexPosition
              read FCurrentRecordPositionInIndex
              write FCurrentRecordPositionInIndex;
    property IsClientCursor: Boolean read FIsClientCursor write FIsClientCursor;
    property CreateTableStarted: Boolean read FCreateTableStarted write FCreateTableStarted;
  end; // TACRCursor



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseSession
//
////////////////////////////////////////////////////////////////////////////////

  // manager for TACRSession
  TACRSessionComponentManager = class(TObject)
  end;

  // Session object base class
  TACRSessionNamedObject = class(TObject)
   private
    FName: String;
   public
    constructor Create(Name: String);
   public
    property Name: String read FName write FName;
  end;

  // Sequence last value stored in user session
  TACRSessionNamedObjectSequenceValue = class(TACRSessionNamedObject)
   public
    FValue: TACRSequenceValue;
   public
    constructor Create(ValueName: String; Value: TACRSequenceValue);
   public
    property Value: TACRSequenceValue read FValue write FValue;
  end;

  // for database component
  TACRBaseSession = class (TObject)
   private
    FSessionHandle:           TACRSessionComponentManager;
    FSessionNamedObjectList:  TACRSortedStringPtrArray;
   protected
    FDatabaseName:            String; // name of database
    FDatabaseFileName:        String; // database file name
    FSessionID:               TACRSessionID;
    FReadOnly:                Boolean;
    FExclusive:               Boolean;
    FInMemory:                Boolean;
    FOptions:                 TACROptions;
    FLockParams:              TACRLockParams;
    FCryptoParams:            TACRCryptoParams;

   protected
    // db connected?
    function GetConnected: Boolean;  virtual;  abstract;
    // connect / disconnect
    procedure SetConnected(Value: Boolean); virtual; abstract;
   public
    FConnectParams:           TACRConnectParams;

   public
    // constructor
    constructor Create;
    // destructor
    destructor Destroy; override;

    // create database
    procedure CreateDatabase; virtual;
    // delete database
    procedure DeleteDatabase; virtual;
    // rename database
    procedure RenameDatabase(NewDatabaseFileName: String); virtual;
    // flush file buffers
    procedure FlushFileBuffers; virtual; abstract;
    // return database format version
    function GetFormatVersion: Double; virtual; abstract;
    // return total number of pages
    function GetTotalPageCount : Integer; virtual; abstract;
    // return number of free pages
    function GetFreePageCount : Integer; virtual; abstract;
    // return true if database is encrypted
    function IsDatabaseEncrypted: Boolean; virtual; abstract;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword: Boolean; virtual; abstract;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid: Boolean; virtual; abstract;
    // check if database exists
    function GetDatabaseExists: Boolean; virtual; abstract;
    procedure GetTablesList(List: TStrings); virtual; abstract;
    function TableExists(TableName: String): Boolean; virtual; abstract;
    // makes Exe database from edb file
    procedure MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string); virtual; abstract;
    // removes database file from executable database file
    procedure RemoveDatabaseFromExe; virtual; abstract;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile: Boolean; virtual; abstract;

    //------------- session variables and sequence values ----------------------
    // Get Named object from session
    function GetNamedObject(ObjectName: String): TACRSessionNamedObject;
    // Set Named object to session
    procedure SetSequenceValue(const Name: TACRObjectName; const Value: TACRSequenceValue);

    //------------------------ Transactions ------------------------------------
   protected
    // retrun true if database has active transaction
    function GetInTransaction: Boolean; virtual; abstract;
   public
    // start a transaction
    procedure StartTransaction; virtual; abstract;
    // apply changes made by transaction
    procedure Commit(FlushFileBuffers: Boolean = True); virtual; abstract;
    // cancel changes made by transaction
    procedure Rollback; virtual; abstract;

    //------------------------ ClientServer ------------------------------------
    // for client and server sessions
    procedure ReceiveData(Buffer: PChar; BufferSize: Integer); virtual;
    procedure OnDisconnect; virtual;
    procedure RemoveAllLocks; virtual;

   public
    property Connected: Boolean read GetConnected write SetConnected default False;
    property DatabaseFileName: string read FDatabaseFileName write FDatabaseFileName;
    property DatabaseName: string read FDatabaseName write FDatabaseName;
    property SessionID: TACRSessionID read FSessionID write FSessionID;
    property InTransaction: Boolean read GetInTransaction;
    property ReadOnly: Boolean read FReadOnly write FReadOnly;
    property Exclusive: Boolean read FExclusive write FExclusive;
    property InMemory: Boolean read FInMemory write FInMemory;
    property SessionComponentManager: TACRSessionComponentManager
             read FSessionHandle write FSessionHandle;
    property LockParams: TACRLockParams read FLockParams write FLockParams;
    property Options: TACROptions read FOptions write FOptions;
    property CryptoParams: TACRCryptoParams read FCryptoParams write FCryptoParams;
    property ConnectParams: TACRConnectParams read FConnectParams write FConnectParams;
  end; // TACRBaseSession


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLProcessor
//
////////////////////////////////////////////////////////////////////////////////


  TACRSQLProcessor = class (TObject)
   protected
    FACRQuery:      TDataSet;
    FReadOnly:      Boolean;
    FRequestLive:   Boolean;
    FInMemory:      Boolean;
    FRowsAffected:  TACRRecordNo;

    FSqlText:       String;
    FSQLParams:     TACRSQLParams;
    FCursor:        TACRCursor;

   public
    constructor Create(Query: TDataSet);
    destructor Destroy; override;

    function OpenQuery: TACRCursor; virtual;
    procedure ExecuteQuery;

    procedure PrepareStatement(SQLText: PChar);

   public
    property ReadOnly: Boolean read FReadOnly write FReadOnly;
    property RequestLive: Boolean read FRequestLive write FRequestLive;
    property InMemory: Boolean read FInMemory write FInMemory;
    property RowsAffected: TACRRecordNo read FRowsAffected;
    property SQLParams: TACRSQLParams read FSQLParams;
    property ACRQuery: TDataset read FACRQuery;
    property SQLText: String read FSQLText;
  end; // TACRSQLProcessor


implementation

uses

// Accuracer units
     ACRExpressions,
     ACRBaseEngine;



////////////////////////////////////////////////////////////////////////////////
//
// TACRMetaObjectDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRMetaObjectDef.Create;
begin
  FName       := '';
  FObjectID   := OBJECTID_IS_NULL;
  FSchemaName := '';
  FSchemaID   := OBJECTID_IS_NULL;
end;//Create


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRMetaObjectDef.Assign(Source: TACRMetaObjectDef);
begin
  FName := Source.FName;
  FObjectID := Source.FObjectID;
  FSchemaName := Source.FSchemaName;
  FSchemaID := Source.FSchemaID;
end;//Assign


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRMetaObjectDef.LoadFromStream(Stream: TStream);
var Len: Byte;
begin
  LoadDataFromStream(Len,sizeof(Len),Stream,10166);
  SetLength(FName,Len);
  LoadDataFromStream(PChar(@FName[1])^,Len,Stream,10167);
  LoadDataFromStream(FObjectID,sizeof(ObjectID),Stream,10169);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRMetaObjectDef.SaveToStream(Stream: TStream);
var Len: Byte;
begin
  Len := Length(Name);
  SaveDataToStream(Len,Sizeof(Len),Stream,10164);
  SaveDataToStream(PChar(@FName[1])^,Len,Stream,10165);
  SaveDataToStream(FObjectID,sizeof(ObjectID),Stream,10168);
end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRSequenceDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSequenceDef.Create;
begin
  inherited Create;
  FDataType  := bftUnSignedInt32;
  FMinValue  := 0;
  FMaxValue  := High(Cardinal);
  FLastValue := 0;
  FIncrement := 1;
  FCycled    := true;
end;//Create


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRSequenceDef.Assign(Source: TACRMetaObjectDef);
begin
  inherited Assign(Source);
  FDataType  := TACRSequenceDef(Source).FDataType;
  FMinValue  := TACRSequenceDef(Source).FMinValue;
  FMaxValue  := TACRSequenceDef(Source).FMaxValue;
  FLastValue := TACRSequenceDef(Source).FLastValue;
  FIncrement := TACRSequenceDef(Source).FIncrement;
  FCycled    := TACRSequenceDef(Source).FCycled;
end;//Assign


//------------------------------------------------------------------------------
// GetNextVal
//------------------------------------------------------------------------------
function TACRSequenceDef.GetNextVal: TACRSequenceValue;
begin
  FLastValue := FLastValue + FIncrement;
  if (FLastValue > FMaxValue) then
    if (FCycled) then
      FLastValue := FMinValue
    else
      raise EACRException.Create(30009, ErrorGSequenceOverflow, [FName]);
  Result := FLastValue;
end;//GetNextVal


//------------------------------------------------------------------------------
// load sequence
//------------------------------------------------------------------------------
procedure TACRSequenceDef.LoadFromStream(Stream: TStream);
begin
  inherited LoadFromStream(Stream);
  LoadDataFromStream(FDataType,sizeof(FDataType),Stream,10212);
  LoadDataFromStream(FMinValue,sizeof(FMinValue),Stream,10213);
  LoadDataFromStream(FMaxValue,sizeof(FMaxValue),Stream,10214);
  LoadDataFromStream(FLastValue,sizeof(FLastValue),Stream,10215);
  LoadDataFromStream(FIncrement,sizeof(FIncrement),Stream,10216);
  LoadDataFromStream(FCycled,sizeof(FCycled),Stream,10217);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save sequence
//------------------------------------------------------------------------------
procedure TACRSequenceDef.SaveToStream(Stream: TStream);
begin
  inherited SaveToStream(Stream);
  SaveDataToStream(FDataType,sizeof(FDataType),Stream,10206);
  SaveDataToStream(FMinValue,sizeof(FMinValue),Stream,10207);
  SaveDataToStream(FMaxValue,sizeof(FMaxValue),Stream,10208);
  SaveDataToStream(FLastValue,sizeof(FLastValue),Stream,10209);
  SaveDataToStream(FIncrement,sizeof(FIncrement),Stream,10210);
  SaveDataToStream(FCycled,sizeof(FCycled),Stream,10211);
end; // LoadFromStream



////////////////////////////////////////////////////////////////////////////////
//
// TACRFieldDef
//
////////////////////////////////////////////////////////////////////////////////



//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRFieldDef.Create;
begin
  inherited Create;
  FBaseFieldType := bftUnknown;
  FAdvancedFieldType := aftUnknown;
  FFieldSize := 0;
  FDiskDataSize := 0;
  FMemoryDataSize := 0;
  FDiskOffset := 0;
  FMemoryOffset := 0;
  FFieldNoReference := 0;

  // Default Value
  //FDefaultValueType := dvtNull;
  FSequenceDefObjectId := OBJECTID_IS_NULL;
  FDefaultValue := TACRVariant.Create;

  // Blob data
  FBLOBCompressionAlgorithm := acaNone;
  FBLOBCompressionMode := 0;
  FBLOBBlockSize := DefaultBLOBBlockSize;

  // Autoinc settings
  FAutoincIncrement := 1;
  FAutoincInitialValue := 0;
  FAutoincMinValue  := 0;//Low(Int64);
  FAutoincMaxValue  := High(Int64);
  FAutoincCycled    := False;

end;//Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRFieldDef.Destroy;
begin
  FDefaultValue.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// SetFieldDefData ( set Advanced Field Type )
//------------------------------------------------------------------------------
procedure TACRFieldDef.SetFieldDefDataType(AdvancedFieldType: TACRAdvancedFieldType;
                                           FieldSize: Integer);
begin
  FAdvancedFieldType := AdvancedFieldType;
  FBaseFieldType := AdvancedFieldTypeToBaseFieldType(AdvancedFieldType);
  if FBaseFieldType = bftUnknown then
   raise EACRException.Create(30007,ErrorGUnsupportedDataType,
                                    [AftToStr(AdvancedFieldType)]);
  FFieldSize := FieldSize;
  RecalcInternalSizes;
end;//SetFieldDefData


//------------------------------------------------------------------------------
// SetFieldDefData ( set Advanced Field Type )
//------------------------------------------------------------------------------
procedure TACRFieldDef.SetFieldDefDataType(BaseFieldType: TACRBaseFieldType;
                                           FieldSize: Integer);
begin
  FBaseFieldType := BaseFieldType;
  FFieldSize := FieldSize;

  FAdvancedFieldType := BaseFieldTypeToAdvancedFieldType(BaseFieldType);
  if FAdvancedFieldType = aftUnknown then
   raise EACRException.Create(30008,ErrorGUnsupportedDataType,
                                    [BftToStr(BaseFieldType)]);
  RecalcInternalSizes;
end;//SetFieldDefData



//------------------------------------------------------------------------------
// RecalcInternalSizes
//------------------------------------------------------------------------------
procedure TACRFieldDef.RecalcInternalSizes;
begin
  // FMemoryDataSize ...
  FMemoryDataSize := GetDataSizeInMemory(FBaseFieldType, FFieldSize);

  // FDiskDataSize ...
  FDiskDataSize := FMemoryDataSize;
  if (IsBLOBFieldType(FBaseFieldType) or (IsVarcharFieldType(FBaseFieldType))) then
    FDiskDataSize := SizeOf(TACRRecordID);
end;//RecalcInternalSizes



//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRFieldDef.Assign(Source: TACRMetaObjectDef);
var s: TACRFieldDef;
begin
  s := Source as TACRFieldDef;
  inherited Assign(Source);
  FBaseFieldType := s.FBaseFieldType;
  FAdvancedFieldType := s.FAdvancedFieldType;
  FFieldSize := s.FFieldSize;
  FDiskDataSize := s.FDiskDataSize;
  FMemoryDataSize := s.FMemoryDataSize;
  FDiskOffset := s.FDiskOffset;
  FMemoryOffset := s.FMemoryOffset;

  // Default Value
  //FDefaultValueType := s.FDefaultValueType;
  FSequenceDefObjectId := s.FSequenceDefObjectId;
  FDefaultValue.Assign(s.FDefaultValue);

  // Blob data
  FBLOBCompressionAlgorithm := s.FBLOBCompressionAlgorithm;
  FBLOBCompressionMode      := s.FBLOBCompressionMode;
  FBLOBBlockSize            := s.FBLOBBlockSize;

  // Autoinc settings
  FAutoincIncrement := s.FAutoincIncrement;
  FAutoincInitialValue := s.FAutoincInitialValue;
  FAutoincMinValue  := s.FAutoincMinValue;
  FAutoincMaxValue  := s.FAutoincMaxValue;
  FAutoincCycled    := s.FAutoincCycled;

end;//Assign


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRFieldDef.LoadFromStream(Stream: TStream);
begin
  inherited LoadFromStream(Stream);

  LoadDataFromStream(FBaseFieldType,Sizeof(FBaseFieldType),Stream,10187);
  LoadDataFromStream(FAdvancedFieldType,Sizeof(FAdvancedFieldType),Stream,10188);
  LoadDataFromStream(FFieldSize,Sizeof(FFieldSize),Stream,10189);

  FDefaultValue.SetNull(FBaseFieldType);
  if (IsBlobFieldType(FBaseFieldType)
{$IFNDEF SQLMEMTABLE}
// v. 2.00 and 1.xx bug fix with not saving varchar compression in Accuracer
      or (IsVarcharFieldType(FBaseFieldType) and (FEngineVersion >= (2.10 - 0.001)))
{$ENDIF}      
      ) then
   begin
    // load blob params
    LoadDataFromStream(FBLOBCompressionAlgorithm,Sizeof(FBLOBCompressionAlgorithm),Stream,10192);
    LoadDataFromStream(FBLOBCompressionMode,Sizeof(FBLOBCompressionMode),Stream,10193);
    LoadDataFromStream(FBLOBBlockSize,Sizeof(FBLOBBlockSize),Stream,10194);
   end;

  //if (FDefaultValueType = dvtSequence) then
  if IsAutoincFieldType(FAdvancedFieldType) then
   begin
    // sequence
    LoadDataFromStream(FSequenceDefObjectId,Sizeof(FSequenceDefObjectId),Stream,10195);
    LoadDataFromStream(FAutoincIncrement,SizeOf(FAutoincIncrement),Stream,10690);
    LoadDataFromStream(FAutoincInitialValue,SizeOf(FAutoincInitialValue),Stream,10691);
    LoadDataFromStream(FAutoincMinValue,SizeOf(FAutoincMinValue),Stream,10692);
    LoadDataFromStream(FAutoincMaxValue,SizeOf(FAutoincMaxValue),Stream,10693);
    LoadDataFromStream(FAutoincCycled,SizeOf(FAutoincCycled),Stream,10694);
   end
  else
    // load default value
    FDefaultValue.LoadFromStream(Stream);

 RecalcInternalSizes;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRFieldDef.SaveToStream(Stream: TStream);
begin
  inherited SaveToStream(Stream);

  SaveDataToStream(FBaseFieldType,Sizeof(FBaseFieldType),Stream,10170);
  SaveDataToStream(FAdvancedFieldType,Sizeof(FAdvancedFieldType),Stream,10171);
  SaveDataToStream(FFieldSize,Sizeof(FFieldSize),Stream,10172);

  if (IsBlobFieldType(FBaseFieldType)
{$IFNDEF SQLMEMTABLE}
// v. 2.00 and 1.xx bug fix with not saving varchar compression in Accuracer
      or (IsVarcharFieldType(FBaseFieldType) and (FEngineVersion >= (2.10 - 0.001)))
{$ENDIF}
      ) then
   begin
    // save blob params
    SaveDataToStream(FBLOBCompressionAlgorithm,Sizeof(FBLOBCompressionAlgorithm),Stream,10175);
    SaveDataToStream(FBLOBCompressionMode,Sizeof(FBLOBCompressionMode),Stream,10176);
    SaveDataToStream(FBLOBBlockSize,Sizeof(FBLOBBlockSize),Stream,10177);
   end;

  //if (FDefaultValueType = dvtSequence) then
  if IsAutoincFieldType(FAdvancedFieldType) then
   begin
    // sequence
    SaveDataToStream(FSequenceDefObjectId,Sizeof(FSequenceDefObjectId),Stream,10178);
    SaveDataToStream(FAutoincIncrement,SizeOf(FAutoincIncrement),Stream,10685);
    SaveDataToStream(FAutoincInitialValue,SizeOf(FAutoincInitialValue),Stream,10686);
    SaveDataToStream(FAutoincMinValue,SizeOf(FAutoincMinValue),Stream,10687);
    SaveDataToStream(FAutoincMaxValue,SizeOf(FAutoincMaxValue),Stream,10688);
    SaveDataToStream(FAutoincCycled,SizeOf(FAutoincCycled),Stream,10689);
   end
  else
    // save default value
    FDefaultValue.SaveToStream(Stream);

end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRIndexColumn
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRIndexColumn.LoadFromStream(Stream: TStream);
var Len: Byte;
begin
  LoadDataFromStream(Len,Sizeof(Len),Stream,10358);
  SetLength(FFieldName,Len);
  LoadDataFromStream(PChar(@FFieldName[1])^,Len,Stream,10359);
  LoadDataFromStream(FDescending,Sizeof(FDescending),Stream,10360);
  LoadDataFromStream(FCaseInsensitive,Sizeof(FCaseInsensitive),Stream,10361);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRIndexColumn.SaveToStream(Stream: TStream);
var Len: Byte;
begin
  Len := Length(FFieldName);
  SaveDataToStream(Len,Sizeof(Len),Stream,10354);
  SaveDataToStream(PChar(@FFieldName[1])^,Len,Stream,10355);
  SaveDataToStream(FDescending,Sizeof(FDescending),Stream,10356);
  SaveDataToStream(FCaseInsensitive,Sizeof(FCaseInsensitive),Stream,10357);
end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRIndexDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// get index column
//------------------------------------------------------------------------------
function TACRIndexDef.GetIndexColumn(Index: Integer): TACRIndexColumn;
begin
  Result := FIndexColumns[Index];
end;// GetIndexColumn


//------------------------------------------------------------------------------
// get column count
//------------------------------------------------------------------------------
function TACRIndexDef.GetColumnCount: Integer;
begin
  Result := Length(FIndexColumns);
end;// GetColumnCount


//------------------------------------------------------------------------------
// set column count
//------------------------------------------------------------------------------
procedure TACRIndexDef.SetColumnCount(Value: Integer);
var
  oldCount, i: Integer;
begin
  oldCount := Length(FIndexColumns);
  if (Value > oldCount) then
   begin
     SetLength(FIndexColumns, Value);
     for i := oldCount to Value-1 do
      FIndexColumns[i] := TACRIndexColumn.Create;
   end
  else
  if (Value < oldCount) then
   begin
     for i := Value to oldCount-1 do
      FIndexColumns[i].Free;
     SetLength(FIndexColumns, Value);
   end;
end;// SetColumnCount




//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRIndexDef.Create;
begin
  FUnique := False;
  FPrimary := False;
  FIndexType := itBTree;
  FIndexColumns := nil;
  FRootPageNo := INVALID_PAGE_NO;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRIndexDef.Destroy;
begin
  ColumnCount := 0;
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// assign another IndexDef
//------------------------------------------------------------------------------
procedure TACRIndexDef.Assign(Source: TACRMetaObjectDef);
var
  i: Integer;
begin
  inherited Assign(Source);
  IndexType := TACRIndexDef(Source).IndexType;
  Unique := TACRIndexDef(Source).Unique;
  Primary := TACRIndexDef(Source).Primary;
  FRootPageNo := TACRIndexDef(Source).RootPageNo;
  ColumnCount := TACRIndexDef(Source).ColumnCount;
  FTemporary := TACRIndexDef(Source).FTemporary;
  for i := 0 to ColumnCount-1 do
   begin
    Columns[i].FieldName := TACRIndexDef(Source).Columns[i].FieldName;
    Columns[i].Descending := TACRIndexDef(Source).Columns[i].Descending;
    Columns[i].CaseInsensitive := TACRIndexDef(Source).Columns[i].CaseInsensitive;
   end;
end;// Assign


//------------------------------------------------------------------------------
// assign by names
//------------------------------------------------------------------------------
procedure TACRIndexDef.AssignByNames(FieldNames, AscDescList, CaseSensitivityList: TStringList);
var i: integer;
begin
 if (FieldNames.Count <> AscDescList.Count) then
  raise EACRException.Create(10278,ErrorLDifferentListsLength,
    [FieldNames.Count,AscDescList.Count]);
 if (FieldNames.Count <> CaseSensitivityList.Count) then
  raise EACRException.Create(10279,ErrorLDifferentListsLength,
    [FieldNames.Count,CaseSensitivityList.Count]);
  ColumnCount := FieldNames.Count;
  for i := 0 to ColumnCount-1 do
   begin
    Columns[i].FieldName := FieldNames[i];
    Columns[i].Descending := (AscDescList[i] = ACR_DESC);
    Columns[i].CaseInsensitive := (CaseSensitivityList[i] = ACR_NO_CASE);
   end;
end;


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRIndexDef.LoadFromStream(Stream: TStream);
var i: Integer;
begin
  inherited LoadFromStream(Stream);

  LoadDataFromStream(FIndexType,Sizeof(FIndexType),Stream,10349);
  LoadDataFromStream(FUnique,Sizeof(FUnique),Stream,10350);
  LoadDataFromStream(FPrimary,Sizeof(FPrimary),Stream,10351);
  FTemporary := False;
  LoadDataFromStream(i,Sizeof(i),Stream,10352);
  ColumnCount := i;
  LoadDataFromStream(FRootPageNo,Sizeof(FRootPageNo),Stream,10353);
  for i := 0 to ColumnCount-1 do
    Columns[i].LoadFromStream(Stream);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRIndexDef.SaveToStream(Stream: TStream);
var i: Integer;
begin
  inherited SaveToStream(Stream);
  SaveDataToStream(FIndexType,Sizeof(FIndexType),Stream,10344);
  SaveDataToStream(FUnique,Sizeof(FUnique),Stream,10345);
  SaveDataToStream(FPrimary,Sizeof(FPrimary),Stream,10346);
  i := ColumnCount;
  SaveDataToStream(i,Sizeof(i),Stream,10347);
  SaveDataToStream(FRootPageNo,Sizeof(FRootPageNo),Stream,10348);
  for i := 0 to ColumnCount-1 do
    Columns[i].SaveToStream(Stream);
end; // SaveToStream


//------------------------------------------------------------------------------
// return -1 if field was not found, otherwise return column index
//------------------------------------------------------------------------------
function TACRIndexDef.FindField(FieldName: String): Integer;
var s: String;
    i: Integer;
begin
  s := AnsiUpperCase(FieldName);
  Result := -1;
  for i := 0 to ColumnCount-1 do
    if (AnsiUpperCase(Columns[i].FieldName) = s) then
      begin
        Result := i;
        break;
      end;
end; // FindField


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRConstraintDef.Assign(Source: TACRMetaObjectDef);
begin
  inherited Assign(Source);
  FConstraintType := TACRConstraintDef(Source).FConstraintType;
end;//Assign



////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefNotNull
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRConstraintDefNotNull.Create;
begin
  FConstraintType := ctNotNull;
end;//Create
	
 
//------------------------------------------------------------------------------ 
// Assign
//------------------------------------------------------------------------------
procedure TACRConstraintDefNotNull.Assign(Source: TACRMetaObjectDef); 
begin 
  inherited Assign(Source);
  FTableSchemaName := TACRConstraintDefNotNull(Source).FTableSchemaName;
  FTableSchemaObjectID := TACRConstraintDefNotNull(Source).FTableSchemaObjectID; 
  FTableName := TACRConstraintDefNotNull(Source).FTableName;
  FTableObjectID := TACRConstraintDefNotNull(Source).FTableObjectID; 
  FColumnName := TACRConstraintDefNotNull(Source).FColumnName;
  FColumnObjectID := TACRConstraintDefNotNull(Source).FColumnObjectID; 
end;//Assign
 
//------------------------------------------------------------------------------
// SetNames 
//------------------------------------------------------------------------------ 
procedure TACRConstraintDefNotNull.SetNames(ColumnName, TableName, SchemaName: String);
begin 
  FColumnName := ColumnName; 
  FTableName := TableName; 
  FSchemaName := SchemaName; 
end;//SetNames 
	
	
	
//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------ 
procedure TACRConstraintDefNotNull.LoadFromStream(Stream: TStream);
var Len: Byte; 
begin
  inherited LoadFromStream(Stream); 
  LoadDataFromStream(FTableSchemaObjectID,sizeof(FTableSchemaObjectID),Stream,10249); 
  LoadDataFromStream(FTableObjectID,sizeof(FTableObjectID),Stream,10250);
  LoadDataFromStream(FColumnObjectID,sizeof(FColumnObjectID),Stream,10251); 
	
  LoadDataFromStream(Len,sizeof(Len),Stream,10252); 
  SetLength(FTableSchemaName,Len-1); 
  LoadDataFromStream(PChar(@FTableSchemaName)^,Len,Stream,10253);
	
  LoadDataFromStream(Len,sizeof(Len),Stream,10254); 
  SetLength(FTableName,Len-1); 
  LoadDataFromStream(PChar(@FTableName)^,Len,Stream,10255); 
 
  LoadDataFromStream(Len,sizeof(Len),Stream,10256); 
  SetLength(FColumnName,Len-1); 
  LoadDataFromStream(PChar(@FColumnName)^,Len,Stream,10257); 
end; // LoadFromStream
	
 
//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------ 
procedure TACRConstraintDefNotNull.SaveToStream(Stream: TStream);
var Len: Byte;
begin 
  inherited SaveToStream(Stream); 
 
  SaveDataToStream(FTableSchemaObjectID,sizeof(FTableSchemaObjectID),Stream,10240);
  SaveDataToStream(FTableObjectID,sizeof(FTableObjectID),Stream,10241);
  SaveDataToStream(FColumnObjectID,sizeof(FColumnObjectID),Stream,10242); 
	
  Len := Length(FTableSchemaName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,10243); 
  SaveDataToStream(PChar(@FTableSchemaName)^,Len,Stream,10244); 
	
  Len := Length(FTableName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,10245);
  SaveDataToStream(PChar(@FTableName)^,Len,Stream,10246); 
 
  Len := Length(FColumnName)+1; 
  SaveDataToStream(Len,Sizeof(Len),Stream,10247); 
  SaveDataToStream(PChar(@FColumnName)^,Len,Stream,10248);
end; // SaveToStream 
 
 
////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefCheck
// 
////////////////////////////////////////////////////////////////////////////////
	
	
//------------------------------------------------------------------------------ 
// Constructor 
//------------------------------------------------------------------------------
constructor TACRConstraintDefCheck.Create;
begin
  FConstraintType := ctCheck;
  FMinValue := TACRVariant.Create;
  FMaxValue := TACRVariant.Create;
end;//Create 
	
	
//------------------------------------------------------------------------------ 
// Destructor
//------------------------------------------------------------------------------
destructor TACRConstraintDefCheck.Destroy; 
begin 
  FMinValue.Free; 
  FMaxValue.Free; 
  inherited;
end;//Destroy 
 
	
//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRConstraintDefCheck.Assign(Source: TACRMetaObjectDef); 
begin 
  inherited Assign(Source);
  FTableSchemaName := TACRConstraintDefCheck(Source).FTableSchemaName;
  FTableSchemaObjectID := TACRConstraintDefCheck(Source).FTableSchemaObjectID; 
  FTableName := TACRConstraintDefCheck(Source).FTableName; 
  FTableObjectID := TACRConstraintDefCheck(Source).FTableObjectID; 
  FColumnName := TACRConstraintDefCheck(Source).FColumnName;
  FColumnObjectID := TACRConstraintDefCheck(Source).FColumnObjectID;
  FMinValue.Assign(TACRConstraintDefCheck(Source).FMinValue);
  FMaxValue.Assign(TACRConstraintDefCheck(Source).FMaxValue);
end;//Assign
 
	
//------------------------------------------------------------------------------ 
// SetNames
//------------------------------------------------------------------------------
procedure TACRConstraintDefCheck.SetNames(ColumnName, TableName,
  SchemaName: String);
begin
  FColumnName := ColumnName;
  FTableName := TableName;
  FSchemaName := SchemaName;
end;//SetNames


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefCheck.LoadFromStream(Stream: TStream);
var Len: Byte;
begin
  inherited LoadFromStream(Stream);
  LoadDataFromStream(FTableSchemaObjectID,sizeof(FTableSchemaObjectID),Stream,10230);
  LoadDataFromStream(FTableObjectID,sizeof(FTableObjectID),Stream,10231);
  LoadDataFromStream(FColumnObjectID,sizeof(FColumnObjectID),Stream,10232);

  LoadDataFromStream(Len,sizeof(Len),Stream,10233);
  SetLength(FTableSchemaName,Len-1);
  LoadDataFromStream(PChar(@FTableSchemaName)^,Len,Stream,10234);

  LoadDataFromStream(Len,sizeof(Len),Stream,10235);
  SetLength(FTableName,Len-1);
  LoadDataFromStream(PChar(@FTableName)^,Len,Stream,10236);

  LoadDataFromStream(Len,sizeof(Len),Stream,10237);
  SetLength(FColumnName,Len-1);
  LoadDataFromStream(PChar(@FColumnName)^,Len,Stream,10238);

  FMinValue.LoadFromStream(Stream);
  FMaxValue.LoadFromStream(Stream);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefCheck.SaveToStream(Stream: TStream);
var Len: Byte;
begin
  inherited SaveToStream(Stream);

  SaveDataToStream(FTableSchemaObjectID,sizeof(FTableSchemaObjectID),Stream,10227);
  SaveDataToStream(FTableObjectID,sizeof(FTableObjectID),Stream,10228);
  SaveDataToStream(FColumnObjectID,sizeof(FColumnObjectID),Stream,10229);

  Len := Length(FTableSchemaName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,10221);
  SaveDataToStream(PChar(@FTableSchemaName)^,Len,Stream,10222);

  Len := Length(FTableName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,10223);
  SaveDataToStream(PChar(@FTableName)^,Len,Stream,10224);

  Len := Length(FColumnName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,10225);
  SaveDataToStream(PChar(@FColumnName)^,Len,Stream,10226);

   FMinValue.SaveToStream(Stream);
   FMaxValue.SaveToStream(Stream);
end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefUnique
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRConstraintDefUnique.Create;
begin
  inherited;
  FConstraintType := ctUnique;
  SetLength(Columns,0);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRConstraintDefUnique.Destroy;
begin
  SetLength(Columns,0);
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRConstraintDefUnique.Assign(Source: TACRMetaObjectDef);
var i,L: Integer;
begin
  inherited Assign(Source);
  L := Length(TACRConstraintDefUnique(Source).Columns);
  SetLength(Columns, L);
  for i:=0 to L-1 do
    Columns[i] := TACRConstraintDefUnique(Source).Columns[i];

  FTableSchemaName := TACRConstraintDefUnique(Source).FTableSchemaName;
  FTableSchemaObjectID := TACRConstraintDefUnique(Source).FTableSchemaObjectID;
  FTableName := TACRConstraintDefUnique(Source).FTableName;
  FTableObjectID := TACRConstraintDefUnique(Source).FTableObjectID;
  FIndexName := TACRConstraintDefUnique(Source).FIndexName;
  FIndexObjectID := TACRConstraintDefUnique(Source).FIndexObjectID;

end;//Assign


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefUnique.LoadFromStream(Stream: TStream);
var i,j, Len: Integer;
begin
  inherited LoadFromStream(Stream);

  LoadDataFromStream(FTableSchemaObjectID, sizeof(FTableSchemaObjectID), Stream, 30303);
  LoadDataFromStream(FTableObjectID,  sizeof(FTableObjectID),  Stream, 30304);
  LoadDataFromStream(FIndexObjectID,  sizeof(FIndexObjectID),  Stream, 30317);

  LoadDataFromStream(Len,sizeof(Len),Stream,30305);
  SetLength(FTableSchemaName,Len-1);
  LoadDataFromStream(PChar(@FTableSchemaName)^,Len,Stream,30306);

  LoadDataFromStream(Len,sizeof(Len),Stream,30307);
  SetLength(FTableName,Len-1);
  LoadDataFromStream(PChar(@FTableName)^,Len,Stream,30308);

  LoadDataFromStream(Len,sizeof(Len),Stream,30321);
  SetLength(FIndexName,Len-1);
  LoadDataFromStream(PChar(@FIndexName)^,Len,Stream,30322);

  LoadDataFromStream(j,sizeof(j),Stream,30309);
  SetLength(Columns,j);
  for i:=0 to j-1 do
    begin
      LoadDataFromStream(Columns[i].ColumnObjectID,sizeof(Columns[i].ColumnObjectID),Stream,30310);
      LoadDataFromStream(Len,sizeof(Len),Stream,30311);
      SetLength(Columns[i].ColumnName, Len);
      LoadDataFromStream(PChar(@Columns[i].ColumnName)^,Len,Stream,30312);
    end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefUnique.SaveToStream(Stream: TStream);
var i,j, Len: Integer;
begin
  inherited SaveToStream(Stream);

  SaveDataToStream(FTableSchemaObjectID, sizeof(FTableSchemaObjectID), Stream, 30313);
  SaveDataToStream(FTableObjectID,  sizeof(FTableObjectID),  Stream, 30314);
  SaveDataToStream(FIndexObjectID,  sizeof(FIndexObjectID),  Stream, 30318);

  Len := Length(FTableSchemaName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,30315);
  SaveDataToStream(PChar(@FTableSchemaName)^,Len,Stream,30316);

  Len := Length(FTableName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,30317);
  SaveDataToStream(PChar(@FTableName)^,Len,Stream,30318);

  Len := Length(FIndexName)+1;
  SaveDataToStream(Len,Sizeof(Len),Stream,30323);
  SaveDataToStream(PChar(@FIndexName)^,Len,Stream,30324);

  j := Length(Columns);
  SaveDataToStream(j,sizeof(j),Stream,30319);
  for i:=0 to j-1 do
    begin
      SaveDataToStream(Columns[i].ColumnObjectID,sizeof(Columns[i].ColumnObjectID),Stream,30320);
      Len := Length(Columns[i].ColumnName)+1;
      SaveDataToStream(Len,sizeof(Len),Stream,30321);
      SaveDataToStream(PChar(@Columns[i].ColumnName)^,Len,Stream,30322);
    end;
end;//SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRMetaObjectDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRConstraintDefPrimary.Create;
begin
  inherited;
  FConstraintType := ctPK;
end;//Create



////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefForeignKeyAction
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRConstraintDefForeignKeyAction.Create;
begin
  inherited;
  FConstraintType := ctFKAction;
  FMatchType := cfkmtDefault;
  FDeleteAction := cfkaDefault;
  FUpdateAction := cfkaDefault;
  SetLength(Columns,0);
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRConstraintDefForeignKeyAction.Destroy;
begin
  SetLength(Columns,0);
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRConstraintDefForeignKeyAction.Assign(Source: TACRMetaObjectDef);
var i,L: Integer;
begin
  inherited Assign(Source);
  L := Length(TACRConstraintDefForeignKeyAction(Source).Columns);
  SetLength(Columns, L);
  for i:=0 to L-1 do
    Columns[i] := TACRConstraintDefForeignKeyAction(Source).Columns[i];
  FReferencedTableName := TACRConstraintDefForeignKeyAction(Source).FReferencedTableName;
  FReferencedTableObjectID := TACRConstraintDefForeignKeyAction(Source).FReferencedTableObjectID;
  FReferencedFKName := TACRConstraintDefForeignKeyAction(Source).FReferencedFKName;
  FReferencedFKObjectID := TACRConstraintDefForeignKeyAction(Source).FReferencedFKObjectID;
  FDeleteAction := TACRConstraintDefForeignKeyAction(Source).DeleteAction;
  FUpdateAction := TACRConstraintDefForeignKeyAction(Source).UpdateAction;
  FMatchType := TACRConstraintDefForeignKeyAction(Source).MatchType;
end;//Assign


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefForeignKeyAction.LoadFromStream(Stream: TStream);
var i,j, Len: Integer;
begin
  inherited LoadFromStream(Stream);

  LoadDataFromStream(Len,sizeof(Len),Stream,11398);
  SetLength(FReferencedTableName,Len);
  LoadDataFromStream(PChar(@FReferencedTableName[1])^,Len,Stream,11399);
  LoadDataFromStream(FReferencedTableObjectID,Sizeof(FReferencedTableObjectID),Stream,11400);
  LoadDataFromStream(Len,sizeof(Len),Stream,11439);
  SetLength(FReferencedFKName,Len);
  LoadDataFromStream(PChar(@FReferencedFKName[1])^,Len,Stream,11440);
  LoadDataFromStream(FReferencedFKObjectID,Sizeof(FReferencedFKObjectID),Stream,11441);
  LoadDataFromStream(FDeleteAction,Sizeof(FDeleteAction),Stream,11402);
  LoadDataFromStream(FUpdateAction,Sizeof(FUpdateAction),Stream,11403);
  LoadDataFromStream(FMatchType,Sizeof(FMatchType),Stream,11401);

  LoadDataFromStream(j,sizeof(j),Stream,11404);
  SetLength(Columns,j);
  for i := 0 to j-1 do
    begin
      LoadDataFromStream(Columns[i].ColumnObjectID,sizeof(Columns[i].ColumnObjectID),Stream,11405);
      LoadDataFromStream(Len,sizeof(Len),Stream,11406);
      SetLength(Columns[i].ColumnName, Len);
      LoadDataFromStream(PChar(@Columns[i].ColumnName[1])^,Len,Stream,11407);
    end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefForeignKeyAction.SaveToStream(Stream: TStream);
var i,j, Len: Integer;
begin
  inherited SaveToStream(Stream);
  Len := Length(FReferencedTableName);
  SaveDataToStream(Len,Sizeof(Len),Stream,11408);
  SaveDataToStream(PChar(@FReferencedTableName[1])^,Len,Stream,11409);
  SaveDataToStream(FReferencedTableObjectID,SizeOf(FReferencedTableObjectID),Stream,11410);
  Len := Length(FReferencedFKName);
  SaveDataToStream(Len,Sizeof(Len),Stream,11436);
  SaveDataToStream(PChar(@FReferencedFKName[1])^,Len,Stream,11437);
  SaveDataToStream(FReferencedFKObjectID,SizeOf(FReferencedFKObjectID),Stream,11438);
  SaveDataToStream(FDeleteAction,Sizeof(FDeleteAction),Stream,11412);
  SaveDataToStream(FUpdateAction,Sizeof(FUpdateAction),Stream,11413);
  SaveDataToStream(FMatchType,Sizeof(FMatchType),Stream,11411);
  j := Length(Columns);
  SaveDataToStream(j,sizeof(j),Stream,11414);
  for i:=0 to j-1 do
    begin
      SaveDataToStream(Columns[i].ColumnObjectID,SizeOf(Columns[i].ColumnObjectID),Stream,11415);
      Len := Length(Columns[i].ColumnName);
      SaveDataToStream(Len,sizeof(Len),Stream,11416);
      SaveDataToStream(PChar(@Columns[i].ColumnName[1])^,Len,Stream,11417);
    end;
end;//SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefForeignKey
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRConstraintDefForeignKey.Create;
begin
  inherited;
  FConstraintType := ctFK;
end;//Create


////////////////////////////////////////////////////////////////////////////////
//
// TACRMetaObjectDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.LoadFromStream(Stream: TStream);
begin
 Clear;
 LoadDataFromStream(FLoadedItemCount,sizeof(FLoadedItemCount),Stream,10159);
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.SaveToStream(Stream: TStream);
var ItemCount:  Integer;
begin
 ItemCount := Self.Count;
 SaveDataToStream(ItemCount,sizeof(ItemCount),Stream,10158);
end; // SaveToStream


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRMetaObjectDefs.Create;
begin
  FDefsList := TACRSortedStringPtrArray.Create;
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRMetaObjectDefs.Destroy;
begin
  Clear;
  FDefsList.Free;
end;//Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.Assign(Source: TACRMetaObjectDefs);
var i,k: Integer;
begin
  Clear;
  for i:=0 to Source.Count-1 do
   begin
    k := GetDefNumberByName(Source[i].Name);
    if (k >= 0) then
      raise EACRException.Create(10424,ErrorLDuplicateFieldName,
        [Source[i].Name]);
    InternalAddCreated.Assign(Source.Items[i]);
   end;
end;//Assign


//------------------------------------------------------------------------------
// GetCount
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.GetCount: Integer;
begin
  Result := FDefsList.Count;
end;//GetCount



//------------------------------------------------------------------------------
// GetDef
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.GetDef(Index: Integer): TACRMetaObjectDef;
begin
  Result := TACRMetaObjectDef(FDefsList[Index]);
end;//GetDef


//------------------------------------------------------------------------------
// SetDef
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.SetDef(Index: Integer;
  Value: TACRMetaObjectDef);
begin
  TACRMetaObjectDef(FDefsList[Index]).Free;
  FDefsList[Index] := Value;
end;//SetDef


//------------------------------------------------------------------------------
// GetDefNumberByName
//    ( if name not found, then Result = -1 )
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.GetDefNumberByName(Name: String): Integer;
var i: Integer;
begin
  Result := -1;
  for i:=0 to FDefsList.Count -1 do
    if (AnsiUpperCase(TACRMetaObjectDef(FDefsList[i]).Name) = AnsiUpperCase(Name)) then
      begin
        Result := i;
        break;
      end;
end;//GetDefNumberByName


//------------------------------------------------------------------------------
// GetDefByName
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.GetDefByName(Name: String): TACRMetaObjectDef;
var i: Integer;
begin
  i := GetDefNumberByName(Name);
  if i = -1 then
    Result := nil
  else
    Result := TACRMetaObjectDef(Items[i]);
end;//GetDefByName


//------------------------------------------------------------------------------
// GetDefNumberByObjectId
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.GetDefNumberByObjectId(id: TACRObjectID): Integer;
var i: Integer;
begin
  Result := -1;
  for i:=0 to FDefsList.Count -1 do
    if (TACRMetaObjectDef(FDefsList[i]).ObjectID = id) then
      begin
        Result := i;
        break;
      end;
end;//GetDefNumberByObjectId


//------------------------------------------------------------------------------
// GetDefByByObjectId
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.GetDefByObjectId(id: TACRObjectID): TACRMetaObjectDef;
var i: Integer;
begin
  i := GetDefNumberByObjectId(id);
  if i <> -1 then
    Result := Items[i]
  else
    Result := nil;
end;//GetDefByByObjectId


//------------------------------------------------------------------------------
// Add
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.Add(MetaObjectDef: TACRMetaObjectDef);
begin
  FDefsList.Add(MetaObjectDef);
end;//Add


//------------------------------------------------------------------------------
// internal add created
//------------------------------------------------------------------------------
function TACRMetaObjectDefs.InternalAddCreated: TACRMetaObjectDef;
begin
  raise EACRException.Create(30339, ErrorGMethodNotOverrided,
                                            ['InternalAddCreated', classname]);
end; // InternalAddCreated


//------------------------------------------------------------------------------
// Delete
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.Delete(Index: Integer);
begin
 if FDefsList[Index] <> nil then
    TACRMetaObjectDef(FDefsList[Index]).Free;
 FDefsList.Delete(Index);
end;//Delete


//------------------------------------------------------------------------------
// Insert
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.Insert(Index: Integer; MetaObjectDef: TACRMetaObjectDef);
begin
  FDefsList.Insert(Index, MetaObjectDef);
end;//Insert


//------------------------------------------------------------------------------
// Clear
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.Clear;
begin
  while (Count > 0) do
    Delete(0);
  FDefsList.Clear;
end;//Clear



//------------------------------------------------------------------------------
// Move
//------------------------------------------------------------------------------
procedure TACRMetaObjectDefs.Move(CurIndex, NewIndex: Integer);
begin
  FDefsList.Move(CurIndex, NewIndex);
end;//Move




////////////////////////////////////////////////////////////////////////////////
//
// TACRIndexDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// get index def
//------------------------------------------------------------------------------
function TACRIndexDefs.GetIndexDef(Index: Integer): TACRIndexDef;
begin
  Result := TACRIndexDef(GetDef(Index));
end;// GetIndexDef


//------------------------------------------------------------------------------
// set index def
//------------------------------------------------------------------------------
procedure TACRIndexDefs.SetIndexDef(Index: Integer; Value: TACRIndexDef);
begin
  SetDef(Index, Value);
end;// SetIndexDef


//------------------------------------------------------------------------------
// InternalAddCreated
//------------------------------------------------------------------------------
function TACRIndexDefs.InternalAddCreated: TACRMetaObjectDef;
begin
  Result := TACRIndexDef.Create;
  Add(Result);
end;//InternalAddCreated


//------------------------------------------------------------------------------
// AscDesc and CaseSensitivity lists should contain constants ACR_ASC, ACR_DESC, ACR_NO_CASE, ACR_CASE
//------------------------------------------------------------------------------
function TACRIndexDefs.IsIndexExists(FieldNames, AscDescList, CaseSensitivityList: TStringList): Boolean;
begin
  Result := not (FindIndex(FieldNames,AscDescList,CaseSensitivityList) = INVALID_OBJECT_ID);
end; // IsIndexExists


//------------------------------------------------------------------------------
// find index
//------------------------------------------------------------------------------
function TACRIndexDefs.FindIndex(FieldNames, AscDescList, CaseSensitivityList: TStringList): TACRObjectID;
var i,j,n:  Integer;
    bOK:    Boolean;
begin
 Result := INVALID_OBJECT_ID;
 if (FieldNames = nil) then
  raise EACRException.Create(10389,ErrorLNilPointer);
 if (AscDescList = nil) then
  raise EACRException.Create(10390,ErrorLNilPointer);
 if (CaseSensitivityList = nil) then
  raise EACRException.Create(10391,ErrorLNilPointer);
 if (FieldNames.Count <> AscDescList.Count) then
  raise EACRException.Create(10276,ErrorLDifferentListsLength,
    [FieldNames.Count,AscDescList.Count]);
 if (FieldNames.Count <> CaseSensitivityList.Count) then
  raise EACRException.Create(10277,ErrorLDifferentListsLength,
    [FieldNames.Count,CaseSensitivityList.Count]);
 n := 0;   
 for i := 0 to Self.Count - 1 do
  begin
   if (Items[i].GetColumnCount < FieldNames.Count) then
    continue
   else
    n := FieldNames.Count-1;
   bOK := True;
   for j := 0 to n do
    if (AnsiUpperCase(FieldNames[j]) = AnsiUpperCase(Items[i].Columns[j].FieldName)) then
     begin
     if (
          ((AscDescList[j] = ACR_ASC) and (Items[i].Columns[j].FDescending)) or
          ((AscDescList[j] = ACR_DESC) and (not Items[i].Columns[j].FDescending)) or
          ((CaseSensitivityList[j] = ACR_CASE) and (Items[i].Columns[j].FCaseInsensitive)) or
          ((CaseSensitivityList[j] = ACR_NO_CASE) and (not Items[i].Columns[j].FCaseInsensitive))
         ) then
      begin
        bOK := False;
        break;
      end;
     end
    else
     begin
      bOK := False;
      break;
     end;
   if (bOK) then
    begin
     Result := Items[i].ObjectID;
     break;
    end;
  end;
end; // FindIndex


//------------------------------------------------------------------------------
// primary index
//------------------------------------------------------------------------------
function TACRIndexDefs.FindPrimaryIndex: TACRIndexDef;
var i: Integer;
begin
  Result := nil;
  for i := 0 to Self.Count-1 do
   if (GetIndexDef(i).Primary) then
    begin
     Result := GetIndexDef(i);
     break;
    end;
end; // FindPrimaryIndex


//------------------------------------------------------------------------------
// AddCreated
//------------------------------------------------------------------------------
function TACRIndexDefs.AddCreated: TACRIndexDef;
begin
  Result := TACRIndexDef(InternalAddCreated);
end;//AddCreated


//------------------------------------------------------------------------------
// GetIndexDefByIndexName
//------------------------------------------------------------------------------
function TACRIndexDefs.GetIndexDefByName(Name: String): TACRIndexDef;
var
  i: Integer;
begin
  i := GetDefNumberByName(Name);
  if i = -1 then
    Result := nil
  else
    Result := TACRIndexDef(Items[i]);
end;//GetIndexDefByIndexName


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRIndexDefs.LoadFromStream(Stream: TStream);
var IndexDef: TACRIndexDef;
    i:        Integer;
begin
 inherited LoadFromStream(Stream);
 for i := 0 to FLoadedItemCount-1 do
  begin
   IndexDef := TACRIndexDef.Create;
   IndexDef.LoadFromStream(Stream);
   Add(IndexDef);
  end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRIndexDefs.SaveToStream(Stream: TStream);
var i:    Integer;
    qty:  Integer;
begin
 qty := 0;
 for i := 0 to Count-1 do
  if (not Items[i].FTemporary) then
   Inc(qty);
 SaveDataToStream(qty,sizeof(qty),Stream,10343);
 for i := 0 to Count-1 do
  if (not Items[i].FTemporary) then
   TACRIndexDef(Items[i]).SaveToStream(Stream);
end; // SaveToStream




////////////////////////////////////////////////////////////////////////////////
//
// TACRFieldDefs
//
////////////////////////////////////////////////////////////////////////////////



//------------------------------------------------------------------------------
// GetFieldDef
//------------------------------------------------------------------------------
function TACRFieldDefs.GetDef(Index: Integer): TACRFieldDef;
begin
  Result := TACRFieldDef(FDefsList[Index]);
end;//GetFieldDef


//------------------------------------------------------------------------------
// SetFieldDef
//------------------------------------------------------------------------------
procedure TACRFieldDefs.SetDef(Index: Integer; Value: TACRFieldDef);
begin
  TACRFieldDef(FDefsList[Index]).Free;
  FDefsList[Index] := Value;
end;//SetFieldDef


//------------------------------------------------------------------------------
// InternalAddCreated
//------------------------------------------------------------------------------
function TACRFieldDefs.InternalAddCreated: TACRMetaObjectDef;
begin
  Result := TACRFieldDef.Create;
  Add(Result);
end;//InternalAddCreated


//------------------------------------------------------------------------------
// AddCreated
//------------------------------------------------------------------------------
function TACRFieldDefs.AddCreated: TACRFieldDef;
begin
  Result := TACRFieldDef(InternalAddCreated);
end;//AddCreated


//------------------------------------------------------------------------------
// Recalc FieldOffsets in children fields
//------------------------------------------------------------------------------
procedure TACRFieldDefs.RecalcFieldOffsets;
var i: integer;
    CurMemOffset, CurDiskOffset: Integer;
    FieldDef: TACRFieldDef;
begin
  // Init Offsets
  FVarcharOrBLOBFieldsExists := False;
  FAutoIncFieldsExists := False;
  FDefaultValuesExists := False;
  CurMemOffset := Count div 8;
  if (Count mod 8 > 0) then Inc(CurMemOffset);
  CurDiskOffset := Count div 8;
  if (Count mod 8 > 0) then Inc(CurDiskOffset);
  // Fill offsets
  for i:=0 to FDefsList.Count-1 do
   begin
     FieldDef := FDefsList[i];
     if (not FVarcharOrBLOBFieldsExists) then
       if ((IsBLOBFieldType(FieldDef.BaseFieldType)) or
           (IsVarcharFieldType(FieldDef.BaseFieldType))) then
        FVarcharOrBLOBFieldsExists := True;
     if (not FAutoIncFieldsExists) then
       if (IsAutoincFieldType(FieldDef.AdvancedFieldType)) then
        FAutoIncFieldsExists := True;
     if (not FDefaultValuesExists) then
       if (not FieldDef.DefaultValue.IsNull) then
        FDefaultValuesExists := True;
     FieldDef.FMemoryOffset := CurMemOffset;
     FieldDef.FDiskOffset := CurDiskOffset;
     CurMemOffset := CurMemOffset + FieldDef.FMemoryDataSize;
     CurDiskOffset := CurDiskOffset + FieldDef.FDiskDataSize;
   end;
end;//RecalcFieldOffsets


//------------------------------------------------------------------------------
// GetMemoryRecordBufferSize
//------------------------------------------------------------------------------
function TACRFieldDefs.GetMemoryRecordBufferSize: Integer;
var i: integer;
begin
  Result := 0;
  // 1 bit for each field ( NULL flags )
  Result := Result + Integer(Count) div 8;
  if (Count mod 8 > 0) then Inc(Result);
  for i:=0 to Count-1 do
    Result := Result + Items[i].FMemoryDataSize;
end;//GetMemoryRecordBufferSize


//------------------------------------------------------------------------------
// GetFieldDefByFieldName
//------------------------------------------------------------------------------
function TACRFieldDefs.GetFieldDefByName(Name: String): TACRFieldDef;
begin
  Result := TACRFieldDef(GetDefByName(Name));
end;//GetFieldDefByFieldName


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRFieldDefs.LoadFromStream(Stream: TStream);
var
    FieldDef: TACRFieldDef;
    i:        Integer;
begin
  inherited LoadFromStream(Stream);
 for i := 0 to FLoadedItemCount-1 do
   begin
   FieldDef := TACRFieldDef.Create;
   FieldDef.EngineVersion := FEngineVersion;
   FieldDef.LoadFromStream(Stream);
   Add(FieldDef);
   end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRFieldDefs.SaveToStream(Stream: TStream);
var i: Integer;
begin
 inherited SaveToStream(Stream);
 for i := 0 to Count-1 do
  begin
   TACRFieldDef(Items[i]).EngineVersion := FEngineVersion;
   TACRFieldDef(Items[i]).SaveToStream(Stream);
  end;
end; // SaveToStream



////////////////////////////////////////////////////////////////////////////////
//
// TACRConstraintDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// AddNotNull
//------------------------------------------------------------------------------
function TACRConstraintDefs.AddNotNull: TACRConstraintDefNotNull;
begin
  Result := TACRConstraintDefNotNull.Create;
  Add(Result);
end;//AddNotNull


//------------------------------------------------------------------------------
// AddCheck
//------------------------------------------------------------------------------
function TACRConstraintDefs.AddCheck: TACRConstraintDefCheck;
begin
  Result := TACRConstraintDefCheck.Create;
  Add(Result);
end;//AddCheck


//------------------------------------------------------------------------------
// Add PK
//------------------------------------------------------------------------------
function TACRConstraintDefs.AddPK: TACRConstraintDefPrimary;
begin
  Result := TACRConstraintDefPrimary.Create;
  Add(Result);
end;//AddPK


//------------------------------------------------------------------------------
// Add Unique
//------------------------------------------------------------------------------
function TACRConstraintDefs.AddUnique: TACRConstraintDefUnique;
begin
  Result := TACRConstraintDefUnique.Create;
  Add(Result);
end;//AddUnique


//------------------------------------------------------------------------------
// create FK
//------------------------------------------------------------------------------
function TACRConstraintDefs.AddFK: TACRConstraintDefForeignKey;
begin
  Result := TACRConstraintDefForeignKey.Create;
  Add(Result);
  FForeignKeysExists := True;
end; // AddFK


//------------------------------------------------------------------------------
// create FKAction (master table action for FK in detail table)
//------------------------------------------------------------------------------
function TACRConstraintDefs.AddFKAction: TACRConstraintDefForeignKeyAction;
begin
  Result := TACRConstraintDefForeignKeyAction.Create;
  Add(Result);
  FForeignKeysActionsExists := True;
end; // AddFKAction


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRConstraintDefs.Create;
begin
 inherited;
 FForeignKeysExists := False;
 FForeignKeysActionsExists := False;
end; // Create


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRConstraintDefs.Assign(Source: TACRMetaObjectDefs);
var i: Integer;
begin
  Clear;
  FForeignKeysExists := False;
  FForeignKeysActionsExists := False;
  for i:=0 to Source.Count-1 do
   begin
    case TACRConstraintDef(Source.Items[i]).ConstraintType of
     ctNotNull:
       AddNotNull.Assign(TACRConstraintDefNotNull(Source.Items[i]));
     ctCheck:
       AddCheck.Assign(TACRConstraintDefCheck(Source.Items[i]));
     ctPK:
       AddPK.Assign(TACRConstraintDefPrimary(Source.Items[i]));
     ctUnique:
       AddUnique.Assign(TACRConstraintDefUnique(Source.Items[i]));
     ctFK:
       AddFK.Assign(TACRConstraintDefForeignKey(Source.Items[i]));
     ctFKAction:
       AddFKAction.Assign(TACRConstraintDefForeignKeyAction(Source.Items[i]));
     else
       raise EACRException.Create(30035, ErrorGNotImplementedYet);
    end;
   end;
end;//Assign


//------------------------------------------------------------------------------
// delete constraint
//------------------------------------------------------------------------------
procedure TACRConstraintDefs.Delete(Index: Integer);
var i: Integer;
begin
  inherited Delete(Index);
  FForeignKeysExists := False;
  FForeignKeysActionsExists := False;
  for i := 0 to GetCount-1 do
   if (TACRConstraintDef(FDefsList[i]).FConstraintType = ctFK) then
    FForeignKeysExists := True
   else
   if (TACRConstraintDef(FDefsList[i]).FConstraintType = ctFKAction) then
    FForeignKeysActionsExists := True;
end; // Delete


//------------------------------------------------------------------------------
// GetDef
//------------------------------------------------------------------------------
function TACRConstraintDefs.GetDef(Index: Integer): TACRConstraintDef;
begin
  Result := TACRConstraintDef(FDefsList[Index]);
end;//GetDef


//------------------------------------------------------------------------------
// SetDef
//------------------------------------------------------------------------------
procedure TACRConstraintDefs.SetDef(Index: Integer; Value: TACRConstraintDef);
begin
  if FDefsList[Index] <> nil then
    TACRConstraintDef(FDefsList[Index]).Free;
  FDefsList[Index] := Value;
end;//SetDef


//------------------------------------------------------------------------------
// return true if actions with update action <> NoAction exists
//------------------------------------------------------------------------------
function TACRConstraintDefs.GetForeignKeysActionsUpdateExists: Boolean;
var i: Integer;
begin
  Result := FForeignKeysActionsExists;
  if (Result) then
   begin
    Result := False;
    for i := 0 to Count-1 do
     if (TACRConstraintDef(FDefsList.Items[i]).ConstraintType = ctFKAction) then
      if (TACRConstraintDefForeignKeyAction(FDefsList.Items[i]).UpdateAction <>
            cfkaNoAction) then
        begin
         Result := True;
         break;
        end;
   end;
end; // GetForeignKeysActionsUpdateExists


//------------------------------------------------------------------------------
// return true if actions with delete action <> NoAction exists
//------------------------------------------------------------------------------
function TACRConstraintDefs.GetForeignKeysActionsDeleteExists: Boolean;
var i: Integer;
begin
  Result := FForeignKeysActionsExists;
  if (Result) then
   begin
    Result := False;
    for i := 0 to Count-1 do
     if (TACRConstraintDef(FDefsList.Items[i]).ConstraintType = ctFKAction) then
      if (TACRConstraintDefForeignKeyAction(FDefsList.Items[i]).DeleteAction <>
            cfkaNoAction) then
        begin
         Result := True;
         break;
        end;
   end;
end; // GetForeignKeysActionsDeleteExists


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefs.LoadFromStream(Stream: TStream);
var i:              Integer;
    ConstraintDef:  TACRConstraintDef;
    ConstraintType: TACRConstraintType;
begin
  FForeignKeysExists := False;
  FForeignKeysActionsExists := False;
  inherited LoadFromStream(Stream);
  for i := 0 to FLoadedItemCount - 1 do
    begin
      LoadDataFromStream(ConstraintType,sizeof(ConstraintType),Stream,10220);
      case ConstraintType of
        ctNotNull:
         ConstraintDef := AddNotNull;
        ctCheck:
         ConstraintDef := AddCheck;
        ctPK:
         ConstraintDef := AddPK;
        ctUnique:
         ConstraintDef := AddUnique;
        ctFK:
         ConstraintDef := AddFK;
        ctFKAction:
         ConstraintDef := AddFKAction;
        else
          raise EACRException.Create(30316, ErrorGUnknownConstrainType,
                                     [Integer(ConstraintType)]);
      end;
      ConstraintDef.LoadFromStream(Stream);
    end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRConstraintDefs.SaveToStream(Stream: TStream);
var i: Integer;
begin
  inherited SaveToStream(Stream);
  for i := 0 to Count-1 do
   begin
    SaveDataToStream(Items[i].FConstraintType,sizeof(Items[i].ConstraintType),
      Stream,10219);
    Items[i].SaveToStream(Stream);
   end;
end; // SaveToStream


////////////////////////////////////////////////////////////////////////////////
//
// TACRSequenceDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// GetDef
//------------------------------------------------------------------------------
function TACRSequenceDefs.GetDef(Index: Integer): TACRSequenceDef;
begin
  Result := TACRSequenceDef(FDefsList[Index]);
end;//GetDef


//------------------------------------------------------------------------------
// SetDef
//------------------------------------------------------------------------------
procedure TACRSequenceDefs.SetDef(Index: Integer;
  Value: TACRSequenceDef);
begin
  TACRSequenceDef(FDefsList[Index]).Free;
  FDefsList[Index] := Value;
end;//SetDef


//------------------------------------------------------------------------------
// InternalAddCreated
//------------------------------------------------------------------------------
function TACRSequenceDefs.InternalAddCreated: TACRMetaObjectDef;
begin
  Result := TACRSequenceDef.Create;
  Add(Result);
end;//InternalAddCreated


//------------------------------------------------------------------------------
// AddCreated
//------------------------------------------------------------------------------
function TACRSequenceDefs.AddCreated: TACRSequenceDef;
begin
  Result := TACRSequenceDef(InternalAddCreated);
end;//AddCreated


//------------------------------------------------------------------------------
// GetSequenceDefByName
//------------------------------------------------------------------------------
function TACRSequenceDefs.GetSequenceDefByName(Name: String): TACRSequenceDef;
var i: Integer;
begin
  i := GetDefNumberByName(Name);
  if i = -1 then
    Result := nil
  else
    Result := TACRSequenceDef(Items[i]);
end;//GetSequenceDefByName





////////////////////////////////////////////////////////////////////////////////
//
// TACRCursor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// set index name
//------------------------------------------------------------------------------
procedure TACRCursor.SetIndexName(Value: String);
var
  ACRIndexDef: TACRIndexDef;
begin
  FIndexName := Value;
  ACRIndexDef := FIndexDefs.GetIndexDefByName(Value);
  if (ACRIndexDef <> nil) then
   begin
     FIndexID := ACRIndexDef.ObjectID;
   end
  else
   FIndexID := INVALID_OBJECT_ID; 
end;// SetIndexName


//------------------------------------------------------------------------------
// Rename Field by Field Index in FieldDefs
//------------------------------------------------------------------------------
procedure TACRCursor.RenameField(FieldName, NewFieldName: String);
var
  fd: TACRFieldDef;
begin
  // Check Field Exists
  fd := FFieldDefs.GetFieldDefByName(FieldName);
  if (fd = nil) then
    raise EACRException.Create(30345, ErrorGFieldWithNameNotFound, [FieldName]);

  // Check For Duplicate FieldName
  if ( FFieldDefs.GetFieldDefByName(NewFieldName) <> nil ) then
   raise EACRException.Create(30346, ErrorGCannotRenameField,
                                     [FieldName, NewFieldName]);

  fd.Name := NewFieldName;
end;//RenameField


//------------------------------------------------------------------------------
// update table definitions (fields, indexes, constraints)
//------------------------------------------------------------------------------
procedure TACRCursor.UpdateTableDefinitions;
begin
;
end; // UpdateTableDefinitions


//------------------------------------------------------------------------------
// GetIndexDefs
//------------------------------------------------------------------------------
function TACRCursor.GetIndexDefs: TACRIndexDefs;
begin
  Result := FIndexDefs;
end;// GetIndexDefs


//------------------------------------------------------------------------------
// check field value and if not null move data from RecordBuffer to Buffer
//------------------------------------------------------------------------------
function TACRCursor.GetFieldData(
                          FieldNo:      Integer; // field no
                          Buffer:       Pointer; // buffer
                          RecordBuffer: TACRRecordBuffer // record buffer
                                      ): Boolean;
begin
 if (FieldNo >= VisibleFieldDefs.Count) then
   raise EACRException.Create(10005,ErrorLInvalidFieldNumber,[FieldNo,VisibleFieldDefs.Count]);
 Result := ACRBaseEngine.GetFieldData(VisibleFieldDefs[FieldNo].FieldNoReference,
                                      FieldDefs,Buffer,RecordBuffer);
end; // GetFieldData


//------------------------------------------------------------------------------
// set field data from Buffer to RecordBuffer
//------------------------------------------------------------------------------
procedure TACRCursor.SetFieldData(
                            FieldNo:      Integer;
                            Buffer:       Pointer;
                            RecordBuffer: TACRRecordBuffer // record buffer
                                      );
begin
 if (FieldNo >= VisibleFieldDefs.Count) then
   raise EACRException.Create(10006,ErrorLInvalidFieldNumber,[FieldNo,VisibleFieldDefs.Count]);
 ACRBaseEngine.SetFieldData(VisibleFieldDefs[FieldNo].FieldNoReference,
                                      FieldDefs,Buffer,RecordBuffer);
end; // SetFieldData


//------------------------------------------------------------------------------
// get field value
//------------------------------------------------------------------------------
procedure TACRCursor.GetFieldValue(Value: TACRVariant; FieldNo: Integer; DirectAccess: Boolean; CopyFlag: Boolean = True);
var IsNull: Boolean;
    Offset: Integer;
    Buffer: PChar;

 procedure GetBLOBValue(V: TACRVariant; aFieldNo: Integer);
 var
    bs:     TACRStream;
    size:   Int64;
    zero:   Word;
    l:      Integer;
 begin
   zero := 0;
   V.SetNull;
   V.DataType := FieldDefs[aFieldNo].BaseFieldType;
   bs := Self.InternalCreateBlobStream(False,aFieldNo,bomRead);
   try
     size := bs.Size;
     if (size > 0) then
      begin
       if (V.DataType = bftWideClob) then
        l := 2
       else
        l := 1;
       V.SetDataSize(size+l);
       bs.ReadBuffer(V.pData^,size);
       Move(zero,PChar(V.pData + size)^,l);
      end;
   finally
     InternalCloseBLOB(aFieldNo);
//     bs.Free;
   end;
 end; // GetBLOBValue


begin
 if (DirectAccess) then
  begin
   if (FieldNo >= FieldDefs.Count) then
     raise EACRException.Create(10317,ErrorLInvalidFieldNumber,
              [FieldNo,FieldDefs.Count]);
   IsNull := CheckNullFlag(FieldNo, CurrentRecordBuffer);
   if (IsNull) then
    Value.SetNull
   else
    begin
     if (IsBLOBFieldType(FieldDefs[FieldNo].FBaseFieldType)) then
      GetBLOBValue(Value,FieldNo)
     else
      begin
       Offset := FieldDefs[FieldNo].MemoryOffset;
       Buffer := PChar(CurrentRecordBuffer + Offset);
       Value.SetData(
                            Buffer,
                            FieldDefs[FieldNo].MemoryDataSize,
                            FieldDefs[FieldNo].FBaseFieldType,
                            CopyFlag
                           );
       end;
    end;
  end // access to base field defs, without projection
 else
  begin
   if (FieldNo >= VisibleFieldDefs.Count) then
     raise EACRException.Create(10737,ErrorLInvalidFieldNumber,
              [FieldNo,VisibleFieldDefs.Count]);
   IsNull := CheckNullFlag(VisibleFieldDefs[FieldNo].FieldNoReference, CurrentRecordBuffer);
   if (IsNull) then
    Value.SetNull
   else
    begin
     if (IsBLOBFieldType(VisibleFieldDefs[FieldNo].FBaseFieldType)) then
      GetBLOBValue(Value,VisibleFieldDefs[FieldNo].FFieldNoReference)
     else
      begin
       Offset := VisibleFieldDefs[FieldNo].MemoryOffset;
       Buffer := PChar(CurrentRecordBuffer + Offset);
       Value.SetData(
                            Buffer,
                            VisibleFieldDefs[FieldNo].MemoryDataSize,
                            VisibleFieldDefs[FieldNo].FBaseFieldType,
                            CopyFlag
                           );
      end;
    end;
  end; // access to fields using projection
end; // GetFieldValue


//------------------------------------------------------------------------------
// get field value
//------------------------------------------------------------------------------
procedure TACRCursor.SetFieldValue(
                            Value:        TACRVariant;
                            FieldNo:      Integer;
                            DirectAccess: Boolean;
                            RecordBuffer: TACRRecordBuffer = nil
                                  );
var Buffer:     PChar;
    Offset:     Integer;
begin
 if (RecordBuffer = nil) then
   RecordBuffer := CurrentRecordBuffer;
 if (DirectAccess) then
  begin
   if (FieldNo >= FieldDefs.Count) then
     raise EACRException.Create(10318,ErrorLInvalidFieldNumber,
            [FieldNo,FieldDefs.Count]);
   if (Value.IsNull) then
    begin
     SetNullFlag(True,FieldNo,RecordBuffer);
    end
   else
    begin
     SetNullFlag(False,FieldNo,RecordBuffer);
     if (IsBLOBFieldType(FieldDefs[FieldNo].FBaseFieldType)) then
       raise EACRException.Create(11589,ErrorLSetBLOBFieldValueIsNotAllowed,[FieldNo,FTableName])
     else
      begin
       Offset := FieldDefs[FieldNo].MemoryOffset;
       Buffer := PChar(RecordBuffer + Offset);
       Value.Cast(FieldDefs[FieldNo].FBaseFieldType);
       Value.CopyDataToAddress(Buffer,FieldDefs[FieldNo].MemoryDataSize);
      end;
    end;
  end // access to base field defs, without projection
 else
  begin
   if (FieldNo >= VisibleFieldDefs.Count) then
     raise EACRException.Create(10738,ErrorLInvalidFieldNumber,
            [FieldNo,VisibleFieldDefs.Count]);
   if (Value.IsNull) then
    begin
     SetNullFlag(True,VisibleFieldDefs[FieldNo].FieldNoReference,RecordBuffer);
    end
   else
    begin
     SetNullFlag(False,VisibleFieldDefs[FieldNo].FieldNoReference,RecordBuffer);
     if (IsBLOBFieldType(VisibleFieldDefs[FieldNo].FBaseFieldType)) then
       raise EACRException.Create(11591,ErrorLSetBLOBFieldValueIsNotAllowed,[FieldNo,FTableName])
     else
      begin
       Offset := VisibleFieldDefs[FieldNo].MemoryOffset;
       Buffer := PChar(RecordBuffer + Offset);
       Value.Cast(VisibleFieldDefs[FieldNo].FBaseFieldType);
       Value.CopyDataToAddress(Buffer,VisibleFieldDefs[FieldNo].MemoryDataSize);
      end;
    end;
  end; // access to fields using projection
end; // SetFieldValue


//------------------------------------------------------------------------------
// allocate record buffer and set null flags
//------------------------------------------------------------------------------
function TACRCursor.AllocateRecordBuffer: TACRRecordBuffer;
begin
 Result := MemoryManager.GetMem(RecordBufferSize);
end; // AllocateRecordBuffer


//------------------------------------------------------------------------------
// free record buffer
//------------------------------------------------------------------------------
procedure TACRCursor.FreeRecordBuffer(var Buffer: TACRRecordBuffer);
begin
 MemoryManager.FreeAndNilMem(Buffer);
end; // FreeRecordBuffer


//------------------------------------------------------------------------------
// allocate record buffer and set null flags
//------------------------------------------------------------------------------
function TACRCursor.AllocateKeyRecordBuffer: TACRRecordBuffer;
begin
  Result := MemoryManager.GetMem(KeyBufferSize);
end;


//------------------------------------------------------------------------------
// initialize record buffer
//------------------------------------------------------------------------------
procedure TACRCursor.InternalInitKeyBuffer(RecordBuffer: TACRRecordBuffer);
begin
  FillChar(RecordBuffer^,KeyBufferSize,0);
  InternalInitRecord(RecordBuffer,False);
end;


//------------------------------------------------------------------------------
// go to first record (before first record to BOF)
//------------------------------------------------------------------------------
procedure TACRCursor.InternalFirst;
begin
 FirstPosition := True;
 LastPosition := False;
end; // InternalFirst


//------------------------------------------------------------------------------
// go to last record (after last record to EOF)
//------------------------------------------------------------------------------
procedure TACRCursor.InternalLast;
begin
 FirstPosition := False;
 LastPosition := True;
end; // InternalLast


//------------------------------------------------------------------------------
// save position
//------------------------------------------------------------------------------
function TACRCursor.SavePosition: Pointer;
begin
// Result := New(PACRCursorPos);
 Result := MemoryManager.GetMem(SizeOf(TACRCursorPos));
 PACRCursorPos(Result)^.FirstPosition := Self.FirstPosition;
 PACRCursorPos(Result)^.LastPosition := Self.LastPosition;
 PACRCursorPos(Result)^.RecordID := Self.CurrentRecordID;
end; // SavePosition


//------------------------------------------------------------------------------
// restore position
//------------------------------------------------------------------------------
procedure TACRCursor.RestorePosition(Pos: Pointer);
begin
 Self.FirstPosition := PACRCursorPos(Pos)^.FirstPosition;
 Self.LastPosition := PACRCursorPos(Pos)^.LastPosition;
 Self.CurrentRecordID := PACRCursorPos(Pos)^.RecordID;
end; // RestorePosition


//------------------------------------------------------------------------------
// free position
//------------------------------------------------------------------------------
procedure TACRCursor.FreePosition(var Pos: Pointer);
begin
 if (Pos <> nil) then
  MemoryManager.FreeAndNilMem(Pos);
end; // FreePosition


//------------------------------------------------------------------------------
// return true if index applied
//------------------------------------------------------------------------------
function TACRCursor.IsIndexApplied: Boolean;
begin
  Result := (IndexID <> INVALID_OBJECT_ID);
end; // IsIndexApplied


//------------------------------------------------------------------------------
// return true if distinct applied
//------------------------------------------------------------------------------
function TACRCursor.IsDistinctApplied: Boolean;
begin
  Result := (FDistinctFieldCount > 0);
end; // IsDistinctApplied


//------------------------------------------------------------------------------
// return true if filter applied
//------------------------------------------------------------------------------
function TACRCursor.IsFilterApplied: Boolean;
begin
  Result := ((FilterExpression <> nil) or (FilterRecord <> nil) or
             (SQLFilterExpression <> nil));
end; // IsFilterApplied


//------------------------------------------------------------------------------
// return true if range is applied
//------------------------------------------------------------------------------
function TACRCursor.IsRangeApplied: Boolean;
begin
  Result := ((FRangeStartBuffer <> nil) or (FRangeEndBuffer <> nil));
end; // IsRangeApplied


//------------------------------------------------------------------------------
// reset range
//------------------------------------------------------------------------------
function TACRCursor.IsViewRestricted: Boolean;
begin
  Result := (IsFilterApplied or IsRangeApplied or IsDistinctApplied);
end; // IsViewRestricted


////////////////////////////////////////////////////////////////////////////////
//
// TACRSessionNamedObject
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSessionNamedObject.Create(Name: String);
begin
  FName := Name;
end;//Create



////////////////////////////////////////////////////////////////////////////////
//
// TACRSessionNamedObjectSequenceValue
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSessionNamedObjectSequenceValue.Create(ValueName: String;
  Value: TACRSequenceValue);
begin
  inherited Create(ValueName);
  FValue := Value;
end;//Create



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseSession
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRBaseSession.Create;
begin
  FSessionNamedObjectList := TACRSortedStringPtrArray.Create;
  FSessionID := INVALID_SESSION_ID;
end;//Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRBaseSession.Destroy;
var i: integer;
begin
  SetConnected(False);
  if (Length(FCryptoParams.Password) > 0) then
   FillChar(FCryptoParams.Password[1],Length(FCryptoParams.Password),$00);
  FCryptoParams.Password := '';
  if (Length(FConnectParams.CryptoInfo.Password) > 0) then
   FillChar(FConnectParams.CryptoInfo.Password[1],Length(FConnectParams.CryptoInfo.Password),$00);
  FConnectParams.CryptoInfo.Password := '';
  FillChar(FCryptoParams,SizeOf(FCryptoParams),$00);
  FillChar(FConnectParams,SizeOf(FConnectParams),$00);
  for i:=0 to FSessionNamedObjectList.Count -1 do
    TACRSessionNamedObject(FSessionNamedObjectList[i]).Free;
  FSessionNamedObjectList.Free;
end;//destroy


//------------------------------------------------------------------------------
// create database
//------------------------------------------------------------------------------
procedure TACRBaseSession.CreateDatabase;
begin
  raise EACRException.Create(10888,ErrorLOperationIsNotSupported);
end; // CreateDatabase


//------------------------------------------------------------------------------
// delete database
//------------------------------------------------------------------------------
procedure TACRBaseSession.DeleteDatabase;
begin
  raise EACRException.Create(10889,ErrorLOperationIsNotSupported);
end; // DeleteDatabase


//------------------------------------------------------------------------------
// rename database
//------------------------------------------------------------------------------
procedure TACRBaseSession.RenameDatabase(NewDatabaseFileName: String);
begin
  raise EACRException.Create(10890,ErrorLOperationIsNotSupported);
end; // RenameDatabase


//------------------------------------------------------------------------------
// Get Named object from session
//------------------------------------------------------------------------------
function TACRBaseSession.GetNamedObject(ObjectName: String): TACRSessionNamedObject;
var i: integer;
begin
  Result := nil;
  for i:=0 to FSessionNamedObjectList.Count -1 do
   if (ObjectName = TACRSessionNamedObject(FSessionNamedObjectList[i]).Name) then
    begin
     Result := TACRSessionNamedObject(FSessionNamedObjectList[i]);
     break;
    end;
end;//GetNamedObject


//------------------------------------------------------------------------------
// set sequence value
//------------------------------------------------------------------------------
procedure TACRBaseSession.SetSequenceValue(const Name:  TACRObjectName;
                                           const Value: TACRSequenceValue);
var
  OldValue: TACRSessionNamedObject;
begin
  OldValue := GetNamedObject(Name);
  if (OldValue = nil) then
   begin
    // create new value
    OldValue := TACRSessionNamedObjectSequenceValue.Create(Name,Value);
    FSessionNamedObjectList.Add(OldValue);
   end
  else
   begin
    TACRSessionNamedObjectSequenceValue(OldValue).Value := Value;
   end;
end; // SetSequenceValue


//------------------------------------------------------------------------------
// for client and server sessions
//------------------------------------------------------------------------------
procedure TACRBaseSession.ReceiveData(Buffer: PChar; BufferSize: Integer);
begin
end; // ReceiveData


//------------------------------------------------------------------------------
// for client and server sessions
//------------------------------------------------------------------------------
procedure TACRBaseSession.OnDisconnect;
begin
end; // OnDisconnect


//------------------------------------------------------------------------------
// for TACRLocalSession called by TACRServerSession
//------------------------------------------------------------------------------
procedure TACRBaseSession.RemoveAllLocks;
begin
end; // RemoveAllLocks


////////////////////////////////////////////////////////////////////////////////
//
// TACRSQLProcessor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRSQLProcessor.Create(Query: TDataSet);
begin
  FSQLParams := TACRSQLParams.Create;
  FACRQuery := Query;
  FCursor := nil;
  FSqlText := '';
end;//Create


//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
destructor TACRSQLProcessor.Destroy;
begin
  FSQLParams.Free;
  inherited;
end;//Destroy


//------------------------------------------------------------------------------
// OpenQuery
//------------------------------------------------------------------------------
function TACRSQLProcessor.OpenQuery: TACRCursor;
begin
  FRowsAffected := 0;
  Result := nil;
end;//OpenQuery


//------------------------------------------------------------------------------
// ExecuteQuery
//------------------------------------------------------------------------------
procedure TACRSQLProcessor.ExecuteQuery;
begin
  OpenQuery;
end;//ExecuteQuery


//------------------------------------------------------------------------------
// PrepareStatement
//------------------------------------------------------------------------------
procedure TACRSQLProcessor.PrepareStatement(SQLText: PChar);
begin
  FSqlText := SQLText;
end;//PrepareStatement



initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRBase> initialized');
{$ENDIF}

end.

