unit ACRStrUtils;

interface

uses SysUtils
{$IFDEF MSWINDOWS}
     ,Windows
{$ENDIF}
{$IFDEF LINUX}
     ,Libc
{$ENDIF}
     ;

type                        

  PACRCharTable = ^TACRCharTable;
  TACRCharTable = array[0..255] of Char;
  PACRByteTable = ^TACRByteTable;
  TACRByteTable = array[0..255] of Byte;

var
  ToUpperChars, ToLowerChars: TACRCharTable;
  OrderedChars: TACRByteTable;


{
  function StrComp(P1, P2: PChar): Integer; overload;
  function StrComp(P1, P2: PChar; OrderTable: PACRByteTable): Integer; overload;
  function StrCompL(P1, P2: PChar; MaxLen: Cardinal): Integer; overload;
  function StrCompL(P1, P2: PChar; MaxLen: Cardinal; OrderTable: PACRByteTable): Integer; overload;

  function StrIComp(P1, P2: PChar): Integer; overload;
  function StrIComp(P1, P2: PChar; OrderTable: PACRByteTable; UpperTable: TACRCharTable): Integer; overload;
  function StrICompL(P1, P2: PChar; MaxLen: Cardinal): Integer; overload;
  function StrICompL(P1, P2: PChar; MaxLen: Cardinal; OrderTable: PACRByteTable; UpperTable: TACRCharTable): Integer; overload;
}



  procedure ChangeCaseCustom(Buffer: PChar; PTable: PACRCharTable); register;
  procedure ToLowerCase(Buffer: PChar); register;
  procedure ToUpperCase(Buffer: PChar); register;


implementation


procedure ChangeCaseCustom(Buffer: PChar; PTable: PACRCharTable);
asm
        TEST    EAX, EAX                                   // Buffer = nil ?
        JE      @exit
        TEST    EDX, EDX                                   // Table = nil ?
        JE      @exit

@loop:  MOVZX   ECX, BYTE PTR [EAX]
        TEST    CL, CL                                     // End of String ?
        JE      @exit
        MOV     CL, BYTE PTR [EDX + ECX]
        MOV     BYTE PTR [EAX], CL
        INC     EAX
        JMP     @loop
@exit:
end;



procedure ToLowerCase(Buffer: PChar);
asm
        TEST    EAX, EAX                                   // Buffer = nil ?
        JE      @exit
@loop:  MOVZX   ECX, BYTE PTR [EAX]
        TEST    CL, CL                                     // End of String ?
        JE      @exit
        MOV     CL, BYTE PTR [ToLowerChars + ECX]
        MOV     BYTE PTR [EAX], CL
        INC     EAX
        JMP     @loop
@exit:
end;


procedure ToUpperCase(Buffer: PChar);
asm
        TEST    EAX, EAX                                   // Buffer = nil ?
        JE      @exit
@loop:  MOVZX   ECX, BYTE PTR [EAX]
        TEST    CL, CL                                     // End of String ?
        JE      @exit
        MOV     CL, BYTE PTR [ToUpperChars + ECX]
        MOV     BYTE PTR [EAX], CL
        INC     EAX
        JMP     @loop
@exit:
end;



procedure InitOrderedChars;
var i,j: Integer;
  a1, a2: array[0..1] of Byte;
  p1, p2: PChar;
  n: Byte;
begin
  a1[1] := 0;
  a2[1] := 0;
  p1 := @a1;
  p2 := @a2;
  for i:=0 to 255 do
    begin
      n := 0;
      a1[0] := i;
      for j:=0 to 255 do
        begin
          a2[0] := j;
          if (AnsiStrLComp(p1, p2, 1) > 0) then
            Inc(n);
        end;
      OrderedChars[i] := n;
    end;
end;


var i: Integer;
initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRSecurity> initialization started');
{$ENDIF}

{$IFDEF MSWINDOWS}
  for i:=0 to 255 do
    ToUpperChars[i] := chr(i);
  move(ToUpperChars, ToLowerChars, 256);
  CharUpperBuff(PAnsiChar(@ToUpperChars), 256);
  CharLowerBuff(PAnsiChar(@ToLowerChars), 256);
{$ENDIF}

{$IFDEF LINUX}
  for i:=0 to 255 do
   begin
    ToUpperChars[i] := chr(toupper(i));
    ToLowerChars[i] := chr(tolower(i));
   end;
{$ENDIF}

  InitOrderedChars;

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRSecurity> initialization finished');
{$ENDIF}

end.
