unit ACRTypes;

interface

{$I ACRVer.inc}

uses
     SysUtils, Classes,
{$IFDEF MSWINDOWS}
     Controls,
     Windows,
{$ENDIF}

// Accuracer units
{$IFDEF LINUX}
     ACRLinux,
{$ENDIF}
     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRMemory,
     ACRConst,
     ACRExcept;


type

//------------------------------------------------------------------------------
// Encryption types
//------------------------------------------------------------------------------

  TACRCryptoKey = packed record
   Key:             array [0..ACR_MAX_KEY] of Byte;
   KeySize:         Word; // 0 by default
  end;

  TACRCryptoInfo = packed record
   KeyInfo:         TACRCryptoKey;
   InitVector:      array [0..ACR_MAX_VECTOR] of Byte;
   Password:        String; // ACRDefaultPassword by default
   CryptoAlgorithm: Byte;  // acr_Cipher_None by Default
   CryptoMode:      Byte;  // acr_CTS by Default
   UseInitVector:   Boolean; // False by default
  end;

  TACRCryptoControlBlock = packed record
   Data:            array [0..ACR_MAX_CONTROL_BLOCK] of Byte;
  end; // 256

  TACRCryptoHeader = packed record
   ControlBlock:      TACRCryptoControlBlock;
   ControlBlockCRC:   Cardinal;
   CryptoAlgorithm:   Byte;
   CryptoMode:        Byte;
   CryptoAskPassword: Byte; // ask password (1) or key (0)
   Reserverd:         array[0..4] of Byte;
  end; // 268

 TACRCryptoParams = TACRCryptoInfo;

//------------------------------------------------------------------------------
// general types
//------------------------------------------------------------------------------

 TACRIntegerArray = class;
 
 TACRErrorCode = Integer;
 TACRPageNo = Integer;
 PACRPageNo = ^TACRPageNo;
 TACRPageBuffer = PChar;
 TACRGetRecordMode = (grmCurrent, grmNext, grmPrior);
 TACRGetRecordResult = (grrOK, grrBOF, grrEOF, grrError);
 TACRBLOBOpenMode = (bomWrite, bomRead, bomReadWrite);
 TACRRecordBuffer = PChar;
 TACRState = Integer;
 TACRObjectID = Integer;
 TACRTableID = TACRObjectID;
 TACRObjectName = ShortString;
 TACRSessionID = Integer;
 TACRRecordNo = Int64;
 TACRPageItemID = Word;
 TACRRecordID = packed record
  PageNo:        TACRPageNo; // page number or record number (disk engine or memory, temporary engine)
  PageItemNo:    TACRPageItemID;
 end;
 PACRRecordID = ^TACRRecordID;
 TACRLockParams = packed record
  RetryCount:   Integer;
  Delay:        Integer;
 end;
 TACROptions = packed record
  MaxSessionCount:            Integer;
  PageSize:                   Integer;
  ExtentPageCount:            Integer;
  RandomSearchRetryCount:     Integer;
 end;


//------------------------------------------------------------------------------
// network types
//------------------------------------------------------------------------------

 TACRConnectParams = packed record
    ConnectionParamsTunning:          Boolean;
    PingClients:                      Boolean;
    CompressionAlgorithm:             Byte;
    CompressionMode:                  Byte;
    CryptoInfo:                       TACRCryptoInfo;
    RemoteHost:                       String;
    RemotePort:                       Integer;
    LocalPort:                        Integer;
    ServerID:                         Integer;
    PacketSize:                       Integer;
    MaxThreadCount:                   Integer;
    TestPacketCount:                  Integer;
    ConnectRetryCount:                Integer;
    ConnectDelay:                     Integer;
    StartReceiveTimeOut:              Integer;
    ReceiveTimeOut:                   Integer;
    ReceiveSleep:                     Integer;
    ServerReceiveTimeOut:             Integer;
    ServerReceiveSleep:               Integer;
    MinSendTimeOut:                   Integer;
    SendTimeOut:                      Integer;
    MinServerSendTimeOut:             Integer;
    ServerSendTimeOut:                Integer;
    WaitForSendSleep:                 Integer;
    ServerWaitForSendSleep:           Integer;
    ResendDelay:                      Integer;
    RequestDelay:                     Integer;
    ServerResendDelay:                Integer;
    ServerRequestDelay:               Integer;
    DisconnectRetryCount:             Integer;
    DisconnectDelay:                  Integer;
    WaitForTimeOut:                   Integer;
    WaitForMessagesSend:              Integer;
    WaitForServerSessionThreadTimeOut:Integer;
    ThreadsTerminateDelay:            Integer;
    ServerThreadsTerminateDelay:      Integer;
    ServerSessionTerminatorSleep:     Integer;
    WaitForPingAnswer:                Integer;
    ServerPingSleep:                  Integer;
 end;
 PACRConnectParams = ^TACRConnectParams;

 TACRMessageType = (aamtText,aamtBinary,aamtStream);

//------------------------------------------------------------------------------
// disk types
//------------------------------------------------------------------------------

 TACRIsPageSynchronizationRequired =
  function (PageState: Cardinal): Boolean of object;

 TACRPageRecordCount = Word;
 PACRPageRecordCount = ^TACRPageRecordCount;
 TACRDBFileType = (dbftUnknown,dbftTablesList,dbftActiveSessionsList);
 TACRLockType = (ltIS,ltS,ltIRW,ltRW,ltU,ltX);
 TACRLockObjectType = (lotDatabase,lotTable,lotRecord);
 TACRTablePageType = (tptRowStart,tptRowContinue,tptVarchar,tptBLOB);

 TACRDBHeader = packed record
  Signature:        TACRSignature;
  HeaderSize:       Word;
  Version:          Double;
  PageSize:         Word;
  ExtentPageCount:  Word;
  TotalPageCount:   TACRPageNo;
  State:            Cardinal;
  CryptoHeader:     TACRCryptoHeader; // for database encryption
 end;
 TACRLockedBytes = packed record
  LockedByteCount:  Word;
 end;

 TACRPageTypeID = Word;

 TACRDiskPageHeader = packed record
  Signature:        Array [0..3] of Char;
  State:            Integer;
  PageType:         TACRPageTypeID;
  NextPageNo:       TACRPageNo;
  CRC32:            Longword;
  CRCType:          Byte;
  HashType:         Byte;
  Cipherype:        Byte;
  MACType:          Byte;
  ObjectID:         TACRObjectID;
  RecordID:         TACRRecordID;
  Reserved:         Array [0..7] of Byte;
 end; // 40 bytes
 PACRDiskPageHeader = ^TACRDiskPageHeader;

 TACRBackupDescHeader = packed record
  Date:                 TDateTime;
  FileSize:             Int64;
  TableCount:           Integer;
  DescLength:           Integer;
  Size:                 Integer; // size of description and tables list
  // description
  // tables names
 end; // TACRBackupDescHeader

 TACRBackupHeader = packed record
  Signature:            Array [0..3] of Char;
  BlockSize:            Integer;
  CompressionAlgorithm: Byte;
  CompressionMode:      Byte;
  CryptoHeader:         TACRCryptoHeader; // for database encryption
  DescHeader:           TACRBackupDescHeader;
 end; // TACRBackupHeader


 TACRBlockHeader = packed record
  CompressedSize:   Integer;
  UncompressedSize: Integer;
 end; // TACRBlockHeader


  // ACRDatabaseFile Mode Types
  TACRShareMode = (smExclusive, smShared);
  TACRAccessMode = (amReadOnly, amReadWrite);
 
 // Any list header
 PACRListHeader = ^TACRListHeader;
 TACRListHeader = packed record
   Count:       Integer;
   ItemSize:    Integer;
   NextPageNo:  TACRPageNo; // or INVALID_PAGE_NO
 end;// 12 byte

 // Header for internal file
 PACRInternalFileHeader = ^TACRInternalFileHeader;
 TACRInternalFileHeader = packed record
   Signature:       Array [0..3] of Char;
   FileSize:        Integer;
   CompressionMode: Byte;
   Reserved:        Array [0..2] of Char;
 end; // 12 bytes

 // System Directory List Item
 TACRSystemDirectoryListItem = packed record
   FileID:      TACRDBFileType;
   FirstPageNo: TACRPageNo;
 end;// 6 byte

 
 // Table List Item
 PACRTableListItem = ^TACRTableListItem;
 TACRTableListItem = packed record
   TableName:                           ShortString;
   TableID:                             TACRTableID;
   TableMetaDataFilePageNo:             TACRPageNo;
   MostUpdatedTableMetadataFilePageNo:  TACRPageNo;
   PFSPageMapFilePageNo:                TACRPageNo;
   LockFilePageNo:                      TACRPageNo;
   CryptoHeader:                        TACRCryptoHeader; // for table encryption  
 end;
 
 TACRTablePFSPageMapItem = packed record
   PageNo:          TACRPageNo;
   PageRecordCount: Integer;
 end;

 TACRTablePFSItem = packed record
   PageType:        TACRTablePageType;
   PageRecordCount: Word; // Record Count for RowStart, otherwise Free Space
                          // $FFFF - empty or other table page
 end;

 // blob or varchar
 TACRDataItem = packed record
    ItemID:       Word;
    DataSize:     Word;
 end;
 TACRDataItemsMap = array of TACRDataItem;

 TACRFileData = packed record
    Size:   Integer;
    Buffer: PChar;
    Pages:  TACRIntegerArray;
 end;

 TACRTableLockInfo = packed record
    NumLocksX:      Integer;
    NumLocksIS:     Integer;
    NumLocksS:      Integer;
    NumLocksIRW:    Integer;
    NumLocksRW:     Integer;
    IsRecordLocked: Boolean;
    LockedRecordID: TACRRecordID;
 end;
 PACRTableLockInfo = ^TACRTableLockInfo;


 TACRSessionLockItem = packed record
   SessionID:     TACRSessionID;
   TableLockInfo: TACRTableLockInfo;
 end;
 PACRSessionLockItem = ^TACRSessionLockItem;

 TACRTableLockItemForSession = packed record
   TableName:     String;
   TableLockInfo: TACRTableLockInfo;
 end;
 PACRTableLockItemForSession = ^TACRTableLockItemForSession;
//------------------------------------------------------------------------------
// SQL types
//------------------------------------------------------------------------------


 // join type
 TACRJoinType = (ajtCross, ajtInner, ajtLeftOuter,
                 ajtRightOuter, ajtFullOuter);

 // union type
 TACRUnionType = (autUnion, autIntersect, autExcept);

 // table | joined table | subquery
 TACRTableType = (attTable, attJoinedTable, attSubQuery);

 // table | joined table | subquery
 TACRQueryExprType = (qetSelect, qetUnion, qetExcept, qetIntersect);




 // field types
 TACRBaseFieldType = (
                        bftUnknown,

                        bftChar,
                        bftWideChar,
                        bftVarchar,
                        bftWideVarchar,

                        bftSignedInt8,        // Shortint
                        bftSignedInt16,       // Smallint
                        bftSignedInt32,       // Integer
                        bftSignedInt64,       // Int64
                        bftUnsignedInt8,      // Byte
                        bftUnsignedInt16,     // Word
                        bftUnsignedInt32,     // Cardinal
                        //bftUnSignedInt64,

                        bftSingle,
                        bftDouble,
                        bftExtended,

                        bftDate,
                        bftTime,
                        bftDateTime,

                        bftBlob,
                        bftClob,
                        bftWideClob,

                        bftLogical,
                        bftCurrency,

                        bftBytes,
                        bftVarBytes,
                        //bftTimeStamp, == bftDateTime 
                        bftBFile
                     );


  TACRAdvancedFieldType = (
    aftUnknown,

    aftChar,            // = bftChar
    aftString,          // = bftVarChar

    aftWideChar,        // = bftChar
    aftWideString,      // = bftWideVarChar

    aftShortint,        // = bftSignedInt8
    aftSmallint,        // = bftSignedInt16
    aftInteger,         // = bftSignedInt32
    aftLargeint,        // = bftSignedInt64
    aftByte,            // = bftUnsignedInt8
    aftWord,            // = bftUnsignedInt16
    aftCardinal,        // = bftUnsignedInt32

    aftAutoInc,         // = bftSignedInt32
    aftAutoIncShortint, // = bftSignedInt8
    aftAutoIncSmallint, // = bftSignedInt16
    aftAutoIncInteger,  // = bftSignedInt32
    aftAutoIncLargeint, // = bftSignedInt64
    aftAutoIncByte,     // = bftUnsignedInt8
    aftAutoIncWord,     // = bftUnsignedInt16
    aftAutoIncCardinal, // = bftUnsignedInt32


    aftSingle,          // = bftSingle
    aftDouble,          // = bftDouble
    aftExtended,        // = bftExtended

    aftBoolean,         // = bftLogical

    aftCurrency,        // = bftCurrency

    aftDate,            // = bftDate
    aftTime,            // = bftTime
    aftDateTime,        // = bftDateTime
    aftTimeStamp,       // = bftDateTime

    aftBytes,           // = bftBytes
    aftVarBytes,        // = bftVarBytes

    aftBlob,            // = bftBlob
    aftGraphic,         // = bftBlob
    aftMemo,            // = bftClob
    aftFormattedMemo,   // = bftClob
    aftWideMemo         // = bftWideClob

//    aftGuid             // = bftChar(XX)
//aftArray,
//    aftParadoxOle,
//    aftDBaseOle,
//    aftTypedBinary,
//    aftCursor,
//    aftReference,
//    aftDataSet,
//    aftVariant,
//    aftInterface,
//    aftIDispatch,
    );

 //TACRDefaultValueType = (dvtNull, dvtConst, dvtSequence, dvtFunction, dvtQuery);
 TACRCompareResult = (cmprEqual, cmprLower, cmprGreater,
                      cmprLeftNull, cmprRightNull, cmprBothNull);

//------------------------------------------------------------------------------
// record types
//------------------------------------------------------------------------------

 // used in ACRMemEngine for saving / loading table to stream
 TACRMemoryTableFileHeader = packed record
  Signature:            array [0..3] of char ;
  Version:              Double;
  RecordCount:          TACRRecordNo;
  // compression params
  UncompressedSize:     Int64;
  NumBlocks:            Int64;
  BlockSize:            Integer;
  CompressionAlgorithm: Byte;
  CompressionMode:      Byte;
  NameLength:           Byte;
 end; // 43 bytes

 TACRIndexPosition = TObject;
 TACRSearchInfo = Pointer;

  // bookmark information type
  TACRBookmarkFlag = (abfCurrent, abfBOF, abfEOF, abfInserted);
  TACRBookmarkInfo = packed record
   BookmarkData: TACRRecordID;      // record ID
   BookmarkFlag: TACRBookmarkFlag;  // bookmark flag
  end;
  PACRBookmarkInfo = ^TACRBookmarkInfo;

 TACRSequenceValue = Int64;

 TACRTimeoutParams = record
  RetryCount:   Word;
  Timeout:      Integer;
 end;

 //------------------------- Search types --------------------------------------
  TACRKeyIndex = (kiLookup, kiRangeStart, kiRangeEnd, kiCurRangeStart,
    kiCurRangeEnd, kiSave);
  TACRSearchCondition = (scNone,scEqual,scLower,scGreater,scLowerEqual,scGreaterEqual);
  TACRKeySearchType = (kstFirst,kstLast,kstAny);

  PACRKeyBuffer = ^TACRKeyBuffer;
  TACRKeyBuffer = packed record
    Modified: Boolean;
    Exclusive: Boolean;
    FieldCount: Integer;
  end;

  TACRNavigationInfo = packed record
    // in
    SessionID:        TACRSessionID; // for disk engine only; INVALID_SESSION_ID for memory and temp engines
    GetRecordMode:    TACRGetRecordMode;
    IndexID:          TACRObjectID; // current index id (INVALID_OBJECT_ID if no active index)
    // in and out
    RecordID:         TACRRecordID;
    RecordBuffer:     TACRRecordBuffer;
    FirstPosition:    Boolean;
    LastPosition:     Boolean;
    // out
    GetRecordResult:  TACRGetRecordResult;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRScanSearchCondition
//
////////////////////////////////////////////////////////////////////////////////

  TACRScanSearchCondition = class(TObject)
   public
    Condition:                 TACRSearchCondition;
    KeyRecordBuffer:           TACRRecordBuffer; // if nil then Expression will be used
    KeyFieldCount:             Integer;
    IndexID:                   TACRObjectID;
    Expression:                Pointer; // if nil then KeyBuffer will be used

    constructor Create;
    destructor Destroy; override;
    procedure Assign(ScanSearchCondition: TACRScanSearchCondition);
  end;

  TACRExtractedConditionInfo = packed record
    KeyRecordBuffer:           TACRRecordBuffer;
    ExtractedExpressionNode:   Pointer;
    Expression:                Pointer; // if nil then KeyBuffer will be used
  end;
  PACRExtractedConditionInfo = ^TACRExtractedConditionInfo;

////////////////////////////////////////////////////////////////////////////////
//
// TACRScanSearchConditionArray
//
////////////////////////////////////////////////////////////////////////////////

  TACRScanSearchConditionArray = class(TObject)
   public
    Items:    array of TACRScanSearchCondition;
    Count:    Integer;

    constructor Create;
    destructor Destroy; override;
    procedure Delete(ItemNo: Integer);
    procedure AddExpression(Expression: Pointer);
    procedure AddCondition(ScanSearchCondition: TACRScanSearchCondition);
    procedure ExtractConditionsFromExpressions(IndexDefs: Pointer; ExtractedConditionsInfo: TList);
    procedure ReturnConditionsToExpressions(ExtractedConditionsInfo: TList);
  end;

 // ORDER BY clause element
 TACRSortSpecification = record
   TableName:       String;   // table or its pseudonym
   ColumnName:      String;  // field name or pseudonym
   ColumnNumber:    Integer; // field number (1,2 ...)
   Descending:      Boolean; // ASC | DESC
   CaseInsensitive: Boolean; // NOCASE | CASE
 end;
  

//----------------------- Field Types ------------------------------------------

 PACRDate = ^TACRDate;
 TACRDate = Integer;  // Number of days from Christmas ( -5 883 516 .. 5 883 516 )

 PACRTime = ^TACRTime;
 TACRTime = Integer; // Number of milliseconds from day begin ( 0 .. 24*60*60*1000 )

 PACRDateTime = ^TACRDateTime;
 TACRDateTime = packed record  // Total 8 byte
   Date: TACRDate;  // Number of days from Christmas ( -5 883 516 .. 5 883 516 )
   Time: TACRTime; // Number of milliseconds from day begin ( 0 .. 24*60*60*1000 )
 end;//TACRDateTime                                                = 86 400 000

 PACRLogical = ^TACRLogical;
 //TACRLogical = ByteBool;
 TACRLogical = WordBool; //TDataset using WordBool

 PACRCurrency = ^TACRCurrency;
 //TACRCurrency = Currency;
 TACRCurrency = Double;  // Overflow fix

//------------------------------------------------------------------------------
// Delphi 4,5 types
//------------------------------------------------------------------------------

{$IFNDEF D6H}
 PWord = ^Word;
 PInteger = ^Integer;
 PByte = ^Byte;
 PCardinal = ^Cardinal;
{$ENDIF}


TACRPagesArray = TACRIntegerArray;

//------------------------------------------------------------------------------
// PageManager types
//------------------------------------------------------------------------------
  TACRSessionPageInfo = record
    SessionID:          TACRSessionID;
    DirtyPages:         TList;
    AddedPageNumbers:   TACRPagesArray;
    RemovedPageNumbers: TACRPagesArray;
  end;

////////////////////////////////////////////////////////////////////////////////
//
// TACRIntegerArray
//
////////////////////////////////////////////////////////////////////////////////

 TACRIntegerArray = class(TObject)
   public
     Items:          array of Integer;
     ItemCount:      Integer;
     AllocBy:        Integer;
     deAllocBy:      Integer;
     MaxAllocBy:     Integer;
     AllocItemCount: Integer;

     constructor Create(
      size: Integer = 0;
      DefaultAllocBy: Integer = 10;
      MaximumAllocBy: Integer = 100
     );
     destructor Destroy; override;
     procedure SetSize(newSize: Integer);
     procedure Assign(v: TACRIntegerArray);
     procedure Append(value: Integer);
     procedure Remove(value: Integer);
     procedure Insert(ItemNo: Integer; value: Integer);
     procedure Delete(ItemNo: Integer);
     procedure MoveTo(itemNo, newItemNo: Integer);
     procedure CopyTo(
                      var ar: array of Integer;
                      itemNo, iCount: Integer
                      );
     function IsValueExists(value: Integer): Boolean;
     function Add(value: Integer): Boolean;
 end; // TACRIntegerArray


////////////////////////////////////////////////////////////////////////////////
//
// TACRThreadIntArray
//
////////////////////////////////////////////////////////////////////////////////

 TACRThreadIntArray = class(TACRIntegerArray)
   private
    FCSect:         TRTLCriticalSection;
   public
    constructor Create(
      size: Integer = 0;
      DefaultAllocBy: Integer = 10;
      MaximumAllocBy: Integer = 100
                        );
    destructor Destroy; override;
    procedure Lock;
    procedure Unlock;
 end; // TACRThreadIntArray



////////////////////////////////////////////////////////////////////////////////
//
// TACRRecordIDArray
//
////////////////////////////////////////////////////////////////////////////////


 TACRRecordIDArray = class(TObject)
   private
     FSorted:        Boolean;
     FInMemory:      Boolean;
     FTemporary:     Boolean;
     FMaxRecordID:   TACRRecordID;
   public
     Items:          array of TACRRecordID;
     ItemCount:      Integer;
     AllocBy:        Integer;
     deAllocBy:      Integer;
     MaxAllocBy:     Integer;
     AllocItemCount: Integer;

     constructor Create(
      Size:           Integer = 0;
      DefaultAllocBy: Integer = 10;
      MaximumAllocBy: Integer = 100
     );
     destructor Destroy; override;
     procedure SetSize(newSize: Integer);
     procedure Assign(v: TACRRecordIDArray);
     procedure Append(var value: TACRRecordID);
     procedure Insert(ItemNo: Integer; const value: TACRRecordID);
     procedure Delete(ItemNo: Integer);
    protected
     // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
     function CompareRecordIDMemory(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
     // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
     function CompareRecordIDTemporary(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
     // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
     function CompareRecordIDDisk(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
    public
     // return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
     function CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
    protected 
     // return -1 if not found, otherwise return position in Items array [0..ItemCount-1]
     function FindRecordByIDSorted(const RecordID: TACRRecordID): Integer;
    public
     procedure Sort;
     // return -1 if not found, otherwise return position in Items array [0..ItemCount-1]
     function FindRecordByID(const RecordID: TACRRecordID): Integer;
    public
     property Sorted: Boolean read FSorted write FSorted;
     property InMemory: Boolean read FInMemory write FInMemory;
     property Temporary: Boolean read FTemporary write FTemporary;
     property MaxRecordID: TACRRecordID read FMaxRecordID write FMaxRecordID;
 end; // TACRRecordIDArray


////////////////////////////////////////////////////////////////////////////////
//
// TACRPtrArray
//
////////////////////////////////////////////////////////////////////////////////
(*
 TACRPtrArray = class(TObject)
   private
    FItems:           array of Pointer;
    FSizes:           array of Integer;
    FOverallSize:     Integer;
    FItemCount:       Integer;
    FAllocBy:         Integer;
    FDeAllocBy:       Integer;
    FMaxAllocBy:      Integer;
    FAllocItemCount:  Integer;
    FMemoryManager:   TACRMemoryManager;
   public
//     constructor Create(
//                        Size, MaxAllocBy, AllocBy, DeAllocaBy: Integer = 0;
//                        MemoryManager: TACRMemoryManager = nil
//                         );
//     destructor Destroy; override;
     procedure SetSize(NewSize: Integer); virtual; abstract;
     procedure Append(
                      Buffer:     Pointer;
                      BufferSize: Integer;
                      CopyFlag:   Boolean = False // to copy buffer contents or not
                      ); virtual; abstract;
 end; // TACRPtrArray
*)


////////////////////////////////////////////////////////////////////////////////
//
// TACRSortedPtrArray
//
////////////////////////////////////////////////////////////////////////////////

 TACRSortedPtrArray = class(TObject)
  private
   KeyItems: array of integer;
   ValueItems: array of integer;
   ItemCount: integer;
   AllocBy: integer;
   deAllocBy: integer;
   MaxAllocBy: integer;
   AllocItemCount: integer;

   function FindPositionForInsert(key: Integer) : Integer;
   function FindPosition(key: Integer): Integer;
   procedure InsertByPosition(ItemNo: integer; key, value: integer);
   procedure DeleteByPosition(ItemNo: integer);

  public
   constructor Create(size: integer=0);
   procedure SetSize(newSize: integer);
   function Find(key: Integer) : Integer;
   procedure Insert(key, value: integer);
   procedure Delete(key: integer);
 end;//TACRSortedPtrArray

 TACRSortedStringPtrArray = class(TList);



////////////////////////////////////////////////////////////////////////////////
//
// Compressed Stream Block Headers Array and related stuff
//
////////////////////////////////////////////////////////////////////////////////


 TACRBLOBDescriptor = packed record
  BlockSize:             Integer; // block size
  NumBlocks:             Int64;  // number of blocks
  UncompressedSize:      Int64; // size of uncompressed data
  StartPosition:         Int64; // position in stream where first block header is stored
  CompressionAlgorithm:  Byte; // compression algorithm
  CompressionMode:       Byte; // compression mode (0-9)
  Reserved:              array [0..1] of Byte;
 end; // TACRBLOBDescriptor - 32 bytes
 PACRBLOBDescriptor = ^TACRBLOBDescriptor;

 // blob descriptor that stores for each blob value
 TACRPartialBLOBDescriptor = packed record
  NumBlocks:             Integer; // number of blocks
  UncompressedSize:      Int64; // size of uncompressed data
 end; // TACRPartialBLOBDescriptor - 8 bytes
 PACRPartialBLOBDescriptor = ^TACRPartialBLOBDescriptor;

 // blob descriptor that stores for each blob value
 TACRPartialTemporaryBLOBDescriptor = packed record
  NumBlocks:             Integer; // number of blocks
  UncompressedSize:      Int64; // size of uncompressed data
  CompressedSize:        Int64; // size of compressed data
 end; // TACRPartialTemporaryBLOBDescriptor - 8 bytes
 PACRPartialTemporaryBLOBDescriptor = ^TACRPartialTemporaryBLOBDescriptor;


 TACRCompressedStreamBlockHeader = packed record
       CompressedSize:      Integer; // packed block size
       UncompressedSize:    Integer; // unpacked block size
       Crc32:               Cardinal; // check sum for this block page
       OffsetToNextHeader:  Int64;  // offset from beginning of the compressed stream
                                    // to next block header
 end; // 20

 TACRCompressedStreamBlockHeadersArray = class
    private
     AllocBy:             Int64;
     DeAllocBy:           Int64;
     MaxAllocBy:          Int64;
     AllocItemCount:      Int64;
    public
     Items:               array of TACRCompressedStreamBlockHeader;
     Positions:           array of Int64; // block positions
     ItemCount:           Int64; // all files quantity (including deleted files)
    public
     constructor Create;
     destructor Destroy; override;
     procedure SetSize(NewSize: Int64);
     procedure AppendItem(Value: TACRCompressedStreamBlockHeader; Pos: Int64);
     function FindPosition(Pos: Int64) : Integer;
   end; // TACRCompressedStreamBlockHeadersArray


////////////////////////////////////////////////////////////////////////////////
//
// TACRBitsArray
//
////////////////////////////////////////////////////////////////////////////////


  TACRBitsArray = class (TObject)
   private
    FBits:              PChar;
    FBitCount:          Integer; // total number of bits
    FNonZeroBitCount:   Integer; // number of bits equal to 1
    FBitsTable:         array [0..255] of byte;
   protected
    function GetNonZeroBitCount: Integer;
    procedure SetSize(NewSize: Integer);
   public
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    constructor Create;
    destructor Destroy; override;
    function GetBit(BitNo: Integer): Boolean;
    procedure SetBit(BitNo: Integer; Value: Boolean);
    // returns number of bit = 1 in FBits array by bit position
    function GetBitNoByBitPosition(BitPosition: Integer): Integer;
    // returns position of bit = 1 by bit no in FBits array
    function GetBitPositionByBitNo(BitNo: Integer): Integer;
    procedure SetAllBits;
    function Find(
                  Value:        Boolean;
                  var BitNo:    Integer
                 ): Boolean;
   public
    property Size: Integer read FBitCount write SetSize;
    property NonZeroBitCount: Integer read FNonZeroBitCount;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// Bits functions
//
////////////////////////////////////////////////////////////////////////////////


  // return true if null flag is set (bit = 1)
  function CheckNullFlag(
          BitNo:       Integer; // number of bit to check
          NullFlags:   PChar // pointer to bits of null flags
                      ): Boolean;

  // set or clear a null flag
  procedure SetNullFlag(
          ToSet:       Boolean; // if true - set bit = 1, otherwise set bit = 0
          BitNo:       Integer; // number of bit to check
          NullFlags:   PChar // pointer to bits of null flags
                     );


////////////////////////////////////////////////////////////////////////////////
//
// Other functions
//
////////////////////////////////////////////////////////////////////////////////


  // returns temporary name
  function GetTemporaryName(Prefix: String): String;
  
implementation

uses ACRExpressions, ACRCompression, Math;


////////////////////////////////////////////////////////////////////////////////
//
// TACRScanSearchCondition
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRScanSearchCondition.Create;
begin
  KeyRecordBuffer := nil;
  IndexID := INVALID_OBJECT_ID;
  Expression := nil;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRScanSearchCondition.Destroy;
begin
  inherited;
end;// Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRScanSearchCondition.Assign(ScanSearchCondition: TACRScanSearchCondition);
begin
  Condition := ScanSearchCondition.Condition;
  KeyRecordBuffer := ScanSearchCondition.KeyRecordBuffer;
  KeyFieldCount := ScanSearchCondition.KeyFieldCount;
  IndexID := ScanSearchCondition.IndexID;
  Expression := ScanSearchCondition.Expression;
end;// Assign



////////////////////////////////////////////////////////////////////////////////
//
// TACRScanSearchConditionArray
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRScanSearchConditionArray.Create;
begin
  Items := nil;
  Count := 0;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRScanSearchConditionArray.Destroy;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    Items[i].Free;
  SetLength(Items, 0);
end;// Destroy


//------------------------------------------------------------------------------
// Delete
//------------------------------------------------------------------------------
procedure TACRScanSearchConditionArray.Delete(ItemNo: Integer);
begin
 if (ItemNo < 0) or (ItemNo >= Count) then
  raise EACRException.Create(10413,ErrorLInvalidItemNumber);
 if (Items[ItemNo] = nil) then
  raise EACRException.Create(10414,ErrorLNilPointer);
 Items[ItemNo].Free; 
 if (ItemNo < Count - 1) then
  Move(Items[ItemNo+ 1 ],Items[ItemNo],
      (Count - ItemNo-1) * SizeOf(TACRScanSearchCondition));
 Dec(Count);
 SetLength(Items,Count);
end; // Delete


//------------------------------------------------------------------------------
// AddExpression
//------------------------------------------------------------------------------
procedure TACRScanSearchConditionArray.AddExpression(Expression: Pointer);
begin
  SetLength(Items, Count+1);
  Items[Count] := TACRScanSearchCondition.Create;
  Items[Count].Expression := Expression;
  Items[Count].Condition := scNone;
  Items[Count].KeyRecordBuffer := nil;
  Items[Count].KeyFieldCount := 0;
  Items[Count].IndexID := INVALID_OBJECT_ID;
  Inc(Count);
end;// AddExpression


//------------------------------------------------------------------------------
// AddCondition
//------------------------------------------------------------------------------
procedure TACRScanSearchConditionArray.AddCondition(
                  ScanSearchCondition:  TACRScanSearchCondition
                             );
begin
  if ((ScanSearchCondition.Expression = nil) and
      (ScanSearchCondition.IndexID = INVALID_ID4)) then
   raise EACRException.Create(10388,ErrorLInvalidScanConditionNoIndex,
      [ScanSearchCondition.IndexID,Count]);
  SetLength(Items, Count+1);
  Items[Count] := TACRScanSearchCondition.Create;
  Items[Count].Assign(ScanSearchCondition);
  Inc(Count);
end;// AddCondition


//------------------------------------------------------------------------------
// ExtractConditionsFromExpressions
//------------------------------------------------------------------------------
procedure TACRScanSearchConditionArray.ExtractConditionsFromExpressions(
                                     IndexDefs: Pointer;
                                     ExtractedConditionsInfo: TList
                                     );
var
  i: Integer;
begin
  for i := 0 to Count-1 do
   if (Items[i].Expression <> nil) then
    begin
      TACRExpression(Items[i].Expression).ExtractIndexScanConditions(
          IndexDefs,
          Self,
          ExtractedConditionsInfo
                                                                     );
    end;
end;// ExtractConditionsFromExpressions


//------------------------------------------------------------------------------
// ReturnConditionsToExpressions
//------------------------------------------------------------------------------
procedure TACRScanSearchConditionArray.ReturnConditionsToExpressions(ExtractedConditionsInfo: TList);
var
  i: Integer;
begin
  for i := 0 to ExtractedConditionsInfo.Count-1 do
   begin
    if (PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.Expression = nil) then
     raise EACRException.Create(20060, ErrorANilPointer);
    if (PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.
          ExtractedExpressionNode = nil) then
     raise EACRException.Create(20061, ErrorANilPointer);
    if (PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.KeyRecordBuffer = nil) then
     raise EACRException.Create(20062, ErrorANilPointer);

    TACRExpression(PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.
        Expression).AddNode(
        PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.
        ExtractedExpressionNode
                         );
    MemoryManager.FreeAndNilMem(
                              PACRExtractedConditionInfo(ExtractedConditionsInfo[i])^.
                              KeyRecordBuffer);

    Dispose(ExtractedConditionsInfo[i]);
   end;
end;// ReturnConditionsToExpressions



////////////////////////////////////////////////////////////////////////////////
//
// TACRIntegerArray
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Construct array of specified size
//------------------------------------------------------------------------------
constructor TACRIntegerArray.Create(
  size: Integer;
  DefaultAllocBy: Integer;
  MaximumAllocBy: Integer
  );
begin
 AllocBy := DefaultAllocBy; // default alloc
 deAllocBy := DefaultAllocBy; // default dealloc
 MaxAllocBy := MaximumAllocBy; // max alloc
 AllocItemCount := 0;
 SetSize(size);
end;//TACRIntegerArray.Create


//------------------------------------------------------------------------------
// Destruct array (free mem)
//------------------------------------------------------------------------------
destructor TACRIntegerArray.Destroy;
begin
 Items := nil;
 inherited Destroy;
end;//TACRIntegerArray.Destroy;


//------------------------------------------------------------------------------
// Set length of array to specified size
//------------------------------------------------------------------------------
procedure TACRIntegerArray.Assign(v: TACRIntegerArray);
var
  i: Integer;
begin
  SetSize(v.ItemCount);
  for i := 0 to ItemCount-1 do
    items[i] := v.items[i];
end;// Assign


//------------------------------------------------------------------------------
// Set length of array to specified size
//------------------------------------------------------------------------------
procedure TACRIntegerArray.SetSize(newSize: Integer);
begin
 if (newSize = 0) then
  begin
   ItemCount := 0;
   allocItemCount := 0;
   Items := nil;
   Exit;
  end;

 if (newSize > allocItemCount) then
  begin
     AllocBy := AllocBy * 2;
     if (AllocBy > MaxAllocBy) then
      AllocBy := MaxAllocBy;
     if (allocItemCount + AllocBy > newSize) then
      allocItemCount := allocItemCount + AllocBy
     else
      allocItemCount := newSize;
     SetLength(Items,allocItemCount);
  end
 else
  if (newSize < ItemCount) then
   if (allocItemCount-newSize > deAllocBy) then
    begin
     deAllocBy := deAllocBy * 2;
     if (deAllocBy > MaxAllocBy) then
      deAllocBy := MaxAllocBy;
     SetLength(Items,newSize);
     allocItemCount := newSize;
    end;

 ItemCount := newSize;
end;//TACRIntegerArray.SetSize


//------------------------------------------------------------------------------
// inserts an element to the end of items array
//------------------------------------------------------------------------------
procedure TACRIntegerArray.Append(value: Integer);
begin
 Inc(ItemCount);
 SetSize(itemCount);
 Items[itemCount-1] := value;
end;//TACRIntegerArray.Append


//------------------------------------------------------------------------------
// Remove first item = value
//------------------------------------------------------------------------------
procedure TACRIntegerArray.Remove(value: Integer);
var
  i, j: Integer;
begin
 j := -1;
 for i := 0 to ItemCount-1 do 
  if Items[i] = value then 
   begin
    j := i;
    break; 
   end; 
 if j > -1 then
   Delete(j);
end;//TACRIntegerArray.Remove 
	
 
//------------------------------------------------------------------------------
// Insert an element into specified position 
//------------------------------------------------------------------------------
procedure TACRIntegerArray.Insert(itemNo: Integer; value: Integer); 
begin
 inc(ItemCount); 
 SetSize(ItemCount); 
 if (itemCount <= 1) then
  items[0] := value 
 else 
 if (itemNo >= itemCount-1) 
  then 
   items[itemCount-1] := value 
  else
   begin
    Move(items[itemNo],items[itemNo+1],
        (itemCount - itemNo-1) * sizeOf(Integer));
    items[itemNo] := value;
   end; 
end;//TACRIntegerArray.Insert
 
	
//------------------------------------------------------------------------------ 
// Delete an element at specified position 
//------------------------------------------------------------------------------
procedure TACRIntegerArray.Delete(itemNo: Integer); 
begin
 if (itemNo < itemCount-1) then 
  Move(items[itemNo+1],items[itemNo], 
      (itemCount - itemNo-1) * sizeOf(Integer));
 dec(ItemCount);
 SetSize(ItemCount); 
end;//TACRIntegerArray.Delete 
 
 
//------------------------------------------------------------------------------ 
// moves element to new position 
//------------------------------------------------------------------------------ 
procedure TACRIntegerArray.MoveTo(itemNo, newItemNo: Integer);
var value : Integer;
begin 
 if (itemNo = newItemNo) then
  Exit;
 if (itemNo - newItemNo = 1) or (newItemNo-itemNo = 1) then 
  begin
   value := items[itemNo];
   items[itemNo] := items[newItemNo]; 
   items[newItemNo] := value; 
   Exit; 
  end;
 if (itemNo > newItemNo) then
  begin 
   value := items[itemNo];
   Move(PChar(items[newItemNo]),PChar(items[newItemNo+1]),
        (itemNo-newItemNo) * sizeof(Integer)); 
   items[newItemNo] := value; 
  end
 else
  begin
     value := items[ItemNo]; 
     Move(PChar(items[ItemNo+1]),PChar(items[ItemNo]), 
        (newItemNo-ItemNo-1) * sizeof(Integer)); 
     items[newItemNo-1] := value; 
  end;
end;// MoveTo(itemNo, newItemNo : Integer); 
 
 
//------------------------------------------------------------------------------
// copies itemCount elements to ar from ItmeNo
//------------------------------------------------------------------------------
procedure TACRIntegerArray.CopyTo( 
                      var ar: array of Integer;
                      itemNo, iCount: Integer
                             );
begin 
 if (itemCount > 0) then 
  Move (PChar(items[itemNo]),PChar(ar[0]),sizeOf(Integer)*iCount);
end;// CopyTo(ar : array of Integer; itemNo,itemCount : Integer);
	
	
//------------------------------------------------------------------------------
// returns true if value exists in Items array
//------------------------------------------------------------------------------ 
function TACRIntegerArray.IsValueExists(value: Integer): Boolean;
var i: Integer;
begin 
 Result := false;
 for i := 0 to ItemCount-1 do
  if Items[i] = value then 
   begin 
    Result := true; 
    break; 
   end;
end; // IsValueExists 
 
	
//------------------------------------------------------------------------------
// adds value if it is not existing, otherwise returns false
//------------------------------------------------------------------------------
function TACRIntegerArray.Add(value: Integer): Boolean; 
begin 
 Result := not (IsValueExists(value));
 if Result then
   Append(value); 
end; // Add 
 
	
	
////////////////////////////////////////////////////////////////////////////////
//
// TACRThreadIntArray
// 
////////////////////////////////////////////////////////////////////////////////
 
	
//------------------------------------------------------------------------------
// Construct array of specified size
//------------------------------------------------------------------------------
constructor TACRThreadIntArray.Create(
                                size: Integer;
                                DefaultAllocBy: Integer;
                                MaximumAllocBy: Integer
                                    );
begin
  InitializeCriticalSection(FCSect);
  inherited Create(size, DefaultAllocBy, MaximumAllocBy);
end;//TACRThreadIntArray.Create


//------------------------------------------------------------------------------
// Destruct array (free mem)
//------------------------------------------------------------------------------
destructor TACRThreadIntArray.Destroy;
begin
  inherited Destroy;
  DeleteCriticalSection(FCSect);
end;//TACRThreadIntArray.Destroy;


//------------------------------------------------------------------------------
// Lock
//------------------------------------------------------------------------------
procedure TACRThreadIntArray.Lock;
begin
  EnterCriticalSection(FCSect);
end;//TACRThreadIntArray.Lock;


//------------------------------------------------------------------------------
// Unlock
//------------------------------------------------------------------------------
procedure TACRThreadIntArray.Unlock;
begin
  LeaveCriticalSection(FCSect);
end;//TACRThreadIntArray.Unlock;



////////////////////////////////////////////////////////////////////////////////
//
// TACRRecordIDArray
//
////////////////////////////////////////////////////////////////////////////////



//------------------------------------------------------------------------------
// Construct array of specified size
//------------------------------------------------------------------------------
constructor TACRRecordIDArray.Create(
  Size:           Integer;
  DefaultAllocBy: Integer;
  MaximumAllocBy: Integer
  );
begin
 AllocBy := DefaultAllocBy; // default alloc
 deAllocBy := DefaultAllocBy; // default dealloc
 MaxAllocBy := MaximumAllocBy; // max alloc
 AllocItemCount := 0;
 SetSize(Size);
 FSorted := True;
end;//TACRRecordIDArray.Create


//------------------------------------------------------------------------------
// Destruct array (free mem)
//------------------------------------------------------------------------------
destructor TACRRecordIDArray.Destroy;
begin
 Items := nil;
 inherited Destroy;
end;//TACRRecordIDArray.Destroy;


//------------------------------------------------------------------------------
// Set length of array to specified size
//------------------------------------------------------------------------------
procedure TACRRecordIDArray.Assign(v: TACRRecordIDArray);
var
  i: Integer;
begin
  SetSize(v.ItemCount);
  for i := 0 to ItemCount-1 do
    items[i] := v.items[i];
end;// Assign


//------------------------------------------------------------------------------
// Set length of array to specified size
//------------------------------------------------------------------------------
procedure TACRRecordIDArray.SetSize(newSize: Integer);
begin
 if (newSize = 0) then
  begin
   ItemCount := 0;
   allocItemCount := 0;
   Items := nil;
   Exit;
  end;
 if (newSize > allocItemCount) then
  begin
   AllocBy := AllocBy * 2;
   if (AllocBy > MaxAllocBy) then
    AllocBy := MaxAllocBy;
   if (allocItemCount + AllocBy > newSize) then
    allocItemCount := allocItemCount + AllocBy
   else
    allocItemCount := newSize;
   SetLength(Items,allocItemCount);
  end
 else
  if (newSize < ItemCount) then
   if (allocItemCount-newSize > deAllocBy) then
    begin
     deAllocBy := deAllocBy * 2;
     if (deAllocBy > MaxAllocBy) then
      deAllocBy := MaxAllocBy;
     SetLength(Items,newSize);
     allocItemCount := newSize;
    end;
 ItemCount := newSize;
end;//TACRRecordIDArray.SetSize


//------------------------------------------------------------------------------
// inserts an element to the end of items array
//------------------------------------------------------------------------------
procedure TACRRecordIDArray.Append(var value: TACRRecordID);
begin
 Inc(ItemCount);
 SetSize(itemCount);
 Items[itemCount-1] := value;
end;//TACRRecordIDArray.Append


//------------------------------------------------------------------------------
// Insert an element into specified position
//------------------------------------------------------------------------------
procedure TACRRecordIDArray.Insert(itemNo: Integer; const value: TACRRecordID);
begin
 inc(ItemCount);
 SetSize(ItemCount);
 if (itemCount <= 1) then
  items[0] := value
 else
 if (itemNo >= itemCount-1)
  then
   items[itemCount-1] := value
  else
   begin
    Move(items[itemNo],items[itemNo+1],
        (itemCount - itemNo-1) * sizeOf(TACRRecordID));
    items[itemNo] := value;
   end;
end;//TACRRecordIDArray.Insert


//------------------------------------------------------------------------------
// Delete an element at specified position
//------------------------------------------------------------------------------
procedure TACRRecordIDArray.Delete(itemNo: Integer);
begin
 if (itemNo < itemCount-1) then
  Move(items[itemNo+1],items[itemNo],
      (itemCount - itemNo-1) * sizeOf(TACRRecordID));
 dec(ItemCount);
 SetSize(ItemCount);
end;//TACRRecordIDArray.Delete


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRRecordIDArray.CompareRecordIDMemory(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
begin
 if (RecordID1.PageNo = RecordID2.PageNo) then
  Result := 0
 else
 if (RecordID1.PageNo > RecordID2.PageNo) then
  Result := 1
 else
  Result := -1;
end; // CompareRecordIDMemory


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRRecordIDArray.CompareRecordIDTemporary(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
var id1,id2: TACRRecordNo;
begin
 id1 := 0;
 id2 := 0;
 Move(RecordID1,id1,sizeof(RecordID1));
 Move(RecordID2,id2,sizeof(RecordID2));
 if (id1 = id2) then
  Result := 0
 else
 if (id1 > id2) then
  Result := 1
 else
  Result := -1;
end; // CompareRecordIDTemporary


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRRecordIDArray.CompareRecordIDDisk(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
begin
 if ((RecordID1.PageNo = RecordID2.PageNo) and
     (RecordID1.PageItemNo = RecordID2.PageItemNo)) then
  Result := 0
 else
 if (RecordID1.PageNo > RecordID2.PageNo) then
  Result := 1
 else
 if (RecordID1.PageNo < RecordID2.PageNo) then
  Result := -1
 else
 if (RecordID1.PageItemNo > RecordID2.PageItemNo) then
  Result := 1
 else
  Result := -1
end; // CompareRecordIDDisk


//------------------------------------------------------------------------------
// return 0 if id1 = id2, 1 if id1>id2, -1 if id1<id2
//------------------------------------------------------------------------------
function TACRRecordIDArray.CompareRecordID(const RecordID1: TACRRecordID; const RecordID2: TACRRecordID): Integer;
begin
  if (FInMemory) then
   Result := CompareRecordIDMemory(RecordID1,RecordID2)
  else
  if (FTemporary) then
   Result := CompareRecordIDTemporary(RecordID1,RecordID2)
  else
   Result := CompareRecordIDDisk(RecordID1,RecordID2);
end; // CompareRecordID


//------------------------------------------------------------------------------
// return -1 if not found, otherwise return position in Items array [0..ItemCount-1]
//------------------------------------------------------------------------------
function TACRRecordIDArray.FindRecordByIDSorted(const RecordID: TACRRecordID): Integer;
var
    i,dx,f,oldRes: Integer;
    cmpRecBuf_res: Integer;
begin
  Result := -1;
  i := ItemCount shr 1;
  dx := i;
  if (ItemCount > 0) then
    begin
     f := 0;
     cmpRecBuf_res := 2;
     while (true) do
      begin
       dx := dx shr 1;
       if (dx < 1) then dx := 1;
       oldRes := cmpRecBuf_res;
       cmpRecBuf_res := -CompareRecordID(Items[i],RecordID);
       if (cmpRecBuf_res < 0) then
        begin
        //  element in index array for record buffer should be higher then current buffer (+->0)
         i := i - dx;
        end
       else
       if (cmpRecBuf_res > 0) then
        begin
        //  element in index array for record buffer should be lower then current buffer (0->+)
         i := i + dx;
        end
       else // index values are equal
        begin
         Result := i;
         break;
        end;
       if  (i < 0) and (dx = 1) then
        begin
         break;
        end;
       if  (i > ItemCount-1) and (dx = 1) then
        begin
         break;
        end;

       if  i > ItemCount-1 then
        i := ItemCount-1;
       if  i < 0 then
        i := 0;

       if (dx = 1) and (f > 1) then
        begin
         // dx minimum
         cmpRecBuf_res := -CompareRecordID(Items[i],RecordID);
         if (cmpRecBuf_res = oldRes) then
          continue;
         break;
        end;// last step
       if (cmpRecBuf_res <> oldRes) and (dx = 1) and (oldRes <> 2) then
        inc(f);
      end;//while dx
    end; // if reccount > 0
end; // FindRecordByIDSorted


//------------------------------------------------------------------------------
// sort array by record id - ascending
//------------------------------------------------------------------------------
procedure TACRRecordIDArray.Sort;
var
 aLo, aHi : Integer;

 procedure QuickSort (
                    var iLo, iHi : Integer
                    );
  var
    Lo, Hi:  Integer;
    T, Mid:  TACRRecordID;
  begin
    Lo := iLo;
    Hi := iHi;
    Mid := Items[(Lo + Hi) shr 1];
    repeat
     while (CompareRecordID(Items[Lo],Mid) < 0) and (Lo < iHi) do
      Inc(Lo);
     while (CompareRecordID(Items[Hi],Mid) > 0) and (Hi > 0) do
       Dec(Hi);
      if (Lo <= Hi) then
       begin
        T := Items[Lo];
        Items[Lo] := Items[Hi];
        Items[Hi] := T;
        Inc(Lo);
        Dec(Hi);
       end;
    until (Lo > Hi);
    if (Hi > iLo) then
     begin
      // check infinite recurse
      if (iHi = Hi) then
       raise EACRException.Create(11329,ErrorLErrorSoringRecordsByID,[Hi,Lo,iHi,iLo,ItemCount]);
      QuickSort(iLo, Hi);
     end;
    if (Lo < iHi) then
      QuickSort(Lo, iHi);
  end; //QuickSort
begin
  if (ItemCount > 1) then
   begin
    aLo := 0;
    aHi := ItemCount-1;
    QuickSort (aLo, aHi);
   end;
  FSorted := True;
end; // Sort


//------------------------------------------------------------------------------
// return -1 if not found, otherwise return position in Items array [0..ItemCount-1]
//------------------------------------------------------------------------------
function TACRRecordIDArray.FindRecordByID(const RecordID: TACRRecordID): Integer;
var i: Integer;
begin
  Result := -1;
  if (FSorted and (ItemCount > 10)) then
   begin
    Result := FindRecordByIDSorted(RecordID);
   end
  else
   begin
    for i := 0 to ItemCount-1 do
      if ((Items[i].PageNo = RecordID.PageNo) and (Items[i].PageItemNo = RecordID.PageItemNo)) then
       begin
        Result := i;
        break;
       end;
   end;
end; // FindRecordByID


////////////////////////////////////////////////////////////////////////////////
//
// TACRCompressedStreamBlockHeadersArray
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Construct array of specified size
//------------------------------------------------------------------------------
constructor TACRCompressedStreamBlockHeadersArray.Create;
begin
 AllocBy := 10; // default alloc
 DeAllocBy := 10; // default alloc
 MaxAllocBy := 10000; // max alloc
 AllocItemCount := 0;
 ItemCount := 0;
 SetSize(0);
end; // Create


//------------------------------------------------------------------------------
// Destruct array (free mem)
//------------------------------------------------------------------------------
destructor TACRCompressedStreamBlockHeadersArray.Destroy;
begin
 SetSize(0);
 inherited Destroy;
end;//Destroy;


//------------------------------------------------------------------------------
// Set length of array to specified size
//------------------------------------------------------------------------------
procedure TACRCompressedStreamBlockHeadersArray.SetSize(NewSize: Int64);
begin
 if (NewSize = 0) then
  begin
   ItemCount := 0;
   allocItemCount := 0;
   Items := nil;
   Positions := nil;
   Exit;
  end;

 if (NewSize > allocItemCount) then
  begin
     AllocBy := AllocBy * 2;
     if (AllocBy > MaxAllocBy) then
      AllocBy := MaxAllocBy;
     if (allocItemCount + AllocBy > NewSize) then
      allocItemCount := allocItemCount + AllocBy
     else
      allocItemCount := NewSize;
     SetLength(Items,allocItemCount);
     SetLength(Positions,allocItemCount);
  end
 else
  if (NewSize < ItemCount) then
   if (allocItemCount-NewSize > deAllocBy) then
    begin
     deAllocBy := deAllocBy * 2;
     if (deAllocBy > MaxAllocBy) then
      deAllocBy := MaxAllocBy;
     SetLength(Items,NewSize);
     SetLength(Positions,NewSize);
     allocItemCount := NewSize;
    end;
 ItemCount := NewSize;
end;// SetSize


//------------------------------------------------------------------------------
// Finds block containing specified position in user data
//------------------------------------------------------------------------------
function TACRCompressedStreamBlockHeadersArray.FindPosition(Pos: Int64) : Integer;
var i,dx,f,
    oldRes,res: Int64;

 function Compare: Integer;
 begin
  //---------------------------- start of compare -----------------------------------
       // by parent
       if (Positions[i] = pos) then
        Result := 0
       else
        if (Positions[i] < pos) then
         Result := 1
        else
         Result := -1;
  //---------------------------- end of compare -----------------------------------
 end;

begin

 i := ItemCount shr 1;
 dx := i;
 Result := 0;
 if (ItemCount <= 0) then
  begin
   Result := 0;
   Exit;
  end;
  f := 0;
  res := 2;
  while (true) do
   begin
    dx := dx shr 1;
    if (dx < 1) then dx := 1;
     oldRes := res;
     // compare, ascending
     res := Compare;
    if (res < 0) then
     begin
      //  element, specified by value should be higher then current element (+->0)
      i := i - dx;
     end
    else
    if (res > 0) then
     begin
      //  element, specified by value should be lower then current element (+->0)
      i := i + dx;
     end
    else
     begin
      // values are equal
      Result := i;
      break;
     end;
    if  (i < 0) and (dx = 1) then
     begin
      // equal not found
      Result := 0;
      break;
     end;
    if  (i > ItemCount-1) and (dx = 1) then
     begin
      // equal not found
      Result := ItemCount;
      break;
     end;

    if  (i > ItemCount-1) then
     i := ItemCount-1;
    if  i < 0 then
     i := 0;

    if (dx = 1) and (f > 1) then
     begin
      // dx minimum
      // compare, ascending
      res := Compare;
      if (res < 0) and (oldRes > 0) then
       Result := i;
      if (res > 0) and (oldRes < 0) then
       Result := i+1;
      if (res = oldRes) then
       continue;
      break;
     end;// last step
    if (res <> oldRes) and (dx = 1) and (oldRes <> 2) then
     inc(f);
  end;//while dx
 if (Result >= ItemCount) then
     Result := ItemCount-1;
 if (Result > 0) then
  if (Positions[Result] > pos) then
   dec(Result);
 if (Result < 0) then
   Result := 0;
end; //FindPosition


//------------------------------------------------------------------------------
// Insert an element into specified position
//------------------------------------------------------------------------------
procedure TACRCompressedStreamBlockHeadersArray.AppendItem(Value: TACRCompressedStreamBlockHeader; Pos: Int64);
begin
 Inc(ItemCount);
 SetSize(ItemCount);
 Items[ItemCount-1] := value;
 Positions[ItemCount-1] := pos;
end; // AppendItem


////////////////////////////////////////////////////////////////////////////////
//
// TACRSortedPtrArray
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Construct array of specified size
//------------------------------------------------------------------------------
constructor TACRSortedPtrArray.Create(size: integer);
begin
 AllocBy := 1000; // default alloc
 deAllocBy := 1000; // default dealloc
 MaxAllocBy := 10000; // max alloc
 AllocItemCount := 0;
 SetSize(size);
end;//Create


//------------------------------------------------------------------------------
// Delete an element by specified key
//------------------------------------------------------------------------------
procedure TACRSortedPtrArray.Delete(key: integer);
var pos : Integer;
begin
 if (itemCount <= 0) then
  raise Exception.Create('TaaSortedIntArray.Delete - no elements in array!');
 if (itemCount = 1) then
  DeleteByPosition(0)
 else
  begin
   pos := FindPosition(key);
   if (pos < 0) then
     raise Exception.Create('TaaSortedIntArray.Delete - element not found, key = '+
      InttoStr(key)+', itemCount = '+InttoStr(itemCount)+'!');
   DeleteByPosition(pos);
  end;
end;//Delete


//------------------------------------------------------------------------------
// Delete an element at specified position
//------------------------------------------------------------------------------
procedure TACRSortedPtrArray.DeleteByPosition(ItemNo: integer);
begin
 if (itemNo < itemCount-1) then
  begin
   Move(KeyItems[itemNo+1],KeyItems[itemNo],
       (itemCount - itemNo-1) * sizeOf(integer));
   Move(ValueItems[itemNo+1],ValueItems[itemNo],
       (itemCount - itemNo-1) * sizeOf(integer));
  end;
 dec(ItemCount);
 SetSize(ItemCount);
end;//DeleteByPosition


//------------------------------------------------------------------------------
// Finds value for specified key
// returns -1 if element was not found
//------------------------------------------------------------------------------
function TACRSortedPtrArray.Find(key: Integer): Integer;
begin
 Result := FindPositionForInsert(key);
 if (Result >= itemCount) or (Result < 0) then
  Result := -1
 else
  if (KeyItems[Result] <> key) then
   Result := -1
 else
  Result := ValueItems[Result];
end;//Find


//------------------------------------------------------------------------------
// Finds position for insert element
// returns -1 if element was not found
//------------------------------------------------------------------------------
function TACRSortedPtrArray.FindPosition(key: Integer): Integer;
begin
 Result := FindPositionForInsert(key);
 if (Result >= itemCount) or (Result < 0) then
  Result := -1
 else
  if (KeyItems[Result] <> key) then
   Result := -1;
end;//FindPosition


//------------------------------------------------------------------------------
// Finds position for insert element
//------------------------------------------------------------------------------
function TACRSortedPtrArray.FindPositionForInsert(key: Integer): Integer;
var i,dx,f,
    oldRes,res : Integer;
begin
 i := itemCount shr 1;
 dx := i;
 Result := itemCount;
 if (itemCount > 0) then
 begin
  f := 0;
  res := 2;
  while (true) do
   begin
    dx := dx shr 1;
    if (dx < 1) then dx := 1;
     oldRes := res;
     // compare, ascending
     if (KeyItems[i] = key) then
      res := 0
     else
      if (KeyItems[i] < key) then
       res := 1
      else
       res := -1;
    if (res < 0) then
     begin
      //  element, specified by value should be higher then current element (+->0)
      i := i - dx;
     end
    else
    if (res > 0) then
     begin
      //  element, specified by value should be lower then current element (+->0)
      i := i + dx;
     end
    else // values are equal
     begin
      Result := i;
      break;
     end;
    if  (i < 0) and (dx = 1) then
     begin
      Result := 0;
      break;
     end;
    if  (i > itemCount-1) and (dx = 1) then
     begin
      Result := itemCount;
      break;
     end;

    if  (i > itemCount-1) then
     i := itemCount-1;
    if  i < 0 then
     i := 0;

    if (dx = 1) and (f > 1) then
     begin
      // dx minimum
      // compare, ascending
      if (KeyItems[i] = key) then
       res := 0
      else
       if (KeyItems[i] < key) then
        res := 1
       else
        res := -1;

      if (res < 0) and (oldRes > 0) then
       Result := i;
      if (res > 0) and (oldRes < 0) then
       Result := i+1;
      if (res = oldRes) then
       continue;
      break;
     end;// last step
    if (res <> oldRes) and (dx = 1) and (oldRes <> 2) then
     inc(f);
  end;//while dx
 end; // if itemCount > 0
end;//FindPositionForInsert


//------------------------------------------------------------------------------
// Insert an element into specified position
//------------------------------------------------------------------------------
procedure TACRSortedPtrArray.Insert(key, value: integer);
var pos : Integer;
begin
 if (itemCount <= 0) then
  InsertByPosition(0,key,value)
 else
  if (itemCount = 1) then
   begin
    if (KeyItems[0] <= key) then
     InsertByPosition(1,key,value)
    else
     InsertByPosition(0,key,value);
   end
  else
   begin
    pos := FindPositionForInsert(key);
    InsertByPosition(pos,key,value);
   end;
end;//Insert


//------------------------------------------------------------------------------
// Insert an element into specified position
//------------------------------------------------------------------------------
procedure TACRSortedPtrArray.InsertByPosition(ItemNo, key, value: integer);
begin
 inc(ItemCount);
 SetSize(ItemCount);
 if (itemCount <= 1) then
  begin
   KeyItems[0] := key;
   ValueItems[0] := value;
  end
 else
 if (itemNo >= itemCount-1)
  then
   begin
    KeyItems[itemCount-1] := key;
    ValueItems[itemCount-1] := value;
   end
  else
   begin
    Move(KeyItems[itemNo],KeyItems[itemNo+1],
        (itemCount - itemNo-1) * sizeOf(integer));
    Move(ValueItems[itemNo],ValueItems[itemNo+1],
        (itemCount - itemNo-1) * sizeOf(integer));
    KeyItems[itemNo] := key;
    ValueItems[itemNo] := value;
   end;
end;//InsertByPosition


//------------------------------------------------------------------------------
// Set length of array to specified size
//------------------------------------------------------------------------------
procedure TACRSortedPtrArray.SetSize(newSize: integer);
begin
 if (newSize = 0) then
  begin
   ItemCount := 0;
   allocItemCount := 0;
   KeyItems := nil;
   ValueItems := nil;
   Exit;
  end;

 if (newSize > allocItemCount) then
  begin
     AllocBy := AllocBy * 2;
     if (AllocBy > MaxAllocBy) then
      AllocBy := MaxAllocBy;
     if (allocItemCount + AllocBy > newSize) then
      allocItemCount := allocItemCount + AllocBy
     else
      allocItemCount := newSize;
     SetLength(KeyItems,allocItemCount);
     SetLength(ValueItems,allocItemCount);
  end
 else
  if (newSize < ItemCount) then
   if (allocItemCount-newSize > deAllocBy) then
    begin
     deAllocBy := deAllocBy * 2;
     if (deAllocBy > MaxAllocBy) then
      deAllocBy := MaxAllocBy;
     SetLength(KeyItems,newSize);
     SetLength(ValueItems,newSize);
     allocItemCount := newSize;
    end;

 ItemCount := newSize;
end;//SetSize


////////////////////////////////////////////////////////////////////////////////
//
// TACRBitsArray
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return number of bits = 1
//------------------------------------------------------------------------------
function TACRBitsArray.GetNonZeroBitCount: Integer;
var i: Integer;
begin
 Result := 0;
 if (FBitCount > 0) then
  for i := 0 to FBitCount - 1 do
   if (GetBit(i)) then
    Inc(Result);
end; // GetNonZeroBitCount


//------------------------------------------------------------------------------
// set new size
//------------------------------------------------------------------------------
procedure TACRBitsArray.SetSize(NewSize: Integer);
var
  SizeInBytes: Integer;
begin
  SizeInBytes := (NewSize div 8) + Integer((NewSize mod 8) > 0);
  if (NewSize = 0) then
   begin
    if (FBits <> nil) then
     MemoryManager.FreeAndNilMem(FBits);
    FNonZeroBitCount := 0;
   end // empty array
  else
   begin
    if (FBits = nil) then
     FBits := MemoryManager.AllocMem(SizeInBytes)
    else
     MemoryManager.ReallocMemAndClearTail(FBits, SizeInBytes);
   end;// not empty array
  if (NewSize < FBitCount) then
   begin
    if (NewSize = 0) then
     FNonZeroBitCount := 0
    else
     begin
      FBitCount := NewSize;
      FNonZeroBitCount := GetNonZeroBitCount;
     end;
   end;
  FBitCount := NewSize;
end;// SetSize


//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRBitsArray.LoadFromStream(Stream: TStream);
var aBitsSize: Integer;
begin
 LoadDataFromStream(FBitCount,Sizeof(FBitCount),Stream,10401);
 LoadDataFromStream(FNonZeroBitCount,Sizeof(FNonZeroBitCount),Stream,10426);
 SetSize(FBitCount);
 if (FBitCount > 0) then
  begin
   aBitsSize := FBitCount div 8;
   if ((FBitCount mod 8) > 0) then
    Inc(aBitsSize);
   LoadDataFromStream(FBits^,aBitsSize,Stream,10402);
  end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRBitsArray.SaveToStream(Stream: TStream);
var aBitsSize: Integer;
begin
 SaveDataToStream(FBitCount,Sizeof(FBitCount),Stream,10398);
 SaveDataToStream(FNonZeroBitCount,Sizeof(FNonZeroBitCount),Stream,10427);
 if (FBitCount > 0) then
  begin
   if (FBits = nil) then
    raise EACRException.Create(10399,ErrorLNilPointer);
   aBitsSize := FBitCount div 8;
   if ((FBitCount mod 8) > 0) then
    Inc(aBitsSize);
   SaveDataToStream(FBits^,aBitsSize,Stream,10400);
  end;
end; // SaveToStream


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRBitsArray.Create;
var
    i,c: Byte;
begin
 FBits := nil;
 SetSize(0);
 for i := 0 to 255 do
  begin
   c := 0;
   if ((i and 1) <> 0) then Inc(c);
   if ((i and 2) <> 0) then Inc(c);
   if ((i and 4) <> 0) then Inc(c);
   if ((i and 8) <> 0) then Inc(c);
   if ((i and 16) <> 0) then Inc(c);
   if ((i and 32) <> 0) then Inc(c);
   if ((i and 64) <> 0) then Inc(c);
   if ((i and 128) <> 0) then Inc(c);
   FBitsTable[i] := c;
  end;
end;// Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRBitsArray.Destroy;
begin
 if (FBits <> nil) then
  MemoryManager.FreeAndNilMem(FBits);
end;// Destroy


//------------------------------------------------------------------------------
// get bit value
//------------------------------------------------------------------------------
function TACRBitsArray.GetBit(BitNo: Integer): Boolean;
begin
  if (BitNo >= FBitCount) then
   raise EACRException.Create(10394,ErrorLInvalidBitNo,[BitNo,FBitCount]);
  Result := CheckNullFlag(BitNo, FBits);
end;// GetBit


//------------------------------------------------------------------------------
// set bit value
//------------------------------------------------------------------------------
procedure TACRBitsArray.SetBit(BitNo: Integer; Value: Boolean);
var Bit: Boolean;
begin
  if (BitNo >= FBitCount) then
   raise EACRException.Create(10395,ErrorLInvalidBitNo,[BitNo,FBitCount]);
  Bit := CheckNullFlag(BitNo,FBits);
  if (Bit <> Value) then
   begin
    if (Value) then
     Inc(FNonZeroBitCount)
    else
     Dec(FNonZeroBitCount);
    SetNullFlag(Value, BitNo, FBits);
   end;
end;// SetBit


//------------------------------------------------------------------------------
// returns number of bit = 1 in FBits array by bit position
//------------------------------------------------------------------------------
function TACRBitsArray.GetBitNoByBitPosition(BitPosition: Integer): Integer;
var i,n:    Integer;
    b,k,l:  Byte;
begin
 if (BitPosition >= FNonZeroBitCount) then
  raise EACRException.Create(10428,ErrorLInvalidBitNo,
    [BitPosition,FNonZeroBitCount]);

 if (FBitCount = FNonZeroBitCount) then
  begin
   Result := BitPosition;
   Exit;
  end;


 i := 0; // byte number
 n := 0; // bits count
 while (n <= BitPosition) do
  begin
   b := PByte(FBits + i)^;
   if (n + FBitsTable[b] > BitPosition) then
    break;
   Inc(n,FBitsTable[b]);
   Inc(i);
  end;
 Result := i * 8;
 if (Result > FBitCount) then
  raise EACRException.Create(10456,ErrorLInvalidBitNo,[BitPosition,FBitCount]);
 b := PByte(FBits + i)^;
 l := 7;
 if (i = (FBitCount div 8)) then
  l := (FBitCount mod 8)-1;
 for k := 0 to l do
  begin
   if ((b and (1 shl k)) <> 0) then Inc(n);
   if (n > BitPosition) then Break;
//   if (n >= BitPosition) then Break;
   Inc(Result);
  end;
end; // GetBitNoByBitPosition


//------------------------------------------------------------------------------
// returns position of bit = 1 by bit no in FBits array
//------------------------------------------------------------------------------
function TACRBitsArray.GetBitPositionByBitNo(BitNo: Integer): Integer;
var i,j:    Integer;
    b,k:      Byte;
begin
 if (BitNo >= FBitCount) then
  raise EACRException.Create(10429,ErrorLInvalidBitNo,
    [BitNo,FBitCount]);
 if (FBitCount = FNonZeroBitCount) then
  begin
   Result := BitNo;
   Exit;
  end;
 Result := 0;
 i := BitNo div 8;
 for j := 0 to i-1 do
  begin
   b := PByte(FBits + j)^;
   Inc(Result,FBitsTable[b]);
  end;
 b := PByte(FBits + i)^;
 j := BitNo mod 8;
 for k := 0 to j do
  if ((b and (1 shl k)) <> 0) then
   Inc(Result);
 if (Result > 0) then
  Dec(Result);
{
 // number of byte with flags
 i := Integer(BitNo) div 8;
 Result := 0; // bits count
 if (i > 0) then
   for j := 0 to i-1 do
     for k := 7 downto 0 do
      begin
       b := PByte(FBits + j)^;
       if ((b and (1 shl k)) <> 0) then Inc(Result);
      end;
 // scan last byte
 b := PByte(FBits + i)^;
 n := Integer(BitNo) mod 8;
 for k := 0 to n do
   if ((b and (1 shl k)) <> 0) then Inc(Result);
}
end; // GetBitPositionByBitNo


//------------------------------------------------------------------------------
// set all bits to 1
//------------------------------------------------------------------------------
procedure TACRBitsArray.SetAllBits;
var
    SizeInBytes: Integer;
begin
 if (FBitCount > 0) then
  begin
    SizeInBytes := (FBitCount div 8) + Integer((FBitCount mod 8) > 0);
    FillChar(FBits^,SizeInBytes,$FF);
  end;
  FNonZeroBitCount := FBitCount;
end; // SetAllBits


//------------------------------------------------------------------------------
// find any bit with specified value
//------------------------------------------------------------------------------
function TACRBitsArray.Find(
                  Value:        Boolean;
                  var BitNo:    Integer
                           ): Boolean;
var
  i: Integer;
begin
  Result := False;
  if (FBitCount > 0) then
   for i := 0 to FBitCount-1 do
    if (GetBit(i) = Value) then
     begin
       Result := True;
       BitNo := i;
       break;
     end
end;// Find


////////////////////////////////////////////////////////////////////////////////
//
// Bits functions
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// return true if null flag is set (bit = 1)
//------------------------------------------------------------------------------
function CheckNullFlag(
          BitNo:       Integer; // number of bit to check
          NullFlags:   PChar // pointer to bits of null flags
                      ): Boolean;
begin
 if ((PByte(NullFlags+(BitNo div 8))^) and
		 (1 shl (BitNo mod 8)) <> 0) then
  Result := True
 else
  Result := False;
end; // CheckNullFlag


//------------------------------------------------------------------------------
// set or clear a null flag
//------------------------------------------------------------------------------
procedure SetNullFlag(
          ToSet:       Boolean; // if true - set bit = 1, otherwise set bit = 0
          BitNo:       Integer; // number of bit to check
          NullFlags:   PChar // pointer to bits of null flags
                     );
begin
  NullFlags := PChar(NullFlags + (BitNo div 8));
  if (ToSet) then
    PByte(NullFlags)^ := PByte(NullFlags)^ or
		(1 shl (BitNo mod 8))
  else
    PByte(NullFlags)^ := PByte(NullFlags)^ and
		(not (1 shl (BitNo mod 8)));
end; // SetNullFlag


function GetTemporaryName(Prefix: String): String;
var x: Integer;
begin
 x := Random(MAXINT);
 Result := Prefix + IntToStr(x);
end; // GetTemporaryName

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRTypes> initialized');
{$ENDIF}

end.
