unit ACRTypesRoutines;

interface

uses
     ACRTypes,
     ACRConverts,
     ACRMemory,
     ACRExcept,
     ACRConst;


//  Return DataSize
function GetDataSizeInMemory(DataType: TACRBaseFieldType; Length: Integer = 0): Integer; overload;

//  Return DataSize
function GetDataSizeInMemory(DataType: TACRBaseFieldType; Data: Pointer): Integer; overload;

// Allocate Memory and Set Data
procedure AllocMemAndSetData(DataType: TACRBaseFieldType; const Value; out Buffer; out BufferSize: Integer);


implementation


//------------------------------------------------------------------------------
//  Return DataSize
//------------------------------------------------------------------------------
function GetDataSizeInMemory(DataType: TACRBaseFieldType; Length: Integer): Integer;
begin
  case DataType of
    bftChar,            // AsciZ
    bftVarchar:         Result := Length + 1;
    bftWideChar,
    bftWideVarchar:     Result := (Length + 1) * 2;

    bftSignedInt8:      Result := SizeOf(Shortint);
    bftSignedInt16:     Result := SizeOf(Smallint);
    bftSignedInt32:     Result := SizeOf(Integer);
    bftSignedInt64:     Result := SizeOf(Int64);
    bftUnSignedInt8:    Result := SizeOf(Byte);
    bftUnSignedInt16:   Result := SizeOf(Word);
    bftUnSignedInt32:   Result := SizeOf(Cardinal);

    bftSingle:          Result := SizeOf(Single);
    bftDouble:          Result := SizeOf(Double);
    bftExtended:        Result := SizeOf(Extended);

    bftDate:            Result := SizeOf(TACRDate);
    bftTime:            Result := SizeOf(TACRTime);
    bftDateTime:        Result := SizeOf(TACRDateTime);

    bftBlob,
    bftClob,
    bftWideClob:        Result := SizeOf(Int64); // 8

    bftLogical:         Result := SizeOf(TACRLogical);
    bftCurrency:        Result := SizeOf(TACRCurrency);
    bftBytes:           Result := Length;
    bftVarBytes:        Result := Length;
    else
      raise EACRException.Create(30050, ErrorGUnknownDataType, [BftToStr(DataType)]);
  end;
end;//GetDataSizeInMemory


//------------------------------------------------------------------------------
//  Return DataSize
//------------------------------------------------------------------------------
function GetDataSizeInMemory(DataType: TACRBaseFieldType; Data: Pointer): Integer; overload;
begin
//  Result := -1;
  case DataType of
    bftChar,
    bftVarchar:
        Result := GetDataSizeInMemory(DataType, Length(PChar(Data)));
    bftWideChar,
    bftWideVarchar:
        Result := GetDataSizeInMemory(DataType, Length(PWideChar(Data)));
    else
        Result := GetDataSizeInMemory(DataType);
  end;
end;//GetDataSizeInMemory


//------------------------------------------------------------------------------
// Allocate Memory and Set Data
//------------------------------------------------------------------------------
procedure AllocMemAndSetData(DataType: TACRBaseFieldType;
                             const Value;
                             out Buffer;
                             out BufferSize: Integer);
var len: Integer;
begin
  Pointer(Buffer) := nil;
  try
    case DataType of
      bftChar,
      bftVarchar:
        begin
          len := Length(PChar(Value));
          if (len = 0) then
           begin
            BufferSize := 0;
           end
          else
           begin
            BufferSize := GetDataSizeInMemory(DataType,len);
            Pointer(Buffer) := MemoryManager.GetMem(BufferSize);
            //Move(PChar(Value)^, PChar(Buffer)^, BufferSize);
            Move(PChar(String(Value))^, PChar(Buffer)^, BufferSize);
           end;
        end;
      bftWideChar,
      bftWideVarchar:
        begin
          len := Length(PWideChar(Value));
          if (len = 0) then
           begin
            BufferSize := 0;
           end
          else
           begin
            BufferSize := GetDataSizeInMemory(DataType,len);
            Pointer(Buffer) := MemoryManager.GetMem(BufferSize);
            Move(PWideChar(Value)^, PChar(Buffer)^, BufferSize);
           end;
        end;
      else
        begin
          BufferSize := GetDataSizeInMemory(DataType);
          Pointer(Buffer) := MemoryManager.GetMem(BufferSize);
          case BufferSize of
            1:      PByte(Buffer)^ := Byte(Value);
            2:      PWord(Buffer)^ := Word(Value);
            4:      PInteger(Buffer)^ := Integer(Value);
            8:      PInt64(Buffer)^ := Int64(Value);
            10:     PExtended(Buffer)^ := Extended(Value);
            else    Move(Value, PChar(Buffer)^, BufferSize);
          end;
        end;
    end;
  except
    if Pointer(Buffer) <> nil then
      raise EACRException.Create(30054, ErrorGCannotCopyDataIntoNewMemory);
  end;
end;//AllocMemAndSetData

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRTypesRoutines> initialized');
{$ENDIF}

end.

