unit ACRConverts;

interface

{$I ACRVer.inc}

uses SysUtils,
     DB,
{$IFDEF MSWINDOWS}
     Controls,
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Libc,
{$ENDIF}
     ACRTypes,
     ACRStrUtils,
     ACRDecUtil,
     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRExcept,
     ACRMemory,
     ACRConst;

const
  WildCardMultipleChar = '%';
  WildCardSingleChar = '_';

type

{$IFDEF LINUX}
  // Delphi7 Controls.pas
  TDate = type TDateTime;
  TTime = type TDateTime;
{$ENDIF}

  TSQLFieldType = packed record
    SqlName: String;
    AdvancedFieldType: TACRAdvancedFieldType;
    BaseFieldType: TACRBaseFieldType;
    FieldType: TFieldType;
  end;

  TACRFieldTypeDesc = packed record
    Name: String;
    AdvancedFieldType: TACRAdvancedFieldType;
  end;

const ACRFieldTypes: array[1..35] of TACRFieldTypeDesc = (
(Name: 'Char'; AdvancedFieldType:     aftChar),
(Name: 'Varchar'; AdvancedFieldType:     aftString),

(Name: 'WideChar'; AdvancedFieldType:     aftWideChar),
(Name: 'WideVarchar'; AdvancedFieldType:     aftWideString),

(Name: 'ShortInt'; AdvancedFieldType:     aftShortint),
(Name: 'SmallInt'; AdvancedFieldType:     aftSmallint),
(Name: 'Integer'; AdvancedFieldType:     aftInteger),
(Name: 'LargeInt'; AdvancedFieldType:     aftLargeint),
(Name: 'Byte'; AdvancedFieldType:     aftByte),
(Name: 'Word'; AdvancedFieldType:     aftWord),
(Name: 'Cardinal'; AdvancedFieldType:     aftCardinal),

(Name: 'AutoInc'; AdvancedFieldType:     aftAutoInc),
(Name: 'AutoIncShortInt'; AdvancedFieldType:     aftAutoIncShortint),
(Name: 'AutoIncSmallInt'; AdvancedFieldType:     aftAutoIncSmallint),
(Name: 'AutoIncInteger'; AdvancedFieldType:     aftAutoIncInteger),
(Name: 'AutoIncLargeInt'; AdvancedFieldType:     aftAutoIncLargeint),
(Name: 'AutoIncByte'; AdvancedFieldType:     aftAutoIncByte),
(Name: 'AutoIncWord'; AdvancedFieldType:     aftAutoIncWord),
(Name: 'AutoIncCardinal'; AdvancedFieldType:     aftAutoIncCardinal),

(Name: 'Single'; AdvancedFieldType:     aftSingle),
(Name: 'Double'; AdvancedFieldType:     aftDouble),
(Name: 'Extended'; AdvancedFieldType:     aftExtended),

(Name: 'Boolean'; AdvancedFieldType:     aftBoolean),

(Name: 'Currency'; AdvancedFieldType:     aftCurrency),

(Name: 'Date'; AdvancedFieldType:     aftDate),
(Name: 'Time'; AdvancedFieldType:     aftTime),
(Name: 'DateTime'; AdvancedFieldType:     aftDateTime),
(Name: 'TimeStamp'; AdvancedFieldType:     aftTimeStamp),

(Name: 'Memo'; AdvancedFieldType:     aftMemo),
(Name: 'WideMemo'; AdvancedFieldType:     aftWideMemo),
(Name: 'FormattedMemo'; AdvancedFieldType:     aftFormattedMemo),
(Name: 'BLOB'; AdvancedFieldType:     aftBlob),
(Name: 'Graphic'; AdvancedFieldType:     aftGraphic),

(Name: 'Bytes'; AdvancedFieldType:     aftBytes),
(Name: 'VarBytes'; AdvancedFieldType:     aftVarBytes)
);

const
  SQLFieldTypes: array[1..58] of TSQLFieldType = (

    (SqlName: 'CHAR';           AdvancedFieldType: aftChar;           BaseFieldType: bftChar;         FieldType: ftFixedChar),
    (SqlName: 'FIXEDCHAR';      AdvancedFieldType: aftChar;           BaseFieldType: bftChar;         FieldType: ftFixedChar),
    (SqlName: 'STRING';         AdvancedFieldType: aftString;         BaseFieldType: bftVarchar;      FieldType: ftString),
    (SqlName: 'VARCHAR';        AdvancedFieldType: aftString;         BaseFieldType: bftVarchar;      FieldType: ftString),
    (SqlName: 'VARCHAR2';       AdvancedFieldType: aftString;         BaseFieldType: bftVarchar;      FieldType: ftString),
    (SqlName: 'WIDECHAR';       AdvancedFieldType: aftWideChar;       BaseFieldType: bftWideChar;     FieldType: ftWideString),
    (SqlName: 'FIXEDWIDECHAR';  AdvancedFieldType: aftWideChar;       BaseFieldType: bftWideChar;     FieldType: ftWideString),
    (SqlName: 'WIDESTRING';     AdvancedFieldType: aftWideString;     BaseFieldType: bftWideVarchar;  FieldType: ftWideString),
    (SqlName: 'WIDEVARCHAR';    AdvancedFieldType: aftWideString;     BaseFieldType: bftWideVarchar;  FieldType: ftWideString),

    (SqlName: 'SIGNEDINT16';    AdvancedFieldType: aftSmallint;       BaseFieldType: bftSignedInt16;  FieldType: ftSmallint),
    (SqlName: 'SMALLINT';       AdvancedFieldType: aftSmallint;       BaseFieldType: bftSignedInt16;  FieldType: ftSmallint),
    (SqlName: 'SIGNEDINT8';     AdvancedFieldType: aftShortint;       BaseFieldType: bftSignedInt8;   FieldType: ftSmallint),
    (SqlName: 'SHORTINT';       AdvancedFieldType: aftShortint;       BaseFieldType: bftSignedInt8;   FieldType: ftSmallint),
    (SqlName: 'SIGNEDINT32';    AdvancedFieldType: aftInteger;        BaseFieldType: bftSignedInt32;  FieldType: ftInteger),
    (SqlName: 'INTEGER';        AdvancedFieldType: aftInteger;        BaseFieldType: bftSignedInt32;  FieldType: ftInteger),
    (SqlName: 'SIGNEDINT64';    AdvancedFieldType: aftLargeint;       BaseFieldType: bftSignedInt64;  FieldType: ftLargeint),
    (SqlName: 'LARGEINT';       AdvancedFieldType: aftLargeint;       BaseFieldType: bftSignedInt64;  FieldType: ftLargeint),
    (SqlName: 'INT64';          AdvancedFieldType: aftLargeint;       BaseFieldType: bftSignedInt64;  FieldType: ftLargeint),
    (SqlName: 'UNSIGNEDINT16';  AdvancedFieldType: aftWord;           BaseFieldType: bftUnsignedInt16;FieldType: ftWord),
    (SqlName: 'WORD';           AdvancedFieldType: aftWord;           BaseFieldType: bftUnsignedInt16;FieldType: ftWord),
    (SqlName: 'UNSIGNEDINT8';   AdvancedFieldType: aftByte;           BaseFieldType: bftUnsignedInt8; FieldType: ftWord),
    (SqlName: 'BYTE';           AdvancedFieldType: aftByte;           BaseFieldType: bftUnsignedInt8; FieldType: ftWord),
    (SqlName: 'UNSIGNEDINT32';  AdvancedFieldType: aftCardinal;       BaseFieldType: bftUnsignedInt32;FieldType: ftLargeint),
    (SqlName: 'CARDINAL';       AdvancedFieldType: aftCardinal;       BaseFieldType: bftUnsignedInt32;FieldType: ftLargeint),

    (SqlName: 'AUTOINC';        AdvancedFieldType: aftAutoInc;        BaseFieldType: bftSignedInt32;  FieldType: ftAutoInc),
    (SqlName: 'AUTOINCSHORTINT';AdvancedFieldType: aftAutoIncShortint;BaseFieldType: bftSignedInt8;   FieldType: ftAutoInc),
    (SqlName: 'AUTOINCSMALLINT';AdvancedFieldType: aftAutoIncSmallint;BaseFieldType: bftSignedInt16;  FieldType: ftAutoInc),
    (SqlName: 'AUTOINCINTEGER'; AdvancedFieldType: aftAutoIncInteger; BaseFieldType: bftSignedInt32;  FieldType: ftAutoInc),
    (SqlName: 'AUTOINCLARGEINT';AdvancedFieldType: aftAutoIncLargeint;BaseFieldType: bftSignedInt64;  FieldType: ftLargeInt),
    (SqlName: 'AUTOINCBYTE';    AdvancedFieldType: aftAutoIncByte;    BaseFieldType: bftUnsignedInt8; FieldType: ftAutoInc),
    (SqlName: 'AUTOINCWORD';    AdvancedFieldType: aftAutoIncWord;    BaseFieldType: bftUnsignedInt16;FieldType: ftAutoInc),
    (SqlName: 'AUTOINCCARDINAL';AdvancedFieldType: aftAutoIncCardinal;BaseFieldType: bftUnsignedInt32;FieldType: ftAutoInc),

    (SqlName: 'FLOAT';          AdvancedFieldType: aftDouble;         BaseFieldType: bftDouble;       FieldType: ftFloat),
    (SqlName: 'SINGLE';         AdvancedFieldType: aftSingle;         BaseFieldType: bftSingle;       FieldType: ftFloat),
    (SqlName: 'DOUBLE';         AdvancedFieldType: aftDouble;         BaseFieldType: bftDouble;       FieldType: ftFloat),
    (SqlName: 'EXTENDED';       AdvancedFieldType: aftExtended;       BaseFieldType: bftExtended;     FieldType: ftFloat),

    (SqlName: 'LOGICAL';        AdvancedFieldType: aftBoolean;        BaseFieldType: bftLogical;      FieldType: ftBoolean),
    (SqlName: 'BOOLEAN';        AdvancedFieldType: aftBoolean;        BaseFieldType: bftLogical;      FieldType: ftBoolean),
    (SqlName: 'BOOL';           AdvancedFieldType: aftBoolean;        BaseFieldType: bftLogical;      FieldType: ftBoolean),
    (SqlName: 'BIT';            AdvancedFieldType: aftBoolean;        BaseFieldType: bftLogical;      FieldType: ftBoolean),

    (SqlName: 'CURRENCY';        AdvancedFieldType: aftCurrency;      BaseFieldType: bftCurrency;     FieldType: ftCurrency),
    (SqlName: 'MONEY';          AdvancedFieldType: aftCurrency;      BaseFieldType: bftCurrency;     FieldType: ftCurrency),

    (SqlName: 'DATE';            AdvancedFieldType: aftDate;          BaseFieldType: bftDate;         FieldType: ftDate),
    (SqlName: 'TIME';            AdvancedFieldType: aftTime;          BaseFieldType: bftTime;         FieldType: ftTime),
    (SqlName: 'DATETIME';        AdvancedFieldType: aftDateTime;      BaseFieldType: bftDateTime;     FieldType: ftDateTime),

{$IFDEF D6H}
    (SqlName: 'TIMESTAMP';       AdvancedFieldType: aftTimeStamp;     BaseFieldType: bftDateTime;     FieldType: ftTimeStamp),
{$ELSE}
    (SqlName: 'TIMESTAMP';       AdvancedFieldType: aftTimeStamp;     BaseFieldType: bftDateTime;     FieldType: ftUnknown),
{$ENDIF}

    (SqlName: 'BYTES';           AdvancedFieldType: aftBytes;         BaseFieldType: bftBytes;        FieldType: ftBytes),
    (SqlName: 'VARBYTES';        AdvancedFieldType: aftVarBytes;      BaseFieldType: bftVarBytes;     FieldType: ftVarBytes),

    (SqlName: 'BLOB';            AdvancedFieldType: aftBlob;          BaseFieldType: bftBlob;         FieldType: ftBlob),
    (SqlName: 'BLOB';            AdvancedFieldType: aftBlob;          BaseFieldType: bftBlob;         FieldType: ftParadoxOle),
    (SqlName: 'BLOB';            AdvancedFieldType: aftBlob;          BaseFieldType: bftBlob;         FieldType: ftTypedBinary),
    (SqlName: 'BLOB';            AdvancedFieldType: aftBlob;          BaseFieldType: bftBlob;         FieldType: ftDBaseOle),
    (SqlName: 'GRAPHIC';         AdvancedFieldType: aftGraphic;       BaseFieldType: bftBlob;         FieldType: ftGraphic),
    (SqlName: 'MEMO';            AdvancedFieldType: aftMemo;          BaseFieldType: bftClob;         FieldType: ftMemo),
    (SqlName: 'CLOB';            AdvancedFieldType: aftMemo;          BaseFieldType: bftClob;         FieldType: ftMemo),
    (SqlName: 'FORMATTEDMEMO';   AdvancedFieldType: aftFormattedMemo; BaseFieldType: bftClob;         FieldType: ftFmtMemo),
    (SqlName: 'WIDEMEMO';        AdvancedFieldType: aftWideMemo;      BaseFieldType: bftWideClob;     FieldType: ftMemo),
    (SqlName: 'WIDECLOB';        AdvancedFieldType: aftWideMemo;      BaseFieldType: bftWideClob;     FieldType: ftMemo)
  );

  // Convert AdvancedFieldType to BasicFieldType
  function AdvancedFieldTypeToBaseFieldType(AdvancedFieldType: TACRAdvancedFieldType): TACRBaseFieldType;
  // Convert BaseFieldType to AdvancedFieldType
  function BaseFieldTypeToAdvancedFieldType(BaseFieldType: TACRBaseFieldType): TACRAdvancedFieldType;

  // Converts TFieldType to TACRAdvancedFieldType
  function FieldTypeToACRAdvFieldType(FieldType: TFieldType): TACRAdvancedFieldType;
  // Converts TACRAdvancedFieldType to TFieldType
  function ACRAdvFieldTypeToFieldType(AdvancedFieldType: TACRAdvancedFieldType): TFieldType;

  // AdvancedFieldType for print
  function AftToStr(AdvancedFieldType: TACRAdvancedFieldType): String;
  // BaseFieldType for print
  function BftToStr(BaseFieldType: TACRBaseFieldType): String;


  // Convert ACRDate to Date
  function ACRDateToDate(ACRDate: TACRDate): TDate; overload;
  // Convert Date to ACRDate
  function DateToACRDate(Date: TDate): TACRDate; overload;

  // Convert ACRTime to Time
  function ACRTimeToTime(ACRTime: TACRTime): TTime; overload;
  // Convert Time to ACRTime
  function TimeToACRTime(Time: TTime): TACRTime; overload;

  // Convert ACRDateTime to DateTime
  function ACRDateTimeToDateTime(ACRDateTime: TACRDateTime): TDateTime; overload;
  // Convert DateTime to ACRDateTime
  function DateTimeToACRDateTime(DateTime: TDateTime): TACRDateTime; overload;

  // return true if field type is a BLOB field type
  function IsBLOBFieldType(FieldType: TACRBaseFieldType): Boolean; overload;
  function IsBLOBFieldType(FieldType: TACRAdvancedFieldType): Boolean; overload;

  // return true if field type is a varchar field type
  function IsVarcharFieldType(FieldType: TACRBaseFieldType): Boolean; overload;
  function IsVarcharFieldType(FieldType: TACRAdvancedFieldType): Boolean; overload;

  // return true if field type is a string field type, but not wide string
  function IsStringFieldType(FieldType: TACRBaseFieldType): Boolean; overload;
  function IsStringFieldType(FieldType: TACRAdvancedFieldType): Boolean; overload;

  // return true if field type is a wide string field type
  function IsWideStringFieldType(FieldType: TACRBaseFieldType): Boolean; overload;
  function IsWideStringFieldType(FieldType: TACRAdvancedFieldType): Boolean; overload;

  // return true if field type is bytes field type
  function IsBytesFieldType(FieldType: TACRBaseFieldType): Boolean; overload;
  function IsBytesFieldType(FieldType: TACRAdvancedFieldType): Boolean; overload;

  // return true if field type is Autoinc field type
  function IsAutoincFieldType(FieldType: TACRAdvancedFieldType): Boolean; overload;

  // return true if DataType is numeric
  function IsNumericFieldType(FieldType: TACRBaseFieldType): Boolean;

  // return true if DataType is TateTime, Time, Date, TimeStamp
  function IsDateTimeFieldType(FieldType: TACRBaseFieldType): Boolean;

  // Result = Can cast this type
  function IsConvertableFieldType(FieldType: TACRAdvancedFieldType): boolean;

  // Like '%_' compare for String
  function IsStrMatchPattern(StrPtr: PChar; PatternPtr: PChar; bIgnoreCase:boolean): Boolean;
  // Like '%_' compare for WideString
  function IsWideStrMatchPattern(StrPtr: PWideChar; PatternPtr: PWideChar; bIgnoreCase:boolean): Boolean;

  function GetStrLength(Buffer: PChar; FieldType: TACRAdvancedFieldType): Integer;

  function GetReservedWord(Word: TReservedWord): String;

  // return HEX string converted from the source buffer
  function ACRBinaryToHEX(Data: PChar; Size: Integer): String;
  // return MIME64 string converted from the source buffer
  function ACRBinaryToMIME64(Data: PChar; Size: Integer): String;
  // allocate buffer and return binary data converted from HEX
  function ACRHEXToBinary(const HEXValue: String): PChar;
  // allocate buffer and return binary data converted from MIME64
  function ACRMIME64ToBinary(const MIME64Value: String): PChar;

implementation

//------------------------------------------------------------------------------
// Convert AdvancedFieldType to BasicFieldType
//------------------------------------------------------------------------------
function AdvancedFieldTypeToBaseFieldType(AdvancedFieldType: TACRAdvancedFieldType): TACRBaseFieldType;
var i: Integer;
begin
  Result := bftUnknown;
  for i:=Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].AdvancedFieldType = AdvancedFieldType then
      begin
        Result := SQLFieldTypes[i].BaseFieldType;
        break;
      end;
end;//AdvancedFieldTypeToBasicFieldType


//------------------------------------------------------------------------------
// Convert BaseFieldType to AdvancedFieldType
//------------------------------------------------------------------------------
function BaseFieldTypeToAdvancedFieldType(BaseFieldType: TACRBaseFieldType): TACRAdvancedFieldType;
var i: Integer;
begin
  Result := aftUnknown;
  for i:=Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].BaseFieldType = BaseFieldType then
      begin
        Result := SQLFieldTypes[i].AdvancedFieldType;
        break;
      end;
end;//BaseFieldTypeToAdvancedFieldType


//------------------------------------------------------------------------------
// Converts TFieldType to TACRAdvancedFieldType
//------------------------------------------------------------------------------
function FieldTypeToACRAdvFieldType(FieldType: TFieldType): TACRAdvancedFieldType;
var i: Integer;
begin
  Result := aftUnknown;
  for i:=Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].FieldType = FieldType then
      begin
        Result := SQLFieldTypes[i].AdvancedFieldType;
        break;
      end;
  if (Result = aftUnknown) then
    raise EACRException.Create(30289, ErrorGUnknownDataType,
                               [IntToStr(Integer(FieldType))]); 
end;//FieldTypeToACRAdvFieldType


//------------------------------------------------------------------------------
// Converts TACRAdvancedFieldType to TFieldType
//------------------------------------------------------------------------------
function ACRAdvFieldTypeToFieldType(AdvancedFieldType: TACRAdvancedFieldType): TFieldType;
var i: Integer;
begin
  Result := ftUnknown;
  for i:=Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].AdvancedFieldType = AdvancedFieldType then
      begin
        Result := SQLFieldTypes[i].FieldType;
        break;
      end;
end;//ACRAdvFieldTypeToFieldType


//------------------------------------------------------------------------------
// AdvancedFieldType for print
//------------------------------------------------------------------------------
function AftToStr(AdvancedFieldType: TACRAdvancedFieldType): String;
var i: Integer;
begin
  Result := 'Unknown';
  for i:=Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].AdvancedFieldType = AdvancedFieldType then
      begin
        Result := SQLFieldTypes[i].SqlName;
        break;
      end;
end;//AftToStr


//------------------------------------------------------------------------------
// BaseFieldType for print
//------------------------------------------------------------------------------
function BftToStr(BaseFieldType: TACRBaseFieldType): String;
var i: Integer;
begin
  Result := 'Unknown';
  for i:=Low(SQLFieldTypes) to High(SQLFieldTypes) do
    if SQLFieldTypes[i].BaseFieldType = BaseFieldType then
      begin
        Result := SQLFieldTypes[i].SqlName;
        break;
      end;
end;//BftToStr


//------------------------------------------------------------------------------
// Convert ACRDate to Date
//------------------------------------------------------------------------------
function ACRDateToDate(ACRDate: TACRDate): TDate;
begin 
  Result := ACRDate - DateDelta; 
end;//ACRDateToDate
	
 
//------------------------------------------------------------------------------ 
// Convert Date to ACRDate
//------------------------------------------------------------------------------
function DateToACRDate(Date: TDate): TACRDate; 
begin
  Result := Trunc(Date) + DateDelta; 
end;//DateToACRDate
 
	
//------------------------------------------------------------------------------ 
// Convert ACRTime to Time
//------------------------------------------------------------------------------ 
function ACRTimeToTime(ACRTime: TACRTime): TTime; 
begin
  Result := Frac(ACRTime / (24*60*60*1000)); 
end;//ACRTimeToTime 
 
 
//------------------------------------------------------------------------------ 
// Convert Time to ACRTime
//------------------------------------------------------------------------------
function TimeToACRTime(Time: TTime): TACRTime;
begin
  //Result := Trunc(Frac(Time) * (24*60*60*1000));
  Result := Round(Frac(Time) * (24*60*60*1000)); 
end;//TimeToACRTime
 
	
//------------------------------------------------------------------------------ 
// Convert ACRDateTime to DateTime 
//------------------------------------------------------------------------------
function ACRDateTimeToDateTime(ACRDateTime: TACRDateTime): TDateTime; 
begin
  Result := ACRDateToDate(ACRDateTime.Date) + ACRTimeToTime(ACRDateTime.Time); 
end;//ACRDateTimeToDateTime 
	
	
//------------------------------------------------------------------------------ 
// Convert DateTime to ACRDateTime 
//------------------------------------------------------------------------------ 
function DateTimeToACRDateTime(DateTime: TDateTime): TACRDateTime; 
begin 
  Result.Date := DateToACRDate(DateTime); 
  Result.Time := TimeToACRTime(DateTime); 
end;//ACRDateTimeToDateTime
	
 
//------------------------------------------------------------------------------
// return true if field type is a BLOB field type
//------------------------------------------------------------------------------ 
function IsBLOBFieldType(FieldType: TACRBaseFieldType): Boolean;
begin
  Result := FieldType in [bftBLOB, bftClob, bftWideClob]; 
end; // IsBLOBFieldType 
 
	
//------------------------------------------------------------------------------
// return true if field type is a BLOB field type 
//------------------------------------------------------------------------------
function IsBLOBFieldType(FieldType: TACRAdvancedFieldType): Boolean;
begin 
  Result := FieldType in [aftBLOB, aftGraphic, aftMemo, aftFormattedMemo, aftWideMemo]; 
end; // IsBLOBFieldType
	
	
//------------------------------------------------------------------------------ 
// return true if field type is a varchar field type 
//------------------------------------------------------------------------------ 
function IsVarcharFieldType(FieldType: TACRBaseFieldType): Boolean; 
begin
  Result := FieldType in [bftWideVarchar, bftVarchar]; 
end; // IsVarcharFieldType 
 
	
//------------------------------------------------------------------------------
// return true if field type is a varchar field type
//------------------------------------------------------------------------------ 
function IsVarcharFieldType(FieldType: TACRAdvancedFieldType): Boolean;
begin
  Result := FieldType in [aftWideString, aftString];
end; // IsVarcharFieldType 
 
	
//------------------------------------------------------------------------------
// return true if field type is a string field type, but not wide string
//------------------------------------------------------------------------------
function IsStringFieldType(FieldType: TACRBaseFieldType): Boolean;
begin
  Result := FieldType in [bftChar, bftWideChar, bftVarchar, bftWideVarchar , bftClob, bftWideClob]; 
end; // IsStringFieldType
	
 
//------------------------------------------------------------------------------
// return true if field type is a string field type, but not wide string
//------------------------------------------------------------------------------ 
function IsStringFieldType(FieldType: TACRAdvancedFieldType): Boolean; 
begin 
  Result := FieldType in [aftChar, aftString, aftWideChar, aftWideString ]; // , aftMemo, aftFormattedMemo, aftWideMemo 
end; // IsStringFieldType
 
 
//------------------------------------------------------------------------------
// return true if field type is a wide string field type
//------------------------------------------------------------------------------
function IsWideStringFieldType(FieldType: TACRBaseFieldType): Boolean;
begin 
  Result := FieldType in [bftWideChar, bftWideVarchar, bftWideClob]; 
end; // IsStringFieldType
	
 
//------------------------------------------------------------------------------ 
// return true if field type is a wide string field type 
//------------------------------------------------------------------------------
function IsWideStringFieldType(FieldType: TACRAdvancedFieldType): Boolean;
begin
  Result := FieldType in [aftWideChar, aftWideString, aftWideMemo];
end; // IsWideStringFieldType
 
	
//------------------------------------------------------------------------------ 
// return true if field type is a Bytes field type, but not wide Bytes
//------------------------------------------------------------------------------
function IsBytesFieldType(FieldType: TACRBaseFieldType): Boolean;
begin
  Result := FieldType in [bftBytes, bftVarBytes];
end; // IsBytesFieldType


//------------------------------------------------------------------------------
// return true if field type is a Bytes field type, but not wide Bytes
//------------------------------------------------------------------------------
function IsBytesFieldType(FieldType: TACRAdvancedFieldType): Boolean;
begin
  Result := FieldType in [aftBytes, aftVarBytes];
end; // IsBytesFieldType


//------------------------------------------------------------------------------
// return true if field type is a Autoinc field type, but not wide Autoinc
//------------------------------------------------------------------------------
function IsAutoincFieldType(FieldType: TACRAdvancedFieldType): Boolean;
begin
  Result := FieldType in [aftAutoInc,
                          aftAutoIncShortint,
                          aftAutoIncSmallint,
                          aftAutoIncInteger,
                          aftAutoIncLargeint,
                          aftAutoIncByte,
                          aftAutoIncWord,
                          aftAutoIncCardinal];
end; // IsAutoincFieldType


//------------------------------------------------------------------------------
// return true if DataType is numeric
//------------------------------------------------------------------------------
function IsNumericFieldType(FieldType: TACRBaseFieldType): Boolean;
begin
  Result := FieldType in [bftSignedInt8, bftSignedInt16, bftSignedInt32, bftSignedInt64,
                          bftUnsignedInt8, bftUnsignedInt16, bftUnsignedInt32,
                          bftSingle, bftDouble, bftExtended, bftCurrency];
end;//IsNumericDataType


//------------------------------------------------------------------------------
// return true if DataType is TateTime, Time, Date, TimeStamp
//------------------------------------------------------------------------------
function IsDateTimeFieldType(FieldType: TACRBaseFieldType): Boolean;
begin
  Result := FieldType in [bftDate, bftTime, bftDateTime];
end;//IsDateTimeFieldType

//------------------------------------------------------------------------------
// Result = Can cast this type
//------------------------------------------------------------------------------
function IsConvertableFieldType(FieldType: TACRAdvancedFieldType): boolean;
begin
  Result := FieldType in [aftChar, aftString, aftWideChar, aftWideString,
                          aftShortint, aftSmallint, aftAutoInc, aftInteger,
                          aftLargeint, aftByte, aftWord, aftCardinal,
                          aftSingle, aftDouble, aftExtended,
                          aftBoolean,
                          aftCurrency, 
                          aftDate, aftTime, aftDateTime, aftTimeStamp];
end; // IsConvertableFieldType


//------------------------------------------------------------------------------
// IsStrMatchPattern
//------------------------------------------------------------------------------
function IsStrMatchPattern(StrPtr: PChar; PatternPtr: PChar; bIgnoreCase:boolean): Boolean;
var i : integer;
    bEQ: Boolean;
    tmp1, tmp2: array [0..1] of char;
begin
  tmp1[0]:=#0; tmp1[1]:=#0;
  tmp2[0]:=#0;tmp2[1]:=#0;
  repeat
      if (StrComp(PatternPtr,WildCardMultipleChar)=0) then
       begin
         Result:=True;
         exit;
       end
      else if (StrPtr^=#0) and (PatternPtr^ <> #0) then
       begin
         Result:=False;
         exit;
       end
      else if (StrPtr^=#0) then
       begin
         Result:=True;
         exit;
       end
      else
         begin
           case PatternPtr^ of
            WildCardMultipleChar:
               begin
                for i:=0 to Length(StrPtr)-1 do
                 begin
                  if IsStrMatchPattern(StrPtr+i,PatternPtr+1,bIgnoreCase) then
                   begin
                    Result := True;
                    exit;
                   end;
                 end;
                Result := False;
                exit;
               end;
            WildCardSingleChar:
               begin
                inc(StrPtr);
                inc(PatternPtr);
               end;
            else
               begin
                tmp1[0] := StrPtr^;
                tmp2[0] := PatternPtr^;
                if (bIgnoreCase) then
{$IFDEF MSWINDOWS}
                 //bEQ := (Q_AnsiCompTextL(tmp1, tmp2, 1) = 0)
                 bEQ := (CompareString(LOCALE_USER_DEFAULT,
                         NORM_IGNORECASE + SORT_STRINGSORT,
//                         SORT_STRINGSORT,
                         tmp1, 1, tmp2, 1) - 2) = 0
{$ENDIF}
{$IFDEF LINUX}
                 bEQ := (strncmp(tmp1, tmp2, 1)) = 0
{$ENDIF}
                else
{$IFDEF MSWINDOWS}
                 //bEQ := (Q_AnsiCompStrL(tmp1, tmp2, 1) = 0);
                 bEQ := (CompareString(LOCALE_USER_DEFAULT,
//                         NORM_IGNORECASE + SORT_STRINGSORT,
                         SORT_STRINGSORT,
                         tmp1, 1, tmp2, 1) - 2) = 0;
{$ENDIF}
{$IFDEF LINUX}
                 bEQ := (strncmp(tmp1, tmp2, 1)) = 0;
{$ENDIF}
                if (bEQ) then
                 begin
                  inc(StrPtr);
                  inc(PatternPtr);
                 end
                else
                 begin
                  Result:=False;
                  exit;
                 end;
               end;
           end; // case
         end; // non-simple cases
  until false;
end;// IsStrMatchPattern


//------------------------------------------------------------------------------
// Like '%_' compare for WideString
//------------------------------------------------------------------------------
function IsWideStrMatchPattern(StrPtr: PWideChar; PatternPtr: PWideChar; bIgnoreCase:boolean): Boolean;
var i: integer;
    bOk: boolean;
{$IFDEF LINUX}
    UCS4_S1, UCS4_S2: UCS4String;
{$ENDIF}
begin
  repeat
{$IFDEF MSWINDOWS}
      if (CompareStringW(LOCALE_USER_DEFAULT,
                        NORM_IGNORECASE + SORT_STRINGSORT,
                        PatternPtr,
                        Length(PatternPtr),
                        WideChar(WildCardMultipleChar),
                        1)-2 = 0) then
{$ENDIF}
{$IFDEF LINUX}
      UCS4_S1 := WideStringToUCS4String(PatternPtr);
      UCS4_S2 := WideStringToUCS4String(WildCardMultipleChar);
  // glibc 2.1.2 / 2.1.3 implementations of wcscoll() and wcsxfrm()
  // have severe capacity limits.  Comparing two 100k strings may
  // exhaust the stack and kill the process.
  // Fixed in glibc 2.1.91 and later.
      SetLastError(0);
      if (wcscoll(PUCS4Chars(UCS4_S1), PUCS4Chars(UCS4_S2)) = 0) then
{$ENDIF}
       begin
         Result:=True;
         exit;
       end
      else if (StrPtr^=#0) and (PatternPtr^ <> #0) then
       begin
         Result:=False;
         exit;
       end
      else if (StrPtr^=#0) then
       begin
         Result:=True;
         exit;
       end
      else
         begin
           case PatternPtr^ of
            WildCardMultipleChar:
               begin
                for i:=0 to Length(StrPtr)-1 do
                 begin
                  if IsWideStrMatchPattern(StrPtr+i,PatternPtr+1,bIgnoreCase) then
                   begin
                    Result := True;
                    exit;
                   end;
                 end;
                Result := False;
                exit;
               end;
            WildCardSingleChar:
               begin
                inc(StrPtr);
                inc(PatternPtr);
               end;
            else
               begin
                bOk := false;
                if bIgnoreCase then
                 begin
{$IFDEF MSWINDOWS}
                  if (Windows.CompareStringW(LOCALE_USER_DEFAULT,
                        NORM_IGNORECASE + SORT_STRINGSORT,
                        PatternPtr, 1, StrPtr,1)-2 = 0) then
{$ENDIF}
{$IFDEF LINUX} // case not ignored
                  UCS4_S1 := WideStringToUCS4String(PatternPtr);
                  UCS4_S2 := WideStringToUCS4String(StrPtr);
                  if (wcscoll(PUCS4Chars(UCS4_S1), PUCS4Chars(UCS4_S2))
                         = 0) then
{$ENDIF}
                   bOk := true;
                 end
                else
                 begin
{$IFDEF MSWINDOWS}
                  if (Windows.CompareStringW(LOCALE_USER_DEFAULT,
                        SORT_STRINGSORT,
                        PatternPtr, 1, StrPtr,1)-2 = 0) then
{$ENDIF}
{$IFDEF LINUX}
                  UCS4_S1 := WideStringToUCS4String(PatternPtr);
                  UCS4_S2 := WideStringToUCS4String(StrPtr);
                  if (wcscoll(PUCS4Chars(UCS4_S1), PUCS4Chars(UCS4_S2))
                         = 0) then
{$ENDIF}
                   bOk := true;
                 end;

                if (bOk) then
                 begin
                  inc(StrPtr);
                  inc(PatternPtr);
                 end
                else
                 begin
                  Result:=False;
                  exit;
                 end;
               end;
           end; // case
         end; // non-simple cases
  until false;
end;// IsWideStrMatchPattern


//------------------------------------------------------------------------------
// return string value length
//------------------------------------------------------------------------------
function GetStrLength(Buffer: PChar; FieldType: TACRAdvancedFieldType): Integer;
begin
 if (IsWideStringFieldType(FieldType)) then
  Result := Length(WideCharToString(PWideChar(Buffer))) * 2
 else
  Result := StrLen(Buffer);
end; // GetStrLength


//------------------------------------------------------------------------------
// return reserved word text
//------------------------------------------------------------------------------
function GetReservedWord(Word: TReservedWord): String;
begin
  Result := ACRSQLReservedWords[Integer(Word)];
end; // GetReservedWord


//------------------------------------------------------------------------------
// return HEX string converted from the source buffer
//------------------------------------------------------------------------------
function ACRBinaryToHEX(Data: PChar; Size: Integer): String;
begin
  Result := StrToFormat(Data,Size,fmtHEX);
end; // ACRBinaryToHEX


//------------------------------------------------------------------------------
// return MIME64 string converted from the source buffer
//------------------------------------------------------------------------------
function ACRBinaryToMIME64(Data: PChar; Size: Integer): String;
begin
  Result := StrToFormat(Data,Size,fmtMIME64);
end; // ACRBinaryToMIME64


//------------------------------------------------------------------------------
// allocate buffer and return binary data converted from HEX
//------------------------------------------------------------------------------
function ACRHEXToBinary(const HEXValue: String): PChar;
var s: String;
    l: Integer;
begin
  s := FormatToStr(PChar(@HEXValue[1]),Length(HEXValue),fmtHEX);
  l := Length(s);
  if (l <= 0) then
   Result := nil
  else
   begin
     Result := MemoryManager.GetMem(l);
     Move(PChar(@s[1])^,Result^,l);
   end;
end; // ACRHEXToBinary


//------------------------------------------------------------------------------
// allocate buffer and return binary data converted from MIME64
//------------------------------------------------------------------------------
function ACRMIME64ToBinary(const MIME64Value: String): PChar;
var s: String;
    l: Integer;
begin
  s := FormatToStr(PChar(@MIME64Value[1]),Length(MIME64Value),fmtMIME64);
  l := Length(s);
  if (l <= 0) then
   Result := nil
  else
   begin
     Result := MemoryManager.GetMem(l);
     Move(PChar(@s[1])^,Result^,l);
   end;
end; // ACRMIME64ToBinary


initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRConverts> initialized');
{$ENDIF}

end.
