unit ACRLocalEngine;

{$I ACRVer.inc}

interface

uses SysUtils, Classes,

// Accuracer units

     ACRRelationalAlgebra,
     //{$IFDEF DEBUG_LOG}
     //ACRDebug,
     //{$ENDIF}
     ACRExcept,
     ACRBase,
     ACRBaseEngine,
     {$IFDEF MEMORY_ENGINE}
     ACRMemEngine,
     {$ENDIF}
     {$IFDEF TEMPORARY_ENGINE}
     ACRTempEngine,
     {$ENDIF}
     {$IFDEF DISK_ENGINE}
     ACRDiskEngine,
     {$ENDIF}
{$IFNDEF D5H}
     ACRD4Routines,
{$ENDIF}
     ACRCompression,
     ACRTypes,
     //ACRSQLProcessor,
     ACRExpressions,
     ACRConst,
     ACRVariant,
     //ACRConverts,
     ACRMemory;


type


////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalBLOBStream
//
////////////////////////////////////////////////////////////////////////////////

  // local BLOB stream
  TACRLocalBLOBStream = class (TACRStream)
   private
    FOpenMode:                  TACRBLOBOpenMode;
    FTemporaryStream:           TACRStream;
    FUserBLOBStream:            TACRStream;
    FFieldNo:                   Integer;
    FCursor:                    TACRCursor;
    FPosition:                  Int64;
    FDoNotFreeCompressedStream: Boolean;
   protected
    // sets new size of the stream
    procedure InternalSetSize(const NewSize: Int64);
    // sets new size of the stream
    procedure SetSize(NewSize: Longint);
    {$IFDEF D6H}
      overload;
    {$ENDIF}
      override;
    {$IFDEF D6H}
    procedure SetSize(const NewSize: Int64); overload; override;
    {$ENDIF}
   public
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint;
    {$IFDEF D6H}
            overload;
    {$ENDIF}
      override;
    {$IFDEF D6H}
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; overload; override;
    {$ENDIF}
    constructor Create(
                        TemporaryStream: TACRStream;
                        Cursor: TACRCursor;
                        OpenMode: TACRBLOBOpenMode;
                        FieldNo: Integer
                      );
    destructor Destroy; override;
   public
    // blob stream interface
   public
    property Cursor: TACRCursor read FCursor;
    property FieldNo: Integer read FFieldNo;
    property OpenMode: TACRBLOBOpenMode read FOpenMode write FOpenMode;
    property TemporaryStream: TACRStream read FTemporaryStream write FTemporaryStream;
    property UserBLOBStream: TACRStream read FUserBLOBStream write FUserBLOBStream;
    property DoNotFreeCompressedStream: Boolean read FDoNotFreeCompressedStream write FDoNotFreeCompressedStream;
  end; // TACRLocalBLOBtream


////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalCursor
//
////////////////////////////////////////////////////////////////////////////////


  TACRLocalCursor = class (TACRCursor)
   private
    FTableData:                     TACRTableData;
    FDatabaseData:                  TACRDatabaseData;
    FSavedPosition:                 Pointer;
    FSavedRecordBuffer:             TACRRecordBuffer;
    FDoNotCloseTableData:           Boolean;
    FDoNotUnlockTable:              Boolean;
   public
    constructor Create;
    destructor Destroy; override;
    // create table
    procedure CreateTable(
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                         ); override;
    procedure DeleteTable(Cascade: Boolean = false); override;
    procedure EmptyTable; override;
    procedure AddForeignKey(ConstraintDef: TACRConstraintDefForeignKey); override;
    procedure DeleteConstraint(Name: String; Cascade: Boolean; FKPartialDelete: Boolean); override;
    procedure RenameTable(NewTableName: String); override;
    procedure RenameField(FieldName, NewFieldName: String); override;
    procedure UpdateTableDefinitions; override;
   protected
    procedure TryToCopyRecords(NewCursor: TACRLocalCursor; var Log: String);
   public
    function RepairTable(var Log: String; NewSession: Pointer = nil): Boolean; override;
    procedure LoadTableFromStream(
                        Stream:               TStream
                       ); override;
    procedure SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      ); override;
   protected
    function GetDropTableCommand(aTableName: String): String;
    function GetAutoIncFieldDefinition(FieldDef: TACRFieldDef): String;
    function GetBLOBFieldDefinition(FieldDef: TACRFieldDef): String;
    function GetFieldDefinition(FieldDef: TACRFieldDef; UseBracketsForNames:  Boolean): String;
    function GetCreateTableCommand(aTableName: String; UseBracketsForNames:  Boolean): String;
    function GetCreateIndexCommand(aTableName: String; IndexName: String; IndexDef: TACRIndexDef; UseBracketsForNames:  Boolean): String;
    function GetDropIndexCommand(aTableName: String; IndexName: String): String;
    function GetSQLBLOBFieldValue(FieldNo: Integer): String;
    function GetSQLFieldValue(var v: TACRVariant; FieldNo: Integer; ExportBLOBFields: Boolean): String;
    function GetInsertCommand(var v: TACRVariant; aTableName: String; UseBracketsForNames:  Boolean; ExportBLOBFields: Boolean): String;
   public
    function ExportTableToSQL(
                              ExportStructure:      Boolean;
                              AddDropTableCommand:  Boolean;
                              ExportIndexes:        Boolean;
                              AddDropIndexCommand:  Boolean;
                              ExportData:           Boolean;
                              ExportBLOBFields:     Boolean;
                              UseBracketsForNames:  Boolean
                            ): String; override;
    procedure InternalInitFieldDefs; override;
    procedure OpenTableByFieldDefs(
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                       ); override;
    procedure OpenTable(aTableData: TACRTableData);
    procedure CloseTable; override;

    // initialize record buffer
    procedure InternalInitRecord(RecordBuffer: TACRRecordBuffer;
                InsertMode: Boolean); override;

    // index operations
    function GetIndexDefs: TACRIndexDefs; override;
    procedure AddIndex(IndexDef: TACRIndexDef); override;
    procedure DeleteIndex(Name: String); override;
    procedure DeleteAllIndexes; override;
    // return index name of the index or '' if not found
    function FindIndex(FieldNamesList, AscDescList,
              CaseSensitivityList: TStringList): String; override;
    // return index name
    function FindOrCreateIndex(FieldNamesList, AscDescList,
              CaseSensitivityList: TStringList): String; override;

    //---------------------------------------------------------------------------
    // navigation & bookmark methods
    //---------------------------------------------------------------------------
    // return true if current record exists
    function IsRecordExists: Boolean; override;
    function GetRecordBuffer(
              GetRecordMode:  TACRGetRecordMode
              ): TACRGetRecordResult; override;
    function GetRecordCount: TACRRecordNo; override;
    // go to record
    procedure SetRecNo(Value: TACRRecordNo); override;
    function GetRecNo: TACRRecordNo; override;

    //---------------------------------------------------------------------------
    // insert, edit, post, delete methods
    //---------------------------------------------------------------------------

    // insert record
    procedure InternalInsert; override;
    // edit record
    procedure InternalEdit; override;
    // cancels updates
    procedure InternalCancel(ToInsert: Boolean); override;
    // insert or update record
    procedure InternalPost(ToInsert: Boolean); override;
    // delete record
    procedure InternalDelete; override;
    procedure DeleteVisibleRecords; override;
    procedure UpdateVisibleRecords(FieldNames:   TStringList;
                                   values:       array of TACRVariant;
                                   SkipFKCheck:  Boolean = False
                                   ); override;

    //---------------------------------------------------------------------------
    // search & filter methods
    //---------------------------------------------------------------------------

    // disable record bitmap
    procedure DisableRecordBitmap; override;
    // set distinct
    procedure ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList); override;
    // apply projection
    procedure ApplyProjection(FieldNamesList, AliasList: TStringList); override;
    procedure ActivateFilters(
                              FilterText:      String;
                              CaseInsensitive: Boolean;
                              PartialKey:      Boolean
                            ); override;
    procedure DeactivateFilters; override;
    function Locate(
                    const KeyFields: String;
                    const KeyValues: Variant;
                    CaseInsensitive: Boolean;
                    PartialKey:      Boolean
                   ): Boolean; override;
    function FindKey(SearchCondition: TACRSearchCondition): Boolean; override;
    procedure ResetRange; override;
    procedure ApplyRange(
                          StartBuffer, EndBuffer: TACRRecordBuffer;
                          StartKeyFieldCount:     Integer;
                          EndKeyFieldCount:       Integer;
                          StartExclusive:         Boolean;
                          EndExclusive:           Boolean
                        ); override;
    // set SQL Filter
    procedure SetSQLFilter(FilterExpr: TObject); override;


    //---------------------------------------------------------------------------
    // BLOB methods
    //---------------------------------------------------------------------------

    function InternalCreateBlobStream(
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ):TACRStream; override;

    procedure InternalCloseBLOB(FieldNo: Integer); override;

    // clear blob streams
    procedure ClearBLOBStreams(WriteOnly: Boolean = False); override;

    function LastAutoincValue(FieldNo: Integer): Int64; override;
    procedure SetLastAutoincValue(Value: Int64; FieldNo: Integer); override;

    procedure LockTableForRead; override;
    procedure UnlockTableForRead; override;

    procedure LockTableS(ErrorCode: Integer); override;
    procedure UnlockTableS(ErrorCode: Integer); override;
    procedure LockTableIRW(ErrorCode: Integer); override;
    procedure UnlockTableIRW(ErrorCode: Integer); override;
    procedure LockTableRW(ErrorCode: Integer); override;
    procedure UnlockTableRW(ErrorCode: Integer); override;

    function TryToLockTableS: Boolean; override;
    function TryToUnlockTableS: Boolean; override;
    function TryToLockTableIRW: Boolean; override;
    function TryToUnlockTableIRW: Boolean; override;
    function TryToLockTableRW: Boolean; override;
    function TryToUnlockTableRW: Boolean; override;

    procedure SaveCurrentPosition;
    procedure RestoreSavedPosition;
    procedure FreeSavedPosition;
   public
    property TableData: TACRTableData read FTableData;
    property DatabaseData: TACRDatabaseData read FDatabaseData;
    property DoNotUnlockTable: Boolean read FDoNotUnlockTable write FDoNotUnlockTable;
  end; // TACRLocalCursor


////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalSession
//
////////////////////////////////////////////////////////////////////////////////


  TACRLocalSession = class (TACRBaseSession)
   private
    FDatabaseData:  TACRDatabaseData;
    FTransaction:   TACRTransaction;

    function IsAppropriateDatabaseData(DBData: TACRDatabaseData): Boolean;
    function FindDatabaseData: TACRDatabaseData;
    function CreateDatabaseData: TACRDatabaseData;
    function FindOrCreateDatabaseData: TACRDatabaseData;

   protected
    // db connected?
    function GetConnected: Boolean; override;
    // connect / disconnect
    procedure SetConnected(Value: boolean); override;

   public
    // check if database exists
    function GetDatabaseExists: Boolean; override;
    // create database
    procedure CreateDatabase; override;
    // delete database
    procedure DeleteDatabase; override;
    // rename database
    procedure RenameDatabase(NewDatabaseFileName: String); override;
    // flush file buffers
    procedure FlushFileBuffers; override;
    // return database format version
    function GetFormatVersion: Double; override;
    // return total number of pages
    function GetTotalPageCount : Integer; override;
    // return number of free pages
    function GetFreePageCount : Integer; override;
    // return true if database is encrypted
    function IsDatabaseEncrypted: Boolean; override;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword: Boolean; override;
    // makes Exe database from adb file
    procedure MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string); override;
    // removes database file from executable database file
    procedure RemoveDatabaseFromExe; override;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile: Boolean; override;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid: Boolean; override;
    procedure GetTablesList(List: TStrings); override;
    function TableExists(TableName: String): Boolean; override;
   protected
    // retrun true if database has active transaction
    function GetInTransaction: Boolean; override;
   public
    constructor Create;
    // start a transaction
    procedure StartTransaction; override;
    // apply changes made by transaction
    procedure Commit(FlushFileBuffers: Boolean = True); override;
    // cancel changes made by transaction
    procedure Rollback; override;
    procedure RemoveAllLocks; override;
   public
    property DatabaseData: TACRDatabaseData read FDatabaseData;
    property Transaction: TACRTransaction read FTransaction;
  end; // TACRLocalSession

var
  DBDataList: TThreadList;

implementation

{$IFDEF D6H}
uses DateUtils;
{$ELSE}
procedure DecodeDateTime(const AValue: TDateTime; out AYear, AMonth, ADay,
  AHour, AMinute, ASecond, AMilliSecond: Word);
begin
  DecodeDate(AValue, AYear, AMonth, ADay);
  DecodeTime(AValue, AHour, AMinute, ASecond, AMilliSecond);
end;
{$ENDIF}

////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalBLOBStream
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// sets new size of the stream
//------------------------------------------------------------------------------
procedure TACRLocalBLOBStream.InternalSetSize(const NewSize: Int64);
begin
 if (OpenMode = bomRead) then
  raise EACRException.Create(10116,ErrorLCannotWriteToReadOnlyStream);
 FTemporaryStream.Size := NewSize;
end; // InternalSetSize


//------------------------------------------------------------------------------
// sets new size of the stream
//------------------------------------------------------------------------------
procedure TACRLocalBLOBStream.SetSize(NewSize: Longint);
begin
 InternalSetSize(NewSize);
end; // SetSize


{$IFDEF D6H}
//------------------------------------------------------------------------------
// sets new size of the stream
//------------------------------------------------------------------------------
procedure TACRLocalBLOBStream.SetSize(const NewSize: Int64);
begin
 InternalSetSize(NewSize);
end; // SetSize
{$ENDIF}


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRLocalBLOBStream.Read(var Buffer; Count: Longint): Longint;
begin
 Result := FTemporaryStream.Read(Buffer,Count);
end; // Read


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRLocalBLOBStream.Write(const Buffer; Count: Longint): Longint;
begin
 if (OpenMode = bomRead) then
  raise EACRException.Create(10115,ErrorLCannotWriteToReadOnlyStream);
 Result := FTemporaryStream.Write(Buffer,Count);
end; // Write


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRLocalBLOBStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
 Result := FTemporaryStream.Seek(Offset,Origin);
end; // Seek


{$IFDEF D6H}
//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRLocalBLOBStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
begin
 Result := FTemporaryStream.Seek(Offset,Origin);
end; // Seek
{$ENDIF}


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
constructor TACRLocalBLOBStream.Create(
                        TemporaryStream: TACRStream;
                        Cursor: TACRCursor;
                        OpenMode: TACRBLOBOpenMode;
                        FieldNo: Integer
                                       );
begin
 UserBLOBStream := nil;
 FPosition := 0;
 FTemporaryStream := TemporaryStream;
 FCursor := Cursor;
 FFieldNo := FieldNo;
 FOpenMode := OpenMode;
 FDoNotFreeCompressedStream := False;
 inherited Create;
end; // Create


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
destructor TACRLocalBLOBStream.Destroy;
begin
 if (not FDoNotFreeCompressedStream) then
   TACRCompressedBLOBStream(FTemporaryStream).CompressedStream.Free;
 FTemporaryStream.Free;
 inherited;
end; // Destroy


////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalCursor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRLocalCursor.Create;
begin
 FRandomOrder := False;
 FDoNotUnlockTable := False;
 FMemoryTableAllocBy := ACRDefaultMemoryTableAllocBy;
 FCreateTableStarted := False;
 FSavedPosition := nil;
 FSavedRecordBuffer := nil;
 CurrentRecordBuffer := nil;
 FDatabaseData := nil;
 FTableData := nil;
 FFieldDefs := nil;
 FVisibleFieldDefs := nil;
 FIndexDefs := nil;
 IsOpen := False;
 FBLOBStreams := nil;
 FIndexName := '';
 FIndexID := INVALID_OBJECT_ID;
 FilterExpression := nil;
 SQLFilterExpression := nil;
 FilterRecord := nil;
 FSettingProjection := False;
 DistinctFieldCount := 0;
 RecordBitmap := nil;
 FDoNotCloseTableData := False;
 FRepair := False;
 FIsClientCursor := False;
end; // Create


//------------------------------------------------------------------------------
// destroy
//------------------------------------------------------------------------------
destructor TACRLocalCursor.Destroy;
begin
 CloseTable;
 inherited;
end; // Destroy


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.CreateTable
                         (
                          FieldDefs: TACRFieldDefs;
                          IndexDefs: TACRIndexDefs;
                          ConstraintDefs: TACRConstraintDefs
                          );
var i: Integer;
begin
  CloseTable;
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   if (InMemory or Temporary) then
    if (FDatabaseData.TableExists(FTableName)) then
     DeleteTable;
   FTableData := FDatabaseData.FindOrCreateTableData(Self);
   if (FTableData = nil) then
    raise EACRException.Create(10523,ErrorLNilPointer);
   for i := 0 to IndexDefs.Count - 1 do
    IndexDefs[i].RootPageNo := INVALID_PAGE_NO; 
   FTableData.CreateTable(Self,FieldDefs,IndexDefs,ConstraintDefs);
  finally
   FDatabaseData.Unlock;
  end;
end;// CreateTable


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DeleteTable(Cascade: Boolean);
begin
  CloseTable;
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   if (InMemory or Temporary) then
     FTableData := FDatabaseData.FindTableData(Self)
   else
    begin
     if (Session.InTransaction) then
      raise EACRException.Create(10838,ErrorLCannotPerformThisOperationInsideATransaction);
     FTableData := FDatabaseData.CreateTableData(Self);
    end;
   if (FTableData = nil) then
    raise EACRException.Create(10027,ErrorLTableDataNotFound,[TableName]);
   FTableData.DeleteTable(Session,Cascade);
   if (InMemory or Temporary) then
    FTableData := nil
   else
    FreeAndNil(FTableData);
  finally
   FDatabaseData.Unlock;
  end;
end; // DeleteTable;


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.EmptyTable;
begin
  CloseTable;
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   if (InMemory or Temporary) then
     FTableData := FDatabaseData.FindTableData(Self)
   else
     FTableData := FDatabaseData.CreateTableData(Self);
   if (FTableData = nil) then
    raise EACRException.Create(10073,ErrorLTableDataNotFound,[TableName]);
   FTableData.EmptyTable(Self);
   if (InMemory or Temporary) then
    FTableData := nil
   else
    FreeAndNil(FTableData);
  finally
   FDatabaseData.Unlock;
  end;
end; // EmptyTable


//------------------------------------------------------------------------------
// add foreign key
//------------------------------------------------------------------------------
procedure TACRLocalCursor.AddForeignKey(ConstraintDef: TACRConstraintDefForeignKey);
begin
  if (FTableData = nil) then
   raise EACRException.Create(11582,ErrorLTableDataNotFound,[TableName]);
  if (not FExclusive) then
   raise EACRException.Create(11583,ErrorLTableIsNotOpenedExclusively,[FTableName]);
  FTableData.AddForeignKey(Self,ConstraintDef); 
end; // AddForeignKey


//------------------------------------------------------------------------------
// delete constraint (FK,FKAction or PK) and write changes to MetaData file
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DeleteConstraint(Name: String; Cascade: Boolean; FKPartialDelete: Boolean);
begin
  if (FTableData = nil) then
   raise EACRException.Create(11484,ErrorLTableDataNotFound,[TableName]);
  if (not FExclusive) then
   raise EACRException.Create(11486,ErrorLTableIsNotOpenedExclusively,[FTableName]);
  FTableData.DeleteConstraint(Self,Name,Cascade,FKPartialDelete);
end; // DeleteConstraint


//------------------------------------------------------------------------------
// rename table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.RenameTable(NewTableName: String);
begin
  CloseTable;
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   if (InMemory or Temporary) then
     FTableData := FDatabaseData.FindTableData(Self)
   else
    begin
     if (Session.InTransaction) then
      raise EACRException.Create(10839,ErrorLCannotPerformThisOperationInsideATransaction);
     FTableData := FDatabaseData.CreateTableData(Self);
    end;
   if (FTableData = nil) then
    raise EACRException.Create(10151,ErrorLTableDataNotFound,[TableName]);
   FTableData.RenameTable(Self,NewTableName);
   FTableName := NewTableName;
   if (InMemory or Temporary) then
    FTableData := nil
   else
    FreeAndNil(FTableData);
  finally
   FDatabaseData.Unlock;
  end;
end; // RenameTable


//------------------------------------------------------------------------------
// Rename Field
//------------------------------------------------------------------------------
procedure TACRLocalCursor.RenameField(FieldName, NewFieldName: String);
begin
  CloseTable;
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   if (InMemory or Temporary) then
     FTableData := FDatabaseData.FindTableData(Self)
   else
    begin
     if (Session.InTransaction) then
      raise EACRException.Create(11165,ErrorLCannotPerformThisOperationInsideATransaction);
     FTableData := FDatabaseData.CreateTableData(Self);
    end;
   if (FTableData = nil) then
    raise EACRException.Create(11166,ErrorLTableDataNotFound,[TableName]);
   FTableData.OpenTable(Self);
   try
     FTableData.RenameField(Self,FieldName, NewFieldName);
   finally
     FTableData.CloseTable(Self);
   end;
   if (InMemory or Temporary) then
    FTableData := nil
   else
    FreeAndNil(FTableData);
  finally
   FDatabaseData.Unlock;
  end;
end;//RenameField


//------------------------------------------------------------------------------
// update table definitions (fields, indexes, constraints)
//------------------------------------------------------------------------------
procedure TACRLocalCursor.UpdateTableDefinitions;
begin
 if (FIsOpen) then
  begin
   if (FIndexDefs <> nil) then
    FIndexDefs.Free;
   FIndexDefs := TACRIndexDefs.Create;
   FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
   if (FConstraintDefs <> nil) then
    FConstraintDefs.Free;
   FConstraintDefs := TACRConstraintDefs.Create;
   FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
   InternalInitFieldDefs;
  end;
end; // UpdateTableDefinitions


//------------------------------------------------------------------------------
// copy records
//------------------------------------------------------------------------------
procedure TACRLocalCursor.TryToCopyRecords(NewCursor: TACRLocalCursor; var Log: String);
var GetRecordResult: TACRGetRecordResult;
    bStart:           Boolean;
    bCalculateMinMax: Boolean;
    minmax:           array of Int64;
    CurrentValue:     TACRVariant;
    i:                Integer;

  procedure CopyRecord;
  var i:          Integer;
      bs,bs1:     TACRStream;
      BLOBExists: Boolean;
  begin
    Move(FCurrentRecordBuffer^,NewCursor.FCurrentRecordBuffer^,FRecordBufferSize);
    BLOBExists := False;
    NewCursor.InternalInsert;
    try
     if (FieldDefs.VarcharOrBLOBFieldsExists) then
      for i := 0 to FieldDefs.Count - 1 do
       if (IsBLOBFieldType(FieldDefs[i].BaseFieldType)) then
        begin
         BLOBExists := True;
         bs1 := NewCursor.InternalCreateBlobStream(True,i,bomWrite);
         try
           bs := Self.InternalCreateBlobStream(False,i,bomRead);
           try
             bs1.LoadFromStream(bs);
             bs1.Modified := True;
           finally
             bs.Free;
           end;
         except
           bs1.Size := 0;
           bs1.Modified := True;
         end;
         NewCursor.InternalCloseBLOB(i);
        end;
     NewCursor.InternalPost(True);
     if (bCalculateMinMax) then
      begin
       for i := 0 to FieldDefs.Count - 1 do
        if (IsAutoincFieldType(FieldDefs[i].AdvancedFieldType) and
            (not FieldDefs[i].AutoincCycled)) then
         begin
          NewCursor.GetFieldValue(CurrentValue,i,True,True);
          if (FieldDefs[i].AutoincIncrement >= 0) then
           begin
            if (bStart) then
             minmax[i] := CurrentValue.AsInt64
            else
             if (CurrentValue.AsInt64 > minmax[i]) then
              minmax[i] := CurrentValue.AsInt64;
           end // increment >= 0
          else
           begin
            if (bStart) then
             minmax[i] := CurrentValue.AsInt64
            else
             if (CurrentValue.AsInt64 < minmax[i]) then
              minmax[i] := CurrentValue.AsInt64;
           end; // increment < 0
         end;
       bStart := False;
      end;
     if (BLOBExists) then
      FBLOBStreams.Clear;
    except
     on e: Exception do
      Log := Log + ErrorLRepairTableCannotInsertRecord + FTableName + #13#10+ e.Message + #13#10;
    end;
  end;

  procedure TryToGetRecordsFromFirst;
  begin
   FirstPosition := True;
   LastPosition := False;
   repeat
    GetRecordResult := GetRecordBuffer(grmNext);
    if (GetRecordResult = grrOK) then
     CopyRecord
    else
     if (GetRecordResult = grrError) then
      raise EACRException.Create(11167,ErrorLRepairTableCannotGetRecord,
            [CurrentRecordID.PageNo,CurrentRecordID.PageItemNo,FTableName]);
   until (GetRecordResult = grrEOF);
  end; // TryToGetRecordsFromFirst


  procedure TryToGetRecordsFromLast(LastRecordID: TACRRecordID; bGetAll: Boolean);
  begin
   FirstPosition := False;
   LastPosition := True;
   repeat
    GetRecordResult := GetRecordBuffer(grmPrior);
    if (not bGetAll) then
     if ((CurrentRecordID.PageNo = LastRecordID.PageNo) and
         (CurrentRecordID.PageItemNo = LastRecordID.PageItemNo)) then
      break;
    if (GetRecordResult = grrOK) then
     begin
      NewCursor.CurrentRecordBuffer := FCurrentRecordBuffer;
      NewCursor.InternalInsert;
      try
       NewCursor.InternalPost(True);
      except
       on e: Exception do
        Log := Log + ErrorLRepairTableCannotInsertRecord + FTableName + #13#10+ e.Message + #13#10;
      end;
     end
    else
     if (GetRecordResult = grrError) then
      raise EACRException.Create(11168,ErrorLRepairTableCannotGetRecord,
            [CurrentRecordID.PageNo,CurrentRecordID.PageItemNo,FTableName]);
   until (GetRecordResult = grrBOF);
  end; // TryToGetRecordsFromLast

begin
  Self.Repair := True;
  bStart := True;
  bCalculateMinMax := FieldDefs.AutoIncFieldsExists;
  if (bCalculateMinMax) then
   begin
    SetLength(minmax,FieldDefs.Count);
    CurrentValue := TACRVariant.Create;
   end;
  try
    FCurrentRecordBuffer := AllocateRecordBuffer;
    NewCursor.FCurrentRecordBuffer := NewCursor.AllocateRecordBuffer;
    try
      // try to get records from beginning of the table
      try
        TryToGetRecordsFromFirst;
      except
       on e: Exception do
        begin
          Log := Log + ErrorLRepairTableCannotGetRecordsFromFirst + FTableName + #13#10 + e.Message + #13#10;
          try
           TryToGetRecordsFromLast(CurrentRecordID,FirstPosition);
          except
           on e: Exception do
            begin
             Log := Log + ErrorLRepairTableCannotGetRecordsFromLast + FTableName + #13#10+ e.Message + #13#10;
            end;
          end;
        end;
      end;
    finally
      Self.Repair := False;
      Self.FreeRecordBuffer(FCurrentRecordBuffer);
      NewCursor.FreeRecordBuffer(NewCursor.FCurrentRecordBuffer);
    end;
  finally
   if (bCalculateMinMax) then
    begin
     if (not bStart) then
       for i := 0 to FieldDefs.Count - 1 do
        if (IsAutoincFieldType(FieldDefs[i].AdvancedFieldType) and
            (not FieldDefs[i].AutoincCycled)) then
         NewCursor.SetLastAutoincValue(minmax[i],i);
     CurrentValue.Free;
     minmax := nil;
    end;
  end;
end; // TryToCopyRecords


//------------------------------------------------------------------------------
// Repair table
//------------------------------------------------------------------------------
function TACRLocalCursor.RepairTable(var Log: String; NewSession: Pointer): Boolean;
var NewCursor:  TACRLocalCursor;
begin
  if (FInMemory or FTemporary) then
   Result := True
  else
   begin
    if (not FExclusive) then
     raise EACRException.Create(11165,ErrorLTableIsNotOpenedExclusively,[FTableName]);
    try
      Self.OpenTableByFieldDefs(nil,nil,nil);
      Self.InternalInitFieldDefs;
    except
      Result := False;
      Exit;
    end;
    NewCursor := TACRLocalCursor.Create;
    try
      NewCursor.InMemory := False;
      NewCursor.Temporary := False;
      NewCursor.Exclusive := True;
      if (NewSession <> nil) then
       NewCursor.Session := NewSession
      else
       NewCursor.Session := Self.Session;
      repeat
       NewCursor.TableName := GetTemporaryName(FTableName+'_REPAIRED_');
      until (not NewCursor.Session.TableExists(NewCursor.TableName));
      // create table
      try
        NewCursor.CreateTable(Self.FFieldDefs,Self.IndexDefs,Self.ConstraintDefs);
        NewCursor.OpenTableByFieldDefs(nil,nil,nil);
        NewCursor.InternalInitFieldDefs;
        Result := True;
      except
        Log := Log + ErrorLRepairTableCannotCreateTable + NewCursor.TableName + #13#10;
        Result := False;
        Exit;
      end;
      try
        TryToCopyRecords(NewCursor,Log);
        // update sequence values
{        
        if (NewCursor.GetRecordCount > 0) then
          for i := 0 to NewCursor.FieldDefs.Count - 1 do
           if (IsAutoincFieldType(NewCursor.FieldDefs.Items[i].AdvancedFieldType)) then
            if (not NewCursor.FieldDefs.Items[i].AutoincCycled) then
             begin
              x := NewCursor.LastAutoincValue(i)+NewCursor.FieldDefs.Items[i].AutoincIncrement;
              y := x;
              Name := NewCursor.FieldDefs.Items[i].Name;
              while (NewCursor.Locate(Name,y,false,false)) do
               Inc(y);
              if (x <> y) then
               NewCursor.SetLastAutoincValue(y-NewCursor.FieldDefs.Items[i].AutoincIncrement,i);
             end;
}
      finally
        NewCursor.CloseTable;
        Self.CloseTable;
        Self.DeleteTable;
        NewCursor.RenameTable(FTableName);
      end;
    finally
      NewCursor.Free;
    end;
   end;
end; // RepairTable


//------------------------------------------------------------------------------
// load table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.LoadTableFromStream(
                        Stream:               TStream
                       );
begin
//  CloseTable;
  FCreateTableStarted := True;
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   FTableData := FDatabaseData.FindOrCreateTableData(Self);
   if (FTableData = nil) then
    raise EACRException.Create(10154,ErrorLTableDataNotFound,[TableName]);
   FTableData.LoadTableFromStream(Stream);
   FTableName := FTableData.TableName;
   FTableData := nil;
  finally
   FCreateTableStarted := False;
   FDatabaseData.Unlock;
  end;
end; // LoadTableFromStream;


//------------------------------------------------------------------------------
// save table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.SaveTableToStream(
                        Stream:               TStream;
                        CompressionAlgorithm: TACRCompressionAlgorithm = acaNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
begin
  FDatabaseData := TACRLocalSession(Session).DatabaseData;
  FDatabaseData.Lock;
  try
   FTableData := FDatabaseData.FindTableData(Self);
   if (FTableData = nil) then
    raise EACRException.Create(10154,ErrorLTableDataNotFound,[TableName]);
   FTableData.SaveTableToStream(Stream,CompressionAlgorithm,CompressionMode,BlockSize);
   FTableData.CloseTable(Self);
   FTableData := nil;
  finally
   FDatabaseData.Unlock;
  end;
end; // SaveTableToStream


//------------------------------------------------------------------------------
// drop table
//------------------------------------------------------------------------------
function TACRLocalCursor.GetDropTableCommand(aTableName: String): String;
begin
  Result := GetReservedWord(rwDROP) + Space +
            GetReservedWord(rwTABLE) + Space +
            aTableName + Space + GetReservedWord(rwCASCADE) + SemiColon + Crlf;
end; // GetDropTableCommand


//------------------------------------------------------------------------------
// get autoinc field definition
//------------------------------------------------------------------------------
function TACRLocalCursor.GetAutoIncFieldDefinition(FieldDef: TACRFieldDef): String;
var Cycled, MinValue, MaxValue: String;
begin
  if (FieldDef.AutoincCycled) then
   Cycled := GetReservedWord(rwCYCLED)
  else
   Cycled := GetReservedWord(rwNOCYCLED);
  if (FieldDef.AutoincMinValue = 0) then
   MinValue := GetReservedWord(rwNOMINVALUE)
  else
   MinValue := GetReservedWord(rwMINVALUE) + Space +
               IntToStr(FieldDef.AutoincMinValue);
  if (FieldDef.AutoincMaxValue = High(Int64)) then
   MaxValue := GetReservedWord(rwNOMAXVALUE)
  else
   MaxValue := GetReservedWord(rwMAXVALUE) + Space +
               IntToStr(FieldDef.AutoincMaxValue);
  Result := Result + Space + GetFieldTypeSQLName(aftAutoInc) + Space + LeftParenthesis +
            GetFieldTypeSQLName(FieldDef.AdvancedFieldType) + Space +
            GetReservedWord(rwINITIALVALUE) + Space +
            IntToStr(FieldDef.AutoincInitialValue)+ Space +
            GetReservedWord(rwINCREMENT) + Space +
            IntToStr(FieldDef.AutoincIncrement)+ Space +
            MinValue + Space + MaxValue + Space + Cycled +
            RightParenthesis;
end; // GetAutoIncFieldDefinition


//------------------------------------------------------------------------------
// blob field definition
//------------------------------------------------------------------------------
function TACRLocalCursor.GetBLOBFieldDefinition(FieldDef: TACRFieldDef): String;
begin
  Result := GetReservedWord(rwBLOBBLOCKSIZE) + Space +
            IntToStr(FieldDef.BLOBBlockSize) + Space +
            GetReservedWord(rwBLOBCOMPRESSIONALGORITHM) + Space +
            GetCompressionAlgorithmSQLName(FieldDef.BLOBCompressionAlgorithm) + Space +
            GetReservedWord(rwBLOBCOMPRESSIONMODE) + Space +
            IntToStr(FieldDef.BLOBCompressionMode);
end; // GetBLOBFieldDefinition


//------------------------------------------------------------------------------
// field definition for create table
//------------------------------------------------------------------------------
function TACRLocalCursor.GetFieldDefinition(FieldDef:             TACRFieldDef;
                                            UseBracketsForNames:  Boolean): String;
var i:      Integer;
    s:      String;
    check:  TACRConstraintDefCheck;
begin
  // get name
  if (UseBracketsForNames) then
    Result := LeftBracket + FieldDef.Name + RightBracket
  else
    Result := FieldDef.Name;
 // get data type
 if (IsAutoincFieldType(FieldDef.AdvancedFieldType)) then
  Result := Result + Space + GetAutoIncFieldDefinition(FieldDef)
 else
  begin
   Result := Result + Space + GetFieldTypeSQLName(FieldDef.AdvancedFieldType);
   if (IsStringFieldType(FieldDef.AdvancedFieldType) or
       IsBytesFieldType(FieldDef.AdvancedFieldType)) then
    Result := Result + Space + LeftParenthesis +
              IntToStr(FieldDef.FieldSize) + RightParenthesis;
  end;
 if (IsBLOBFieldType(FieldDef.AdvancedFieldType)) then
  Result := Result + Space + GetBLOBFieldDefinition(FieldDef);
 if (not FieldDef.DefaultValue.IsNull) then
  begin
    if (FieldDef.DefaultValue.IsStringDataType) then
     s := AnsiQuotedStr(FieldDef.DefaultValue.AsString,SingleQuote)
    else
     s := FieldDef.DefaultValue.AsString;
    Result := Result + Space + GetReservedWord(rwDEFAULT) + Space + s;
  end;
 // check if field is required or has minimum or maximum value restrictions
 s := AnsiUpperCase(FieldDef.Name);
 for i := 0 to FConstraintDefs.Count - 1 do
  if (AnsiUpperCase(TACRConstraintDefNotNull(FConstraintDefs[i]).ColumnName) = s) then
   begin
    if (FConstraintDefs[i].ConstraintType = ctNotNull) then
     Result := Result + Space + GetReservedWord(rwNOT) + Space + GetReservedWord(rwNULL)
    else
     if (FConstraintDefs[i].ConstraintType = ctCheck) then
      begin
       check := TACRConstraintDefCheck(FConstraintDefs[i]);
       if (not check.MinValue.IsNull) then
        begin
          if (check.MinValue.IsStringDataType) then
           s := AnsiQuotedStr(check.MinValue.AsString,SingleQuote)
          else
           s := check.MinValue.AsString;
          Result := Result + Space + GetReservedWord(rwMINVALUE) + Space + s;
        end;
       if (not check.MaxValue.IsNull) then
        begin
          if (check.MaxValue.IsStringDataType) then
           s := AnsiQuotedStr(check.MaxValue.AsString,SingleQuote)
          else
           s := check.MaxValue.AsString;
          Result := Result + Space + GetReservedWord(rwMAXVALUE) + Space + s;
        end;
      end; // Check
   end;
end; // GetFieldDefinition


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
function TACRLocalCursor.GetCreateTableCommand(aTableName: String; UseBracketsForNames:  Boolean): String;
var i,j:        Integer;
    FieldName:  String;
    IndexName:  String;
    TableName:  String;
    KeyName:    String;
    MatchType:  String;
    Action:     String;
    IndexDef:   TACRIndexDef;
    FKDef:      TACRConstraintDefForeignKey;
begin
  Result := GetReservedWord(rwCREATE) + Space +
            GetReservedWord(rwTABLE) + Space +
            aTableName + Space +
            LeftParenthesis + Crlf;
  for i := 0 to FFieldDefs.Count - 1 do
   begin
    if (i > 0) then
     Result := Result + Comma + Crlf;
    Result := Result + Tab;
    Result := Result +
      GetFieldDefinition(FFieldDefs[i],UseBracketsForNames);
   end;
  // add primary key
  for i := 0 to FIndexDefs.Count - 1 do
   if (FIndexDefs[i].Primary) then
    begin
     IndexDef := FIndexDefs[i];
     if (UseBracketsForNames) then
      IndexName := LeftBracket + IndexDef.Name + RightBracket
     else
      IndexName := IndexDef.Name;
     Result := Result + Comma + Crlf + GetReservedWord(rwPRIMARY) + Space +
               GetReservedWord(rwKEY) + Space +
               IndexName + Space + LeftParenthesis;
     for j := 0 to IndexDef.ColumnCount - 1 do
      begin
       if (UseBracketsForNames) then
        FieldName := LeftBracket + IndexDef.Columns[j].FieldName + RightBracket
       else
        FieldName := IndexDef.Columns[j].FieldName;
       if (IndexDef.Columns[j].Descending) then
        FieldName := FieldName + Space + GetReservedWord(rwDESC);
       if (IndexDef.Columns[j].CaseInsensitive) then
        FieldName := FieldName + Space + GetReservedWord(rwNOCASE);
       if (j = 0) then
        Result := Result + FieldName
       else
        Result := Result + Comma + Space + FieldName;
      end;
      Result := Result + RightParenthesis;
    end; // Primary index
  // foreign keys
  for i := 0 to FConstraintDefs.Count-1 do
   if (FConstraintDefs.Items[i].ConstraintType = ctFK) then
    begin
     FKDef := TACRConstraintDefForeignKey(FConstraintDefs.Items[i]);
     if (UseBracketsForNames) then
      KeyName := LeftBracket + FKDef.Name + RightBracket
     else
      KeyName := FKDef.Name;
     Result := Result + Comma + Crlf + GetReservedWord(rwFOREIGN) + Space +
               GetReservedWord(rwKEY) + Space +
               KeyName + Space + LeftParenthesis;
     for j := 0 to High(FKDef.Columns) do
      begin
       if (UseBracketsForNames) then
        FieldName := LeftBracket + FKDef.Columns[j].ColumnName + RightBracket
       else
        FieldName := FKDef.Columns[j].ColumnName;
       if (j = 0) then
        Result := Result + FieldName
       else
        Result := Result + Comma + Space + FieldName;
      end;
     if (UseBracketsForNames) then
      TableName := LeftBracket + FKDef.ReferencedTableName + RightBracket
     else
      TableName := FKDef.ReferencedTableName;

     MatchType := '';
     if (FKDef.MatchType = cfkmtFull) then
      MatchType := GetReservedWord(rwMATCH) + Space + GetReservedWord(rwFULL) + Space
     else
     if (FKDef.MatchType = cfkmtPartial) then
      MatchType := GetReservedWord(rwMATCH) + Space + GetReservedWord(rwPARTIAL) + Space;

     Result := Result + RightParenthesis + Space +
               GetReservedWord(rwREFERENCES) + Space + TableName + Space +
               MatchType;
     case FKDef.DeleteAction of
      cfkaCascade: Action := GetReservedWord(rwCASCADE);
      cfkaSetNull: Action := GetReservedWord(rwSET) + Space + GetReservedWord(rwNULL);
      cfkaSetDefault: Action := GetReservedWord(rwSET) + Space + GetReservedWord(rwDEFAULT);
      cfkaNoAction: Action := GetReservedWord(rwNO) + Space + GetReservedWord(rwACTION);
     end;
     Result := Result + GetReservedWord(rwON) + Space + GetReservedWord(rwDELETE)
               + Space + Action + Space;
     case FKDef.UpdateAction of
      cfkaCascade: Action := GetReservedWord(rwCASCADE);
      cfkaSetNull: Action := GetReservedWord(rwSET) + Space + GetReservedWord(rwNULL);
      cfkaSetDefault: Action := GetReservedWord(rwSET) + Space + GetReservedWord(rwDEFAULT);
      cfkaNoAction: Action := GetReservedWord(rwNO) + Space + GetReservedWord(rwACTION);
     end;
     Result := Result + GetReservedWord(rwON) + Space + GetReservedWord(rwUPDATE)
               + Space + Action + Space;
    end;
  Result := Result + Crlf + RightParenthesis + SemiColon + Crlf;
end; // GetCreateTableCommand


//------------------------------------------------------------------------------
// create index
//------------------------------------------------------------------------------
function TACRLocalCursor.GetCreateIndexCommand(aTableName:          String;
                                               IndexName:           String;
                                               IndexDef:            TACRIndexDef;
                                               UseBracketsForNames: Boolean): String;
var Unique:     String;
    i:          Integer;
    FieldName:  String;
begin
  Unique := '';
  if (IndexDef.Unique) then
   Unique := GetReservedWord(rwUNIQUE) + Space;
  Result := GetReservedWord(rwCREATE) + Space + Unique +
            GetReservedWord(rwINDEX) + Space + IndexName + Space +
            GetReservedWord(rwON) + Space + aTableName + Space + LeftParenthesis + Crlf;
  for i := 0 to IndexDef.ColumnCount - 1 do
   begin
    if (UseBracketsForNames) then
     FieldName := LeftBracket + IndexDef.Columns[i].FieldName + RightBracket
    else
     FieldName := IndexDef.Columns[i].FieldName;
    if (IndexDef.Columns[i].Descending) then
     FieldName := FieldName + Space + GetReservedWord(rwDESC);
    if (IndexDef.Columns[i].CaseInsensitive) then
     FieldName := FieldName + Space + GetReservedWord(rwNOCASE);
    if (i > 0) then
     Result := Result + Comma + Crlf;
    Result := Result  + Tab + FieldName;
   end;
  Result := Result + Crlf + RightParenthesis + SemiColon + Crlf;  
end; // GetCreateIndexCommand


//------------------------------------------------------------------------------
// drop index
//------------------------------------------------------------------------------
function TACRLocalCursor.GetDropIndexCommand(aTableName: String; IndexName: String): String;
begin
  Result := GetReservedWord(rwDROP) + Space +
            GetReservedWord(rwINDEX) + Space +
            GetReservedWord(rwIF) + Space +
            GetReservedWord(rwEXISTS) + Space +
            aTableName + Dot + IndexName + SemiColon + Crlf;
end; // GetDropIndexCommand


//------------------------------------------------------------------------------
// get BLOB field value
//------------------------------------------------------------------------------
function TACRLocalCursor.GetSQLBLOBFieldValue(FieldNo: Integer): String;
var bs:   TACRStream;
    buf:  PChar;
    size: Integer;

  procedure CloseBlobStream;
  begin
   bs.Free;
   bs := nil;
  end;


begin
  bs := Self.InternalCreateBlobStream(False,FieldNo,bomRead);
  try
   size := bs.Size;
   if (size <= 0) then
    begin
     Result := 'NULL';
     CloseBlobStream;
     Exit;
    end;
   buf := MemoryManager.GetMem(size);
   try
     bs.ReadBuffer(buf^,size);
     CloseBlobStream;
     Result := GetReservedWord(rwTOBLOB) + Space + LeftParenthesis +
               SingleQuote + ACRBinaryToMIME64(buf,size) + SingleQuote + Comma +
               GetReservedWord(rwMIME64) + RightParenthesis;
   finally
     MemoryManager.FreeAndNilMem(buf);
   end;
  finally
   if (bs <> nil) then
    CloseBlobStream;
  end;
end; // GetBLOBFieldValue


//------------------------------------------------------------------------------
// get field value
//------------------------------------------------------------------------------
function TACRLocalCursor.GetSQLFieldValue(var v: TACRVariant; FieldNo: Integer; ExportBLOBFields: Boolean): String;
var c: Char;
    y,m,d,h,n,s,z: Word;
begin
  GetFieldValue(v,FieldNo,True,False);
  if (v.IsNull) then
   Result := 'NULL'
  else
  if (v.IsDateTimeDataType) then
   begin
    DecodeDateTime(v.AsTDateTime,y,m,d,h,n,s,z);
    Result := 'TODATE('''+IntToStr(m)+'/'+IntToStr(d)+'/'+IntToStr(y)+' '+
                      IntToStr(h)+':'+IntToStr(n)+':'+IntToStr(s)+':'+IntToStr(z)+
                      ''',''M/D/YYYY H24:N:S:Z'')';
   end
  else
  if (v.IsStringDataType) then
   Result := AnsiQuotedStr(v.AsString,SingleQuote)
  else
  if (v.IsNumericDataType) then
   begin
    c := DecimalSeparator;
    DecimalSeparator := '.';
    try
      Result := v.AsString;
    finally
      DecimalSeparator := c;
    end;
   end
  else
   Result := v.AsString;
end; // GetSQLFieldValue


//------------------------------------------------------------------------------
// insert
//------------------------------------------------------------------------------
function TACRLocalCursor.GetInsertCommand(var v: TACRVariant; aTableName: String; UseBracketsForNames:  Boolean; ExportBLOBFields: Boolean): String;
var FieldsList:   String;
    i:            Integer;
    FieldValues:  String;
    bBlobFound:   Boolean;
    FieldName:    String;
begin
  FieldsList := '';
  FieldValues := '';
  if (not ExportBLOBFields) then
   begin
    bBlobFound := False;
    for i := 0 to FieldDefs.Count - 1 do
     if (IsBLOBFieldType(FieldDefs[i].AdvancedFieldType)) then
       bBlobFound := True
     else
      begin
       if (UseBracketsForNames) then
        FieldName := LeftBracket + FieldDefs[i].Name + RightBracket
       else
        FieldName := FieldDefs[i].Name;
       if (FieldsList <> '') then
        FieldsList := FieldsList + Comma + Space + FieldName
       else
        FieldsList := FieldName;
      end;
    if (not bBlobFound) then
     FieldsList := ''
    else
     FieldsList := LeftParenthesis + FieldsList + RightParenthesis + Space;
   end;
  for i := 0 to FieldDefs.Count - 1 do
   begin
    if ((not ExportBLOBFields) and (IsBLOBFieldType(FieldDefs[i].AdvancedFieldType))) then
     continue;
    if (i > 0) then
     FieldValues := FieldValues + Comma + Crlf;
    FieldValues := FieldValues + Tab;
    if (IsBLOBFieldType(FieldDefs[i].AdvancedFieldType)) then
     begin
      FieldValues := FieldValues + GetSQLBLOBFieldValue(i);
      bBlobFound := True;
     end
    else
     FieldValues := FieldValues + GetSQLFieldValue(v,i,ExportBLOBFields);
   end;
  Result := GetReservedWord(rwINSERT) + Space +
            GetReservedWord(rwINTO) + Space +
            aTableName + Space +
            FieldsList +
            GetReservedWord(rwVALUES) + Space + LeftParenthesis + Crlf +
            FieldValues + Crlf +
            RightParenthesis + SemiColon + Crlf;
  if (bBlobFound and ExportBLOBFields) then
    FBLOBStreams.Clear;
end; // GetInsertCommand


//------------------------------------------------------------------------------
// export table to SQL
//------------------------------------------------------------------------------
function TACRLocalCursor.ExportTableToSQL(
                            ExportStructure:      Boolean;
                            AddDropTableCommand:  Boolean;
                            ExportIndexes:        Boolean;
                            AddDropIndexCommand:  Boolean;
                            ExportData:           Boolean;
                            ExportBLOBFields:     Boolean;
                            UseBracketsForNames:  Boolean
                        ): String;
var aTableName:   String;
    IndexName:    String;
    i:            Integer;
    OldRecBuffer: PChar;
    Buf:          PChar;
    OldPos:       Pointer;
    v:            TACRVariant;
begin
  Result := '';
  if (UseBracketsForNames) then
   aTableName := LeftBracket + FTableName + RightBracket
  else
   aTableName := FTableName;
  if (FInMemory) then
   aTableName := GetReservedWord(rwMEMORY)+ Space + aTableName;
  if (ExportStructure) then
   begin
    if (AddDropTableCommand) then
     Result := Result + GetDropTableCommand(aTableName);
    Result := Result + GetCreateTableCommand(aTableName,UseBracketsForNames);
   end;
  if (ExportIndexes) then
   begin
    for i := 0 to FIndexDefs.Count - 1 do
     begin
      if (FIndexDefs[i].Primary) then continue;
      if (UseBracketsForNames) then 
       IndexName := LeftBracket + FIndexDefs[i].Name + RightBracket 
      else
       IndexName := FIndexDefs[i].Name;
      if (AddDropIndexCommand) then 
       Result := Result + GetDropIndexCommand(aTableName,IndexName); 
      Result := Result + GetCreateIndexCommand(aTableName,IndexName,FIndexDefs[i],UseBracketsForNames);
     end;
   end; 
  if (ExportData) then
   begin 
    OldPos := SavePosition;
    OldRecBuffer := CurrentRecordBuffer; 
    CurrentRecordBuffer := AllocateRecordBuffer;
    v := TACRVariant.Create; 
    ClearBLOBStreams(False);
    try 
      InternalFirst; 
      while (GetRecordBuffer(grmNext) = grrOK) do
       Result := Result + GetInsertCommand(v,aTableName,UseBracketsForNames,ExportBLOBFields); 
    finally 
      v.Free; 
      Buf := CurrentRecordBuffer; 
      FreeRecordBuffer(Buf); 
      RestorePosition(OldPos);
      FreePosition(OldPos);
      CurrentRecordBuffer := OldRecBuffer;
    end;
   end; // Export Data
end; // ExportTableToSQL 
	
 
//------------------------------------------------------------------------------
// open table 
//------------------------------------------------------------------------------ 
procedure TACRLocalCursor.InternalInitFieldDefs;
var 
    i: Integer;
begin 
 if (FFieldDefs <> nil) then 
  FFieldDefs.Free;
 FFieldDefs := TACRFieldDefs.Create;
 if (not FSettingProjection) then 
  begin 
   if (FVisibleFieldDefs <> nil) then 
    FVisibleFieldDefs.Free; 
   FVisibleFieldDefs := TACRFieldDefs.Create; 
  end; 
 
	
{$IFDEF DEBUG_TRACE_DATASET}
aaWriteToLog('TACRLocalCursor.InternalInitFieldDefs, TableData.FieldDefs.Count = '+ 
IntToStr(FTableData.FieldManager.FieldDefs.Count));
{$ENDIF}
 
	
 FFieldDefs.Assign(FTableData.FieldManager.FieldDefs);
 FFieldDefs.RecalcFieldOffsets; 
 
 FieldValuesOffset := FFieldDefs[0].MemoryOffset; 
 BookmarkOffset := FFieldDefs[FFieldDefs.Count-1].MemoryOffset +
                   FFieldDefs[FFieldDefs.Count-1].MemoryDataSize;
 KeyOffset := BookmarkOffset; 
 KeyBufferSize := BookmarkOffset + sizeof(TACRKeyBuffer);
 RecordSize := BookmarkOffset + SizeOf(TACRBookmarkInfo);
 RecordBufferSize := RecordSize; 
 CalculatedFieldsOffset := RecordSize; 
	
 if (not FSettingProjection) then
  begin
   FVisibleFieldDefs.Assign(FTableData.FieldManager.FieldDefs); 
   // create default fields order 
   for i := 0 to VisibleFieldDefs.Count-1 do 
    FVisibleFieldDefs[i].FieldNoReference := i; 
  end;
 
 // Constraints... 
 if (FConstraintDefs <> nil) then 
   FConstraintDefs.Free;
 FConstraintDefs := TACRConstraintDefs.Create;
 FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
end; // InternalInitFieldDefs 
	
	
//------------------------------------------------------------------------------
// open table 
//------------------------------------------------------------------------------ 
procedure TACRLocalCursor.OpenTableByFieldDefs(
                                    FieldDefs: TACRFieldDefs;
                                    IndexDefs: TACRIndexDefs;
                                    ConstraintDefs: TACRConstraintDefs
                                   );
begin
{$IFDEF DEBUG_TRACE_DATASET} 
aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs start');
{$ENDIF}
 FreeSavedPosition; 
 if (not IsOpen) then
  begin
    FDatabaseData := TACRLocalSession(Session).DatabaseData; 
    FDatabaseData.Lock; 
    try 
     if (InMemory or Temporary) then 
       FTableData := FDatabaseData.FindTableData(Self)
     else 
       FTableData := FDatabaseData.FindOrCreateTableData(Self); 
{$IFDEF DEBUG_TRACE_DATASET}
if (FTableData = nil) then
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs tabledata = nil')
else
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs tabledata = '+FTableData.TableName); 
{$ENDIF} 
     if (FTableData = nil) then
      begin
        if ({IsDesignMode and }InMemory) then 
         begin 
{$IFDEF DEBUG_TRACE_DATASET} 
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs creating tabledata...');
{$ENDIF}
          FTableData := FDatabaseData.CreateTableData(Self);
{$IFDEF DEBUG_TRACE_DATASET}
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs creating table...');
{$ENDIF} 
          FTableData.CreateTable(Self,FieldDefs,IndexDefs, ConstraintDefs);
{$IFDEF DEBUG_TRACE_DATASET} 
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs table created!');
{$ENDIF}
         end
        else
         raise EACRException.Create(10028,ErrorLTableDataNotFound,[TableName]);
      end;
{$IFDEF DEBUG_TRACE_DATASET}
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs opening tabledata...');
{$ENDIF}
     FTableData.OpenTable(Self);
{$IFDEF DEBUG_TRACE_DATASET}
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs tabledata opened successfully!');
{$ENDIF}
     FirstPosition := True;
     LastPosition := False;

     if (FIndexDefs <> nil) then
      FIndexDefs.Free;
     FIndexDefs := TACRIndexDefs.Create;
{$IFDEF DEBUG_TRACE_DATASET}
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs indexdefs = '+IntToStr(FTableData.IndexManager.IndexDefs.Count));
{$ENDIF}
     FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
{$IFDEF DEBUG_TRACE_DATASET}
  aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs indexdefs2 = '+IntToStr(FIndexDefs.Count));
{$ENDIF}
     if (FConstraintDefs <> nil) then
      FConstraintDefs.Free;
     FConstraintDefs := TACRConstraintDefs.Create;
     FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
     if (FBLOBStreams <> nil) then
      FBLOBStreams.Free;
     FBLOBStreams := TList.Create;
     if (RecordBitmap <> nil) then
      TACRRecordBitmap(RecordBitmap).Free;
     RecordBitmap := TACRRecordBitmap.Create(FTableData);
     IsOpen := True;
     FDoNotCloseTableData := False;
    finally
     FDatabaseData.Unlock;
    end;
  end;
{$IFDEF DEBUG_TRACE_DATASET}
aaWriteToLog('TACRLocalCursor.OpenTableByFieldDefs finish');
{$ENDIF}
end;// OpenTable


//------------------------------------------------------------------------------
// open table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.OpenTable(aTableData: TACRTableData);
begin
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.OpenTable(aTableData) start');
{$ENDIF}
 FreeSavedPosition;
 FTableData := aTableData;
 FirstPosition := True;
 LastPosition := False;
 if (FIndexDefs <> nil) then
  FIndexDefs.Free;
 FIndexDefs := TACRIndexDefs.Create;
 FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
 if (FConstraintDefs <> nil) then
  FConstraintDefs.Free;
 FConstraintDefs := TACRConstraintDefs.Create;
 FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
 if (FBLOBStreams <> nil) then
  FBLOBStreams.Free;
 FBLOBStreams := TList.Create;
 InternalInitFieldDefs;
 if (RecordBitmap <> nil) then
  TACRRecordBitmap(RecordBitmap).Free;
 RecordBitmap := TACRRecordBitmap.Create(FTableData);
 IsOpen := True;
 FDoNotCloseTableData := True;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.OpenTable(aTableData) finish');
{$ENDIF}
end; // OpenTable


//------------------------------------------------------------------------------
// close table
//------------------------------------------------------------------------------
procedure TACRLocalCursor.CloseTable;
begin
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable start');
{$ENDIF}
 FreeSavedPosition;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 0');
{$ENDIF}
 if (FTableData <> nil) then
  begin
   if (not FDoNotCloseTableData) then
    begin
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 01');
{$ENDIF}
     FTableData.CloseTable(Self);
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 02');
{$ENDIF}
     if (FTemporary) then
      FTableData := nil;
    end;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 03');
{$ENDIF}

   if (IsDesignMode) and (FInMemory) and (FTableData <> nil) and (FTableData.CursorList.Count <= 0) then
     FTableData.DeleteTable(Session,True)
   else
     FTableData := nil;
  end;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 1');
{$ENDIF}
 if (RecordBitmap <> nil) then
  TACRRecordBitmap(RecordBitmap).Free;
 RecordBitmap := nil;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 2');
{$ENDIF}
 if (FilterExpression <> nil) then
  TACRExpression(FilterExpression).Free;
 FilterExpression := nil;
 if (FSQLFilterExpression <> nil) then
  TACRExpression(FSQLFilterExpression).Free;
 FSQLFilterExpression := nil; 
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 3');
{$ENDIF}
 if (FFieldDefs <> nil) then
   FFieldDefs.Free;
 FFieldDefs := nil;
 if (not FSettingProjection) then
  begin
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 4');
{$ENDIF}
   if (FVisibleFieldDefs <> nil) then
    FVisibleFieldDefs.Free;
   FVisibleFieldDefs := nil;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 44');
{$ENDIF}
  end;
 if (FIndexDefs <> nil) then
   FIndexDefs.Free;
 FIndexDefs := nil;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable 5');
{$ENDIF}
 ClearBLOBStreams(False);
 if (FBLOBStreams <> nil) then
   FBLOBStreams.Free;
 FBLOBStreams := nil;
 // Constraints
 if (FConstraintDefs <> nil) then
   FConstraintDefs.Free;
 FConstraintDefs := nil;
{$IFDEF DEBUG_TRACE_DATASET}
 aaWriteToLog('TACRLocalCursor.CloseTable finish');
{$ENDIF}
end; // CloseTable


//------------------------------------------------------------------------------
// initialize record buffer
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalInitRecord(RecordBuffer: TACRRecordBuffer; InsertMode: Boolean);
var x,y: Integer;
begin
 // set null flags for all fields
 y := FFieldDefs.Count;
 x := y div 8;
 if (y mod 8 > 0) then
  Inc(x);
 FillChar(RecordBuffer^,x,$FF);
 // set Default Values...
 if ((not FTemporary) and (InsertMode)) then
   FTableData.ApplyDefaultValuesToRecordBuffer(Self,RecordBuffer,InsertMode);
end; // InternalInitRecord


//------------------------------------------------------------------------------
// GetIndexDefs
//------------------------------------------------------------------------------
function TACRLocalCursor.GetIndexDefs: TACRIndexDefs;
begin
  if (FIndexDefs <> nil) then
   FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
  Result := inherited GetIndexDefs;
end;// GetIndexDefs


//------------------------------------------------------------------------------
// add index
//------------------------------------------------------------------------------
procedure TACRLocalCursor.AddIndex(IndexDef: TACRIndexDef);
begin
  if (FIndexDefs.GetIndexDefByName(IndexDef.Name) <> nil) then
   raise EACRException.Create(20022, ErrorAIndexAlreadyExists, [IndexDef.Name]);
  FTableData.AddIndex(IndexDef, Self);
  FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
  //if (not FTemporary) then
  FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
end;// AddIndex


//------------------------------------------------------------------------------
// DeleteIndex
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DeleteIndex(Name: String);
begin
  if (FIndexDefs.GetIndexDefByName(Name) = nil) then
   raise EACRException.Create(20021, ErrorACannotDeleteIndex, [Name]);
  // Delete Index
  FTableData.DeleteIndex(FIndexDefs.GetIndexDefByName(Name).ObjectID, Self);
  FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
  //if (not FTemporary) then
  FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
end;// DeleteIndex


//------------------------------------------------------------------------------
// DeleteAllIndexes
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DeleteAllIndexes;
begin
  FTableData.DeleteAllIndexes(Self);
  FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
  //if (not FTemporary) then
   FConstraintDefs.Assign(FTableData.ConstraintManager.ConstraintDefs);
end;// DeleteAllIndexes


//------------------------------------------------------------------------------
// return index name of the index or '' if not found
//------------------------------------------------------------------------------
function TACRLocalCursor.FindIndex(FieldNamesList, AscDescList, CaseSensitivityList: TStringList): String;
begin
  Result := FTableData.FindIndex(Self, FieldNamesList, AscDescList, CaseSensitivityList);
end; // FindIndex


//------------------------------------------------------------------------------
// return index name
//------------------------------------------------------------------------------
function TACRLocalCursor.FindOrCreateIndex(FieldNamesList, AscDescList, CaseSensitivityList: TStringList): String;
begin
  raise EACRException.Create(11394,ErrorLOperationIsNotSupported);
  Result := FTableData.FindOrCreateIndex(Self, FieldNamesList, AscDescList, CaseSensitivityList);
  FIndexDefs.Assign(FTableData.IndexManager.IndexDefs);
end; // FindOrCreateIndex


//------------------------------------------------------------------------------
// return true if current record exists
//------------------------------------------------------------------------------
function TACRLocalCursor.IsRecordExists: Boolean;
begin
  Result := FTableData.IsRecordExists(Self);
end; // IsRecordExists


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
function TACRLocalCursor.GetRecordBuffer(
              GetRecordMode:  TACRGetRecordMode
              ): TACRGetRecordResult;
begin
  Result := FTableData.GetRecordBuffer(Self,GetRecordMode);
end; // GetRecordBuffer


//------------------------------------------------------------------------------
// returns record count
//------------------------------------------------------------------------------
function TACRLocalCursor.GetRecordCount: TACRRecordNo;
begin
  Result := FTableData.GetRecordCount(Self);
end; // GetRecordCount


//------------------------------------------------------------------------------
// go to record
//------------------------------------------------------------------------------
procedure TACRLocalCursor.SetRecNo(Value: Int64);
begin
  FTableData.SetRecNo(Self,Value);
end; // SetRecNo


//------------------------------------------------------------------------------
// return current record number
//------------------------------------------------------------------------------
function TACRLocalCursor.GetRecNo: Int64;
begin
  Result := FTableData.GetRecNo(Self);
end; // GetRecNo


//------------------------------------------------------------------------------
// insert record
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalInsert;
begin
;
end; // InternalInsert


//------------------------------------------------------------------------------
// edit record
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalEdit;
begin
  ErrorCode := ACR_ERR_OK;
  if (FTableData is TACRTemporaryTableData) then
    ErrorCode := ACR_ERR_UPDATE_RECORD_PROHIBITED
  else
    FTableData.EditRecord(Self);
end; // InternalEdit


//------------------------------------------------------------------------------
// cancels updates
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalCancel(ToInsert: Boolean);
begin
  ErrorCode := ACR_ERR_OK;
  if (FTableData is TACRTemporaryTableData) then
    ErrorCode := ACR_ERR_CANCEL_PROHIBITED
  else
    begin
     ClearBLOBStreams(True);
     FTableData.CancelRecord(Self,ToInsert);
    end;
end; // InternalCancel


//------------------------------------------------------------------------------
// update record
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalPost(ToInsert: Boolean);
var TempCursor: TACRCursor;
    Bookmark:   PACRBookmarkInfo;
begin
 // Check Constraints
 try
   FTableData.CheckConstraints(
          Self,
          Session.SessionID, CurrentRecordBuffer, EditRecordBuffer,
          ToInsert, CurrentRecordID);
 except
   on e: Exception do
     begin
       ErrorCode := ACR_ERR_CONSTRAINT_VIOLATED;
       ErrorMessage := e.Message;
       Exit;
     end;
 end;
 TempCursor := Self;
 ClearBLOBStreams(True);
 if (ToInsert) then
  begin
   // insert
   if (FTableData.InsertRecord(TempCursor)) then
    ErrorCode := ACR_ERR_OK
   else
    begin
      ErrorCode := ACR_ERR_INSERT_RECORD;
      ErrorMessage := ErrorLAddingRecord;
    end;
   // write bookmark info to new record buffer
   Bookmark := PACRBookmarkInfo(CurrentRecordBuffer + BookmarkOffset);
   Bookmark^.BookmarkData := CurrentRecordID;
   Bookmark^.BookmarkFlag := abfInserted;
  end // insert
 else
  begin
   // update
   Move(PChar(FCurrentRecordBuffer + BookmarkOffset)^,
    CurrentRecordID,sizeof(TACRRecordID));
   if (FTableData.UpdateRecord(TempCursor)) then
    ErrorCode := ACR_ERR_OK
   else
    begin
     ErrorCode := ACR_ERR_Update_RECORD;
     ErrorMessage := ErrorLUpdatingRecord;
    end;
  end; // update;
end; // InternalPost


//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalDelete;
begin
 if (FTableData is TACRTemporaryTableData) then
  ErrorCode := ACR_ERR_DELETE_RECORD_PROHIBITED
 else
  begin
   // delete
   Move(PChar(FCurrentRecordBuffer + BookmarkOffset)^,
    CurrentRecordID,sizeof(TACRRecordID));
   if (FTableData.DeleteRecord(Self)) then
    ErrorCode := ACR_ERR_OK
   else
    ErrorCode := ACR_ERR_DELETE_RECORD;
  end;
end; // InternalDelete


//------------------------------------------------------------------------------
// delete all visible records
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DeleteVisibleRecords;
begin
  FTableData.DeleteVisibleRecords(Self);
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// update visible records
//------------------------------------------------------------------------------
procedure TACRLocalCursor.UpdateVisibleRecords(FieldNames:   TStringList;
                                               values:       array of TACRVariant;
                                               SkipFKCheck:  Boolean = False
                                               );
begin
  FTableData.UpdateVisibleRecords(Self,FieldNames,values,SkipFKCheck);
end; // UpdateVisibleRecords

//------------------------------------------------------------------------------
// disable record bitmap
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DisableRecordBitmap;
begin
 if (RecordBitmap <> nil) then
  TACRRecordBitmap(RecordBitmap).Active := False;
end; // DisableRecordBitmap


//------------------------------------------------------------------------------
// activate filters
//------------------------------------------------------------------------------
procedure TACRLocalCursor.ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList);
begin
  raise EACRException.Create(11393,ErrorLOperationIsNotSupported);
  FIndexName := FindOrCreateIndex(FieldNamesList, AscDescList, CaseSensitivityList);
  SetIndexName(FIndexName);
  DistinctFieldCount := FieldNamesList.Count;
  DisableRecordBitmap;
end; // ApplyDistinct


//------------------------------------------------------------------------------
// apply projection
//------------------------------------------------------------------------------
procedure TACRLocalCursor.ApplyProjection(FieldNamesList, AliasList: TStringList);
var i,j: Integer;
begin
  if (FieldNamesList.Count <> AliasList.Count) then
    raise EACRException.Create(10321,ErrorLDifferentListsLength,
            [FieldNamesList.Count,AliasList.Count]);
  if (FieldNamesList.Count = 0) then
    raise EACRException.Create(10322,ErrorLNoFieldsInProjection);
  FVisibleFieldDefs.Clear;
  for i := 0 to FieldNamesList.Count -1 do
   begin
    j := FFieldDefs.GetDefNumberByName(FieldNamesList[i]);
    if (j < 0) then
      raise EACRException.Create(10323,ErrorLCannotFindField,[FieldNamesList[i]]);
    FVisibleFieldDefs.AddCreated.Assign(FFieldDefs[j]);
    FVisibleFieldDefs[FVisibleFieldDefs.Count-1].FieldNoReference := j;
    if (AliasList[i] <> '') then
     FVisibleFieldDefs[FVisibleFieldDefs.Count-1].Name := AliasList[i];
   end;
end;


//------------------------------------------------------------------------------
// activate filters
//------------------------------------------------------------------------------
procedure TACRLocalCursor.ActivateFilters(
                          FilterText:      String;
                          CaseInsensitive: Boolean;
                          PartialKey:      Boolean
                        );
begin
  if (FilterExpression <> nil) then
    TACRExpression(FilterExpression).Free;
  FilterExpression := TACRExpression.Create;
  TACRExpression(FilterExpression).ParseForFilter(Self,FilterText,CaseInsensitive,PartialKey);
  DisableRecordBitmap;
end; // ActivateFilters


//------------------------------------------------------------------------------
// deactivate filters
//------------------------------------------------------------------------------
procedure TACRLocalCursor.DeactivateFilters;
begin
  if (FilterExpression <> nil) then
    TACRExpression(FilterExpression).Free;
  FilterExpression := nil;
  DisableRecordBitmap;
end; // DeactivateFilters


//------------------------------------------------------------------------------
// locate
//------------------------------------------------------------------------------
function TACRLocalCursor.Locate(
                const KeyFields: String;
                const KeyValues: Variant;
                CaseInsensitive: Boolean;
                PartialKey:      Boolean
               ): Boolean;
var SearchExpression: TACRExpression;
begin
  SearchExpression := TACRExpression.Create;
  try
    SearchExpression.ParseForLocate(Self,KeyFields,KeyValues,CaseInsensitive,PartialKey);
    Result := FTableData.Locate(Self,SearchExpression);
  finally
    SearchExpression.Free;
  end;
end; // Locate


//------------------------------------------------------------------------------
// find key
//------------------------------------------------------------------------------
function TACRLocalCursor.FindKey(SearchCondition: TACRSearchCondition): Boolean;
begin
  Result := FTableData.FindKey(Self,SearchCondition);
end; // FindKey


//------------------------------------------------------------------------------
// reset range
//------------------------------------------------------------------------------
procedure TACRLocalCursor.ResetRange;
begin
  RangeStartBuffer := nil;
  RangeEndBuffer := nil;
  DisableRecordBitmap;
end; // ResetRange


//------------------------------------------------------------------------------
// apply range
//------------------------------------------------------------------------------
procedure TACRLocalCursor.ApplyRange(
                          StartBuffer, EndBuffer: TACRRecordBuffer;
                          StartKeyFieldCount:     Integer;
                          EndKeyFieldCount:       Integer;
                          StartExclusive:         Boolean;
                          EndExclusive:           Boolean
                        );
begin
  RangeStartBuffer := StartBuffer;
  RangeEndBuffer := EndBuffer;
  RangeStartKeyFieldCount := StartKeyFieldCount;
  RangeEndKeyFieldCount := EndKeyFieldCount;
  RangeStartExclusive := StartExclusive;
  RangeEndExclusive := EndExclusive;
  DisableRecordBitmap;
end; // ApplyRange


//------------------------------------------------------------------------------
// set SQL Filter
//------------------------------------------------------------------------------
procedure TACRLocalCursor.SetSQLFilter(FilterExpr: TObject);
begin
  SQLFilterExpression := FilterExpr;
  if SQLFilterExpression <> nil then
    TACRExpression(SQLFilterExpression).AssignCursor(self);
end; // SetSQLFilter



//------------------------------------------------------------------------------
// create blob stream
//------------------------------------------------------------------------------
function TACRLocalCursor.InternalCreateBlobStream(
              ToInsert: Boolean;
              FieldNo:  Integer;
              OpenMode: TACRBLOBOpenMode
              ):TACRStream;
var FieldNumber:      Integer;
    i:                Integer;
    LocalBLOBStream:  TACRLocalBLOBStream;
begin
 if (FieldNo >= FVisibleFieldDefs.Count) then
  raise EACRException.Create(10110,ErrorLInvalidFieldNumber,
    [FieldNo,FVisibleFieldDefs.Count]);
 if (FBLOBStreams = nil) then
  raise EACRException.Create(10113,ErrorLNilPointer);
 FieldNumber := FVisibleFieldDefs.Items[FieldNo].FieldNoReference;

 // find existing blob stream
 if ((OpenMode = bomReadWrite) or (OpenMode = bomWrite)) then
  for i := 0 to FBLOBStreams.Count -1 do
   begin
    LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
    if (LocalBLOBStream = nil) then
      raise EACRException.Create(10114,ErrorLNilPointer);
    if (LocalBLOBStream.FFieldNo = FieldNumber) then
     if ((LocalBLOBStream.OpenMode = bomReadWrite) or
         (LocalBLOBStream.OpenMode = bomWrite)) then
      raise EACRException.Create(10115,ErrorLBLOBFieldAlreadyOpened,[FieldNumber,
        FFieldDefs.Items[FieldNumber].Name]);
   end;

 Result := FTableData.InternalCreateBlobStream(Self,ToInsert,FieldNumber,OpenMode);
 if (Result <> nil) then
  FBLOBStreams.Add(Result);
end; // InternalCreateBlobStream


//------------------------------------------------------------------------------
// close blob
//------------------------------------------------------------------------------
procedure TACRLocalCursor.InternalCloseBLOB(FieldNo: Integer);
var FieldNumber:       Integer;
    i:                 Integer;
    LocalBLOBStream:   TACRLocalBLOBStream;
begin
 if (FieldNo >= FVisibleFieldDefs.Count) then
  raise EACRException.Create(10118,ErrorLInvalidFieldNumber,
    [FieldNo,FVisibleFieldDefs.Count]);
 if (FBLOBStreams = nil) then
  raise EACRException.Create(10119,ErrorLNilPointer);
 FieldNumber := FVisibleFieldDefs.Items[FieldNo].FieldNoReference;

 i := 0;
 while (i < FBLOBStreams.Count) do
  begin
   LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
   if (LocalBLOBStream = nil) then
      raise EACRException.Create(10120,ErrorLNilPointer);
   if (LocalBLOBStream.FieldNo = FieldNumber) then
    begin
     if ((LocalBLOBStream.OpenMode = bomWrite) or
       (LocalBLOBStream.OpenMode = bomReadWrite)) then
      begin
        FTableData.WriteBLOBFieldToRecordBuffer(Self,FieldNumber,LocalBLOBStream);
      end;

     if (LocalBLOBStream.UserBLOBStream <> nil) then
          LocalBLOBStream.UserBLOBStream.Free;
     LocalBLOBStream.Free;
     FBLOBStreams.Delete(i);
     Dec(i);
    end;
   Inc(i);
  end;
end; // InternalCloseBLOB


//------------------------------------------------------------------------------
// clear blob streams
//------------------------------------------------------------------------------
procedure TACRLocalCursor.ClearBLOBStreams(WriteOnly: Boolean = False);
var i:                  Integer;
    LocalBLOBStream:    TACRLocalBLOBStream;
begin
 i := 0;
 if (FBLOBStreams <> nil) then
   while (i < FBLOBStreams.Count) and (FBLOBStreams.Count > 0) do
    begin
     LocalBLOBStream := TACRLocalBLOBStream(FBLOBStreams.Items[i]);
     if (LocalBLOBStream  = nil) then
      raise EACRException.Create(10121,ErrorLNilPointer);
     if ((Not WriteOnly) or
         ((WriteOnly) and
          ((LocalBLOBStream.OpenMode = bomWrite) or
           (LocalBLOBStream.OpenMode = bomReadWrite)))) then
       begin
        if (LocalBLOBStream.UserBLOBStream <> nil) then
          LocalBLOBStream.UserBLOBStream.Free
        else
         begin
          LocalBLOBStream.Free;
          FBLOBStreams.Delete(i);
         end;
        Dec(i);
       end;
     Inc(i);
    end;
end; // ClearBLOBStreams


//------------------------------------------------------------------------------
//LastAutoincValue
//------------------------------------------------------------------------------
function TACRLocalCursor.LastAutoincValue(FieldNo: Integer): Int64;
begin
  Result := FTableData.LastAutoincValue(FieldNo, FSession);
end;//LastAutoincValue


//------------------------------------------------------------------------------
//SetLastAutoincValue
//------------------------------------------------------------------------------
procedure TACRLocalCursor.SetLastAutoincValue(Value: Int64; FieldNo: Integer);
begin
  FTableData.SetLastAutoincValue(Value, FieldNo, Self);
end;//SetLastAutoincValue


//------------------------------------------------------------------------------
// lock table for read
//------------------------------------------------------------------------------
procedure TACRLocalCursor.LockTableForRead;
begin
  FTableData.LockTableForRead(Session.SessionID);
end; // LockTableForRead;


//------------------------------------------------------------------------------
// unlock table for read
//------------------------------------------------------------------------------
procedure TACRLocalCursor.UnlockTableForRead;
begin
  FTableData.UnlockTableForRead(Session.SessionID);
end; // UnlockTableForRead;


//------------------------------------------------------------------------------
// lock table in S mode
//------------------------------------------------------------------------------
procedure TACRLocalCursor.LockTableS(ErrorCode: Integer);
begin
 if (not Session.InTransaction) then
  if (not FTableData.TryToLockTableS) then
   raise EACRException.Create(ErrorCode,ErrorLCannotLockTable,['S',FTableName]);
end; // LockTableS


//------------------------------------------------------------------------------
// unlock table in S mode
//------------------------------------------------------------------------------
procedure TACRLocalCursor.UnlockTableS(ErrorCode: Integer);
begin
 if (not Session.InTransaction) then
  if (not FTableData.TryToUnlockTableS) then
   raise EACRException.Create(ErrorCode,ErrorLCannotUnlockTable,['S',FTableName]);
end; // UnlockTableS


//------------------------------------------------------------------------------
// lock table in IRW mode
//------------------------------------------------------------------------------
procedure TACRLocalCursor.LockTableIRW(ErrorCode: Integer);
begin
 if (not Session.InTransaction) then
  if (not FTableData.TryToLockTableIRW) then
   raise EACRException.Create(ErrorCode,ErrorLCannotLockTable,['IRW',FTableName]);
end; // LockTableIRW


//------------------------------------------------------------------------------
// unlock table in IRW mode
//------------------------------------------------------------------------------
procedure TACRLocalCursor.UnlockTableIRW(ErrorCode: Integer);
begin
 if (not Session.InTransaction) then
  if (not FTableData.TryToUnlockTableIRW) then
   raise EACRException.Create(ErrorCode,ErrorLCannotUnlockTable,['IRW',FTableName]);
end; // UnlockTableIRW


//------------------------------------------------------------------------------
// lock table in RW mode
//------------------------------------------------------------------------------
procedure TACRLocalCursor.LockTableRW(ErrorCode: Integer);
begin
 if (not Session.InTransaction) then
  if (not FTableData.TryToLockTableRW) then
   raise EACRException.Create(ErrorCode,ErrorLCannotLockTable,['RW',FTableName]);
end; // LockTableRW


//------------------------------------------------------------------------------
// unlock table in RW mode
//------------------------------------------------------------------------------
procedure TACRLocalCursor.UnlockTableRW(ErrorCode: Integer);
begin
 if (not Session.InTransaction) then
  if (not FTableData.TryToUnlockTableRW) then
   raise EACRException.Create(ErrorCode,ErrorLCannotUnlockTable,['RW',FTableName]);
end; // UnockTableRW


//------------------------------------------------------------------------------
// TryToLockTableS
//------------------------------------------------------------------------------
function TACRLocalCursor.TryToLockTableS: Boolean;
begin
 if (Session.InTransaction) then
  Result := True
 else
  Result := FTableData.TryToLockTableS;
end; // TryToLockTableS


//------------------------------------------------------------------------------
// TryToUnlockTableS
//------------------------------------------------------------------------------
function TACRLocalCursor.TryToUnlockTableS: Boolean;
begin
 if (Session.InTransaction) then
  Result := True
 else
  Result := FTableData.TryToUnlockTableS;
end; // TryToUnlockTableS


//------------------------------------------------------------------------------
// TryToLockTableIRW
//------------------------------------------------------------------------------
function TACRLocalCursor.TryToLockTableIRW: Boolean;
begin
 if (Session.InTransaction) then
   Result := True
 else
  Result := FTableData.TryToLockTableIRW;
end; // TryToLockTableIRW


//------------------------------------------------------------------------------
// TryToUnlockTableIRW
//------------------------------------------------------------------------------
function TACRLocalCursor.TryToUnlockTableIRW: Boolean;
begin
 if (Session.InTransaction) then
   Result := True
 else
   Result := FTableData.TryToUnlockTableIRW;
end; // TryToUnlockTableIRW


//------------------------------------------------------------------------------
// TryToLockTableRW
//------------------------------------------------------------------------------
function TACRLocalCursor.TryToLockTableRW: Boolean;
begin
 if (Session.InTransaction) then
   Result := True
 else
   Result := FTableData.TryToLockTableRW;
end; // TryToLockTableRW


//------------------------------------------------------------------------------
// TryToUnlockTableRW
//------------------------------------------------------------------------------
function TACRLocalCursor.TryToUnlockTableRW: Boolean;
begin
 if (Session.InTransaction) then
   Result := True
 else
   Result := FTableData.TryToUnlockTableRW;
end; // TryToUnlockTableRW


//------------------------------------------------------------------------------
// save current position
//------------------------------------------------------------------------------
procedure TACRLocalCursor.SaveCurrentPosition;
begin
  FreeSavedPosition;
  if (FCurrentRecordBuffer <> nil) then
   begin
    FSavedRecordBuffer := AllocateRecordBuffer;
    Move(FCurrentRecordBuffer^, FSavedRecordBuffer^,RecordBufferSize);
   end;
  FSavedPosition := SavePosition;
end; // SaveCurrentPosition


//------------------------------------------------------------------------------
// restore saved position
//------------------------------------------------------------------------------
procedure TACRLocalCursor.RestoreSavedPosition;
begin
  if (FSavedPosition <> nil) then
   begin
    RestorePosition(FSavedPosition);
    if (FSavedRecordBuffer <> nil) then
     begin
      if (FCurrentRecordBuffer <> nil) then
       Move(FSavedRecordBuffer^,FCurrentRecordBuffer^,RecordBufferSize)
     end
    else
     FCurrentRecordBuffer := nil;
    FreeSavedPosition;
   end;
end; // RestoreSavedPosition


//------------------------------------------------------------------------------
// free saved position
//------------------------------------------------------------------------------
procedure TACRLocalCursor.FreeSavedPosition;
begin
  if (FSavedPosition <> nil) then
   begin
    FreePosition(FSavedPosition);
    if (FSavedRecordBuffer <> nil) then
     FreeRecordBuffer(FSavedRecordBuffer);
   end;
end; // FreeSavedPosition


////////////////////////////////////////////////////////////////////////////////
//
// TACRLocalSession
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// does DBData correspond to this local sesion?
//------------------------------------------------------------------------------
function TACRLocalSession.IsAppropriateDatabaseData(DBData: TACRDatabaseData): Boolean;
begin
  Result := False;
  if ((DatabaseName <> ACRMemoryDatabaseName) and
      (DatabaseName <> ACRTemporaryDatabaseName)) then
   Result := (AnsiUpperCase(DBData.DatabaseName) =
              AnsiUpperCase(DatabaseFileName))
  else
  if (DatabaseName = ACRTemporaryDatabaseName) then
   Result := (AnsiUpperCase(DBData.DatabaseName) =
              AnsiUpperCase(ACRTemporaryDatabaseName))
  else
  if (AnsiUpperCase(DatabaseName) = AnsiUpperCase(ACRMemoryDatabaseName)) then
   Result := (AnsiUpperCase(DBData.DatabaseName) =
              AnsiUpperCase(ACRMemoryDatabaseName));
end;// IsAppropriateDatabaseData


//------------------------------------------------------------------------------
// FindDatabaseData
//------------------------------------------------------------------------------
function TACRLocalSession.FindDatabaseData: TACRDatabaseData;
var
  DBDatas: TList;
  i: integer;
begin
  DBDatas := DBDataList.LockList;
  Result := nil;
  try
    for i:=0 to DBDatas.Count-1 do
     if (IsAppropriateDatabaseData(DBDatas.Items[i])) then
      begin
       Result := DBDatas.Items[i];
       break;
      end;
  finally
    DBDataList.UnlockList;
  end;
end;// FindDatabaseData


//------------------------------------------------------------------------------
// create new database data and add it to global list
//------------------------------------------------------------------------------
function TACRLocalSession.CreateDatabaseData: TACRDatabaseData;
{$IFDEF DISK_ENGINE}
var
  DBDatas: TList;
{$ENDIF}
begin
{$IFDEF DISK_ENGINE}
  Result := TACRDiskDatabaseData.Create;
  TACRDiskDatabaseData(Result).DatabaseName := DatabaseFileName;
//  TACRDiskDatabaseData(Result).Options := Options;
  TACRDiskDatabaseData(Result).CryptoParams := CryptoParams;
  DBDatas := DBDataList.LockList;
  DBDatas.Add(Result);
  DBDataList.UnlockList;
{$ENDIF}
end;// CreateDatabaseData


//------------------------------------------------------------------------------
// find or create database data
//------------------------------------------------------------------------------
function TACRLocalSession.FindOrCreateDatabaseData: TACRDatabaseData;
begin
  DBDataList.LockList;
  try
    Result := FindDatabaseData;
    if (Result = nil) then
      Result := CreateDatabaseData;
  finally
    DBDataList.UnlockList;
  end;
end;// FindOrCreateDatabaseData


//------------------------------------------------------------------------------
// db connected?
//------------------------------------------------------------------------------
function TACRLocalSession.GetConnected: Boolean;
begin
  Result := (FDatabaseData <> nil);
end;// GetConnected


//------------------------------------------------------------------------------
// connect / disconnect
//------------------------------------------------------------------------------
procedure TACRLocalSession.SetConnected(Value: boolean);
begin
 if Value <> GetConnected then
  begin
    DBDataList.LockList;
    try
     if Value then
      begin
        FTransaction := nil;
        FDatabaseData := FindOrCreateDatabaseData;
        try
          FDatabaseData.ConnectSession(Self);
        except
          FDatabaseData.FreeIfNoSessionsConnected;// implemented for disk DB only
          FDatabaseData := nil;
          raise;
        end;
      end
     else
      begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRLocalSession.SetConnected(False) starting...'+
              ', SessionID = '+IntToStr(FSessionID)+
              ', DatabaseData = '+IntToStr(Integer(FDatabaseData))+
              ', Transaction = '+IntToStr(Integer(FTransaction)));
 {$ENDIF}
        if (FDatabaseData <> nil) then
         begin
          if (FTransaction <> nil) then
           begin
            FTransaction.Rollback;
            FTransaction.Free;
            FTransaction := nil;
           end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRLocalSession.SetConnected(False) disconnecting session...'+
              ', SessionID = '+IntToStr(FSessionID)+
              ', DatabaseData.DatabaseName = '+IntToStr(Integer(FDatabaseData.DatabaseName))+
              ', Transaction = '+IntToStr(Integer(FTransaction)));
 {$ENDIF}
          FDatabaseData.DisconnectSession(Self);
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRLocalSession.SetConnected(False) disconnecting session... OK'+
              ', DatabaseData = '+IntToStr(Integer(FDatabaseData))+
              ', Transaction = '+IntToStr(Integer(FTransaction)));
 {$ENDIF}
          FDatabaseData.FreeIfNoSessionsConnected;// implemented for disk DB only
          FDatabaseData := nil;
         end;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRLocalSession.SetConnected(False) finished'+
              ', DatabaseData = '+IntToStr(Integer(FDatabaseData))+
              ', Transaction = '+IntToStr(Integer(FTransaction)));
 {$ENDIF}
      end;
    finally
     DBDataList.UnlockList;
    end;
  end;
end;// SetConnected


//------------------------------------------------------------------------------
// check if database exists
//------------------------------------------------------------------------------
function TACRLocalSession.GetDatabaseExists: Boolean;
begin
  Result := FileExists(FDatabaseFileName);
end; // GetDatabaseExists


//------------------------------------------------------------------------------
// create database
//------------------------------------------------------------------------------
procedure TACRLocalSession.CreateDatabase;
begin
  DBDataList.LockList;
  try
    if (FindDatabaseData <> nil) then
      raise EACRException.Create(20092, ErrorADatabaseAlreadyOpen);
    FDatabaseData := CreateDatabaseData;
    try
      FDatabaseData.CreateDatabase(Self);
    finally
      FDatabaseData.FreeIfNoSessionsConnected;// implemented for disk DB only
      FDatabaseData := nil;
    end;
  finally
    DBDataList.UnlockList;
  end;
end;// CreateDatabase


//------------------------------------------------------------------------------
// delete database
//------------------------------------------------------------------------------
procedure TACRLocalSession.DeleteDatabase;
begin

end;// DeleteDatabase


//------------------------------------------------------------------------------
// rename database
//------------------------------------------------------------------------------
procedure TACRLocalSession.RenameDatabase(NewDatabaseFileName: String);
begin

end;// RenameDatabase


//------------------------------------------------------------------------------
// flush file buffers
//------------------------------------------------------------------------------
procedure TACRLocalSession.FlushFileBuffers;
begin
  if (FDatabaseData = nil) then
    raise EACRException.Create(11235,ErrorLNilPointer);
  FDatabaseData.FlushFileBuffers;  
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// return database format version
//------------------------------------------------------------------------------
function TACRLocalSession.GetFormatVersion: Double;
begin
 DBDataList.LockList;
 try
  if (FDatabaseData = nil) then
   FDatabaseData := FindOrCreateDatabaseData;
  Result := FDatabaseData.GetFormatVersion(Self);
 finally
   DBDataList.UnlockList;
 end;
end;// GetFormatVersion


//------------------------------------------------------------------------------
// get count of free pages
//------------------------------------------------------------------------------
function TACRLocalSession.GetFreePageCount: Integer;
begin
 DBDataList.LockList;
 try
  if (FDatabaseData = nil) then
   FDatabaseData := FindOrCreateDatabaseData;
  Result := FDatabaseData.GetFreePageCount(Self);
 finally
   DBDataList.UnlockList;
 end;
end;// GetFreePageCount


//------------------------------------------------------------------------------
// get total count of pages
//------------------------------------------------------------------------------
function TACRLocalSession.GetTotalPageCount: Integer;
begin
 DBDataList.LockList;
 try
  if (FDatabaseData = nil) then
   FDatabaseData := FindOrCreateDatabaseData;
  Result := FDatabaseData.GetTotalPageCount(Self);
 finally
   DBDataList.UnlockList;
 end;
end;// GetTotalPageCount


//------------------------------------------------------------------------------
// return true if database is encrypted
//------------------------------------------------------------------------------
function TACRLocalSession.IsDatabaseEncrypted: Boolean;
var DBDataNil: Boolean;
    FreeDB:    Boolean;
begin
 DBDataList.LockList;
 try
  DBDataNil := (FDatabaseData = nil);
  if (DBDataNil) then
   begin
    FDatabaseData := FindDatabaseData;
    FreeDB := (FDatabaseData = nil);
    if (FreeDB) then
     FDatabaseData := CreateDatabaseData;
   end;
  try
    Result := FDatabaseData.IsDatabaseEncrypted(Self);
  finally
    if (DBDataNil) then
     begin
      if (FreeDB) then
        FDatabaseData.Free;
      FDatabaseData := nil;
     end;
  end;
 finally
   DBDataList.UnlockList;
 end;
end; // IsDatabaseEncrypted


//------------------------------------------------------------------------------
// return true if database is encrypted by password or by key
//------------------------------------------------------------------------------
function TACRLocalSession.IsDatabaseEncryptedByPassword: Boolean;
var DBDataNil: Boolean;
    FreeDB:    Boolean;
begin
 DBDataList.LockList;
 try
  DBDataNil := (FDatabaseData = nil);
  if (DBDataNil) then
   begin
    FDatabaseData := FindDatabaseData;
    FreeDB := (FDatabaseData = nil);
    if (FreeDB) then
     FDatabaseData := CreateDatabaseData;
   end;
  try
    Result := FDatabaseData.IsDatabaseEncryptedByPassword(Self);
  finally
    if (DBDataNil) then
     begin
      if (FreeDB) then
        FDatabaseData.Free;
      FDatabaseData := nil;
     end;
  end;
 finally
   DBDataList.UnlockList;
 end;
end; // IsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// makes Exe database from adb file
//------------------------------------------------------------------------------
procedure TACRLocalSession.MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string);
var DBDataNil: Boolean;
    FreeDB:    Boolean;
begin
 DBDataList.LockList;
 try
  DBDataNil := (FDatabaseData = nil);
  if (DBDataNil) then
   begin
    FDatabaseData := FindDatabaseData;
    FreeDB := (FDatabaseData = nil);
    if (FreeDB) then
     FDatabaseData := CreateDatabaseData;
   end;
  try
    FDatabaseData.MakeExeDatabase(Self,ExeFileName,ExeDatabaseFileName);
  finally
    if (DBDataNil) then
     begin
      if (FreeDB) then
        FDatabaseData.Free;
      FDatabaseData := nil;
     end;
  end;
 finally
   DBDataList.UnlockList;
 end;
end; // MakeExeDatabase


//------------------------------------------------------------------------------
// removes database file from executable database file
//------------------------------------------------------------------------------
procedure TACRLocalSession.RemoveDatabaseFromExe;
var DBDataNil: Boolean;
    FreeDB:    Boolean;
begin
 DBDataList.LockList;
 try
  DBDataNil := (FDatabaseData = nil);
  if (DBDataNil) then
   begin
    FDatabaseData := FindDatabaseData;
    FreeDB := (FDatabaseData = nil);
    if (FreeDB) then
     FDatabaseData := CreateDatabaseData;
   end;
  try
    DatabaseData.RemoveDatabaseFromExe(Self);
  finally
    if (DBDataNil) then
     begin
      if (FreeDB) then
        FDatabaseData.Free;
      FDatabaseData := nil;
     end;
  end;
 finally
   DBDataList.UnlockList;
 end;
end; // RemoveDatabaseFromExe


//------------------------------------------------------------------------------
// returns true if this file is an Accuracer database
//------------------------------------------------------------------------------
function TACRLocalSession.IsAccuracerDatabaseFile: Boolean;
var DBDataNil: Boolean;
    FreeDB:    Boolean;
begin
 DBDataList.LockList;
 try
  DBDataNil := (FDatabaseData = nil);
  if (DBDataNil) then
   begin
    FDatabaseData := FindDatabaseData;
    FreeDB := (FDatabaseData = nil);
    if (FreeDB) then
     FDatabaseData := CreateDatabaseData;
   end;
  try
    Result := FDatabaseData.IsAccuracerDatabaseFile(Self);
  finally
    if (DBDataNil) then
     begin
      if (FreeDB) then
        FDatabaseData.Free;
      FDatabaseData := nil;
     end;
  end;
 finally
   DBDataList.UnlockList;
 end;
end; // IsAccuracerDatabaseFile


//------------------------------------------------------------------------------
// return true if CryptoParams are valid
//------------------------------------------------------------------------------
function TACRLocalSession.IsCryptoParamsValid: Boolean;
var DBDataNil: Boolean;
    FreeDB:    Boolean;
begin
 DBDataList.LockList;
 try
  DBDataNil := (FDatabaseData = nil);
  if (DBDataNil) then
   begin
    FDatabaseData := FindDatabaseData;
    FreeDB := (FDatabaseData = nil);
    if (FreeDB) then
     FDatabaseData := CreateDatabaseData;
   end;
  try
    Result := FDatabaseData.IsCryptoParamsValid(Self);
  finally
    if (DBDataNil) then
     begin
      if (FreeDB) then
        FDatabaseData.Free;
      FDatabaseData := nil;
     end;
  end;
 finally
   DBDataList.UnlockList;
 end;
end; // IsCryptoParamsValid


//------------------------------------------------------------------------------
// get list of database tables
//------------------------------------------------------------------------------
procedure TACRLocalSession.GetTablesList(List: TStrings);
begin
  FDatabaseData := FindOrCreateDatabaseData;
  FDatabaseData.GetTablesList(List);
end;// GetTablesList


//------------------------------------------------------------------------------
// TableExists
//------------------------------------------------------------------------------
function TACRLocalSession.TableExists(TableName: String): Boolean;
begin
  FDatabaseData := FindOrCreateDatabaseData;
  Result := FDatabaseData.TableExists(TableName);
end;// TableExists


//------------------------------------------------------------------------------
// retrun true if database has active transaction
//------------------------------------------------------------------------------
function TACRLocalSession.GetInTransaction: Boolean;
begin
  Result := (FTransaction <> nil);
end; // GetInTransaction


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRLocalSession.Create;
begin
  inherited;
  FDatabaseData := nil;
end; // Create


//------------------------------------------------------------------------------
// start a transaction
//------------------------------------------------------------------------------
procedure TACRLocalSession.StartTransaction;
begin
 DBDataList.LockList;
 try
   if (FDatabaseData = nil) then
    raise EACRException.Create(10813,ErrorLNilPointer);
   if (FTransaction <> nil) then
    raise EACRException.Create(10814,ErrorLTransactionAlreadyStarted,[FDatabaseData.DatabaseName]);
   if ((FDatabaseData.DatabaseName = ACRMemoryDatabaseName) or
       (FDatabaseData.DatabaseName = ACRTemporaryDatabaseName)) then
    raise EACRException.Create(10831,ErrorLTransactionOnNotDiskDatabase);
   FTransaction := TACRTransaction.Create(Self,FDatabaseData);
 finally
   DBDataList.UnlockList;
 end;
end; // StartTransaction


//------------------------------------------------------------------------------
// apply changes made by transaction
//------------------------------------------------------------------------------
procedure TACRLocalSession.Commit(FlushFileBuffers: Boolean);
begin
  if (FDatabaseData = nil) then
    raise EACRException.Create(10815,ErrorLNilPointer);
  if (FTransaction = nil) then
   raise EACRException.Create(10816,ErrorLTransactionIsNotStarted,[FDatabaseData.DatabaseName]);
  FTransaction.Commit(FlushFileBuffers);
  FreeAndNil(FTransaction);
end; // Commit


//------------------------------------------------------------------------------
// cancel changes made by transaction
//------------------------------------------------------------------------------
procedure TACRLocalSession.Rollback;
begin
  if (FDatabaseData = nil) then
    raise EACRException.Create(10817,ErrorLNilPointer);
  if (FTransaction = nil) then
   raise EACRException.Create(10818,ErrorLTransactionIsNotStarted,[FDatabaseData.DatabaseName]);
  FTransaction.Rollback;
  FreeAndNil(FTransaction);
end; // Rollback


//------------------------------------------------------------------------------
// remove all session locks - called by server session disconnect
//------------------------------------------------------------------------------
procedure TACRLocalSession.RemoveAllLocks;
begin
  if (FDatabaseData <> nil) then
   FDatabaseData.RemoveAllLocks(Self.SessionID);
end; // RemoveAllLocks


initialization

{$IFDEF DEBUG_LOG_INIT}
// ShowMessage('ACRLocalEngine initialization started');
aaWriteToLog('ACRLocalEngine> initialization started');
{$ENDIF}

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRLocalEngine> initialization finished');
{$ENDIF}

end.
