unit ACRVariant;

{$I ACRVer.inc}

interface

uses SysUtils, Classes,

{$IFDEF MSWINDOWS}
     Controls,
     Windows,
{$ENDIF}
{$IFDEF LINUX}
     Libc,
     Types,
{$ENDIF}

     Math,

// Accuracer units

{$IFNDEF D6H}
     ACRD4Routines,
{$ENDIF}
     {$IFDEF DEBUG_LOG}
     ACRDebug,
     {$ENDIF}
     ACRCompression,
     ACRTypes,
     ACRTypesRoutines,
     ACRConst,
     ACRConverts,
     ACRMemory,
     ACRExcept;

type
{$IFDEF LINUX}
  // Windows.pas
  LCID = DWORD;
const
  LOCALE_USER_DEFAULT =  $400;
  SORT_STRINGSORT     = $1000;
  NORM_IGNORECASE     =     1;
  // Delphi7 Controls.pas
type
  TDate = type TDateTime;
  TTime = type TDateTime;
{$ENDIF}

 TACRVariant = class(TObject)
  private
   FDataType:     TACRBaseFieldType;    // type of data
   FDataSize:     Integer;              // size of data value
   FPData:        PChar;                // Binary data representation
   FIsDataLinked: ByteBool;             // freemem required?
   FIsNull:       ByteBool;             // Abstract Boolean Flag
  public
   procedure LoadFromStream(Stream: TStream); virtual;
   procedure SaveToStream(Stream: TStream); virtual;
   // Constructor
   constructor Create(DataType: TACRBaseFieldType = bftUnknown);
   // Destructor
   destructor Destroy; override;
   // SetNull
   procedure SetNull(DataType: TACRBaseFieldType = bftUnknown);
   // Free Buffer
   procedure Clear(DataType: TACRBaseFieldType = bftUnknown);
   // Set data size
   procedure SetDataSize(NewSize: Integer);
   // Assign data from source (or make link to Source.Data)
   procedure Assign(Source: TACRVariant; CopyDataFlag: boolean = true); virtual;
   // Set Data (or make link to Data)
   procedure SetData(Buffer: Pointer; Size: Integer;
                     DataType: TACRBaseFieldType;
                     CopyDataFlag: boolean = true);
   // Copy Data to Address
   function CopyDataToAddress(Buffer: Pointer; MaxSize: Integer = -1): boolean;
   // Cast to new data type
   procedure Cast(NewDataType: TACRBaseFieldType);
   // return true if DataType is numeric
   function IsNumericDataType: Boolean;
   // return true if DataType is String
   function IsStringDataType: Boolean;
   // return true if DataType is WideString
   function IsWideStringDataType: Boolean;
   // return true if DataType is BLOB
   function IsBlobDataType: Boolean;
   // return true if DataType is Time, Date, DateTime
   function IsDateTimeDataType: Boolean;

   // return Length of String or -1 (if not IsStringType)
   function StrLen: Integer;

   // data := -data ( if not number then raise)
   procedure InvertValue;
   // data := data + value
   procedure Add(Value: TACRVariant);
   // data := data - value
   procedure Sub(Value: TACRVariant);
   // data := data * value
   procedure Mul(Value: TACRVariant);
   // data := data / value
   procedure Division(Value: TACRVariant);

   // Compare with another Variants
   function Compare(
                    Value: TACRVariant;
                    TrueFalseNullLogic: boolean = True;
                    CaseInsensitive: boolean = True;
                    PartialKey: boolean = False
                   ): TACRCompareResult;
  private
   // Set Data Type
   procedure SetDataType(DataType: TACRBaseFieldType);
   // CastResultToType
   procedure CastResultToType(NewDataType: TACRBaseFieldType; out Result);
   // Cast value to empty FPData
   procedure CastAndSetData(const Value; ValueType: TACRBaseFieldType);
   // raise EACRException if FPData = nil
   procedure CheckDataForNull;

  private
   // return (FPData = nil)
   function GetIsNull: Boolean;

   // Set typed Data to Variant
   procedure SetDataValue(const value; ValueType: TACRBaseFieldType);
   // Set typed Data from Variant
   procedure GetDataValue(out Value; ValueType: TACRBaseFieldType);

   // Set Data for Signed Int8 (Shortint)
   procedure SetDataAsSignedInt8(Value: Shortint);
   // Get Data for Signed Int8 (Shortint)
   function  GetDataAsSignedInt8: Shortint;

   // Set Data for Signed Int16 (Smallint)
   procedure SetDataAsSignedInt16(Value: Smallint);
   // Get Data for Signed Int16 (Smallint)
   function  GetDataAsSignedInt16: Smallint;

   // Set Data for Signed Int32 (Integer)
   procedure SetDataAsSignedInt32(Value: Integer);
   // Get Data for Signed Int32 (Integer)
   function  GetDataAsSignedInt32: Integer;

   // Set Data for Signed Int64 (Int64)
   procedure SetDataAsSignedInt64(Value: Int64);
   // Get Data for Signed Int64 (Int64)
   function  GetDataAsSignedInt64: Int64;

   // Set Data for Unsigned Int8 (Byte)
   procedure SetDataAsUnsignedInt8(Value: Byte);
   // Get Data for Unsigned Int8 (Byte)
   function  GetDataAsUnsignedInt8: Byte;

   // Set Data for Unsigned Int16 (Word)
   procedure SetDataAsUnsignedInt16(Value: Word);
   // Get Data for Unsigned Int16 (Word)
   function  GetDataAsUnsignedInt16: Word;

   // Set Data for Unsigned Int32 (Cardinal)
   procedure SetDataAsUnsignedInt32(Value: Cardinal);
   // Get Data for Unsigned Int32 (Cardinal)
   function  GetDataAsUnsignedInt32: Cardinal;


   // Set Data for Char, Varchar (String)
   procedure SetDataAsString(Value: String);
   // Get Data as String
   function  GetDataAsString: String;

   // Set Data for WideChar, WideVarchar (WideString)
   procedure SetDataAsWideString(Value: WideString);
   // Get Data as WideString
   function  GetDataAsWideString: WideString;


   // Set Data for Single
   procedure SetDataAsSingle(Value: Single);
   // Get Data for Single
   function  GetDataAsSingle: Single;

   // Set Data for Double
   procedure SetDataAsDouble(Value: Double);
   // Get Data for Double
   function  GetDataAsDouble: Double;

   // Set Data for Extended
   procedure SetDataAsExtended(Value: Extended);
   // Get Data for Extended
   function  GetDataAsExtended: Extended;


   // Set Data for AcrDate
   procedure SetDataAsACRDate(Value: TACRDate);
   // Get Data for AcrDate
   function  GetDataAsACRDate: TACRDate;
   // Set Data for TDate
   procedure SetDataAsTDate(Value: TDate);
   // Get Data for TDate
   function  GetDataAsTDate: TDate;


   // Set Data for AcrTime
   procedure SetDataAsACRTime(Value: TACRTime);
   // Get Data for AcrTime
   function  GetDataAsACRTime: TACRTime;
   // Set Data for TTime
   procedure SetDataAsTTime(Value: TTime);
   // Get Data for TTime
   function  GetDataAsTTime: TTime;

   // Set Data for AcrDateTime
   procedure SetDataAsACRDateTime(Value: TACRDateTime);
   // Get Data for AcrDateTime
   function  GetDataAsACRDateTime: TACRDateTime;
   // Set Data for TDateTime
   procedure SetDataAsTDateTime(Value: TDateTime);
   // Get Data for TDateTime
   function  GetDataAsTDateTime: TDateTime;

   // Set Data for Boolean
   procedure SetDataAsBoolean(Value: TACRLogical);
   // Get Data for Boolean
   function  GetDataAsBoolean: TACRLogical;


   // Set Data for Currency
   procedure SetDataAsCurrency(Value: TACRCurrency);
   // Get Data for Currency
   function  GetDataAsCurrency: TACRCurrency;

   // Set Data from Borland Variant type
   procedure SetDataAsVariant(Value: Variant);
   // Get Data to Borland Variant type
   function GetDataAsVariant: Variant;

  public
   property DataType: TACRBaseFieldType read FDataType write SetDataType;
   property DataSize: Integer read FDataSize;
   property IsNull: Boolean read GetIsNull;
   property pData: PChar read FPData;
   property IsDataLinked: ByteBool read FIsDataLinked;
  public
   property AsShortint: Shortint      read GetDataAsSignedInt8    write SetDataAsSignedInt8;
   property AsSmallint: Smallint      read GetDataAsSignedInt16   write SetDataAsSignedInt16;
   property AsInteger: Integer        read GetDataAsSignedInt32   write SetDataAsSignedInt32;
   property AsInt64: Int64            read GetDataAsSignedInt64   write SetDataAsSignedInt64;
   property AsByte: Byte              read GetDataAsUnsignedInt8  write SetDataAsUnsignedInt8;
   property AsWord: Word              read GetDataAsUnsignedInt16 write SetDataAsUnsignedInt16;
   property AsCardinal: Cardinal      read GetDataAsUnsignedInt32 write SetDataAsUnsignedInt32;
   property AsLongWord: Cardinal      read GetDataAsUnsignedInt32 write SetDataAsUnsignedInt32;

   property AsSingle: Single          read GetDataAsSingle        write SetDataAsSingle;
   property AsDouble: Double          read GetDataAsDouble        write SetDataAsDouble;
   property AsExtended: Extended      read GetDataAsExtended      write SetDataAsExtended;

   property AsString: String          read GetDataAsString        write SetDataAsString;
   property AsWideString: WideString  read GetDataAsWideString    write SetDataAsWideString;

   property AsACRDate: TACRDate       read GetDataAsACRDate       write SetDataAsACRDate;
   property AsACRTime: TACRTime       read GetDataAsACRTime       write SetDataAsACRTime;
   property AsACRDateTime: TACRDateTime read GetDataAsACRDateTime write SetDataAsACRDateTime;
   property AsTDate: TDate            read GetDataAsTDate         write SetDataAsTDate;
   property AsTTime: TTime            read GetDataAsTTime         write SetDataAsTTime;
   property AsTDateTime: TDateTime    read GetDataAsTDateTime     write SetDataAsTDateTime;

   property AsBoolean: TACRLogical    read GetDataAsBoolean       write SetDataAsBoolean;
   property AsLogical: TACRLogical    read GetDataAsBoolean       write SetDataAsBoolean;
   property AsCurrency: TACRCurrency  read GetDataAsCurrency      write SetDataAsCurrency;

   property AsVariant: Variant        read GetDataAsVariant       write SetDataAsVariant;
 end;


// Return Common DataType for 2 types, or ftUnknown
function GetCommonDataType(a,b: TACRBaseFieldType): TACRBaseFieldType; overload;
// Return Common DataType for 2 types, or ftUnknown
function GetCommonDataType(a,b: TACRAdvancedFieldType): TACRAdvancedFieldType; overload;

// Allocate new buffer and convert data to it
function CastToNewBuffer(
                          const Buffer;
                          const DataType: TACRBaseFieldType;
                          const NewDataType: TACRBaseFieldType
                        ): PChar; overload;


// Allocate new buffer and convert data to it
function CastToNewBuffer(
                          const Buffer;
                          const DataType: TACRBaseFieldType;
                          const NewDataType: TACRBaseFieldType;
                          out NewDataSize: Integer
                        ): PChar; overload;

// Compare 2 values
function CompareValueBuffers(
                              Buffer1, Buffer2: Pointer;
                              BaseFieldType1, BaseFieldType2: TACRBaseFieldType;
                              IsField1Null: Boolean = false;
                              IsField2Null: Boolean = false;
                              PartialCompareLength: Integer = -1;
                              IgnoreCase: Boolean = false;
                              LocaleID: LCID = LOCALE_USER_DEFAULT
                            ): TACRCompareResult;

// Compare 2 values in sense of order (-1,0,1)
function CompareValueBuffersForOrder(
                              Buffer1, Buffer2: Pointer;
                              BaseFieldType1, BaseFieldType2: TACRBaseFieldType;
                              IsField1Null: Boolean = false;
                              IsField2Null: Boolean = false;
                              PartialCompareLength: Integer = -1;
                              IgnoreCase: Boolean = false;
                              LocaleID: LCID = LOCALE_USER_DEFAULT
                            ): Integer;

implementation

uses
{$IFDEF D6H}
     Variants,
{$ENDIF}     
     ACRMain;



//------------------------------------------------------------------------------
// load from stream
//------------------------------------------------------------------------------
procedure TACRVariant.LoadFromStream(Stream: TStream);
begin
  Clear;
  LoadDataFromStream(FDataType,Sizeof(FDataType),Stream,10183);
  LoadDataFromStream(FIsNull,Sizeof(FIsNull),Stream,10184);
  if (not FIsNull) then
   begin
    LoadDataFromStream(FDataSize,Sizeof(FDataSize),Stream,10185);
    FPData := MemoryManager.GetMem(FDataSize);
    LoadDataFromStream(FPData^,FDataSize,Stream,10186);
   end;
end; // LoadFromStream


//------------------------------------------------------------------------------
// save to stream
//------------------------------------------------------------------------------
procedure TACRVariant.SaveToStream(Stream: TStream);
begin
  SaveDataToStream(FDataType,Sizeof(FDataType),Stream,10179);
  SaveDataToStream(FIsNull,Sizeof(FIsNull),Stream,10180);
  if (not FIsNull) then
   begin
    SaveDataToStream(FDataSize,Sizeof(FDataSize),Stream,10181);
    SaveDataToStream(FPData^,FDataSize,Stream,10182);
   end;
end; // SaveToStream


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRVariant.Create(DataType: TACRBaseFieldType);
begin
  FPData := nil;
  FDataType := DataType;
  FDataSize := 0;
  FIsDataLinked := false;
  FIsNull := True;
end;//Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRVariant.Destroy;
begin
  Clear;
end;//Destroy


//------------------------------------------------------------------------------
// SetNull
//------------------------------------------------------------------------------
procedure TACRVariant.SetNull(DataType: TACRBaseFieldType);
begin
  FIsNull := True;
end;//SetNull


//------------------------------------------------------------------------------
// Free Buffer
//------------------------------------------------------------------------------
procedure TACRVariant.Clear(DataType: TACRBaseFieldType);
begin
  if ((FPData <> nil) and (not FIsDataLinked)) then
      MemoryManager.FreeAndNilMem(FPData);
  FPData := nil;
  FDataType := DataType;
  FDataSize := 0;
  FIsDataLinked := false;
  FIsNull := True;
end;//Clear


//------------------------------------------------------------------------------
// Set data size
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataSize(NewSize: Integer);
begin
  if (NewSize = 0) then
   SetNull
  else
   begin
    FDataSize := NewSize;
    if (FPData = nil) then
     FPData := MemoryManager.GetMem(FDataSize)
    else
     MemoryManager.ReallocMem(FPData,FDataSize,False);
    FIsNull := False; 
   end;
end; // SetDataSize


//------------------------------------------------------------------------------
// Set Data Type
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataType(DataType: TACRBaseFieldType);
begin
  if not IsNull then
    try
      Cast(DataType);
    except
      Clear(DataType);
    end
  else
   if (FPData <> nil) then
    Clear(DataType)
   else
    FDataType := DataType;
end;//SetDataType


//------------------------------------------------------------------------------
// CastResultToType
//------------------------------------------------------------------------------
procedure TACRVariant.CastResultToType(NewDataType: TACRBaseFieldType; out Result);
var
  p: Pointer;
  TmpSize: Integer;
begin
  p := CastToNewBuffer(FPData, FDataType, NewDataType, TmpSize);
  case NewDataType of
    bftChar,
    bftVarChar:
      begin
        SetLength(String(Result), TmpSize-1);
        Move(PChar(p)^, PChar(Result)^, TmpSize);
      end;
    bftWideChar, bftWideVarChar:
      begin
        SetLength(WideString(Result), (TmpSize) div 2 - 1);
        Move(PWideChar(p)^, PWideChar(Result)^, TmpSize);
      end;
    else
      Move(p^, Result, TmpSize);
  end;
  MemoryManager.FreeAndNilMem(p);
end;//CastResultToType


//------------------------------------------------------------------------------
// Cast value to empty FPData
//------------------------------------------------------------------------------
procedure TACRVariant.CastAndSetData(const Value; ValueType: TACRBaseFieldType);
var p: Pointer;
begin
  if ((FPData <> nil) and (not FIsDataLinked)) then
    MemoryManager.FreeAndNilMem(FPData);
  if (ValueType in [bftChar, bftVarChar, bftWideChar, bftWideVarchar]) then
    p := Pointer(Value)
  else
    p := @Value;
  FPData := CastToNewBuffer(p, ValueType, FDataType, FDataSize);
end;//CastAndSetData


//------------------------------------------------------------------------------
// raise EACRException if FPData = nil
//------------------------------------------------------------------------------
procedure TACRVariant.CheckDataForNull;
begin
 if IsNull then
   raise EACRException.Create(30016, ErrorGValueIsNull);
end;//CheckDataForNull


//------------------------------------------------------------------------------
// Assign data from source (or make link to Source.Data)
//------------------------------------------------------------------------------
procedure TACRVariant.Assign(Source: TACRVariant; CopyDataFlag: boolean = true);
begin
  Clear;
  FDataType := Source.FDataType;
  FDataSize := Source.FDataSize;
  FIsDataLinked := not CopyDataFlag;
  FIsNull := Source.FIsNull;

  if (not Source.IsNull) then
    if (CopyDataFlag) then
      begin
        FPData := MemoryManager.GetMem(FDataSize);
        Move(Source.FPData^, FPData^, FDataSize);
      end
    else
      FPData := Source.FPData;
end;//Assign


//------------------------------------------------------------------------------
// Set Data (or make link to Data)
//------------------------------------------------------------------------------
procedure TACRVariant.SetData(Buffer: Pointer; Size: Integer;
  DataType: TACRBaseFieldType; CopyDataFlag: boolean);
begin
  if (CopyDataFlag) then
    begin
      if ((FPData <> nil) and
          (not FIsDataLinked) and
          (Size = FDataSize)) then
      // Using Old Buffer (FPData)
      else
        begin
          Clear;
          FPData := MemoryManager.GetMem(Size);
        end
    end
  else
    begin
      Clear;
      FPData := Buffer;
    end;

  FDataType := DataType;
  FDataSize := Size;
  FIsNull := False;
  FIsDataLinked := not CopyDataFlag;
  if (CopyDataFlag) then
    Move(Buffer^, FPData^, FDataSize);
end;//SetData


//------------------------------------------------------------------------------
// Copy Data to Address
//------------------------------------------------------------------------------
function TACRVariant.CopyDataToAddress(Buffer: Pointer; MaxSize: Integer): boolean;
var
  L: Integer;
begin
  if (IsNull) then
    begin
      Result := False;
    end
  else
    begin
      L := DataSize;
      if (MaxSize <> -1) then
       if (L > MaxSize) then
        raise EACRException.Create(20056, ErrorABufferSizeExceeded);

      Move(FPData^, Buffer^, L);
      Result := True;
    end;
end;//CopyDataToAddress




//------------------------------------------------------------------------------
// return (FPData = nil)
//------------------------------------------------------------------------------
function TACRVariant.GetIsNull: Boolean;
begin
  Result := FIsNull or (FPData = nil);
end;//GetIsNull


//------------------------------------------------------------------------------
// Set typed Data to Variant
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataValue(const value; ValueType: TACRBaseFieldType);
begin
  if ( (FDataType = ValueType) and
       (FPData <> nil) and
       (not IsStringDataType) and
       (not FIsDataLinked) ) then
   begin
    Move(Value, FPData^, FDataSize)
   end
  else
   begin
    if (FPData <> nil) then
      Clear(FDataType);
    if FDataType = bftUnknown then
      FDataType := ValueType;
    if FDataType = ValueType then
      AllocMemAndSetData(FDataType, Value, FPData, FDataSize)
    else
      CastAndSetData(Value, ValueType);
  end;
  if (FPData = nil) then
   FIsNull := True
  else
   FIsNull := False;
  FIsDataLinked := False;
end;//SetDataValue


//------------------------------------------------------------------------------
// Set typed Data from Variant
//------------------------------------------------------------------------------
procedure TACRVariant.GetDataValue(out Value; ValueType: TACRBaseFieldType);
var Len:    Integer;
    w:      Word;
    Offset: Integer;
begin
  CheckDataForNull;
  if FDataType = ValueType then
   begin
     case ValueType of
      bftChar,
      bftVarchar,
      bftClob,
      bftBlob:
        begin
          SetLength(String(Value), FDataSize-1);
          Move(PChar(FPData)^, PChar(Value)^, FDataSize-1);
          Len := SysUtils.StrLen(PChar(String(Value)));
          if (Len < (FDataSize - 1)) then
           SetLength(String(Value), Len);
        end;
      bftWideChar,
      bftWideClob,
      bftWideVarchar:
        begin
          Offset := GetStrLength(FPData,aftWideString);
          if (Offset >= FDataSize) then
           Offset := FDataSize;
          SetLength(WideString(Value), Offset+2);
          Move(PWideChar(FPData)^, PWideChar(Value)^, Offset);
          w := 0;
          Move(w,PChar(PChar(Value) + Offset)^,SizeOf(w));
        end;
      else
        CopyDataToAddress(@Value);
     end;
   end
  else
    CastResultToType(ValueType, Value);
end;//GetDataValue


//------------------------------------------------------------------------------
// Set Data for Int8 (Shortint)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsSignedInt8(Value: Shortint);
begin
  SetDataValue(Value, bftSignedInt8);
end;//SetDataAsSignedInt8


//------------------------------------------------------------------------------
// Get Data for Int8 (Shortint)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsSignedInt8: Shortint;
begin
  GetDataValue(Result, bftSignedInt8);
end;//GetDataAsSignedInt8



//------------------------------------------------------------------------------
// Set Data for Signed Int16 (Smallint)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsSignedInt16(Value: Smallint);
begin
  SetDataValue(Value, bftSignedInt16);
end;//SetDataAsSignedInt16


//------------------------------------------------------------------------------
// Get Data for Signed Int16 (Smallint)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsSignedInt16: Smallint;
begin
  GetDataValue(Result, bftSignedInt16);
end;



//------------------------------------------------------------------------------
// Set Data for Signed Int32 (Integer)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsSignedInt32(Value: Integer);
begin
  SetDataValue(Value, bftSignedInt32);
end;//SetDataAsSignedInt32


//------------------------------------------------------------------------------
// Get Data for Signed Int32 (Integer)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsSignedInt32: Integer;
begin
  GetDataValue(Result, bftSignedInt32);
end;//GetDataAsSignedInt32



//------------------------------------------------------------------------------
// Set Data for Signed Int64 (Int64)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsSignedInt64(Value: Int64);
begin
  SetDataValue(Value, bftSignedInt64);
end;//SetDataAsSignedInt64


//------------------------------------------------------------------------------
// Get Data for Signed Int64 (Int64)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsSignedInt64: Int64;
begin
  GetDataValue(Result, bftSignedInt64);
end;//GetDataAsSignedInt64


//------------------------------------------------------------------------------
// Set Data for Unsigned Int8 (Byte)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsUnsignedInt8(Value: Byte);
begin
  SetDataValue(Value, bftUnsignedInt8);
end;//SetDataAsUnsignedInt8


//------------------------------------------------------------------------------
// Get Data for Unsigned Int8 (Byte)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsUnsignedInt8: Byte;
begin
  GetDataValue(Result, bftUnsignedInt8);
end;//GetDataAsUnsignedInt8


//------------------------------------------------------------------------------
// Set Data for Unsigned Int16 (Word)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsUnsignedInt16(Value: Word);
begin
  SetDataValue(Value, bftUnsignedInt16);
end;//SetDataAsUnsignedInt16


//------------------------------------------------------------------------------
// Get Data for Unsigned Int16 (Word)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsUnsignedInt16: Word;
begin
  GetDataValue(Result, bftUnsignedInt16);
end;//GetDataAsUnsignedInt16


//------------------------------------------------------------------------------
// Set Data for Unsigned Int32 (Cardinal)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsUnsignedInt32(Value: Cardinal);
begin
  SetDataValue(Value, bftUnsignedInt32);
end;//SetDataAsUnsignedInt32


//------------------------------------------------------------------------------
// Get Data for Unsigned Int32 (Cardinal)
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsUnsignedInt32: Cardinal;
begin
  GetDataValue(Result, bftUnsignedInt32);
end;//GetDataAsUnsignedInt32


//------------------------------------------------------------------------------
// Set Data for Char, Varchar (String)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsString(Value: String);
begin
  SetDataValue(Value, bftVarchar);
end;//SetDataAsString


//------------------------------------------------------------------------------
// Get Data as String
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsString: String;
begin
  GetDataValue(Result, bftVarchar);
end;//GetDataAsString


//------------------------------------------------------------------------------
// Set Data for WideChar, WideVarchar (WideString)
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsWideString(Value: WideString);
begin
  SetDataValue(Value, bftWideVarchar);
end;//SetDataAsWideString


//------------------------------------------------------------------------------
// Get Data as WideString
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsWideString: WideString;
begin
  GetDataValue(Result, bftWideVarchar);
end;//GetDataAsWideString


//------------------------------------------------------------------------------
// Set Data for Single
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsSingle(Value: Single);
begin
  SetDataValue(Value, bftSingle);
end;//SetDataAsSingle


//------------------------------------------------------------------------------
// Get Data for Single
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsSingle: Single;
begin
  GetDataValue(Result, bftSingle);
end;//GetDataAsSingle


//------------------------------------------------------------------------------
// Set Data for Double
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsDouble(Value: Double);
begin
  SetDataValue(Value, bftDouble);
end;//SetDataAsDouble


//------------------------------------------------------------------------------
// Get Data for Double
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsDouble: Double;
begin
  GetDataValue(Result, bftDouble);
end;//GetDataAsDouble


//------------------------------------------------------------------------------
// Set Data for Extended
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsExtended(Value: Extended);
begin
  SetDataValue(Value, bftExtended);
end;//SetDataAsExtended


//------------------------------------------------------------------------------
// Get Data for Extended
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsExtended: Extended;
begin
  GetDataValue(Result, bftExtended);
end;//GetDataAsExtended




//------------------------------------------------------------------------------
// Set Data for AcrDate
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsACRDate(Value: TACRDate);
begin
  SetDataValue(Value, bftDate);
end;//SetDataAsACRDate


//------------------------------------------------------------------------------
// Get Data for AcrDate
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsACRDate: TACRDate;
begin
  GetDataValue(Result, bftDate);
end;//GetDataAsACRDate


//------------------------------------------------------------------------------
// Set Data for TDate
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsTDate(Value: TDate);
begin
  SetDataAsACRDate(DateToACRDate(Value));
end;//SetDataAsTDate


//------------------------------------------------------------------------------
// Get Data for TDate
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsTDate: TDate;
begin
  Result := ACRDateToDate(GetDataAsACRDate);
end;//GetDataAsTDate


//------------------------------------------------------------------------------
// Set Data for AcrTime
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsACRTime(Value: TACRTime);
begin
  SetDataValue(Value, bftTime);
end;//SetDataAsACRTime


//------------------------------------------------------------------------------
// Get Data for AcrTime
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsACRTime: TACRTime;
begin
  GetDataValue(Result, bftTime);
end;//GetDataAsACRTime


//------------------------------------------------------------------------------
// Set Data for TTime
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsTTime(Value: TTime);
begin
  SetDataAsACRTime(TimeToACRTime(Value));
end;//SetDataAsTTime


//------------------------------------------------------------------------------
// Get Data for AcrTime
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsTTime: TTime;
begin
  Result := ACRTimeToTime(GetDataAsACRTime);
end;//GetDataAsTTime



//------------------------------------------------------------------------------
// Set Data for AcrDateTime
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsACRDateTime(Value: TACRDateTime);
begin
  SetDataValue(Value, bftDateTime);
end;//SetDataAsACRDateTime


//------------------------------------------------------------------------------
// Get Data for AcrDateTime
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsACRDateTime: TACRDateTime;
begin
  GetDataValue(Result, bftDateTime);
end;//GetDataAsACRDateTime


//------------------------------------------------------------------------------
// Set Data for TDateTime
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsTDateTime(Value: TDateTime);
begin
  SetDataAsACRDateTime(DateTimeToACRDateTime(Value));
end;//SetDataAsTDateTime


//------------------------------------------------------------------------------
// Get Data for TDateTime
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsTDateTime: TDateTime;
begin
  Result := ACRDateTimeToDateTime(GetDataAsACRDateTime);
end;//GetDataAsTDateTime



procedure TACRVariant.SetDataAsBoolean(Value: TACRLogical);
begin
  SetDataValue(Value, bftLogical);
end;


//------------------------------------------------------------------------------
// Get Data for Boolean
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsBoolean: TACRLogical;
begin
  GetDataValue(Result, bftLogical);
end;//GetDataAsBoolean


//------------------------------------------------------------------------------
// Set Data for Currency
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsCurrency(Value: TACRCurrency);
begin
  SetDataValue(Value, bftCurrency);
end;//SetDataAsCurrency


//------------------------------------------------------------------------------
// Get Data for Currency
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsCurrency: TACRCurrency;
begin
  GetDataValue(Result, bftCurrency);
end;//GetDataAsCurrency


//------------------------------------------------------------------------------
// Set Data from Borland Variant type
//------------------------------------------------------------------------------
procedure TACRVariant.SetDataAsVariant(Value: Variant);
begin
  case VarType(Value) of
    //varEmpty
    //varNull
    varSmallint:
      AsSmallint := Value;
    varInteger:
      AsInteger := Value;
    varSingle:
      AsSingle := Value;
    varDouble:
      AsDouble := Value;
    varCurrency:
      AsCurrency := Value;
    varDate:
      AsTDateTime := Value;
    varOleStr:
      AsWideString := Value;
    //varDispatch
    //varError
    varBoolean: AsBoolean := Value;
    //varVariant
    //varUnknown
    varShortInt:
      AsShortInt := Value;
    varWord:
      AsWord := Value;
    varLongWord:
      AsCardinal := Value;
   {$IFDEF D6H}
    varInt64:
      AsInt64 := Value;
    {$ENDIF}
    varByte:
      AsByte := Value;
    //varStrArg
    //varString,
    varString .. varString + $1F:
      AsString := Value;
    //varAny
    //varTypeMask
    //varArray
    //varByRef
    varNull:
      SetNull(FDataType);
    else
      raise EACRException.Create(30113, ErrorGUnsupportedVariantType, [IntToStr(VarType(Value))]);
  end;
end;//SetDataAsVariant


//------------------------------------------------------------------------------
// Get Data to Borland Variant type
//------------------------------------------------------------------------------
function TACRVariant.GetDataAsVariant: Variant;
begin
  case FDataType of
    bftChar,
    bftVarchar:
      Result := AsString;
    bftWideChar,
    bftWideVarchar:
      Result := AsWideString;


    bftSignedInt8:
      Result := AsShortint;
    bftSignedInt16:
      Result := AsSmallint;
    bftSignedInt32:
      Result := AsInteger;
   {$IFDEF D6H}
    bftSignedInt64:
      Result := AsInt64;
    {$ENDIF}
    bftUnsignedInt8:
      Result := AsByte;
    bftUnsignedInt16:
      Result := AsWord;
    bftUnsignedInt32:
      Result := AsCardinal;

    bftSingle:
      Result := AsSingle;
    bftDouble:
      Result := AsDouble;
    bftExtended:
      Result := AsExtended;

    bftDate:
      Result := AsTDate;
    bftTime:
      Result := AsTTime;
    bftDateTime:
      Result := AsTDateTime;

    bftLogical:
      Result := AsLogical;

    bftCurrency:
      Result := AsCurrency;

    bftBlob:
      Result := AsString;
      
    else
      // Check for null
      if (IsNull) then
        Result := Null
      else
      raise EACRException.Create(30114, ErrorGUnsupportedVariantType, [BftToStr(FDataType)]);
  end;
end;//GetDataAsVariant


//------------------------------------------------------------------------------
// Cast
//------------------------------------------------------------------------------
procedure TACRVariant.Cast(NewDataType: TACRBaseFieldType);
var p: Pointer;
    NewSize: Integer;
begin
  if (FDataType = NewDataType) then
    Exit;

  if ( IsNull ) then
    begin
      Clear(NewDataType);
    end
  else
    begin
      p := CastToNewBuffer(FPData, FDataType, NewDataType, NewSize);
      Clear(NewDataType);
      FIsNull := False;
      FDataType := NewDataType;
      FPData := p;
      FDataSize := NewSize;
    end;
end;//Cast


//------------------------------------------------------------------------------
// return true if DataType is numeric
//------------------------------------------------------------------------------
function TACRVariant.IsNumericDataType: Boolean;
begin
  Result := IsNumericFieldType(FDataType);
end;//IsNumericDataType


//------------------------------------------------------------------------------
// return true if DataType is String
//------------------------------------------------------------------------------
function TACRVariant.IsStringDataType: Boolean;
begin
  Result := IsStringFieldType(FDataType);
end;//IsStringDataType


//------------------------------------------------------------------------------
// return true if DataType is WideString
//------------------------------------------------------------------------------
function TACRVariant.IsWideStringDataType: Boolean;
begin
  Result := IsWideStringFieldType(FDataType);
end;//IsStringDataType


//------------------------------------------------------------------------------
// return true if DataType is BLOB
//------------------------------------------------------------------------------
function TACRVariant.IsBlobDataType: Boolean;
begin
  Result := IsBLOBFieldType(FDataType);
end;//IsBlobDataType


//------------------------------------------------------------------------------
// return true if DataType is Time, Date, DateTime
//------------------------------------------------------------------------------
function TACRVariant.IsDateTimeDataType: Boolean;
begin
  Result := IsDateTimeFieldType(FDataType);
end;//IsDateTimeFieldType


//------------------------------------------------------------------------------
// return Length of String or -1 (if not IsStringType)
//------------------------------------------------------------------------------
function TACRVariant.StrLen: Integer;
begin
  if (IsStringDataType and (not IsNull)) then
    if (IsWideStringDataType) then
      Result := ((FDataSize) div 2) - 1
    else
      Result := FDataSize - 1
  else
    Result := -1;
end;


//------------------------------------------------------------------------------
// value := -value ( if not number then raise)
//------------------------------------------------------------------------------
procedure TACRVariant.InvertValue;
begin
  if not IsNull then
    case FDataType of
      bftSignedInt8:      AsShortint := -AsShortint;
      bftSignedInt16:     AsSmallint := -AsSmallint;
      bftSignedInt32:     AsInteger := -AsInteger;
      bftSignedInt64:     AsInt64 := -AsInt64;

      bftUnsignedInt8:    AsShortint := -AsByte;
      bftUnsignedInt16:   AsSmallint := -AsWord;
      bftUnsignedInt32:   AsInteger := -AsCardinal;

      bftSingle:          AsSingle := -AsSingle;
      bftDouble:          AsDouble := -AsDouble;
      bftExtended:        AsExtended := -AsExtended;

      bftDate:            AsACRDate := -AsACRDate;

      bftLogical:         AsLogical := not AsLogical;
      bftCurrency:        AsCurrency := -AsCurrency;
      else
        raise EACRException.Create(30123, ErrorGCannotInvertValue, [BftToStr(FDataType)]);
    end;
end;//InvertValue


{$IFOPT Q+}
 {$DEFINE RESET_OVERFLOW}
 {$Q-}
{$ENDIF}
//------------------------------------------------------------------------------
// data := data + value
//------------------------------------------------------------------------------
procedure TACRVariant.Add(Value: TACRVariant);
var
  dt: TACRDateTime;
begin
  if Value.IsNull then
    SetNull(FDataType)
  else
    if not IsNull then
      begin
        Cast(GetCommonDataType(FDataType, Value.DataType));
        case FDataType of
          bftChar,
          bftVarchar:         AsString := AsString + Value.AsString;
          bftWideChar,
          bftWideVarchar:     AsWideString := AsWideString + Value.AsWideString;

          bftSignedInt8:      AsShortint := AsShortint + Value.AsShortint;
          bftSignedInt16:     AsSmallint := AsSmallint + Value.AsSmallint;
          bftSignedInt32:     AsInteger := AsInteger + Value.AsInteger;
          bftSignedInt64:     AsInt64 := AsInt64 + Value.AsInt64;

          bftUnsignedInt8:    AsByte := AsByte + Value.AsByte;
          bftUnsignedInt16:   AsWord := AsWord + Value.AsWord;
          bftUnsignedInt32:   AsCardinal := AsCardinal + Value.AsCardinal;

          bftSingle:          AsSingle := AsSingle + Value.AsSingle;
          bftDouble:          AsDouble := AsDouble + Value.AsDouble;
          bftExtended:        AsExtended := AsExtended + Value.AsExtended;

          bftDate:            AsACRDate := AsACRDate + Value.AsACRDate;
          bftTime:            AsACRTime := AsACRTime + Value.AsACRTime;
          bftDateTime:        begin
                                dt := AsACRDateTime;
                                dt.Date := dt.Date + Value.AsACRDateTime.Date;
                                dt.Time := dt.Time + Value.AsACRDateTime.Time;
                                if (dt.Time >= MILSECS_IN_DAY) then
                                 begin
                                  dt.Date := dt.Date + dt.Time div MILSECS_IN_DAY;
                                  dt.Time := dt.Time mod MILSECS_IN_DAY;
                                 end;
                                AsACRDateTime := dt;
                              end;

          bftLogical:         AsLogical := AsLogical or Value.AsLogical;
          bftCurrency:        AsCurrency := AsCurrency + Value.AsCurrency;
          else
            raise EACRException.Create(30124, ErrorGNotNumericDataType, ['+']);
        end;
      end
    else//If IsNull
      Assign(Value);

end;//Add


//------------------------------------------------------------------------------
// data := data - value
//------------------------------------------------------------------------------
procedure TACRVariant.Sub(Value: TACRVariant);
var
  dt: TACRDateTime;
begin
  if Value.IsNull then
    SetNull(FDataType)
  else
    if not IsNull then
      begin
        Cast(GetCommonDataType(FDataType, Value.DataType));
        case FDataType of
          bftSignedInt8:      AsShortint := AsShortint - Value.AsShortint;
          bftSignedInt16:     AsSmallint := AsSmallint - Value.AsSmallint;
          bftSignedInt32:     AsInteger := AsInteger - Value.AsInteger;
          bftSignedInt64:     AsInt64 := AsInt64 - Value.AsInt64;

          bftUnsignedInt8:    AsByte := AsByte - Value.AsByte;
          bftUnsignedInt16:   AsWord := AsWord - Value.AsWord;
          bftUnsignedInt32:   AsCardinal := AsCardinal - Value.AsCardinal;

          bftSingle:          AsSingle := AsSingle - Value.AsSingle;
          bftDouble:          AsDouble := AsDouble - Value.AsDouble;
          bftExtended:        AsExtended := AsExtended - Value.AsExtended;

          bftDate:            AsACRDate := AsACRDate - Value.AsACRDate;
          bftTime:            AsACRTime := AsACRTime - Value.AsACRTime;
          bftDateTime:        begin
                                dt := AsACRDateTime;
                                dt.Date := dt.Date - Value.AsACRDateTime.Date;
                                dt.Time := dt.Time - Value.AsACRDateTime.Time;
                                if (dt.Time >= MILSECS_IN_DAY) then
                                 begin
                                  dt.Date := dt.Date + dt.Time div MILSECS_IN_DAY;
                                  dt.Time := dt.Time mod MILSECS_IN_DAY;
                                 end;
                                AsACRDateTime := dt;
                              end;

          bftLogical:         AsLogical := AsLogical or Value.AsLogical;
          bftCurrency:        AsCurrency := AsCurrency - Value.AsCurrency;

          else
            raise EACRException.Create(30125, ErrorGNotNumericDataType, ['-']);
        end;
      end;
end;//Sub


//------------------------------------------------------------------------------
// data := data * value
//------------------------------------------------------------------------------
procedure TACRVariant.Mul(Value: TACRVariant);
var
  dt: TACRDateTime;
begin
  if Value.IsNull then
    SetNull(FDataType)
  else
    if not IsNull then
      begin
        Cast(GetCommonDataType(FDataType, Value.DataType));
        case FDataType of
          bftSignedInt8:      AsShortint := AsShortint * Value.AsShortint;
          bftSignedInt16:     AsSmallint := AsSmallint * Value.AsSmallint;
          bftSignedInt32:     AsInteger := AsInteger * Value.AsInteger;
          bftSignedInt64:     AsInt64 := AsInt64 * Value.AsInt64;

          bftUnsignedInt8:    AsByte := AsByte * Value.AsByte;
          bftUnsignedInt16:   AsWord := AsWord * Value.AsWord;
          bftUnsignedInt32:   AsCardinal := AsCardinal * Value.AsCardinal;

          bftSingle:          AsSingle := AsSingle * Value.AsSingle;
          bftDouble:          AsDouble := AsDouble * Value.AsDouble;
          bftExtended:        AsExtended := AsExtended * Value.AsExtended;

          bftDate:            AsACRDate := AsACRDate * Value.AsACRDate;
          bftTime:            AsACRTime := AsACRTime * Value.AsACRTime;
          bftDateTime:        begin
                                dt := AsACRDateTime;
                                dt.Date := dt.Date * Value.AsACRDateTime.Date;
                                dt.Time := dt.Time * Value.AsACRDateTime.Time;
                                if (dt.Time >= MILSECS_IN_DAY) then
                                 begin
                                  dt.Date := dt.Date + dt.Time div MILSECS_IN_DAY;
                                  dt.Time := dt.Time mod MILSECS_IN_DAY;
                                 end;
                                AsACRDateTime := dt;
                              end;

          bftLogical:         AsLogical := AsLogical or Value.AsLogical;
          bftCurrency:        AsCurrency := AsCurrency * Value.AsCurrency;
          else
            raise EACRException.Create(30126, ErrorGNotNumericDataType, ['*']);
        end;
      end;
end;//Mul


//------------------------------------------------------------------------------
// data := data / value
//------------------------------------------------------------------------------
procedure TACRVariant.Division(Value: TACRVariant);
var
  dt: TACRDateTime;
begin
  if Value.IsNull then
    SetNull(FDataType)
  else
    if not IsNull then
      begin
        try
          Cast(GetCommonDataType(FDataType, Value.DataType));
          case FDataType of
            bftSignedInt8:      AsShortint := AsShortint div Value.AsShortint;
            bftSignedInt16:     AsSmallint := AsSmallint div Value.AsSmallint;
            bftSignedInt32:     AsInteger := AsInteger div Value.AsInteger;
            bftSignedInt64:     AsInt64 := AsInt64 div Value.AsInt64;

            bftUnsignedInt8:    AsByte := AsByte div Value.AsByte;
            bftUnsignedInt16:   AsWord := AsWord div Value.AsWord;
            bftUnsignedInt32:   AsCardinal := AsCardinal div Value.AsCardinal;

            bftSingle:          AsSingle := AsSingle / Value.AsSingle;
            bftDouble:          AsDouble := AsDouble / Value.AsDouble;
            bftExtended:        AsExtended := AsExtended / Value.AsExtended;

            bftDate:            AsACRDate := AsACRDate div Value.AsACRDate;
            bftTime:            AsACRTime := AsACRTime div Value.AsACRTime;
            bftDateTime:        begin
                                  dt := AsACRDateTime;
                                  dt.Date := dt.Date div Value.AsACRDateTime.Date;
                                  dt.Time := dt.Time div Value.AsACRDateTime.Time;
                                  if (dt.Time >= MILSECS_IN_DAY) then
                                   begin
                                    dt.Date := dt.Date + dt.Time div MILSECS_IN_DAY;
                                    dt.Time := dt.Time mod MILSECS_IN_DAY;
                                   end;
                                  AsACRDateTime := dt;
                                end;

            bftLogical:         AsLogical := AsLogical or Value.AsLogical;
            bftCurrency:        AsCurrency := AsCurrency / Value.AsCurrency;
            else
              raise EACRException.Create(30127, ErrorGNotNumericDataType, ['/']);
          end;
        except
          on EACRException do raise;
          on e: Exception do
            raise EACRException.Create(30128, ErrorGDivisionError, [e.Message]);
      end;
    end;
end;//Division
 {$IFDEF RESET_OVERFLOW}
  {$Q+}
 {$ENDIF}


//------------------------------------------------------------------------------
// Compare with another Variants
//------------------------------------------------------------------------------
function TACRVariant.Compare(Value: TACRVariant; TrueFalseNullLogic, CaseInsensitive, PartialKey: boolean): TACRCompareResult;
var
  i: Integer;
begin
  if (PartialKey) then
    i := min(StrLen, Value.StrLen)
  else
    i := -1;
  Result := CompareValueBuffers(FPData, Value.FPData,
                                FDataType, Value.DataType,
                                IsNull, Value.IsNull,
                                i, CaseInsensitive);
  if (not TrueFalseNullLogic) then
   case Result of
    cmprBothNull:
      Result := cmprEqual;
    cmprLeftNull:
      Result := cmprLower;
    cmprRightNull:
      Result := cmprGreater;
   end;
end;//Compare


//------------------------------------------------------------------------------
// Return Common DataType for 2 types, or ftUnknown
//------------------------------------------------------------------------------
function GetCommonDataType(a,b: TACRBaseFieldType): TACRBaseFieldType;
begin
{  if (a = b) then
    begin
      Result := a;
      Exit;
    end;
}
  Result := bftUnknown;
  case a of
    bftChar:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;
        bftClob:             Result := bftChar;
        bftWideClob:         Result := bftWideChar;

        bftSignedInt8:       Result := bftChar;
        bftSignedInt16:      Result := bftChar;
        bftSignedInt32:      Result := bftChar;
        bftSignedInt64:      Result := bftChar;
        bftUnsignedInt8:     Result := bftChar;
        bftUnsignedInt16:    Result := bftChar;
        bftUnsignedInt32:    Result := bftChar;

        bftSingle:           Result := bftChar;
        bftDouble:           Result := bftChar;
        bftExtended:         Result := bftChar;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftChar;
        bftCurrency:         Result := bftChar;

      end;
    bftWideChar:
      case b of
        bftChar:             Result := bftWideChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftWideChar;
        bftWideVarchar:      Result := bftWideVarchar;
        bftClob:             Result := bftChar;
        bftWideClob:         Result := bftWideChar;

        bftSignedInt8:       Result := bftWideChar;
        bftSignedInt16:      Result := bftWideChar;
        bftSignedInt32:      Result := bftWideChar;
        bftSignedInt64:      Result := bftWideChar;
        bftUnsignedInt8:     Result := bftWideChar;
        bftUnsignedInt16:    Result := bftWideChar;
        bftUnsignedInt32:    Result := bftWideChar;

        bftSingle:           Result := bftWideChar;
        bftDouble:           Result := bftWideChar;
        bftExtended:         Result := bftWideChar;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftWideChar;
        bftCurrency:         Result := bftWideChar;

      end;
    bftVarchar:
      case b of
        bftChar:             Result := bftVarchar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;
        bftClob:             Result := bftChar;
        bftWideClob:         Result := bftWideChar;

        bftSignedInt8:       Result := bftVarchar;
        bftSignedInt16:      Result := bftVarchar;
        bftSignedInt32:      Result := bftVarchar;
        bftSignedInt64:      Result := bftVarchar;
        bftUnsignedInt8:     Result := bftVarchar;
        bftUnsignedInt16:    Result := bftVarchar;
        bftUnsignedInt32:    Result := bftVarchar;

        bftSingle:           Result := bftVarchar;
        bftDouble:           Result := bftVarchar;
        bftExtended:         Result := bftVarchar;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftVarchar;
        bftCurrency:         Result := bftVarchar;

      end;
    bftWideVarchar:
      case b of
        bftChar:             Result := bftWideVarchar;
        bftWideChar:         Result := bftWideVarchar;
        bftVarchar:          Result := bftWideVarchar;
        bftWideVarchar:      Result := bftWideVarchar;
        bftClob:             Result := bftChar;
        bftWideClob:         Result := bftWideChar;

        bftSignedInt8:       Result := bftWideVarchar;
        bftSignedInt16:      Result := bftWideVarchar;
        bftSignedInt32:      Result := bftWideVarchar;
        bftSignedInt64:      Result := bftWideVarchar;
        bftUnsignedInt8:     Result := bftWideVarchar;
        bftUnsignedInt16:    Result := bftWideVarchar;
        bftUnsignedInt32:    Result := bftWideVarchar;

        bftSingle:           Result := bftWideVarchar;
        bftDouble:           Result := bftWideVarchar;
        bftExtended:         Result := bftWideVarchar;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftWideVarchar;
        bftCurrency:         Result := bftWideVarchar;

      end;
    bftClob:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;
        bftClob:             Result := bftClob;
        bftWideClob:         Result := bftWideClob;
       else
        Result := bftUnknown;
      end;
    bftWideClob:
      case b of
        bftChar:             Result := bftWideChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftWideVarchar;
        bftWideVarchar:      Result := bftWideVarchar;
        bftClob:             Result := bftWideClob;
        bftWideClob:         Result := bftWideClob;
       else
        Result := bftUnknown;
      end;
    bftSignedInt8:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt8;
        bftSignedInt16:      Result := bftSignedInt16;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftSignedInt16;
        bftUnsignedInt16:    Result := bftUnsignedInt16;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftSignedInt8;
        bftCurrency:         Result := bftCurrency;

      end;
    bftSignedInt16:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt16;
        bftSignedInt16:      Result := bftSignedInt16;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftSignedInt16;
        bftUnsignedInt16:    Result := bftUnsignedInt16;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftSignedInt16;
        bftCurrency:         Result := bftCurrency;

      end;
    bftSignedInt32:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt32;
        bftSignedInt16:      Result := bftSignedInt32;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftSignedInt32;
        bftUnsignedInt16:    Result := bftSignedInt32;
        bftUnsignedInt32:    Result := bftSignedInt64;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftSignedInt32;
        bftCurrency:         Result := bftCurrency;

      end;
    bftSignedInt64:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt64;
        bftSignedInt16:      Result := bftSignedInt64;
        bftSignedInt32:      Result := bftSignedInt64;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftSignedInt64;
        bftUnsignedInt16:    Result := bftSignedInt64;
        bftUnsignedInt32:    Result := bftSignedInt64;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftSignedInt64;
        bftCurrency:         Result := bftCurrency;

      end;
    bftUnsignedInt8:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt16;
        bftSignedInt16:      Result := bftSignedInt16;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftUnsignedInt8;
        bftUnsignedInt16:    Result := bftUnsignedInt16;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftUnsignedInt8;
        bftCurrency:         Result := bftCurrency;

      end;
    bftUnsignedInt16:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt32;
        bftSignedInt16:      Result := bftSignedInt32;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftUnsignedInt16;
        bftUnsignedInt16:    Result := bftUnsignedInt16;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftUnsignedInt16;
        bftCurrency:         Result := bftCurrency;

      end;
    bftUnsignedInt32:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt64;
        bftSignedInt16:      Result := bftSignedInt64;
        bftSignedInt32:      Result := bftSignedInt64;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftUnsignedInt32;
        bftUnsignedInt16:    Result := bftUnsignedInt32;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftUnsignedInt32;
        bftCurrency:         Result := bftCurrency;

      end;

    bftSingle:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSingle;
        bftSignedInt16:      Result := bftSingle;
        bftSignedInt32:      Result := bftSingle;
        bftSignedInt64:      Result := bftSingle;
        bftUnsignedInt8:     Result := bftSingle;
        bftUnsignedInt16:    Result := bftSingle;
        bftUnsignedInt32:    Result := bftSingle;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftSingle;
        bftCurrency:         Result := bftCurrency;

      end;
    bftDouble:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftDouble;
        bftSignedInt16:      Result := bftDouble;
        bftSignedInt32:      Result := bftDouble;
        bftSignedInt64:      Result := bftDouble;
        bftUnsignedInt8:     Result := bftDouble;
        bftUnsignedInt16:    Result := bftDouble;
        bftUnsignedInt32:    Result := bftDouble;

        bftSingle:           Result := bftDouble;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftDouble;
        bftCurrency:         Result := bftCurrency;

      end;
    bftExtended:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftExtended;
        bftSignedInt16:      Result := bftExtended;
        bftSignedInt32:      Result := bftExtended;
        bftSignedInt64:      Result := bftExtended;
        bftUnsignedInt8:     Result := bftExtended;
        bftUnsignedInt16:    Result := bftExtended;
        bftUnsignedInt32:    Result := bftExtended;

        bftSingle:           Result := bftExtended;
        bftDouble:           Result := bftExtended;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftExtended;
        bftCurrency:         Result := bftCurrency;

      end;

    bftDate:
      case b of
        bftChar:             Result := bftDate;
        bftWideChar:         Result := bftDate;
        bftVarchar:          Result := bftDate;
        bftWideVarchar:      Result := bftDate;

        bftSignedInt8:       Result := bftDate;
        bftSignedInt16:      Result := bftDate;
        bftSignedInt32:      Result := bftDate;
        bftSignedInt64:      Result := bftDate;
        bftUnsignedInt8:     Result := bftDate;
        bftUnsignedInt16:    Result := bftDate;
        bftUnsignedInt32:    Result := bftDate;

        bftSingle:           Result := bftDate;
        bftDouble:           Result := bftDate;
        bftExtended:         Result := bftDate;

        bftDate:             Result := bftDate;
        bftTime:             Result := bftDateTime;
        bftDateTime:         Result := bftDate;
//        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftUnknown;
        bftCurrency:         Result := bftUnknown;

      end;
    bftTime:
      case b of
        bftChar:             Result := bftTime;
        bftWideChar:         Result := bftTime;
        bftVarchar:          Result := bftTime;
        bftWideVarchar:      Result := bftTime;

        bftSignedInt8:       Result := bftTime;
        bftSignedInt16:      Result := bftTime;
        bftSignedInt32:      Result := bftTime;
        bftSignedInt64:      Result := bftTime;
        bftUnsignedInt8:     Result := bftTime;
        bftUnsignedInt16:    Result := bftTime;
        bftUnsignedInt32:    Result := bftTime;

        bftSingle:           Result := bftTime;
        bftDouble:           Result := bftTime;
        bftExtended:         Result := bftTime;

        bftDate:             Result := bftDateTime;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftTime;
//        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftUnknown;
        bftCurrency:         Result := bftUnknown;

      end;
    bftDateTime:
      case b of
        bftChar:             Result := bftDateTime;
        bftWideChar:         Result := bftDateTime;
        bftVarchar:          Result := bftDateTime;
        bftWideVarchar:      Result := bftDateTime;

        bftSignedInt8:       Result := bftDateTime;
        bftSignedInt16:      Result := bftDateTime;
        bftSignedInt32:      Result := bftDateTime;
        bftSignedInt64:      Result := bftDateTime;
        bftUnsignedInt8:     Result := bftDateTime;
        bftUnsignedInt16:    Result := bftDateTime;
        bftUnsignedInt32:    Result := bftDateTime;

        bftSingle:           Result := bftDateTime;
        bftDouble:           Result := bftDateTime;
        bftExtended:         Result := bftDateTime;

//        bftDate:             Result := bftDateTime;
//        bftTime:             Result := bftDateTime;
        bftDate:             Result := bftDate;
        bftTime:             Result := bftTime;
        bftDateTime:         Result := bftDateTime;

        bftLogical:          Result := bftUnknown;
        bftCurrency:         Result := bftUnknown;

      end;

    bftLogical:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt8;
        bftSignedInt16:      Result := bftSignedInt16;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftUnsignedInt8;
        bftUnsignedInt16:    Result := bftUnsignedInt16;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftUnknown;
        bftTime:             Result := bftUnknown;
        bftDateTime:         Result := bftUnknown;

        bftLogical:          Result := bftLogical;
        bftCurrency:         Result := bftUnknown;

      end;
    bftCurrency:
      case b of
        bftChar:             Result := bftChar;
        bftWideChar:         Result := bftWideChar;
        bftVarchar:          Result := bftVarchar;
        bftWideVarchar:      Result := bftWideVarchar;

        bftSignedInt8:       Result := bftSignedInt8;
        bftSignedInt16:      Result := bftSignedInt16;
        bftSignedInt32:      Result := bftSignedInt32;
        bftSignedInt64:      Result := bftSignedInt64;
        bftUnsignedInt8:     Result := bftUnsignedInt8;
        bftUnsignedInt16:    Result := bftUnsignedInt16;
        bftUnsignedInt32:    Result := bftUnsignedInt32;

        bftSingle:           Result := bftSingle;
        bftDouble:           Result := bftDouble;
        bftExtended:         Result := bftExtended;

        bftDate:             Result := bftUnknown;
        bftTime:             Result := bftUnknown;
        bftDateTime:         Result := bftUnknown;

        bftLogical:          Result := bftUnknown;
        bftCurrency:         Result := bftCurrency;

      end;

  end;
end;//GetCommonDataType


//------------------------------------------------------------------------------
// Return Common DataType for 2 types, or ftUnknown
//------------------------------------------------------------------------------
function GetCommonDataType(a,b: TACRAdvancedFieldType): TACRAdvancedFieldType;
var
   b1,b2, CommonBft: TACRBaseFieldType;
begin
  b1 := AdvancedFieldTypeToBaseFieldType(a);
  b2 := AdvancedFieldTypeToBaseFieldType(b);
  CommonBft := GetCommonDataType(b1, b2);

  if (CommonBft = b1) then
    Result := a
  else if (CommonBft = b2) then
    Result := b
  else
    Result := BaseFieldTypeToAdvancedFieldType(CommonBft);
    
end;//GetCommonDataType

//------------------------------------------------------------------------------
// Allocate new buffer and convert data to it
//------------------------------------------------------------------------------
function CastToNewBuffer(
                          const Buffer;
                          const DataType: TACRBaseFieldType;
                          const NewDataType: TACRBaseFieldType
                        ): PChar;
var
   Size: Integer;
begin
  Result := CastToNewBuffer(Buffer, DataType, NewDataType, Size);
end;//CastToNewBuffer


//------------------------------------------------------------------------------
// Allocate new buffer and convert data to it
//------------------------------------------------------------------------------
function CastToNewBuffer(
                          const Buffer;
                          const DataType: TACRBaseFieldType;
                          const NewDataType: TACRBaseFieldType;
                          out NewDataSize: Integer
                        ): PChar;
var

  s: String;
  ws: WideString;
  si8:  Shortint;
  si16: Smallint;
  si32: Integer; 
  si64: Int64; 
  ui8:  Byte;
  ui16: Word;
  ui32: Cardinal; 
  sng:  Single; 
  dbl:  Double;
  ext:  Extended;
  dt:   TACRDateTime; 
  d:    TACRDate;
  t:    TACRTime; 
  b:    TACRLogical;
  cur:  TACRCurrency; 
	
  oldSeparator: char; 
	
begin 
  Result := nil; 
  try
    case DataType of 
      bftChar, 
      bftVarchar:  //  from String 
        case NewDataType of 
          bftChar, 
          bftVarchar:
            begin
              AllocMemAndSetData(bftVarchar, Buffer, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar: 
            begin
              ws := WideString(PChar(Buffer)); 
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end; 
          bftSignedInt8: 
            begin
              si8 := StrToInt(PChar(Buffer)); 
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end; 
          bftSignedInt16: 
            begin
              si16 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize); 
            end; 
          bftSignedInt32: 
            begin 
              si32 := StrToInt(PChar(Buffer)); 
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize); 
            end; 
          bftSignedInt64:
            begin
              si64 := StrToInt64(PChar(Buffer)); 
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8: 
            begin
              ui8 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize); 
            end; 
          bftUnsignedInt16: 
            begin
              ui16 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize); 
            end;
          bftUnsignedInt32:
            begin 
              ui32 := StrToInt(PChar(Buffer)); 
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;
	
          bftSingle: 
            begin 
              oldSeparator := DecimalSeparator; 
              DecimalSeparator := '.'; 
              try
                sng := StrToFloat(PChar(Buffer)); 
              finally 
                DecimalSeparator := oldSeparator; 
              end;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble: 
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try 
                dbl := StrToFloat(PChar(Buffer)); 
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin 
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try 
                ext := StrToFloat(PChar(Buffer));
              finally
                DecimalSeparator := oldSeparator; 
              end; 
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize); 
            end; 
	
          bftDate: 
            begin 
              d := DateToACRDate(StrToDate(PChar(Buffer)));
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin 
              t := TimeToACRTime(StrToTime(PChar(Buffer))); 
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime: 
            begin 
              dt := DateTimeToACRDateTime(StrToDateTime(PChar(Buffer))); 
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
	
          bftLogical:
            begin
              s := AnsiUpperCase(PChar(Buffer)); 
              b := (s = '1') or (s = 'TRUE');
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize); 
            end;
          bftCurrency:
            begin
              cur := StrToCurr(PChar(Buffer));
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
          bftBlob,bftClob,bftWideClob:
            begin
              AllocMemAndSetData(bftVarchar, PChar(Buffer), Result, NewDataSize);
            end;
        end;

      bftWideChar,
      bftWideVarchar:  // from WideString
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := String(PWideChar(Buffer));
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := WideString(PWideChar(Buffer));
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := StrToInt(PWideChar(Buffer));
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := StrToInt(PWideChar(Buffer));
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := StrToInt(PWideChar(Buffer));
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := StrToInt64(PWideChar(Buffer));
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := StrToInt(PWideChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := StrToInt(PWideChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := StrToInt(PWideChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try
                sng := StrToFloat(PWideChar(Buffer));
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try
                dbl := StrToFloat(PWideChar(Buffer));
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try
                ext := StrToFloat(PWideChar(Buffer));
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := DateToACRDate(StrToDate(PWideChar(Buffer)));
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := TimeToACRTime(StrToTime(PWideChar(Buffer)));
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt := DateTimeToACRDateTime(StrToDateTime(PWideChar(Buffer)));
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;

          bftLogical:
            begin
              s := AnsiUpperCase(PWideChar(Buffer));
              b := (s = '1') or (s = 'TRUE');
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := StrToCurr(PWideChar(Buffer));
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
          bftBlob,bftClob,bftWideClob:
            begin
              ws := WideString(PWideChar(Buffer));
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
        end;


      bftSignedInt8:  // from Shortint
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PShortint(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PShortint(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PShortint(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PShortint(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PShortint(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PShortint(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PShortint(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PShortint(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PShortint(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PShortint(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PShortint(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PShortint(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PShortint(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PShortint(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PShortint(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PShortint(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PShortint(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;                        

      bftSignedInt16:  // fro Smallint
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PSmallint(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PSmallint(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PSmallint(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PSmallint(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PSmallint(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PSmallint(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PSmallint(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PSmallint(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PSmallint(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PSmallint(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PSmallint(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;


      bftSignedInt32:  // from Integer
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PInteger(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PInteger(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PInteger(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PInteger(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PInteger(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PInteger(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PInteger(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PInteger(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PInteger(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PInteger(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PInteger(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PInteger(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PInteger(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PInteger(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PInteger(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PInteger(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PInteger(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftSignedInt64:  // from int64
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PInt64(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PInt64(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PInt64(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PInt64(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PInt64(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PInt64(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PInt64(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PInt64(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PInt64(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PInt64(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PInt64(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PInt64(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PInt64(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PInt64(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PInt64(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PInt64(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PInt64(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;


      bftUnsignedInt8:  // from Byte
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PByte(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PByte(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PByte(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PByte(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PByte(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PByte(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PByte(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PByte(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PByte(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PByte(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PByte(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PByte(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PByte(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PByte(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PByte(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PByte(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PByte(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftUnsignedInt16:  // from Word
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PWord(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PWord(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PWord(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PWord(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PWord(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PWord(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PWord(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PWord(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PWord(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PWord(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PWord(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PWord(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PWord(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PWord(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PWord(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PWord(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PWord(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftUnsignedInt32:  // from Cardinal
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := IntToStr(PCardinal(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := IntToStr(PCardinal(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PCardinal(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PCardinal(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PCardinal(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PCardinal(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PCardinal(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PCardinal(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PCardinal(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PCardinal(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PCardinal(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftSingle:  // from Single
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := FloatToStr(PSingle(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := FloatToStr(PSingle(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PSingle(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PSingle(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PSingle(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := Trunc(PSingle(Buffer)^) mod MILSECS_IN_DAY;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := Trunc(Frac(PSingle(Buffer)^)*MILSECS_IN_DAY);
              dt.Date := Trunc(PSingle(Buffer)^);
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PSingle(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PSingle(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftDouble:  // From Double
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := FloatToStr(PDouble(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := FloatToStr(PDouble(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PDouble(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PDouble(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PDouble(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := Trunc(PDouble(Buffer)^) mod MILSECS_IN_DAY;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := Trunc(Frac(PDouble(Buffer)^)*MILSECS_IN_DAY);
              dt.Date := Trunc(PDouble(Buffer)^);
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PDouble(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PDouble(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftExtended: // From Extended
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := FloatToStr(PExtended(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := FloatToStr(PExtended(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PExtended(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PExtended(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PExtended(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := Trunc(PExtended(Buffer)^) mod MILSECS_IN_DAY;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := Trunc(Frac(PExtended(Buffer)^)*MILSECS_IN_DAY);
              dt.Date := Trunc(PExtended(Buffer)^);
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PExtended(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PExtended(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;


      bftDate:  // from Date
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              d := PACRDate(Buffer)^;
              if (d = DATE_IS_NULL) then
                s := ''
              else
                s := DateToStr(ACRDateToDate(d));
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              d := PACRDate(Buffer)^;
              if (d = DATE_IS_NULL) then
                ws := ''
              else
                ws := DateToStr(ACRDateToDate(d));
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PACRDate(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PACRDate(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PACRDate(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PACRDate(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PACRDate(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PACRDate(Buffer)^ mod MILSECS_IN_DAY;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := TIME_IS_NULL;
              dt.Date := PACRDate(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PACRDate(Buffer)^ = DATE_IS_NULL);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PACRDate(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;


      bftTime:  // from Time
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              t := PACRTime(Buffer)^;
              if (t = TIME_IS_NULL) then
                s := ''
              else
                s := TimeToStr(ACRTimeToTime(t));
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              t := PACRTime(Buffer)^;
              if (t = TIME_IS_NULL) then
                ws := ''
              else
                ws := TimeToStr(ACRTimeToTime(T));
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PACRTime(Buffer)^;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PACRTime(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PACRTime(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PACRTime(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PACRTime(Buffer)^;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PACRTime(Buffer)^;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt.Time := PACRTime(Buffer)^;
              dt.Date := DATE_IS_NULL;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not (PACRTime(Buffer)^ = TIME_IS_NULL);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PACRTime(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;


      bftDateTime:  // from DateTime
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              dt := PACRDateTime(Buffer)^;
              if (dt.Date = DATE_IS_NULL) and (dt.Time = TIME_IS_NULL) then
                s := ''
              else
                s := DateTimeToStr(ACRDateTimeToDateTime(dt));
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              dt := PACRDateTime(Buffer)^;
              if (dt.Date = DATE_IS_NULL) and (dt.Time = TIME_IS_NULL) then
                ws := ''
              else
                ws := DateTimeToStr(ACRDateTimeToDateTime(dt));
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := 0;
              if (PACRDateTime(Buffer)^.Date <> DATE_IS_NULL) then
                sng := sng + PACRDateTime(Buffer)^.Date;
              if (PACRDateTime(Buffer)^.Time <> TIME_IS_NULL) then
                sng := sng + MILSECS_IN_DAY / PACRDateTime(Buffer)^.Time;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := 0;
              if (PACRDateTime(Buffer)^.Date <> DATE_IS_NULL) then
                dbl := dbl + PACRDateTime(Buffer)^.Date;
              if (PACRDateTime(Buffer)^.Time <> TIME_IS_NULL) then
                dbl := dbl + MILSECS_IN_DAY / PACRDateTime(Buffer)^.Time;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := 0;
              if (PACRDateTime(Buffer)^.Date <> DATE_IS_NULL) then
                ext := ext + PACRDateTime(Buffer)^.Date;
              if (PACRDateTime(Buffer)^.Time <> TIME_IS_NULL) then
                ext := ext + MILSECS_IN_DAY / PACRDateTime(Buffer)^.Time;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := PACRDateTime(Buffer)^.Date;
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := PACRDateTime(Buffer)^.Time;
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt := PACRDateTime(Buffer)^;
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;
          bftLogical:
            begin
              b := not ((PACRDateTime(Buffer)^.Date = DATE_IS_NULL) and
                        (PACRDateTime(Buffer)^.Time = TIME_IS_NULL) );
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              ext := 0;
              if (PACRDateTime(Buffer)^.Date <> DATE_IS_NULL) then
                ext := ext + PACRDateTime(Buffer)^.Date;
              if (PACRDateTime(Buffer)^.Time <> TIME_IS_NULL) then
                ext := ext + MILSECS_IN_DAY / PACRDateTime(Buffer)^.Time;
              cur := FloatToCurr(ext);
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;


      bftLogical:  // From Logical
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := BoolToStr(PACRLogical(Buffer)^, true);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := BoolToStr(PACRLogical(Buffer)^, true);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              if PACRLogical(Buffer)^ then
                si8 := 1
              else
                si8 := 0;
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              if PACRLogical(Buffer)^ then
                si16 := 1
              else
                si16 := 0;
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              if PACRLogical(Buffer)^ then
                si32 := 1
              else
                si32 := 0;
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              if PACRLogical(Buffer)^ then
                si64 := 1
              else
                si64 := 0;
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              if PACRLogical(Buffer)^ then
                ui8 := 1
              else
                ui8 := 0;
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              if PACRLogical(Buffer)^ then
                ui16 := 1
              else
                ui16 := 0;
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              if PACRLogical(Buffer)^ then
                ui32 := 1
              else
                ui32 := 0;
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              if PACRLogical(Buffer)^ then
                sng := 1
              else
                sng := 0;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              if PACRLogical(Buffer)^ then
                dbl := 1
              else
                dbl := 0;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              if PACRLogical(Buffer)^ then
                ext := 1
              else
                ext := 0;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:;      // No CAST
          bftTime:;
          bftDateTime:;

          bftLogical:
            begin
              b := PACRLogical(Buffer)^;
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency: ; // No CAST
        end;

      bftCurrency:  // from Currency
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              s := CurrToStr(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftVarchar, s, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := CurrToStr(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := Trunc(PACRCurrency(Buffer)^);
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              sng := PACRCurrency(Buffer)^;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              dbl := PACRCurrency(Buffer)^;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              ext := PACRCurrency(Buffer)^;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:;     // No CAST
          bftTime:;
          bftDateTime:;
          bftLogical:
            begin
              b := not (PACRCurrency(Buffer)^ = 0);
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := PACRCurrency(Buffer)^;
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
        end;

      bftBlob,bftClob,bftWideClob:  //  from Blob
        case NewDataType of
          bftChar,
          bftVarchar:
            begin
              AllocMemAndSetData(bftVarchar, Buffer, Result, NewDataSize);
            end;
          bftWideChar,
          bftWideVarchar:
            begin
              ws := WideString(PChar(Buffer));
              AllocMemAndSetData(bftWideVarchar, ws, Result, NewDataSize);
            end;
          bftSignedInt8:
            begin
              si8 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftSignedInt8, si8, Result, NewDataSize);
            end;
          bftSignedInt16:
            begin
              si16 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftSignedInt16, si16, Result, NewDataSize);
            end;
          bftSignedInt32:
            begin
              si32 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftSignedInt32, si32, Result, NewDataSize);
            end;
          bftSignedInt64:
            begin
              si64 := StrToInt64(PChar(Buffer));
              AllocMemAndSetData(bftSignedInt64, si64, Result, NewDataSize);
            end;
          bftUnsignedInt8:
            begin
              ui8 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt8, ui8, Result, NewDataSize);
            end;
          bftUnsignedInt16:
            begin
              ui16 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt16, ui16, Result, NewDataSize);
            end;
          bftUnsignedInt32:
            begin
              ui32 := StrToInt(PChar(Buffer));
              AllocMemAndSetData(bftUnsignedInt32, ui32, Result, NewDataSize);
            end;

          bftSingle:
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try
                sng := StrToFloat(PChar(Buffer));
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftSingle, sng, Result, NewDataSize);
            end;
          bftDouble:
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try
                dbl := StrToFloat(PChar(Buffer));
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftDouble, dbl, Result, NewDataSize);
            end;
          bftExtended:
            begin
              oldSeparator := DecimalSeparator;
              DecimalSeparator := '.';
              try
                ext := StrToFloat(PChar(Buffer));
              finally
                DecimalSeparator := oldSeparator;
              end;
              AllocMemAndSetData(bftExtended, ext, Result, NewDataSize);
            end;

          bftDate:
            begin
              d := DateToACRDate(StrToDate(PChar(Buffer)));
              AllocMemAndSetData(bftDate, d, Result, NewDataSize);
            end;
          bftTime:
            begin
              t := TimeToACRTime(StrToTime(PChar(Buffer)));
              AllocMemAndSetData(bftTime, t, Result, NewDataSize);
            end;
          bftDateTime:
            begin
              dt := DateTimeToACRDateTime(StrToDateTime(PChar(Buffer)));
              AllocMemAndSetData(bftDateTime, dt, Result, NewDataSize);
            end;

          bftLogical:
            begin
              s := AnsiUpperCase(PChar(Buffer));
              b := (s = '1') or (s = 'TRUE');
              AllocMemAndSetData(bftLogical, b, Result, NewDataSize);
            end;
          bftCurrency:
            begin
              cur := StrToCurr(PChar(Buffer));
              AllocMemAndSetData(bftCurrency, cur, Result, NewDataSize);
            end;
          bftBlob,bftClob,bftWideClob:
            begin
              AllocMemAndSetData(bftVarchar, PChar(Buffer), Result, NewDataSize);
            end;
        end;

      //else raise EACRException.Create
    end;
  except
    raise EACRException.Create(30051, ErrorGCastError,
                                   [BftToStr(DataType), BftToStr(NewDataType)]);
  end;

  if (Result = nil) then
    raise EACRException.Create(30052, ErrorGCastError,
                                  [BftToStr(DataType), BftToStr(NewDataType)]);
end;//Cast



//------------------------------------------------------------------------------
// Compare 2 values
//------------------------------------------------------------------------------
function CompareValueBuffers(
                              Buffer1, Buffer2: Pointer;
                              BaseFieldType1, BaseFieldType2: TACRBaseFieldType;
                              IsField1Null: Boolean = false;
                              IsField2Null: Boolean = false;
                              PartialCompareLength: Integer = -1;
                              IgnoreCase: Boolean = false;
                              LocaleID: LCID = LOCALE_USER_DEFAULT
                            ): TACRCompareResult;
var
   Value1,Value2 : Pointer;
   NewBuffer1, NewBuffer2: Pointer;

   CommonType: TACRBaseFieldType;
   StrCmpResult: Integer;
   StrCmpFlags: DWORD;
   //Variant1, Variant2: TACRVariant;
{$IFDEF LINUX}
   Str1, Str2 : String;
   WStr1, WStr2 : WideString;
{$ENDIF}
begin
 Result := cmprEqual;
 // Check NULLs
 if (IsField1Null or IsField2Null) then
  begin
     if (IsField1Null and IsField2Null) then
      result := cmprBothNull
     else
     if (IsField1Null) then
       result := cmprLeftNull
     else
       result := cmprRightNull;
   Exit;
  end;

  NewBuffer1 := nil;
  NewBuffer2 := nil;
  try

  // Cast both to common Type
  if (BaseFieldType1 = BaseFieldType2) then
   begin
     Value1 := Buffer1;
     Value2 := Buffer2;
     CommonType := BaseFieldType1;
   end
  else
   begin
     CommonType := GetCommonDataType(BaseFieldType1, BaseFieldType2);
     // Cannot compare
     if (CommonType = bftUnknown) then
       raise EACRException.Create(30044, ErrorGCatnnotCompareIncompatibleDataTypes,
                                  [bftToStr(BaseFieldType1), bftToStr(BaseFieldType2)]);

     // Cast 1-st parameter
     if (CommonType <> BaseFieldType1) then
       begin
         NewBuffer1 := CastToNewBuffer(Buffer1, BaseFieldType1, CommonType);
         Value1 := NewBuffer1;
       end
     else
       Value1 := Buffer1;

     // Cast 2-nd parameter
     if (CommonType <> BaseFieldType2) then
       begin
         NewBuffer2 := CastToNewBuffer(Buffer2, BaseFieldType2, CommonType);
         Value2 := NewBuffer2;
       end
     else
       Value2 := Buffer2;

   end;//Cast both to common Type

  if ((Value1 = nil) or (Value2 = nil)) then
    raise EACRException.Create(30140, ErrorGValueIsNull);

  case CommonType of
    bftChar,
    bftVarchar:
      begin
        StrCmpFlags := SORT_STRINGSORT;
        if IgnoreCase then
          StrCmpFlags := StrCmpFlags + NORM_IGNORECASE;
{$IFDEF MSWINDOWS}
        StrCmpResult := Windows.CompareString(LocaleID, StrCmpFlags,
                                              Value1, PartialCompareLength,
                                              Value2, PartialCompareLength);
{$ENDIF}
{$IFDEF LINUX}
        if (PartialCompareLength > 0) then
          begin
           if IgnoreCase then
             StrCmpResult := SysUtils.AnsiStrLIComp(
                                            Value1, Value2,
                                            PartialCompareLength
                                                       )
            else
             StrCmpResult := SysUtils.AnsiStrLComp(
                                            Value1, Value2,
                                            PartialCompareLength
                                                      );
          end
         else
          begin
{ TODO -oAlex : uncomment it, debug and comment your code}
{
 if (IgnoreCase) then
  StrCmpResult := SysUtils.WideCompareText(WideString(PWideChar(Value1)^),
                                           WideString(PWideChar(Value2)^))
 else
  StrCmpResult := SysUtils.WideCompareStr(WideString(PWideChar(Value1)^),
                                           WideString(PWideChar(Value2)^));
}
           SetLength(Str1, MAX_STRING_LENGTH);
           SetLength(Str2, MAX_STRING_LENGTH);
           StrCopy(PChar(Str1), Value1);
           StrCopy(PChar(Str2), Value2);
           if IgnoreCase then
             StrCmpResult := SysUtils.AnsiCompareText(Str1, Str2)
            else
             StrCmpResult := SysUtils.AnsiCompareStr(Str1, Str2);
          end;
        if (StrCmpResult < 0) then
          StrCmpResult := 1 else
        if (StrCmpResult = 0) then
          StrCmpResult := 2
         else
          StrCmpResult := 3; // (StrCmpResult > 0)
{$ENDIF}
        case StrCmpResult of
          1:   Result := cmprLower;
          2:   Result := cmprEqual;
          3:   Result := cmprGreater;
            else raise EACRException.Create(30045, ErrorGUnknownReturnCode, [StrCmpResult]);
        end;
      end;
    bftWideChar,
    bftWideVarchar:
      begin
        StrCmpFlags := SORT_STRINGSORT;
        if IgnoreCase then
          StrCmpFlags := StrCmpFlags + NORM_IGNORECASE;
{$IFDEF MSWINDOWS}
        StrCmpResult := Windows.CompareStringW(LocaleID, StrCmpFlags,
                                               Value1, PartialCompareLength,
                                               Value2,PartialCompareLength);
{$ENDIF}
{$IFDEF LINUX}
{ TODO -oAlex : uncomment it, debug and comment your code}
{
 if (IgnoreCase) then
  StrCmpResult := SysUtils.WideCompareText(WideString(PWideChar(Value1)^),
                                           WideString(PWideChar(Value2)^))
 else
  StrCmpResult := SysUtils.WideCompareStr(WideString(PWideChar(Value1)^),
                                           WideString(PWideChar(Value2)^));
}

        if (PartialCompareLength > 0) then
          begin
           SetLength(WStr1, PartialCompareLength * 2);
           SetLength(WStr2, PartialCompareLength * 2);
           Move(Value1^, PChar(WStr1)^, PartialCompareLength * 2);
           Move(Value2^, PChar(WStr2)^, PartialCompareLength * 2);
          end
         else
          begin
           SetLength(WStr1, MAX_STRING_LENGTH);
           SetLength(WStr2, MAX_STRING_LENGTH);
           StrCopy(PChar(WStr1), Value1);
           StrCopy(PChar(WStr2), Value2);
          end;
        if IgnoreCase then
          StrCmpResult := SysUtils.WideCompareText(WStr1, WStr2) // WideAnsiUpperCase in SysUtils does not work correctly
         else
          StrCmpResult := SysUtils.WideCompareStr(WStr1, WStr2);
        if (StrCmpResult < 0) then
          StrCmpResult := 1 else
        if (StrCmpResult = 0) then
          StrCmpResult := 2
         else
          StrCmpResult := 3; // (StrCmpResult > 0)
{$ENDIF}
        case StrCmpResult of
          1:   Result := cmprLower;
          2:   Result := cmprEqual;
          3:   Result := cmprGreater;
            else raise EACRException.Create(30046, ErrorGUnknownReturnCode, [StrCmpResult]);
        end;
      end;
    bftSignedInt8:
      begin
        if (pShortint(Value1)^ = pShortint(Value2)^) then
          Result := cmprEqual
        else
          if (pShortint(Value1)^ > pShortint(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftSignedInt16:
      begin
        if (pSmallInt(Value1)^ = pSmallInt(Value2)^) then
          Result := cmprEqual
        else
          if (pSmallInt(Value1)^ > pSmallInt(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftSignedInt32:
      begin
        if (pInteger(Value1)^ = pInteger(Value2)^) then
          Result := cmprEqual
        else
          if (pInteger(Value1)^ > pInteger(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftSignedInt64:
      begin
        if (pInt64(Value1)^ = pInt64(Value2)^) then
          Result := cmprEqual
        else
          if (pInt64(Value1)^ > pInt64(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftUnSignedInt8:
      begin
        if (pByte(Value1)^ = pByte(Value2)^) then
          Result := cmprEqual
        else
          if (pByte(Value1)^ > pByte(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftUnSignedInt16:
      begin
        if (pWord(Value1)^ = pWord(Value2)^) then
          Result := cmprEqual
        else
          if (pWord(Value1)^ > pWord(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftUnSignedInt32:
      begin
        if (pLongword(Value1)^ = pLongword(Value2)^) then
          Result := cmprEqual
        else
          if (pLongword(Value1)^ > pLongword(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftSingle:
      begin
        if (pSingle(Value1)^ = pSingle(Value2)^) then
          Result := cmprEqual
        else
          if (pSingle(Value1)^ > pSingle(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftDouble:
      begin
        if (pDouble(Value1)^ = pDouble(Value2)^) then
          Result := cmprEqual
        else
          if (pDouble(Value1)^ > pDouble(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftExtended:
      begin
        if (pExtended(Value1)^ = pExtended(Value2)^) then
          Result := cmprEqual
        else
          if (pExtended(Value1)^ > pExtended(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftDate:
      begin
        if (PACRDate(Value1)^ = PACRDate(Value2)^) then
          Result := cmprEqual
        else
          if (PACRDate(Value1)^ > PACRDate(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftTime:
      begin
        if (PACRTime(Value1)^ = PACRTime(Value2)^) then
          Result := cmprEqual
        else
          if (PACRTime(Value1)^ > PACRTime(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftDateTime:
      begin
        // Comparing Date perts
        if (PACRDateTime(Value1)^.Date = PACRDateTime(Value2)^.Date) then
          // Time part IS NULL --> result = ???
          //if ( (PACRDateTime(Value1)^.Time = TIME_IS_NULL) or
          //     (PACRDateTime(Value2)^.Time = TIME_IS_NULL) then ;

          // date1 = date2. Comparing times
          if (PACRDateTime(Value1)^.Time = PACRDateTime(Value2)^.Time) then
            // Time1 = Time 2
            Result := cmprEqual
          else
            // Time1 <> Time 2
            if (PACRDateTime(Value1)^.Time > PACRDateTime(Value2)^.Time) then
              Result := cmprGreater
            else
              Result := cmprLower
        else
          // Date1 <> Date2
          if (PACRTime(Value1)^ > PACRTime(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftLogical:
      begin
        if (PACRLogical(Value1)^ = PACRLogical(Value2)^) then
          Result := cmprEqual
        else
          if (PACRLogical(Value1)^ > PACRLogical(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;
    bftCurrency:
      begin
        if (PACRCurrency(Value1)^ = PACRCurrency(Value2)^) then
          Result := cmprEqual
        else
          if (PACRCurrency(Value1)^ > PACRCurrency(Value2)^) then
            Result := cmprGreater
          else
            Result := cmprLower;
      end;

    else
        raise EACRException.Create(30047, ErrorGUnknownDataType, [BftToStr(CommonType)]);
    end;
  finally
    if NewBuffer1 <> nil then
      MemoryManager.FreeAndNilMem(NewBuffer1);
    if NewBuffer2 <> nil then
      MemoryManager.FreeAndNilMem(NewBuffer2);
  end;
end;//CompareValueBuffers


//------------------------------------------------------------------------------
// Compare 2 values in sense of order (-1,0,1)
//------------------------------------------------------------------------------
function CompareValueBuffersForOrder(
                              Buffer1, Buffer2: Pointer;
                              BaseFieldType1, BaseFieldType2: TACRBaseFieldType;
                              IsField1Null: Boolean = false;
                              IsField2Null: Boolean = false;
                              PartialCompareLength: Integer = -1;
                              IgnoreCase: Boolean = false;
                              LocaleID: LCID = LOCALE_USER_DEFAULT
                            ): Integer;
var
  cmpResult: TACRCompareResult;
begin
  cmpResult := CompareValueBuffers(
              Buffer1, Buffer2, BaseFieldType1, BaseFieldType2,
              IsField1Null, IsField2Null, PartialCompareLength,
              IgnoreCase, LocaleID);
  case cmpResult of
   cmprBothNull, cmprEqual:    Result := 0;
   cmprLeftNull, cmprLower:    Result := -1;
   cmprRightNull, cmprGreater: Result := 1;
  else
   Result := 0;
  end;
end;// CompareValueBuffersForOrder

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRVariant> initialized');
{$ENDIF}

end.

