unit ACRConnection;

interface

{$I ACRVer.inc}


{$IFDEF DEBUG_LOG_NETWORK_THREADS}
{DEFINE LOG_CLIENT_THREADS}
{$DEFINE LOG_SERVER_THREADS}
{$ENDIF}


{DEFINE LOGIC_TEST}
{DEFINE CONNECTION_TEST}
{$IFNDEF CONNECTION_TEST}
 {DEFINE NETWORK_TEST}
{$ENDIF CONNECTION_TEST}

{$IFNDEF API_NETWORK}
 {$IFDEF D6H}
  {$DEFINE ProcessMessages} // Needs for Indy only
 {$ENDIF D6H}
{$ENDIF API_NETWORK}

uses
  Classes, Math, SysUtils,
{$IFDEF MSWINDOWS}
  Windows,
  Forms,
{$ENDIF}
{$IFDEF LINUX}
  Libc,
  QForms,
{$ENDIF}

{$IFNDEF API_NETWORK}
 {$IFDEF D6H}
  IdBaseComponent, IdComponent, IdUDPBase, IdUDPClient, IdUDPServer, IdSocketHandle,
 {$ELSE}
  NMUDP,
 {$ENDIF}
{$ELSE}
// Accuracer units
  ACRNetwork,
  ACRCrypto,
{$ENDIF}

{$IFDEF DEBUG_LOG}
  ACRDebug,
{$ENDIF}
  ACRCompression,
  ACRBase,
  ACRTypes,
  ACRConst,
  ACRExcept,
  ACRLinux,
  ACRCriticalSection,
  ACRBaseEngine,
  ACRMemory;

const

{******************************************************************************}
// ControlCode (-128..127):
{******************************************************************************}
// From Client
                     //  8...9  -- reserwed for Accuracer commands
 ACRNewRequest = 10; // 10...32 -- reserwed for MsgCommunicator commands
                     // 33...63 -- reserwed for custom commands
 ACRConnect = 1;
// From Server
 ACRConnected = 2;
// Both
 ACRNoAction = 0;
 ACRMessage = 3;
// ACRLastPacket: 7th bit = 1, i.e. byte = x1xxxxxx, i.e. code >= 64
 ACRLastPacket = 64;
 ACREcho = 4;
 ACRTunning = 5;
 ACRServerSessionTunning = 6;
 ACRPing = 7;
// single packet:
 ACRAllPacketsReceived = -1;
 ACRPacketResendRequest = -2;
 ACRMessageReceived = -3;
 ACRMessagePacketResendRequest = -4;
 ACRDisconnect = -55;

{$IFDEF MsgCommunicator}
// Both
 ACRGetUserInfo = 10;
 ACRRegisterNewUser = 11;
 ACRAddUserToContacts = 12;
 ACRRemoveUserFromContacts = 13;
 ACRGetContacts = 14;
 ACRUpdateUserInfo = 15;
 ACRGetUserID = 16;
 ACRFindUsers = 17;
 ACRFindMessages = 18;
 ACRIsUserExisting = 19;
 ACRIsUserOnline = 20;
 ACRUpdateUserInContacts = 21;
{$ENDIF}

{******************************************************************************}


{******************************************************************************}
// Sending Status - Fof Commands and Messages
{******************************************************************************}
  ACRNotSent = 0;
  ACRSent = -1;
{******************************************************************************}

{******************************************************************************}
// Receiving Status
{******************************************************************************}
 ACRNo = 0;
 ACRStart = 1;
 ACRNotFull = 2;
 ACRFull = -1;
{******************************************************************************}

{******************************************************************************}
// Session Control Codes
{******************************************************************************}
 ACRExecute = 0;
 ACRSuspend = 1;
 ACRTerminate = -1;
{******************************************************************************}


{******************************************************************************}
// Session Status
{******************************************************************************}
 ACRVacant = 0;
 ACRInUse = -1;
{******************************************************************************}

 ACRNoResending = 999;

{******************************************************************************}


////////////////////////////////////////////////////////////////////////////////
// Error Codes
////////////////////////////////////////////////////////////////////////////////

{******************************************************************************}
// location
{******************************************************************************}
  // client
  AcrClientListenerThread =          70;
  AcrClientResendRequestThread =     71;
  AcrClientMsgResendRequestThread =  72;

  AcrClientConnectionManager =       79;

  // server
  AcrServerListenerThread =          80;
  AcrServerResendRequestThread =     81;
  AcrServerMsgResendRequestThread =  82;
  AcrServerSessionThread =           83;
  AcrServerSessionMsgThread =        84;
  AcrServerSessionDisconnectThread = 85;
  AcrServerSessionTerminatorThread = 86;
  AcrServerPingClientsThread =       87;

  AcrServerTerminateCommandThreads = 90;
  AcrServerTerminateMessageThreads = 91;

  AcrServerDeleteSession           = 92;

  AcrServerConnectionManager =       99;

{******************************************************************************}
// functions
{******************************************************************************}
  // client
  AcrClntEchoRecv = 1;
  AcrClntEchoSend = 2;
  AcrClntEchoDND  = 3;
  AcrClntRecvDND  = 4;
  AcrClntMsgDND   = 5;

  // server
  AcrSrvrRecv     = 31;
  AcrSrvrRecvDND  = 32;
  AcrSrvrEchoSend = 33;
  AcrSrvrEchoDND  = 34;
  AcrSrvrMsgRecv  = 35;
  AcrSrvrMsgDND   = 36;



////////////////////////////////////////////////////////////////////////////////
// To avoid multiple realloc
////////////////////////////////////////////////////////////////////////////////

 ACRDefaultPacketsInAnswer = 8;
 ACRDefaultPacketsInRequest = 2;
 ACRDefaultMsgPacketsInAnswer = 1;
 ACRDefaultMsgPacketsInRequest = 1;

// Signatures
{$IFDEF MsgCommunicator}
 ACRClientPacketSign = 'MSG1';
 ACRServerPacketSign = 'MSG2';
{$ELSE}
 ACRClientPacketSign = 'ACR1';
 ACRServerPacketSign = 'ACR2';
{$ENDIF}

type

  PThread = ^TThread;

  TACRSessionsArray = array of TACRBaseSession;

  TACRNetwork = class;
  TACRBaseConnectionManager = class;
{$IFDEF CLIENT_VERSION}
  TACRClientConnectionManager = class;
  TACRClientResendRequestThread = class;
  TACRClientMsgResendRequestThread = class;
{$ENDIF}
{$IFDEF SERVER_VERSION}
  TACRServerConnectionManager = class;
  TACRServerSessionThread = class;
  TACRServerSessionMsgThread = class;
  TACRServerSessionDisconnectThread = class;
  TACRServerResendRequestThread = class;
  TACRServerMsgResendRequestThread = class;
{$ENDIF}

  TACRControlCode = ShortInt;
  TACRNetworkClientID = Integer;
  TACRConnectionID = Integer;

  TACRPacketHeader = packed record
    CheckSum:           Integer;
    Signature:          array [0..3] of Char;
    Recepient:          TACRNetworkClientID;
    Sender:             TACRNetworkClientID;
    ConnectionID:       TACRConnectionID;
    SessionID:          TACRSessionID;
    CurrentRequestID:   Integer;
    PacketID:           Integer;
    ControlCode:        TACRControlCode;
  end;
  PACRPacketHeader = ^TACRPacketHeader;

  TACRPacket = packed record
    Buffer:             PChar;
    BufferSize:         Integer;
  end;
  PACRPacket = ^TACRPacket;

  TACRConnectionParams = packed record
    PacketSize:                   Integer;
    CompressionAlgorithm:         Byte;
    CompressionMode:              Byte;
  end;
  PACRConnectionParams = ^TACRConnectionParams;

{$IFDEF CLIENT_VERSION}
  TACRClntSession = packed record
    Session:                TACRBaseSession;
    ServerSessionID:        TACRSessionID;
    CurrentRequestID:       Integer;
    Packets:                TList;
    ListeningThreads:       TThreadList;
    MsgListeningThreads:    TThreadList;
    ResendRequestThread:    TACRClientResendRequestThread;
    MsgResendRequestThread: TACRClientMsgResendRequestThread;
    AnswerStatus:           Integer;
    SendStatus:             Integer;
    PacketIDsToResend:      TACRIntegerArray;
    ClientMessageID:        Integer;
    ServerMessageID:        Integer;
    MsgPackets:             TList;
    MsgPacketIDsToResend:   TACRIntegerArray;
    MsgSendStatus:          Integer;
    MsgReceiveStatus:       Integer;
    ConnectionID:           TACRConnectionID;
    RemoteConnectionID:     TACRConnectionID;
    ControlCode:            Shortint;
    MsgControlCode:         Shortint;
    Status:                 Shortint;
    AnswerTime:             DWORD;
  end;
  PACRClntSession = ^TACRClntSession;

  TACRClntConnection = packed record
    ConnectionID:       TACRConnectionID;
    Network:            TACRNetwork;
  end;
  PACRClntConnection = ^TACRClntConnection;
{$ENDIF}

{$IFDEF SERVER_VERSION}
  TACRSrvrSession = packed record
    Session:                      TACRBaseSession;
    CurrentRequestID:             Integer;
    Packets:                      TThreadList;
    ReceiveStatus:                Shortint;               // for ReceiveBuffer
    SendStatus:                   Shortint;               // for SendBuffer
    MsgReceiveStatus:             Shortint;
    MsgSendStatus:                Shortint;
    PacketIDsToResend:            TACRThreadIntArray;
    ClientMessageID:              Integer;
    ServerMessageID:              Integer;
    MsgPackets:                   TThreadList;
    MsgPacketIDsToResend:         TACRThreadIntArray;
    ClientSessionID:              TACRSessionID;
    ConnectionID:                 TACRConnectionID;
    RemoteClientID:               TACRNetworkClientID;
    RemoteHost:                   String;
    RemotePort:                   Integer;
    Application:                  String;
    ListeningThreads:             TThreadList;
    MsgListeningThreads:          TThreadList;
    Thread:                       TACRServerSessionThread;
    MsgThread:                    TACRServerSessionMsgThread;
    DisconnectThread:             TACRServerSessionDisconnectThread;
    ControlCode:                  Shortint;
    MsgControlCode:               Shortint;
    Status:                       Shortint;
    Connected:                    Boolean; // used in Ping thread to avoid sending to non-specified remote address
    AnswerTime:                   DWORD;
    LastSendPingTime:             DWORD;
    LastReceivePingTime:          DWORD;
  end;

  PACRSrvrSession = ^TACRSrvrSession;
{$ENDIF}



////////////////////////////////////////////////////////////////////////////////
//
// TACRNetwork
//
////////////////////////////////////////////////////////////////////////////////

  TACRNetwork = class (TObject)
   private
    FManager:             TACRBaseConnectionManager;
    FLocalClient:         TACRNetworkClientID;
{$IFNDEF API_NETWORK}
    FLocalHost:           String;
    FRemoteHost:          String;
    FRemotePort:          Integer;
{$ENDIF API_NETWORK}
    function RndClientID: TACRNetworkClientID;
    procedure SetRemoteHost(Host: String);
    function GetRemoteHost: String;
    procedure SetRemotePort(Port: Integer);
    function GetRemotePort: Integer;
    procedure SetLocalHost(Host: String);
    function GetLocalHost: String;
    procedure SetLocalPort(Port: Integer);
    function GetLocalPort: Integer;
   public
{$IFDEF API_NETWORK}
    FACRNetwork:          TACRapiNetwork;
{$ELSE}
 {$IFDEF D6H}
//    FNetworkClient:       TIdUDPClient;
    FNetworkServer:       TIdUDPServer;
 {$ELSE}
    FTransport:           TNMUDP;
 {$ENDIF}
{$ENDIF}
   protected
    procedure CreateNetwork;
    procedure FreeNetwork;
    procedure ReCreateNetwork;
   public
    constructor Create(ConnectionManager: TACRBaseConnectionManager);
    destructor Destroy; override;
    procedure SendBuffer(
                          Buffer: PChar;
                          Count:  Integer
                         );
    procedure OnDisconnect(
                             FromHost:  String;
                             FromPort:  Integer
                             );
{$IFDEF API_NETWORK}
    procedure OnDataReceived(
                             Buffer:    PChar;
                             Count:     Integer;
                             FromHost:  String;
                             FromPort:  Integer
                             );
{$ELSE}
 {$IFDEF D6H}
    procedure OnDataReceived(
                             Sender:   TObject;
                             AData:    TStream;
                             ABinding: TIdSocketHandle
                             );
 {$ELSE}
    procedure OnDataReceived(
                             Sender: TComponent;
                             BufferSize: Integer;
                             FromIP: string;
                             Port: integer
                             );
 {$ENDIF}
{$ENDIF}
   public
    property RemoteHost: String read GetRemoteHost write SetRemoteHost;
    property RemotePort: Integer read GetRemotePort write SetRemotePort;
    property LocalHost: String read GetLocalHost write SetLocalHost;
    property LocalPort: Integer read GetLocalPort write SetLocalPort;
    property LocalClientID: Integer read FLocalClient write FLocalClient;
  end; // TACRNetwork



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

  TACRBaseConnectionManager = class (TObject)
   private
{$IFDEF CONNECTION_TEST}
    FOtherManager:    TObject;
{$ELSE}
 {$IFDEF NETWORK_TEST}
    FOtherManager:    TObject;
    FClient:          Boolean;
 {$ENDIF}
{$ENDIF}
    FCSect:           TRTLCriticalSection;
    FSessionID:       TACRSessionID;
    FListenerStoped:  Boolean;
   public
    FSessions:        TThreadList;
    FThreadCount:     Integer;
   protected
    procedure IncThreadCount(Ignore: Boolean = False);
    procedure DecThreadCount(Ignore: Boolean = False);
    procedure SetThreadCount(Value: Integer);
   private
    procedure NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); virtual; abstract;
    procedure OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); virtual; abstract;
    procedure CompressAndEncryptBuffer(
                        Session:              TACRBaseSession;
                        InBuffer:             PChar;
                        InBufferSize:         Integer;
                        var OutBuffer:        PChar;
                        var OutBufferSize:    Integer
                                        );
    function DecompressAndDecryptBuffer(
                        Session:              TACRBaseSession;
                        var Buffer:           PChar;
                        var BufferSize:       Integer
                                        ): Boolean;
   public
    property ThreadCount: Integer read FThreadCount write SetThreadCount;
  end; // TACRBaseConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

{$IFDEF CLIENT_VERSION}
  TACRClientConnectionManager = class (TACRBaseConnectionManager)
   private
    FConnections:     TThreadList;
    FConnectionID:    TACRConnectionID;
    FApplication:     String;
    FLiveSessions:    Integer;
    FMaxThreadCount:  Integer;
    FReceiveTimeOut:  Integer;
    procedure PacketResendRequest(
                               Buffer:        PChar;
                               Network:       TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
    procedure SendConnectRequest(ClientSession: PACRClntSession);
    procedure SendDisconnectRequest(ClientSession: PACRClntSession;
                                    WaitForAnswer: Boolean = True);
    procedure SendPing(ClientSession: PACRClntSession);
    procedure SendAcknowledgement(ClientSession: PACRClntSession;
                                  Msg: Boolean = False;
                                  CurrentRequestID:     Integer = -1);
{$IFDEF MsgCommunicator}
    function IsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
    procedure SendConnectAckn(
                          ClientSession:        PACRClntSession;
                          ClientConnection:     PACRClntConnection;
                          CurrentRequestID:     Integer = -1
                              );
{$ENDIF MsgCommunicator}
    procedure DoDisconnect(Session: TACRBaseSession);
    procedure TerminateAllSessionThreads(ClientSession: PACRClntSession);
    procedure TerminateMessageThreads(ClientSession: PACRClntSession);
    procedure TerminateCommandThreads(ClientSession: PACRClntSession);
    procedure DeleteSession(ClientSession: PACRClntSession);
    procedure OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure DoSendBuffer(
                          ClientSession:    PACRClntSession;
                          ClientConnection: PACRClntConnection;
                          Buffer:           PChar;
                          BufferSize:       Integer;
                          Code:             Integer = ACRNewRequest
                         );
    // if encryption algorithm <> acr_Cipher_None then allocate buffer, fill it and return size
    // otherwise return BufferSize = 0
    procedure CreateAuthorizationBuffer(
                          CryptoInfo:     TACRCryptoInfo;
                          out Buffer:     PChar;
                          out BufferSize: Integer
                                        );
   public
    constructor Create;
    destructor Destroy; override;
    procedure TuneConnectionParamaters(ClientSession:  PACRClntSession);
    procedure Connect(Session: TACRBaseSession;
                          ListenOnly: Boolean = False;
                          Tune: Boolean = True
                          );
    procedure Disconnect(Session: TACRBaseSession; ListenOnly: Boolean = False);
    procedure DisconnectAll;
    procedure SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNewRequest
                          );
    procedure ReceiveBuffer(
                          Session:        TACRBaseSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
    procedure SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                          );
    procedure ReceiveMessage(
                          ClientSession:  PACRClntSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
  end; // TACRClientConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientListenerThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientListenerThread = class(TThread)
  private
    FManager:       TACRClientConnectionManager;
    FNetwork:       TACRNetwork;
    Buffer:         PChar;
    BufferSize:     Integer;
    FromHost:       String;
    FromPort:       Integer;
    Sessions:       TList;
    ClientSession:  PACRClntSession;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:       TACRClientConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       Network:       TACRNetwork;
                       FromHost:      String;
                       FromPort:      Integer
                       );
    destructor Destroy; override;
  end;// TACRClientListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientResendRequestThread = class(TThread)
  private
    FManager:           TACRClientConnectionManager;
    FClientSession:     PACRClntSession;
    FBuffer:            PChar;
    FNetwork:           TACRNetwork;
    FHost:              String;
    FPort:              Integer;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                       );
    destructor Destroy; override;
  end;// TACRClientResendRequestThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientMsgResendRequestThread = class(TThread)
  private
    FManager:           TACRClientConnectionManager;
    FClientSession:     PACRClntSession;
    FBuffer:            PChar;
    FNetwork:           TACRNetwork;
    FHost:              String;
    FPort:              Integer;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                       );
    destructor Destroy; override;
  end;// TACRClientMsgResendRequestThread

{$ENDIF} // Client



{$IFDEF SERVER_VERSION} // Server

////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionTerminatorThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerSessionTerminatorThread = class(TThread)
  private
    FManager:                 TACRServerConnectionManager;
    FTerminatedSessions:      TThreadList;
    Error:                    String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                          Manager:          TACRServerConnectionManager
                       );
    destructor Destroy; override;
  public
  end;// TACRServerSessionThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerPingClientsThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerPingClientsThread = class(TThread)
  private
    FManager:                 TACRServerConnectionManager;
    Error:                    String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                          Manager:          TACRServerConnectionManager
                       );
    destructor Destroy; override;
  public
  end;// TACRServerSessionThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerConnectionManager
//
////////////////////////////////////////////////////////////////////////////////


  TACRServerConnectionManager = class (TACRBaseConnectionManager)
   private
    FServer:                TComponent;
    ResendRequestThread:    TACRServerResendRequestThread;
    MsgResendRequestThread: TACRServerMsgResendRequestThread;
    SessionTerminator:      TACRServerSessionTerminatorThread;
    PingClientsThread:      TACRServerPingClientsThread;
   public
    FNetwork:       TACRNetwork;
//    property Network: TACRNetwork read FNetwork write FNetwork;
   private
    function SessionsCount: Integer;
    procedure WaitForThread(
                          Thread:        PThread;
                          TimeOut:       DWORD
                            );
    function TerminateServerThread(
                          Thread:        PThread;
                          ErrProcess:    Integer;
                          ErrObject:     String;
                          WaitTimeOut:   DWORD = ACRServerThreadsTerminateDelay
                                      ): Boolean;
    procedure TerminateAllSessionThreads(ServerSession: PACRSrvrSession);
    procedure TerminateMessageThreads(ServerSession: PACRSrvrSession);
    procedure TerminateCommandThreads(ServerSession: PACRSrvrSession);
    procedure EnableNextCommand(ServerSession: PACRSrvrSession);
    procedure PacketResendRequest(
                               Buffer:      PChar;
                               Network:     TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
    procedure DeleteSession(ServerSession: PACRSrvrSession);
    procedure DoDisconnect(Session: TACRBaseSession);
    procedure DoSendBuffer(
                          ServerSession:    PACRSrvrSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNoAction
                                                  );
    procedure SendDisconnectRequest(ServerSession: PACRSrvrSession;
                                    WaitForAnswer: Boolean = True);
    procedure SendConnectAckn(
                          ServerSession:        PACRSrvrSession;
                          CurrentRequestID:     Integer = -1
                              );
    procedure SendPing(ServerSession: PACRSrvrSession);
    procedure SendAcknowledgement(
                          ServerSession:        PACRSrvrSession;
                          Msg:                  Boolean = False;
                          CurrentRequestID:     Integer = -1
                              );
    procedure OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure WaitForServerSessionThread(
                 ServerSession:    PACRSrvrSession;
                 TimeOut:          DWORD = ACRWaitForServerSessionThreadTimeOut
                                          );
    procedure WaitForServerSessionMsgThread(
                 ServerSession:    PACRSrvrSession;
                 TimeOut:          DWORD = ACRWaitForServerSessionThreadTimeOut
                                          );
    procedure CommandReceived(
                          ServerSession:    PACRSrvrSession;
                          ControlCode:      TACRControlCode;
                          CurrentRequestID: Integer
                              );
    procedure MessageReceived(
                               ServerSession:  PACRSrvrSession;
                               ControlCode:    TACRControlCode = ACRMessage
                              );
    // if authorization buffer is valid (encrypted by the same crypto settings as in CryptoInfo)
    // then return true else return false
    function IsAuthorizationBufferValid(
                          CryptoInfo: TACRCryptoInfo;
                          Buffer:     PChar;
                          BufferSize: Integer
                                        ): Boolean;
   public
    constructor Create(Server: TComponent);
    destructor Destroy; override;
    procedure Disconnect(Session: TACRBaseSession);
    procedure DisconnectAll(WaitForAllDisconected: Boolean = True);
    procedure SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNoAction
                          );
    procedure ReceiveBuffer(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
    procedure SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                          );
    procedure ReceiveMessage(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
    function GetClientInfo(
                          Session:          TACRBaseSession;
                          var Host:         String;
                          var Port:         Integer;
                          var Application:  String
                            ): Boolean;
    // fill array with server session object connected to this server
    procedure GetClientsList(var Clients: TACRSessionsArray);
    procedure TerminateSession(Session: TACRBaseSession);
  end; // TACRClientConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerSessionThread = class(TThread)
  private
    FServerSession:           PACRSrvrSession;
    FManager:                 TACRServerConnectionManager;
    FCode:                    TACRControlCode;
    Error:                    String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          Code:             Integer = ACRNoAction
                       );
    destructor Destroy; override;
  public
  end;// TACRServerSessionThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionMsgThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerSessionMsgThread = class(TThread)
  private
    FServerSession:           PACRSrvrSession;
    FManager:                 TACRServerConnectionManager;
    FCode:                    TACRControlCode;
    Error:                    String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          Code:             Integer = ACRNoAction
                       );
    destructor Destroy; override;
  public
  end;// TACRServerSessionMsgThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionDisconnectThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerSessionDisconnectThread = class(TThread)
  private
    FServerSession:           PACRSrvrSession;
    FManager:                 TACRServerConnectionManager;
    FCurrentRequestID:        Integer;
    Error:                    String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          CurrentRequestID: Integer = -1
                       );
    destructor Destroy; override;
  public
  end;// TACRServerSessionDisconnectThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerListenerThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerListenerThread = class(TThread)
  private
    FManager:       TACRServerConnectionManager;
    Buffer:         PChar;
    BufferSize:     Integer;
    FromHost:       String;
    FromPort:       Integer;
    ServerSession:  PACRSrvrSession;
    SessionFound:   Boolean;
    Error:          String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:       TACRServerConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       FromHost:      String;
                       FromPort:      Integer
                       );
    destructor Destroy; override;
  end;// TACRServerListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerResendRequestThread = class(TThread)
  private
    FManager:           TACRServerConnectionManager;
    Error:              String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRServerConnectionManager
                       );
    destructor Destroy; override;
  end;// TACRServerResendRequestThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerMsgResendRequestThread = class(TThread)
  private
    FManager:           TACRServerConnectionManager;
    Error:              String;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRServerConnectionManager
                       );
    destructor Destroy; override;
  end;// TACRServerMsgResendRequestThread

{$ENDIF} // Server



////////////////////////////////////////////////////////////////////////////////
//
// Procedures and Functions
//
////////////////////////////////////////////////////////////////////////////////

  function CheckSum(Buffer: PChar; BufferSize: Integer): Integer;

{$IFNDEF MsgCommunicator}
{IFDEF SERVER_VERSION}
{$ENDIF MsgCommunicator}
  function fnIsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
{$IFNDEF MsgCommunicator}
{ENDIF SERVER_VERSION}
{$ENDIF MsgCommunicator}


implementation

uses
{$IFDEF SERVER_VERSION}
  ACRServer,
{$ENDIF}
{IFDEF MsgCommunicator}
 {$IFDEF CLIENT_VERSION}
  ACRClient,
 {$ENDIF}
{ENDIF}
{$IFNDEF MsgCommunicator}
  ACRCommunication,
{$ENDIF}
  ACRMain;

////////////////////////////////////////////////////////////////////////////////
// CheckSum
////////////////////////////////////////////////////////////////////////////////
function CheckSum(Buffer: PChar; BufferSize: Integer): Integer;
begin
  // Start Point = Pointer(Buffer)+4
  // Count (in Bytes) = BufferSize-4
//  Result := ACRCountCRC(0, PChar(Buffer+4), BufferSize-4)
  Result := 0;
end;



////////////////////////////////////////////////////////////////////////////////
//
// TACRNetwork
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// CreateNetwork
//------------------------------------------------------------------------------
procedure TACRNetwork.CreateNetwork;
begin
{$IFDEF API_NETWORK}
  FACRNetwork := TACRapiNetwork.Create;
  FACRNetwork.OnDataReceived := OnDataReceived;
  FACRNetwork.OnDisconnect := OnDisconnect;
{$ELSE}
 {$IFDEF D6H}
//  FNetworkClient := TIdUDPClient.Create(nil);
  FNetworkServer := TIdUDPServer.Create(nil);
  FNetworkServer.OnUDPRead := OnDataReceived;
 {$ELSE}
  FTransport := TNMUDP.Create(nil);
  FTransport.OnDataReceived := OnDataReceived;
 {$ENDIF D6H}
{$ENDIF API_NETWORK}
end; // CreateNetwork


//------------------------------------------------------------------------------
// FreeNetwork
//------------------------------------------------------------------------------
procedure TACRNetwork.FreeNetwork;
begin
{$IFDEF API_NETWORK}
  if FACRNetwork <> nil then
    FACRNetwork.Free;
  FACRNetwork := nil;
{$ELSE}
 {$IFDEF D6H}
//  FNetworkClient.Free;
  FNetworkServer.Free;
 {$ELSE}
  FTransport.Free;
 {$ENDIF D6H}
{$ENDIF}
end; // FreeNetwork


//------------------------------------------------------------------------------
// ReCreateNetwork
//------------------------------------------------------------------------------
procedure TACRNetwork.ReCreateNetwork;
var
  lRemoteHost:       String;
  lLocalHost:        String;
  lRemotePort:       Integer;
  lLocalPort:        Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> START');
{$ENDIF}
  EnterCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> CS entered');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> RemoteHost="'+RemoteHost+'"');
{$ENDIF}
  lRemoteHost := RemoteHost;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> RemotePort='+IntToStr(RemotePort));
{$ENDIF}
  lRemotePort := RemotePort;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> LocalHost="'+LocalHost+'"');
{$ENDIF}
  lLocalHost := LocalHost;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> LocalPort='+IntToStr(LocalPort));
{$ENDIF}
  lLocalPort := LocalPort;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> Free...');
{$ENDIF}
  FreeNetwork;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> Create...');
{$ENDIF}
  CreateNetwork;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> Set...');
{$ENDIF}
  LocalPort := lLocalPort;
  RemotePort := lRemotePort;
  RemoteHost := lRemoteHost;
  LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.ReCreateNetwork> FINISH');
{$ENDIF}
end; // ReCreateNetwork


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRNetwork.Create(ConnectionManager: TACRBaseConnectionManager);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.Create');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

  inherited Create;
  FManager := ConnectionManager;
  FLocalClient := RndClientID;
  CreateNetwork;

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRNetwork.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.Destroy');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

  FreeNetwork;
  inherited Destroy;

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;// Destoy


//------------------------------------------------------------------------------
// RndClientID
//------------------------------------------------------------------------------
function TACRNetwork.RndClientID: TACRNetworkClientID;
label generate;
begin
  Randomize;
generate:
  Result := Random(MAXINT); // - (MAXINT div 2);
  if (Result >= 0) and (Result < 1024) then // Reserved for ServerID
{ TODO -oAll: Add to DevGuide:
ServerID must be unique for each server working at the one machine.
We recomend to set ServerID from the region 0...1023.
}
   goto generate;
end;// RndClientID


//------------------------------------------------------------------------------
// SetRemoteHost
//------------------------------------------------------------------------------
procedure TACRNetwork.SetRemoteHost(Host: String);
begin
{$IFDEF API_NETWORK}
  FACRNetwork.RemoteHost := Host;
{$ELSE}
 {$IFDEF D6H}
  FRemoteHost := Host;
//  FNetworkClient.Host := Host;
 {$ELSE}
  FTransport.RemoteHost := Host;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetRemoteHost
//------------------------------------------------------------------------------
function TACRNetwork.GetRemoteHost: String;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.RemoteHost;
{$ELSE}
 {$IFDEF D6H}
  Result := FRemoteHost;
//  Result := FNetworkClient.Host;
 {$ELSE}
  Result := FTransport.RemoteHost;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SetRemotePort
//------------------------------------------------------------------------------
procedure TACRNetwork.SetRemotePort(Port: Integer);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SetRemotePort');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

{$IFDEF API_NETWORK}
  FACRNetwork.RemotePort := Port;
{$ELSE}
 {$IFDEF D6H}
  FRemotePort := Port;
//  FNetworkClient.Port := Port;
 {$ELSE}
  FTransport.RemotePort := Port;
 {$ENDIF}
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetRemotePort
//------------------------------------------------------------------------------
function TACRNetwork.GetRemotePort: Integer;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.RemotePort;
{$ELSE}
 {$IFDEF D6H}
  Result := FRemotePort;
//  Result := FNetworkClient.Port;
 {$ELSE}
  Result := FTransport.RemotePort;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SetLocalPort
//------------------------------------------------------------------------------
procedure TACRNetwork.SetLocalPort(Port: Integer);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SetLocalPort');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

{$IFDEF API_NETWORK}
  FACRNetwork.LocalPort := Port;
{$ELSE}
 {$IFDEF D6H}
  FNetworkServer.Active := False;
  FNetworkServer.DefaultPort := Port;
  FNetworkServer.Active := True;
 {$ELSE}
  FTransport.LocalPort := Port;
 {$ENDIF}
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SetLocalHost
//------------------------------------------------------------------------------
procedure TACRNetwork.SetLocalHost(Host: String);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SetLocalHost');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

{$IFDEF API_NETWORK}
  FACRNetwork.LocalHost := Host;
{$ELSE}
  FLocalHost := Host;
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetLocalHost
//------------------------------------------------------------------------------
function TACRNetwork.GetLocalHost: String;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.LocalHost;
{$ELSE}
  Result := FLocalHost;
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetLocalPort
//------------------------------------------------------------------------------
function TACRNetwork.GetLocalPort: Integer;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.LocalPort;
{$ELSE}
 {$IFDEF D6H}
  Result := FNetworkServer.DefaultPort;
 {$ELSE}
  Result := FTransport.LocalPort;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRNetwork.SendBuffer(
                          Buffer: PChar;
                          Count:  Integer);
var
  Header: PACRPacketHeader;
{$IFNDEF D6H}
  Buff: array [0..ACRMaxPacketSize] of Char;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
  buf: PChar;
  i: Integer;
{$ENDIF}
{$IFDEF NETWORK_TEST}
  Connections: TList;
  Connection:  PACRClntConnection;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
  str: String;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SendBuffer');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

  Header := PACRPacketHeader(Buffer);
  Header.Sender := FLocalClient;
  Header.CheckSum := CheckSum(Buffer, Count);
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
if LocalPort = ACRDefaultClientPort then
str := 'Client'
else
if LocalPort = ACRDefaultServerPort then
str := 'Server'
else
str := IntToStr(LocalPort);
str := str + '>>> '
+IntToStr(Header.CurrentRequestID)+' : '
+IntToStr(Header.PacketID)+' / '
+IntToStr(Header.ControlCode)
+' >>> '+RemoteHost+':'+IntToStr(RemotePort)
+' - '+IntToStr(GetTickCount)+' msec'
//+' <'+IntToStr(Header.Recepient)+'>'
;
aaWriteToLog(str);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------');
aaWriteToLog('ACRConnection>TACRNetwork.SendBuffer('+IntToHex(Integer(@Buffer),6)+', '+IntToStr(Count)+')');
aaWriteToLog('Sender    : '+IntToStr(FLocalClient));
aaWriteToLog('-----------');
aaWriteToLog('RemoteHost: '+RemoteHost);
aaWriteToLog('RemotePort: '+IntToStr(RemotePort));
aaWriteToLog('LocalHost : '+LocalHost);
aaWriteToLog('LocalPort : '+IntToStr(LocalPort));
aaWriteToLog('>>> Head:');
aaWriteToLog('CheckSum         '+IntToStr(Header.CheckSum));
aaWriteToLog('Signature        '+Header.Signature);
aaWriteToLog('Recepient        '+IntToStr(Header.Recepient));
aaWriteToLog('Sender           '+IntToStr(Header.Sender));
aaWriteToLog('ConnectionID     '+IntToStr(Header.ConnectionID));
aaWriteToLog('SessionID        '+IntToStr(Header.SessionID));
aaWriteToLog('CurrentRequestID '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('PacketID         '+IntToStr(Header.PacketID));
aaWriteToLog('ControlCode      '+IntToStr(Header.ControlCode));
(*
aaWriteToLog('>>> Data:');
buf:=MemoryManager.GetMem(Count-SizeOf(TACRPacketHeader)+1);
Move(PChar(Integer(Buffer)+SizeOf(TACRPacketHeader))^, buf^, Count-SizeOf(TACRPacketHeader));
for i:= 0 to Count-SizeOf(TACRPacketHeader)-1 do
  if (buf+i)^=#0 then
    (buf+i)^:='0';
(buf+Count-SizeOf(TACRPacketHeader))^:=#0;
aaWriteToLog(buf);
aaWriteToLog('###');
MemoryManager.FreeAndNilMem(buf);
*)
{$ENDIF}

{******************************************************************************}
// Send Buffer
{******************************************************************************}
//  EnterCriticalSection(FCSect);
{$IFDEF NETWORK_TEST}
  if FManager.FClient then
   if Assigned(TACRServerConnectionManager(FManager.FOtherManager).FNetwork.OnDataReceived) then
    TACRServerConnectionManager(FManager.FOtherManager).FNetwork.OnDataReceived(Buffer, Count, LocalHost, LocalPort)
  else
   begin
    Connections:=TACRClientConnectionManager(FManager.FOtherManager).FConnections.LockList;
    try
     Connection := Connections.Items[0];
    finally
     TACRClientConnectionManager(FManager.FOtherManager).FConnections.UnlockList;
    end;
    Connection.Network.OnDataReceived(Buffer, Count, LocalHost, LocalPort);
   end;
{$ELSE}
 {$IFDEF API_NETWORK}
  FACRNetwork.SendBuffer(Buffer, Count);
//  sleep(0); // TRY TO REMOVE!!!
 {$ELSE}
  {$IFDEF D6H}
  FNetworkServer.SendBuffer(FRemoteHost, FRemotePort, Buffer^, Count);
//  FNetworkClient.SendBuffer(Buffer^, Count);
  {$ELSE}
  Move(Buffer^, Buff[0], Count);
  FTransport.SendBuffer(Buff, Count);
  {$ENDIF D6H}
 {$ENDIF API_NETWORK}
{$ENDIF NETWORK_TEST}
//  LeaveCriticalSection(FCSect);
{******************************************************************************}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end; // SendBuffer


//------------------------------------------------------------------------------
// OnDisconnect
//------------------------------------------------------------------------------
procedure TACRNetwork.OnDisconnect(
                             FromHost:  String;
                             FromPort:  Integer
                             );
begin
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect> START');
{$ENDIF}
 try
  if (FManager = nil)
  or FManager.FListenerStoped
  then
   begin
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect> Manager not exist - FINISH');
{$ENDIF}
    Exit;
   end;
  ReCreateNetwork;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect> FManager.OnDisconnect...');
{$ENDIF}
  FManager.OnDisconnect(self, FromHost, FromPort);
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect> FINISH');
{$ENDIF}
end;// OnDisconnect


//------------------------------------------------------------------------------
// OnDataReceived
//------------------------------------------------------------------------------
{$IFDEF API_NETWORK}
procedure TACRNetwork.OnDataReceived(
                             Buffer:    PChar;
                             Count:     Integer;
                             FromHost:  String;
                             FromPort:  Integer
                             );
{$ELSE}
 {$IFDEF D6H}
procedure TACRNetwork.OnDataReceived(
                             Sender:   TObject;
                             AData:    TStream;
                             ABinding: TIdSocketHandle
                             );
 {$ELSE}
procedure TACRNetwork.OnDataReceived(
                             Sender: TComponent;
                             BufferSize: Integer;
                             FromIP: string;
                             Port: integer
                             );
 {$ENDIF D6H}
{$ENDIF API_NETWORK}
var
  Buf: PChar;
  BufSize: Integer;
{$IFDEF API_NETWORK}
{$ELSE}
 {$IFDEF D6H}
  FromHost: String;
  FromPort: Integer;
 {$ELSE}
  Buffer: array [0..ACRMaxPacketSize] of Char;
 {$ENDIF}
{$ENDIF}
  Header: PACRPacketHeader;
{$IFDEF DEBUG_LOG_NETWORK}
  buff: pchar;
  i: integer;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
  str: String;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.OnDataReceived');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

//  EnterCriticalSection(FCSect);
{$IFDEF API_NETWORK}
  Buf := MemoryManager.GetMem(Count);
  Move(Buffer^, Buf^, Count);
  BufSize := Count;
{$ELSE}
 {$IFDEF D6H}
  Buf := MemoryManager.GetMem(ACRMaxPacketSize);
  BufSize := AData.Read(Buf^, ACRMaxPacketSize);
  FromHost := ABinding.PeerIP;
  FromPort := ABinding.PeerPort;
 {$ELSE}
  Buf := MemoryManager.GetMem(BufferSize);
  FTransport.ReadBuffer(Buffer, BufferSize);
  Move(Buffer[0], Buf^, BufferSize);
  BufSize := BufferSize;
 {$ENDIF D6H}
{$ENDIF API_NETWORK}
//  LeaveCriticalSection(FCSect);
  Header := PACRPacketHeader(Buf);
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
if LocalPort = ACRDefaultClientPort then
str := 'Client'
else
if LocalPort = ACRDefaultServerPort then
str := 'Server'
else
str := IntToStr(LocalPort);
str := str + '<<< '
+IntToStr(Header.CurrentRequestID)+' : '
+IntToStr(Header.PacketID)+' / '
+IntToStr(Header.ControlCode)
+' <<< '+FromHost+':'+IntToStr(FromPort)
+' - '+IntToStr(GetTickCount)+' msec'
//+' <'+IntToStr(Header.Recepient)+'>'
;
aaWriteToLog(str);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------');
aaWriteToLog('ACRConnection>TACRNetwork.OnDataReceived('+IntToHex(Integer(@Buf),6)+', '+IntToStr(BufSize)+')');
aaWriteToLog('Recepient : '+IntToStr(FLocalClient));
aaWriteToLog('-----------');
 {$IFDEF API_NETWORK}
aaWriteToLog('RemoteHost: '+FromHost);
aaWriteToLog('RemotePort: '+IntToStr(FromPort));
 {$ELSE}
  {$IFDEF D6H}
aaWriteToLog('RemoteHost: '+FromHost);
aaWriteToLog('RemotePort: '+IntToStr(FromPort));
  {$ELSE}
aaWriteToLog('RemoteHost: '+FromIP);
aaWriteToLog('RemotePort: '+IntToStr(Port));
  {$ENDIF D6H}
 {$ENDIF API_NETWORK}
aaWriteToLog('LocalHost : '+LocalHost);
aaWriteToLog('LocalPort : '+IntToStr(LocalPort));
aaWriteToLog('>>> Head:');
aaWriteToLog('CheckSum         '+IntToStr(Header.CheckSum));
aaWriteToLog('Signature        '+Header.Signature);
aaWriteToLog('Recepient        '+IntToStr(Header.Recepient));
aaWriteToLog('Sender           '+IntToStr(Header.Sender));
aaWriteToLog('ConnectionID     '+IntToStr(Header.ConnectionID));
aaWriteToLog('SessionID        '+IntToStr(Header.SessionID));
aaWriteToLog('CurrentRequestID '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('PacketID         '+IntToStr(Header.PacketID));
aaWriteToLog('ControlCode      '+IntToStr(Header.ControlCode));
(*
aaWriteToLog('>>> Data:');
buff:=MemoryManager.GetMem(BufSize-SizeOf(TACRPacketHeader)+1);
Move(PChar(Integer(Buf)+SizeOf(TACRPacketHeader))^, buff^, BufSize-SizeOf(TACRPacketHeader));
for i:= 0 to BufSize-SizeOf(TACRPacketHeader)-1 do
  if (buff+i)^=#0 then
    (buff+i)^:='0';
(buff+BufSize-SizeOf(TACRPacketHeader))^:=#0;
aaWriteToLog(buff);
aaWriteToLog('###');
MemoryManager.FreeAndNilMem(buff);
*)
{$ENDIF DEBUG_LOG_NETWORK}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount)+ ' - buffer prepared');
{$ENDIF}
  if (Header.Recepient = FLocalClient)
{$IFDEF MsgCommunicator}
  or (
//      (Header.Recepient = Integer(MSG_INVALID_USER_ID)) and
      ((Header.ControlCode = ACRConnect) or (Header.ControlCode = ACRConnect+ACRLastPacket))
      and
      (Header.Signature = ACRClientPacketSign)
      )
{$ENDIF MsgCommunicator}
  then
{$IFDEF API_NETWORK}
    FManager.NetworkListener(Buf, BufSize, Self, FromHost, FromPort);
{$ELSE}
 {$IFDEF D6H}
    FManager.NetworkListener(Buf, BufSize, Self, FromHost, FromPort);
 {$ELSE}
    FManager.NetworkListener(Buf, BufSize, Self, FromIP, Port);
 {$ENDIF D6H}

{$ENDIF API_NETWORK}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount)+' - end of parsing');
aaWriteToLog('');
{$ENDIF}
end;
// TACRNetwork



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// TACRBaseConnectionManager.IncThreadCount
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.IncThreadCount(Ignore: Boolean);
begin
  if not Ignore then
    EnterCriticalSection(FCSect);
  FThreadCount := FThreadCount+1;
  if not Ignore then
    LeaveCriticalSection(FCSect);
end;// TACRBaseConnectionManager.IncThreadCount

//------------------------------------------------------------------------------
// TACRBaseConnectionManager.DecThreadCount
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.DecThreadCount(Ignore: Boolean);
begin
  if not Ignore then
    EnterCriticalSection(FCSect);
  FThreadCount := FThreadCount-1;
  if not Ignore then
    LeaveCriticalSection(FCSect);
end;// TACRBaseConnectionManager.DecThreadCount

//------------------------------------------------------------------------------
// TACRBaseConnectionManager.SetThreadCount
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.SetThreadCount(Value: Integer);
begin
  EnterCriticalSection(FCSect);
  FThreadCount := Value;
  LeaveCriticalSection(FCSect);
end;// TACRBaseConnectionManager.SetThreadCount

//------------------------------------------------------------------------------
// compresses and encrypts buffer
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.CompressAndEncryptBuffer(
                        Session:              TACRBaseSession;
                        InBuffer:             PChar;
                        InBufferSize:         Integer;
                        var OutBuffer:        PChar;
                        var OutBufferSize:    Integer
                                                              );
begin
 try
  ACRBaseEngine.CompressAndEncryptBuffer(Session.ConnectParams.CryptoInfo,
                                         Session.ConnectParams.CompressionAlgorithm,
                                         Session.ConnectParams.CompressionMode,
                                         InBuffer,  InBufferSize,
                                         OutBuffer, OutBufferSize);
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> CompressAndEncryptBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
end; // CompressAndEncryptBuffer


//------------------------------------------------------------------------------
// decompresses and decrypts buffer
//------------------------------------------------------------------------------
function TACRBaseConnectionManager.DecompressAndDecryptBuffer(
                        Session:              TACRBaseSession;
                        var Buffer:           PChar;
                        var BufferSize:       Integer
                                                              ): Boolean;
begin
 try
  Result := ACRBaseEngine.DecompressAndDecryptBuffer(
                                Session.ConnectParams.CryptoInfo,
                                Session.ConnectParams.CompressionAlgorithm,
                                Session.ConnectParams.CompressionMode,
                                Buffer,
                                BufferSize);
 except
  on E: Exception do
   begin
    Result := False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> DecompressAndDecryptBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
   end;
 end;
end; // DecompressAndDecryptBuffer



{$IFDEF CLIENT_VERSION}

////////////////////////////////////////////////////////////////////////////////
//
// TACRClientConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientConnectionManager.Create;
begin
  InitializeCriticalSection(FCSect);
  inherited Create;
  FListenerStoped := False;
  FSessions := TThreadList.Create;
  FConnections := TThreadList.Create;
  FSessionID := MAXINT;
  FConnectionID := -MAXINT;
  FLiveSessions := 0;
  FApplication := '';
  FApplication := ParamStr(0);
  if FApplication = '' then
    FApplication := 'UNKNOWN';
  ThreadCount := 0;
{$IFDEF CONNECTION_TEST}
//  FOtherManager :=
{$ELSE}
 {$IFDEF NETWORK_TEST}
//  FOtherManager :=
  FClient := True;
 {$ENDIF}
{$ENDIF}
  FMaxThreadCount := ACRMaxThreadCount;
  FReceiveTimeOut := ACRReceiveTimeOut; 
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRClientConnectionManager.Create> Application: '+FApplication);
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientConnectionManager.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.Destroy> Application: '+FApplication);
{$ENDIF}
  DisconnectAll;
  FListenerStoped := True;
  EnterCriticalSection(FCSect);
  FConnections.Free;
  FSessions.Free;
  FSessions := nil;
  LeaveCriticalSection(FCSect);
{ TODO -oAlex : Write ThreadCount to ErrorLog }
  DeleteCriticalSection(FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.Destroy - FINISHED');
{$ENDIF}
end;// Destroy


//------------------------------------------------------------------------------
// PacketResendRequest
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.PacketResendRequest(
                               Buffer:        PChar;
                               Network:       TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
var
  Header:          PACRPacketHeader;
begin
  Header := Pointer(Buffer);
  if Msg then
    Header.ControlCode := ACRMessagePacketResendRequest
  else
    Header.ControlCode := ACRPacketResendRequest;
  Header.Signature := ACRClientPacketSign;
  if (PacketID >= 0) then
    Header.PacketID := PacketID
  else
    Header.Recepient := Header.Sender;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('TACRClientConnectionManager.PacketResendRequest> asks to resend packet # '+IntToStr(Header.PacketID));
{$ENDIF}
  EnterCriticalSection(FCSect);
  try
   Network.RemoteHost := RemoteHost;
   Network.RemotePort := RemotePort;
   Network.SendBuffer(Buffer, SizeOf(TACRPacketHeader));
  finally
   LeaveCriticalSection(FCSect);
  end;
end;// PacketResendRequest


//------------------------------------------------------------------------------
// OnDisconnect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
var
  Sessions:             TList;
  Session:              TACRBaseSession;
  ClientSession:        PACRClntSession;
//  Connections:          TList;
//  ClientConnection:     PACRClntConnection;
  i, j:                 Integer;
//  Found:                Boolean;
begin
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - START');
{$ENDIF}
{
  Connections:=FConnections.LockList;
  try
   j := Connections.Count-1;
   for i:=0 to j do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.Network = FNetwork then
      begin
       Found := True;
       break;
      end;
    end;
  finally
   FConnections.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(FNetwork)]);
  Found := False;
}
  Sessions:=FSessions.LockList;
  try
   j := Sessions.Count-1;
   for i:=j downto 0 do
    begin
     ClientSession := Sessions.Items[i];
//     if ClientSession.ConnectionID = ClientConnection.ConnectionID then
     if  (FromHost = ClientSession.Session.ConnectParams.RemoteHost)
     and (FromPort = ClientSession.Session.ConnectParams.RemotePort)
     then
      begin
//       Found := True;
       Session := ClientSession.Session;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - DoDisconnect');
{$ENDIF}
       FSessions.UnlockList;
       DoDisconnect(Session);
       Sessions := FSessions.LockList;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - Session.OnDisconnect');
{$ENDIF}
       if Session<>nil then
         TACRClientSession(Session).DoCloseSessionOnNetworkError;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - Session.OnDisconnect FINISHED');
{$ENDIF}
{$IFDEF MsgCommunicator}
       if Session<>TACRClientSession(Session).FClient.FDefaultSession then
        if Session<>TACRClientSession(Session).FClient.FDefaultServerSession then
         Session.Free;
 {$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - Session Freed!');
 {$ENDIF}
{$ENDIF}
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
{
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(ClientConnection.ConnectionID)]);
}
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - FINISH');
{$ENDIF}
end; // OnDisconnect


//------------------------------------------------------------------------------
// NetworkListener
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
var
  StartTime:        DWORD;
  Error:            String;
  Sessions:         TList;
  ClientSession:    PACRClntSession;
begin
  if not FListenerStoped then
   begin
    StartTime := GetTickCount;
    while ((GetTickCount-StartTime) < FReceiveTimeOut) do
     begin
      if ThreadCount<FMaxThreadCount then
       begin
        TACRClientListenerThread.Create(self, Buffer, BufferSize, Network, FromHost, FromPort);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRClientConnectionManager.NetworkListener> Listener Thread Created ! - '+IntToStr(GetTickCount));
{$ENDIF}
        Exit;
       end;
      sleep(0);
     end;
     Error :=
                  ErrorRCannotCreateListenerThread+
                  IntToStr(FMaxThreadCount);
     Sessions := FSessions.LockList;
     try
       if Sessions.Count > 0 then
        begin
         ClientSession := Sessions.Items[0];
         if ClientSession.Session <> nil then
           TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrServerListenerThread,40516,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FSessions.UnlockList;
      end;
   end;
  MemoryManager.FreeAndNilMem(Buffer);
end;// NetworkListener


//------------------------------------------------------------------------------
// ReceiveMessage
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.ReceiveMessage(
                          ClientSession:        PACRClntSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
var
  i:                    Integer;
  Packet:               PACRPacket;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT has started to receive message');
{$ENDIF}
  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);
  if ClientSession.MsgReceiveStatus<>ACRFull then
   begin
    while (ClientSession.MsgReceiveStatus<>ACRStart)
    and (ClientSession.MsgReceiveStatus<>ACRFull)
    do // Wait for starting answer receive
     begin
      if ClientSession.MsgControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ClientSession.Session.ConnectParams.ReceiveSleep);
     end;
    StartTime := GetTickCount;
    while ClientSession.MsgReceiveStatus<>ACRFull do // Wait for all packets to arrive
     begin
      if ClientSession.MsgControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
      if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.ReceiveTimeOut then
        raise EACRException.Create(40026, ErrorRTimeoutFullReceive,
                                [ClientSession.ServerSessionID,
                                ClientSession.Session.ConnectParams.ReceiveTimeOut]);
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ClientSession.Session.ConnectParams.ReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT are receiving message from SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
//  TerminateMessageThreads(ClientSession);
  EnterCriticalSection(FCSect);
  BufferSize := 0;
  for i := 0 to ClientSession.MsgPackets.Count - 1 do
   begin
    Packet := ClientSession.MsgPackets.Items[i];
    BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
   end;
  Buffer := MemoryManager.GetMem(BufferSize);
  pBuf := Buffer;
  for i := 0 to ClientSession.MsgPackets.Count - 1 do
   begin
    Packet := ClientSession.MsgPackets.Items[i];
    if Packet <> nil then
     begin
      if Packet.Buffer <> nil then
       begin
        Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
        inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
        MemoryManager.FreeAndNilMem(Packet.Buffer);
       end;
      Dispose(Packet);
      ClientSession.MsgPackets.Items[i] := nil;
     end;
   end;
  ClientSession.MsgPackets.Count := 0;
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('CLIENT HAS RECEIVED Message FROM SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
  inc(ClientSession.ServerMessageID);
  ClientSession.MsgReceiveStatus := ACRNo;
  EnterCriticalSection(FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FCSect);
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientConnectionManager.ReceiveMessage - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end; // except
end; // ReceiveMessage


//------------------------------------------------------------------------------
// ReceiveBuffer
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.ReceiveBuffer(
                          Session:        TACRBaseSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
label
  Loop;
var
  Sessions:             TList;
  ClientSession:        PACRClntSession;
  SessionFound:         Boolean;
  i:                    Integer;
  Packet:               PACRPacket;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> STARTED - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
 try

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT has started to receive answer');
{$ENDIF}
  SessionFound := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if (ClientSession.Session = Session)
     then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);

  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);

  StartTime := GetTickCount;
  if ClientSession.AnswerStatus<>ACRFull then
   begin
    while (ClientSession.AnswerStatus<>ACRStart)
    and (ClientSession.AnswerStatus<>ACRFull)
    do // Wait for starting answer receive
     begin
      if ClientSession.ControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
      if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.StartReceiveTimeOut then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40025, ErrorRTimeoutFullReceive,
                                [ClientSession.ServerSessionID,
                                ClientSession.Session.ConnectParams.ReceiveTimeOut]);
       end;
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ClientSession.Session.ConnectParams.ReceiveSleep);
     end;

    StartTime := GetTickCount;
    while ClientSession.AnswerStatus<>ACRFull do // Wait for all packets to arrive
     begin
      if ClientSession.ControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
      if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.ReceiveTimeOut then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40025, ErrorRTimeoutFullReceive,
                            [ClientSession.ServerSessionID,
                            ClientSession.Session.ConnectParams.ReceiveTimeOut]);
       end;
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ClientSession.Session.ConnectParams.ReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received all packets and starts to extract buffer from SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}

  TerminateCommandThreads(ClientSession);

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CommandThreads Terminated');
{$ENDIF}
  EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CS Entered');
{$ENDIF}
  BufferSize := 0;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packets.Count = '+IntToStr(ClientSession.Packets.Count));
{$ENDIF}
  for i := 0 to ClientSession.Packets.Count - 1 do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packet # '+IntToStr(i));
{$ENDIF}
    Packet := ClientSession.Packets.Items[i];
    BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packet.BufferSize = '+IntToStr(Packet.BufferSize));
{$ENDIF}
   end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> BufferSize = '+IntToStr(BufferSize));
{$ENDIF}
   if BufferSize > 0 then
     Buffer := MemoryManager.GetMem(BufferSize)
   else
     Buffer := nil;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Buffer memory got');
{$ENDIF}
  pBuf := Buffer;
  for i := 0 to ClientSession.Packets.Count - 1 do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packets.Count = '+IntToStr(ClientSession.Packets.Count));
{$ENDIF}
    Packet := ClientSession.Packets.Items[i];
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packet # '+IntToStr(i));
{$ENDIF}
    if Packet <> nil then
     begin
      if Packet.Buffer <> nil then
       begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Move...');
{$ENDIF}
        if (Packet.BufferSize > SizeOf(TACRPacketHeader)) then
         begin
          Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> inc...');
{$ENDIF}
          inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
         end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> inced'+IntToStr(Packet.BufferSize-SizeOf(TACRPacketHeader)));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> free packet buffer...');
{$ENDIF}
        MemoryManager.FreeAndNilMem(Packet.Buffer);
       end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Dispose packet...');
{$ENDIF}
      Dispose(Packet);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> packet = nil...');
{$ENDIF}
      ClientSession.Packets.Items[i] := nil;
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> ClientSession.Packets.Count := 0 ...');
{$ENDIF}
  ClientSession.Packets.Count := 0;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CS leave...');
{$ENDIF}
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CS left');
{$ENDIF}
  DecompressAndDecryptBuffer(ClientSession.Session, Buffer, BufferSize);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> DecompressAndDecryptBuffer - OK');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('CLIENT HAS RECEIVED BUFFER FROM SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
  ClientSession.AnswerStatus := ACRNo;

  EnterCriticalSection(FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FCSect);

 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientConnectionManager.ReceiveBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end; // except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT has finished answer receiving');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('CLIENT<<< '+IntToStr(ClientSession.CurrentRequestID)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}
end; // ReceiveBuffer


//------------------------------------------------------------------------------
// SendConnectRequest
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendConnectRequest(ClientSession: PACRClntSession);
var
  Buf, Buffer:                    PChar;
  BufSize, BufferSize,
  SizeApp, SizeParams:            Integer;
  StartTime:                      DWORD;
  Retry:                          Integer;
  ConnectionParams:               TACRConnectionParams;
  ServerID:                       Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT SendConnectRequest started');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);
  ClientSession.AnswerStatus := ACRNo;
  ClientSession.CurrentRequestID := 0;
  ConnectionParams.PacketSize := ClientSession.Session.ConnectParams.PacketSize;
  ConnectionParams.CompressionAlgorithm := ClientSession.Session.ConnectParams.CompressionAlgorithm;
  ConnectionParams.CompressionMode := ClientSession.Session.ConnectParams.CompressionMode;
  CreateAuthorizationBuffer(ClientSession.Session.ConnectParams.CryptoInfo, Buffer, BufferSize);
//  IsAuthorizationBufferValid(ClientSession.Session.ConnectParams.CryptoInfo, Buffer, BufferSize);
  SizeApp := Length(FApplication) + 1;
  SizeParams := SizeOf(ConnectionParams);
  BufSize := SizeParams + SizeOf(BufferSize) + BufferSize + SizeApp;
{$IFDEF MsgCommunicator}
  inc(BufSize, SizeOf(ClientSession.Session.FUserID));
{$ENDIF}
  Buf := MemoryManager.GetMem(BufSize);
{$IFDEF MsgCommunicator}
  Move(ClientSession.Session.FUserID, Buf^, SizeOf(ClientSession.Session.FUserID));
  inc(Buf, SizeOf(ClientSession.Session.FUserID));
{$ENDIF}
  try
   Move(ConnectionParams, Buf^, SizeParams);
   Move(BufferSize, PChar(Buf + SizeParams)^, SizeOf(BufferSize));
   if BufferSize > 0 then
    begin
     Move(Buffer^, PChar(Buf + SizeParams + SizeOf(BufferSize))^, BufferSize);
     MemoryManager.FreeAndNilMem(Buffer);
    end;
   StrPCopy(Buf + SizeParams + SizeOf(BufferSize) + BufferSize, FApplication);
   Retry := 0;
{$IFDEF MsgCommunicator}
   dec(Buf, SizeOf(ClientSession.Session.FUserID));
{$ENDIF}
   repeat
    if ClientSession.ControlCode = ACRTerminate then
     begin
      EnterCriticalSection(FCSect);
      dec(ClientSession.Status);
      LeaveCriticalSection(FCSect);
      Exit;
     end;
{$IFDEF MsgCommunicator}
    if TACRClientSession(ClientSession.Session).Direct then
       ClientSession.Session.FConnectParams.ServerID :=
          TACRClientSession(ClientSession.Session).RemoteUser.UserID; // for connect request only
{$ENDIF}
    try
     SendBuffer(ClientSession.Session, Buf, BufSize, ACRConnect);
    except
      EnterCriticalSection(FCSect);
      dec(ClientSession.Status);
      LeaveCriticalSection(FCSect);
//      raise;
      Exit;
    end;
{$IFDEF MsgCommunicator}
    EnterCriticalSection(FCSect);
    if TACRClientSession(ClientSession.Session).Direct then
      if ClientSession.Session.FConnectParams.ServerID = TACRClientSession(ClientSession.Session).RemoteUser.UserID then // Header.Sender from the answer is not set
        ClientSession.Session.FConnectParams.ServerID := MSG_INVALID_USER_ID;  // allow listen for ACRConnected answer
    LeaveCriticalSection(FCSect);
{$ENDIF}
    StartTime := GetTickCount;
    while ((GetTickCount - StartTime) < ClientSession.Session.ConnectParams.ConnectDelay) do // pause
     if ClientSession.SendStatus <> ACRSent then
      begin
{$IFDEF ProcessMessages}
       Application.ProcessMessages;
{$ENDIF ProcessMessages}
       Sleep(ClientSession.Session.ConnectParams.WaitForSendSleep);
      end
     else
      Exit;
    inc(Retry);
   until  (Retry > ClientSession.Session.ConnectParams.ConnectRetryCount);
   if ClientSession.SendStatus <> ACRSent then
    begin
     ClientSession.Status := ACRVacant;
     ServerID := ClientSession.Session.ConnectParams.ServerID;
{$IFDEF MsgCommunicator}
     Direct := TACRClientSession(ClientSession.Session).Direct;
     if Direct then
       ServerID := TACRClientSession(ClientSession.Session).RemoteUser.UserID
     else
       ServerID := TACRClientSession(ClientSession.Session).FClient.ConnectionParams.ConnectParams.ServerID;
{$ENDIF}
     DoDisconnect(ClientSession.Session);
{$IFDEF MsgCommunicator}
     if Direct then
       raise EACRException.Create(40023, ErrorRCannotConnect,
                           ['user', ServerID,
                            ClientSession.Session.ConnectParams.ConnectRetryCount,
                            ClientSession.Session.ConnectParams.ConnectDelay])
     else
{$ENDIF}
       raise EACRException.Create(40023, ErrorRCannotConnect,
                           ['server', ServerID,
                            ClientSession.Session.ConnectParams.ConnectRetryCount,
                            ClientSession.Session.ConnectParams.ConnectDelay]);
    end;
  finally
   MemoryManager.FreeAndNilMem(Buf);
   EnterCriticalSection(FCSect);
   dec(ClientSession.Status);
   LeaveCriticalSection(FCSect);
  end;
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> SendConnectRequest - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end; // except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT SendConnectRequest finished');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
end;// SendConnectRequest


//------------------------------------------------------------------------------
// Connect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.Connect(Session: TACRBaseSession;
                                              ListenOnly: Boolean = False;
                                              Tune: Boolean = True
                                              );
var
  Connections:          TList;
  Sessions:             TList;
  ClientConnection:     PACRClntConnection;
  ClientSession:        PACRClntSession;
//  MaxHeaderSize:        Integer;
  ConnectionID,
  SessionID:            TACRSessionID;
  Buffer:               PChar;
  BufferSize:           Integer;
  i:                    Integer;
  ConnectionFound:      Boolean;
  SessionFound:         Boolean;
  Network:              TACRNetwork;
  ConnectionParams:     PACRConnectionParams;
begin
  if ListenOnly then
    Session.SessionID := INVALID_SESSION_ID
  else
   begin
    EnterCriticalSection(FCSect);
    Session.SessionID := FSessionID;
    dec(FSessionID);
    if FSessionID = INVALID_SESSION_ID then
      dec(FSessionID);
    LeaveCriticalSection(FCSect);
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT #'+IntToStr(Session.SessionID)+' TRY TO CONNECT TO SERVER #'+IntToStr(Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
{
// To send connect request in one packet:
  MaxHeaderSize := Max (
   SizeOf(TACRPacketHeader) + SizeOf(TACRConnectionParams)
//    + SizeOf(TACRConnectionID)  // set by client
   + SizeOf(TACRSessionID)        // set by server
   ,
   SizeOf(TACRPacketHeader) + SizeOf(TACRConnectionParams)
    + SizeOf(FApplication)
                        );
  if (Session.ConnectParams.PacketSize < MaxHeaderSize) then
   raise EACRException.Create(40020, ErrorRPacketSizeTooSmall,
                                    [Session.ConnectParams.PacketSize,
                                     MaxHeaderSize]);
}
  SessionFound := False;
// Does Session exist?
  Sessions:=FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if ClientSession.Session = Session then
      begin
       SessionFound := True;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
// No. Does Connection exist?
   begin
    ConnectionFound := False;
    Connections:=FConnections.LockList;
    try
     for i:=0 to Connections.Count-1 do
      begin
       ClientConnection := Connections.Items[i];
       if  (ClientConnection.Network.LocalPort = Session.ConnectParams.LocalPort)
       then
        begin
         ConnectionFound := True;
         break;
        end;
      end;
    finally
     FConnections.UnlockList;
    end;
    if not ConnectionFound then
     begin // No Session, no Connection
// Create new connection
      Network := nil;
      Connections:=FConnections.LockList;
      try
       for i:=0 to Connections.Count-1 do
        begin
         ClientConnection := Connections.Items[i];
         if (ClientConnection.Network.LocalPort = Session.ConnectParams.LocalPort)
{
         and
            (ClientConnection.Network.LocalHost := Session.ConnectParams.LocalHost)
}
         then
          begin
           Network := ClientConnection.Network;
           break;
          end;
        end;
      finally
       FConnections.UnlockList;
      end;
      New(ClientConnection);
      EnterCriticalSection(FCSect);
      inc(FConnectionID);
      ClientConnection.ConnectionID := FConnectionID;
      LeaveCriticalSection(FCSect);
      if Network = nil then
       begin
        IncThreadCount;
        ClientConnection.Network := TACRNetwork.Create(self);
        ClientConnection.Network.LocalPort := Session.ConnectParams.LocalPort;
        Session.FConnectParams.LocalPort := ClientConnection.Network.LocalPort; // if port is already in use
       end
      else
        ClientConnection.Network := Network;
// Add new connection to Connections list
      Connections.Add(ClientConnection);
     end; // Set new connection
// No Session, Connection Exists
    New(ClientSession);
    ClientSession.AnswerTime := 0;
    ClientSession.Session := Session;
    ClientSession.ServerSessionID := ClientSession.Session.SessionID; // Header.SessionID
    ClientSession.ConnectionID := ClientConnection.ConnectionID;
    ClientSession.RemoteConnectionID := ClientConnection.ConnectionID;
    ClientSession.CurrentRequestID := 0;
    ClientSession.ClientMessageID := 0;
    ClientSession.ServerMessageID := 0;
    ClientSession.Packets := TList.Create;
    ClientSession.Packets.Capacity := ACRDefaultPacketsInAnswer; // Allocate some place in list
    ClientSession.Packets.Clear; // Set Items to nil
    ClientSession.MsgPackets := TList.Create;
    ClientSession.MsgPackets.Capacity := ACRDefaultMsgPacketsInAnswer; // Allocate some place in list
    ClientSession.MsgPackets.Clear; // Set Items to nil
    ClientSession.PacketIDsToResend := TACRIntegerArray.Create;
    ClientSession.MsgPacketIDsToResend := TACRIntegerArray.Create;
    ClientSession.ListeningThreads := TThreadList.Create;
    ClientSession.MsgListeningThreads := TThreadList.Create;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           ListeningThreads created in @ClientSession='+IntToStr(Integer(ClientSession)));
{$ENDIF}
    ClientSession.ResendRequestThread := nil;
    ClientSession.MsgResendRequestThread := nil;
    ClientSession.AnswerStatus := ACRNo;
    ClientSession.ControlCode := ACRExecute;
    ClientSession.MsgControlCode := ACRExecute;
    ClientSession.Status := ACRVacant;
// Add new Session
    Sessions:=FSessions.LockList;
    try
     Sessions.Add(ClientSession);
    finally
     FSessions.UnlockList;
    end;
    if not ListenOnly then
     begin
// Connect to server
    EnterCriticalSection(FCSect);
    inc(ClientSession.Status);
    LeaveCriticalSection(FCSect);
    SendConnectRequest(ClientSession);
    if ClientSession.Status <= ACRVacant then
     begin
      DoDisconnect(Session);
      raise EACRException.Create(40039, ErrorRCannotConnetToServer,
                                [Session.ConnectParams.RemoteHost,
                                 Session.ConnectParams.RemotePort,
                                 Session.ConnectParams.LocalPort,
                                 Session.ConnectParams.ServerID]);
     end;
    if ClientSession.ControlCode = ACRTerminate then
     begin
      EnterCriticalSection(FCSect);
      dec(ClientSession.Status);
      LeaveCriticalSection(FCSect);
      Exit;
     end;
    ClientSession.PacketIDsToResend.SetSize(0); // Do not request broken packets
    ReceiveBuffer(Session, Buffer, BufferSize);
// Get connection parameters
    ConnectionParams := PACRConnectionParams(Buffer);
    ClientSession.Session.FConnectParams.PacketSize := ConnectionParams.PacketSize;
    ClientSession.Session.FConnectParams.CompressionAlgorithm := ConnectionParams.CompressionAlgorithm;
    ClientSession.Session.FConnectParams.CompressionMode := ConnectionParams.CompressionMode;
    // SessionID
    Move((Buffer+SizeOf(TACRConnectionParams))^, SessionID, SizeOf(SessionID));
    ConnectionID := -MAXINT;
{$IFDEF MsgCommunicator}
    if (BufferSize>=(SizeOf(TACRConnectionParams))+SizeOf(SessionID)+SizeOf(ConnectionID)) then
      Move((Buffer+SizeOf(TACRConnectionParams)+SizeOf(SessionID))^, ConnectionID, SizeOf(SessionID));
{$ENDIF}
    MemoryManager.FreeAndNilMem(Buffer);
    ClientSession.ServerSessionID := SessionID;
    if (ConnectionID = -MAXINT) then
      ClientSession.RemoteConnectionID := ClientSession.ConnectionID
    else
      ClientSession.RemoteConnectionID := ConnectionID;
    EnterCriticalSection(FCSect);
    dec(ClientSession.Status);
    LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('ClientSession.ServerSessionID = '+IntToStr(ClientSession.ServerSessionID));
{$ENDIF}
      if Tune then
        TuneConnectionParamaters(ClientSession);
     end; // Connect to server
   end; //Session not Found

  // Update ClientConnectionManager variables with sessions settings
  FMaxThreadCount := ClientSession.Session.ConnectParams.MaxThreadCount;
  FReceiveTimeOut := ClientSession.Session.ConnectParams.ReceiveTimeOut;

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT #'+IntToStr(Session.SessionID)+
              ' CONNECTED TO SERVER #'+IntToStr(Session.ConnectParams.ServerID)+
              ' SESSION #'+IntToStr(ClientSession.ServerSessionID));
aaWriteToLog('==============================================================');
{$ENDIF}

end; // Connect


//------------------------------------------------------------------------------
// TuneConnectionParamaters
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TuneConnectionParamaters(ClientSession:  PACRClntSession);
var
  Buffer:               PChar;
  BufferSize:           Integer;
  i:                    Integer;
  StartTime:            DWORD;
begin
// test network speed for connection parameters tunning
    if ClientSession.Session.ConnectParams.ConnectionParamsTunning then
     begin
      StartTime := GetTickCount;
      BufferSize := (ClientSession.Session.ConnectParams.PacketSize-SizeOf(TACRPacketHeader)) * ClientSession.Session.ConnectParams.TestPacketCount;
      Buffer := MemoryManager.GetMem(BufferSize);
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.Connect - NETWORK TUNNING:');
aaWriteToLog('Buffer size = '+IntToStr(BufferSize)+' Bytes');
{$ENDIF}
      try
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.Connect - NETWORK TUNNING - Sending...');
{$ENDIF}
      SendBuffer(ClientSession.Session,Buffer,BufferSize,ACRTunning);
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.Connect - NETWORK TUNNING - Sent!');
{$ENDIF}
      finally
       MemoryManager.FreeAndNilMem(Buffer);
      end;
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.Connect - NETWORK TUNNING - Receiving...');
{$ENDIF}
      ReceiveBuffer(ClientSession.Session,Buffer,BufferSize);
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.Connect - NETWORK TUNNING - Received!');
{$ENDIF}
      ClientSession.AnswerTime := GetTickCount - StartTime;
      ClientSession.Session.FConnectParams.SendTimeOut := ClientSession.AnswerTime * 4;
      if ClientSession.Session.ConnectParams.SendTimeOut < ClientSession.Session.ConnectParams.MinSendTimeOut then
        ClientSession.Session.FConnectParams.SendTimeOut := ClientSession.Session.ConnectParams.MinSendTimeOut;
      ClientSession.Session.FConnectParams.ReceiveTimeOut := ClientSession.Session.ConnectParams.SendTimeOut;
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.Connect - NETWORK TUNNING:');
aaWriteToLog('---------------------------------------------------------------');
aaWriteToLog('Packet size = '+IntToStr(ClientSession.Session.ConnectParams.PacketSize)+' Bytes');
aaWriteToLog('Test packets count = '+IntToStr(ACRTestPacketCount));
aaWriteToLog('Answer time  = '+IntToStr(ClientSession.AnswerTime)+' msec');
aaWriteToLog('New timeout  = '+IntToStr(FSendTimeOut)+' msec');
aaWriteToLog('===============================================================');
{$ENDIF}
      BufferSize := SizeOf(ClientSession.AnswerTime);
      Buffer := MemoryManager.GetMem(BufferSize);
      try
       Move(ClientSession.AnswerTime,Buffer^,BufferSize);
       SendBuffer(ClientSession.Session,Buffer,BufferSize,ACRServerSessionTunning);
      finally
       MemoryManager.FreeAndNilMem(Buffer);
      end;
     end;
// end of network testing
end; // TuneConnectionParamaters


//------------------------------------------------------------------------------
// SendPing
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendPing(
                          ClientSession: PACRClntSession
                          );
var
  Header:               PACRPacketHeader;
  Connections:          TList;
  ClientConnection:     PACRClntConnection;
  i:                    Integer;
begin
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
 try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - START');
{$ENDIF}
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   Header.ControlCode := ACRPing;
   Header.Signature := ACRClientPacketSign;
   Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
   Header.ConnectionID := ClientSession.RemoteConnectionID;
   Header.SessionID := ClientSession.ServerSessionID;
   Header.PacketID := 0;
   Header.CurrentRequestID := 0;
   Connections:=FConnections.LockList;
   try
    for i:=0 to Connections.Count-1 do
     begin
      ClientConnection := Connections.Items[i];
      if (ClientConnection.ConnectionID = ClientSession.ConnectionID) then
       break;
     end;
   finally
    FConnections.UnlockList;
   end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - Enter...');
{$ENDIF}
   EnterCriticalSection(FCSect);
   try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - Entered!');
{$ENDIF}
    ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
    ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - Send...');
aaWriteToLog('TACRClientConnectionManager.SendPing> ClientSession.Session.ConnectParams.RemoteHost="'+ClientSession.Session.ConnectParams.RemoteHost+'"');
aaWriteToLog('TACRClientConnectionManager.SendPing> ClientSession.Session.ConnectParams.RemotePort='+IntToStr(ClientSession.Session.ConnectParams.RemotePort));
{$ENDIF}
    ClientConnection.Network.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - Sent!');
{$ENDIF}
   finally
    LeaveCriticalSection(FCSect);
   end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - Left!');
{$ENDIF}
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendPing - FINISH');
{$ENDIF}
 end;
end;// SendPing


//------------------------------------------------------------------------------
// SendAcknowledgement
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendAcknowledgement(
                          ClientSession: PACRClntSession;
                          Msg: Boolean = False;
                          CurrentRequestID:     Integer = -1
                          );
var
  Header:               PACRPacketHeader;
  Connections:          TList;
  ClientConnection:     PACRClntConnection;
  i:                    Integer;
begin
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
 try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - START');
{$ENDIF}
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   if Msg then
     Header.ControlCode := ACRMessageReceived
   else
     Header.ControlCode := ACRAllPacketsReceived;
   Header.Signature := ACRClientPacketSign;
   Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
   Header.ConnectionID := ClientSession.RemoteConnectionID;
   Header.SessionID := ClientSession.ServerSessionID;
{$IFDEF MsgCommunicator}
   if TACRClientSession(ClientSession.Session).Direct then
     Header.SessionID := ClientSession.Session.SessionID;
{$ENDIF}
   Header.PacketID := 0;
   if CurrentRequestID>=0 then
     Header.CurrentRequestID := CurrentRequestID
   else
     if Msg then
       Header.CurrentRequestID := ClientSession.ServerMessageID
     else
       Header.CurrentRequestID := ClientSession.CurrentRequestID;
   Connections:=FConnections.LockList;
   try
    for i:=0 to Connections.Count-1 do
     begin
      ClientConnection := Connections.Items[i];
      if (ClientConnection.ConnectionID = ClientSession.ConnectionID) then
       break;
     end;
   finally
    FConnections.UnlockList;
   end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Enter...');
{$ENDIF}
   EnterCriticalSection(FCSect);
   try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Entered!');
{$ENDIF}
    ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
    ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Send...');
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement> ClientSession.Session.ConnectParams.RemoteHost="'+ClientSession.Session.ConnectParams.RemoteHost+'"');
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement> ClientSession.Session.ConnectParams.RemotePort='+IntToStr(ClientSession.Session.ConnectParams.RemotePort));
{$ENDIF}
    ClientConnection.Network.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Sent!');
{$ENDIF}
   finally
    LeaveCriticalSection(FCSect);
   end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Left!');
{$ENDIF}
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - FINISH');
{$ENDIF}
 end;
end;// SendAcknowledgement


{$IFDEF MsgCommunicator}
//------------------------------------------------------------------------------
// SendConnectAckn
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendConnectAckn(
                          ClientSession:        PACRClntSession;
                          ClientConnection:     PACRClntConnection;
                          CurrentRequestID:     Integer = -1
                                                      );
var
  Buffer:                         PChar;
  Buf:                            PChar;
  BufferSize,
  BufSize,
  SizeSID, SizeCID, SizeParams:   Integer;
  ConnectionParams:               TACRConnectionParams;
  RequestID:                      Integer;
begin
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
 try
{
  ConnectionParams.LocalPort := ClientSession.RemotePort;
  ConnectionParams.RemoteHost := FNetwork.LocalHost;
  ConnectionParams.RemotePort := FNetwork.LocalPort;
  ConnectionParams.ServerID := FNetwork.FLocalClient;
}
  ConnectionParams.PacketSize := ClientSession.Session.ConnectParams.PacketSize;
  ConnectionParams.CompressionAlgorithm := ClientSession.Session.ConnectParams.CompressionAlgorithm;
  ConnectionParams.CompressionMode := ClientSession.Session.ConnectParams.CompressionMode;
  SizeParams := SizeOf(ConnectionParams);
  SizeSID := SizeOf(ClientSession.Session.SessionID);
  SizeCID := SizeOf(ClientSession.ConnectionID);
  BufferSize := SizeParams + SizeSID + SizeCID;
  Buffer := MemoryManager.GetMem(BufferSize);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER: SessionID='+IntToStr(ClientSession.Session.SessionID));
aaWriteToLog('SizeSID='+IntToStr(SizeSID)+', SizeParams='+IntToStr(SizeParams)+', BufferSize='+IntToStr(BufferSize));
{$ENDIF}
  try
   Move(ConnectionParams, Buffer^, SizeParams);
   Move(ClientSession.Session.SessionID, (Buffer+SizeParams)^, SizeSID);
   Move(ClientSession.ConnectionID, (Buffer+SizeParams+SizeSID)^, SizeCID);

   if CurrentRequestID >= 0 then
    begin
     RequestID := ClientSession.CurrentRequestID;
     ClientSession.CurrentRequestID := CurrentRequestID;
    end;

   CompressAndEncryptBuffer(ClientSession.Session, Buffer, BufferSize, Buf, BufSize);
   try
    DoSendBuffer(ClientSession, ClientConnection, Buf, BufSize, ACRConnected);
   finally
    if Buf<>Buffer then
      MemoryManager.FreeAndNilMem(Buf);
   end;

   if CurrentRequestID >= 0 then
     ClientSession.CurrentRequestID := RequestID;

   ClientSession.AnswerStatus := ACRNo;
  finally
   MemoryManager.FreeAndNilMem(Buffer);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
 end;
end;// SendConnectAckn


//------------------------------------------------------------------------------
// IsAuthorizationBufferValid
//------------------------------------------------------------------------------
function TACRClientConnectionManager.IsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
begin
  Result := fnIsAuthorizationBufferValid(CryptoInfo, Buffer, BufferSize);
end;// IsAuthorizationBufferValid
{$ENDIF MsgCommunicator}


//------------------------------------------------------------------------------
// SendDisconnectRequest
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendDisconnectRequest(
                                    ClientSession: PACRClntSession;
                                    WaitForAnswer: Boolean = True);
var
  Header:               PACRPacketHeader;
  Connections:          TList;
  ClientConnection:     PACRClntConnection;
  Retry, i:             Integer;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> START');
{$ENDIF}
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
try
  ClientSession.AnswerStatus := ACRNo;
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   Header.ControlCode := ACRDisconnect;
   Header.Signature := ACRClientPacketSign;
   Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
   Header.ConnectionID := ClientSession.RemoteConnectionID;
   Header.SessionID := ClientSession.ServerSessionID;
{$IFDEF MsgCommunicator}
   if TACRClientSession(ClientSession.Session).Direct then
     Header.SessionID := ClientSession.Session.SessionID;
{$ENDIF}
   Header.PacketID := 0;
   Header.CurrentRequestID := ClientSession.CurrentRequestID;
   Connections:=FConnections.LockList;
   try
    for i:=0 to Connections.Count-1 do
     begin
      ClientConnection := Connections.Items[i];
      if (ClientConnection.ConnectionID = ClientSession.ConnectionID) then
        break;
     end;
   finally
    FConnections.UnlockList;
   end;
   Retry := 0;
   ClientSession.SendStatus := ACRNotSent;
   try
    repeat
     if ClientSession.Session.ConnectParams.RemoteHost = '' then
       Exit; // Session already freed
     EnterCriticalSection(FCSect);
     try
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> RemoteHost='+ClientSession.Session.ConnectParams.RemoteHost);
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> RemotePort='+IntToStr(ClientSession.Session.ConnectParams.RemotePort));
{$ENDIF}
      ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
      ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
      if clientSession.ControlCode = ACRTerminate then
        raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
      ClientConnection.Network.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
     finally
      LeaveCriticalSection(FCSect);
     end;
     if not WaitForAnswer then
       Exit;
     StartTime := GetTickCount;
     while ((GetTickCount - StartTime) < ClientSession.Session.ConnectParams.DisconnectDelay) do // pause
      if ClientSession.SendStatus <> ACRSent then
       begin
        Sleep(ClientSession.Session.ConnectParams.WaitForSendSleep);
{$IFDEF ProcessMessages}
        Application.ProcessMessages;
{$ENDIF ProcessMessages}
       end
      else
       Exit;
     inc(Retry);
    until  (Retry > ClientSession.Session.ConnectParams.DisconnectRetryCount);
   if ClientSession.SendStatus <> ACRSent then
{$IFDEF MsgCommunicator}
    if TACRClientSession(ClientSession.Session).Direct then
      raise EACRException.Create(40024, ErrorRCannotDisconnect,
                   ['client session with user', TACRClientSession(ClientSession.Session).UserID,
                    ClientSession.Session.ConnectParams.DisconnectRetryCount,
                    ClientSession.Session.ConnectParams.DisconnectDelay])
    else
{$ENDIF}
      raise EACRException.Create(40024, ErrorRCannotDisconnect,
                   ['server session', ClientSession.ServerSessionID,
                    ClientSession.Session.ConnectParams.DisconnectRetryCount,
                    ClientSession.Session.ConnectParams.DisconnectDelay]);
   finally
{$IFDEF DEBUG_LOG_NETWORK}
if ClientSession.SendStatus <> ACRSent then
aaWriteToLog('CLIENT: SendDisconnectRequest> Answer from server received')
else
aaWriteToLog('CLIENT: SendDisconnectRequest> Answer from server was not received');
{$ENDIF}
   end;
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> FINISH');
{$ENDIF}
 end;
end;// SendDisconnectRequest


//------------------------------------------------------------------------------
// TerminateMessageThreads
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TerminateMessageThreads(ClientSession: PACRClntSession);
var
  i:                    Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
  Error:                String;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> START');
{$ENDIF}
  // Terminate MsgResendRequestThread
  if (ClientSession.MsgResendRequestThread <> nil) then
    ClientSession.MsgResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   if (ClientSession.MsgResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.ThreadsTerminateDelay then
       begin
        Error := 'message resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
        ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
       end;
     end;
   sleep(0);
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> message resend thread terminated');
{$ENDIF}
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
   if ClientSession.MsgListeningThreads = nil then
    begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> MsgListeningThreads = nil - FINISH');
{$ENDIF}
     Exit;
    end;
   // Terminate message threads
   Threads := ClientSession.MsgListeningThreads.LockList;
   try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> MsgListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ClientSession.MsgListeningThreads.UnlockList;
   end;
   // Check
   Threads := ClientSession.MsgListeningThreads.LockList;
   try
    if (Threads.Count = 0)
     then break
     else
      begin
       if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.ThreadsTerminateDelay then
        begin
         Error := IntToStr(Threads.Count)+' message listening threads';
         raise EACRException.Create(40036, ErrorRThreadHangs,
          ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
        end;
      end;
   finally
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> MsgListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    ClientSession.MsgListeningThreads.UnlockList;
   end;
   sleep(0);
  until False;
end; // TerminateMessageThreads


//------------------------------------------------------------------------------
// TerminateCommandThreads
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TerminateCommandThreads(ClientSession: PACRClntSession);
var
  i:                    Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
  Error:                String;
begin
{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> START - Terminate ResendRequestThread > '+IntToStr(GetTickCount));
{$ENDIF}

  // Terminate ResendRequestThread
  if (ClientSession.ResendRequestThread <> nil) then
    ClientSession.ResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   if (ClientSession.ResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.ThreadsTerminateDelay then
       begin
        Error := 'command resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
          ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
       end;
     end;
   sleep(0);
  until False;

{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> Terminate Listening Threads > '+IntToStr(GetTickCount));
{$ENDIF}

  if ClientSession.ListeningThreads = nil then
   begin
{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> ListeningThreads = nil - FINISH');
{$ENDIF}
    Exit;
   end;
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
   // Terminate all threads
   Threads := ClientSession.ListeningThreads.LockList;
   try
{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ClientSession.ListeningThreads.UnlockList;
   end;

{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> Check Terminate Listening Threads > '+IntToStr(GetTickCount));
{$ENDIF}

   // Check
   Threads := ClientSession.ListeningThreads.LockList;
   try
    if (Threads.Count <= FLiveSessions)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.ThreadsTerminateDelay then
       begin
        Error := IntToStr(Threads.Count)+' command listening threads';
        raise EACRException.Create(40036, ErrorRThreadHangs,
          ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
       end;
     end;
   finally
{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    ClientSession.ListeningThreads.UnlockList;
   end;
   sleep(0);
  until False;

{$IFDEF DEBUG_LOG_CLIENT_TERMINATE}
aaWriteToLog('TerminateCommandThreads finished > '+IntToStr(GetTickCount));
{$ENDIF}
end; // TerminateCommandThreads


//------------------------------------------------------------------------------
// TerminateAllSessionThreads
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TerminateAllSessionThreads(ClientSession: PACRClntSession);
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateAllSessionThreads> START');
{$ENDIF}
  TerminateCommandThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateAllSessionThreads> Command threads terminated!');
{$ENDIF}
  TerminateMessageThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateAllSessionThreads> FINISH');
{$ENDIF}
end; // TerminateAllSessionThreads


//------------------------------------------------------------------------------
// DeleteSession
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DeleteSession(ClientSession: PACRClntSession);
var
  StartTime:            DWORD;
  i, Count:             Integer;
  Packet:               PACRPacket;
begin
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> START');
{$ENDIF}
  // block using ServerSession
  EnterCriticalSection(FCSect);
  // no threads - free list
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ListeningThreads will delete in @ClientSession='+IntToStr(Integer(ClientSession)));
{$ENDIF}
  // free list
  ClientSession.ListeningThreads.Free;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ListeningThreads have been deleted');
{$ENDIF}
  ClientSession.ListeningThreads := nil;
  // free command packets
  Count := ClientSession.Packets.Count;
  for i := 0 to Count - 1 do
   begin
    Packet := ClientSession.Packets.Items[i];
    if Packet <> nil then
     begin
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Packet # 0: Free Buffer...');
{$ENDIF}
      if Packet.Buffer <> nil then
        MemoryManager.FreeAndNilMem(Packet.Buffer);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Packet # 0: Dispose Packet...');
{$ENDIF}
      Dispose(Packet);
     end;
   end;
  ClientSession.Packets.Free;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Packets have been deleted');
{$ENDIF}
  // free list
  ClientSession.MsgListeningThreads.Free;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Messages ListeningThreads have been deleted');
{$ENDIF}
  ClientSession.MsgListeningThreads := nil;
  // free message packets
  Count := ClientSession.MsgPackets.Count;
  for i := 0 to Count - 1 do
   begin
    Packet := ClientSession.MsgPackets.Items[i];
    if Packet <> nil then
     begin
      if Packet.Buffer <> nil then
        MemoryManager.FreeAndNilMem(Packet.Buffer);
      Dispose(Packet);
     end;
   end;
  ClientSession.MsgPackets.Free;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Messages Packets have been deleted');
{$ENDIF}
  // free
  ClientSession.PacketIDsToResend.Free;
  ClientSession.MsgPacketIDsToResend.Free;
  // wait for vacant ClientSession
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> wait for vacant ClientSession...');
{$ENDIF}
  StartTime := GetTickCount;
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> CS leaved');
{$ENDIF}
  repeat
    ClientSession.ControlCode := ACRTerminate;
    ClientSession.MsgControlCode := ACRTerminate;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ACRTerminate is Set');
{$ENDIF}
    sleep(0);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ClientSession.Status = '+IntToStr(ClientSession.Status));
{$ENDIF}
    if (ClientSession.Status = ACRVacant) then
      break;
  until ((GetTickCount-StartTime) >= ClientSession.Session.ConnectParams.WaitForTimeOut);
  if (ClientSession.Status > ACRVacant) then
    raise EACRException.Create(40038, ErrorRWaitForTimeOut,
                                ['client session',
                                ClientSession.Session.ConnectParams.WaitForTimeOut]);
  // free session
  EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ClientSession.Status = '+IntToStr(ClientSession.Status));
{$ENDIF}
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Dispose...');
{$ENDIF}
  Dispose(ClientSession);
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> FINISH');
{$ENDIF}
end; // DeleteSession


//------------------------------------------------------------------------------
// DoDisconnect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DoDisconnect(Session: TACRBaseSession);
var
  Connections:          TList;
  Sessions:             TList;
  ClientConnection:     PACRClntConnection;
  ClientSession:        PACRClntSession;
  ConnectionID:         TACRConnectionID;
  i, Count:             Integer;
  NoSessions:           Boolean;
  NoConnections:        Boolean;
begin
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> START');
{$ENDIF}
  Sessions:=FSessions.LockList;
  try
   Count := Sessions.Count;
   for i:=Count-1 downto 0 do
    begin
    if (i<0)
    or (i>=Sessions.Count)
    then
       break;
     ClientSession := Sessions.Items[i];
     if ClientSession.Session = Session then
      begin
       ConnectionID := ClientSession.ConnectionID;
       // Terminate all threads
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> TerminateAllSessionThreads');
{$ENDIF}
       TerminateAllSessionThreads(ClientSession);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> All sessions terminated');
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Remove SessionID='+IntToStr(ClientSession.Session.SessionID));
{$ENDIF}
       FSessions.Remove(ClientSession);
       FSessions.UnlockList;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> DeleteSession...');
{$ENDIF}
       DeleteSession(ClientSession);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Session Deleted!');
{$ENDIF}
       Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Sessions Lockeded!');
{$ENDIF}
       break;
      end;
    end;
//   Sessions.Capacity := Sessions.Count;
// Search for another session using the same connection
   NoSessions := True;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Sessions Count = '+IntToStr(Count));
{$ENDIF}
   Count := Sessions.Count;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Sessions Count = '+IntToStr(Count));
{$ENDIF}
   for i:=0 to Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if ClientSession.ConnectionID = ConnectionID then
      begin
       NoSessions := False;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;

  NoConnections := True;  // Suppose that it is not availabdle to have another connection
                          // with the same localport
{
// Search for another connection with the same localport (= the same network)
  Connections:=FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if (ClientConnection.ConnectionID = ConnectionID)
     then
      begin
       Network := ClientConnection.Network;
       break;
      end;
    end;
   NoConnections := True;
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if (ClientConnection.ConnectionID <> ConnectionID)
     and (ClientConnection.Network = Network)
     then
      begin
       NoConnections := False;
       break;
      end;
    end;
  finally
   FConnections.UnlockList;
  end;
}

// There is no another session who use this connection
 if NoSessions then
 begin
  Connections:=FConnections.LockList;
  Count := Connections.Count;
  try
   for i:=Count-1 downto 0 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.ConnectionID = ConnectionID then
      begin
       if NoConnections then
        if ClientConnection.Network <> nil then
         begin
          ClientConnection.Network.Free;
          DecThreadCount;
         end;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Freed connection!');
{$ENDIF}
       ClientConnection.Network := nil;
       Dispose(Connections.Items[i]);
       Connections.Delete(i);
      end;
    end;
  finally
   FConnections.UnlockList;
  end;
 end;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> FINISH');
{$ENDIF}
end; // DoDisconnect


//------------------------------------------------------------------------------
// Disconnect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.Disconnect(Session: TACRBaseSession;
                                                  ListenOnly: Boolean = False);
var
  Sessions:             TList;
  ClientSession:        PACRClntSession;
  i:                    Integer;
  Found:                Boolean;
begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT #'+IntToStr(Session.SessionID)+' TRY TO DISCONNECT FROM SERVER #'+IntToStr(Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}

  if not ListenOnly then
   begin // Send disconnect request
    Found := False;
    Sessions:=FSessions.LockList;
    try
     for i:=Sessions.Count-1 downto 0 do
      begin
       ClientSession := Sessions.Items[i];
       if ClientSession.Session = Session then
        begin
         Found := True;
         break;
        end;
      end;
    finally
     FSessions.UnlockList;
    end;
    if not Found then
      raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
    inc(ClientSession.CurrentRequestID);
    SendDisconnectRequest(ClientSession);
   end; // Send disconnect request

  DoDisconnect(Session);

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT DISCONNECTED FROM SERVER');
aaWriteToLog('==============================================================');
{$ENDIF}

end; // Disconnect


//------------------------------------------------------------------------------
// DisconnectAll
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DisconnectAll;
var
  Sessions:             TList;
  Connections:          TList;
  ClientSession:        PACRClntSession;
  ClientConnection:     PACRClntConnection;
  i:                    Integer;
begin
{$IFDEF CONNECTION_TEST}
  Exit;
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DisconnectAll> START');
{$ENDIF}
 Sessions:=FSessions.LockList;
 try
  while Sessions.Count>=1 do
//  for i:=Sessions.Count-1 downto 0 do
   begin
    ClientSession := Sessions.Items[0];
    SendDisconnectRequest(ClientSession, False);
    // Terminate all threads
    TerminateAllSessionThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DisconnectAll> All sessions terminated');
{$ENDIF}
    FSessions.Remove(ClientSession);
    FSessions.UnlockList;
    DeleteSession(ClientSession);
    Sessions:=FSessions.LockList;
   end;
//  Sessions.Capacity :=0;
 finally
   FSessions.UnlockList;
 end;

 Connections:=FConnections.LockList;
 try
  for i:=Connections.Count-1 downto 0 do
   begin
    ClientConnection := Connections.Items[i];
    if ClientConnection.Network <> nil then
     begin
      ClientConnection.Network.Free;
      DecThreadCount;
     end;
    ClientConnection.Network := nil;
    Dispose(Connections.Items[i]);
    Connections.Delete(i);
   end;
//  Connections.Capacity :=0;
 finally
  FConnections.UnlockList;
 end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DisconnectAll> FINISH');
{$ENDIF}
end; // DisconnectAll


//------------------------------------------------------------------------------
// SendMessage
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                                                  );
{
var
  Buf:  PChar;
}
begin
 try
{
  if (TACRCompressionAlgorithm(Session.ConnectParams.CompressionAlgorithm) <> acaNone)
  or (Session.ConnectParams.CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)
  then
   begin
    Buf := MemoryManager.GetMem(BufferSize);
    Move(Buffer^,Buf^,BufferSize);
    SendBuffer(Session, Buf, BufferSize, ACRMessage);
    MemoryManager.FreeAndNilMem(Buf);
   end
  else
}
  SendBuffer(Session, Buffer, BufferSize, ACRMessage);
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientConnectionManager.SendMessage - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // SendMessage


//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNewRequest
                                                  );
var
  Connections:          TList;
  Sessions:             TList;
  ClientConnection:     PACRClntConnection;
  ClientSession:        PACRClntSession;
  ConnectionID:         TACRConnectionID;
  i:                    Integer;
  Found:                Boolean;
  Buf:                  PChar;
  BufSize:              Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - START');
{$ENDIF}
 try
  if (Session.ConnectParams.PacketSize < SizeOf(TACRPacketHeader)) then
   raise EACRException.Create(40020, ErrorRPacketSizeTooSmall,
                                    [Session.ConnectParams.PacketSize,
                                     SizeOf(TACRPacketHeader)]);
//  ClientSession := FindSession(Session);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - started');
{$ENDIF}
  Found := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if ClientSession.Session = Session then
       begin
        ConnectionID := ClientSession.ConnectionID;
        Found := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
  Found := False;
  Connections := FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.ConnectionID = ConnectionID then
       begin
        Found := True;
        break;
       end;
    end;
  finally
   FConnections.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);

{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('CLIENT>>> '+IntToStr(ClientSession.CurrentRequestID + 1)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}

  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);

  if (Code = ACRConnect)
  or (Code = ACRTunning)
  then
   begin
    Buf := Buffer;
    BufSize := BufferSize;
   end
  else
   begin  // not connect or tunning request, evaluate appropriate parameters
    CompressAndEncryptBuffer(ClientSession.Session, Buffer, BufferSize, Buf, BufSize);

// tune connection parameters
    if ClientSession.Session.ConnectParams.ConnectionParamsTunning then
     begin
      // i = packets count
      i := (BufSize div (ClientSession.Session.ConnectParams.PacketSize-SizeOf(TACRPacketHeader)));
      if i>ClientSession.Session.ConnectParams.TestPacketCount then
       begin
        if ClientSession.AnswerTime=0 then
          TuneConnectionParamaters(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('Packets count in request,           i = '+IntToStr(i));
aaWriteToLog('sqrt(i*0.1) = '+FloatToStr(sqrt(i*0.1)));
aaWriteToLog('expectancy count of packets to resend = '+FloatToStr( (sqrt(i*0.1)*0.5+1)*(i*0.3) ));
aaWriteToLog('total packets planed to send          = '+IntToStr(trunc((sqrt(i*0.1)*0.5+1)*(i*0.3)+i)));
aaWriteToLog('times more = '+IntToStr(trunc(((sqrt(i*0.1)*0.5+1)*(i*0.3)+i)/ACRTestPacketCount)+1));
aaWriteToLog('answer time / 2 ='+IntToStr(ClientSession.AnswerTime div 2));
aaWriteToLog('time planned    ='+IntToStr((trunc( ((sqrt(i*0.1)*0.5+1)*(i*0.3)+i)/ACRTestPacketCount)+1)*(ClientSession.AnswerTime div 2)));
aaWriteToLog('assurance factor, 1+log10(i)='+IntToStr(1+trunc(log10(i))));
{$ENDIF}
        ClientSession.Session.FConnectParams.SendTimeOut := (trunc( ((sqrt(i*0.1)*0.5+1)*(i*0.3)+i)/ClientSession.Session.ConnectParams.TestPacketCount)+1)*(ClientSession.AnswerTime div 2);
        if ClientSession.AnswerTime < 1000 then
          ClientSession.Session.FConnectParams.SendTimeOut := ClientSession.Session.ConnectParams.SendTimeOut * (1+trunc(log10(i)));
       end;
      if ClientSession.Session.ConnectParams.SendTimeOut < ClientSession.Session.ConnectParams.MinSendTimeOut then
        ClientSession.Session.FConnectParams.SendTimeOut := ClientSession.Session.ConnectParams.MinSendTimeOut;
      ClientSession.Session.FConnectParams.ReceiveTimeOut := ClientSession.Session.ConnectParams.SendTimeOut * 2;
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - NETWORK TUNNING:');
aaWriteToLog('---------------------------------------------------------------');
aaWriteToLog('Buffer size = '+IntToStr(BufferSize)+' Bytes (original)');
aaWriteToLog('Buffer size = '+IntToStr(BufSize)+' Bytes (compressed)');
aaWriteToLog('Packet size = '+IntToStr(ClientSession.Session.ConnectParams.PacketSize)+' Bytes');
aaWriteToLog('---------------------------------------------------------------');
aaWriteToLog('Packets count to send = '+IntToStr(i));
aaWriteToLog('Test packets count    = '+IntToStr(ACRTestPacketCount));
aaWriteToLog('Answer time = '+IntToStr(ClientSession.AnswerTime)+' msec');
aaWriteToLog('New timeout = '+IntToStr(FSendTimeOut)+' msec');
aaWriteToLog('===============================================================');
{$ENDIF}
     end;
   end;
  try
   DoSendBuffer(ClientSession, ClientConnection, Buf, BufSize, Code);
  finally
   if Buf <> Buffer then
    if Code <> ACRConnect then
     MemoryManager.FreeAndNilMem(Buf);
  end;

  if ClientSession.Status <= ACRVacant then
    raise EACRException.Create(40040, ErrorRCannotSendToServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);

  EnterCriticalSection(FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FCSect);

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - FINISH');
{$ENDIF}

 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRClientConnectionManager.SendBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // SendBuffer


//------------------------------------------------------------------------------
// DoSendBuffer
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DoSendBuffer(
                          ClientSession:    PACRClntSession;
                          ClientConnection: PACRClntConnection;
                          Buffer:           PChar;
                          BufferSize:       Integer;
                          Code:             Integer = ACRNewRequest
                                                  );
var
  Header:               PACRPacketHeader;
  Packets:              TList;
  Packet:               PACRPacket;
  BytesSent, DataSize:  Integer;
  i:                    Integer;

procedure WaitForCommandSent;
var
  PacketID,
  i:                    Integer;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
  j:                    Integer;
{$ENDIF}
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT waits for all packets');
{$ENDIF}
  if not ((Code = ACRConnect) or (Code = (ACRConnect + ACRLastPacket))) then
   begin
(*
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRReceiveSleep); // To avoid delay if full answer is already received
*)
    i := ClientSession.PacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ClientSession.SendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRResendDelay);
      if ClientSession.PacketIDsToResend.ItemCount = 0 then
       begin
        StartSleepTime := GetTickCount;
        while ((GetTickCount - StartSleepTime) < ClientSession.Session.ConnectParams.ResendDelay) do  // sleep after last packet send
         begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> WaitForCommandSent - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
          if ClientSession.SendStatus = ACRSent then
            Exit;
          if ClientSession.ControlCode = ACRTerminate then
           begin
            dec(ClientSession.Status);
            raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
           end;
          if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.SendTimeOut then
           begin
            dec(ClientSession.Status);
            raise EACRException.Create(40077, ErrorRTimeOutSending,
              ['command', ClientSession.Session.ConnectParams.ServerID,
              ClientSession.Session.ConnectParams.SendTimeOut]);
           end;
          Sleep(ClientSession.Session.ConnectParams.WaitForSendSleep);
         end; // sleep
       end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT has '+IntToStr(ClientSession.PacketIDsToResend.ItemCount)+' packets to resend:');
for j:=0 to ClientSession.PacketIDsToResend.ItemCount-1 do
aaWriteToLog(IntToStr(ClientSession.PacketIDsToResend.Items[j]));
{$ENDIF}
      dec(i);
      if (i<0) then
        i := ClientSession.PacketIDsToResend.ItemCount - 1;
      EnterCriticalSection(FCSect);
      try
       if (ClientSession.PacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ClientSession.PacketIDsToResend.Items[i];
           if PacketID >= Packets.Count then
             Continue;
          end
         else // No packets to resend - Resend last packet
           PacketID := Packets.Count - 1;
         Packet := Packets.Items[PacketID];
         // resend packet
         ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
         ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
         if ClientSession.ControlCode = ACRTerminate then
          begin
           dec(ClientSession.Status);
           raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
          end;
         try
          ClientConnection.Network.SendBuffer(Packet.Buffer, Packet.BufferSize);
         except
          dec(ClientSession.Status);
          raise;
         end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT resent packet # '+IntToStr(PacketID));
{$ENDIF}
        end;
      finally
       if  (ClientSession.PacketIDsToResend.ItemCount > 0)
       and (ClientSession.PacketIDsToResend.ItemCount > i)
       then
         ClientSession.PacketIDsToResend.Delete(i);
       LeaveCriticalSection(FCSect);
      end; // finally
     end; // loop
   end; // no Connect
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT received all the packets');
{$ENDIF}
end; // WaitForCommandSent

procedure WaitForMessageSent;
var
  PacketID,
  i:                    Integer;
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
(*
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRReceiveSleep); // To avoid delay if full answer is already received
*)
    i := ClientSession.MsgPacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ClientSession.MsgSendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRResendDelay);
      if ClientSession.MsgPacketIDsToResend.ItemCount = 0 then
       begin
        StartSleepTime := GetTickCount;
        while ((GetTickCount - StartSleepTime) < ClientSession.Session.ConnectParams.ResendDelay) do  // sleep after last packet send
         begin
          if ClientSession.SendStatus = ACRSent then
            Exit;
          if ClientSession.MsgControlCode = ACRTerminate then
           begin
            dec(ClientSession.Status);
            raise EACRException.Create(40081, ErrorRClientMsgSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
           end;
          if (GetTickCount - StartTime) > ClientSession.Session.ConnectParams.SendTimeOut then
           begin
            EnterCriticalSection(FCSect);
            dec(ClientSession.Status);
            LeaveCriticalSection(FCSect);
            raise EACRException.Create(40077, ErrorRTimeOutSending,
              ['message', ClientSession.Session.ConnectParams.ServerID,
                          ClientSession.Session.ConnectParams.SendTimeOut]);
           end;
          Sleep(ClientSession.Session.ConnectParams.WaitForSendSleep);
         end; // sleep
       end;
      dec(i);
      if i<0 then
        i := ClientSession.MsgPacketIDsToResend.ItemCount - 1;
      EnterCriticalSection(FCSect);
      try
       if (ClientSession.MsgPacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ClientSession.MsgPacketIDsToResend.Items[i];
           if PacketID >= Packets.Count then
             Continue;
          end
         else // No packets to resend - Resend last packet
           PacketID := Packets.Count - 1;
         Packet := Packets.Items[PacketID];
         // resend packet
         ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
         ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
         if ClientSession.MsgControlCode = ACRTerminate then
          begin
           dec(ClientSession.Status);
           raise EACRException.Create(40081, ErrorRClientMsgSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
          end;
         try
          ClientConnection.Network.SendBuffer(Packet.Buffer, Packet.BufferSize);
         except
          dec(ClientSession.Status);
          raise;
         end;
        end;
      finally
       if  (ClientSession.MsgPacketIDsToResend.ItemCount > 0)
       and (ClientSession.MsgPacketIDsToResend.ItemCount > i)
       then
         ClientSession.MsgPacketIDsToResend.Delete(i);
       LeaveCriticalSection(FCSect);
      end; // finally
     end;
end; // WaitForMessageSent

procedure CheckTermination;
begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> BEGIN');
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> ClientSession.Status = '+IntToStr(ClientSession.Status));
{$ENDIF}
  try
        if Code = ACRMessage then
         begin
          if ClientSession.MsgControlCode = ACRTerminate then
           begin
            dec(ClientSession.Status);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> 40081');
{$ENDIF}
            raise EACRException.Create(40081, ErrorRClientMsgSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
           end;
         end
        else
          if ClientSession.ControlCode = ACRTerminate then
           begin
            dec(ClientSession.Status);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> 40080');
{$ENDIF}
            raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
           end;
  finally
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> ClientSession.Status = '+IntToStr(ClientSession.Status));
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> END');
{$ENDIF}
  end;
end; // CheckTermination

begin // DoSendBuffer
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> START - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  DataSize := ClientSession.Session.ConnectParams.PacketSize - SizeOf(TACRPacketHeader);
  Header.Signature := ACRClientPacketSign;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> Code='+IntToStr(Code));
{$ENDIF}
  Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
  Header.Sender := ClientConnection.Network.LocalClientID;
//  Header.ConnectionID := ClientConnection.ConnectionID;
  Header.ConnectionID := ClientSession.RemoteConnectionID;
////  if Code=ACRConnect then
  Header.SessionID := ClientSession.ServerSessionID;
{$IFDEF MsgCommunicator}
  if TACRClientSession(ClientSession.Session).Direct then
   if Code<>ACRConnected then
    Header.SessionID := ClientSession.Session.SessionID;
{$ENDIF}
  Header.PacketID := 0;
  Header.ControlCode := Code;
  if (Code >= ACRNewRequest)
  or (Code = ACREcho)
  or (Code = ACRTunning)
  or (Code = ACRServerSessionTunning)
  then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT IS SENDING NEW REQUEST TO SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
    EnterCriticalSection(FCSect);
    inc(ClientSession.CurrentRequestID);
    ClientSession.Packets.Count := 0; // Delete old packets receved since ReceiveBuffer finished
    ClientSession.AnswerStatus := ACRNo;
    LeaveCriticalSection(FCSect);
   end;
  Header.CurrentRequestID := ClientSession.CurrentRequestID;
  if Code = ACRMessage then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT IS SENDING MESSAGE # '+IntToStr(ClientSession.ClientMessageID)+'TO SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
    Header.CurrentRequestID := ClientSession.ClientMessageID;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> create packets list...');
{$ENDIF}
  Packets := TList.Create;
  try
    BytesSent := 0;
    i := DataSize;
    if Code = ACRMessage then
     begin
      ClientSession.MsgSendStatus := ACRNotSent;
     end
    else
      ClientSession.SendStatus := ACRNotSent;
    repeat
//    while BytesSent < BufferSize do
//     begin // Create and send all packets
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> create packet #'+IntToStr(Header.PacketID));
{$ENDIF}
      New(Packet);
      Packets.Add(Packet);
      Packet.Buffer := MemoryManager.GetMem(ClientSession.Session.ConnectParams.PacketSize);
      if BytesSent + DataSize > BufferSize then
       DataSize := BufferSize - BytesSent;
      Packet.BufferSize := DataSize + SizeOf(TACRPacketHeader);
      if BytesSent + DataSize = BufferSize then
       Header.ControlCode := Header.ControlCode+ACRLastPacket;
      Move(Header^, Packet.Buffer^, SizeOf(TACRPacketHeader));
      Move(Pointer(Integer(Buffer)+Header.PacketID*i)^, (Packet.Buffer+SizeOf(TACRPacketHeader))^, DataSize);
      inc(Header.PacketID);
      // send packet
      EnterCriticalSection(FCSect);
      try
       ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
       ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
       CheckTermination;
       try
        ClientConnection.Network.SendBuffer(Packet.Buffer, Packet.BufferSize);
       except
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise;
       end;
      finally
       LeaveCriticalSection(FCSect);
      end;
//      sleep(0); // DO NOT UNCOMMENT!!! - The fastest speed is without sleep !!! - for single-packet requests
      if ((Header.PacketID) mod 16)=15 then
        sleep(0); // For huge packets requests
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> packet sent');
{$ENDIF}
      // packet has been sent
      BytesSent := BytesSent + DataSize;
//     end;
    until BytesSent >= BufferSize;
    if Code = ACRMessage then
     begin
      WaitForMessageSent;
      inc(ClientSession.ClientMessageID);
     end
    else
      if Code <> ACRConnected then
        WaitForCommandSent;
  finally
  // Remove all sent packets and free memory
    for i:= 0 to Packets.Count - 1 do
     begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> free packet #'+IntToStr(i));
{$ENDIF}
      Packet := Packets.Items[i];
      MemoryManager.FreeAndNilMem(Packet.Buffer);
      Dispose(Packet);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> freed');
{$ENDIF}
     end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> free packets...');
{$ENDIF}
    Packets.Free;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> freed');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> free header...');
{$ENDIF}
    MemoryManager.FreeAndNilMem(Header);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> freed');
{$ENDIF}
   EnterCriticalSection(FCSect);
   dec(ClientSession.Status);
   LeaveCriticalSection(FCSect);
   CheckTermination;
  end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT sent buffer');
aaWriteToLog('==============================================================');
{$ENDIF}
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer - FINISH');
{$ENDIF}
end; // DoSendBuffer


//------------------------------------------------------------------------------
// if encryption algorithm <> acr_Cipher_None then allocate buffer, fill it and return size
// otherwise return BufferSize = 0
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.CreateAuthorizationBuffer(
                      CryptoInfo:     TACRCryptoInfo;
                      out Buffer:     PChar;
                      out BufferSize: Integer
                                    );
var ms:           TACRMemoryStream;
    buf:          PChar;
    crc32:        Cardinal;
    size:         Integer;
begin
  if CryptoInfo.CryptoAlgorithm = ACR_Cipher_None then
   begin
    BufferSize := 0;
    Exit;
   end;
  size := ACRDefaultAuthorizationBufferSize;
  buf := MemoryManager.GetMem(size);
  ms := TACRMemoryStream.Create;
  try
    ACRGenerateRandomBuffer(buf,size);
    crc32 := ACRCountCRC(0,buf,size);
    SaveDataToStream(size,SizeOf(crc32),ms,11319);
    SaveDataToStream(crc32,SizeOf(crc32),ms,11320);
    ACREncryptBuffer(CryptoInfo,buf,size);
    ms.WriteBuffer(buf^,size);
    BufferSize := ms.Size;
    Buffer := ms.Buffer;
    ms.Buffer := nil;
  finally
    ms.Free;
    MemoryManager.FreeAndNilMem(buf);
  end;
end; // CreateAuthorizationBuffer

// TACRClientConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientListenerThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientListenerThread.Create(
                       Manager:       TACRClientConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       Network:       TACRNetwork;
                       FromHost:      String;
                       FromPort:      Integer
                                            );
var
  Error:        String;
begin
 try
  ClientSession := nil;
  FManager := Manager;
  FManager.IncThreadCount;
  FNetwork := Network;
  self.Buffer := Buffer;
  self.BufferSize := BufferSize;
  self.FromHost := FromHost;
  self.FromPort := FromPort;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    if (ClientSession = nil)
    or (ClientSession.Session = nil)
    then
     begin
      Error:=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         ClientSession := Sessions.Items[0];
         if ClientSession.Session <> nil then
           TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrClientListenerThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrClientListenerThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientListenerThread.Destroy;
var
  Error:        String;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if ClientSession <> nil then
   begin
    if ClientSession.ListeningThreads <> nil then
      ClientSession.ListeningThreads.Remove(self);
    if ClientSession.MsgListeningThreads <> nil then
      ClientSession.MsgListeningThreads.Remove(self);
   end;
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT LISTENER THREAD - FINISHED');
{$ENDIF}
  if (FManager<>nil) then
    FManager.DecThreadCount;
 except
  on E: Exception do
   begin
    if (ClientSession = nil)
    or (ClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         ClientSession := Sessions.Items[0];
         if ClientSession.Session <> nil then
           TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrClientListenerThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrClientListenerThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message);
   end;
 end;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRClientListenerThread.Execute;
label
  KillPacket,
  IsMsgFull;
var
  Error:                String;
  Header:               PACRPacketHeader;
  Packet:               PACRPacket;
  Connections:          TList;
{$IFDEF MsgCommunicator}
  Session:              TACRBaseSession;
  ClientSession2:       PACRClntSession;
{$ENDIF}
  ClientConnection:     PACRClntConnection;
  AllPacketsReceived,
  SessionFound,
  ConnectionFound:      Boolean;
  Buf:                  PChar;
  BufSize:              Integer;
  i:                    Integer;
{$IFDEF MsgCommunicator}
  AuBufSize:            Integer;
  UserID:               Cardinal;
  ConnectionParams:     PACRConnectionParams;
  Application:          String;
{$ENDIF}
  CurrentRequestID:     Integer;
begin
//EnterCriticalSection(FManager.FCSect);
try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRClientListenerThread.Execute> START - '+IntToStr(GetTickCount));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** CLIENT LISTENER START *****');
{$ENDIF}
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
  Header := Pointer(Buffer);
// Check for errors
  if CheckSum(Buffer, BufferSize) <> Header.CheckSum then
   begin
    FManager.PacketResendRequest(Buffer, ClientConnection.Network, FromHost, FromPort);
    goto KillPacket;
   end;
// Check sign
  if (Header.Signature <> ACRServerPacketSign)
{$IFDEF MsgCommunicator}
  and (Header.Signature <> ACRClientPacketSign)
{$ENDIF}
  then
    goto KillPacket;
{$IFDEF MsgCommunicator}
  if    // first packet in multi-packet connect request
  ((Header.ControlCode=ACRConnect) and (Header.PacketID=0))
  or    // single-packet connect request
  ((Header.ControlCode=(ACRConnect+ACRLastPacket)) and (Header.PacketID=0))
  then
   begin // Connect
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connect');
{$ENDIF}
    // is direct connection enabled?
    SessionFound := False;
    Sessions := FManager.FSessions.LockList;
    try
     for i:=0 to Sessions.Count-1 do
      begin
       ClientSession2 := Sessions.Items[i];
       if (ClientSession2.ServerSessionID = INVALID_SESSION_ID) then
        if TACRClientSession(ClientSession2.Session)<>nil then
         if TACRClientSession(ClientSession2.Session).UserID = Header.Recepient then // for connect request only
          if TACRClientSession(ClientSession2.Session).Direct then
           if ClientSession2.Session.Connected then
            if (ClientSession2.Session.SessionID = INVALID_SESSION_ID) then // special listening session
             begin
              Session := ClientSession2.Session;
              SessionFound := True;
              break;
             end;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
    if not SessionFound then
      goto KillPacket;
    // does connection already exist?
    SessionFound := False;
    Sessions := FManager.FSessions.LockList;
    try
     for i:=0 to Sessions.Count-1 do
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Check session '+IntToStr(i)+' of '+IntToStr(Sessions.Count));
{$ENDIF}
       ClientSession := Sessions.Items[i];
       if (ClientSession.ServerSessionID = Header.SessionID)
       and (ClientSession.RemoteConnectionID = Header.ConnectionID)
       then
         if (ClientSession.Session = nil) // just added, not finished
          then SessionFound := True
         else if (ClientSession.Session.SessionID = INVALID_SESSION_ID) // buffer received, not finished
          then SessionFound := True
         else if (ClientSession.Session.ConnectParams.ServerID = Header.Sender) // not first connect - session already exist
          then SessionFound := True;
         if SessionFound then
          begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ClientSession),6)+', Client SessionID='+IntToStr(ClientSession.ServerSessionID));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
if ClientSession.Session <> nil then
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ClientSession),6)+', Server SessionID='+IntToStr(ClientSession.Session.SessionID));
{$ENDIF}
           break;
          end;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
    if not SessionFound then
     begin // First connect
// new ClientSession
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New ClientSession');
{$ENDIF}
  ConnectionFound := False;
  Connections := FManager.FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.Network = FNetwork then
       begin
        ConnectionFound := True;
        break;
       end;
    end;
  finally
   FManager.FConnections.UnlockList;
  end;
  if not ConnectionFound then
    goto KillPacket;
      New(ClientSession);
      ClientSession.AnswerTime := 0;
      ClientSession.Session := nil;
      FManager.FSessions.Add(ClientSession);
      ClientSession.AnswerStatus := ACRNo;
      ClientSession.Status := ACRVacant;
      ClientSession.RemoteConnectionID := Header.ConnectionID;
      ClientSession.ConnectionID := ClientConnection.ConnectionID;
      ClientSession.ServerSessionID := Header.SessionID;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ClientSession),6)+', Client SessionID='+IntToStr(ClientSession.ServerSessionID));
{$ENDIF}
//      ClientSession.Thread := nil;
//      ClientSession.MsgThread := nil;
      ClientSession.ListeningThreads := TThreadList.Create;
      ClientSession.ControlCode := ACRExecute;
//      ClientSession.ListeningThreads.Add(self);
      ClientSession.MsgListeningThreads := TThreadList.Create;
      ClientSession.MsgControlCode := ACRExecute;
// Create PacketIDsToResend
      ClientSession.PacketIDsToResend := TACRThreadIntArray.Create;
      ClientSession.MsgPacketIDsToResend := TACRThreadIntArray.Create;
      ClientSession.ResendRequestThread := nil;
      ClientSession.MsgResendRequestThread := nil;
// Create new packets list
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packets list create');
{$ENDIF}
      ClientSession.Packets := TList.Create;
      ClientSession.Packets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
      ClientSession.MsgPackets := TList.Create;
      ClientSession.MsgPackets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
// add new session to list
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': add new session to list');
{$ENDIF}
      SessionFound := True;
     end
    else // not first connect - reset and resend connection info
     begin
      // Is buffer receiving now?
      if (ClientSession.AnswerStatus = ACRStart)
      or (ClientSession.AnswerStatus = ACRFull)
      or (ClientSession.Session = nil) // just added, not finished
      or (ClientSession.Session.SessionID = INVALID_SESSION_ID) // buffer received, not finished
      then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer is receiving now');
{$ENDIF}
        goto KillPacket;
       end;
      if (ClientSession.ControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Not Execute - Terminate ListeningThread, ClientSession.ControlCode='+IntToStr(ClientSession.ControlCode));
{$ENDIF}
        goto KillPacket;
       end;
      if (ClientSession.CurrentRequestID > Header.CurrentRequestID) then // packet from old connection request
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - SendAckn');
{$ENDIF}
//        Header.Recepient := Header.Sender;
        if ClientSession.Session<>nil then
          FManager.SendConnectAckn(ClientSession, ClientConnection, Header.CurrentRequestID);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - Terminate ListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
//      ClientSession.MsgThread := nil;
//      ClientSession.ListeningThreads.Add(self);
// Set PacketIDsToResend
//      ClientSession.PacketIDsToResend.Lock;
      ClientSession.PacketIDsToResend.SetSize(0);
//      ClientSession.PacketIDsToResend.Unlock;
//      ClientSession.MsgPacketIDsToResend.Lock;
      ClientSession.MsgPacketIDsToResend.SetSize(0);
//      ClientSession.MsgPacketIDsToResend.Unlock;
(*
// Delete old packets from list
      Packets := ClientSession.Packets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ClientSession.Packets.UnlockList;
      end;
      Packets := ClientSession.MsgPackets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ClientSession.Packets.UnlockList;
      end;
*)
     end;
// Put this first packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New Packet');
{$ENDIF}
    New(Packet);
    Packet.Buffer := Buffer;
    Packet.BufferSize := BufferSize;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufferSize='+IntToStr(BufferSize));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet add');
{$ENDIF}
    ClientSession2.Packets.Add(Packet);
    ClientSession.CurrentRequestID := Header.CurrentRequestID;
    ClientSession.ClientMessageID := 0;
    ClientSession.ServerMessageID := 0;
    Session.FConnectParams.RemoteHost := FromHost;
    Session.FConnectParams.RemotePort := FromPort;
    EnterCriticalSection(FManager.FCSect);
    Session.FConnectParams.ServerID := Header.Sender; // Allow following communication
    LeaveCriticalSection(FManager.FCSect);
// Get connection parameters
    if Header.ControlCode=ACRConnect+ACRLastPacket then
     ClientSession2.AnswerStatus := ACRFull;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSessionThread.Create');
{$ENDIF}
    FManager.ReceiveBuffer(ClientSession2.Session, Buf, BufSize);
    ClientSession.Session := Session;
    ClientSession.ListeningThreads.Add(self);
    UserID := TACRClientSession(Session).UserID;
    Move(Buf^, UserID, SizeOf(UserID));
    inc(Buf, SizeOf(UserID));
    Move(PChar(Buf + SizeOf(TACRConnectionParams))^, AuBufSize, SizeOf(AuBufSize));
    if not FManager.IsAuthorizationBufferValid(
              Session.ConnectParams.CryptoInfo,
              PChar(Buf + SizeOf(TACRConnectionParams) + SizeOf(AuBufSize)),
              AuBufSize
                                               )
    then
     begin
      FManager.SendDisconnectRequest(ClientSession, False);
      FManager.FSessions.Remove(ClientSession);
      FManager.DeleteSession(ClientSession);
      Exit;
     end;
    // Get ConnectParams
    ConnectionParams := PACRConnectionParams(Buf);
    // Get client Application name
    SetLength(Application, BufSize-SizeOf(TACRConnectionParams)
      -SizeOf(AuBufSize)-AuBufSize-1-SizeOf(UserID));
    StrCopy(PChar(Application), Buf+SizeOf(TACRConnectionParams)+SizeOf(AuBufSize)+AuBufSize);
    // Free
    dec(Buf, SizeOf(UserID));
    ClientSession.Session := Session.ConnectUser(UserID, FromHost, FromPort);
    if ClientSession.Session = nil then
     begin
      ClientSession.Session := Session;
      FManager.SendDisconnectRequest(ClientSession, False);
      FManager.FSessions.Remove(ClientSession);
      FManager.DeleteSession(ClientSession);
      Exit;
     end
    else
     begin
      ClientSession.AnswerStatus := ACRStart;
//      ClientSession.Session.FConnectParams.CryptoInfo := TACRServer(FManager.FServer).CryptoParams.GetCryptoParams;
// Set SessionID
      EnterCriticalSection(FManager.FCSect);
      ClientSession.Session.FConnectParams.RemoteHost := FromHost;
      ClientSession.Session.FConnectParams.RemotePort := FromPort;
      ClientSession.Session.FConnectParams.ServerID := Session.FConnectParams.ServerID;
      ClientSession.Session.SessionID := FManager.FSessionID;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ClientSession),6)+', Server SessionID='+IntToStr(ClientSession.Session.SessionID));
{$ENDIF}
      dec(FManager.FSessionID);
      if FManager.FSessionID = INVALID_SESSION_ID then
        dec(FManager.FSessionID);
      LeaveCriticalSection(FManager.FCSect);
      ClientSession.Session.FConnectParams.PacketSize := ConnectionParams.PacketSize;
      ClientSession.Session.FConnectParams.CompressionAlgorithm := ConnectionParams.CompressionAlgorithm;
      ClientSession.Session.FConnectParams.CompressionMode := ConnectionParams.CompressionMode;
//      Header.Recepient := Header.Sender;
      FManager.SendConnectAckn(ClientSession, ClientConnection);
     end;
    MemoryManager.FreeAndNilMem(Buf);
    Exit;
   end; // Connect
{$ENDIF MsgCommunicator}
  Header.Recepient := Header.Sender;
// verify ConnectionID existing
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('           Header.PacketID         = '+IntToStr(Header.PacketID));
aaWriteToLog('           Header.ControlCode      = '+IntToStr(Header.ControlCode));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connections');
{$ENDIF}
  ConnectionFound := False;
  Connections := FManager.FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.ConnectionID = Header.ConnectionID then
       begin
        ConnectionFound := True;
        break;
       end;
    end;
  finally
   FManager.FConnections.UnlockList;
  end;
  if not ConnectionFound then
    goto KillPacket;
// verify SessionID in this connection
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sessions');
{$ENDIF}
  SessionFound := False;
  Sessions := FManager.FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':');
aaWriteToLog('           Sessions.Count='+IntToStr(Sessions.Count));
aaWriteToLog('           i='+IntToStr(i));
{$ENDIF}
     ClientSession := Sessions.Items[i];
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           @ClientSession='+IntToStr(Integer(ClientSession)));
aaWriteToLog('           ClientSession.ConnectionID='+IntToStr(Integer(ClientSession.ConnectionID)));
aaWriteToLog('           Header.ConnectionID='+IntToStr(Integer(Header.ConnectionID)));
aaWriteToLog('           ClientSession.ServerSessionID='+IntToStr(Integer(ClientSession.ServerSessionID)));
aaWriteToLog('           ClientSession.Session.SessionID='+IntToStr(Integer(ClientSession.Session.SessionID)));
aaWriteToLog('           Header.ServerSessionID='+IntToStr(Integer(Header.SessionID)));
{$ENDIF}
{
     if (ClientSession.ConnectionID = Header.ConnectionID)
     and (
               (ClientSession.ServerSessionID = Header.SessionID)
            or (Header.ControlCode = ACRConnected)
            or (Header.ControlCode = ACRConnected+ACRLastPacket)
          )
}
     if (ClientSession.ConnectionID = Header.ConnectionID)
     and (
          (ClientSession.ServerSessionID = Header.SessionID)
{
          or
          (
            (ClientSession.ServerSessionID = INVALID_SESSION_ID)
            and
            (
             (Header.ControlCode = ACRConnected)
             or
             (Header.ControlCode = ACRConnected+ACRLastPacket)
            ) // and
          ) // or
}
         ) // and
     then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FManager.FSessions.UnlockList;
  end;
{$IFDEF LOG_CLIENT_THREADS}
if SessionFound then
aaWriteToLog('           Session Found')
else
aaWriteToLog('           Session Not Found')
;
{$ENDIF}
  if not SessionFound then
    goto KillPacket;

  // is listener stoped?
  if (Header.ControlCode = ACRMessage)
  or (Header.ControlCode = ACRMessage + ACRLastPacket)
  then
    if ClientSession.MsgControlCode <> ACRExecute then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Messages Listeninig blocked');
{$ENDIF}
      goto KillPacket;
     end
  else
    if ClientSession.ControlCode <> ACRExecute then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Commands Listeninig blocked');
{$ENDIF}
      goto KillPacket;
     end;
{
  EnterCriticalSection(FManager.FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FManager.FCSect);
}

// verify Sender
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sender');
{$ENDIF}
  if ClientSession.Session.ConnectParams.ServerID <> Header.Sender then
   begin
{$IFDEF MsgCommunicator}
    if (ClientSession.Session.ConnectParams.ServerID = Integer(MSG_INVALID_USER_ID)) // new direct session
    and ((Header.ControlCode = ACRConnected)                // just connected
      or (Header.ControlCode = ACRConnected+ACRLastPacket))
    then
     begin
      ClientSession.Session.FConnectParams.ServerID := Header.Sender;
     end
    else
{$ENDIF}
      goto KillPacket;
   end;
// check CurrentRequestID
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
{$ENDIF}
  if (Header.ControlCode = ACRMessageReceived)
  or (Header.ControlCode = ACRMessagePacketResendRequest)
  then
   begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ClientMessageID = '+IntToStr(ClientSession.ClientMessageID));
{$ENDIF}
    if ClientSession.ClientMessageID <> Header.CurrentRequestID then
      goto KillPacket;
   end
  else
  if (Header.ControlCode = ACRMessage)
  or (Header.ControlCode = (ACRMessage + ACRLastPacket))
  then
   begin
    if ClientSession.ServerMessageID <> Header.CurrentRequestID then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID = '+IntToStr(ClientSession.ServerMessageID));
{$ENDIF}
      if ClientSession.ServerMessageID > Header.CurrentRequestID then // old message
        FManager.SendAcknowledgement(ClientSession,True,Header.CurrentRequestID);
      goto KillPacket;
     end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Message Listening Thread in '+IntToStr(Integer(ClientSession)));
{$ENDIF}
    if ClientSession.MsgListeningThreads = nil then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': MsgListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
      ClientSession.MsgListeningThreads.Add(self);
   end
  else
   begin
    if Header.ControlCode<>ACRDisconnect then
    if Header.ControlCode <> ACRPing then
     if ClientSession.CurrentRequestID <> Header.CurrentRequestID then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.CurrentRequestID = '+IntToStr(ClientSession.CurrentRequestID));
{$ENDIF}
       if ClientSession.CurrentRequestID > Header.CurrentRequestID then // old comand answer
         FManager.SendAcknowledgement(ClientSession,False,Header.CurrentRequestID);
       goto KillPacket;
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Listening Thread in '+IntToStr(Integer(ClientSession)));
{$ENDIF}
    if ClientSession.ListeningThreads = nil then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
      ClientSession.ListeningThreads.Add(self);
   end;
// Process service packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Process service packet');
{$ENDIF}
  if (Header.ControlCode = ACRConnected)
  or (Header.ControlCode = (ACRConnected+ACRLastPacket))
    then ClientSession.ServerSessionID := Header.SessionID; // To allow disconnect immediately
  AllPacketsReceived := False;
  case Header.ControlCode of
   ACRDisconnect:
    begin
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog('Listener Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRDisconnect');
{$ENDIF}
     if Header.SessionID = INVALID_SESSION_ID then
       goto KillPacket;
     inc(FManager.FLiveSessions);
     ClientSession.ControlCode := ACRTerminate;
     sleep(0);
     ClientSession.CurrentRequestID := Header.CurrentRequestID;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendACKN');
{$ENDIF}
     FManager.SendAcknowledgement(ClientSession);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session.OnDisconnect');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect);
     try
      if ClientSession.Session <> nil then
        ClientSession.Session.OnDisconnect;
     finally
      LeaveCriticalSection(FManager.FCSect);
     end;
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TerminateAllSessionThreads');
{$ENDIF}
     if FManager.FListenerStoped then
       goto KillPacket;
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
}
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': FSessions.Remove...');
{$ENDIF}
     if not FManager.FListenerStoped then
       if FManager.FSessions <> nil then
         FManager.FSessions.Remove(ClientSession);
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DeleteSession...');
{$ENDIF}
     if not FManager.FListenerStoped then
       FManager.DeleteSession(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT DISCONNECTED FROM SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
{$ENDIF}
     if not FManager.FListenerStoped then
       dec(FManager.FLiveSessions);
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
     goto KillPacket;
}
{$IFDEF DEBUG_LOG_CLIENT_DISCONNECT}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': End of Disconnect');
{$ENDIF}
    end;
   ACRPing:
    begin
     FManager.SendPing(ClientSession);
     goto KillPacket;
    end;
   ACRPacketResendRequest:
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRPacketResendRequest');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect);
     ClientSession.PacketIDsToResend.Add(Header.PacketID);
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received Command Packet Resend Request ID='+IntToStr(Header.PacketID));
{$ENDIF}
     goto KillPacket;
    end;
   ACRMessagePacketResendRequest:
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRMessagePacketResendRequest');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect);
     ClientSession.MsgPacketIDsToResend.Add(Header.PacketID);
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received Message Packet Resend Request ID='+IntToStr(Header.PacketID));
{$ENDIF}
     goto KillPacket;
    end;
   ACRAllPacketsReceived:
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRAllPacketsReceived');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect);
     ClientSession.SendStatus := ACRSent;
     ClientSession.PacketIDsToResend.SetSize(0);
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received Command Acknowledgement');
{$ENDIF}
     sleep(0);
     goto KillPacket;
    end;
   ACRMessageReceived:
    begin
     EnterCriticalSection(FManager.FCSect);
     ClientSession.MsgSendStatus := ACRSent;
     ClientSession.MsgPacketIDsToResend.SetSize(0);
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received Message Acknowledgement');
{$ENDIF}
     goto KillPacket;
    end;
   ACRMessage,
   (ACRMessage+ACRLastPacket):
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': message packet');
{$ENDIF}
     CurrentRequestID := Header.CurrentRequestID;
//     EnterCriticalSection(FManager.FCSect);
     if ClientSession.ServerMessageID = Header.CurrentRequestID then
     if (ClientSession.MsgReceiveStatus = ACRFull) then // lost acknowledgement
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
       FManager.SendAcknowledgement(ClientSession, True);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACKN Sent');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet while the answer received fully');
{$ENDIF}
//       LeaveCriticalSection(FManager.FCSect);
       MemoryManager.FreeAndNilMem(Buffer);
       goto IsMsgFull;
//       goto KillPacket; // Do not replace correct packets with doubles
      end;
//     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Enter...');
{$ENDIF}
      EnterCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Entered!');
{$ENDIF}
// Is this packet already stored?
      if ClientSession.MsgPackets.List <> nil then // list exists
       if ((Header.PacketID+1) <= ClientSession.MsgPackets.Count) then // old packet
       if (ClientSession.MsgPackets.Items[Header.PacketID] <> nil) then // already received
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.MsgPackets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
         MemoryManager.FreeAndNilMem(Buffer);
         goto IsMsgFull;
//         goto KillPacket; // Do not replace correct packets with doubles
        end;
// Add correct received data packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New packet');
if ((Header.PacketID+1) <= ClientSession.MsgPackets.Count) then // old packet
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.MsgPackets.Items[Header.PacketID])));
{$ENDIF}
     if ClientSession.MsgPackets = nil then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets = nil');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
     if Terminated then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
// Add packet
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
     if ClientSession.MsgPackets.Count < (Header.PacketID + 1) then // New Packet - Allocate
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Allocate');
aaWriteToLog('           MsgPackets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
{$ENDIF}
       ClientSession.MsgPackets.Count := (Header.PacketID + 1); // List fills hole by Nils
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           MsgPackets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add packet');
{$ENDIF}
     ClientSession.MsgPackets.Items[Header.PacketID] := Packet;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           MsgPackets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': packet added');
{$ENDIF}
{
     if Terminated then
      begin
        Dispose(Packet);
        LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
      end;
}
// Are all packets received?
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.MsgPackets.Items[Header.PacketID])));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all packets received?');
{$ENDIF}
     if
//     (ClientSession.MsgPackets.Count > (Header.PacketID + 1)) or // repeated packet
     (Header.ControlCode >= ACRLastPacket)
     then // Last packet
      begin
       AllPacketsReceived := True;
       for i:=0 to ClientSession.MsgPackets.Count-1 do
        if ClientSession.MsgPackets.Items[i] = nil then
         begin
          AllPacketsReceived := False;
          break;
         end;
      end;
     if AllPacketsReceived then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
       ClientSession.MsgReceiveStatus := ACRFull;  // Allow to extract buffer
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
      end
     else
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
       ClientSession.MsgReceiveStatus := ACRStart;
      end;
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Left!');
{$ENDIF}
// Is it the last packet?
     if Header.ControlCode >= ACRLastPacket then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       Header.ControlCode := Header.ControlCode - ACRLastPacket;
       if ClientSession.MsgReceiveStatus <> ACRFull then // Last packet received, but answer is not full
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Start resending thread');
{$ENDIF}
         if ClientSession.MsgResendRequestThread = nil then
           ClientSession.MsgResendRequestThread := TACRClientMsgResendRequestThread.Create(FManager, ClientSession, Buffer, ClientConnection.Network, FromHost, FromPort);
        end; // last packet - answer is not full
      end;
     if Terminated then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated!');
{$ENDIF}
        Exit;
      end;
// Is Answer full?
 IsMsgFull:
      if ClientSession.MsgReceiveStatus = ACRFull then
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full answer');
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
         if ClientSession.MsgResendRequestThread <> nil then
           ClientSession.MsgResendRequestThread.Terminate;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - start');
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
         FManager.SendAcknowledgement(ClientSession, True, CurrentRequestID);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - finish');
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
         repeat
          if ClientSession.MsgResendRequestThread = nil then
            break
          else
            ClientSession.MsgResendRequestThread.Terminate;
          sleep(0);
         until False;
     if Terminated then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated!');
{$ENDIF}
        Exit;
      end;
     if ClientSession.MsgControlCode<>ACRExecute then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Paused or Terminated!');
{$ENDIF}
        Exit;
      end;
     if ClientSession.ServerMessageID<>CurrentRequestID then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Already received!');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Header.CurrentRequestID       ='+IntToStr(Header.CurrentRequestID));
{$ENDIF}
        Exit;
      end;
         FManager.ReceiveMessage(ClientSession, Buf, BufSize);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DecompressAndDecryptBuffer');
{$ENDIF}
         FManager.DecompressAndDecryptBuffer(ClientSession.Session, Buf, BufSize);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT: Call ClientSession.ReceiveMessage');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
         TACRNetworkSession(ClientSession.Session).ReceiveMessage(Buf, BufSize);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(ClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(ClientSession.MsgReceiveStatus));
{$ENDIF}
        end; // answer is full
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End data packet process
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
}
     Exit;
    end; // message packet process
   ACRNewRequest,
   ACRConnected,
   ACRLastPacket..(ACRMessage+ACRLastPacket-1),
   (ACRMessage+ACRLastPacket+1)..127,
   ACRNoAction:
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Data packet');
{$ENDIF}
     if (ClientSession.CurrentRequestID = Header.CurrentRequestID) then
     if (ClientSession.AnswerStatus = ACRFull) then // lost acknowledgement
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Send ACKN');
{$ENDIF}
       FManager.SendAcknowledgement(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet while the answer received fully');
{$ENDIF}
       goto KillPacket; // Do not replace correct packets with doubles
      end;
     if (ClientSession.CurrentRequestID > Header.CurrentRequestID) then // old request
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Send ACKN');
{$ENDIF}
       FManager.SendAcknowledgement(ClientSession, False, Header.CurrentRequestID);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet while the answer received fully');
{$ENDIF}
       goto KillPacket; // Do not replace correct packets with doubles
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Enter...');
{$ENDIF}
      EnterCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Entered!');
{$ENDIF}
// Is this packet already stored?
      if ClientSession.Packets.List <> nil then // list exists
       if ((Header.PacketID+1) <= ClientSession.Packets.Count) then // old packet
       if (ClientSession.Packets.Items[Header.PacketID] <> nil) then // already received
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.Packets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
         goto KillPacket; // Do not replace correct packets with doubles
        end;
// Add correct received data packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New packet');
if ((Header.PacketID+1) <= ClientSession.Packets.Count) then // old packet
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.Packets.Items[Header.PacketID])));
{$ENDIF}
     if ClientSession.Packets = nil then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets = nil');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
     if Terminated then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
// Add packet
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufferSize='+IntToStr(BufferSize));
{$ENDIF}
     if ClientSession.Packets.Count < (Header.PacketID + 1) then // New Packet - Allocate
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Allocate');
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
{$ENDIF}
       ClientSession.Packets.Count := (Header.PacketID + 1); // List fills hole by Nils
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add packet');
{$ENDIF}
     ClientSession.Packets.Items[Header.PacketID] := Packet;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': packet added');
{$ENDIF}
{
    if Terminated then
      begin
        Dispose(Packet);
        LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
      end;
}
// Are all packets received?
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.Packets.Items[Header.PacketID])));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all packets received?');
{$ENDIF}
     if
//     (ClientSession.Packets.Count > (Header.PacketID + 1)) or // repeated packet
     (Header.ControlCode >= ACRLastPacket)
     then // Last packet
      begin
       AllPacketsReceived := True;
       for i:=0 to ClientSession.Packets.Count-1 do
        if ClientSession.Packets.Items[i] = nil then
         begin
          AllPacketsReceived := False;
          break;
         end;
      end;
     if AllPacketsReceived then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
       ClientSession.AnswerStatus := ACRFull  // Allow to extract buffer
      end
     else
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
       ClientSession.AnswerStatus := ACRStart;
      end;
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Left!');
{$ENDIF}
// Is it the last packet?
     if Header.ControlCode >= ACRLastPacket then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       Header.ControlCode := Header.ControlCode - ACRLastPacket;
       if ClientSession.AnswerStatus <> ACRFull then // Last packet received, but answer is not full
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Start resending thread');
{$ENDIF}
         if ClientSession.ResendRequestThread = nil then
           ClientSession.ResendRequestThread := TACRClientResendRequestThread.Create(FManager, ClientSession, Buffer, ClientConnection.Network, FromHost, FromPort);
        end; // last packet - answer is not full
      end;
// Is Answer full?
       if ClientSession.AnswerStatus = ACRFull then
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full answer');
{$ENDIF}
         if ClientSession.ResendRequestThread <> nil then
          ClientSession.ResendRequestThread.Terminate;
// To allow GetAnswer=Connected while Connect procedure
         if (Header.ControlCode = ACRConnected)
          then
           begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRConnected');
{$ENDIF}
            ClientSession.SendStatus := ACRSent;
            sleep(0); // TRY TO REMOVE!!!
{$IFDEF MsgCommunicator}
            if TACRClientSession(ClientSession.Session).Direct then
              ClientSession.Session.ConnectedUser(
                TACRClientSession(ClientSession.Session).RemoteUser.UserID,
                FromHost, FromPort
                                                );
 {$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ConnectedUser!');
 {$ENDIF}
{$ENDIF}
           end
          else
           begin
{
            if Terminated then
             begin
              Dispose(Packet);
//              LeaveCriticalSection(FManager.FCSect);
              goto KillPacket;
             end;
}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - start');
{$ENDIF}
            FManager.SendAcknowledgement(ClientSession);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - finish');
{$ENDIF}
           end;
         if (Header.ControlCode = ACREcho)
         or (Header.ControlCode = ACRTunning)
         then
          begin
           FManager.ReceiveBuffer(ClientSession.Session,Buf,BufSize);
           FManager.SendBuffer(ClientSession.Session,Buf,BufSize,ACRNoAction);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT SENT ECHO TO SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
           MemoryManager.FreeAndNilMem(Buf);
          end;
        end; // answer is full
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End data packet process
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
}
     Exit;
    end // data packet process
   else
     raise EACRException.Create(40022, ErrorRUnknownControlCode, [Header.ControlCode]);
     goto KillPacket;
   end; // packet process
KillPacket:
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': KillPacket');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT KILLED PACKET');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
  MemoryManager.FreeAndNilMem(Buffer);
{
  EnterCriticalSection(FManager.FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FManager.FCSect);
}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
finally
//LeaveCriticalSection(FManager.FCSect);
end;

 except
  on E: Exception do
   begin
    if (ClientSession = nil)
    or (ClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         ClientSession := Sessions.Items[0];
         if ClientSession.Session <> nil then
           TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrClientListenerThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(ClientSession.Session).DoOnError(
                  AcrClientListenerThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message);
    exit;
   end;
 end;
end;// Execute



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientResendRequestThread.Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                                            );
var
  Error:        String;
  Header:       PACRPacketHeader;
  Sessions:     TList;
begin
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESEND REQUEST THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 try
  FManager := Manager;
  FManager.IncThreadCount;
  FClientSession := ClientSession;
  FBuffer := MemoryManager.AllocMem(SizeOf(TACRPacketHeader));
  Move(Buffer^,FBuffer^,SizeOf(TACRPacketHeader));
  Header := PACRPacketHeader(FBuffer);
  Header.Recepient := FClientSession.Session.ConnectParams.ServerID;
  FNetwork := Network;
  FHost := FromHost;
  FPort := FromPort;
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESEND REQUEST THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - STARTED');
{$ENDIF}
 except
  on E: Exception do
   begin
    if (ClientSession = nil)
    or (ClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         ClientSession := Sessions.Items[0];
         if ClientSession.Session <> nil then
           TACRNetworkSession(ClientSession.Session).DoOnError(
                  ACRClientResendRequestThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(ClientSession.Session).DoOnError(
                  ACRClientResendRequestThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientResendRequestThread.Destroy;
var
  Error:        String;
  Sessions:     TList;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT ResendRequest THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  MemoryManager.FreeAndNilMem(FBuffer);
  EnterCriticalSection(FManager.FCSect);
  if FClientSession.ResendRequestThread <> nil then
    FClientSession.ResendRequestThread := nil;
  LeaveCriticalSection(FManager.FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT ResendRequest THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
 except
  on E: Exception do
   begin
    if (FClientSession = nil)
    or (FClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         FClientSession := Sessions.Items[0];
         if FClientSession.Session <> nil then
           TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else 
      TACRNetworkSession(FClientSession.Session).DoOnError( 
                  ACRClientMsgResendRequestThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+ 
                  ErrorRDestroy+ 
                  E.Message);
   end;
 end; 
end; // Destroy
 
	
//------------------------------------------------------------------------------ 
// Execute
//------------------------------------------------------------------------------ 
procedure TACRClientResendRequestThread.Execute;
var 
  Error:              String; 
  i, j, Count:        Integer;
  AllPacketsReceived: Boolean; 
  Sessions:           TList; 
begin 
 try 
{$IFDEF DEBUG_LOG_NETWORK_RESENDING} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** CLIENT RESEND REQUEST START *****');
{$ENDIF}
  repeat
   AllPacketsReceived := True;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRClientResendRequestThread'); 
{$ENDIF}
   j := 0; 
   Count := FClientSession.Packets.Count;
   for i:=0 to Count-1 do 
    begin 
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i)); 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF} 
     if Terminated then 
      begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count)); 
{$ENDIF} 
       Exit; 
      end; 
     EnterCriticalSection(FManager.FCSect); 
     if FClientSession.Packets <> nil then 
     if FClientSession.Packets.Count > i then 
     if FClientSession.Packets.Items[i] = nil then
      begin
       LeaveCriticalSection(FManager.FCSect); 
       AllPacketsReceived := False;
       inc(j);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketResendRequest, i='+IntToStr(i));
{$ENDIF}
       FManager.PacketResendRequest(FBuffer, FNetwork, FHost, FPort, i); 
{$IFDEF DEBUG_LOG_NETWORK_RESENDING} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been requested'); 
{$ENDIF}
       Sleep(FClientSession.Session.ConnectParams.ReceiveSleep);
       EnterCriticalSection(FManager.FCSect); 
      end;
     LeaveCriticalSection(FManager.FCSect);
    end; 
   if AllPacketsReceived then 
    begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF} 
     FClientSession.AnswerStatus := ACRFull; 
//     FManager.SendAcknowledgement(FClientSession); 
    end; 
   if FClientSession.Session.ConnectParams.ConnectionParamsTunning then
     Sleep((FClientSession.AnswerTime div 16) * j) 
   else 
     Sleep(FClientSession.Session.ConnectParams.RequestDelay); 
  until FClientSession.AnswerStatus = ACRFull;
 except
  on E: Exception do
   begin 
    if (FClientSession = nil)
    or (FClientSession.Session = nil)
    then
     begin 
      Error := 
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try 
       if Sessions.Count > 0 then
        begin
         FClientSession := Sessions.Items[0]; 
         if FClientSession.Session <> nil then
           TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientResendRequestThread,-1, 
                  Error) 
         else 
          begin 
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('=================================================================='); 
aaWriteToLog(Error); 
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else 
        begin 
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error); 
aaWriteToLog('=================================================================='); 
{$ENDIF} 
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else 
      TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientResendRequestThread,-1, 
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message);
   end;
 end;
end;// Execute


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientMsgResendRequestThread.Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                                            );
var
  Error:        String;
  Header:       PACRPacketHeader;
  Sessions:     TList;
begin
 try
  FManager := Manager;
  FManager.IncThreadCount;
  FClientSession := ClientSession;
  FBuffer := MemoryManager.AllocMem(SizeOf(TACRPacketHeader));
  Move(Buffer^,FBuffer^,SizeOf(TACRPacketHeader));
  Header := PACRPacketHeader(FBuffer);
  Header.Recepient := FClientSession.Session.ConnectParams.ServerID;
  FNetwork := Network;
  FHost := FromHost;
  FPort := FromPort;
  inherited Create(False);
  Priority := tpNormal; //tpLowest;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    if (ClientSession = nil)
    or (ClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         ClientSession := Sessions.Items[0];
         if ClientSession.Session <> nil then
           TACRNetworkSession(ClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(ClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientMsgResendRequestThread.Destroy;
var
  Error:        String;
  Sessions:     TList;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  MemoryManager.FreeAndNilMem(FBuffer);
  EnterCriticalSection(FManager.FCSect);
  if FClientSession.MsgResendRequestThread <> nil then
    FClientSession.MsgResendRequestThread := nil;
  LeaveCriticalSection(FManager.FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESENDING THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
 except
  on E: Exception do
   begin
    if (FClientSession = nil)
    or (FClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         FClientSession := Sessions.Items[0];
         if FClientSession.Session <> nil then
           TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message);
   end;
 end;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRClientMsgResendRequestThread.Execute;
var
  Error:              String;
  i, j, Count:        Integer;
  AllPacketsReceived: Boolean;
  Sessions:           TList;
begin
 try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** CLIENT RESEND REQUEST START *****');
{$ENDIF}
  repeat
   AllPacketsReceived := True;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRClientMsgResendingThread');
{$ENDIF}
   j := 0;
   Count := FClientSession.MsgPackets.Count;
   for i:=0 to Count-1 do
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF}
     if Terminated then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF}
       Exit;
      end;
//     Sleep(0);
     EnterCriticalSection(FManager.FCSect);
     if FClientSession.MsgPackets <> nil then
     if FClientSession.MsgPackets.Count > i then
     if FClientSession.MsgPackets.Items[i] = nil then
      begin
       LeaveCriticalSection(FManager.FCSect);
       AllPacketsReceived := False;
       inc(j);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketResendRequest, i='+IntToStr(i));
{$ENDIF}
       FManager.PacketResendRequest(FBuffer, FNetwork, FHost, FPort, i, True);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been resent');
{$ENDIF}
//       Sleep(ACRRequestDelay);
       Sleep(FClientSession.Session.ConnectParams.ReceiveSleep);
       EnterCriticalSection(FManager.FCSect);
      end;
     LeaveCriticalSection(FManager.FCSect);
    end;
   if AllPacketsReceived then
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
     FClientSession.MsgReceiveStatus := ACRFull;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID ='+IntToStr(FClientSession.ServerMessageID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.MsgReceiveStatus='+IntToStr(FClientSession.MsgReceiveStatus));
{$ENDIF}
//     FManager.SendAcknowledgement(FClientSession, True);
    end;
   if FClientSession.Session.ConnectParams.ConnectionParamsTunning then
     Sleep((FClientSession.AnswerTime div 16) * j)
   else
     Sleep(FClientSession.Session.ConnectParams.RequestDelay);
  until FClientSession.MsgReceiveStatus = ACRFull;
 except
  on E: Exception do
   begin
    if (FClientSession = nil)
    or (FClientSession.Session = nil)
    then
     begin
      Error :=
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  ' Session does not exist. '+
                  E.Message;
      Sessions := FManager.FSessions.LockList;
      try
       if Sessions.Count > 0 then
        begin
         FClientSession := Sessions.Items[0];
         if FClientSession.Session <> nil then
           TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  Error)
         else
          begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
          end;
        end
       else
        begin
{$IFDEF DEBUG_ONERROR}
aaWriteToLog('==================================================================');
aaWriteToLog(Error);
aaWriteToLog('==================================================================');
{$ENDIF}
        end;
      finally
       FManager.FSessions.UnlockList;
      end;
     end
    else
      TACRNetworkSession(FClientSession.Session).DoOnError(
                  ACRClientMsgResendRequestThread,-1,
                  ErrorRClient+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message);
   end;
 end;
end;// Execute

{$ENDIF} // Client



{$IFDEF SERVER_VERSION}

////////////////////////////////////////////////////////////////////////////////
//
// TACRServerConnectionManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerConnectionManager.Create(Server: TComponent);
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Create> @Server='+IntToStr(Integer(FServer)));
{$ENDIF}
  InitializeCriticalSection(FCSect);
  inherited Create;
  FSessions := TThreadList.Create;
  FSessionID := -MAXINT;
  FServer := Server;
  FNetwork := TACRNetwork.Create(self);
  ThreadCount := 1;
  if Server = nil then
   begin
    FNetwork.FLocalClient := 1;
    FNetwork.LocalPort := ACRDefaultServerPort;
   end
  else
   begin
    FNetwork.FLocalClient := TACRServer(FServer).ServerID;
    FNetwork.LocalHost := TACRServer(FServer).LocalHost;
    FNetwork.LocalPort := TACRServer(FServer).LocalPort;
   end;
  FListenerStoped := False;
  ResendRequestThread := TACRServerResendRequestThread.Create(self);
  MsgResendRequestThread := TACRServerMsgResendRequestThread.Create(self);
  SessionTerminator := TACRServerSessionTerminatorThread.Create(self);
  if Server <> nil then
    if TACRServer(FServer).NetworkSettings.PingClients then
      PingClientsThread := TACRServerPingClientsThread.Create(self);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Create> FINISHED');
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerConnectionManager.Destroy;
var
  Error:        String;
  StartTime:    DWord;
  Abnormal:     Boolean;
  Delay:        Integer;
begin
  Abnormal := False;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> @Server='+IntToStr(Integer(FServer)));
{$ENDIF}
  FListenerStoped := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> DisconnectAll...');
{$ENDIF}
  DisconnectAll(False);
// Terminate ResendRequestThread
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> Terminate ResendRequestThread');
{$ENDIF}
{
  if ResendRequestThread <> nil then
    ResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   sleep(0);
   if (ResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ACRServerThreadsTerminateDelay then
       begin
        Error := 'command resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
                                          ['ServerConnectionManager',
                                                                  Error]);
       end;
     end;
  until False;
}
  if not TerminateServerThread(@ResendRequestThread,AcrServerConnectionManager,ErrorRResendRequestThread) then
    Abnormal := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> command resend thread terminated');
{$ENDIF}
  if not TerminateServerThread(@MsgResendRequestThread,AcrServerConnectionManager,ErrorRMsgResendRequestThread) then
    Abnormal := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> message resend thread terminated');
{$ENDIF}
  if not TerminateServerThread(@SessionTerminator,AcrServerConnectionManager,ErrorRServerSessionTerminatorThread) then
    Abnormal := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> session terminator thread terminated');
{$ENDIF}
  if not TerminateServerThread(@PingClientsThread,AcrServerConnectionManager,ErrorRServerPingClientsThread) then
    Abnormal := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> Ping Clients thread terminated');
{$ENDIF}
// Wait for Sessions Threads
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy> Wait for Sessions Threads');
{$ENDIF}
  if FServer = nil then
    Delay := ACRServerThreadsTerminateDelay
  else
    Delay := TACRServer(FServer).NetworkSettings.ServerThreadsTerminateDelay;
  StartTime := GetTickCount;
  while (GetTickCount - StartTime) < Delay do
   begin
    sleep(0);
    if (ThreadCount=1) then
      break;
   end;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy - Delete Sessions and Free Network...');
{$ENDIF}
  if not Abnormal then
    EnterCriticalSection(FCSect);
  if not Abnormal then
    FSessions.Free;
  FSessions := nil;
  if FNetwork <> nil then
   begin
    FNetwork.Free;
    DecThreadCount(Abnormal);
   end;
  FNetwork := nil;
  if not Abnormal then
    LeaveCriticalSection(FCSect);
  DeleteCriticalSection(FCSect);
// Check for threads
  if ThreadCount <> 0 then
    TACRServer(FServer).DoOnConnectionError(
                  AcrServerConnectionManager,40515,
                  ErrorRServer+
                  ErrorRDestroy+
                  ErrorRThreadsLeft+
                  IntToStr(ThreadCount)
                  );
  inherited Destroy;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Destroy - FINISHED');
{$ENDIF}
end;// Destoy


//------------------------------------------------------------------------------
// PacketResendRequest
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.PacketResendRequest(
                               Buffer:        PChar;
                               Network:       TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
var
  Header:          PACRPacketHeader;
begin
  Header := Pointer(Buffer);
  Header.Signature := ACRServerPacketSign;
  if Msg then
    Header.ControlCode := ACRMessagePacketResendRequest
  else
    Header.ControlCode := ACRPacketResendRequest;
  if (PacketID >= 0) then
    Header.PacketID := PacketID
  else
    Header.Recepient := Header.Sender;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('SERVER asks to resend packet # '+IntToStr(Header.PacketID));
{$ENDIF}
  EnterCriticalSection(FCSect);
  try
   Network.RemoteHost := RemoteHost;
   Network.RemotePort := RemotePort;
   Network.SendBuffer(Buffer, SizeOf(TACRPacketHeader));
  finally
   LeaveCriticalSection(FCSect);
  end;
end;// PacketResendRequest


//------------------------------------------------------------------------------
// OnDisconnect
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
{$IFDEF MsgCommunicator}
  SSessions:            array of TACRServerSession;
{$ENDIF}
begin
{ TODO -oAlex : Re-write OnDisconnect to use DoDisconnect. In MsgCommunicator DisconnectUser move to Session.Destroy}

{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.OnDisconnect> START');
aaWriteToLog('OnDisconnect> Broken client: '+FromHost+':'+IntToStr(FromPort));
{$ENDIF}
 try
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> FSessions');
aaWriteToLog(IntToStr(Integer(FSessions)));
 {$ENDIF DEBUG_LOG_SERVER_DISCONNECT}
{$IFDEF MsgCommunicator}
  SetLength(SSessions, 0);
{$ENDIF}
  Sessions:=FSessions.LockList;
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> FSessions Locked!');
 {$ENDIF DEBUG_LOG_SERVER_DISCONNECT}
   try
    for i:=Sessions.Count-1 downto 0 do
     begin
      if (i<0)
      or (i>=Sessions.Count)
      then
        break;
      ServerSession := Sessions.Items[i];
      if (ServerSession.RemoteHost = FromHost)
      and (ServerSession.RemotePort = FromPort)
      then
       begin
(* REMOVED IN ACR --------------------------------------------------------------
{$IFDEF MsgCommunicator}
        SetLength(SSessions, Length(SSessions)+1);
        SSessions[Length(SSessions)-1] := TACRServerSession(ServerSession.Session);
{$ENDIF}
REMOVED IN ACR -------------------------------------------------------------- *)
        EnterCriticalSection(FCSect);
        if (ServerSession.ControlCode <> ACRTerminate)
        or (ServerSession.MsgControlCode <> ACRTerminate)
        then
         begin
          // Terminate sending threads
          ServerSession.ControlCode := ACRTerminate;
          ServerSession.MsgControlCode := ACRTerminate;
          LeaveCriticalSection(FCSect);
          // Block send message
{$IFDEF MsgCommunicator}
          EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
          FSessions.UnlockList;
          // Terminate listening threads
          TerminateAllSessionThreads(ServerSession);
(*
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> Terminate command listening threads...');
{$ENDIF}
          TerminateCommandThreads(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> Terminate message listening threads...');
{$ENDIF}
          TerminateMessageThreads(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> All sessions terminated');
{$ENDIF}
*)
          // delete session from list
          FSessions.Remove(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> Removed SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
{$IFDEF MsgCommunicator}
          LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> Session CS left');
{$ENDIF}
          // free memory and session
          DeleteSession(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> Session deleted');
{$ENDIF}
          Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> Sessions locked');
{$ENDIF}
         end
        else
          LeaveCriticalSection(FCSect);
       end;
     end;
   finally
    FSessions.UnlockList;
   end;

(*
  EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> OLD: FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
{$ENDIF}
  if FNetwork <> nil then
   begin
    FNetwork.Free;
    DecThreadCount;
   end;
  FNetwork := nil;
  FNetwork := TACRNetwork.Create(self);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> NEW: FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
{$ENDIF}
  if FServer = nil then
   begin
    FNetwork.FLocalClient := 1;
    FNetwork.LocalPort := ACRDefaultServerPort;
   end
  else
   begin
    FNetwork.FLocalClient := TACRServer(FServer).ServerID;
    FNetwork.LocalHost := TACRServer(FServer).LocalHost;
    FNetwork.LocalPort := TACRServer(FServer).LocalPort;
   end;
  LeaveCriticalSection(FCSect);
*)

{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog(IntToStr(GetTickCount)+':');
aaWriteToLog('OnDisconnect> FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
{$ENDIF}

(* REMOVED IN ACR --------------------------------------------------------------
{$IFDEF MsgCommunicator}
  if not FListenerStoped then
   // Send off-line events
   for i:=0 to Length(SSessions)-1 do
    begin
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> DisconnectUser...');
 {$ENDIF}
     SSessions[i].DisconnectUser;
{
     ServerSession := Sessions.Items[0];
     TACRServerSession(ServerSession.Session).DisconnectUser(Users.Items[i]);
}
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> DisconnectUser finished');
 {$ENDIF}
    end;
{$ENDIF}
--------------------------------------------------------------- REMOVED IN ACR*)
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('OnDisconnect> FINISH');
{$ENDIF DEBUG_LOG_SERVER_DISCONNECT}
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerConnectionManager.OnDisconnect - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
end; // OnDisconnect


//------------------------------------------------------------------------------
// NetworkListener
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
var
  StartTime:        DWORD;
begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.NetworkListener> START - '+IntToStr(GetTickCount));
{$ENDIF}
  if not FListenerStoped then
   begin
    StartTime := GetTickCount;
    while ((GetTickCount-StartTime) < TACRServer(FServer).NetworkSettings.ServerReceiveTimeOut) do
     begin
      if ThreadCount<TACRServer(FServer).NetworkSettings.MaxThreadCount then
       begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.NetworkListener> Create Listener Thread... - '+IntToStr(GetTickCount));
{$ENDIF}
        TACRServerListenerThread.Create(self, Buffer, BufferSize, FromHost, FromPort);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.NetworkListener> Listener Thread Created ! - '+IntToStr(GetTickCount));
{$ENDIF}
        Exit;
       end;
      sleep(0);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.NetworkListener> Sleeped ! - '+IntToStr(GetTickCount));
{$ENDIF}
     end;
    TACRServer(FServer).DoOnConnectionError(
                  AcrServerListenerThread,40516,
                  ErrorRCannotCreateListenerThread+
                  IntToStr(TACRServer(FServer).NetworkSettings.MaxThreadCount));
   end;
  MemoryManager.FreeAndNilMem(Buffer);
end;// NetworkListener


//------------------------------------------------------------------------------
// WaitForThreadFinish
//------------------------------------------------------------------------------
function TACRServerConnectionManager.TerminateServerThread(
                          Thread:        PThread;
                          ErrProcess:    Integer;
                          ErrObject:     String;
                          WaitTimeOut:   DWORD = ACRServerThreadsTerminateDelay
                                      ): Boolean;
var
  Error:          String;
begin
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> START');
{$ENDIF}
 try
  Result := False;
  if Thread^ = nil then
   begin
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Thread is not started - OK');
{$ENDIF}
    Result := True;
    Exit;
   end;
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Ask to terminate Thread...');
{$ENDIF}
  // AskToTerminate
  Thread^.Terminate;
  WaitForThread(Thread, WaitTimeOut);
  if Thread^ = nil then
   begin
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Terminated!');
{$ENDIF}
    Result := True;
    Exit;
   end
  else
   begin // ForceToTerminate
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Not Terminated!');
{$ENDIF}
    Error:= ErrorRServer+ErrObject+IntToStr(Integer(Thread^.ThreadID))+'/'+IntToStr(Integer(Thread^.Handle))+ErrorRExecute;
    TACRServer(FServer).DoOnConnectionError(ErrProcess,40517,
                    Error+ErrorRCannotTerminateThread+IntToStr(WaitTimeOut));
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Force to terminate Thread...');
{$ENDIF}
    Result := TerminateThread(Thread^.Handle, 0);
    if not Result then
     begin
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Not Terminated! - TerminateThread failed');
{$ENDIF}
      TACRServer(FServer).DoOnConnectionError(ErrProcess,40518,
                    Error+'TerminateThread failed - '+ErrorRCannotKillThread+IntToStr(GetLastError));
     end;
    if Thread^.Handle <> 0 then
      Result := CloseHandle(Thread^.Handle);
    if Result then
     begin
      if ErrObject = ErrorRListenerThread then
       begin
        if TACRServerListenerThread(Thread^).SessionFound then
          if TACRServerListenerThread(Thread^).ServerSession.ListeningThreads <> nil then
            TACRServerListenerThread(Thread^).ServerSession.ListeningThreads.Remove(Thread^);
       end
      else
      if ErrObject = ErrorRListenerMsgThread then
       begin
        if TACRServerListenerThread(Thread^).SessionFound then
          if TACRServerListenerThread(Thread^).ServerSession.MsgListeningThreads <> nil then
            TACRServerListenerThread(Thread^).ServerSession.MsgListeningThreads.Remove(Thread^);
       end;
      Thread^ := nil;
      DecThreadCount(True);
     end
    else
     begin
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Not Terminated! - CloseHandle failed');
{$ENDIF}
      TACRServer(FServer).DoOnConnectionError(ErrProcess,40518,
                    Error+'CloseHandle failed - '+ErrorRCannotKillThread+IntToStr(GetLastError));
     end;
    Result := False;
   end; // ForceToTerminate
 finally
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateServerThread> Terminated!');
{$ENDIF}
 end;
end;


//------------------------------------------------------------------------------
// WaitForThread
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.WaitForThread(
                          Thread:        PThread;
                          TimeOut:       DWORD
                          );
var
  StartTime:      DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.WaitForThread> TimeOut = '+IntToStr(TimeOut));
{$ENDIF}
  StartTime := GetTickCount;
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.WaitForThread> Started at: '+IntToStr(StartTime));
{$ENDIF}
  while ((GetTickCount-StartTime) < TimeOut) do
   begin
    if Thread^ = nil then
      break;
    sleep(0);
   end;
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
if Thread^ = nil then
aaWriteToLog('TACRServerConnectionManager.WaitForThread> Thread Finished at: '+IntToStr(GetTickCount))
else
aaWriteToLog('TACRServerConnectionManager.WaitForThread> Thread Not Finished! Timeout at: '+IntToStr(GetTickCount));
{$ENDIF}
end; // Wait for thread finish


//------------------------------------------------------------------------------
// WaitForServerSessionThread
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.WaitForServerSessionThread(
                          ServerSession:    PACRSrvrSession;
                          TimeOut:          DWORD = ACRWaitForServerSessionThreadTimeOut
                          );
var
  StartTime:      DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.WaitForServerSessionThread> Started at: '+IntToStr(GetTickCount));
{$ENDIF}
//  WaitForThread(@ServerSession.Thread, TimeOut);
  StartTime := GetTickCount;
  while ((GetTickCount-StartTime) < TimeOut) do
   begin
    if ServerSession.ReceiveStatus = ACRNo then
      break;
    sleep(0);
   end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
//if ServerSession.Thread = nil then
if ServerSession.ReceiveStatus = ACRNo then
aaWriteToLog('TACRServerConnectionManager.WaitForServerSessionThread> Finished at: '+IntToStr(GetTickCount))
else
aaWriteToLog('TACRServerConnectionManager.WaitForServerSessionThread> Not Finished! Timeout at: '+IntToStr(GetTickCount));
{$ENDIF}
end; // Wait for server session thread


//------------------------------------------------------------------------------
// WaitForServerSessionMsgThread
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.WaitForServerSessionMsgThread(
                          ServerSession:    PACRSrvrSession;
                          TimeOut:          DWORD = ACRWaitForServerSessionThreadTimeOut
                          );
begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.WaitForServerSessionThread> Started at: '+IntToStr(GetTickCount));
{$ENDIF}
  WaitForThread(@ServerSession.MsgThread, TimeOut);
{
  StartTime := GetTickCount;
  while ((GetTickCount-StartTime) < TimeOut) do
   begin
    if ServerSession.ReceiveStatus = ACRNo then
      break;
    sleep(0);
   end;
}
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
if ServerSession.Thread = nil then
aaWriteToLog('TACRServerConnectionManager.WaitForServerSessionThread> Finished at: '+IntToStr(GetTickCount))
else
aaWriteToLog('TACRServerConnectionManager.WaitForServerSessionThread> Not Finished! Timeout at: '+IntToStr(GetTickCount));
{$ENDIF}
end; // Wait for server session message thread


//------------------------------------------------------------------------------
// CommandReceived
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.CommandReceived(
                          ServerSession:    PACRSrvrSession;
                          ControlCode:      TACRControlCode;
                          CurrentRequestID: Integer
                                                      );
var
  Error:        String;

function IsNotSingle: Boolean;
begin
  Result := False;
   if (ControlCode <> ACRDisconnect) then
     if (CurrentRequestID < ServerSession.CurrentRequestID) then
      begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> (CurrentRequestID < ServerSession.CurrentRequestID) ==> Exit, old command already processed');
{$ENDIF}
       Result := True;
       Exit;
      end;
   if ServerSession.ReceiveStatus = ACRFull then
    begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> (ReceiveStatus = ACRFull) ==> Exit');
{$ENDIF}
     Result := True;
     Exit;
    end;
   ServerSession.ReceiveStatus := ACRFull;
   if ServerSession.ControlCode = ACRExecute then
     ServerSession.ControlCode := ACRSuspend
   else
    begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> (ControlCode <> ACRExecute) ==> Exit');
{$ENDIF}
     Result := True;
     Exit;
    end;
end;

begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> STARTED - GetTickCount='+IntToStr(GetTickCount));
aaWriteToLog(' CurrentRequestID='+IntToStr(ServerSession.CurrentRequestID));
aaWriteToLog(' Code='+IntToStr(ControlCode));
aaWriteToLog(' ReceiveStatus = '+IntToStr(ServerSession.ReceiveStatus));
{$ENDIF}
  if (ControlCode >= ACRLastPacket) then
    ControlCode := ControlCode - ACRLastPacket;
  ServerSession.Thread.FCode := ControlCode;
  if (CurrentRequestID-1)=ServerSession.CurrentRequestID then // next command already received, could be previous was without answer and still executed...
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> previous session still executes...');
{$ENDIF}
    WaitForServerSessionThread(ServerSession);
   end;
  ServerSession.ReceiveStatus := ACRFull;
//  sleep(0); // NEVER SET IT!!!
  Exit;


// Disconnect - wait for session end
  if (ControlCode = ACRDisconnect) then
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> Disconnect - SendAcknowledgement');
{$ENDIF}
    SendAcknowledgement(ServerSession);
    WaitForServerSessionThread(ServerSession);
   end;

// check for single execution and block other calls
  EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> CS Entered!');
{$ENDIF}
  try
  if IsNotSingle then
    Exit;
  finally
   LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> CS Left!');
{$ENDIF}
  end;
  if (ControlCode <> ACRConnect) then
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> SendAcknowledgement');
{$ENDIF}
    if (ControlCode <> ACRDisconnect) then
     begin
      SendAcknowledgement(ServerSession, False, CurrentRequestID);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> SendAcknowledgement - SENT');
{$ENDIF}
     end;
    if CurrentRequestID>ServerSession.CurrentRequestID then
     begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> previous session still executes...');
{$ENDIF}
      WaitForServerSessionThread(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> previous session finished, block...');
{$ENDIF}
      EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> CS Entered!');
{$ENDIF}
      try
       if IsNotSingle then
         Exit;
      finally
       LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> CS Left!');
{$ENDIF}
      end;
     end;
    if ServerSession.Thread = nil then
     begin
      ServerSession.Thread := TACRServerSessionThread.Create(self,
                                                   ServerSession, ControlCode);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> CREATED # '+IntToStr(Integer(ServerSession.Thread.ThreadID))+'/'+IntToStr(Integer(ServerSession.Thread.Handle)));
{$ENDIF}
     end
    else
     begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> Server Session already exists');
{$ENDIF}
      Error:=
                  ErrorRServer+ErrorRServerSessionThread+
                  IntToStr(Integer(ServerSession.Thread.ThreadID))+'/'+IntToStr(Integer(ServerSession.Thread.Handle))+
                  ErrorRServerSessionNotFinished+
                  ' SessionID='+IntToStr(ServerSession.Session.SessionID)+
                  ', current comand: CurrentRequestID='+IntToStr(CurrentRequestID)+
                  ', ControlCode='+IntToStr(ControlCode)+
                  ', previous command still executed: CurrentRequestID='+IntToStr(ServerSession.CurrentRequestID)+
                  ', ControlCode='+IntToStr(ServerSession.Thread.FCode)+'.';
      try
       TACRNetworkSession(ServerSession.Session).DoOnError(AcrServerSessionThread,40514,Error);
      except
       TACRServer(FServer).DoOnConnectionError(AcrServerSessionThread,40514,Error);
      end;
     end;
   end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> FINISHED');
{$ENDIF}
end; // CommandReceived


//------------------------------------------------------------------------------
// MessageReceived
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.MessageReceived(
                                  ServerSession:  PACRSrvrSession;
                                  ControlCode:    TACRControlCode = ACRMessage
                                                      );
begin
  if ServerSession.MsgThread <> nil then
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> previous session still executes...');
{$ENDIF}
    WaitForServerSessionMsgThread(ServerSession);
   end;
  if (ControlCode >= ACRLastPacket) then
    ControlCode := ControlCode - ACRLastPacket;
  ServerSession.MsgReceiveStatus := ACRFull;
  ServerSession.MsgThread := TACRServerSessionMsgThread.Create(self,
                                                 ServerSession, ControlCode);
end; // MessageReceived


//------------------------------------------------------------------------------
// ReceiveMessage
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.ReceiveMessage(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
var
  i:                    Integer;
  Packets:              TList;
  Packet:               PACRPacket;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER are receiving message from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
{$ENDIF}
  if ServerSession.MsgReceiveStatus<>ACRFull then
   begin
    while (ServerSession.MsgReceiveStatus<>ACRStart)
    and (ServerSession.MsgReceiveStatus<>ACRFull)
    do // Wait for starting answer receive
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ServerSession.Session.ConnectParams.ServerReceiveSleep);
     end;
    StartTime := GetTickCount;
    while (ServerSession.MsgReceiveStatus<>ACRFull)
    do // Wait for all packets to arrive
     begin
      if (GetTickCount - StartTime) > ServerSession.Session.ConnectParams.ServerReceiveTimeOut then
       raise EACRException.Create(40026, ErrorRTimeoutFullReceive,
                                [ServerSession.Session.SessionID, ServerSession.Session.ConnectParams.ServerReceiveTimeOut]);
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ServerSession.Session.ConnectParams.ServerReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received request from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
  EnterCriticalSection(FCSect);
  if ServerSession.MsgControlCode = ACRExecute then
    ServerSession.MsgControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);
  Packets := ServerSession.MsgPackets.LockList;
  try
   BufferSize := 0;
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
    end;
   Buffer := MemoryManager.GetMem(BufferSize);
   pBuf := Buffer;
  // Copy to buffer then free MsgPackets
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
        begin
         Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
         inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
         MemoryManager.FreeAndNilMem(Packet.Buffer);
        end;
       Dispose(Packet);
       Packets.Items[i] := nil;
      end;
    end;
   Packets.Count := 0;
  finally
   ServerSession.MsgPackets.UnlockList;
  end;
  inc(ServerSession.ClientMessageID);
  ServerSession.MsgReceiveStatus := ACRNo;
  EnterCriticalSection(FCSect);
  if ServerSession.MsgControlCode = ACRSuspend then
    ServerSession.MsgControlCode := ACRExecute;
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER received full message from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('==============================================================');
{$ENDIF}
end; // ReceiveMessage


//------------------------------------------------------------------------------
// ReceiveBuffer
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.ReceiveBuffer(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
var
  i:                    Integer;
  Packet:               PACRPacket;
  Packets:              TList;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.ReceiveBuffer - START');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER are receiving request from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
{$ENDIF}
  if ServerSession.ReceiveStatus <> ACRFull then
   begin
    while (ServerSession.ReceiveStatus <> ACRStart)
    and (ServerSession.ReceiveStatus <> ACRFull)
    do // Wait for starting answer receive
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ServerSession.Session.ConnectParams.ServerReceiveSleep);
     end;
    StartTime := GetTickCount;
    while (ServerSession.ReceiveStatus <> ACRFull)
    do // Wait for all packets to arrive
     begin
      if (GetTickCount - StartTime) > ServerSession.Session.ConnectParams.ServerReceiveTimeOut then
       raise EACRException.Create(40025, ErrorRTimeoutFullReceive,
                                [ServerSession.Session.SessionID,
                                ServerSession.Session.ConnectParams.ServerReceiveTimeOut]);
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ServerSession.Session.ConnectParams.ServerReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received request from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.ReceiveBuffer> All packets received');
{$ENDIF}
  EnterCriticalSection(FCSect);
  if ServerSession.ControlCode = ACRExecute then
    ServerSession.ControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);
  BufferSize := 0;
  Packets := ServerSession.Packets.LockList;
  try
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
    end;
   if BufferSize > 0 then
     Buffer := MemoryManager.GetMem(BufferSize)
   else
     Buffer := nil;
   pBuf := Buffer;
  // Copy to buffer then free MsgPackets
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
        begin
         if (Packet.BufferSize > SizeOf(TACRPacketHeader)) then
          begin
           Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
           inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
          end;
         MemoryManager.FreeAndNilMem(Packet.Buffer);
        end;
       Dispose(Packet);
       Packets.Items[i] := nil;
      end;
    end;
   Packets.Count := 0;
  finally
   ServerSession.Packets.UnlockList;
  end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER received full buffer from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('==============================================================');
{$ENDIF}
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.ReceiveBuffer - FINISH');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('SERVER<<< '+IntToStr(ServerSession.CurrentRequestID)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}
end; // ReceiveBuffer


//------------------------------------------------------------------------------
// SendConnectAckn
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendConnectAckn(
                          ServerSession:        PACRSrvrSession;
                          CurrentRequestID:     Integer = -1
                                                      );
var
  Buffer,
  Buf:                            PChar;
  BufSize, BufferSize,
  SizeSID, SizeParams:            Integer;
  ConnectionParams:               TACRConnectionParams;
  RequestID:                      Integer;
begin
{
  ConnectionParams.LocalPort := ServerSession.RemotePort;
  ConnectionParams.RemoteHost := FNetwork.LocalHost;
  ConnectionParams.RemotePort := FNetwork.LocalPort;
  ConnectionParams.ServerID := FNetwork.FLocalClient;
}
  ConnectionParams.PacketSize := ServerSession.Session.ConnectParams.PacketSize;
  ConnectionParams.CompressionAlgorithm := ServerSession.Session.ConnectParams.CompressionAlgorithm;
  ConnectionParams.CompressionMode := ServerSession.Session.ConnectParams.CompressionMode;
  SizeParams := SizeOf(ConnectionParams);
  SizeSID := SizeOf(ServerSession.Session.SessionID);
  BufferSize := SizeParams + SizeSID;
  Buffer := MemoryManager.GetMem(BufferSize);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER: SessionID='+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('SizeSID='+IntToStr(SizeSID)+', SizeParams='+IntToStr(SizeParams)+', BufferSize='+IntToStr(BufferSize));
{$ENDIF}
  try
   Move(ConnectionParams, Buffer^, SizeParams);
   Move(ServerSession.Session.SessionID, (Buffer+SizeParams)^, SizeSID);

   if CurrentRequestID >= 0 then
    begin
     RequestID := ServerSession.CurrentRequestID;
     ServerSession.CurrentRequestID := CurrentRequestID;
    end;

   CompressAndEncryptBuffer(ServerSession.Session, Buffer, BufferSize, Buf, BufSize);
   try
    DoSendBuffer(ServerSession, Buf, BufSize, ACRConnected);
   finally
    if Buf<>Buffer then
      MemoryManager.FreeAndNilMem(Buf);
   end;

   if CurrentRequestID >= 0 then
     ServerSession.CurrentRequestID := RequestID;

   ServerSession.ReceiveStatus := ACRNo;
  finally
   MemoryManager.FreeAndNilMem(Buffer);
  end;
end;// SendConnectAckn


//------------------------------------------------------------------------------
// SendDisconnectRequest
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendDisconnectRequest(
                                    ServerSession: PACRSrvrSession;
                                    WaitForAnswer: Boolean = True);
var
  Header:           PACRPacketHeader;
  Retry:            Integer;
  StartTime:        DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.SendDisconnectRequest> START');
{$ENDIF}
 try
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   Header.ControlCode := ACRDisconnect;
   Header.Signature := ACRServerPacketSign;
   Header.Recepient := ServerSession.RemoteClientID;
   EnterCriticalSection(FCSect);
   Header.Sender := FNetwork.FLocalClient;
   LeaveCriticalSection(FCSect);
   Header.ConnectionID := ServerSession.ConnectionID;
   Header.SessionID := ServerSession.Session.SessionID;
   Header.PacketID := 0;
   Header.CurrentRequestID := ServerSession.CurrentRequestID;
   Retry := 0;
   ServerSession.SendStatus := ACRNotSent;
   repeat
     EnterCriticalSection(FCSect);
     try
      FNetwork.RemoteHost := ServerSession.RemoteHost;
      FNetwork.RemotePort := ServerSession.RemotePort;
{
      if ServerSession.ControlCode = ACRTerminate then
        raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
}
      FNetwork.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
     finally
      LeaveCriticalSection(FCSect);
     end;
     if not WaitForAnswer then
       Exit;
     StartTime := GetTickCount;
     while ((GetTickCount - StartTime) < ServerSession.Session.ConnectParams.DisconnectDelay) do // pause
      if ServerSession.SendStatus <> ACRSent then
       begin
        Sleep(ServerSession.Session.ConnectParams.ServerWaitForSendSleep);
{$IFDEF ProcessMessages}
        Application.ProcessMessages;
{$ENDIF ProcessMessages}
       end
      else
       Exit;
     inc(Retry);
   until  (Retry > ServerSession.Session.ConnectParams.DisconnectRetryCount);
   if ServerSession.SendStatus <> ACRSent then
    raise EACRException.Create(40024, ErrorRCannotDisconnect,
                   ['client', ServerSession.ClientSessionID,
                    ServerSession.Session.ConnectParams.DisconnectRetryCount,
                    ServerSession.Session.ConnectParams.DisconnectDelay]);
  finally
   MemoryManager.FreeAndNilMem(Header);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.SendDisconnectRequest> FINISH');
{$ENDIF}
  end;
 except
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerSessionThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
     raise;
    end;
{$ELSE}
  raise;
{$ENDIF}
 end;
end;// SendDisconnectRequest


//------------------------------------------------------------------------------
// DeleteSession
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DeleteSession(ServerSession: PACRSrvrSession);
var
  i:                    Integer;
  Packet:               PACRPacket;
  Packets:              TList;
  Error:                String;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> START');
{$ENDIF}
{$IFDEF MsgCommunicator}
  EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
  EnterCriticalSection(FCSect);
  try

   if ServerSession.CurrentRequestID = ACRTerminate then // <0, illegal value not used
    begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> session is already deleting - FINISH');
{$ENDIF}
     Exit;
    end;
   // block deleting ServerSession
   ServerSession.CurrentRequestID := ACRTerminate;

   // block using ServerSession
   ServerSession.ControlCode := ACRTerminate;
   ServerSession.MsgControlCode := ACRTerminate;
   // free listenning threads lists
   ServerSession.ListeningThreads.Free;
   ServerSession.ListeningThreads := nil;
   ServerSession.MsgListeningThreads.Free;
   ServerSession.MsgListeningThreads := nil;
  finally
   LeaveCriticalSection(FCSect);
  end;
  // wait for the end of current request execution
  TerminateServerThread(@ServerSession.Thread,AcrServerDeleteSession,
        ErrorRServerSessionThread,
        ServerSession.Session.ConnectParams.WaitForServerSessionThreadTimeOut);
{
  WaitForServerSessionThread(ServerSession);
  if ServerSession.Thread <> nil then
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionThread+
                  IntToStr(Integer(ServerSession.Thread.ThreadID))+'/'+IntToStr(Integer(ServerSession.Thread.Handle))+
                  ErrorRExecute;
    try
     TACRNetworkSession(ServerSession.Session).DoOnError(AcrServerDeleteSession,
                  40517,Error+
                  ErrorRCannotTerminateThread+
                  IntToStr(ACRWaitForServerSessionThreadTimeOut));
    except
     TACRServer(FServer).DoOnConnectionError(AcrServerDeleteSession,
                  40517,Error+
                  ErrorRCannotTerminateThread+
                  IntToStr(ACRWaitForServerSessionThreadTimeOut));
    end;
    TerminateThread(ServerSession.Thread.Handle, 0);
    WaitForServerSessionThread(ServerSession, ACRWaitForKillTimeOut);
    if ServerSession.Thread <> nil then
     begin
      try
       TACRNetworkSession(ServerSession.Session).DoOnError(AcrServerDeleteSession,
                  40518,Error+
                  ErrorRCannotKillThread+
                  IntToStr(ACRWaitForKillTimeOut));
      except
       TACRServer(FServer).DoOnConnectionError(AcrServerDeleteSession,
                  40518,Error+
                  ErrorRCannotKillThread+
                  IntToStr(ACRWaitForKillTimeOut));
      end;
     end;
   end;
}
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free Session...');
{$ENDIF}
  // free PacketIDsToResend
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> wait for resending broken packets...');
{$ENDIF}
  repeat
    if ServerSession.PacketIDsToResend.ItemCount = 0 then // wait for resending broken packets
      break;
    sleep(0);
  until False;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free PacketIDsToResend...');
{$ENDIF}
  ServerSession.PacketIDsToResend.Free;
  // free packets
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free packets...');
{$ENDIF}
  Packets := ServerSession.Packets.LockList;
  try
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
         MemoryManager.FreeAndNilMem(Packet.Buffer);
       Dispose(Packet);
      end;
    end;
  finally
   ServerSession.Packets.UnlockList;
  end;
  // free packets list
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free packet list...');
{$ENDIF}
  ServerSession.Packets.Free;
  // Wait for message thread
  TerminateServerThread(@ServerSession.MsgThread,AcrServerDeleteSession,ErrorRServerSessionMsgThread,
                        ServerSession.Session.ConnectParams.WaitForServerSessionThreadTimeOut);
(*
  repeat
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> SessionID='+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.DeleteSession> WaitFor message Session thread...');
aaWriteToLog('TACRServerConnectionManager.DeleteSession> MsgThread='+IntToStr(Integer(ServerSession.MsgThread)));
{$ENDIF}
    if ServerSession.MsgThread = nil then
      break;
    sleep(0);
  until False;
*)
  // free MsgPacketIDsToResend
{$IFDEF MsgCommunicator}
  EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> wait for resending broken message packets...');
{$ENDIF}
  repeat
    if ServerSession.MsgPacketIDsToResend.ItemCount = 0 then // wait for resending broken packets
      break;
    sleep(0);
  until False;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free MsgPacketIDsToResend...');
{$ENDIF}
  ServerSession.MsgPacketIDsToResend.Free;
{$IFDEF MsgCommunicator}
  LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
  // free message packets
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free message packets...');
{$ENDIF}
  Packets := ServerSession.MsgPackets.LockList;
  try
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
         MemoryManager.FreeAndNilMem(Packet.Buffer);
       Dispose(Packet);
      end;
    end;
  finally
   ServerSession.MsgPackets.UnlockList;
  end;
  // free message packets list
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free message packet list...');
{$ENDIF}
  ServerSession.MsgPackets.Free;
  // wait for vacant ServerSession
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> wait for vacant ServerSession...');
{$ENDIF}
  repeat
    // block using ServerSession
    EnterCriticalSection(FCSect);
    ServerSession.ControlCode := ACRTerminate;
    ServerSession.MsgControlCode := ACRTerminate;
    LeaveCriticalSection(FCSect);
    sleep(0);
    if (ServerSession.Status = ACRVacant) then
      break;
  until False;
{$IFDEF MsgCommunicator}
  LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
  // free session
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
  if ServerSession.Session <> nil then
   begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> Free Session...');
{$ENDIF}
    ServerSession.Session.Free;
    ServerSession.Session := nil;
   end;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> Dispose(ServerSession)...');
{$ENDIF}
  Dispose(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> FINISH');
{$ENDIF}
end; // DeleteSession


//------------------------------------------------------------------------------
// DoDisconnect
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DoDisconnect(Session: TACRBaseSession);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> START');
{$ENDIF}
{$IFDEF MsgCommunicator}
  if not FListenerStoped then
    TMsgServerSession(Session).DisconnectUser;
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> LockSessions...');
 {$ENDIF}
{$ENDIF}
  Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> FSessions Locked');
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Sessions.Count='+IntToStr(Sessions.Count));
{$ENDIF}
  try
   for i:=Sessions.Count-1 downto 0 do
    begin
     if (i<0)
     or (i>=Sessions.Count)
     then
       break;
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
      begin
       EnterCriticalSection(FCSect);
       if (ServerSession.ControlCode <> ACRTerminate)
       or (ServerSession.MsgControlCode <> ACRTerminate)
       then
        begin
         // Stop ping
         ServerSession.Connected := False;
         // Terminate sending threads
         ServerSession.ControlCode := ACRTerminate;
         ServerSession.MsgControlCode := ACRTerminate;
         LeaveCriticalSection(FCSect);
{$IFDEF MsgCommunicator}
         // block sendmessage
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Session CS entering...');
 {$ENDIF}
         EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
         FSessions.UnlockList;
         // Terminate listening threads
         TerminateCommandThreads(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Terminate message listening threads...');
{$ENDIF}
         TerminateMessageThreads(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> All sessions terminated');
{$ENDIF}
         // delete session from list
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Remove SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
         FSessions.Remove(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Removed SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
{$IFDEF MsgCommunicator}
         LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
 {$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Session CS left');
 {$ENDIF}
{$ENDIF}
         // free memory and session
         DeleteSession(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Session deleted');
{$ENDIF}
         Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('DoDisconnect> Sessions locked');
{$ENDIF}
        end
       else
         LeaveCriticalSection(FCSect);
(*
       // Terminate all threads
       FSessions.UnlockList;
       TerminateAllSessionThreads(ServerSession);
       // delete session from list
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Session remove session from list...');
{$ENDIF}
       FSessions.Remove(ServerSession);
       // free memory and session
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Delete SessionID='+IntToStr(ServerSession.Session.SessionID)+'...');
{$ENDIF}
       DeleteSession(ServerSession);
       Sessions:=FSessions.LockList;
*)
       break;
      end;
    end;
  finally
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Sessions.Count='+IntToStr(Sessions.Count));
{$ENDIF}
   FSessions.UnlockList;
  end;
end; // DoDisconnect


//------------------------------------------------------------------------------
// TerminateSession
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateSession(Session: TACRBaseSession);
begin
  if SessionTerminator <> nil then
    if SessionTerminator.FTerminatedSessions <> nil then
      SessionTerminator.FTerminatedSessions.Add(Session);
end;// TerminateSession


//------------------------------------------------------------------------------
// Disconnect
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.Disconnect(Session: TACRBaseSession);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
  Found:                Boolean;
begin
  Found := False;
  Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.Disconnect> FSessions Locked');
{$ENDIF}
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
      begin
       Found := True;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);

  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRSuspend;
  ServerSession.MsgControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);

  SendDisconnectRequest(ServerSession);

{
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRTerminate;
  ServerSession.MsgControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
}

  DoDisconnect(Session);
end; // Disconnect


//------------------------------------------------------------------------------
// DisconnectAll
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DisconnectAll(WaitForAllDisconected: Boolean = True);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> START');
{$ENDIF}
if FSessions<>nil then
begin
try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> Lock FSessions...');
{$ENDIF}
 Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> FSessions Locked');
{$ENDIF}
 try
//  while Sessions.Count>=1 do
  for i:=Sessions.Count-1 downto 0 do
   begin
    if (i<0)
    or (i>=Sessions.Count)
    then
      break;
    ServerSession := Sessions.Items[i];
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> EnterCriticalSection...');
{$ENDIF}
    EnterCriticalSection(FCSect);
    if (ServerSession.ControlCode <> ACRTerminate)
    or (ServerSession.MsgControlCode <> ACRTerminate)
    then
     begin
      // Stop ping
      ServerSession.Connected := False;
      // Terminate sending threads
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> Terminate sending threads...');
{$ENDIF}
      ServerSession.ControlCode := ACRTerminate;
      ServerSession.MsgControlCode := ACRTerminate;
      LeaveCriticalSection(FCSect);
      // Block sending message
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> Block sending message...');
{$ENDIF}
{$IFDEF MsgCommunicator}
      EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
      FSessions.UnlockList;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> SendDisconnectRequest');
{$ENDIF}
      // send
      try
       SendDisconnectRequest(ServerSession, False);
      except
       // ignore all exceptions while disconnecting
      end;
      // Terminate all threads
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> TerminateAllSessionThreads...');
{$ENDIF}
      TerminateAllSessionThreads(ServerSession);
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> All sessions terminated');
{$ENDIF}
      // delete session from list
      FSessions.Remove(ServerSession);
      // free memory and session
{$IFDEF MsgCommunicator}
      LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
      DeleteSession(ServerSession);
      Sessions:=FSessions.LockList;
     end
    else
      LeaveCriticalSection(FCSect);
   end;
//  Sessions.Capacity :=0;
  // Wait for all sessions be disconnected
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> WaitForAllDisconected...');
{$ENDIF}
  if WaitForAllDisconected then
    repeat
      Sessions:=FSessions.LockList;
      try
       if Sessions.Count = 0 then
         break;
      finally
       FSessions.UnlockList;
      end;
      sleep(0);
    until False;
 finally
  FSessions.UnlockList;
 end;
except
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
    end;
{$ENDIF}
end;
end;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> FINISH');
{$ENDIF}
end; // DisconnectAll


//------------------------------------------------------------------------------
// SessionsCount
//------------------------------------------------------------------------------
function TACRServerConnectionManager.SessionsCount: Integer;
var
  Sessions:           TList;
begin
  Sessions := FSessions.LockList;
  try
    Result := Sessions.Count;
  finally
    FSessions.UnlockList;
  end;
end;// SessionsCount


//------------------------------------------------------------------------------
// TerminateMessageThreads
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateMessageThreads(ServerSession: PACRSrvrSession);
var
  i:                    Integer;
  Count:                Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
//  Error:                String;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> START');
{$ENDIF}
try
  Threads := ServerSession.MsgListeningThreads.LockList;
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
  try
   for i := 0 to Threads.Count-1 do
    begin
     Thread := Threads.Items[i];
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
     TerminateServerThread(@Thread,AcrServerConnectionManager,ErrorRListenerMsgThread);
    end;
  finally
   ServerSession.MsgListeningThreads.UnlockList;
  end;
(*
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> START');
{$ENDIF}
  if ServerSession.MsgListeningThreads = nil then
   begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> MsgListeningThreads = nil - FINISH');
{$ENDIF}
    Exit;
   end;
  EnterCriticalSection(FCSect);
  ServerSession.MsgControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> MsgListeningThreads.LockList...');
{$ENDIF}
   // Terminate message listening threads
   Threads := ServerSession.MsgListeningThreads.LockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> MsgListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
   try
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ServerSession.MsgListeningThreads.UnlockList;
   end;
   sleep(0);
   // Check
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> CHECKING...');
{$ENDIF}
   Threads := ServerSession.MsgListeningThreads.LockList;
   try
    Count := Threads.Count;
   finally
    ServerSession.MsgListeningThreads.UnlockList;
   end;
   if (Count = 0) then break
   else
    begin
     if (GetTickCount - StartTime) > ACRServerThreadsTerminateDelay then
      begin
       Error := IntToStr(Count)+' message listening threads';
       raise EACRException.Create(40036, ErrorRThreadHangs,
            ['Server session # '+IntToStr(ServerSession.Session.SessionID),
            Error]);
      end;
    end;
  until False;
*)
finally
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> FINISH');
{$ENDIF}
end;
end; // TerminateMessageThreads


//------------------------------------------------------------------------------
// TerminateCommandThreads
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateCommandThreads(ServerSession: PACRSrvrSession);
var
  i:                    Integer;
  Count:                Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> START');
{$ENDIF}
try
  Threads := ServerSession.ListeningThreads.LockList;
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
  try
   for i := 0 to Threads.Count-1 do
    begin
     Thread := Threads.Items[i];
{$IFDEF DEBUG_LOG_SERVER_TERMINATE}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
     TerminateServerThread(@Thread,AcrServerConnectionManager,ErrorRListenerThread);
    end;
  finally
   ServerSession.ListeningThreads.UnlockList;
  end;
(*
  Error:                String;

procedure LogHangedThreads(EMessage: String; ENativeCode: Integer; Kill: Boolean = False);
var
  i:                    Integer;
begin
    Threads := ServerSession.ListeningThreads.LockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    try
     for i := 0 to Threads.Count-1 do
      begin
       Thread := Threads.Items[i];
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
       if (Thread <> nil) then
        begin
         Error:=
                  EMessage+
                  ErrorRServer+ErrorRListenerThread+
                  IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle))+
                  ErrorRExecute;
         if Kill then
           Error := Error+ErrorRCannotKillThread+IntToStr(ACRWaitForKillTimeOut)
         else
           Error := Error+ErrorRCannotTerminateThread+IntToStr(ACRServerThreadsTerminateDelay);
         try
          TACRNetworkSession(ServerSession.Session).DoOnError(AcrServerTerminateCommandThreads,ENativeCode,Error);
         except
          TACRServer(FServer).DoOnConnectionError(AcrServerTerminateCommandThreads,ENativeCode,Error);
         end;
         if not Kill then
           TerminateThread(Thread.Handle, 0);
//         sleep(0);
        end;
      end;
    finally
     ServerSession.ListeningThreads.UnlockList;
    end;
end; // LogHangedThreads

function IsListeningThreadExisting(TimeOut: DWORD): Boolean;
begin
 Result := True;
 Threads := ServerSession.ListeningThreads.LockList;
 try
  Count := Threads.Count;
 finally
  ServerSession.ListeningThreads.UnlockList;
 end;
 if (Count = 0) then
  begin
   Result := False;
   Exit;
  end
 else
  begin
   if (GetTickCount - StartTime) > TimeOut then
    begin
     Error := IntToStr(Count)+' command listening threads';
     raise EACRException.Create(40036, ErrorRThreadHangs,
           ['Server session # '+IntToStr(ServerSession.Session.SessionID),
            Error]);
    end;
  end;
end; // IsListeningThreadExisting

begin //TerminateCommandThreads
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> START');
{$ENDIF}
try // finally
  if ServerSession.ListeningThreads = nil then
   begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> ListeningThreads = nil - FINISH');
{$ENDIF}
    Exit;
   end;
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
 try // except
  StartTime := GetTickCount;
  repeat   // Wait for finishes of all the threads
    // Terminate all command listening threads
    Threads := ServerSession.ListeningThreads.LockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    try
     for i := 0 to Threads.Count-1 do
      begin
       Thread := Threads.Items[i];
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
       if (Thread <> nil) then
         Thread.Terminate;
       sleep(0);
      end;
    finally
     ServerSession.ListeningThreads.UnlockList;
    end;
    sleep(0);
    if not IsListeningThreadExisting(ACRServerThreadsTerminateDelay) then
      Exit;
  until False;
 except
  on E: EACRException do
    LogHangedThreads(E.Message,E.NativeError,False);
 end; // exception
 try
  StartTime := GetTickCount;
  IsListeningThreadExisting(ACRWaitForKillTimeOut);
 except
  on E: EACRException do
    LogHangedThreads(E.Message,E.NativeError,True);
 end; // exception
*)
finally
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> FINISH');
{$ENDIF}
end;
end; // TerminateCommandThreads


//------------------------------------------------------------------------------
// TerminateAllSessionThreads
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateAllSessionThreads(ServerSession: PACRSrvrSession);
begin
  TerminateCommandThreads(ServerSession);
  TerminateMessageThreads(ServerSession);
end; // TerminateAllSessionThreads


//------------------------------------------------------------------------------
// DoSendBuffer
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DoSendBuffer(
                          ServerSession:    PACRSrvrSession;
                          Buffer:           PChar;
                          BufferSize:       Integer;
                          Code:             Integer = ACRNoAction
                                                  );
var
  Header:               PACRPacketHeader;
  Packets:              TList;
  Packet:               PACRPacket;
  BytesSent, DataSize:  Integer;
  PacketID:             Integer;
  StartTime, Delay:     DWORD;
  i:                    Integer;

procedure WaitForCommandSent;
var
  PacketID,
  ControlCode,
  i:                    Integer;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
  j:                    Integer;
{$ENDIF}
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForCommandSent> - START');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  ControlCode := ServerSession.ControlCode;
  ServerSession.ControlCode := ACRExecute;
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server waits for all packets');
{$ENDIF}
  if not ((Code = ACRConnect) or (Code = (ACRConnect + ACRLastPacket))) then
   begin
(*
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRServerReceiveSleep); // To avoid delay if full answer is already received
*)
    i := ServerSession.PacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ServerSession.SendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRServerResendDelay);
      if ServerSession.PacketIDsToResend.ItemCount = 0 then
       begin // no packets to resend -- sleep
        StartSleepTime := GetTickCount;
        while ((GetTickCount - StartSleepTime) < ServerSession.Session.ConnectParams.ServerResendDelay) do  // sleep after last packet send
         begin
          if ServerSession.SendStatus = ACRSent then
            Exit;
          if ServerSession.ControlCode = ACRTerminate then // added as analogue of client
            raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
          if (GetTickCount - StartTime) > ServerSession.Session.ConnectParams.ServerSendTimeOut then
            raise EACRException.Create(40078, ErrorRServerTimeOutSending,
              ['message',
{$IFDEF MsgCommunicator}
              ServerSession.Session.UserID,
{$ELSE}
              ServerSession.Session.SessionID,
{$ENDIF}
              ServerSession.Session.ConnectParams.ServerSendTimeOut]);
          Sleep(ServerSession.Session.ConnectParams.ServerWaitForSendSleep);
         end; // sleep
       end; // no packets to resend
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server has '+IntToStr(ServerSession.PacketIDsToResend.ItemCount)+' packets to resend:');
for j:=0 to ServerSession.PacketIDsToResend.ItemCount-1 do
aaWriteToLog(IntToStr(ServerSession.PacketIDsToResend.Items[j]));
{$ENDIF}
      dec(i);
      if (i<0) then
        i := ServerSession.PacketIDsToResend.ItemCount - 1;
      try
       ServerSession.PacketIDsToResend.Lock;
       if (ServerSession.PacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ServerSession.PacketIDsToResend.Items[i];
           ServerSession.PacketIDsToResend.Unlock;
           if PacketID >= Packets.Count then // Erroneous situation which is must not exist
             Continue;
          end
         else // No packets to resend - Resend last packet
          begin
           ServerSession.PacketIDsToResend.Unlock;
           PacketID := Packets.Count - 1;
          end;
         Packet := Packets.Items[PacketID];
         // resend packet
         EnterCriticalSection(FCSect);
         try
          FNetwork.RemoteHost := ServerSession.RemoteHost;
          FNetwork.RemotePort := ServerSession.RemotePort;
          if ServerSession.ControlCode = ACRTerminate then
            raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
          FNetwork.SendBuffer(Packet.Buffer, Packet.BufferSize);
         finally
          LeaveCriticalSection(FCSect);
         end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server resent packet # '+IntToStr(PacketID));
{$ENDIF}
        end
      else
        ServerSession.PacketIDsToResend.Unlock;
      finally
       ServerSession.PacketIDsToResend.Lock;
       if  (ServerSession.PacketIDsToResend.ItemCount > 0)
       and (ServerSession.PacketIDsToResend.ItemCount > i)
       then
//         ServerSession.PacketIDsToResend.Delete(i);
         ServerSession.PacketIDsToResend.Remove(PacketID);
       ServerSession.PacketIDsToResend.Unlock;
      end; // finally
     end; // loop
   end; // no Connect
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server received all the packets');
{$ENDIF}
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ControlCode;
  LeaveCriticalSection(FCSect);
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForCommandSent> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // WaitForCommandSent

procedure WaitForMessageSent;
var
  PacketID,
  ControlCode,
  i:                    Integer;
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> - START');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  ControlCode := ServerSession.MsgControlCode;
  ServerSession.MsgControlCode := ACRExecute;
  LeaveCriticalSection(FCSect);
(*
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRServerReceiveSleep); // To avoid delay if full answer is already received
*)
    i := ServerSession.MsgPacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ServerSession.MsgSendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRServerResendDelay);
      if ServerSession.MsgPacketIDsToResend.ItemCount = 0 then
       begin
        StartSleepTime := GetTickCount;
        while ((GetTickCount - StartSleepTime) < ServerSession.Session.ConnectParams.ServerResendDelay) do  // sleep after last packet send
         begin
          if ServerSession.MsgSendStatus = ACRSent then
            Exit;
          if ServerSession.MsgControlCode = ACRTerminate then // added as analogue of client
            raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
          if (GetTickCount - StartTime) > ServerSession.Session.ConnectParams.ServerSendTimeOut then
            raise EACRException.Create(40078, ErrorRServerTimeOutSending,
              ['message',
{$IFDEF MsgCommunicator}
              ServerSession.Session.UserID,
{$ELSE}
              ServerSession.Session.SessionID,
{$ENDIF}
              ServerSession.Session.ConnectParams.ServerSendTimeOut]);
          Sleep(ServerSession.Session.ConnectParams.ServerWaitForSendSleep);
         end; // sleep
       end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> New step');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> SessionID='+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.Session='+IntToStr(Integer(ServerSession.Session)));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> i='+IntToStr(i));
{$ENDIF}
      dec(i);
      if i<0 then
        i := ServerSession.MsgPacketIDsToResend.ItemCount - 1;
      try
       ServerSession.MsgPacketIDsToResend.Lock;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> MsgPacketIDsToResend.ItemCount = '+IntToStr(ServerSession.MsgPacketIDsToResend.ItemCount));
{$ENDIF}
       if (ServerSession.MsgPacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ServerSession.MsgPacketIDsToResend.Items[i];
           ServerSession.MsgPacketIDsToResend.Unlock;
           if PacketID >= Packets.Count then
             Continue;
          end
         else // No packets to resend - Resend last packet
          begin
           ServerSession.MsgPacketIDsToResend.Unlock;
           PacketID := Packets.Count - 1;
          end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> PacketID = '+IntToStr(PacketID));
{$ENDIF}
         Packet := Packets.Items[PacketID];
         // resend packet
         EnterCriticalSection(FCSect);
         try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(GetTickCount)+':');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> RemoteHost=');
aaWriteToLog(ServerSession.RemoteHost);
{$ENDIF}
          FNetwork.RemoteHost := ServerSession.RemoteHost;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> RemotePort=');
aaWriteToLog(IntToStr(ServerSession.RemotePort));
{$ENDIF}
          FNetwork.RemotePort := ServerSession.RemotePort;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Sending...');
{$ENDIF}
          if ServerSession.MsgControlCode = ACRTerminate then
            raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
          FNetwork.SendBuffer(Packet.Buffer, Packet.BufferSize);
         finally
          LeaveCriticalSection(FCSect);
         end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Sent!');
{$ENDIF}
        end
      else
        ServerSession.MsgPacketIDsToResend.Unlock;
      finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Locking...');
{$ENDIF}
       ServerSession.MsgPacketIDsToResend.Lock;
       if  (ServerSession.MsgPacketIDsToResend.ItemCount > 0)
       and (ServerSession.MsgPacketIDsToResend.ItemCount > i)
       then
        begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Removing PacketID = '+IntToStr(PacketID));
{$ENDIF}
         ServerSession.MsgPacketIDsToResend.Remove(PacketID);
        end;
       ServerSession.MsgPacketIDsToResend.Unlock;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> UnlLocked');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.Session=');
aaWriteToLog(IntToStr(Integer(ServerSession.Session)));
{$ENDIF}
      end; // finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.Session=');
aaWriteToLog(IntToStr(Integer(ServerSession.Session)));
{$ENDIF}
     end; // loop
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Ending...');
{$ENDIF}
  EnterCriticalSection(FCSect);
  ServerSession.MsgControlCode := ControlCode;
  LeaveCriticalSection(FCSect);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.MsgSendStatus = '+IntToStr(ServerSession.MsgSendStatus));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> SessionID = '+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> End!');
{$ENDIF}
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // WaitForMessageSent

begin // DoSendBuffer
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER are sending buffer...');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 if (ServerSession.Session.ConnectParams.PacketSize < SizeOf(TACRPacketHeader)) then
   raise EACRException.Create(40020, ErrorRPacketSizeTooSmall,
                              [ServerSession.Session.ConnectParams.PacketSize,
                                     SizeOf(TACRPacketHeader)]);

 Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
 try
  DataSize := ServerSession.Session.ConnectParams.PacketSize - SizeOf(TACRPacketHeader);
  Header.Signature := ACRServerPacketSign;
  Header.Recepient := ServerSession.RemoteClientID;
  EnterCriticalSection(FCSect);
  Header.Sender := FNetwork.FLocalClient;
  LeaveCriticalSection(FCSect);
  Header.ConnectionID := ServerSession.ConnectionID;
  if Code = ACRConnected then
    Header.SessionID := ServerSession.ClientSessionID
  else
    Header.SessionID := ServerSession.Session.SessionID;
  Header.PacketID := 0;
  Header.ControlCode := Code;
  if Code = ACRMessage then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER IS SENDING MESSAGE # '+IntToStr(ServerSession.ClientMessageID)+' TO CLIENT ID='+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('==============================================================');
{$ENDIF}
    Header.CurrentRequestID := ServerSession.ServerMessageID;
   end
  else
    Header.CurrentRequestID := ServerSession.CurrentRequestID;
  Packets := TList.Create;
   try
    BytesSent := 0;
    i := DataSize;
    if Code = ACRMessage then
     begin
      ServerSession.MsgSendStatus := ACRNotSent;
     end
    else
      ServerSession.SendStatus := ACRNotSent;
    repeat
//    while BytesSent < BufferSize do
//     begin // Create and send all packets
      New(Packet);
      Packets.Add(Packet);
      Packet.Buffer := MemoryManager.GetMem(ServerSession.Session.ConnectParams.PacketSize);
      if BytesSent + DataSize > BufferSize then
       DataSize := BufferSize - BytesSent;
      Packet.BufferSize := DataSize + SizeOf(TACRPacketHeader);
      if BytesSent + DataSize = BufferSize then
       Header.ControlCode := Header.ControlCode+ACRLastPacket;
      Move(Header^, Packet.Buffer^, SizeOf(TACRPacketHeader));
      Move(Pointer(Integer(Buffer)+Header.PacketID*i)^, Pointer(Integer(Packet.Buffer)+SizeOf(TACRPacketHeader))^, DataSize);
      inc(Header.PacketID);
      // send packet
      EnterCriticalSection(FCSect);
      try
       FNetwork.RemoteHost := ServerSession.RemoteHost;
       FNetwork.RemotePort := ServerSession.RemotePort;
       if Code = ACRMessage then
        begin
         if ServerSession.MsgControlCode = ACRTerminate then
           raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
        end
       else
         if ServerSession.ControlCode = ACRTerminate then
           raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
       FNetwork.SendBuffer(Packet.Buffer, Packet.BufferSize);
      finally
       LeaveCriticalSection(FCSect);
      end;
      sleep(0); // - The fastest speed with sleep !!!
      // packet has been sent
      BytesSent := BytesSent + DataSize;
//     end;
    until BytesSent >= BufferSize;
    // Wait for the end of sending...
    if Code = ACRMessage then
     begin
       if ServerSession.MsgControlCode <> ACRTerminate then
         WaitForMessageSent;
     end
    else
     begin
      if Code <> ACRConnected then
       if ServerSession.ControlCode <> ACRTerminate then
         WaitForCommandSent;
     end;
   finally
    // Remove all sent packets and free memory
    for i:= 0 to Packets.Count - 1 do
     begin
      Packet := Packets.Items[i];
      MemoryManager.FreeAndNilMem(Packet.Buffer);
      Dispose(Packet);
     end;
    Packets.Free;
    // Allow to add new packets
    if Code = ACRMessage then
     begin
      if ServerSession.MsgControlCode = ACRTerminate then
        raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
      inc(ServerSession.ServerMessageID);
      sleep(0);
      ServerSession.MsgPacketIDsToResend.Lock;
      if ServerSession.MsgControlCode <> ACRTerminate then
        ServerSession.MsgPacketIDsToResend.SetSize(0);
      ServerSession.MsgPacketIDsToResend.Unlock;
      ServerSession.MsgReceiveStatus := ACRNo;
     end
    else
     begin
      if ServerSession.ControlCode = ACRTerminate then
        raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
      EnableNextCommand(ServerSession);
     end;
   end; // try
 finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('SendBuffer> FreeAndNilMem(Header)');
{$ENDIF}
  MemoryManager.FreeAndNilMem(Header);
 end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> ServerSession.MsgSendStatus = '+IntToStr(ServerSession.MsgSendStatus));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> End!');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER sent buffer');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // DoSendBuffer


//------------------------------------------------------------------------------
// EnableNextCommand
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.EnableNextCommand(ServerSession: PACRSrvrSession);
begin
      ServerSession.PacketIDsToResend.Lock;
      ServerSession.PacketIDsToResend.SetSize(0);
      ServerSession.PacketIDsToResend.Unlock;
      inc(ServerSession.CurrentRequestID);
      ServerSession.ReceiveStatus := ACRNo;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('ServerSession.ControlCode = '+IntToStr(ServerSession.ControlCode));
{$ENDIF}
      EnterCriticalSection(FCSect);
      if ServerSession.ControlCode = ACRSuspend then
        ServerSession.ControlCode := ACRExecute;
      LeaveCriticalSection(FCSect);
      sleep(0);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('ServerSession.ControlCode = '+IntToStr(ServerSession.ControlCode));
{$ENDIF}
end;// EnableNextCommand


//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNoAction
                                                  );
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  SessionFound:         Boolean;
  Packets:              TList;
  i:                    Integer;
  Buf:                  PChar;
  BufSize:              Integer;
begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.SendBuffer> - START');
{$ENDIF}
  SessionFound := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('SERVER>>> '+IntToStr(ServerSession.CurrentRequestID)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}

  inc(ServerSession.Status);
  try
   CompressAndEncryptBuffer(ServerSession.Session, Buffer, BufferSize, Buf, BufSize);
   try
    DoSendBuffer(ServerSession, Buf, BufSize, Code);
   finally
    if Buf <> Buffer then
      MemoryManager.FreeAndNilMem(Buf);
   end;
   dec(ServerSession.Status);
  except
   dec(ServerSession.Status);
   raise;
  end;

  (*
  if Code <> ACRMessage then
   begin
    ServerSession.ReceiveStatus := ACRNo;
    Packets := ServerSession.Packets.LockList;
    try
     for i:=0 to Packets.Count-1 do
       Packets.Items[i] := nil;
     Packets.Count := 0;
    finally
     ServerSession.Packets.UnlockList;
    end;
   end
  else
   begin
    ServerSession.MsgReceiveStatus := ACRNo;
    Packets := ServerSession.MsgPackets.LockList;
    try
     for i:=0 to Packets.Count-1 do
       Packets.Items[i] := nil;
     Packets.Count := 0;
    finally
     ServerSession.MsgPackets.UnlockList;
    end;
   end;

  DoSendBuffer(ServerSession, Buffer, BufferSize, Code);

  if Code <> ACRMessage then
   begin
//     if ServerSession.ReceiveStatus = ACRNo then
//       inc(ServerSession.CurrentRequestID);
   end
  else
   begin
     inc(ServerSession.ServerMessageID);
   end;
*)
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('TACRServerConnectionManager.SendBuffer> - FINISH');
{$ENDIF}
end; // SendBuffer


//------------------------------------------------------------------------------
// SendMessage
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                                                  );
var
  Buf:            PChar;
  SessionFound:   Boolean;
  Sessions:       TList;
  i:              Integer;
  ServerSession:  PACRSrvrSession;
  StartTime:      DWORD;
begin
  SessionFound := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
  // wait for Session be vacant
  StartTime := GetTickCount;
  repeat
   if Session = nil then
     raise EACRException.Create(40071, ErrorRSessionDeleted);
   if ServerSession.MsgControlCode = ACRTerminate then
     raise EACRException.Create(40072, ErrorRSendSessionTerminated);
   if (GetTickCount - StartTime) > ServerSession.Session.ConnectParams.WaitForMessagesSend then
     raise EACRException.Create(40073, ErrorRTimeoutMessageSent,[
{$IFDEF MsgCommunicator}
          ServerSession.Session.UserID,
{$ELSE}
          ServerSession.Session.SessionID,
{$ENDIF}
          ServerSession.Session.ConnectParams.WaitForMessagesSend]);
   if (ServerSession.Status = ACRVacant) then
      break;
{$IFDEF ProcessMessages}
   Application.ProcessMessages;
{$ENDIF ProcessMessages}
   Sleep(ServerSession.Session.ConnectParams.ServerResendDelay);
  until False;
  // block other threads to send at the same time
{$IFDEF MsgCommunicator}
  EnterCriticalSection(TACRServerSession(Session).FCSect);
{$ENDIF}
  try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - Start');
{$ENDIF}
(*
   if (TACRCompressionAlgorithm(Session.ConnectParams.CompressionAlgorithm) <> acaNone)
   or (Session.ConnectParams.CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)
   then
    begin
     Buf := MemoryManager.GetMem(BufferSize);
     Move(Buffer^,Buf^,BufferSize);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - SendBuffer SessionID='+IntToStr(Session.SessionID));
{$ENDIF}
     SendBuffer(Session, Buf, BufferSize, ACRMessage);
     MemoryManager.FreeAndNilMem(Buf);
    end
   else
*)
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - SendBuffer SessionID='+IntToStr(Session.SessionID));
{$ENDIF}
     SendBuffer(Session, Buffer, BufferSize, ACRMessage);
    end;
  finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - SessionID='+IntToStr(Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.SendMessage> - FINISH');
{$ENDIF}
{$IFDEF MsgCommunicator}
   LeaveCriticalSection(TACRServerSession(Session).FCSect);
{$ENDIF}
  end;
end; // SendMessage


//------------------------------------------------------------------------------
// SendPing
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendPing(
                          ServerSession:        PACRSrvrSession
                                                          );
var
  Header:          PACRPacketHeader;
begin
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   Header.ControlCode := ACRPing;
   Header.Signature := ACRServerPacketSign;
   Header.Recepient := ServerSession.RemoteClientID;
   Header.ConnectionID := ServerSession.ConnectionID;
   Header.SessionID := ServerSession.Session.SessionID;
   Header.PacketID := 0;
   Header.CurrentRequestID := 0;
   EnterCriticalSection(FCSect);
   try
    FNetwork.RemoteHost := ServerSession.RemoteHost;
    FNetwork.RemotePort := ServerSession.RemotePort;
    FNetwork.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
   finally
    LeaveCriticalSection(FCSect);
   end;
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
end;// SendPing


//------------------------------------------------------------------------------
// SendAcknowledgement
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendAcknowledgement(
                          ServerSession:        PACRSrvrSession;
                          Msg:                  Boolean = False;
                          CurrentRequestID:     Integer = -1
                                                          );
var
  Header:          PACRPacketHeader;
begin
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   if Msg then
     Header.ControlCode := ACRMessageReceived
   else
     Header.ControlCode := ACRAllPacketsReceived;
   Header.Signature := ACRServerPacketSign;
   Header.Recepient := ServerSession.RemoteClientID;
   Header.ConnectionID := ServerSession.ConnectionID;
   Header.SessionID := ServerSession.Session.SessionID;
   Header.PacketID := 0;
   if CurrentRequestID>=0 then
     Header.CurrentRequestID := CurrentRequestID
   else
    begin
     if Msg then
       Header.CurrentRequestID := ServerSession.ClientMessageID
     else
       Header.CurrentRequestID := ServerSession.CurrentRequestID;
    end;
   EnterCriticalSection(FCSect);
   try
    FNetwork.RemoteHost := ServerSession.RemoteHost;
    FNetwork.RemotePort := ServerSession.RemotePort;
    FNetwork.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
   finally
    LeaveCriticalSection(FCSect);
   end;
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
end;// SendAcknowledgement


//------------------------------------------------------------------------------
// GetClientInfo
//------------------------------------------------------------------------------
function TACRServerConnectionManager.GetClientInfo(
                          Session:          TACRBaseSession;
                          var Host:         String;
                          var Port:         Integer;
                          var Application:  String
                                                    ): Boolean;
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
  Result := False;
  Sessions:=FSessions.LockList;
  try
   for i:=Sessions.Count-1 downto 0 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
      begin
       Host := ServerSession.RemoteHost;
       Port := ServerSession.RemotePort;
       Application := ServerSession.Application;
       Result := True;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
{
  if not Found then
    raise EACRException.Create(40037, ErrorRSessionNotFound,
                                ['Server', Session.SessionID]);
}
end; // GetClientInfo


//------------------------------------------------------------------------------
// GetClientsList
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.GetClientsList(var Clients: TACRSessionsArray);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
  Sessions:=FSessions.LockList;
  try
   SetLength(Clients, Sessions.Count);
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     Clients[i] := ServerSession.Session;
    end;
  finally
   FSessions.UnlockList;
  end;
end; // GetClientsList


//------------------------------------------------------------------------------
// if authorization buffer is valid (encrypted by the same crypto settings as in CryptoInfo)
// then return true else return false
//------------------------------------------------------------------------------
function TACRServerConnectionManager.IsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
begin
  Result := fnIsAuthorizationBufferValid(CryptoInfo, Buffer, BufferSize);
end; // IsAuthorizationBufferValid

// TACRServerConnectionManager




////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerSessionThread.Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          Code:             Integer = ACRNoAction
                                            );
begin
 try
  FManager := Manager;
  FManager.IncThreadCount;
  FServerSession := ServerSession;
  FCode := Code;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerSessionThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy;
//------------------------------------------------------------------------------
destructor TACRServerSessionThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if FServerSession.Session <> nil then
    FServerSession.Thread := nil;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD - FINISHED');
{$ENDIF}
  if (FManager<>nil) then
    FManager.DecThreadCount;
  inherited Destroy;
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    if (FServerSession = nil)
    or (FServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionThread,-1,
                  Error)
    else
      TACRNetworkSession(FServerSession.Session).DoOnError(
                  AcrServerSessionThread,-1,
                  Error);
   end;
 end;
end;// Destroy;


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerSessionThread.Execute;
var
  Buf:                 PChar;
  BufSize, AuBufSize:  Integer;
  ConnectionParams:    PACRConnectionParams;
begin
try // finally
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - START - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
 repeat
  repeat
   if Terminated then
     Exit;
   if FServerSession.ReceiveStatus = ACRFull then
     break;
   sleep(0);
  until False;
 try // except
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - New Command - GetTickCount='+IntToStr(GetTickCount));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID='+IntToStr(FServerSession.CurrentRequestID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Code='+IntToStr(FCode));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ReceiveStatus = '+IntToStr(FServerSession.ReceiveStatus));
{$ENDIF}
  if (FCode <> ACRMessage)
  and (FCode <> ACRDisconnect)
  then
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ControlCode='+IntToStr(FServerSession.ControlCode));
{$ENDIF}
    EnterCriticalSection(FManager.FCSect);
    if FServerSession.ControlCode = ACRExecute then
      FServerSession.ControlCode := ACRSuspend;
    LeaveCriticalSection(FManager.FCSect);
    try
     if FCode<>ACRConnect then
       FManager.SendAcknowledgement(FServerSession);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement sent! - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
    except
     on E:Exception do
      raise EACRException.Create(40519, ErrorRAckn+E.Message);
    end;
    try
     FManager.ReceiveBuffer(FServerSession, Buf, BufSize);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer received!      - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
    except
     on E:Exception do
      raise EACRException.Create(40506, ErrorRCannotReceive+E.Message);
    end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer received');
{$ENDIF}
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ControlCode='+IntToStr(FServerSession.ControlCode));
{$ENDIF}
   end;
  if (FCode = ACRConnect) then
   begin
    try
(*
    if not FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('Decompress/Decrypt - Error!');
{$ENDIF}
      Exit;
     end;
*)
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': IsAuthorizationBufferValid');
{$ENDIF}
{$IFDEF MsgCommunicator}
    Move(Buf^, FServerSession.Session.FUserID, SizeOf(FServerSession.Session.FUserID));
    inc(Buf, SizeOf(FServerSession.Session.FUserID));
{$ENDIF}
    Move(PChar(Buf + SizeOf(TACRConnectionParams))^, AuBufSize, SizeOf(AuBufSize));
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('SERVER> AuthorizationBuffer:');
aaWriteBufferToLog(PChar(Buf + SizeOf(TACRConnectionParams) + SizeOf(AuBufSize)),AuBufSize);
{$ENDIF}
    if not FManager.IsAuthorizationBufferValid(
              FServerSession.Session.ConnectParams.CryptoInfo,
              PChar(Buf + SizeOf(TACRConnectionParams) + SizeOf(AuBufSize)),
              AuBufSize
                                               )
    then
     begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('IsAuthorizationBufferValid - Invalid!!!');
{$ENDIF}
      try
       if FManager <> nil then
         if FServerSession <> nil then
           if FServerSession.Session <> nil then
             FManager.DoDisconnect(FServerSession.Session);
      except
       on E:Exception do
         raise EACRException.Create(40509, 'ACRConnect section - '+ErrorRDoDisconnect+E.Message);
      end;
      Exit;
     end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('IsAuthorizationBufferValid - OK!');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ConnectionParams');
{$ENDIF}
    try
    // Get ConnectParams
    ConnectionParams := PACRConnectionParams(Buf);
    FServerSession.Session.FConnectParams.PacketSize := ConnectionParams.PacketSize;
    FServerSession.Session.FConnectParams.CompressionAlgorithm := ConnectionParams.CompressionAlgorithm;
    FServerSession.Session.FConnectParams.CompressionMode := ConnectionParams.CompressionMode;
    // Get client Application name
    SetLength(FServerSession.Application, BufSize-SizeOf(TACRConnectionParams)-SizeOf(AuBufSize)-AuBufSize-1
{$IFDEF MsgCommunicator}
    -SizeOf(FServerSession.Session.FUserID)
{$ENDIF}
    );
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteBufferToLog(Buf+SizeOf(TACRConnectionParams)+SizeOf(AuBufSize)+AuBufSize,BufSize-SizeOf(TACRConnectionParams)-SizeOf(AuBufSize)-AuBufSize-1);
{$ENDIF}
    StrCopy(PChar(FServerSession.Application), Buf+SizeOf(TACRConnectionParams)+SizeOf(AuBufSize)+AuBufSize);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('Application="'+FServerSession.Application+'"');
{$ENDIF}
    // Free
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Free Buf');
{$ENDIF}
{$IFDEF MsgCommunicator}
    dec(Buf, SizeOf(FServerSession.Session.FUserID));
{$ENDIF}
    except
     on E:Exception do
      raise EACRException.Create(40510, ErrorRGetConnectParams+E.Message);
    end;
    finally
     MemoryManager.FreeAndNilMem(Buf);
    end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendConnectAckn');
{$ENDIF}
    try
    FManager.SendConnectAckn(FServerSession);
    except
     on E:Exception do
      raise EACRException.Create(40511, 'ACRConnect section - '+ErrorRSendConnectAckn+E.Message);
    end;
{$IFDEF MsgCommunicator}
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ConnectUser...');
{$ENDIF}
    if not TACRServerSession(FServerSession.Session).ConnectUser then
      FManager.Disconnect(FServerSession.Session);
{$ENDIF}
//    inc(FServerSession.CurrentRequestID);
//    ServerSession.Session.Connected := True;
    FServerSession.Connected := True;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER CONNECTED CLIENT #'+IntToStr(FServerSession.RemoteClientID));
aaWriteToLog('PacketSize='+IntToStr(FServerSession.Session.ConnectParams.PacketSize));
aaWriteToLog('CompressionAlgorithm='+IntToStr(FServerSession.Session.ConnectParams.CompressionAlgorithm));
aaWriteToLog('CompressionMode='+IntToStr(FServerSession.Session.ConnectParams.CompressionMode));
aaWriteToLog('CryptoAlgorithm='+IntToStr(FServerSession.Session.ConnectParams.CryptoInfo.CryptoAlgorithm));
aaWriteToLog('Application="'+FServerSession.Application+'"');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
   end
  else
(*
  if (FCode = ACRDisconnect) then
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Disconnect...');
{$ENDIF}
    if FServerSession.Session <> nil then
     begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DoDisconnect');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SessionID='+IntToStr(Integer(FServerSession.Session.SessionID)));
{$ENDIF}
      try
      FManager.DoDisconnect(FServerSession.Session);
      except
       on E:Exception do
        raise EACRException.Create(40509, 'ACRDisconnect section - '+ErrorRDoDisconnect+E.Message);
      end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Disconnected');
{$ENDIF}
     end;
   end
  else
*)
  if (FCode = ACRServerSessionTunning) then
   begin
    try
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRServerSessionTunning...');
{$ENDIF}
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
      Move(Buf^,FServerSession.AnswerTime,BufSize);
    finally
     MemoryManager.FreeAndNilMem(Buf);
    end;
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AnswerTime='+IntToStr(FServerSession.AnswerTime));
{$ENDIF}
    FManager.EnableNextCommand(FServerSession);
{$IFDEF DEBUG_LOG_NETWORK_TUNER}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Tuned!');
{$ENDIF}
   end
  else
  if (FCode = ACREcho)
  or (FCode = ACRTunning)
  then
   begin
    try
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Echo...');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID = '+IntToStr(FServerSession.CurrentRequestID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufSize = '+IntToStr(BufSize));
{$ENDIF}
    if (FCode <> ACRTunning) then
      if not FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
        Exit;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Decompressed BufSize = '+IntToStr(BufSize));
{$ENDIF}
    try
    FManager.SendBuffer(FServerSession.Session, Buf, BufSize, ACRNoAction);
    except
     on E:Exception do
      raise EACRException.Create(40505, 'Echo section, Code='+IntToStr(FCode)+' - '+ErrorRCannotSendEcho+E.Message);
    end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Compressed BufSize = '+IntToStr(BufSize));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID = '+IntToStr(FServerSession.CurrentRequestID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Echo sent');
{$ENDIF}
{ TODO -oLeo :
You must free the buffer you received when it is not needed any more.
Please check the following line to be present in your code: }
    finally
     MemoryManager.FreeAndNilMem(Buf);
    end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER SENT ECHO TO CLIENT #'+IntToStr(FServerSession.RemoteClientID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
   end
  else
(*
  if (FCode = ACRMessage) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call FManager.ReceiveMessage');
{$ENDIF}
    EnterCriticalSection(FManager.FCSect);
    if FServerSession.MsgControlCode = ACRExecute then
      FServerSession.MsgControlCode := ACRSuspend;
    LeaveCriticalSection(FManager.FCSect);
    try
     FManager.SendAcknowledgement(FServerSession,True);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement sent!');
{$ENDIF}
    except
     on E:Exception do
      raise EACRException.Create(40519, ErrorRAckn+E.Message);
    end;
    try
    FManager.ReceiveMessage(FServerSession, Buf, BufSize);
    except
     on E:Exception do
      raise EACRException.Create(40507,'ACRMessage section - '+ErrorRCannotReceiveMsg+E.Message);
    end;
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
      FManager.TerminateMessageThreads(FServerSession);
      FServerSession.MsgControlCode := ACRExecute;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call ServerSession.ReceiveMessage');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}

      try
      TACRNetworkSession(FServerSession.Session).ReceiveMessage(Buf, BufSize);
      except
       on E:Exception do
        raise EACRException.Create(40512,'ACRMessage section - '+ErrorRSessionReceiveMessage+E.Message);
      end;
     end;
    Exit;
   end
  else
*)
   begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - TerminateCommandThreads... - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
//    FManager.TerminateCommandThreads(FServerSession); Does not work in IDE
//    FServerSession.ControlCode := ACRSuspend;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - TerminateCommandThreads OK - GetTickCount='+IntToStr(GetTickCount));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufSize = '+IntToStr(BufSize));
{$ENDIF}
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('Server started new request #'+IntToStr(FServerSession.CurrentRequestID));
aaWriteBufferToLog(Buf, BufSize);
{$ENDIF}
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - Session.ReceiveData - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
      try
{$IFDEF MsgCommunicator}
      FServerSession.Session.ReceiveData(Buf, BufSize);
{$ELSE}
      FServerSession.Session.ReceiveData(Buf, BufSize);
{$ENDIF}
      except
       on E:Exception do
         raise EACRException.Create(40513,'Command section - '+ErrorRSessionReceiveData+E.Message);
      end;
     end;
//    FManager.TerminateCommandThreads(FServerSession); // - Could be better instead of befor ReceiveData 
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: FINISHED EXECUTE SESSION *****************************');
aaWriteToLog('==============================================================');
{$ENDIF}
 except
  on E: EACRException do
   begin
    FServerSession.ReceiveStatus := ACRNo;
    Error:=
                  ErrorRServer+ErrorRServerSessionThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (FServerSession = nil)
    or (FServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(AcrServerSessionThread,E.NativeError,Error)
    else
      TACRNetworkSession(FServerSession.Session).DoOnError(AcrServerSessionThread,E.NativeError,Error);
   end;
  on E: Exception do
   begin
    FServerSession.ReceiveStatus := ACRNo;
    Error:=
                  ErrorRServer+ErrorRServerSessionThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    try
     TACRNetworkSession(FServerSession.Session).DoOnError(AcrServerSessionThread,-1,Error);
    except
     TACRServer(FManager.FServer).DoOnConnectionError(AcrServerSessionThread,-1,Error);
    end;
   end;
 end;
 until False;
finally
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - FINISH');
{$ENDIF}
end;
end; // Execute

// TACRServerSessionThread




////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionMsgThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerSessionMsgThread.Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          Code:             Integer = ACRNoAction
                                            );
begin
 try
  FManager := Manager;
  FManager.IncThreadCount;
  FServerSession := ServerSession;
  FCode := Code;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionMsgThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionMsgThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerSessionMsgThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy;
//------------------------------------------------------------------------------
destructor TACRServerSessionMsgThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if FServerSession.Session <> nil then
    FServerSession.MsgThread := nil;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD - FINISHED');
{$ENDIF}
  if (FManager<>nil) then
    FManager.DecThreadCount;
  inherited Destroy;
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionMsgThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    if (FServerSession = nil)
    or (FServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionMsgThread,-1,
                  Error)
    else
      TACRNetworkSession(FServerSession.Session).DoOnError(
                  AcrServerSessionMsgThread,-1,
                  Error);
   end;
 end;
end;// Destroy;


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerSessionMsgThread.Execute;
var
  Buf:                 PChar;
  BufSize, AuBufSize:  Integer;
  ConnectionParams:    PACRConnectionParams;
begin
try // finally
{$IFDEF DEBUG_LOG_SERVER_SESSION_MESSAGE_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionMsgThread.Execute - START - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
 try // except
{$IFDEF DEBUG_LOG_SERVER_SESSION_MESSAGE_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionMsgThread.Execute - New Command - GetTickCount='+IntToStr(GetTickCount));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID='+IntToStr(FServerSession.CurrentRequestID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Code='+IntToStr(FCode));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ReceiveStatus = '+IntToStr(FServerSession.ReceiveStatus));
{$ENDIF}
  if (FCode = ACRMessage) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call FManager.ReceiveMessage');
{$ENDIF}
    EnterCriticalSection(FManager.FCSect);
    if FServerSession.MsgControlCode = ACRExecute then
      FServerSession.MsgControlCode := ACRSuspend;
    LeaveCriticalSection(FManager.FCSect);
    try
     FManager.SendAcknowledgement(FServerSession,True);
{$IFDEF DEBUG_LOG_SERVER_SESSION_MESSAGE_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement sent!');
{$ENDIF}
    except
     on E:Exception do
      raise EACRException.Create(40519, ErrorRAckn+E.Message);
    end;
    try
    FManager.ReceiveMessage(FServerSession, Buf, BufSize);
    except
     on E:Exception do
      raise EACRException.Create(40507,'ACRMessage section - '+ErrorRCannotReceiveMsg+E.Message);
    end;
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
      FManager.TerminateMessageThreads(FServerSession);
      FServerSession.MsgControlCode := ACRExecute;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call ServerSession.ReceiveMessage');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
      try
      TACRNetworkSession(FServerSession.Session).ReceiveMessage(Buf, BufSize);
      except
       on E:Exception do
        raise EACRException.Create(40512,'ACRMessage section - '+ErrorRSessionReceiveMessage+E.Message);
      end;
     end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: FINISHED EXECUTE SESSION *****************************');
aaWriteToLog('==============================================================');
{$ENDIF}
  end;
 except
  on E: EACRException do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionMsgThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (FServerSession = nil)
    or (FServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(AcrServerSessionMsgThread,E.NativeError,Error)
    else
      TACRNetworkSession(FServerSession.Session).DoOnError(AcrServerSessionMsgThread,E.NativeError,Error);
   end;
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionMsgThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    try
     TACRNetworkSession(FServerSession.Session).DoOnError(AcrServerSessionMsgThread,-1,Error);
    except
     TACRServer(FManager.FServer).DoOnConnectionError(AcrServerSessionMsgThread,-1,Error);
    end;
   end;
 end;
finally
{$IFDEF DEBUG_LOG_SERVER_SESSION_MESSAGE_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionMsgThread.Execute - FINISH');
{$ENDIF}
end;
end; // Execute

// TACRServerSessionMsgThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionDisconnectThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerSessionDisconnectThread.Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          CurrentRequestID: Integer = -1
                                            );
begin
 try
  FManager := Manager;
  FManager.IncThreadCount;
  FServerSession := ServerSession;
  FCurrentRequestID := CurrentRequestID;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionDisconnectThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionDisconnectThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerSessionDisconnectThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy;
//------------------------------------------------------------------------------
destructor TACRServerSessionDisconnectThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if (FManager<>nil) then
    FManager.DecThreadCount;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER SESSION THREAD - FINISHED');
{$ENDIF}
  inherited Destroy;
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionDisconnectThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    if (FServerSession = nil)
    or (FServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionDisconnectThread,-1,
                  Error)
    else
      TACRNetworkSession(FServerSession.Session).DoOnError(
                  AcrServerSessionDisconnectThread,-1,
                  Error);
   end;
 end;
end;// Destroy;


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerSessionDisconnectThread.Execute;
var
  Buf:                 PChar;
  BufSize, AuBufSize:  Integer;
  ConnectionParams:    PACRConnectionParams;
begin
try // finally
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionDisconnectThread.Execute - START - GetTickCount='+IntToStr(GetTickCount));
{$ENDIF}
 try // except
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Disconnect...');
{$ENDIF}
    try
     FManager.SendAcknowledgement(FServerSession, False, FCurrentRequestID);
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement sent!');
{$ENDIF}
    except
     on E:Exception do
      raise EACRException.Create(40519, ErrorRAckn+E.Message);
    end;
    if FServerSession.Session <> nil then
     begin
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DoDisconnect');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SessionID='+IntToStr(Integer(FServerSession.Session.SessionID)));
{$ENDIF}
      try
       FManager.DoDisconnect(FServerSession.Session);
      except
       on E:Exception do
        raise EACRException.Create(40509, ErrorRDoDisconnect+E.Message);
      end;
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Disconnected');
{$ENDIF}
     end;
 except
  on E: EACRException do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionDisconnectThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (FServerSession = nil)
    or (FServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(AcrServerSessionDisconnectThread,E.NativeError,Error)
    else
      TACRNetworkSession(FServerSession.Session).DoOnError(AcrServerSessionDisconnectThread,E.NativeError,Error);
   end;
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionDisconnectThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    try
     TACRNetworkSession(FServerSession.Session).DoOnError(AcrServerSessionDisconnectThread,-1,Error);
    except
     TACRServer(FManager.FServer).DoOnConnectionError(AcrServerSessionDisconnectThread,-1,Error);
    end;
   end;
 end;
finally
{$IFDEF DEBUG_LOG_SERVER_SESSION_THREAD}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionDisconnectThread.Execute - FINISH');
{$ENDIF}
end;
end; // Execute

// TACRServerSessionDisconnectThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionTerminatorThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerSessionTerminatorThread.Create(
                          Manager:          TACRServerConnectionManager
                                            );
begin
 try
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - START - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  FTerminatedSessions := nil;
  FManager := Manager;
  FManager.IncThreadCount;
  FTerminatedSessions := TThreadList.Create;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - STARTED - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionTerminatorThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionTerminatorThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy;
//------------------------------------------------------------------------------
destructor TACRServerSessionTerminatorThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - FINISH - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  if FManager.SessionTerminator <> nil then
    FManager.SessionTerminator := nil;
  FTerminatedSessions.Free;
  FTerminatedSessions := nil;
  inherited Destroy;
  FManager.DecThreadCount;
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - FINISHED - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionTerminatorThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionTerminatorThread,-1,
                  Error);
   end;
 end;
end;// Destroy;


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerSessionTerminatorThread.Execute;
var
  Sessions:             TList;
  Session:              TACRBaseSession;
  i:                    Integer;
begin
try
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
try
{$ENDIF}
 repeat
  if Terminated then
    Exit;
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - Lock Sessions... - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  Sessions := FTerminatedSessions.LockList;
  try
   for i:=Sessions.Count-1 downto 0 do
    begin
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - i='+IntToStr(i)+'- THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
     if Terminated then
       Exit;
     Session := Sessions.Items[i];
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - DoDisconect... - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
     try
      FManager.Disconnect(Session);
//     FManager.DoDisconnect(Session);
     except
      on E:Exception do
       raise EACRException.Create(40520, ErrorRDisconnect+E.Message);
     end;
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - Delete... - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
     Sessions.Delete(i);
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - Session Deleted! - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
    end;
  finally
   FTerminatedSessions.UnlockList;
  end;
{$IFDEF DEBUG_LOG_TERMINATOR}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - Sleep... - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  if FManager.FServer = nil then
    sleep(ACRServerSessionTerminatorSleep)
  else
    sleep(TACRServer(FManager.FServer).NetworkSettings.ServerSessionTerminatorSleep);
 until False;
{$IFDEF DEBUG_LOG_TERMINATOR}
finally
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE FINISHED - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
end;
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerSessionTerminatorThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerSessionTerminatorThread,-1,
                  Error)
    else
      TACRNetworkSession(Session).DoOnError(
                  AcrServerSessionTerminatorThread,-1,
                  Error);
   end;
 end;
end;// Execute



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerPingClientsThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerPingClientsThread.Create(
                          Manager:          TACRServerConnectionManager
                                            );
begin
 try
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - START - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  FManager := Manager;
  FManager.IncThreadCount;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - STARTED - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerPingClientsThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerPingClientsThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy;
//------------------------------------------------------------------------------
destructor TACRServerPingClientsThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - FINISH - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  if FManager.PingClientsThread <> nil then
    FManager.PingClientsThread := nil;
  inherited Destroy;
  FManager.DecThreadCount;
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - FINISHED - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerPingClientsThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerPingClientsThread,-1,
                  Error);
   end;
 end;
end;// Destroy;


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerPingClientsThread.Execute;
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
 try
{$IFDEF DEBUG_LOG_PING}
try
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
 repeat
  if Terminated then
    Exit;
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE - Lock Sessions... - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
{$ENDIF}
  for i:=0 to FManager.SessionsCount do
   begin
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> i='+IntToStr(i));
{$ENDIF}
    if Terminated then
      Exit;
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> EnterCriticalSection...');
{$ENDIF}
    EnterCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> Sessions.LockList...');
{$ENDIF}
    Sessions := FManager.FSessions.LockList;
    try
     if i>= Sessions.Count then
      begin
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> (i='+IntToStr(i)+') >= (Sessions.Count='+IntToStr(Sessions.Count)+')');
{$ENDIF}
       continue;
      end;
     ServerSession := Sessions.Items[i];
     inc(ServerSession.Status);
    finally
     FManager.FSessions.UnlockList;
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> Sessions List Unlocked!');
{$ENDIF}
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> CriticalSection Leaved!');
{$ENDIF}
    end;
    try
(*
     if not ServerSession.Session.Connected then
      begin
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> Session not Connected, i='+IntToStr(i));
{$ENDIF}
       continue;
      end;
*)
     if ServerSession.Connected then
      begin
       if (ServerSession.LastSendPingTime > ServerSession.LastReceivePingTime) then
        begin
         if ((GetTickCount - ServerSession.LastSendPingTime) > ServerSession.Session.ConnectParams.WaitForPingAnswer) then
          begin
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> DoDisconnect...');
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
           FManager.DoDisconnect(ServerSession.Session);
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> Disconnected!');
{$ENDIF}
          end;
        end
       else // ping answer received
        begin
         if ((GetTickCount - ServerSession.LastSendPingTime) > ServerSession.Session.ConnectParams.WaitForPingAnswer) then
          begin
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> Sending Ping...');
{$ENDIF}
           FManager.SendPing(ServerSession);
{$IFDEF DEBUG_LOG_PING}
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> Sent!');
{$ENDIF}
           ServerSession.LastSendPingTime := GetTickCount;
           if ServerSession.LastReceivePingTime = 0 then // first ping
             ServerSession.LastReceivePingTime := ServerSession.LastSendPingTime;
          end;
        end;
      end;
    finally
     dec(ServerSession.Status);
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> sleep...');
{$ENDIF}
     sleep(ServerSession.Session.ConnectParams.ServerPingSleep);
{$IFDEF DEBUG_LOG_PING}
//aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE> next Session');
{$ENDIF}
    end;
   end; // next Session
 until False;
{$IFDEF DEBUG_LOG_PING}
finally
aaWriteToLog('SERVER Ping Clients THREAD - EXECUTE FINISHED - THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
end;
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRServerPingClientsThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerPingClientsThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerPingClientsThread,-1,
                  Error);
   end;
 end;
end;// Execute



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerListenerThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerListenerThread.Create(
                       Manager:       TACRServerConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       FromHost:      String;
                       FromPort:      Integer
                                            );
begin
 try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerListenerThread.Create> START - '+IntToStr(GetTickCount));
{$ENDIF}
  FManager := Manager;
  self.Buffer := Buffer;
  self.BufferSize := BufferSize;
  self.FromHost := FromHost;
  self.FromPort := FromPort;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerListenerThread.Create> inherited... - '+IntToStr(GetTickCount));
{$ENDIF}
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
  FManager.IncThreadCount;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerListenerThread.Create> FINISH - '+IntToStr(GetTickCount));
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerListenerThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerListenerThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerListenerThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if SessionFound then
   begin
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FOUND');
{$ENDIF}
    if ServerSession.ListeningThreads <> nil then
      ServerSession.ListeningThreads.Remove(self);
    if ServerSession.MsgListeningThreads <> nil then
      ServerSession.MsgListeningThreads.Remove(self);
   end;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - inherited Destroy');
{$ENDIF}
  if (FManager<>nil) then
    FManager.DecThreadCount;
  inherited Destroy;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISHED');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerListenerThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerListenerThread,-1,
                  Error);
   end;
 end;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerListenerThread.Execute;
label
  KillPacket;
var
  Header:               PACRPacketHeader;
  Packets:              TList;
  Packet:               PACRPacket;
  Sessions:             TList;
  i:                    Integer;
  Error:                String;
function MsgAllPacketsReceived: Boolean;
var
  i:                    Integer;
  Packets:              TList;
begin
// Are all packets received?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all message packets received?');
{$ENDIF}
 Result := True;
 Packets := ServerSession.MsgPackets.LockList;
 try
  for i:=0 to Packets.Count-1 do
   if Packets.Items[i] = nil then
    begin
     Result := False;
     break;
    end;
  finally
  ServerSession.MsgPackets.UnlockList;
  end;
end; // function AllPacketsReceived: Boolean;

function AllPacketsReceived: Boolean;
var
  i:                    Integer;
  Packets:              TList;
begin
// Are all packets received?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all command packets received?');
{$ENDIF}
 Result := True;
 Packets := ServerSession.Packets.LockList;
 try
  for i:=0 to Packets.Count-1 do
   if Packets.Items[i] = nil then
    begin
     Result := False;
     break;
    end;
  finally
  ServerSession.Packets.UnlockList;
  end;
{$IFDEF LOG_SERVER_THREADS}
if Result then
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes')
else
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
end; // function AllPacketsReceived: Boolean;

begin // TACRServerListenerThread.Execute;
 try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerListenerThread.Execute> START -'+IntToStr(GetTickCount));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: '+IntToStr(GetTickCount)+' > Listener started');
aaWriteToLog('==============================================================');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** SERVER LISTENER START *****');
{$ENDIF}
 try
  Header := Pointer(Buffer);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('           Header.PacketID         = '+IntToStr(Header.PacketID));
aaWriteToLog('           Header.ControlCode      = '+IntToStr(Header.ControlCode));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connections');
{$ENDIF}
//EnterCriticalSection(FManager.FCSect);
{$IFDEF LOG_SERVER_THREADS}
//aaWriteToLog('ENTERED!');
{$ENDIF}
// Check for errors
  if CheckSum(Buffer, BufferSize) <> Header.CheckSum then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid CheckSum');
{$ENDIF}
    FManager.PacketResendRequest(Buffer, FManager.FNetwork, FromHost, FromPort);
    goto KillPacket;
   end;
// Check sign
  if (Header.Signature <> ACRClientPacketSign) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid Signature');
{$ENDIF}
    goto KillPacket;
   end;
// Check ServerID
  if (Header.Recepient <> TACRServer(FManager.FServer).ServerID) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid ServerID');
{$ENDIF}
    goto KillPacket;
   end;
  if    // first packet in multi-packet connect request
  ((Header.ControlCode=ACRConnect) and (Header.PacketID=0))
  or    // single-packet connect request
  ((Header.ControlCode=(ACRConnect+ACRLastPacket)) and (Header.PacketID=0))
  then
   begin // Connect
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connect');
{$ENDIF}
    SessionFound := False;
    Sessions := FManager.FSessions.LockList;
    try
     for i:=0 to Sessions.Count-1 do
      begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Check session '+IntToStr(i)+' of '+IntToStr(Sessions.Count));
{$ENDIF}
       ServerSession := Sessions.Items[i];
       if (ServerSession.ClientSessionID = Header.SessionID)
       and (ServerSession.ConnectionID = Header.ConnectionID)
       and (ServerSession.RemoteClientID = Header.Sender)
       then
        begin
         SessionFound := True;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ServerSession),6)+', Client SessionID='+IntToStr(ServerSession.ClientSessionID));
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ServerSession),6)+', Server SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
         break;
        end;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
    if not SessionFound then
     begin // First connect
// new ServerSession
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New ServerSession');
{$ENDIF}
      New(ServerSession);
      ServerSession.Connected := False;
      ServerSession.AnswerTime := 0;
      ServerSession.LastSendPingTime := 0;
      ServerSession.LastReceivePingTime := 0;
      ServerSession.Status := ACRVacant;
      ServerSession.ReceiveStatus := ACRStart;
      ServerSession.RemoteHost := FromHost;
      ServerSession.RemotePort := FromPort;
      ServerSession.ConnectionID := Header.ConnectionID;
      ServerSession.ClientSessionID := Header.SessionID;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ServerSession),6)+', Client SessionID='+IntToStr(ServerSession.ClientSessionID));
{$ENDIF}
      ServerSession.RemoteClientID := Header.Sender;
      ServerSession.Session := TACRServerSession.Create(FManager.FServer);
      ServerSession.Session.FConnectParams.CryptoInfo := TACRServer(FManager.FServer).CryptoParams.GetCryptoParams;
      ServerSession.Thread := nil;
      ServerSession.MsgThread := nil;
      ServerSession.DisconnectThread := nil;
      ServerSession.ListeningThreads := TThreadList.Create;
      ServerSession.ControlCode := ACRExecute;
      ServerSession.ListeningThreads.Add(self);
      ServerSession.MsgListeningThreads := TThreadList.Create;
      ServerSession.MsgControlCode := ACRExecute;
// Set SessionID
      EnterCriticalSection(FManager.FCSect);
      ServerSession.Session.SessionID := FManager.FSessionID;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ServerSession),6)+', Server SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
      inc(FManager.FSessionID);
      if FManager.FSessionID = INVALID_SESSION_ID then
       inc(FManager.FSessionID);
      LeaveCriticalSection(FManager.FCSect);
// Create PacketIDsToResend
      ServerSession.PacketIDsToResend := TACRThreadIntArray.Create;
      ServerSession.MsgPacketIDsToResend := TACRThreadIntArray.Create;
// Create new packets list
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packets list create');
{$ENDIF}
      ServerSession.Packets := TThreadList.Create;
      Packets := ServerSession.Packets.LockList;
      try
       Packets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
      finally
       ServerSession.Packets.UnlockList;
      end;
      ServerSession.MsgPackets := TThreadList.Create;
      Packets := ServerSession.MsgPackets.LockList;
      try
       Packets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
      finally
       ServerSession.MsgPackets.UnlockList;
      end;
// add new session to list
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': add new session to list');
{$ENDIF}
      FManager.FSessions.Add(ServerSession);
      SessionFound := True;
     end
    else // not first connect - reset and resend connection info
     begin
      if ServerSession.Thread <> nil then // Is buffer receiving now?
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer is receiving now');
{$ENDIF}
        goto KillPacket;
       end;
      if (ServerSession.ControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Not Execute - Terminate ListeningThread, ServerSession.ControlCode='+IntToStr(ServerSession.ControlCode));
{$ENDIF}
        goto KillPacket;
       end;
      if (ServerSession.CurrentRequestID > Header.CurrentRequestID) then // packet from old connection request
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - SendAckn');
{$ENDIF}
        FManager.SendConnectAckn(ServerSession, Header.CurrentRequestID);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - Terminate ListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
      if ServerSession.MsgThread <> nil then
        FManager.TerminateServerThread(@ServerSession.MsgThread,AcrServerListenerThread,ErrorRServerSessionMsgThread,
            ServerSession.Session.ConnectParams.WaitForServerSessionThreadTimeOut);
      if ServerSession.DisconnectThread <> nil then
        FManager.TerminateServerThread(@ServerSession.DisconnectThread,AcrServerListenerThread,ErrorRServerSessionDisconnectThread,
            ServerSession.Session.ConnectParams.WaitForServerSessionThreadTimeOut);
//      ServerSession.MsgThread := nil;
      ServerSession.ListeningThreads.Add(self);
// Set PacketIDsToResend
      ServerSession.PacketIDsToResend.Lock;
      ServerSession.PacketIDsToResend.SetSize(0);
      ServerSession.PacketIDsToResend.Unlock;
      ServerSession.MsgPacketIDsToResend.Lock;
      ServerSession.MsgPacketIDsToResend.SetSize(0);
      ServerSession.MsgPacketIDsToResend.Unlock;
(*
// Delete old packets from list
      Packets := ServerSession.Packets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ServerSession.Packets.UnlockList;
      end;
      Packets := ServerSession.MsgPackets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ServerSession.Packets.UnlockList;
      end;
*)
     end;
// Put this first packet
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New Packet');
{$ENDIF}
    New(Packet);
    Packet.Buffer := Buffer;
    Packet.BufferSize := BufferSize;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet add');
{$ENDIF}
    ServerSession.Packets.Add(Packet);
    ServerSession.CurrentRequestID := Header.CurrentRequestID;
    ServerSession.ClientMessageID := 0;
    ServerSession.ServerMessageID := 0;
// Get connection parameters
    if Header.ControlCode=ACRConnect+ACRLastPacket then
     ServerSession.ReceiveStatus := ACRFull;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSessionThread.Create');
{$ENDIF}
    ServerSession.Thread := TACRServerSessionThread.Create(FManager,
                                                  ServerSession, ACRConnect);
    Sleep(0);
    Exit;
   end; // Connect
// verify ConnectionID existing and SessionID existing in this connection
  SessionFound := False;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sessions');
{$ENDIF}
  Sessions := FManager.FSessions.LockList;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sessions Locked');
{$ENDIF}
  try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           @Sessions='+IntToStr(Integer(Sessions)));
aaWriteToLog('           Sessions.Count='+IntToStr(Integer(Sessions.Count)));
{$ENDIF}
   for i:=0 to Sessions.Count-1 do
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           i='+IntToStr(i));
aaWriteToLog('           Sessions.Count='+IntToStr(Integer(Sessions.Count)));
{$ENDIF}
     ServerSession := Sessions.Items[i];
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           @ServerSession='+IntToStr(Integer(ServerSession)));
{$ENDIF}
     if (ServerSession.ConnectionID = Header.ConnectionID)
     and ( (ServerSession.Session.SessionID = Header.SessionID)
           or ( (ServerSession.ClientSessionID = Header.SessionID)
                and  ( (Header.ControlCode = ACRConnect)
                    or (Header.ControlCode = (ACRConnect + ACRLastPacket))
                     )
              )
         )
     then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FManager.FSessions.UnlockList;
  end;
  if not SessionFound then
  if (Header.ControlCode <> ACRConnect) then // packet is not from Connect request
  if (Header.ControlCode <> (ACRConnect+ACRLastPacket)) then
   begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session not found');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid SessionID or ConnectionID');
{$ENDIF}
    // client try to continue work with new server without loggin first
    // instead of old server it was loggin into
    New(ServerSession);
    ServerSession.Connected := False;
    ServerSession.AnswerTime := 0;
    ServerSession.LastSendPingTime := 0;
    ServerSession.LastReceivePingTime := 0;
    ServerSession.CurrentRequestID := Header.CurrentRequestID;
    ServerSession.RemoteClientID := Header.Sender;
    ServerSession.ConnectionID := Header.ConnectionID;
    ServerSession.Session := TACRServerSession.Create(FManager.FServer);
    ServerSession.Session.SessionID := Header.SessionID;
    ServerSession.RemoteHost := FromHost;
    ServerSession.RemotePort := FromPort;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Try to send ACKN to old client...');
{$ENDIF}
    if Header.ControlCode = ACRDisconnect then
      FManager.SendAcknowledgement(ServerSession)
    else
      FManager.SendDisconnectRequest(ServerSession, False);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session.Free...');
{$ENDIF}
    ServerSession.Session.Free;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Dispose...');
{$ENDIF}
    Dispose(ServerSession);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Old client is disconnected');
{$ENDIF}
    goto KillPacket;
   end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session found');
{$ENDIF}
// verify Sender
  if ServerSession.RemoteClientID <> Header.Sender then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid Sender');
{$ENDIF}
    goto KillPacket;
   end;
// check CurrentRequestID
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
{$ENDIF}
  if (Header.ControlCode = ACRMessageReceived)
  or (Header.ControlCode = ACRMessagePacketResendRequest)
  then
   begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ServerMessageID = '+IntToStr(ServerSession.ServerMessageID));
{$ENDIF}
    if ServerSession.ServerMessageID <> Header.CurrentRequestID then
      goto KillPacket;
   end
  else
  if (Header.ControlCode = ACRMessage)
  or (Header.ControlCode = (ACRMessage + ACRLastPacket))
  then
   begin
    if ServerSession.ClientMessageID > Header.CurrentRequestID then // old message
     begin
      goto KillPacket;
     end;
    if ServerSession.ClientMessageID < Header.CurrentRequestID then // lost acknowledgement from client
      ServerSession.MsgSendStatus := ACRSent;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Listening Thread in '+IntToStr(Integer(ServerSession)));
{$ENDIF}
    if ServerSession.MsgListeningThreads = nil then
     begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': MsgListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
     begin
      if (ServerSession.MsgControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminate ACRListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
      ServerSession.MsgListeningThreads.Add(self);
     end;
   end
  else
   begin
    if Header.ControlCode <> ACRDisconnect then
    if Header.ControlCode <> ACRPing then
      if ServerSession.CurrentRequestID > Header.CurrentRequestID then // old command
       begin
        if ServerSession.CurrentRequestID = (Header.CurrentRequestID+1) then // previous request
          if (Header.ControlCode >= ACRLastPacket) then // lost akcn
            FManager.SendAcknowledgement(ServerSession,False,Header.CurrentRequestID);
        goto KillPacket;
       end;
{
    if ServerSession.CurrentRequestID < Header.CurrentRequestID then // lost acknowledgement from client
      ServerSession.SendStatus := ACRSent;
}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Listening Thread in '+IntToStr(Integer(ServerSession)));
{$ENDIF}
    if ServerSession.ListeningThreads = nil then
     begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
     begin
      if Header.ControlCode <> ACRDisconnect then
      if Header.ControlCode <> ACRPing then
      if (Header.ControlCode <> ACRAllPacketsReceived) then
      if (ServerSession.ControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminate ListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
      ServerSession.ListeningThreads.Add(self);
     end;
   end;
// Process service packet
  case Header.ControlCode of
   ACRDisconnect:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRDisconnect');
{$ENDIF}
//     FManager.CommandReceived(ServerSession,Header.ControlCode,Header.CurrentRequestID);
{
     EnterCriticalSection(FManager.FCSect);
     ServerSession.ControlCode := ACRTerminate;
     ServerSession.MsgControlCode := ACRSuspend;
     LeaveCriticalSection(FManager.FCSect);
}
(*
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement...');
{$ENDIF}
     FManager.SendAcknowledgement(ServerSession, False, Header.CurrentRequestID);
*)
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement has been sent');
{$ENDIF}
     if ServerSession.Session <> nil then // nothing to delete
       if ServerSession.DisconnectThread = nil then // are not deleting
         ServerSession.DisconnectThread :=
              TACRServerSessionDisconnectThread.Create(FManager, ServerSession,
                                                      Header.CurrentRequestID);
     goto KillPacket;
    end;
   ACRPing:
    begin
     ServerSession.LastReceivePingTime := GetTickCount;
     goto KillPacket;
    end;
   ACRPacketResendRequest:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRPacketResendRequest');
{$ENDIF}
     ServerSession.PacketIDsToResend.Lock;
     ServerSession.PacketIDsToResend.Add(Header.PacketID);
     ServerSession.PacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Resend Request');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     goto KillPacket;
    end;
   ACRMessagePacketResendRequest:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRMessagePacketResendRequest');
{$ENDIF}
     ServerSession.MsgPacketIDsToResend.Lock;
     ServerSession.MsgPacketIDsToResend.Add(Header.PacketID);
     ServerSession.MsgPacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Message Resend Request');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     goto KillPacket;
    end;
   ACRAllPacketsReceived:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRAllPacketsReceived');
{$ENDIF}
     ServerSession.SendStatus := ACRSent;
     ServerSession.PacketIDsToResend.Lock;
     ServerSession.PacketIDsToResend.SetSize(0);
     ServerSession.PacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Command Acknowledgement');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     sleep(0);
     goto KillPacket;
    end;
   ACRMessageReceived:
    begin
     ServerSession.MsgSendStatus := ACRSent;
     ServerSession.MsgPacketIDsToResend.Lock;
     ServerSession.MsgPacketIDsToResend.SetSize(0);
     ServerSession.MsgPacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Message Acknowledgement');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     goto KillPacket;
    end;
   ACRMessage,
   (ACRMessage+ACRLastPacket):
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Data packet');
{$ENDIF}
//     if Header.ControlCode < ACRLastPacket then // Enable LastPacket to start session thread
     if (Header.CurrentRequestID <= ServerSession.ClientMessageID) then // lost acknowledgement
     if (ServerSession.MsgReceiveStatus = ACRFull) then // lost acknowledgement
      begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full Answer!');
{$ENDIF}
//       FManager.SendAcknowledgement(ServerSession, True, Header.CurrentRequestID);
       goto KillPacket; // Do not replace correct packets with doubles
      end;
// new request?
// Is this packet already stored?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Is this packet already stored?');
{$ENDIF}
     Packets := ServerSession.MsgPackets.LockList;
     if Packets.List = nil then // list absent
      begin
       ServerSession.MsgPackets.UnlockList;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': list absent');
{$ENDIF}
       goto KillPacket;
      end
     else // list exists
      if ((Header.PacketID+1) <= Packets.Count) then // old packet
       if (Packets.Items[Header.PacketID] <> nil) then // already received
        begin
         ServerSession.MsgPackets.UnlockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(Packets.Count)));
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing message packet');
aaWriteToLog('           @Items='+IntToStr(Integer(Packets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('Server received existing packet');
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('MsgAllPacketsReceived?');
{$ENDIF}
         if MsgAllPacketsReceived then
          begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('Yes');
{$ENDIF}
           FManager.MessageReceived(ServerSession);
          end;
{
         else
           if (Header.ControlCode >= ACRLastPacket) then
             ServerSession.ReceiveStatus := ACRNotFull; // start requesting
}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Old message packet will be killed');
{$ENDIF}
         goto KillPacket; // Do not replace correct packets with doubles
        end;
// First packet in new request?
     if Packets.Count = 0 then
       ServerSession.MsgReceiveStatus := ACRStart;
     ServerSession.MsgPackets.UnlockList;
// Add correct received data packet
     if Terminated then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
        goto KillPacket;
       end;
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
if ((Header.PacketID+1) <= Packets.Count) then // old packet
if (Packets.Items[Header.PacketID] = nil) then // not received yet
aaWriteToLog('SERVER received packet #'+IntToStr(Header.PacketID));
{$ENDIF}
     Packets := ServerSession.MsgPackets.LockList;
     try
      if Packets.Count < (Header.PacketID + 1) then
        Packets.Count := Header.PacketID + 1; // List fills hole by Nil
      if (ServerSession.MsgReceiveStatus <> ACRFull) then
        Packets.Items[Header.PacketID] := Packet
      else
       begin
        Dispose(Packet);
        Packet := nil;
       end;
     finally
      ServerSession.MsgPackets.UnlockList;
     end;
     if Packet = nil then
       goto KillPacket;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': server has added new packet');
{$ENDIF}
(*
     if Terminated then
      begin
//       Dispose(Packet);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
       goto KillPacket;
      end;
*)
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER has added new packet');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// test last packet
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Last Packet');
{$ENDIF}
     if (Header.ControlCode >= ACRLastPacket) then
      begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       if MsgAllPacketsReceived then
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
         ServerSession.MsgReceiveStatus := ACRFull;  // Allow to extract buffer
        end
       else
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
         // start resend requesting
         ServerSession.MsgReceiveStatus := ACRNotFull;
        end;
      end; // Last Packet
// Full Answer
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientMessageID = '+IntToStr(ServerSession.ClientMessageID));
{$ENDIF}
     if ServerSession.MsgReceiveStatus = ACRFull then
       if Terminated then
         goto KillPacket
       else
         FManager.MessageReceived(ServerSession);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End messge packet process
     Exit;
    end; // Message packet process
   ACREcho,
   ACRTunning,
   ACRServerSessionTunning,
   ACRConnect, // for packets with ID>0, not for first packet
   ACRNewRequest,
   ACRLastPacket..(ACRMessage+ACRLastPacket-1),
   (ACRMessage+ACRLastPacket+1)..127,
   ACRNoAction:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Data packet - '+IntToStr(GetTickCount));
{$ENDIF}
// Check for if the previous data received fully:
{
     if Header.ControlCode then
      begin
       discontinue the previous sending;
      end;
Suppose client session should not send new request if it did not receive the previous fully
}
     if (Header.CurrentRequestID <= ServerSession.CurrentRequestID) then // lost acknowledgement
     if (ServerSession.ReceiveStatus = ACRFull) then // lost acknowledgement
      begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full Answer!');
{$ENDIF}
       if (Header.ControlCode <> ACRConnect)
       and (Header.ControlCode <> (ACRConnect+ACRLastPacket))
       then
         FManager.SendAcknowledgement(ServerSession, False, Header.CurrentRequestID);
       goto KillPacket; // Do not replace correct packets with doubles
      end;
// new request?
(*
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('new request?');
{$ENDIF}
     if Header.ControlCode >= ACRNewRequest then
      if ServerSession.CurrentRequestID < Header.CurrentRequestID then
       begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('Yes');
{$ENDIF}
        ServerSession.ReceiveStatus := ACRStart;
        ServerSession.CurrentRequestID := Header.CurrentRequestID;
        Packets := ServerSession.Packets.LockList;
        try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('Packets.Count='+IntToStr(Packets.Count));
{$ENDIF}
         for i:=0 to Packets.Count-1 do
           Packets.Items[i] := nil;
         Packets.Count := 0;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('Packets.Count='+IntToStr(Packets.Count));
{$ENDIF}
        finally
         ServerSession.Packets.UnlockList;
        end;
       end;
*)
// Is this packet already stored?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Is this packet already stored?');
{$ENDIF}
     Packets := ServerSession.Packets.LockList;
     if Packets.List = nil then // list absent
      begin
       ServerSession.Packets.UnlockList;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': list absent');
{$ENDIF}
       goto KillPacket;
      end
     else // list exists
      if ((Header.PacketID+1) <= Packets.Count) then // old packet
       if (Packets.Items[Header.PacketID] <> nil) then // already received
        begin
         ServerSession.Packets.UnlockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(Packets.Count)));
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(Packets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('Server received existing packet');
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('AllPacketsReceived?');
{$ENDIF}
         if AllPacketsReceived then
          begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': All Packets on Repeated Packet - '+IntToStr(GetTickCount));
{$ENDIF}
           FManager.CommandReceived(ServerSession, Header.ControlCode, Header.CurrentRequestID);
          end;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Old command packet will be killed');
{$ENDIF}
         goto KillPacket; // Do not replace correct packets with doubles
        end;
// First packet in new request?
     if Packets.Count = 0 then
      begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': First packet in new request!');
{$ENDIF}
       ServerSession.ReceiveStatus := ACRStart;
      end;
     ServerSession.Packets.UnlockList;
// Add correct received data packet
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add correct received data packet');
{$ENDIF}
     if Terminated then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
        goto KillPacket;
       end;
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
if ((Header.PacketID+1) <= Packets.Count) then // old packet
if (Packets.Items[Header.PacketID] = nil) then // not received yet
aaWriteToLog('SERVER received packet #'+IntToStr(Header.PacketID));
{$ENDIF}
     Packets := ServerSession.Packets.LockList;
     try
      if Packets.Count < (Header.PacketID + 1) then
        Packets.Count := Header.PacketID + 1; // List fills hole by Nil
      if (ServerSession.ReceiveStatus <> ACRFull) then
        Packets.Items[Header.PacketID] := Packet
      else
       begin
        Dispose(Packet);
        Packet := nil;
       end;
     finally
      ServerSession.Packets.UnlockList;
     end;
     if Packet = nil then
       goto KillPacket;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': server has added new packet');
{$ENDIF}
(*
      if Terminated
      then
       begin
//        Dispose(Packet);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
        goto KillPacket;
       end;
*)
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER has added new packet');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// test last packet
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Last Packet');
{$ENDIF}
     if (Header.ControlCode >= ACRLastPacket) then
      begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       if AllPacketsReceived then
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': All Packets on Last Packet - '+IntToStr(GetTickCount));
{$ENDIF}
         FManager.CommandReceived(ServerSession, Header.ControlCode, Header.CurrentRequestID);
        end
       else
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
         // start resend requesting
         ServerSession.ReceiveStatus := ACRNotFull;
        end;
      end; // Last Packet
// End data packet process
     Exit;
    end // data packet process
   else
     raise EACRException.Create(40022, ErrorRUnknownControlCode, [Header.ControlCode]);
   end; // packet process
KillPacket:
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER KILLED PACKET # '+ IntToStr(Header.PacketID)+ ' with control code ' +IntToStr(Header.ControlCode));
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': FreeAndNilMem(Buffer)');
{$ENDIF}
  MemoryManager.FreeAndNilMem(Buffer);
 finally
//LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: '+IntToStr(GetTickCount)+' > Listener finished');
aaWriteToLog('==============================================================');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISHED');
{$ENDIF}
 end;
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRListenerThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerListenerThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerListenerThread,-1,
                  Error);
   end;
 end;
end; // Execute

// TACRServerListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerResendRequestThread.Create(
                       Manager:           TACRServerConnectionManager
                                            );
begin
 try
  FManager := Manager;
  FManager.IncThreadCount;
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('Server RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRResendRequestThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerResendRequestThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerResendRequestThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('Server RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if FManager.ResendRequestThread <> nil then
    FManager.ResendRequestThread := nil;
  inherited Destroy;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('Server RESENDING THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRResendRequestThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerResendRequestThread,-1,
                  Error);
   end;
 end;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerResendRequestThread.Execute;
var
  i, j, k,
  PacketsCount:       Integer;
  Packets:            TList;
  Packet:             PACRPacket;
  AllPacketsReceived: Boolean;
  Sessions:           TList;
  ServerSession:      PACRSrvrSession;
  Header:             TACRPacketHeader;
  PHeader:            PACRPacketHeader;
  Code:               Integer;

function IsRequestNeeded(ServerSession: PACRSrvrSession): Boolean;
begin
  Result := True;
  if ServerSession.ReceiveStatus <> ACRNotFull then
   begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRNotFull <> (ServerSession.ReceiveStatus='+IntToStr(ServerSession.ReceiveStatus)+')');
{$ENDIF}
    Result := False;
    Exit;
   end;
  if ServerSession.ControlCode = ACRTerminate then
   begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ControlCode = ACRTerminate');
{$ENDIF}
    Result := False;
    Exit;
   end;
end;

begin
 try
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** SERVER RESEND REQUEST START *****');
{$ENDIF}
  j:=0;
  repeat
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session # j='+IntToStr(j));
{$ENDIF}
    if Terminated then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sessions loop');
{$ENDIF}
      Exit;
     end;
    if FManager.FServer = nil then
      sleep(1 + (ACRServerRequestDelay
                                              div (FManager.SessionsCount+1)))
    else
      sleep(1 + (TACRServer(FManager.FServer).NetworkSettings.ServerRequestDelay
                                              div (FManager.SessionsCount+1)));
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Lock Sessions...');
{$ENDIF}
    Sessions := FManager.FSessions.LockList;
    try
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Locked!');
{$ENDIF}
     if j >= Sessions.Count then
      begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': j >= (Sessions.Count='+IntToStr(Sessions.Count)+')');
{$ENDIF}
       j := 0;
       continue;
      end;
     ServerSession := Sessions.Items[j];
     inc(ServerSession.Status);
     if not (IsRequestNeeded(ServerSession)) then
      begin
       dec(ServerSession.Status);
       continue;
      end;
    finally
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': UnLock Sessions');
{$ENDIF}
     FManager.FSessions.UnlockList;
    end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerResendRequestThread');
{$ENDIF}
    // Make PacketHeader
    Header.Signature := ACRServerPacketSign;
    Header.Recepient := ServerSession.RemoteClientID;
    EnterCriticalSection(FManager.FCSect);
    Header.Sender := FManager.FNetwork.FLocalClient;
    LeaveCriticalSection(FManager.FCSect);
    Header.ConnectionID := ServerSession.ConnectionID;
    Header.SessionID := ServerSession.Session.SessionID;
    Header.PacketID := 0;
    Header.ControlCode := ACRPacketResendRequest;
    Header.CurrentRequestID := ServerSession.CurrentRequestID;
    // Search absent packet and send request to resend it
    Packets := ServerSession.Packets.LockList;
    AllPacketsReceived := True;
    try
     PacketsCount := Packets.Count;
    finally
     ServerSession.Packets.UnlockList;
    end;
    dec(ServerSession.Status);
    for i:=0 to PacketsCount-1 do
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketsCount='+IntToStr(PacketsCount));
{$ENDIF}
      if Terminated then
       begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in packets loop');
{$ENDIF}
        Exit;
       end;
      if ServerSession.ControlCode = ACRTerminate then
        break;
      inc(ServerSession.Status);
      if ServerSession.Packets <> nil then
       begin
        Packets := ServerSession.Packets.LockList;
        if Packets.List = nil then
         begin
          ServerSession.Packets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Count <= i then
         begin
          ServerSession.Packets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Items[i] <> nil then
         begin
          ServerSession.Packets.UnlockList;
          dec(ServerSession.Status);
         end
        else // send request
         begin
          ServerSession.Packets.UnlockList;
          AllPacketsReceived := False;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketResendRequest, i='+IntToStr(i));
{$ENDIF}
          FManager.PacketResendRequest(@Header, FManager.FNetwork, ServerSession.RemoteHost, ServerSession.RemotePort, i);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been resent');
{$ENDIF}
          if not (IsRequestNeeded(ServerSession)) then
           begin
            dec(ServerSession.Status);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('break');
{$ENDIF}
            break;
           end;
          dec(ServerSession.Status);
          Sleep(ServerSession.Session.ConnectParams.ServerReceiveSleep);
(*
          for k:=0 to (ACRServerRequestDelay div 2) do
           begin // Pause
            if not (IsRequestNeeded(ServerSession)) then
              break;
            if Terminated then
             begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sleep loop');
{$ENDIF}
              Exit;
             end;
            Sleep(2);
           end; // Pause
*)
         end;// send request
       end
      else
       dec(ServerSession.Status);
     end; // next packet
    if AllPacketsReceived then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
{
      EnterCriticalSection(FManager.FCSect);
      if ServerSession.ControlCode = ACRExecute then
        ServerSession.ControlCode := ACRSuspend;
      LeaveCriticalSection(FManager.FCSect);
}
//      ServerSession.ReceiveStatus := ACRFull;
      Packets := ServerSession.Packets.LockList;
      try
       if Packets.Count > 0 then
        begin
         Packet := Packets.Items[0];
         PHeader := PACRPacketHeader(Packet.Buffer);
         Code := PHeader.ControlCode;
        end;
      finally
       ServerSession.Packets.UnlockList;
      end;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': All Packets on Request thread - '+IntToStr(GetTickCount));
{$ENDIF}
       FManager.CommandReceived(ServerSession, Code, ServerSession.CurrentRequestID);
     end;
    inc(j);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Get New Session');
{$ENDIF}
   until False; // Get new Session
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRResendRequestThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerResendRequestThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerResendRequestThread,-1,
                  Error);
   end;
 end;
end;// Execute



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerMsgResendRequestThread.Create(
                       Manager:           TACRServerConnectionManager
                                            );
begin
 try
  FManager := Manager;
  FManager.IncThreadCount;
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('Server MESSAGE RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRMsgResendRequestThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRCreate+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerMsgResendRequestThread,-1,
                  Error);
    Destroy;
   end;
 end;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerMsgResendRequestThread.Destroy;
begin
 try
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('Server MESSAGE RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if FManager.MsgResendRequestThread <> nil then
    FManager.MsgResendRequestThread := nil;
  inherited Destroy;
{$IFDEF DEBUG_LOG_SERVER_DISCONNECT}
aaWriteToLog('Server MESSAGE RESENDING THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRMsgResendRequestThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRDestroy+
                  E.Message;
    TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerMsgResendRequestThread,-1,
                  Error);
   end;
 end;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerMsgResendRequestThread.Execute;
var
  i, j, k,
  PacketsCount:       Integer;
  Packets:            TList;
  AllPacketsReceived: Boolean;
  Sessions:           TList;
  ServerSession:      PACRSrvrSession;
  Header:             TACRPacketHeader;

function IsRequestNeeded(ServerSession: PACRSrvrSession): Boolean;
begin
  Result := True;
  if ServerSession.MsgReceiveStatus <> ACRNotFull then
   begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRNotFull <> (ServerSession.MsgReceiveStatus='+IntToStr(ServerSession.MsgReceiveStatus)+')');
{$ENDIF}
    Result := False;
    Exit;
   end;
  if ServerSession.MsgControlCode = ACRTerminate then
   begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.MsgControlCode = ACRTerminate');
{$ENDIF}
    Result := False;
    Exit;
   end;
end;

begin
 try
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** SERVER MESSAGE RESEND REQUEST START *****');
{$ENDIF}
  j:=0;
  repeat
    if Terminated then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sessions loop');
{$ENDIF}
      Exit;
     end;
    if FManager.FServer = nil then
      sleep(1 + (ACRServerRequestDelay div (FManager.SessionsCount+1)))
    else
      sleep(1 + (TACRServer(FManager.FServer).NetworkSettings.ServerRequestDelay div (FManager.SessionsCount+1)));
    Sessions := FManager.FSessions.LockList;
    try
     if j >= Sessions.Count then
      begin
       j := 0;
       continue;
      end;
     ServerSession := Sessions.Items[j];
     inc(ServerSession.Status);
     if not (IsRequestNeeded(ServerSession)) then
      begin
       dec(ServerSession.Status);
       continue;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerMsgResendRequestThread');
{$ENDIF}
    // Make PacketHeader
    Header.Signature := ACRServerPacketSign;
    Header.Recepient := ServerSession.RemoteClientID;
    EnterCriticalSection(FManager.FCSect);
    Header.Sender := FManager.FNetwork.FLocalClient;
    LeaveCriticalSection(FManager.FCSect);
    Header.ConnectionID := ServerSession.ConnectionID;
    Header.SessionID := ServerSession.Session.SessionID;
    Header.PacketID := 0;
    Header.ControlCode := ACRMessagePacketResendRequest;
    Header.CurrentRequestID := ServerSession.ClientMessageID;
    // Search absent packet and send request to resend it
    Packets := ServerSession.MsgPackets.LockList;
    AllPacketsReceived := True;
    try
     PacketsCount := Packets.Count;
    finally
     ServerSession.MsgPackets.UnlockList;
    end;
    dec(ServerSession.Status);
    for i:=0 to PacketsCount-1 do
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketsCount='+IntToStr(PacketsCount));
{$ENDIF}
      if Terminated then
       begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketsCount='+IntToStr(PacketsCount));
{$ENDIF}
        Exit;
       end;
      if ServerSession.MsgControlCode = ACRTerminate then
        break;
      inc(ServerSession.Status);
      if ServerSession.MsgPackets <> nil then
       begin
        Packets := ServerSession.MsgPackets.LockList;
        if Packets.List = nil then
         begin
          ServerSession.MsgPackets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Count <= i then
         begin
          ServerSession.MsgPackets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Items[i] <> nil then
         begin
          ServerSession.MsgPackets.UnlockList;
          dec(ServerSession.Status);
         end
        else // send request
         begin
          ServerSession.MsgPackets.UnlockList;
          AllPacketsReceived := False;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketMsgResendRequest, i='+IntToStr(i));
{$ENDIF}
          FManager.PacketResendRequest(@Header, FManager.FNetwork, ServerSession.RemoteHost, ServerSession.RemotePort, i, True);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been resent');
{$ENDIF}
          if not (IsRequestNeeded(ServerSession)) then
           begin
            dec(ServerSession.Status);
            break;
           end;
          dec(ServerSession.Status);
          Sleep(ServerSession.Session.ConnectParams.ServerReceiveSleep);
(*
          for k:=0 to (ACRServerRequestDelay div 2) do
           begin // Pause
            if not (IsRequestNeeded(ServerSession)) then
              break;
            if Terminated then
             begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sleep loop');
{$ENDIF}
              Exit;
             end;
            Sleep(2);
           end; // Pause
*)
         end;// send request
       end
      else
       dec(ServerSession.Status);
     end; // next packet
    if AllPacketsReceived then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
{
      EnterCriticalSection(FManager.FCSect);
      if ServerSession.MsgControlCode = ACRExecute then
        ServerSession.MsgControlCode := ACRSuspend;
      LeaveCriticalSection(FManager.FCSect);
}
      ServerSession.MsgReceiveStatus := ACRFull;
//      FManager.MessageReceived(ServerSession);
     end;
    inc(j);
   until False; // Get new Session
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - START');
{$ENDIF}
 except
  on E: Exception do
   begin
    Error:=
                  ErrorRServer+ErrorRMsgResendRequestThread+
                  IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+
                  ErrorRExecute+
                  E.Message;
    if (ServerSession = nil)
    or (ServerSession.Session = nil)
    then
      TACRServer(FManager.FServer).DoOnConnectionError(
                  AcrServerMsgResendRequestThread,-1,
                  Error)
    else
      TACRNetworkSession(ServerSession.Session).DoOnError(
                  AcrServerMsgResendRequestThread,-1,
                  Error);
   end;
 end;
end;// Execute

{$ENDIF} // Server


{$IFNDEF MsgCommunicator}
{IFDEF SERVER_VERSION}
{$ENDIF MsgCommunicator}
//------------------------------------------------------------------------------
// if authorization buffer is valid (encrypted by the same crypto settings as in CryptoInfo)
// then return true else return false
//------------------------------------------------------------------------------
function fnIsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
var ms:             TACRMemoryStream;
    crc32, crc32_1: Cardinal;
    size:           Integer;
begin
  Result := True;
  if BufferSize = 0 then
   begin
    if CryptoInfo.CryptoAlgorithm <> 0 then
      Result:= False;
    Exit;
   end;
  ms := TACRMemoryStream.Create(Buffer, BufferSize);
  try
   try
    LoadDataFromStream(size,SizeOf(crc32),ms,11322);
    LoadDataFromStream(crc32,SizeOf(crc32),ms,11323);
    if (ms.Size - ms.Position <> size) then
     raise EACRException.Create(11325,ErrorLInvalidStreamSize,[ms.Size,(ms.Position+size)]);
   except
     Result := False;
     Exit;
   end;
    ACRDecryptBuffer(CryptoInfo,PChar(Buffer+ms.Position),size);
    crc32_1 := ACRCountCRC(0,PChar(Buffer+ms.Position),size);
    Result := (crc32 = crc32_1);
  finally
    ms.Buffer := nil;
    ms.Free;
  end;
end; // fnIsAuthorizationBufferValid
{$IFNDEF MsgCommunicator}
{ENDIF SERVER_VERSION}
{$ENDIF MsgCommunicator}

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRConnection> initialized');
{$ENDIF}

end.

{ TODO -oLeo : Move these default values to published properties}
(*
  FParams.RemoteHost := ACRDefaultHost;
  FParams.LocalHost := ACRDefaultHost;
{$IFDEF SERVER_VERSION}
  FParams.RemotePort := ACRDefaultClientPort;
  FParams.LocalPort := ACRDefaultServerPort;
{$ENDIF}
{$IFDEF CLIENT_VERSION}
  FParams.RemotePort := ACRDefaultServerPort;
  FParams.LocalPort := ACRDefaultClientPort;
{$ENDIF}
  FParams.PacketSize := ACRDefaultPacketSize;
*)
{ default values has been set - End of Leo's TODO}

{ TODO -oLeo : You must disable possibilities to change ConnectionParams while session is connected }

{ TODO -oLeo : You must disable possibilities to change LocalPort and ServerID while there is active session on the server}



