unit ACRConnection;

interface

{$I ACRVer.inc}


{$IFDEF DEBUG_LOG_NETWORK_THREADS}
{$DEFINE LOG_CLIENT_THREADS}
{$DEFINE LOG_SERVER_THREADS}
{$ENDIF}


{DEFINE LOGIC_TEST}
{DEFINE CONNECTION_TEST}
{$IFNDEF CONNECTION_TEST}
 {DEFINE NETWORK_TEST}
{$ENDIF CONNECTION_TEST}

{$IFNDEF API_NETWORK}
 {$IFDEF D6H}
  {$DEFINE ProcessMessages} // Needs for Indy only
 {$ENDIF D6H}
{$ENDIF API_NETWORK}

uses
  Classes, Math, SysUtils,
{$IFDEF MSWINDOWS}
  Windows,
  Forms,
{$ENDIF}
{$IFDEF LINUX}
//  Libc,
  QForms,
{$ENDIF}

{$IFNDEF API_NETWORK}
 {$IFDEF D6H}
  IdBaseComponent, IdComponent, IdUDPBase, IdUDPClient, IdUDPServer, IdSocketHandle,
 {$ELSE}
  NMUDP,
 {$ENDIF}
{$ELSE}
// Accuracer units
  ACRNetwork,
  ACRCrypto,
{$ENDIF}

{$IFDEF DEBUG_LOG}
  ACRDebug,
{$ENDIF}
  ACRCompression,
  ACRBase,
  ACRTypes,
  ACRConst,
  ACRExcept,
  ACRLinux,
  ACRCriticalSection,
  ACRBaseEngine,
  ACRMemory;

const

 ACRNoResending = 999;

{******************************************************************************}
// Connection parameters
{******************************************************************************}

// Maximum threads number
 ACRMaxThreadCount = 100;

// To connect - Client only
 ACRConnectRetryCount = 20;       // - Client only
 ACRConnectDelay = 500;            // - Client only

// To receive full session buffer
 ACRStartReceiveTimeOut = 20000;    // - Client only, time to wait for the first packet in answer
 ACRReceiveTimeOut = 40000;         // - Client only, time since the first packet to the latest
 ACRReceiveSleep = 1;               // - Client only, time to sleep while receive buffer to allow incoming packets processing, >=0, <= 100
 ACRServerReceiveTimeOut = 60000;   // - Server only
 ACRServerReceiveSleep = 1;         // - Server only   >=0, <= 100

// To send buffer
 ACRSendTimeOut =        30000;        // - Client only
 ACRServerSendTimeOut = 180000;        // - Server only

// To send new request to ask other side to resend broken packet
 ACRResendDelay = 150;//40;       // - Client only, delay (msec) before resend requested packet, >=1, <= 100
 ACRRequestDelay = 0;//20;       // - Client only, delay (msec) before request lost packet, >=1, <= 100
 ACRServerResendDelay = 150;//20;         // - Server only, >=0, <= 100
 ACRServerRequestDelay = 0;//20;       // - Server only, >=0, <= 100

// To Diconnect
 ACRDisconnectRetryCount = 30;     // Both Client and Server
 ACRDisconnectDelay = 100;         // Both Client and Server

{******************************************************************************}

// Wait For Session end
 ACRWaitForTimeOut = 120000;        // Client only

// Wait for end of all other sessions messages sending
 ACRWaitForMessagesSend = 500000;   // Server only

// Wait for other threads finish to free shared object
 ACRThreadsTerminateDelay = 60000;  // Both Client and Server

{******************************************************************************}



{******************************************************************************}
// ControlCode (-128..127):
{******************************************************************************}
// From Client
                     //  5...9  -- reserwed for Accuracer commands
 ACRNewRequest = 10; // 10...32 -- reserwed for MsgCommunicator commands
                     // 33...63 -- reserwed for custom commands
 ACRConnect = 1;
// From Server
 ACRConnected = 2;
// Both
 ACRNoAction = 0;
 ACRMessage = 3;
// ACRLastPacket: 7th bit = 1, i.e. byte = x1xxxxxx, i.e. code >= 64
 ACRLastPacket = 64;
 ACREcho = 4;
// single packet:
 ACRAllPacketsReceived = -1;
 ACRPacketResendRequest = -2;
 ACRMessageReceived = -3;
 ACRMessagePacketResendRequest = -4;
 ACRDisconnect = -55;

{$IFDEF MsgCommunicator}
// Both
 ACRGetUserInfo = 10;
 ACRRegisterNewUser = 11;
 ACRAddUserToContacts = 12;
 ACRRemoveUserFromContacts = 13;
 ACRGetContacts = 14;
 ACRUpdateUserInfo = 15;
 ACRGetUserID = 16;
 ACRFindUsers = 17;
 ACRFindMessages = 18;
 ACRIsUserExisting = 19;
 ACRIsUserOnline = 20;
 ACRUpdateUserInContacts = 21;
{$ENDIF}

{******************************************************************************}


{******************************************************************************}
// Sending Status - Fof Commands and Messages
{******************************************************************************}
  ACRNotSent = 0;
  ACRSent = -1;
{******************************************************************************}

{******************************************************************************}
// Receiving Status
{******************************************************************************}
 ACRNoAnswer = 0;
 ACRStartAnswer = 1;
 ACRAnswerNotFull = 2;
 ACRFullAnswer = -1;
{******************************************************************************}

{******************************************************************************}
// MsgReceiveStatus
{******************************************************************************}
 ACRNotFull = 0;
 ACRFull = -1;
{******************************************************************************}

{******************************************************************************}
// Session Control Codes
{******************************************************************************}
 ACRExecute = 0;
 ACRSuspend = 1;
 ACRTerminate = -1;
{******************************************************************************}


{******************************************************************************}
// Session Status
{******************************************************************************}
 ACRVacant = 0;
 ACRInUse = -1;
{******************************************************************************}


// To avoid multiple realloc
 ACRDefaultPacketsInAnswer = 8;
 ACRDefaultPacketsInRequest = 2;
 ACRDefaultMsgPacketsInAnswer = 1;
 ACRDefaultMsgPacketsInRequest = 1;

// Signatures
{$IFDEF MsgCommunicator}
 ACRClientPacketSign = 'MSG1';
 ACRServerPacketSign = 'MSG2';
{$ELSE}
 ACRClientPacketSign = 'ACR1';
 ACRServerPacketSign = 'ACR2';
{$ENDIF}

type

  TACRSessionsArray = array of TACRBaseSession;

  TACRNetwork = class;
  TACRBaseConnectionManager = class;
{$IFDEF CLIENT_VERSION}
  TACRClientConnectionManager = class;
  TACRClientResendRequestThread = class;
  TACRClientMsgResendRequestThread = class;
{$ENDIF}
{$IFDEF SERVER_VERSION}
  TACRServerConnectionManager = class;
  TACRServerSessionThread = class;
  TACRServerResendRequestThread = class;
  TACRServerMsgResendRequestThread = class;
{$ENDIF}

  TACRControlCode = ShortInt;
  TACRNetworkClientID = Integer;
  TACRConnectionID = Integer;

  TACRPacketHeader = packed record
    CheckSum:           Integer;
    Signature:          array [0..3] of Char;
    Recepient:          TACRNetworkClientID;
    Sender:             TACRNetworkClientID;
    ConnectionID:       TACRConnectionID;
    SessionID:          TACRSessionID;
    CurrentRequestID:   Integer;
    PacketID:           Integer;
    ControlCode:        TACRControlCode;
  end;
  PACRPacketHeader = ^TACRPacketHeader;

  TACRPacket = packed record
    Buffer:             PChar;
    BufferSize:         Integer;
  end;
  PACRPacket = ^TACRPacket;

  TACRConnectionParams = packed record
    PacketSize:                   Integer;
    CompressionAlgorithm:         Byte;
    CompressionMode:              Byte;
  end;
  PACRConnectionParams = ^TACRConnectionParams;

{$IFDEF CLIENT_VERSION}
  TACRClntSession = packed record
    Session:                TACRBaseSession;
    ServerSessionID:        TACRSessionID;
    CurrentRequestID:       Integer;
    Packets:                TList;
    ListeningThreads:       TThreadList;
    MsgListeningThreads:    TThreadList;
    ResendRequestThread:    TACRClientResendRequestThread;
    MsgResendRequestThread: TACRClientMsgResendRequestThread;
    AnswerStatus:           Integer;
    SendStatus:             Integer;
    PacketIDsToResend:      TACRIntegerArray;
    ClientMessageID:        Integer;
    ServerMessageID:        Integer;
    MsgPackets:             TList;
    MsgPacketIDsToResend:   TACRIntegerArray;
    MsgSendStatus:          Integer;
    MsgReceiveStatus:       Integer;
    ConnectionID:           TACRConnectionID;
    RemoteConnectionID:     TACRConnectionID;
    ControlCode:            Shortint;
    MsgControlCode:         Shortint;
    Status:                 Shortint;
  end;
  PACRClntSession = ^TACRClntSession;

  TACRClntConnection = packed record
    ConnectionID:       TACRConnectionID;
    Network:            TACRNetwork;
  end;
  PACRClntConnection = ^TACRClntConnection;
{$ENDIF}

{$IFDEF SERVER_VERSION}
  TACRSrvrSession = packed record
    Session:                      TACRBaseSession;
    CurrentRequestID:             Integer;
    Packets:                      TThreadList;
    AnswerStatus:                 Shortint;               // for ReceiveBuffer
    SendStatus:                   Shortint;               // for SendBuffer
    MsgSendStatus:                Shortint;
    MsgReceiveStatus:             Shortint;
    PacketIDsToResend:            TACRThreadIntArray;
    ClientMessageID:              Integer;
    ServerMessageID:              Integer;
    MsgPackets:                   TThreadList;
    MsgPacketIDsToResend:         TACRThreadIntArray;
    ClientSessionID:              TACRSessionID;
    ConnectionID:                 TACRConnectionID;
    RemoteClientID:               TACRNetworkClientID;
    RemoteHost:                   String;
    RemotePort:                   Integer;
    Application:                  String;
    ControlCode:                  Shortint;
    MsgControlCode:               Shortint;
    ListeningThreads:             TThreadList;
    MsgListeningThreads:          TThreadList;
    Thread:                       TACRServerSessionThread;
    MsgThread:                    TACRServerSessionThread;
    Status:                       Shortint;
  end;

  PACRSrvrSession = ^TACRSrvrSession;
{$ENDIF}



////////////////////////////////////////////////////////////////////////////////
//
// TACRNetwork
//
////////////////////////////////////////////////////////////////////////////////

  TACRNetwork = class (TObject)
   private
    FManager:             TACRBaseConnectionManager;
    FLocalClient:         TACRNetworkClientID;
{$IFNDEF API_NETWORK}
    FLocalHost:           String;
    FRemoteHost:          String;
    FRemotePort:          Integer;
{$ENDIF API_NETWORK}
    function RndClientID: TACRNetworkClientID;
    procedure SetRemoteHost(Host: String);
    function GetRemoteHost: String;
    procedure SetRemotePort(Port: Integer);
    function GetRemotePort: Integer;
    procedure SetLocalHost(Host: String);
    function GetLocalHost: String;
    procedure SetLocalPort(Port: Integer);
    function GetLocalPort: Integer;
   public
{$IFDEF API_NETWORK}
    FACRNetwork:          TACRapiNetwork;
{$ELSE}
 {$IFDEF D6H}
//    FNetworkClient:       TIdUDPClient;
    FNetworkServer:       TIdUDPServer;
 {$ELSE}
    FTransport:           TNMUDP;
 {$ENDIF}
{$ENDIF}
   protected
    procedure CreateNetwork;
    procedure FreeNetwork;
    procedure ReCreateNetwork;
   public
    constructor Create(ConnectionManager: TACRBaseConnectionManager);
    destructor Destroy; override;
    procedure SendBuffer(
                          Buffer: PChar;
                          Count:  Integer
                         );
    procedure OnDisconnect(
                             FromHost:  String;
                             FromPort:  Integer
                             );
{$IFDEF API_NETWORK}
    procedure OnDataReceived(
                             Buffer:    PChar;
                             Count:     Integer;
                             FromHost:  String;
                             FromPort:  Integer
                             );
{$ELSE}
 {$IFDEF D6H}
    procedure OnDataReceived(
                             Sender:   TObject;
                             AData:    TStream;
                             ABinding: TIdSocketHandle
                             );
 {$ELSE}
    procedure OnDataReceived(
                             Sender: TComponent;
                             BufferSize: Integer;
                             FromIP: string;
                             Port: integer
                             );
 {$ENDIF}
{$ENDIF}
   public
    property RemoteHost: String read GetRemoteHost write SetRemoteHost;
    property RemotePort: Integer read GetRemotePort write SetRemotePort;
    property LocalHost: String read GetLocalHost write SetLocalHost;
    property LocalPort: Integer read GetLocalPort write SetLocalPort;
    property LocalClientID: Integer read FLocalClient write FLocalClient;
  end; // TACRNetwork



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

  TACRBaseConnectionManager = class (TObject)
   private
{$IFDEF CONNECTION_TEST}
    FOtherManager:  TObject;
{$ELSE}
 {$IFDEF NETWORK_TEST}
    FOtherManager:  TObject;
    FClient:        Boolean;
 {$ENDIF}
{$ENDIF}
    FCSect:         TRTLCriticalSection;
    FSessionID:     TACRSessionID;
   public
    FSessions:      TThreadList;
    FThreadCount:   Integer;
   protected
    procedure IncThreadCount;
    procedure DecThreadCount;
    procedure SetThreadCount(Value: Integer);
   private
    procedure NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); virtual; abstract;
    procedure OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); virtual; abstract;
    procedure CompressAndEncryptBuffer(
                        Session:              TACRBaseSession;
                        InBuffer:             PChar;
                        InBufferSize:         Integer;
                        var OutBuffer:        PChar;
                        var OutBufferSize:    Integer
                                        );
    function DecompressAndDecryptBuffer(
                        Session:              TACRBaseSession;
                        var Buffer:           PChar;
                        var BufferSize:       Integer
                                        ): Boolean;
   public
    property ThreadCount: Integer read FThreadCount write SetThreadCount;
  end; // TACRBaseConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

{$IFDEF CLIENT_VERSION}
  TACRClientConnectionManager = class (TACRBaseConnectionManager)
   private
    FConnections:     TThreadList;
    FConnectionID:    TACRConnectionID;
    FApplication:     String;
    FLiveSessions:    Integer;
    procedure PacketResendRequest(
                               Buffer:        PChar;
                               Network:       TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
    procedure SendConnectRequest(ClientSession: PACRClntSession);
    procedure SendDisconnectRequest(ClientSession: PACRClntSession;
                                    WaitForAnswer: Boolean = True);
    procedure SendAcknowledgement(ClientSession: PACRClntSession; Msg: Boolean = False);
{$IFDEF MsgCommunicator}
    function IsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
    procedure SendConnectAckn(
                          ClientSession:        PACRClntSession;
                          ClientConnection:     PACRClntConnection;
                          CurrentRequestID:     Integer = -1
                              );
{$ENDIF MsgCommunicator}
    procedure DoDisconnect(Session: TACRBaseSession);
    procedure TerminateAllSessionThreads(ClientSession: PACRClntSession);
    procedure TerminateMessageThreads(ClientSession: PACRClntSession);
    procedure TerminateCommandThreads(ClientSession: PACRClntSession);
    procedure DeleteSession(ClientSession: PACRClntSession);
    procedure OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure DoSendBuffer(
                          ClientSession:    PACRClntSession;
                          ClientConnection: PACRClntConnection;
                          Buffer:           PChar;
                          BufferSize:       Integer;
                          Code:             Integer = ACRNewRequest
                         );
    // if encryption algorithm <> acr_Cipher_None then allocate buffer, fill it and return size
    // otherwise return BufferSize = 0
    procedure CreateAuthorizationBuffer(
                          CryptoInfo:     TACRCryptoInfo;
                          out Buffer:     PChar;
                          out BufferSize: Integer
                                        );
   public
    constructor Create;
    destructor Destroy; override;
    procedure Connect(Session: TACRBaseSession; ListenOnly: Boolean = False);
    procedure Disconnect(Session: TACRBaseSession; ListenOnly: Boolean = False);
    procedure DisconnectAll;
    procedure SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNewRequest
                          );
    procedure ReceiveBuffer(
                          Session:        TACRBaseSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
    procedure SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                          );
    procedure ReceiveMessage(
                          ClientSession:  PACRClntSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
  end; // TACRClientConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientListenerThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientListenerThread = class(TThread)
  private
    FManager:       TACRClientConnectionManager;
    FNetwork:       TACRNetwork;
    Buffer:         PChar;
    BufferSize:     Integer;
    FromHost:       String;
    FromPort:       Integer;
    ClientSession:  PACRClntSession;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:       TACRClientConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       Network:       TACRNetwork;
                       FromHost:      String;
                       FromPort:      Integer
                       );
    destructor Destroy; override;
  end;// TACRClientListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientResendRequestThread = class(TThread)
  private
    FManager:           TACRClientConnectionManager;
    FClientSession:     PACRClntSession;
    FBuffer:            PChar;
    FNetwork:           TACRNetwork;
    FHost:              String;
    FPort:              Integer;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                       );
    destructor Destroy; override;
  end;// TACRClientResendRequestThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRClientMsgResendRequestThread = class(TThread)
  private
    FManager:           TACRClientConnectionManager;
    FClientSession:     PACRClntSession;
    FBuffer:            PChar;
    FNetwork:           TACRNetwork;
    FHost:              String;
    FPort:              Integer;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                       );
    destructor Destroy; override;
  end;// TACRClientMsgResendRequestThread

{$ENDIF} // Client



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerConnectionManager
//
////////////////////////////////////////////////////////////////////////////////


{$IFDEF SERVER_VERSION}
  TACRServerConnectionManager = class (TACRBaseConnectionManager)
   private
    FServer:                TComponent;
    FListenerStoped:        Boolean;
    ResendRequestThread:    TACRServerResendRequestThread;
    MsgResendRequestThread: TACRServerMsgResendRequestThread;
   public
    FNetwork:       TACRNetwork;
//    property Network: TACRNetwork read FNetwork write FNetwork;
   private
    function SessionsCount: Integer;
    procedure TerminateAllSessionThreads(ServerSession: PACRSrvrSession);
    procedure TerminateMessageThreads(ServerSession: PACRSrvrSession);
    procedure TerminateCommandThreads(ServerSession: PACRSrvrSession);
    procedure PacketResendRequest(
                               Buffer:      PChar;
                               Network:     TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
    procedure DeleteSession(ServerSession: PACRSrvrSession);
    procedure DoDisconnect(Session: TACRBaseSession);
    procedure DoSendBuffer(
                          ServerSession:    PACRSrvrSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNoAction
                                                  );
    procedure SendDisconnectRequest(ServerSession: PACRSrvrSession;
                                    WaitForAnswer: Boolean = True);
    procedure SendConnectAckn(
                          ServerSession:        PACRSrvrSession;
                          CurrentRequestID:     Integer = -1
                              );
    procedure SendAcknowledgement(
                          ServerSession:        PACRSrvrSession;
                          Msg:                  Boolean = False;
                          CurrentRequestID:     Integer = -1
                              );
    procedure OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              ); override;
    procedure CommandReceived(
                               ServerSession:  PACRSrvrSession;
                               ControlCode:    TACRControlCode
                              );
    procedure MessageReceived(
                               ServerSession:  PACRSrvrSession
                              );
    // if authorization buffer is valid (encrypted by the same crypto settings as in CryptoInfo)
    // then return true else return false
    function IsAuthorizationBufferValid(
                          CryptoInfo: TACRCryptoInfo;
                          Buffer:     PChar;
                          BufferSize: Integer
                                        ): Boolean;
   public
    constructor Create(Server: TComponent);
    destructor Destroy; override;
    procedure Disconnect(Session: TACRBaseSession);
    procedure DisconnectAll(WaitForAllDisconected: Boolean = True);
    procedure SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNoAction
                          );
    procedure ReceiveBuffer(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
    procedure SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                          );
    procedure ReceiveMessage(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
    function GetClientInfo(
                          Session:          TACRBaseSession;
                          var Host:         String;
                          var Port:         Integer;
                          var Application:  String
                            ): Boolean;
    // fill array with server session object connected to this server
    procedure GetClientsList(var Clients: TACRSessionsArray);
  end; // TACRClientConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerSessionThread = class(TThread)
  private
    FServerSession:           PACRSrvrSession;
    FManager:                 TACRServerConnectionManager;
    FCode:                    TACRControlCode;
  protected
    procedure Execute; override;
  public
    constructor Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          Code:             Integer = ACRNoAction
                       );
    destructor Destroy; override;
  public
  end;// TACRServerSessionThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerListenerThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerListenerThread = class(TThread)
  private
    FManager:       TACRServerConnectionManager;
    Buffer:         PChar;
    BufferSize:     Integer;
    FromHost:       String;
    FromPort:       Integer;
    ServerSession:  PACRSrvrSession;
    SessionFound:   Boolean;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:       TACRServerConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       FromHost:      String;
                       FromPort:      Integer
                       );
    destructor Destroy; override;
  end;// TACRServerListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerResendRequestThread = class(TThread)
  private
    FManager:           TACRServerConnectionManager;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRServerConnectionManager
                       );
    destructor Destroy; override;
  end;// TACRServerResendRequestThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////

  TACRServerMsgResendRequestThread = class(TThread)
  private
    FManager:           TACRServerConnectionManager;
  protected
    procedure Execute; override;
  public
    constructor Create(
                       Manager:           TACRServerConnectionManager
                       );
    destructor Destroy; override;
  end;// TACRServerMsgResendRequestThread

{$ENDIF} // Server


// Procedures and Functions

  function CheckSum(Buffer: PChar; BufferSize: Integer): Integer;

{$IFNDEF MsgCommunicator}
{IFDEF SERVER_VERSION}
{$ENDIF MsgCommunicator}
  function fnIsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
{$IFNDEF MsgCommunicator}
{ENDIF SERVER_VERSION}
{$ENDIF MsgCommunicator}


implementation

uses
{$IFDEF SERVER_VERSION}
  ACRServer,
{$ENDIF}
{$IFDEF MsgCommunicator}
 {$IFDEF CLIENT_VERSION}
  ACRClient,
 {$ENDIF}
{$ENDIF}
{$IFNDEF MsgCommunicator}
  ACRCommunication,
{$ENDIF}
  ACRMain;

////////////////////////////////////////////////////////////////////////////////
// CheckSum
////////////////////////////////////////////////////////////////////////////////
function CheckSum(Buffer: PChar; BufferSize: Integer): Integer;
begin
  // Start Point = Pointer(Buffer)+4
  // Count (in Bytes) = BufferSize-4
//  Result := ACRCountCRC(0, PChar(Buffer+4), BufferSize-4)
  Result := 0;
end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRNetwork
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// CreateNetwork
//------------------------------------------------------------------------------
procedure TACRNetwork.CreateNetwork;
begin
{$IFDEF API_NETWORK}
  FACRNetwork := TACRapiNetwork.Create;
  FACRNetwork.OnDataReceived := OnDataReceived;
  FACRNetwork.OnDisconnect := OnDisconnect;
{$ELSE}
 {$IFDEF D6H}
//  FNetworkClient := TIdUDPClient.Create(nil);
  FNetworkServer := TIdUDPServer.Create(nil);
  FNetworkServer.OnUDPRead := OnDataReceived;
 {$ELSE}
  FTransport := TNMUDP.Create(nil);
  FTransport.OnDataReceived := OnDataReceived;
 {$ENDIF D6H}
{$ENDIF API_NETWORK}
end; // CreateNetwork


//------------------------------------------------------------------------------
// FreeNetwork
//------------------------------------------------------------------------------
procedure TACRNetwork.FreeNetwork;
begin
{$IFDEF API_NETWORK}
  if FACRNetwork <> nil then
    FACRNetwork.Free;
  FACRNetwork := nil;
{$ELSE}
 {$IFDEF D6H}
//  FNetworkClient.Free;
  FNetworkServer.Free;
 {$ELSE}
  FTransport.Free;
 {$ENDIF D6H}
{$ENDIF}
end; // FreeNetwork


//------------------------------------------------------------------------------
// ReCreateNetwork
//------------------------------------------------------------------------------
procedure TACRNetwork.ReCreateNetwork;
var
  lRemoteHost:       String;
  lRemotePort:       Integer;
  lLocalPort:        Integer;
begin
  EnterCriticalSection(FManager.FCSect);
  lRemoteHost := RemoteHost;
  lRemotePort := RemotePort;
  lLocalPort := LocalPort;
  FreeNetwork;
  CreateNetwork;
  LocalPort := lLocalPort;
  RemotePort := lRemotePort;
  RemoteHost := lRemoteHost;
  LeaveCriticalSection(FManager.FCSect);
end; // ReCreateNetwork


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRNetwork.Create(ConnectionManager: TACRBaseConnectionManager);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.Create');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

  inherited Create;
  FManager := ConnectionManager;
  FLocalClient := RndClientID;
  CreateNetwork;

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRNetwork.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.Destroy');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

  FreeNetwork;
  inherited Destroy;

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;// Destoy


//------------------------------------------------------------------------------
// RndClientID
//------------------------------------------------------------------------------
function TACRNetwork.RndClientID: TACRNetworkClientID;
label generate;
begin
  Randomize;
generate:
  Result := Random(MAXINT); // - (MAXINT div 2);
  if (Result >= 0) and (Result < 1024) then // Reserved for ServerID
{ TODO -oAll: Add to DevGuide:
ServerID must be unique for each server working at the one machine.
We recomend to set ServerID from the region 0...1023.
}
   goto generate;
end;// RndClientID


//------------------------------------------------------------------------------
// SetRemoteHost
//------------------------------------------------------------------------------
procedure TACRNetwork.SetRemoteHost(Host: String);
begin
{$IFDEF API_NETWORK}
  FACRNetwork.RemoteHost := Host;
{$ELSE}
 {$IFDEF D6H}
  FRemoteHost := Host;
//  FNetworkClient.Host := Host;
 {$ELSE}
  FTransport.RemoteHost := Host;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetRemoteHost
//------------------------------------------------------------------------------
function TACRNetwork.GetRemoteHost: String;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.RemoteHost;
{$ELSE}
 {$IFDEF D6H}
  Result := FRemoteHost;
//  Result := FNetworkClient.Host;
 {$ELSE}
  Result := FTransport.RemoteHost;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SetRemotePort
//------------------------------------------------------------------------------
procedure TACRNetwork.SetRemotePort(Port: Integer);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SetRemotePort');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

{$IFDEF API_NETWORK}
  FACRNetwork.RemotePort := Port;
{$ELSE}
 {$IFDEF D6H}
  FRemotePort := Port;
//  FNetworkClient.Port := Port;
 {$ELSE}
  FTransport.RemotePort := Port;
 {$ENDIF}
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetRemotePort
//------------------------------------------------------------------------------
function TACRNetwork.GetRemotePort: Integer;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.RemotePort;
{$ELSE}
 {$IFDEF D6H}
  Result := FRemotePort;
//  Result := FNetworkClient.Port;
 {$ELSE}
  Result := FTransport.RemotePort;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SetLocalPort
//------------------------------------------------------------------------------
procedure TACRNetwork.SetLocalPort(Port: Integer);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SetLocalPort');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

{$IFDEF API_NETWORK}
  FACRNetwork.LocalPort := Port;
{$ELSE}
 {$IFDEF D6H}
  FNetworkServer.Active := False;
  FNetworkServer.DefaultPort := Port;
  FNetworkServer.Active := True;
 {$ELSE}
  FTransport.LocalPort := Port;
 {$ENDIF}
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SetLocalHost
//------------------------------------------------------------------------------
procedure TACRNetwork.SetLocalHost(Host: String);
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SetLocalHost');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

{$IFDEF API_NETWORK}
  FACRNetwork.LocalHost := Host;
{$ELSE}
  FLocalHost := Host;
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetLocalHost
//------------------------------------------------------------------------------
function TACRNetwork.GetLocalHost: String;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.LocalHost;
{$ELSE}
  Result := FLocalHost;
{$ENDIF}
end;


//------------------------------------------------------------------------------
// GetLocalPort
//------------------------------------------------------------------------------
function TACRNetwork.GetLocalPort: Integer;
begin
{$IFDEF API_NETWORK}
  Result := FACRNetwork.LocalPort;
{$ELSE}
 {$IFDEF D6H}
  Result := FNetworkServer.DefaultPort;
 {$ELSE}
  Result := FTransport.LocalPort;
 {$ENDIF}
{$ENDIF}
end;


//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRNetwork.SendBuffer(
                          Buffer: PChar;
                          Count:  Integer);
var
  Header: PACRPacketHeader;
{$IFNDEF D6H}
  Buff: array [0..ACRMaxPacketSize] of Char;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
  buf: PChar;
  i: Integer;
{$ENDIF}
{$IFDEF NETWORK_TEST}
  Connections: TList;
  Connection:  PACRClntConnection;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
  str: String;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.SendBuffer');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

  Header := PACRPacketHeader(Buffer);
  Header.Sender := FLocalClient;
  Header.CheckSum := CheckSum(Buffer, Count);
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
if LocalPort = ACRDefaultClientPort then
str := 'Client'
else
if LocalPort = ACRDefaultServerPort then
str := 'Server'
else
str := IntToStr(LocalPort);
str := str + '>>> '
+IntToStr(Header.CurrentRequestID)+' : '
+IntToStr(Header.PacketID)+' / '
+IntToStr(Header.ControlCode)
+' >>> '+RemoteHost+':'+IntToStr(RemotePort)
+' - '+IntToStr(GetTickCount)+'msec'
//+' <'+IntToStr(Header.Recepient)+'>'
;
aaWriteToLog(str);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------');
aaWriteToLog('ACRConnection>TACRNetwork.SendBuffer('+IntToHex(Integer(@Buffer),6)+', '+IntToStr(Count)+')');
aaWriteToLog('Sender    : '+IntToStr(FLocalClient));
aaWriteToLog('-----------');
aaWriteToLog('RemoteHost: '+RemoteHost);
aaWriteToLog('RemotePort: '+IntToStr(RemotePort));
aaWriteToLog('LocalHost : '+LocalHost);
aaWriteToLog('LocalPort : '+IntToStr(LocalPort));
aaWriteToLog('>>> Head:');
aaWriteToLog('CheckSum         '+IntToStr(Header.CheckSum));
aaWriteToLog('Signature        '+Header.Signature);
aaWriteToLog('Recepient        '+IntToStr(Header.Recepient));
aaWriteToLog('Sender           '+IntToStr(Header.Sender));
aaWriteToLog('ConnectionID     '+IntToStr(Header.ConnectionID));
aaWriteToLog('SessionID        '+IntToStr(Header.SessionID));
aaWriteToLog('CurrentRequestID '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('PacketID         '+IntToStr(Header.PacketID));
aaWriteToLog('ControlCode      '+IntToStr(Header.ControlCode));
(*
aaWriteToLog('>>> Data:');
buf:=MemoryManager.GetMem(Count-SizeOf(TACRPacketHeader)+1);
Move(PChar(Integer(Buffer)+SizeOf(TACRPacketHeader))^, buf^, Count-SizeOf(TACRPacketHeader));
for i:= 0 to Count-SizeOf(TACRPacketHeader)-1 do
  if (buf+i)^=#0 then
    (buf+i)^:='0';
(buf+Count-SizeOf(TACRPacketHeader))^:=#0;
aaWriteToLog(buf);
aaWriteToLog('###');
MemoryManager.FreeAndNilMem(buf);
*)
{$ENDIF}

{******************************************************************************}
// Send Buffer
{******************************************************************************}
//  EnterCriticalSection(FCSect);
{$IFDEF NETWORK_TEST}
  if FManager.FClient then
   if Assigned(TACRServerConnectionManager(FManager.FOtherManager).FNetwork.OnDataReceived) then
    TACRServerConnectionManager(FManager.FOtherManager).FNetwork.OnDataReceived(Buffer, Count, LocalHost, LocalPort)
  else
   begin
    Connections:=TACRClientConnectionManager(FManager.FOtherManager).FConnections.LockList;
    try
     Connection := Connections.Items[0];
    finally
     TACRClientConnectionManager(FManager.FOtherManager).FConnections.UnlockList;
    end;
    Connection.Network.OnDataReceived(Buffer, Count, LocalHost, LocalPort);
   end;
{$ELSE}
 {$IFDEF API_NETWORK}
  FACRNetwork.SendBuffer(Buffer, Count);
//  sleep(0); // TRY TO REMOVE!!!
 {$ELSE}
  {$IFDEF D6H}
  FNetworkServer.SendBuffer(FRemoteHost, FRemotePort, Buffer^, Count);
//  FNetworkClient.SendBuffer(Buffer^, Count);
  {$ELSE}
  Move(Buffer^, Buff[0], Count);
  FTransport.SendBuffer(Buff, Count);
  {$ENDIF D6H}
 {$ENDIF API_NETWORK}
{$ENDIF NETWORK_TEST}
//  LeaveCriticalSection(FCSect);
{******************************************************************************}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount));
aaWriteToLog('');
{$ENDIF}
end; // SendBuffer


//------------------------------------------------------------------------------
// OnDisconnect
//------------------------------------------------------------------------------
procedure TACRNetwork.OnDisconnect(
                             FromHost:  String;
                             FromPort:  Integer
                             );
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect> START');
{$ENDIF}
 try
  ReCreateNetwork;
  FManager.OnDisconnect(self, FromHost, FromPort);
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('ACRConnection> TACRNetwork.OnDisconnect> FINISH');
{$ENDIF}
end;// OnDisconnect


//------------------------------------------------------------------------------
// OnDataReceived
//------------------------------------------------------------------------------
{$IFDEF API_NETWORK}
procedure TACRNetwork.OnDataReceived(
                             Buffer:    PChar;
                             Count:     Integer;
                             FromHost:  String;
                             FromPort:  Integer
                             );
{$ELSE}
 {$IFDEF D6H}
procedure TACRNetwork.OnDataReceived(
                             Sender:   TObject;
                             AData:    TStream;
                             ABinding: TIdSocketHandle
                             );
 {$ELSE}
procedure TACRNetwork.OnDataReceived(
                             Sender: TComponent;
                             BufferSize: Integer;
                             FromIP: string;
                             Port: integer
                             );
 {$ENDIF D6H}
{$ENDIF API_NETWORK}
var
  Buf: PChar;
  BufSize: Integer;
{$IFDEF API_NETWORK}
{$ELSE}
 {$IFDEF D6H}
  FromHost: String;
  FromPort: Integer;
 {$ELSE}
  Buffer: array [0..ACRMaxPacketSize] of Char;
 {$ENDIF}
{$ENDIF}
  Header: PACRPacketHeader;
{$IFDEF DEBUG_LOG_NETWORK}
  buff: pchar;
  i: integer;
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
  str: String;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog('TACRNetwork.OnDataReceived');
aaWriteToLog(IntToStr(gettickcount));
{$ENDIF}

//  EnterCriticalSection(FCSect);
{$IFDEF API_NETWORK}
  Buf := MemoryManager.GetMem(Count);
  Move(Buffer^, Buf^, Count);
  BufSize := Count;
{$ELSE}
 {$IFDEF D6H}
  Buf := MemoryManager.GetMem(ACRMaxPacketSize);
  BufSize := AData.Read(Buf^, ACRMaxPacketSize);
  FromHost := ABinding.PeerIP;
  FromPort := ABinding.PeerPort;
 {$ELSE}
  Buf := MemoryManager.GetMem(BufferSize);
  FTransport.ReadBuffer(Buffer, BufferSize);
  Move(Buffer[0], Buf^, BufferSize);
  BufSize := BufferSize;
 {$ENDIF D6H}
{$ENDIF API_NETWORK}
//  LeaveCriticalSection(FCSect);
  Header := PACRPacketHeader(Buf);
{$IFDEF DEBUG_LOG_NETWORK_SHORT}
if LocalPort = ACRDefaultClientPort then
str := 'Client'
else
if LocalPort = ACRDefaultServerPort then
str := 'Server'
else
str := IntToStr(LocalPort);
str := str + '<<< '
+IntToStr(Header.CurrentRequestID)+' : '
+IntToStr(Header.PacketID)+' / '
+IntToStr(Header.ControlCode)
+' <<< '+FromHost+':'+IntToStr(FromPort)
+' - '+IntToStr(GetTickCount)+'msec'
//+' <'+IntToStr(Header.Recepient)+'>'
;
aaWriteToLog(str);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------');
aaWriteToLog('ACRConnection>TACRNetwork.OnDataReceived('+IntToHex(Integer(@Buf),6)+', '+IntToStr(BufSize)+')');
aaWriteToLog('Recepient : '+IntToStr(FLocalClient));
aaWriteToLog('-----------');
 {$IFDEF API_NETWORK}
aaWriteToLog('RemoteHost: '+FromHost);
aaWriteToLog('RemotePort: '+IntToStr(FromPort));
 {$ELSE}
  {$IFDEF D6H}
aaWriteToLog('RemoteHost: '+FromHost);
aaWriteToLog('RemotePort: '+IntToStr(FromPort));
  {$ELSE}
aaWriteToLog('RemoteHost: '+FromIP);
aaWriteToLog('RemotePort: '+IntToStr(Port));
  {$ENDIF D6H}
 {$ENDIF API_NETWORK}
aaWriteToLog('LocalHost : '+LocalHost);
aaWriteToLog('LocalPort : '+IntToStr(LocalPort));
aaWriteToLog('>>> Head:');
aaWriteToLog('CheckSum         '+IntToStr(Header.CheckSum));
aaWriteToLog('Signature        '+Header.Signature);
aaWriteToLog('Recepient        '+IntToStr(Header.Recepient));
aaWriteToLog('Sender           '+IntToStr(Header.Sender));
aaWriteToLog('ConnectionID     '+IntToStr(Header.ConnectionID));
aaWriteToLog('SessionID        '+IntToStr(Header.SessionID));
aaWriteToLog('CurrentRequestID '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('PacketID         '+IntToStr(Header.PacketID));
aaWriteToLog('ControlCode      '+IntToStr(Header.ControlCode));
(*
aaWriteToLog('>>> Data:');
buff:=MemoryManager.GetMem(BufSize-SizeOf(TACRPacketHeader)+1);
Move(PChar(Integer(Buf)+SizeOf(TACRPacketHeader))^, buff^, BufSize-SizeOf(TACRPacketHeader));
for i:= 0 to BufSize-SizeOf(TACRPacketHeader)-1 do
  if (buff+i)^=#0 then
    (buff+i)^:='0';
(buff+BufSize-SizeOf(TACRPacketHeader))^:=#0;
aaWriteToLog(buff);
aaWriteToLog('###');
MemoryManager.FreeAndNilMem(buff);
*)
{$ENDIF DEBUG_LOG_NETWORK}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount)+ ' - buffer prepared');
{$ENDIF}
  if (Header.Recepient = FLocalClient)
{$IFDEF MsgCommunicator}
  or (
//      (Header.Recepient = Integer(MSG_INVALID_USER_ID)) and
      ((Header.ControlCode = ACRConnect) or (Header.ControlCode = ACRConnect+ACRLastPacket))
      and
      (Header.Signature = ACRClientPacketSign)
      )
{$ENDIF MsgCommunicator}
  then
{$IFDEF API_NETWORK}
    FManager.NetworkListener(Buf, BufSize, Self, FromHost, FromPort);
{$ELSE}
 {$IFDEF D6H}
    FManager.NetworkListener(Buf, BufSize, Self, FromHost, FromPort);
 {$ELSE}
    FManager.NetworkListener(Buf, BufSize, Self, FromIP, Port);
 {$ENDIF D6H}

{$ENDIF API_NETWORK}

{$IFDEF DEBUG_LOG_NETWORK_TIME}
aaWriteToLog(IntToStr(gettickcount)+' - end of parsing');
aaWriteToLog('');
{$ENDIF}
end;
// TACRNetwork



////////////////////////////////////////////////////////////////////////////////
//
// TACRBaseConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// TACRBaseConnectionManager.IncThreadCount
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.IncThreadCount;
begin
  EnterCriticalSection(FCSect);
  FThreadCount := FThreadCount+1;
  LeaveCriticalSection(FCSect);
end;// TACRBaseConnectionManager.IncThreadCount

//------------------------------------------------------------------------------
// TACRBaseConnectionManager.DecThreadCount
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.DecThreadCount;
begin
  EnterCriticalSection(FCSect);
  FThreadCount := FThreadCount-1;
  LeaveCriticalSection(FCSect);
end;// TACRBaseConnectionManager.DecThreadCount

//------------------------------------------------------------------------------
// TACRBaseConnectionManager.SetThreadCount
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.SetThreadCount(Value: Integer);
begin
  EnterCriticalSection(FCSect);
  FThreadCount := Value;
  LeaveCriticalSection(FCSect);
end;// TACRBaseConnectionManager.SetThreadCount

//------------------------------------------------------------------------------
// compresses and encrypts buffer
//------------------------------------------------------------------------------
procedure TACRBaseConnectionManager.CompressAndEncryptBuffer(
                        Session:              TACRBaseSession;
                        InBuffer:             PChar;
                        InBufferSize:         Integer;
                        var OutBuffer:        PChar;
                        var OutBufferSize:    Integer
                                                              );
begin
 try
  ACRBaseEngine.CompressAndEncryptBuffer(Session.ConnectParams.CryptoInfo,
                                         Session.ConnectParams.CompressionAlgorithm,
                                         Session.ConnectParams.CompressionMode,
                                         InBuffer,  InBufferSize,
                                         OutBuffer, OutBufferSize);
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> CompressAndEncryptBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
end; // CompressAndEncryptBuffer


//------------------------------------------------------------------------------
// decompresses and decrypts buffer
//------------------------------------------------------------------------------
function TACRBaseConnectionManager.DecompressAndDecryptBuffer(
                        Session:              TACRBaseSession;
                        var Buffer:           PChar;
                        var BufferSize:       Integer
                                                              ): Boolean;
begin
 try
  Result := ACRBaseEngine.DecompressAndDecryptBuffer(
                                Session.ConnectParams.CryptoInfo,
                                Session.ConnectParams.CompressionAlgorithm,
                                Session.ConnectParams.CompressionMode,
                                Buffer,
                                BufferSize);
 except
  on E: Exception do
   begin
    Result := False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> DecompressAndDecryptBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
   end;
 end;
end; // DecompressAndDecryptBuffer



{$IFDEF CLIENT_VERSION}

////////////////////////////////////////////////////////////////////////////////
//
// TACRClientConnectionManager
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientConnectionManager.Create;
begin
  InitializeCriticalSection(FCSect);
  inherited Create;
  FSessions := TThreadList.Create;
  FConnections := TThreadList.Create;
  FSessionID := MAXINT;
  FConnectionID := -MAXINT;
  FLiveSessions := 0;
  FApplication := '';
  FApplication := ParamStr(0);
  if FApplication = '' then
    FApplication := 'UNKNOWN';
  ThreadCount := 0;
{$IFDEF CONNECTION_TEST}
//  FOtherManager :=
{$ELSE}
 {$IFDEF NETWORK_TEST}
//  FOtherManager :=
  FClient := True;
 {$ENDIF}
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('TACRClientConnectionManager.Create> Application: '+FApplication);
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientConnectionManager.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.Destroy> Application: '+FApplication);
{$ENDIF}
  DisconnectAll;
  EnterCriticalSection(FCSect);
  FConnections.Free;
  FSessions.Free;
  FSessions := nil;
  LeaveCriticalSection(FCSect);
{ TODO -oAlex : Write ThreadCount to ErrorLog }
  DeleteCriticalSection(FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.Destroy - FINISHED');
{$ENDIF}
end;// Destroy


//------------------------------------------------------------------------------
// PacketResendRequest
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.PacketResendRequest(
                               Buffer:        PChar;
                               Network:       TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
var
  Header:          PACRPacketHeader;
begin
  Header := Pointer(Buffer);
  if Msg then
    Header.ControlCode := ACRMessagePacketResendRequest
  else
    Header.ControlCode := ACRPacketResendRequest;
  Header.Signature := ACRClientPacketSign;
  if (PacketID >= 0) then
    Header.PacketID := PacketID
  else
    Header.Recepient := Header.Sender;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('TACRClientConnectionManager.PacketResendRequest> asks to resend packet # '+IntToStr(Header.PacketID));
{$ENDIF}
  EnterCriticalSection(FCSect);
  try
   Network.RemoteHost := RemoteHost;
   Network.RemotePort := RemotePort;
   Network.SendBuffer(Buffer, SizeOf(TACRPacketHeader));
  finally
   LeaveCriticalSection(FCSect);
  end;
end;// PacketResendRequest


//------------------------------------------------------------------------------
// OnDisconnect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
var
  Sessions:             TList;
  Session:              TACRBaseSession;
  ClientSession:        PACRClntSession;
//  Connections:          TList;
//  ClientConnection:     PACRClntConnection;
  i, j:                 Integer;
//  Found:                Boolean;
begin
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - START');
{$ENDIF}
{
  Connections:=FConnections.LockList;
  try
   j := Connections.Count-1;
   for i:=0 to j do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.Network = FNetwork then
      begin
       Found := True;
       break;
      end;
    end;
  finally
   FConnections.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(FNetwork)]);
  Found := False;
}
  Sessions:=FSessions.LockList;
  try
   j := Sessions.Count-1;
   for i:=j downto 0 do
    begin
     ClientSession := Sessions.Items[i];
//     if ClientSession.ConnectionID = ClientConnection.ConnectionID then
     if  (FromHost = ClientSession.Session.ConnectParams.RemoteHost)
     and (FromPort = ClientSession.Session.ConnectParams.RemotePort)
     then
      begin
//       Found := True;
       Session := ClientSession.Session;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - DoDisconnect');
{$ENDIF}
       FSessions.UnlockList;
       DoDisconnect(Session);
       Sessions := FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - Session.OnDisconnect');
{$ENDIF}
       if Session<>nil then
         Session.OnDisconnect;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - Session.OnDisconnect FINISHED');
{$ENDIF}
{$IFDEF MsgCommunicator}
       if Session<>TACRClientSession(Session).FClient.FDefaultSession then
        if Session<>TACRClientSession(Session).FClient.FDefaultServerSession then
         Session.Free;
 {$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - Session Freed!');
 {$ENDIF}
{$ENDIF}
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
{
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(ClientConnection.ConnectionID)]);
}
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.OnDisconnect - FINISH');
{$ENDIF}
end; // OnDisconnect


//------------------------------------------------------------------------------
// NetworkListener
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
begin
  repeat
   if ThreadCount<ACRMaxThreadCount then
    break;
   sleep(0);
  until False;
  TACRClientListenerThread.Create(self, Buffer, BufferSize, Network, FromHost, FromPort);
end;// NetworkListener


//------------------------------------------------------------------------------
// ReceiveMessage
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.ReceiveMessage(
                          ClientSession:        PACRClntSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
var
  i:                    Integer;
  Packet:               PACRPacket;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT has started to receive message');
{$ENDIF}
  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);
  if ClientSession.MsgReceiveStatus<>ACRFullAnswer then
   begin
    while (ClientSession.MsgReceiveStatus<>ACRStartAnswer)
    and (ClientSession.MsgReceiveStatus<>ACRFullAnswer)
    do // Wait for starting answer receive
     begin
      if ClientSession.MsgControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRReceiveSleep);
     end;
    StartTime := GetTickCount;
    while ClientSession.MsgReceiveStatus<>ACRFullAnswer do // Wait for all packets to arrive
     begin
      if ClientSession.MsgControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
      if (GetTickCount - StartTime) > ACRReceiveTimeOut then
        raise EACRException.Create(40026, ErrorRTimeoutFullReceive,
                                [ClientSession.ServerSessionID, ACRReceiveTimeOut]);
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT are receiving message from SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
//  TerminateMessageThreads(ClientSession);
  EnterCriticalSection(FCSect);
  BufferSize := 0;
  for i := 0 to ClientSession.MsgPackets.Count - 1 do
   begin
    Packet := ClientSession.MsgPackets.Items[i];
    BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
   end;
  Buffer := MemoryManager.GetMem(BufferSize);
  pBuf := Buffer;
  for i := 0 to ClientSession.MsgPackets.Count - 1 do
   begin
    Packet := ClientSession.MsgPackets.Items[i];
    if Packet <> nil then
     begin
      if Packet.Buffer <> nil then
       begin
        Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
        inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
        MemoryManager.FreeAndNilMem(Packet.Buffer);
       end;
      Dispose(Packet);
      ClientSession.MsgPackets.Items[i] := nil;
     end;
   end;
  ClientSession.MsgPackets.Count := 0;
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('CLIENT HAS RECEIVED Message FROM SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
  ClientSession.MsgReceiveStatus := ACRNoAnswer;
  inc(ClientSession.ServerMessageID);
  EnterCriticalSection(FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FCSect);
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientConnectionManager.ReceiveMessage - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end; // except
end; // ReceiveMessage


//------------------------------------------------------------------------------
// ReceiveBuffer
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.ReceiveBuffer(
                          Session:        TACRBaseSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
label
  Loop;
var
  Sessions:             TList;
  ClientSession:        PACRClntSession;
  SessionFound:         Boolean;
  i:                    Integer;
  Packet:               PACRPacket;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT has started to receive answer');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 try

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT has started to receive answer');
{$ENDIF}
  SessionFound := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if (ClientSession.Session = Session)
     then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);

  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);

  StartTime := GetTickCount;
  if ClientSession.AnswerStatus<>ACRFullAnswer then
   begin
    while (ClientSession.AnswerStatus<>ACRStartAnswer)
    and (ClientSession.AnswerStatus<>ACRFullAnswer)
    do // Wait for starting answer receive
     begin
      if ClientSession.ControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
      if (GetTickCount - StartTime) > ACRStartReceiveTimeOut then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40025, ErrorRTimeoutFullReceive,
                                [ClientSession.ServerSessionID, ACRReceiveTimeOut]);
       end;
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRReceiveSleep);
     end;
    StartTime := GetTickCount;
    while ClientSession.AnswerStatus<>ACRFullAnswer do // Wait for all packets to arrive
     begin
      if ClientSession.ControlCode = ACRTerminate then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40041, ErrorRCannotReceiveFromServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);
       end;
      if (GetTickCount - StartTime) > ACRReceiveTimeOut then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40025, ErrorRTimeoutFullReceive,
                            [ClientSession.ServerSessionID, ACRReceiveTimeOut]);
       end;
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received all packets and starts to extract buffer from SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
  TerminateCommandThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CommandThreads Terminated');
{$ENDIF}
  EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CS Entered');
{$ENDIF}
  BufferSize := 0;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packets.Count = '+IntToStr(ClientSession.Packets.Count));
{$ENDIF}
  for i := 0 to ClientSession.Packets.Count - 1 do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packet # '+IntToStr(i));
{$ENDIF}
    Packet := ClientSession.Packets.Items[i];
    BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packet.BufferSize = '+IntToStr(Packet.BufferSize));
{$ENDIF}
   end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> BufferSize = '+IntToStr(BufferSize));
{$ENDIF}
   if BufferSize > 0 then
     Buffer := MemoryManager.GetMem(BufferSize)
   else
     Buffer := nil;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Buffer memory got');
{$ENDIF}
  pBuf := Buffer;
  for i := 0 to ClientSession.Packets.Count - 1 do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packets.Count = '+IntToStr(ClientSession.Packets.Count));
{$ENDIF}
    Packet := ClientSession.Packets.Items[i];
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Packet # '+IntToStr(i));
{$ENDIF}
    if Packet <> nil then
     begin
      if Packet.Buffer <> nil then
       begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Move...');
{$ENDIF}
        if (Packet.BufferSize > SizeOf(TACRPacketHeader)) then
         begin
          Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> inc...');
{$ENDIF}
          inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
         end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> inced'+IntToStr(Packet.BufferSize-SizeOf(TACRPacketHeader)));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> free packet buffer...');
{$ENDIF}
        MemoryManager.FreeAndNilMem(Packet.Buffer);
       end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> Dispose packet...');
{$ENDIF}
      Dispose(Packet);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> packet = nil...');
{$ENDIF}
      ClientSession.Packets.Items[i] := nil;
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> ClientSession.Packets.Count := 0 ...');
{$ENDIF}
  ClientSession.Packets.Count := 0;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CS leave...');
{$ENDIF}
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> CS left');
{$ENDIF}
  DecompressAndDecryptBuffer(ClientSession.Session, Buffer, BufferSize);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.ReceiveBuffer> DecompressAndDecryptBuffer - OK');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('CLIENT HAS RECEIVED BUFFER FROM SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
  ClientSession.AnswerStatus := ACRNoAnswer;

  EnterCriticalSection(FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FCSect);

 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientConnectionManager.ReceiveBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end; // except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT has finished answer receiving');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('CLIENT<<< '+IntToStr(ClientSession.CurrentRequestID)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}
end; // ReceiveBuffer


//------------------------------------------------------------------------------
// SendConnectRequest
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendConnectRequest(ClientSession: PACRClntSession);
var
  Buf, Buffer:                    PChar;
  BufSize, BufferSize,
  SizeApp, SizeParams:            Integer;
  StartTime:                      DWORD;
  Retry:                          Integer;
  ConnectionParams:               TACRConnectionParams;
  ServerID:                       Integer;
  Direct:                         Boolean;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT SendConnectRequest started');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);
  ClientSession.AnswerStatus := ACRNoAnswer;
  ClientSession.CurrentRequestID := 0;
  ConnectionParams.PacketSize := ClientSession.Session.ConnectParams.PacketSize;
  ConnectionParams.CompressionAlgorithm := ClientSession.Session.ConnectParams.CompressionAlgorithm;
  ConnectionParams.CompressionMode := ClientSession.Session.ConnectParams.CompressionMode;
  CreateAuthorizationBuffer(ClientSession.Session.ConnectParams.CryptoInfo, Buffer, BufferSize);
//  IsAuthorizationBufferValid(ClientSession.Session.ConnectParams.CryptoInfo, Buffer, BufferSize);
  SizeApp := Length(FApplication) + 1;
  SizeParams := SizeOf(ConnectionParams);
  BufSize := SizeParams + SizeOf(BufferSize) + BufferSize + SizeApp;
{$IFDEF MsgCommunicator}
  inc(BufSize, SizeOf(ClientSession.Session.FUserID));
{$ENDIF}
  Buf := MemoryManager.GetMem(BufSize);
{$IFDEF MsgCommunicator}
  Move(ClientSession.Session.FUserID, Buf^, SizeOf(ClientSession.Session.FUserID));
  inc(Buf, SizeOf(ClientSession.Session.FUserID));
{$ENDIF}
  try
   Move(ConnectionParams, Buf^, SizeParams);
   Move(BufferSize, PChar(Buf + SizeParams)^, SizeOf(BufferSize));
   if BufferSize > 0 then
    begin
     Move(Buffer^, PChar(Buf + SizeParams + SizeOf(BufferSize))^, BufferSize);
     MemoryManager.FreeAndNilMem(Buffer);
    end;
   StrPCopy(Buf + SizeParams + SizeOf(BufferSize) + BufferSize, FApplication);
   Retry := 0;
{$IFDEF MsgCommunicator}
   dec(Buf, SizeOf(ClientSession.Session.FUserID));
{$ENDIF}
   repeat
    if ClientSession.ControlCode = ACRTerminate then
     begin
      EnterCriticalSection(FCSect);
      dec(ClientSession.Status);
      LeaveCriticalSection(FCSect);
      Exit;
     end;
{$IFDEF MsgCommunicator}
    if TACRClientSession(ClientSession.Session).Direct then
       ClientSession.Session.FConnectParams.ServerID :=
          TACRClientSession(ClientSession.Session).RemoteUser.UserID; // for connect request only
{$ENDIF}
    try
     SendBuffer(ClientSession.Session, Buf, BufSize, ACRConnect);
    except
      EnterCriticalSection(FCSect);
      dec(ClientSession.Status);
      LeaveCriticalSection(FCSect);
//      raise;
      Exit;
    end;
{$IFDEF MsgCommunicator}
    EnterCriticalSection(FCSect);
    if TACRClientSession(ClientSession.Session).Direct then
      if ClientSession.Session.FConnectParams.ServerID = TACRClientSession(ClientSession.Session).RemoteUser.UserID then // Header.Sender from the answer is not set
        ClientSession.Session.FConnectParams.ServerID := MSG_INVALID_USER_ID;  // allow listen for ACRConnected answer
    LeaveCriticalSection(FCSect);
{$ENDIF}
    StartTime := GetTickCount;
    while ((GetTickCount - StartTime) < ACRConnectDelay) do // pause
     if ClientSession.SendStatus <> ACRSent then
      begin
{$IFDEF ProcessMessages}
       Application.ProcessMessages;
{$ENDIF ProcessMessages}
       Sleep(ACRReceiveSleep);
      end
     else
      Exit;
    inc(Retry);
   until  (Retry > ACRConnectRetryCount);
   if ClientSession.SendStatus <> ACRSent then
    begin
     ClientSession.Status := ACRVacant;
     ServerID := ClientSession.Session.ConnectParams.ServerID;
{$IFDEF MsgCommunicator}
     Direct := TACRClientSession(ClientSession.Session).Direct;
     if Direct then
       ServerID := TACRClientSession(ClientSession.Session).RemoteUser.UserID
     else
       ServerID := TACRClientSession(ClientSession.Session).FClient.ConnectionParams.ConnectParams.ServerID;
{$ENDIF}
     DoDisconnect(ClientSession.Session);
{$IFDEF MsgCommunicator}
     if Direct then
       raise EACRException.Create(40023, ErrorRCannotConnect,
                           ['user', ServerID,
                            ACRConnectRetryCount, ACRConnectDelay])
     else
{$ENDIF}
       raise EACRException.Create(40023, ErrorRCannotConnect,
                           ['server', ServerID,
                            ACRConnectRetryCount, ACRConnectDelay]);
    end;
  finally
   MemoryManager.FreeAndNilMem(Buf);
   EnterCriticalSection(FCSect);
   dec(ClientSession.Status);
   LeaveCriticalSection(FCSect);
  end;
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> SendConnectRequest - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end; // except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT SendConnectRequest finished');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
end;// SendConnectRequest


//------------------------------------------------------------------------------
// Connect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.Connect(Session: TACRBaseSession;
                                              ListenOnly: Boolean = False);
var
  Connections:          TList;
  Sessions:             TList;
  ClientConnection:     PACRClntConnection;
  ClientSession:        PACRClntSession;
//  MaxHeaderSize:        Integer;
  ConnectionID,
  SessionID:            TACRSessionID;
  Buffer:               PChar;
  BufferSize:           Integer;
  i:                    Integer;
  ConnectionFound:      Boolean;
  SessionFound:         Boolean;
  Network:              TACRNetwork;
  ConnectionParams:     PACRConnectionParams;
begin
  if ListenOnly then
    Session.SessionID := INVALID_SESSION_ID
  else
   begin
    EnterCriticalSection(FCSect);
    Session.SessionID := FSessionID;
    dec(FSessionID);
    if FSessionID = INVALID_SESSION_ID then
      dec(FSessionID);
    LeaveCriticalSection(FCSect);
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT #'+IntToStr(Session.SessionID)+' TRY TO CONNECT TO SERVER #'+IntToStr(Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
{
// To send connect request in one packet:
  MaxHeaderSize := Max (
   SizeOf(TACRPacketHeader) + SizeOf(TACRConnectionParams)
//    + SizeOf(TACRConnectionID)  // set by client
   + SizeOf(TACRSessionID)        // set by server
   ,
   SizeOf(TACRPacketHeader) + SizeOf(TACRConnectionParams)
    + SizeOf(FApplication)
                        );
  if (Session.ConnectParams.PacketSize < MaxHeaderSize) then
   raise EACRException.Create(40020, ErrorRPacketSizeTooSmall,
                                    [Session.ConnectParams.PacketSize,
                                     MaxHeaderSize]);
}
  SessionFound := False;
// Does Session exist?
  Sessions:=FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if ClientSession.Session = Session then
      begin
       SessionFound := True;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
// No. Does Connection exist?
   begin
    ConnectionFound := False;
    Connections:=FConnections.LockList;
    try
     for i:=0 to Connections.Count-1 do
      begin
       ClientConnection := Connections.Items[i];
       if  (ClientConnection.Network.LocalPort = Session.ConnectParams.LocalPort)
       then
        begin
         ConnectionFound := True;
         break;
        end;
      end;
    finally
     FConnections.UnlockList;
    end;
    if not ConnectionFound then
     begin // No Session, no Connection
// Create new connection
      Network := nil;
      Connections:=FConnections.LockList;
      try
       for i:=0 to Connections.Count-1 do
        begin
         ClientConnection := Connections.Items[i];
         if (ClientConnection.Network.LocalPort = Session.ConnectParams.LocalPort)
{
         and
            (ClientConnection.Network.LocalHost := Session.ConnectParams.LocalHost)
}
         then
          begin
           Network := ClientConnection.Network;
           break;
          end;
        end;
      finally
       FConnections.UnlockList;
      end;
      New(ClientConnection);
      EnterCriticalSection(FCSect);
      inc(FConnectionID);
      ClientConnection.ConnectionID := FConnectionID;
      LeaveCriticalSection(FCSect);
      if Network = nil then
       begin
        IncThreadCount;
        ClientConnection.Network := TACRNetwork.Create(self);
        ClientConnection.Network.LocalPort := Session.ConnectParams.LocalPort;
        Session.FConnectParams.LocalPort := ClientConnection.Network.LocalPort; // if port is already in use
       end
      else
        ClientConnection.Network := Network;
// Add new connection to Connections list
      Connections.Add(ClientConnection);
     end; // Set new connection
// No Session, Connection Exists
    New(ClientSession);
    ClientSession.Session := Session;
    ClientSession.ServerSessionID := ClientSession.Session.SessionID; // Header.SessionID
    ClientSession.ConnectionID := ClientConnection.ConnectionID;
    ClientSession.RemoteConnectionID := ClientConnection.ConnectionID;
    ClientSession.CurrentRequestID := 0;
    ClientSession.ClientMessageID := 0;
    ClientSession.ServerMessageID := 0;
    ClientSession.Packets := TList.Create;
    ClientSession.Packets.Capacity := ACRDefaultPacketsInAnswer; // Allocate some place in list
    ClientSession.Packets.Clear; // Set Items to nil
    ClientSession.MsgPackets := TList.Create;
    ClientSession.MsgPackets.Capacity := ACRDefaultMsgPacketsInAnswer; // Allocate some place in list
    ClientSession.MsgPackets.Clear; // Set Items to nil
    ClientSession.PacketIDsToResend := TACRIntegerArray.Create;
    ClientSession.MsgPacketIDsToResend := TACRIntegerArray.Create;
    ClientSession.ListeningThreads := TThreadList.Create;
    ClientSession.MsgListeningThreads := TThreadList.Create;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           ListeningThreads created in @ClientSession='+IntToStr(Integer(ClientSession)));
{$ENDIF}
    ClientSession.ResendRequestThread := nil;
    ClientSession.MsgResendRequestThread := nil;
    ClientSession.AnswerStatus := ACRNoAnswer;
    ClientSession.ControlCode := ACRExecute;
    ClientSession.MsgControlCode := ACRExecute;
    ClientSession.Status := ACRVacant;
// Add new Session
    Sessions:=FSessions.LockList;
    try
     Sessions.Add(ClientSession);
    finally
     FSessions.UnlockList;
    end;
    if not ListenOnly then
     begin
// Connect to server
    EnterCriticalSection(FCSect);
    inc(ClientSession.Status);
    LeaveCriticalSection(FCSect);
    SendConnectRequest(ClientSession);
    if ClientSession.Status <= ACRVacant then
     begin
      DoDisconnect(Session);
      raise EACRException.Create(40039, ErrorRCannotConnetToServer,
                                [Session.ConnectParams.RemoteHost,
                                 Session.ConnectParams.RemotePort,
                                 Session.ConnectParams.LocalPort,
                                 Session.ConnectParams.ServerID]);
     end;
    if ClientSession.ControlCode = ACRTerminate then
     begin
      EnterCriticalSection(FCSect);
      dec(ClientSession.Status);
      LeaveCriticalSection(FCSect);
      Exit;
     end;
    ClientSession.PacketIDsToResend.SetSize(0); // Do not request broken packets
    ReceiveBuffer(Session, Buffer, BufferSize);
// Get connection parameters
    ConnectionParams := PACRConnectionParams(Buffer);
    ClientSession.Session.FConnectParams.PacketSize := ConnectionParams.PacketSize;
    ClientSession.Session.FConnectParams.CompressionAlgorithm := ConnectionParams.CompressionAlgorithm;
    ClientSession.Session.FConnectParams.CompressionMode := ConnectionParams.CompressionMode;
    // SessionID
    Move((Buffer+SizeOf(TACRConnectionParams))^, SessionID, SizeOf(SessionID));
    ConnectionID := -MAXINT;
{$IFDEF MsgCommunicator}
    if (BufferSize>=(SizeOf(TACRConnectionParams))+SizeOf(SessionID)+SizeOf(ConnectionID)) then
      Move((Buffer+SizeOf(TACRConnectionParams)+SizeOf(SessionID))^, ConnectionID, SizeOf(SessionID));
{$ENDIF}
    MemoryManager.FreeAndNilMem(Buffer);
    ClientSession.ServerSessionID := SessionID;
    if (ConnectionID = -MAXINT) then
      ClientSession.RemoteConnectionID := ClientSession.ConnectionID
    else
      ClientSession.RemoteConnectionID := ConnectionID;
    EnterCriticalSection(FCSect);
    dec(ClientSession.Status);
    LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('ClientSession.ServerSessionID = '+IntToStr(ClientSession.ServerSessionID));
{$ENDIF}
     end; // Connect to server
   end; //Session not Found

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT #'+IntToStr(Session.SessionID)+
              ' CONNECTED TO SERVER #'+IntToStr(Session.ConnectParams.ServerID)+
              ' SESSION #'+IntToStr(ClientSession.ServerSessionID));
aaWriteToLog('==============================================================');
{$ENDIF}

end; // Connect


//------------------------------------------------------------------------------
// SendAcknowledgement
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendAcknowledgement(
                          ClientSession: PACRClntSession; Msg: Boolean = False);
var
  Header:               PACRPacketHeader;
  Connections:          TList;
  ClientConnection:     PACRClntConnection;
  i:                    Integer;
begin
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
 try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - START');
{$ENDIF}
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   if Msg then
     Header.ControlCode := ACRMessageReceived
   else
     Header.ControlCode := ACRAllPacketsReceived;
   Header.Signature := ACRClientPacketSign;
   Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
   Header.ConnectionID := ClientSession.RemoteConnectionID;
   Header.SessionID := ClientSession.ServerSessionID;
{$IFDEF MsgCommunicator}
   if TACRClientSession(ClientSession.Session).Direct then
     Header.SessionID := ClientSession.Session.SessionID;
{$ENDIF}
   Header.PacketID := 0;
   if Msg then
     Header.CurrentRequestID := ClientSession.ServerMessageID
   else
     Header.CurrentRequestID := ClientSession.CurrentRequestID;
   Connections:=FConnections.LockList;
   try
    for i:=0 to Connections.Count-1 do
     begin
      ClientConnection := Connections.Items[i];
      if (ClientConnection.ConnectionID = ClientSession.ConnectionID) then
       break;
     end;
   finally
    FConnections.UnlockList;
   end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Enter...');
{$ENDIF}
   EnterCriticalSection(FCSect);
   try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Entered!');
{$ENDIF}
    ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
    ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Send...');
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement> ClientSession.Session.ConnectParams.RemoteHost="'+ClientSession.Session.ConnectParams.RemoteHost+'"');
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement> ClientSession.Session.ConnectParams.RemotePort='+IntToStr(ClientSession.Session.ConnectParams.RemotePort));
{$ENDIF}
    ClientConnection.Network.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Sent!');
{$ENDIF}
   finally
    LeaveCriticalSection(FCSect);
   end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - Left!');
{$ENDIF}
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.SendAcknowledgement - FINISH');
{$ENDIF}
 end;
end;// SendAcknowledgement


{$IFDEF MsgCommunicator}
//------------------------------------------------------------------------------
// SendConnectAckn
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendConnectAckn(
                          ClientSession:        PACRClntSession;
                          ClientConnection:     PACRClntConnection;
                          CurrentRequestID:     Integer = -1
                                                      );
var
  Buffer:                         PChar;
  Buf:                            PChar;
  BufferSize,
  BufSize,
  SizeSID, SizeCID, SizeParams:   Integer;
  ConnectionParams:               TACRConnectionParams;
  RequestID:                      Integer;
begin
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
 try
{
  ConnectionParams.LocalPort := ClientSession.RemotePort;
  ConnectionParams.RemoteHost := FNetwork.LocalHost;
  ConnectionParams.RemotePort := FNetwork.LocalPort;
  ConnectionParams.ServerID := FNetwork.FLocalClient;
}
  ConnectionParams.PacketSize := ClientSession.Session.ConnectParams.PacketSize;
  ConnectionParams.CompressionAlgorithm := ClientSession.Session.ConnectParams.CompressionAlgorithm;
  ConnectionParams.CompressionMode := ClientSession.Session.ConnectParams.CompressionMode;
  SizeParams := SizeOf(ConnectionParams);
  SizeSID := SizeOf(ClientSession.Session.SessionID);
  SizeCID := SizeOf(ClientSession.ConnectionID);
  BufferSize := SizeParams + SizeSID + SizeCID;
  Buffer := MemoryManager.GetMem(BufferSize);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER: SessionID='+IntToStr(ClientSession.Session.SessionID));
aaWriteToLog('SizeSID='+IntToStr(SizeSID)+', SizeParams='+IntToStr(SizeParams)+', BufferSize='+IntToStr(BufferSize));
{$ENDIF}
  try
   Move(ConnectionParams, Buffer^, SizeParams);
   Move(ClientSession.Session.SessionID, (Buffer+SizeParams)^, SizeSID);
   Move(ClientSession.ConnectionID, (Buffer+SizeParams+SizeSID)^, SizeCID);

   if CurrentRequestID >= 0 then
    begin
     RequestID := ClientSession.CurrentRequestID;
     ClientSession.CurrentRequestID := CurrentRequestID;
    end;

   CompressAndEncryptBuffer(ClientSession.Session, Buffer, BufferSize, Buf, BufSize);
   try
    DoSendBuffer(ClientSession, ClientConnection, Buf, BufSize, ACRConnected);
   finally
    if Buf<>Buffer then
      MemoryManager.FreeAndNilMem(Buf);
   end;

   if CurrentRequestID >= 0 then
     ClientSession.CurrentRequestID := RequestID;

   ClientSession.AnswerStatus := ACRNoAnswer;
  finally
   MemoryManager.FreeAndNilMem(Buffer);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
 end;
end;// SendConnectAckn


//------------------------------------------------------------------------------
// IsAuthorizationBufferValid
//------------------------------------------------------------------------------
function TACRClientConnectionManager.IsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
begin
  Result := fnIsAuthorizationBufferValid(CryptoInfo, Buffer, BufferSize);
end;// IsAuthorizationBufferValid
{$ENDIF MsgCommunicator}


//------------------------------------------------------------------------------
// SendDisconnectRequest
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendDisconnectRequest(
                                    ClientSession: PACRClntSession;
                                    WaitForAnswer: Boolean = True);
var
  Header:               PACRPacketHeader;
  Connections:          TList;
  ClientConnection:     PACRClntConnection;
  Retry, i:             Integer;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> START');
{$ENDIF}
EnterCriticalSection(FCSect);
inc(ClientSession.Status);
LeaveCriticalSection(FCSect);
try
  ClientSession.AnswerStatus := ACRNoAnswer;
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   Header.ControlCode := ACRDisconnect;
   Header.Signature := ACRClientPacketSign;
   Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
   Header.ConnectionID := ClientSession.RemoteConnectionID;
   Header.SessionID := ClientSession.ServerSessionID;
{$IFDEF MsgCommunicator}
   if TACRClientSession(ClientSession.Session).Direct then
     Header.SessionID := ClientSession.Session.SessionID;
{$ENDIF}
   Header.PacketID := 0;
   Header.CurrentRequestID := ClientSession.CurrentRequestID;
   Connections:=FConnections.LockList;
   try
    for i:=0 to Connections.Count-1 do
     begin
      ClientConnection := Connections.Items[i];
      if (ClientConnection.ConnectionID = ClientSession.ConnectionID) then
        break;
     end;
   finally
    FConnections.UnlockList;
   end;
   Retry := 0;
   ClientSession.SendStatus := ACRNotSent;
   try
    repeat
     if ClientSession.Session.ConnectParams.RemoteHost = '' then
       Exit; // Session already freed
     EnterCriticalSection(FCSect);
     try
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> RemoteHost='+ClientSession.Session.ConnectParams.RemoteHost);
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> RemotePort='+IntToStr(ClientSession.Session.ConnectParams.RemotePort));
{$ENDIF}
      ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
      ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
      if clientSession.ControlCode = ACRTerminate then
        raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
      ClientConnection.Network.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
     finally
      LeaveCriticalSection(FCSect);
     end;
     if not WaitForAnswer then
       Exit;
     StartTime := GetTickCount;
     while ((GetTickCount - StartTime) < ACRDisconnectDelay) do // pause
      if ClientSession.SendStatus <> ACRSent then
       begin
        Sleep(ACRReceiveSleep);
{$IFDEF ProcessMessages}
        Application.ProcessMessages;
{$ENDIF ProcessMessages}
       end
      else
       Exit;
     inc(Retry);
    until  (Retry > ACRDisconnectRetryCount);
   if ClientSession.SendStatus <> ACRSent then
{$IFDEF MsgCommunicator}
    if TACRClientSession(ClientSession.Session).Direct then
      raise EACRException.Create(40024, ErrorRCannotDisconnect,
                   ['client session with user', TACRClientSession(ClientSession.Session).UserID,
                    ACRDisconnectRetryCount, ACRDisconnectDelay])
    else
{$ENDIF}
      raise EACRException.Create(40024, ErrorRCannotDisconnect,
                   ['server session', ClientSession.ServerSessionID,
                    ACRDisconnectRetryCount, ACRDisconnectDelay]);
   finally
{$IFDEF DEBUG_LOG_NETWORK}
if ClientSession.SendStatus <> ACRSent then
aaWriteToLog('CLIENT: SendDisconnectRequest> Answer from server received')
else
aaWriteToLog('CLIENT: SendDisconnectRequest> Answer from server was not received');
{$ENDIF}
   end;
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
 finally
EnterCriticalSection(FCSect);
dec(ClientSession.Status);
LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendDisconnectRequest> FINISH');
{$ENDIF}
 end;
end;// SendDisconnectRequest


//------------------------------------------------------------------------------
// TerminateMessageThreads
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TerminateMessageThreads(ClientSession: PACRClntSession);
var
  i:                    Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
  Error:                String;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> START');
{$ENDIF}
  // Terminate MsgResendRequestThread
  if (ClientSession.MsgResendRequestThread <> nil) then
    ClientSession.MsgResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   sleep(0);
   if (ClientSession.MsgResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
       begin
        Error := 'message resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
        ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
       end;
     end;
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> message resend thread terminated');
{$ENDIF}
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
   // Terminate message threads
   Threads := ClientSession.MsgListeningThreads.LockList;
   try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> MsgListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ClientSession.MsgListeningThreads.UnlockList;
   end;
   sleep(0);
   // Check
   Threads := ClientSession.MsgListeningThreads.LockList;
   try
    if (Threads.Count = 0)
     then break
     else
      begin
       if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
        begin
         Error := IntToStr(Threads.Count)+' message listening threads';
         raise EACRException.Create(40036, ErrorRThreadHangs,
          ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
        end;
      end;
   finally
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateMessageThreads> MsgListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    ClientSession.MsgListeningThreads.UnlockList;
   end;
  until False;
end; // TerminateMessageThreads


//------------------------------------------------------------------------------
// TerminateCommandThreads
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TerminateCommandThreads(ClientSession: PACRClntSession);
var
  i:                    Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
  Error:                String;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> START');
{$ENDIF}
  // Terminate ResendRequestThread
  if (ClientSession.ResendRequestThread <> nil) then
    ClientSession.ResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   sleep(0);
   if (ClientSession.ResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
       begin
        Error := 'command resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
          ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
       end;
     end;
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> command resend thread terminated');
{$ENDIF}
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
   // Terminate all threads
   Threads := ClientSession.ListeningThreads.LockList;
   try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ClientSession.ListeningThreads.UnlockList;
   end;
   sleep(0);
   // Check
   Threads := ClientSession.ListeningThreads.LockList;
   try
    if (Threads.Count <= FLiveSessions)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
       begin
        Error := IntToStr(Threads.Count)+' command listening threads';
        raise EACRException.Create(40036, ErrorRThreadHangs,
          ['client session # '+IntToStr(ClientSession.Session.SessionID),
          Error]);
       end;
     end;
   finally
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
    ClientSession.ListeningThreads.UnlockList;
   end;
  until False;
end; // TerminateCommandThreads


//------------------------------------------------------------------------------
// TerminateAllSessionThreads
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.TerminateAllSessionThreads(ClientSession: PACRClntSession);
var
  StartTime:            DWORD;
  Error:                String;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateAllSessionThreads> START');
{$ENDIF}
  TerminateCommandThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateAllSessionThreads> Command threads terminated!');
{$ENDIF}
  TerminateMessageThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.TerminateAllSessionThreads> FINISH');
{$ENDIF}
end; // TerminateAllSessionThreads


//------------------------------------------------------------------------------
// DeleteSession
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DeleteSession(ClientSession: PACRClntSession);
var
  StartTime:            DWORD;
  i, Count:             Integer;
  Packet:               PACRPacket;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> START');
{$ENDIF}
  // block using ServerSession
  EnterCriticalSection(FCSect);
  // no threads - free list
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ListeningThreads will delete in @ClientSession='+IntToStr(Integer(ClientSession)));
{$ENDIF}
  // free list
  ClientSession.ListeningThreads.Free;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ListeningThreads have been deleted');
{$ENDIF}
  ClientSession.ListeningThreads := nil;
  // free command packets
  Count := ClientSession.Packets.Count;
  for i := 0 to Count - 1 do
   begin
    Packet := ClientSession.Packets.Items[i];
    if Packet <> nil then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Packet # 0: Free Buffer...');
{$ENDIF}
      if Packet.Buffer <> nil then
        MemoryManager.FreeAndNilMem(Packet.Buffer);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Packet # 0: Dispose Packet...');
{$ENDIF}
      Dispose(Packet);
     end;
   end;
  ClientSession.Packets.Free;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Packets have been deleted');
{$ENDIF}
  // free list
  ClientSession.MsgListeningThreads.Free;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Messages ListeningThreads have been deleted');
{$ENDIF}
  ClientSession.MsgListeningThreads := nil;
  // free message packets
  Count := ClientSession.MsgPackets.Count;
  for i := 0 to Count - 1 do
   begin
    Packet := ClientSession.MsgPackets.Items[i];
    if Packet <> nil then
     begin
      if Packet.Buffer <> nil then
        MemoryManager.FreeAndNilMem(Packet.Buffer);
      Dispose(Packet);
     end;
   end;
  ClientSession.MsgPackets.Free;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Messages Packets have been deleted');
{$ENDIF}
  // free
  ClientSession.PacketIDsToResend.Free;
  ClientSession.MsgPacketIDsToResend.Free;
  // wait for vacant ClientSession
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> wait for vacant ClientSession...');
{$ENDIF}
  StartTime := GetTickCount;
  LeaveCriticalSection(FCSect);
  repeat
    ClientSession.ControlCode := ACRTerminate;
    ClientSession.MsgControlCode := ACRTerminate;
    sleep(0);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ClientSession.Status = '+IntToStr(ClientSession.Status));
{$ENDIF}
    if (ClientSession.Status = ACRVacant) then
      break;
  until ((GetTickCount-StartTime) >= ACRWaitForTimeOut);
  if (ClientSession.Status > ACRVacant) then
    raise EACRException.Create(40038, ErrorRWaitForTimeOut,
                                ['client session', ACRWaitForTimeOut]);
  // free session
  EnterCriticalSection(FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> ClientSession.Status = '+IntToStr(ClientSession.Status));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> Dispose...');
{$ENDIF}
  Dispose(ClientSession);
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DeleteSession> FINISH');
{$ENDIF}
end; // DeleteSession


//------------------------------------------------------------------------------
// DoDisconnect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DoDisconnect(Session: TACRBaseSession);
var
  Connections:          TList;
  Sessions:             TList;
  ClientConnection:     PACRClntConnection;
  ClientSession:        PACRClntSession;
  ConnectionID:         TACRConnectionID;
  i, Count:             Integer;
  NoSessions:           Boolean;
  NoConnections:        Boolean;
begin
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> START');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('------------------------------');
aaWriteToLog('CLIENT: DoDisconnect> Started');
aaWriteToLog('------------------------------');
{$ENDIF}
  Sessions:=FSessions.LockList;
  try
   Count := Sessions.Count;
   for i:=Count-1 downto 0 do
    begin
    if (i<0)
    or (i>=Sessions.Count)
    then
       break;
     ClientSession := Sessions.Items[i];
     if ClientSession.Session = Session then
      begin
       ConnectionID := ClientSession.ConnectionID;
       // Terminate all threads
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> TerminateAllSessionThreads');
{$ENDIF}
       TerminateAllSessionThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> All sessions terminated');
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Remove SessionID='+IntToStr(ClientSession.Session.SessionID));
{$ENDIF}
       FSessions.Remove(ClientSession);
       FSessions.UnlockList;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> DeleteSession...');
{$ENDIF}
       DeleteSession(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Session Deleted!');
{$ENDIF}
       Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Sessions Lockeded!');
{$ENDIF}
       break;
      end;
    end;
//   Sessions.Capacity := Sessions.Count;
// Search for another session using the same connection
   NoSessions := True;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Sessions Count = '+IntToStr(Count));
{$ENDIF}
   Count := Sessions.Count;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Sessions Count = '+IntToStr(Count));
{$ENDIF}
   for i:=0 to Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if ClientSession.ConnectionID = ConnectionID then
      begin
       NoSessions := False;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;

  NoConnections := True;  // Suppose that it is not availabdle to have another connection
                          // with the same localport
{
// Search for another connection with the same localport (= the same network)
  Connections:=FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if (ClientConnection.ConnectionID = ConnectionID)
     then
      begin
       Network := ClientConnection.Network;
       break;
      end;
    end;
   NoConnections := True;
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if (ClientConnection.ConnectionID <> ConnectionID)
     and (ClientConnection.Network = Network)
     then
      begin
       NoConnections := False;
       break;
      end;
    end;
  finally
   FConnections.UnlockList;
  end;
}

// There is no another session who use this connection
 if NoSessions then
 begin
  Connections:=FConnections.LockList;
  Count := Connections.Count;
  try
   for i:=Count-1 downto 0 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.ConnectionID = ConnectionID then
      begin
       if NoConnections then
        if ClientConnection.Network <> nil then
         begin
          ClientConnection.Network.Free;
          DecThreadCount;
         end;
{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Freed connection!');
{$ENDIF}
       ClientConnection.Network := nil;
       Dispose(Connections.Items[i]);
       Connections.Delete(i);
      end;
    end;
  finally
   FConnections.UnlockList;
  end;
 end;

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('CLIENT: DoDisconnect> Finished');
aaWriteToLog('------------------------------');
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_ONDISCONNECT}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> FINISH');
{$ENDIF}

end; // DoDisconnect


//------------------------------------------------------------------------------
// Disconnect
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.Disconnect(Session: TACRBaseSession;
                                                  ListenOnly: Boolean = False);
var
  Sessions:             TList;
  ClientSession:        PACRClntSession;
  i:                    Integer;
  Found:                Boolean;
begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT #'+IntToStr(Session.SessionID)+' TRY TO DISCONNECT FROM SERVER #'+IntToStr(Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}

  if not ListenOnly then
   begin // Send disconnect request
    Found := False;
    Sessions:=FSessions.LockList;
    try
     for i:=Sessions.Count-1 downto 0 do
      begin
       ClientSession := Sessions.Items[i];
       if ClientSession.Session = Session then
        begin
         Found := True;
         break;
        end;
      end;
    finally
     FSessions.UnlockList;
    end;
    if not Found then
      raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
    inc(ClientSession.CurrentRequestID);
    SendDisconnectRequest(ClientSession);
   end; // Send disconnect request

  DoDisconnect(Session);

{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT DISCONNECTED FROM SERVER');
aaWriteToLog('==============================================================');
{$ENDIF}

end; // Disconnect


//------------------------------------------------------------------------------
// DisconnectAll
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DisconnectAll;
var
  Sessions:             TList;
  Connections:          TList;
  ClientSession:        PACRClntSession;
  ClientConnection:     PACRClntConnection;
  i:                    Integer;
begin
{$IFDEF CONNECTION_TEST}
  Exit;
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DisconnectAll> START');
{$ENDIF}
 Sessions:=FSessions.LockList;
 try
  while Sessions.Count>=1 do
//  for i:=Sessions.Count-1 downto 0 do
   begin
    ClientSession := Sessions.Items[0];
    SendDisconnectRequest(ClientSession, False);
    // Terminate all threads
    TerminateAllSessionThreads(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DisconnectAll> All sessions terminated');
{$ENDIF}
    FSessions.Remove(ClientSession);
    FSessions.UnlockList;
    DeleteSession(ClientSession);
    Sessions:=FSessions.LockList;
   end;
//  Sessions.Capacity :=0;
 finally
   FSessions.UnlockList;
 end;

 Connections:=FConnections.LockList;
 try
  for i:=Connections.Count-1 downto 0 do
   begin
    ClientConnection := Connections.Items[i];
    if ClientConnection.Network <> nil then
     begin
      ClientConnection.Network.Free;
      DecThreadCount;
     end;
    ClientConnection.Network := nil;
    Dispose(Connections.Items[i]);
    Connections.Delete(i);
   end;
//  Connections.Capacity :=0;
 finally
  FConnections.UnlockList;
 end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DisconnectAll> FINISH');
{$ENDIF}
end; // DisconnectAll


//------------------------------------------------------------------------------
// SendMessage
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                                                  );
var
  Buf:  PChar;
begin
 try
{
  if (TACRCompressionAlgorithm(Session.ConnectParams.CompressionAlgorithm) <> acaNone)
  or (Session.ConnectParams.CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)
  then
   begin
    Buf := MemoryManager.GetMem(BufferSize);
    Move(Buffer^,Buf^,BufferSize);
    SendBuffer(Session, Buf, BufferSize, ACRMessage);
    MemoryManager.FreeAndNilMem(Buf);
   end
  else
}
  SendBuffer(Session, Buffer, BufferSize, ACRMessage);
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientConnectionManager.SendMessage - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // SendMessage


//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNewRequest
                                                  );
var
  Connections:          TList;
  Sessions:             TList;
  ClientConnection:     PACRClntConnection;
  ClientSession:        PACRClntSession;
  ConnectionID:         TACRConnectionID;
  i:                    Integer;
  Found:                Boolean;
  Buf:                  PChar;
  BufSize:              Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - START');
{$ENDIF}
 try
  if (Session.ConnectParams.PacketSize < SizeOf(TACRPacketHeader)) then
   raise EACRException.Create(40020, ErrorRPacketSizeTooSmall,
                                    [Session.ConnectParams.PacketSize,
                                     SizeOf(TACRPacketHeader)]);
//  ClientSession := FindSession(Session);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - started');
{$ENDIF}
  Found := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ClientSession := Sessions.Items[i];
     if ClientSession.Session = Session then
       begin
        ConnectionID := ClientSession.ConnectionID;
        Found := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
  Found := False;
  Connections := FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.ConnectionID = ConnectionID then
       begin
        Found := True;
        break;
       end;
    end;
  finally
   FConnections.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);

{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('CLIENT>>> '+IntToStr(ClientSession.CurrentRequestID + 1)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}

  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);

  if Code <> ACRConnect then
    CompressAndEncryptBuffer(ClientSession.Session, Buffer, BufferSize, Buf, BufSize)
  else
   begin
    Buf := Buffer;
    BufSize := BufferSize;
   end;
  try
   DoSendBuffer(ClientSession, ClientConnection, Buf, BufSize, Code);
  finally
   if Buf <> Buffer then
    if Code <> ACRConnect then
     MemoryManager.FreeAndNilMem(Buf);
  end;

  if ClientSession.Status <= ACRVacant then
    raise EACRException.Create(40040, ErrorRCannotSendToServer,
                                [ClientSession.Session.ConnectParams.RemoteHost,
                                 ClientSession.Session.ConnectParams.RemotePort,
                                 ClientSession.Session.ConnectParams.LocalPort,
                                 ClientSession.Session.ConnectParams.ServerID]);

  EnterCriticalSection(FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FCSect);

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.SendBuffer - FINISH');
{$ENDIF}

 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRClientConnectionManager.SendBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // SendBuffer


//------------------------------------------------------------------------------
// DoSendBuffer
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.DoSendBuffer(
                          ClientSession:    PACRClntSession;
                          ClientConnection: PACRClntConnection;
                          Buffer:           PChar;
                          BufferSize:       Integer;
                          Code:             Integer = ACRNewRequest
                                                  );
var
  Header:               PACRPacketHeader;
  Packets:              TList;
  Packet:               PACRPacket;
  BytesSent, DataSize:  Integer;
  i:                    Integer;

procedure WaitForCommandSent;
var
  PacketID,
  i:                    Integer;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
  j:                    Integer;
{$ENDIF}
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT waits for all packets');
{$ENDIF}
  if not ((Code = ACRConnect) or (Code = (ACRConnect + ACRLastPacket))) then
   begin
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRReceiveSleep); // To avoid delay if full answer is already received
    i := ClientSession.PacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ClientSession.SendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT has '+IntToStr(ClientSession.PacketIDsToResend.ItemCount)+' packets to resend:');
for j:=0 to ClientSession.PacketIDsToResend.ItemCount-1 do
aaWriteToLog(IntToStr(ClientSession.PacketIDsToResend.Items[j]));
{$ENDIF}
      if (GetTickCount - StartTime) > ACRSendTimeOut then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40077, ErrorRTimeOutSending,
          ['command', ClientSession.Session.ConnectParams.ServerID, ACRSendTimeOut]);
       end;
      if ClientSession.ControlCode = ACRTerminate then
        Exit;
      dec(i);
      if (i<0) then
        i := ClientSession.PacketIDsToResend.ItemCount - 1;
      EnterCriticalSection(FCSect);
      try
       if (ClientSession.PacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ClientSession.PacketIDsToResend.Items[i];
           if PacketID >= Packets.Count then
             Continue;
          end
         else // No packets to resend - Resend last packet
           PacketID := Packets.Count - 1;
         Packet := Packets.Items[PacketID];
         // resend packet
         ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
         ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
         if ClientSession.ControlCode = ACRTerminate then
          begin
           dec(ClientSession.Status);
           raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
          end;
         try
          ClientConnection.Network.SendBuffer(Packet.Buffer, Packet.BufferSize);
         except
          dec(ClientSession.Status);
          raise;
         end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT resent packet # '+IntToStr(PacketID));
{$ENDIF}
        end;
      finally
       if  (ClientSession.PacketIDsToResend.ItemCount > 0)
       and (ClientSession.PacketIDsToResend.ItemCount > i)
       then
         ClientSession.PacketIDsToResend.Delete(i);
       LeaveCriticalSection(FCSect);
      end; // finally
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRResendDelay);
      Sleep(0);
      if ClientSession.PacketIDsToResend.ItemCount = 0 then
       StartSleepTime := GetTickCount;
       while ((GetTickCount - StartSleepTime) < ACRResendDelay) do  // sleep after last packet send
        begin
         if ClientSession.SendStatus = ACRSent then
           break;
         Sleep(ACRReceiveSleep);
        end; // sleep
     end;
   end; // no Connect
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('CLIENT received all the packets');
{$ENDIF}
end; // WaitForCommandSent

procedure WaitForMessageSent;
var
  PacketID,
  i:                    Integer;
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRReceiveSleep); // To avoid delay if full answer is already received
    i := ClientSession.MsgPacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ClientSession.MsgSendStatus = ACRSent) do  // resend broken packets
     begin
      if (GetTickCount - StartTime) > ACRSendTimeOut then
       begin
        EnterCriticalSection(FCSect);
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise EACRException.Create(40077, ErrorRTimeOutSending,
          ['message', ClientSession.Session.ConnectParams.ServerID, ACRSendTimeOut]);
       end;
      if ClientSession.MsgControlCode = ACRTerminate then
        Exit;
      dec(i);
      if i<0 then
        i := ClientSession.MsgPacketIDsToResend.ItemCount - 1;
      EnterCriticalSection(FCSect);
      try
       if (ClientSession.MsgPacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ClientSession.MsgPacketIDsToResend.Items[i];
           if PacketID >= Packets.Count then
             Continue;
          end
         else // No packets to resend - Resend last packet
           PacketID := Packets.Count - 1;
         Packet := Packets.Items[PacketID];
         // resend packet
         ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
         ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
         if ClientSession.MsgControlCode = ACRTerminate then
          begin
           dec(ClientSession.Status);
           raise EACRException.Create(40081, ErrorRClientMsgSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
          end;
         try
          ClientConnection.Network.SendBuffer(Packet.Buffer, Packet.BufferSize);
         except
          dec(ClientSession.Status);
          raise;
         end;
        end;
      finally
       if  (ClientSession.MsgPacketIDsToResend.ItemCount > 0)
       and (ClientSession.MsgPacketIDsToResend.ItemCount > i)
       then
         ClientSession.MsgPacketIDsToResend.Delete(i);
       LeaveCriticalSection(FCSect);
      end; // finally
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRResendDelay);
      Sleep(0);
      if ClientSession.MsgPacketIDsToResend.ItemCount = 0 then
       StartSleepTime := GetTickCount;
       while ((GetTickCount - StartSleepTime) < ACRResendDelay) do  // sleep after last packet send
        begin
         if ClientSession.SendStatus = ACRSent then
           break;
         Sleep(ACRReceiveSleep);
        end; // sleep
     end;
end; // WaitForMessageSent

procedure CheckTermination;
begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> BEGIN');
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> ClientSession.Status = '+IntToStr(ClientSession.Status));
{$ENDIF}
  try
        if Code = ACRMessage then
         begin
          if ClientSession.MsgControlCode = ACRTerminate then
           begin
            dec(ClientSession.Status);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> 40081');
{$ENDIF}
            raise EACRException.Create(40081, ErrorRClientMsgSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
           end;
         end
        else
          if ClientSession.ControlCode = ACRTerminate then
           begin
            dec(ClientSession.Status);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> 40080');
{$ENDIF}
            raise EACRException.Create(40080, ErrorRClientSessionTerminated,[Integer(ClientSession.Session.SessionID)]);
           end;
  finally
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> ClientSession.Status = '+IntToStr(ClientSession.Status));
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer.CheckTermination> END');
{$ENDIF}
  end;
end; // CheckTermination

begin // DoSendBuffer
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer - START');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FCSect);
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  DataSize := ClientSession.Session.ConnectParams.PacketSize - SizeOf(TACRPacketHeader);
  Header.Signature := ACRClientPacketSign;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> Code='+IntToStr(Code));
{$ENDIF}
  Header.Recepient := ClientSession.Session.ConnectParams.ServerID;
  Header.Sender := ClientConnection.Network.LocalClientID;
//  Header.ConnectionID := ClientConnection.ConnectionID;
  Header.ConnectionID := ClientSession.RemoteConnectionID;
////  if Code=ACRConnect then
  Header.SessionID := ClientSession.ServerSessionID;
{$IFDEF MsgCommunicator}
  if TACRClientSession(ClientSession.Session).Direct then
   if Code<>ACRConnected then
    Header.SessionID := ClientSession.Session.SessionID;
{$ENDIF}
  Header.PacketID := 0;
  Header.ControlCode := Code;
  if (Code >= ACRNewRequest)
  or (Code = ACREcho)
  then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT IS SENDING NEW REQUEST TO SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
    EnterCriticalSection(FCSect);
    inc(ClientSession.CurrentRequestID);
    ClientSession.Packets.Count := 0; // Delete old packets receved since ReceiveBuffer finished
    ClientSession.AnswerStatus := ACRNoAnswer;
    LeaveCriticalSection(FCSect);
   end;
  Header.CurrentRequestID := ClientSession.CurrentRequestID;
  if Code = ACRMessage then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT IS SENDING MESSAGE # '+IntToStr(ClientSession.ClientMessageID)+'TO SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('==============================================================');
{$ENDIF}
    Header.CurrentRequestID := ClientSession.ClientMessageID;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> create packets list...');
{$ENDIF}
  Packets := TList.Create;
  try
    BytesSent := 0;
    i := DataSize;
    if Code = ACRMessage then
     begin
      ClientSession.MsgSendStatus := ACRNotSent;
     end
    else
      ClientSession.SendStatus := ACRNotSent;
    repeat
//    while BytesSent < BufferSize do
//     begin // Create and send all packets
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> create packet #'+IntToStr(Header.PacketID));
{$ENDIF}
      New(Packet);
      Packets.Add(Packet);
      Packet.Buffer := MemoryManager.GetMem(ClientSession.Session.ConnectParams.PacketSize);
      if BytesSent + DataSize > BufferSize then
       DataSize := BufferSize - BytesSent;
      Packet.BufferSize := DataSize + SizeOf(TACRPacketHeader);
      if BytesSent + DataSize = BufferSize then
       Header.ControlCode := Header.ControlCode+ACRLastPacket;
      Move(Header^, Packet.Buffer^, SizeOf(TACRPacketHeader));
      Move(Pointer(Integer(Buffer)+Header.PacketID*i)^, (Packet.Buffer+SizeOf(TACRPacketHeader))^, DataSize);
      inc(Header.PacketID);
      // send packet
      EnterCriticalSection(FCSect);
      try
       ClientConnection.Network.RemoteHost := ClientSession.Session.ConnectParams.RemoteHost;
       ClientConnection.Network.RemotePort := ClientSession.Session.ConnectParams.RemotePort;
       CheckTermination;
       try
        ClientConnection.Network.SendBuffer(Packet.Buffer, Packet.BufferSize);
       except
        dec(ClientSession.Status);
        LeaveCriticalSection(FCSect);
        raise;
       end;
      finally
       LeaveCriticalSection(FCSect);
      end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> packet sent');
{$ENDIF}
      // packet has been sent
      BytesSent := BytesSent + DataSize;
//     end;
    until BytesSent >= BufferSize;
    if Code = ACRMessage then
     begin
      WaitForMessageSent;
      inc(ClientSession.ClientMessageID);
     end
    else
      if Code <> ACRConnected then
        WaitForCommandSent;
  finally
  // Remove all sent packets and free memory
    for i:= 0 to Packets.Count - 1 do
     begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> free packet #'+IntToStr(i));
{$ENDIF}
      Packet := Packets.Items[i];
      MemoryManager.FreeAndNilMem(Packet.Buffer);
      Dispose(Packet);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> freed');
{$ENDIF}
     end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> free packets...');
{$ENDIF}
    Packets.Free;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> freed');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> free header...');
{$ENDIF}
    MemoryManager.FreeAndNilMem(Header);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer> freed');
{$ENDIF}
   EnterCriticalSection(FCSect);
   dec(ClientSession.Status);
   LeaveCriticalSection(FCSect);
   CheckTermination;
  end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT sent buffer');
aaWriteToLog('==============================================================');
{$ENDIF}
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('TACRClientConnectionManager.DoSendBuffer - FINISH');
{$ENDIF}
end; // DoSendBuffer


//------------------------------------------------------------------------------
// if encryption algorithm <> acr_Cipher_None then allocate buffer, fill it and return size
// otherwise return BufferSize = 0
//------------------------------------------------------------------------------
procedure TACRClientConnectionManager.CreateAuthorizationBuffer(
                      CryptoInfo:     TACRCryptoInfo;
                      out Buffer:     PChar;
                      out BufferSize: Integer
                                    );
var ms:           TACRMemoryStream;
    buf:          PChar;
    crc32:        Cardinal;
    size:         Integer;
begin
  if CryptoInfo.CryptoAlgorithm = ACR_Cipher_None then
   begin
    BufferSize := 0;
    Exit;
   end;
  size := ACRDefaultAuthorizationBufferSize;
  buf := MemoryManager.GetMem(size);
  ms := TACRMemoryStream.Create;
  try
    ACRGenerateRandomBuffer(buf,size);
    crc32 := ACRCountCRC(0,buf,size);
    SaveDataToStream(size,SizeOf(crc32),ms,11319);
    SaveDataToStream(crc32,SizeOf(crc32),ms,11320);
    ACREncryptBuffer(CryptoInfo,buf,size);
    ms.WriteBuffer(buf^,size);
    BufferSize := ms.Size;
    Buffer := ms.Buffer;
    ms.Buffer := nil;
  finally
    ms.Free;
    MemoryManager.FreeAndNilMem(buf);
  end;
end; // CreateAuthorizationBuffer

// TACRClientConnectionManager



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientListenerThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientListenerThread.Create(
                       Manager:       TACRClientConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       Network:       TACRNetwork;
                       FromHost:      String;
                       FromPort:      Integer
                                            );
begin
  FManager := Manager;
  FManager.IncThreadCount;
  FNetwork := Network;
  self.Buffer := Buffer;
  self.BufferSize := BufferSize;
  self.FromHost := FromHost;
  self.FromPort := FromPort;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientListenerThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if ClientSession <> nil then
   begin
    if ClientSession.ListeningThreads <> nil then
      ClientSession.ListeningThreads.Remove(self);
    if ClientSession.MsgListeningThreads <> nil then
      ClientSession.MsgListeningThreads.Remove(self);
   end;
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT LISTENER THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRClientListenerThread.Execute;
label
  KillPacket;
var
  Header:               PACRPacketHeader;
  Packet:               PACRPacket;
  Connections:          TList;
  Sessions:             TList;
  Session:              TACRBaseSession;
  ClientSession2:       PACRClntSession;
  ClientConnection:     PACRClntConnection;
  AllPacketsReceived,
  SessionFound,
  ConnectionFound:      Boolean;
  Buf:                  PChar;
  BufSize, AuBufSize:   Integer;
  i:                    Integer;
  UserID:               Cardinal;
  ConnectionParams:     PACRConnectionParams;
  Application:          String;
begin
//EnterCriticalSection(FManager.FCSect);
try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** CLIENT LISTENER START *****');
{$ENDIF}
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('CLIENT LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
  Header := Pointer(Buffer);
// Check for errors
  if CheckSum(Buffer, BufferSize) <> Header.CheckSum then
   begin
    FManager.PacketResendRequest(Buffer, ClientConnection.Network, FromHost, FromPort);
    goto KillPacket;
   end;
// Check sign
  if (Header.Signature <> ACRServerPacketSign)
{$IFDEF MsgCommunicator}
  and (Header.Signature <> ACRClientPacketSign)
{$ENDIF}
  then
    goto KillPacket;
{$IFDEF MsgCommunicator}
  if    // first packet in multi-packet connect request
  ((Header.ControlCode=ACRConnect) and (Header.PacketID=0))
  or    // single-packet connect request
  ((Header.ControlCode=(ACRConnect+ACRLastPacket)) and (Header.PacketID=0))
  then
   begin // Connect
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connect');
{$ENDIF}
    // is direct connection enabled?
    SessionFound := False;
    Sessions := FManager.FSessions.LockList;
    try
     for i:=0 to Sessions.Count-1 do
      begin
       ClientSession2 := Sessions.Items[i];
       if (ClientSession2.ServerSessionID = INVALID_SESSION_ID) then
        if TACRClientSession(ClientSession2.Session)<>nil then
         if TACRClientSession(ClientSession2.Session).UserID = Header.Recepient then // for connect request only
          if TACRClientSession(ClientSession2.Session).Direct then
           if ClientSession2.Session.Connected then
            if (ClientSession2.Session.SessionID = INVALID_SESSION_ID) then // special listening session
             begin
              Session := ClientSession2.Session;
              SessionFound := True;
              break;
             end;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
    if not SessionFound then
      goto KillPacket;
    // does connection already exist?
    SessionFound := False;
    Sessions := FManager.FSessions.LockList;
    try
     for i:=0 to Sessions.Count-1 do
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Check session '+IntToStr(i)+' of '+IntToStr(Sessions.Count));
{$ENDIF}
       ClientSession := Sessions.Items[i];
       if (ClientSession.ServerSessionID = Header.SessionID)
       and (ClientSession.RemoteConnectionID = Header.ConnectionID)
       then
         if (ClientSession.Session = nil) // just added, not finished
          then SessionFound := True
         else if (ClientSession.Session.SessionID = INVALID_SESSION_ID) // buffer received, not finished
          then SessionFound := True
         else if (ClientSession.Session.ConnectParams.ServerID = Header.Sender) // not first connect - session already exist
          then SessionFound := True;
         if SessionFound then
          begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ClientSession),6)+', Client SessionID='+IntToStr(ClientSession.ServerSessionID));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
if ClientSession.Session <> nil then
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ClientSession),6)+', Server SessionID='+IntToStr(ClientSession.Session.SessionID));
{$ENDIF}
           break;
          end;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
    if not SessionFound then
     begin // First connect
// new ClientSession
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New ClientSession');
{$ENDIF}
  ConnectionFound := False;
  Connections := FManager.FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.Network = FNetwork then
       begin
        ConnectionFound := True;
        break;
       end;
    end;
  finally
   FManager.FConnections.UnlockList;
  end;
  if not ConnectionFound then
    goto KillPacket;
      New(ClientSession);
      ClientSession.Session := nil;
      FManager.FSessions.Add(ClientSession);
      ClientSession.AnswerStatus := ACRNoAnswer;
      ClientSession.Status := ACRVacant;
      ClientSession.RemoteConnectionID := Header.ConnectionID;
      ClientSession.ConnectionID := ClientConnection.ConnectionID;
      ClientSession.ServerSessionID := Header.SessionID;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ClientSession),6)+', Client SessionID='+IntToStr(ClientSession.ServerSessionID));
{$ENDIF}
//      ClientSession.Thread := nil;
//      ClientSession.MsgThread := nil;
      ClientSession.ListeningThreads := TThreadList.Create;
      ClientSession.ControlCode := ACRExecute;
//      ClientSession.ListeningThreads.Add(self);
      ClientSession.MsgListeningThreads := TThreadList.Create;
      ClientSession.MsgControlCode := ACRExecute;
// Create PacketIDsToResend
      ClientSession.PacketIDsToResend := TACRThreadIntArray.Create;
      ClientSession.MsgPacketIDsToResend := TACRThreadIntArray.Create;
      ClientSession.ResendRequestThread := nil;
      ClientSession.MsgResendRequestThread := nil;
// Create new packets list
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packets list create');
{$ENDIF}
      ClientSession.Packets := TList.Create;
      ClientSession.Packets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
      ClientSession.MsgPackets := TList.Create;
      ClientSession.MsgPackets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
// add new session to list
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': add new session to list');
{$ENDIF}
      SessionFound := True;
     end
    else // not first connect - reset and resend connection info
     begin
      // Is buffer receiving now?
      if (ClientSession.AnswerStatus = ACRStartAnswer)
      or (ClientSession.AnswerStatus = ACRFullAnswer)
      or (ClientSession.Session = nil) // just added, not finished
      or (ClientSession.Session.SessionID = INVALID_SESSION_ID) // buffer received, not finished
      then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer is receiving now');
{$ENDIF}
        goto KillPacket;
       end;
      if (ClientSession.ControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Not Execute - Terminate ListeningThread, ClientSession.ControlCode='+IntToStr(ClientSession.ControlCode));
{$ENDIF}
        goto KillPacket;
       end;
      if (ClientSession.CurrentRequestID > Header.CurrentRequestID) then // packet from old connection request
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - SendAckn');
{$ENDIF}
//        Header.Recepient := Header.Sender;
        if ClientSession.Session<>nil then
          FManager.SendConnectAckn(ClientSession, ClientConnection, Header.CurrentRequestID);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - Terminate ListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
//      ClientSession.MsgThread := nil;
//      ClientSession.ListeningThreads.Add(self);
// Set PacketIDsToResend
//      ClientSession.PacketIDsToResend.Lock;
      ClientSession.PacketIDsToResend.SetSize(0);
//      ClientSession.PacketIDsToResend.Unlock;
//      ClientSession.MsgPacketIDsToResend.Lock;
      ClientSession.MsgPacketIDsToResend.SetSize(0);
//      ClientSession.MsgPacketIDsToResend.Unlock;
(*
// Delete old packets from list
      Packets := ClientSession.Packets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ClientSession.Packets.UnlockList;
      end;
      Packets := ClientSession.MsgPackets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ClientSession.Packets.UnlockList;
      end;
*)
     end;
// Put this first packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New Packet');
{$ENDIF}
    New(Packet);
    Packet.Buffer := Buffer;
    Packet.BufferSize := BufferSize;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufferSize='+IntToStr(BufferSize));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet add');
{$ENDIF}
    ClientSession2.Packets.Add(Packet);
    ClientSession.CurrentRequestID := Header.CurrentRequestID;
    ClientSession.ClientMessageID := 0;
    ClientSession.ServerMessageID := 0;
    Session.FConnectParams.RemoteHost := FromHost;
    Session.FConnectParams.RemotePort := FromPort;
    EnterCriticalSection(FManager.FCSect);
    Session.FConnectParams.ServerID := Header.Sender; // Allow following communication
    LeaveCriticalSection(FManager.FCSect);
// Get connection parameters
    if Header.ControlCode=ACRConnect+ACRLastPacket then
     ClientSession2.AnswerStatus := ACRFullAnswer;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSessionThread.Create');
{$ENDIF}
    FManager.ReceiveBuffer(ClientSession2.Session, Buf, BufSize);
    ClientSession.Session := Session;
    ClientSession.ListeningThreads.Add(self);
    UserID := TACRClientSession(Session).UserID;
    Move(Buf^, UserID, SizeOf(UserID));
    inc(Buf, SizeOf(UserID));
    Move(PChar(Buf + SizeOf(TACRConnectionParams))^, AuBufSize, SizeOf(AuBufSize));
    if not FManager.IsAuthorizationBufferValid(
              Session.ConnectParams.CryptoInfo,
              PChar(Buf + SizeOf(TACRConnectionParams) + SizeOf(AuBufSize)),
              AuBufSize
                                               )
    then
     begin
      FManager.SendDisconnectRequest(ClientSession, False);
      FManager.FSessions.Remove(ClientSession);
      FManager.DeleteSession(ClientSession);
      Exit;
     end;
    // Get ConnectParams
    ConnectionParams := PACRConnectionParams(Buf);
    // Get client Application name
    SetLength(Application, BufSize-SizeOf(TACRConnectionParams)
      -SizeOf(AuBufSize)-AuBufSize-1-SizeOf(UserID));
    StrCopy(PChar(Application), Buf+SizeOf(TACRConnectionParams)+SizeOf(AuBufSize)+AuBufSize);
    // Free
    dec(Buf, SizeOf(UserID));
    ClientSession.Session := Session.ConnectUser(UserID, FromHost, FromPort);
    if ClientSession.Session = nil then
     begin
      ClientSession.Session := Session;
      FManager.SendDisconnectRequest(ClientSession, False);
      FManager.FSessions.Remove(ClientSession);
      FManager.DeleteSession(ClientSession);
      Exit;
     end
    else
     begin
      ClientSession.AnswerStatus := ACRStartAnswer;
//      ClientSession.Session.FConnectParams.CryptoInfo := TACRServer(FManager.FServer).CryptoParams.GetCryptoParams;
// Set SessionID
      EnterCriticalSection(FManager.FCSect);
      ClientSession.Session.FConnectParams.RemoteHost := FromHost;
      ClientSession.Session.FConnectParams.RemotePort := FromPort;
      ClientSession.Session.FConnectParams.ServerID := Session.FConnectParams.ServerID;
      ClientSession.Session.SessionID := FManager.FSessionID;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ClientSession),6)+', Server SessionID='+IntToStr(ClientSession.Session.SessionID));
{$ENDIF}
      dec(FManager.FSessionID);
      if FManager.FSessionID = INVALID_SESSION_ID then
        dec(FManager.FSessionID);
      LeaveCriticalSection(FManager.FCSect);
      ClientSession.Session.FConnectParams.PacketSize := ConnectionParams.PacketSize;
      ClientSession.Session.FConnectParams.CompressionAlgorithm := ConnectionParams.CompressionAlgorithm;
      ClientSession.Session.FConnectParams.CompressionMode := ConnectionParams.CompressionMode;
//      Header.Recepient := Header.Sender;
      FManager.SendConnectAckn(ClientSession, ClientConnection);
     end;
    MemoryManager.FreeAndNilMem(Buf);
    Exit;
   end; // Connect
{$ENDIF MsgCommunicator}
  Header.Recepient := Header.Sender;
// verify ConnectionID existing
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('           Header.PacketID         = '+IntToStr(Header.PacketID));
aaWriteToLog('           Header.ControlCode      = '+IntToStr(Header.ControlCode));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connections');
{$ENDIF}
  ConnectionFound := False;
  Connections := FManager.FConnections.LockList;
  try
   for i:=0 to Connections.Count-1 do
    begin
     ClientConnection := Connections.Items[i];
     if ClientConnection.ConnectionID = Header.ConnectionID then
       begin
        ConnectionFound := True;
        break;
       end;
    end;
  finally
   FManager.FConnections.UnlockList;
  end;
  if not ConnectionFound then
    goto KillPacket;
// verify SessionID in this connection
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sessions');
{$ENDIF}
  SessionFound := False;
  Sessions := FManager.FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':');
aaWriteToLog('           Sessions.Count='+IntToStr(Sessions.Count));
aaWriteToLog('           i='+IntToStr(i));
{$ENDIF}
     ClientSession := Sessions.Items[i];
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           @ClientSession='+IntToStr(Integer(ClientSession)));
aaWriteToLog('           ClientSession.ConnectionID='+IntToStr(Integer(ClientSession.ConnectionID)));
aaWriteToLog('           Header.ConnectionID='+IntToStr(Integer(Header.ConnectionID)));
aaWriteToLog('           ClientSession.ServerSessionID='+IntToStr(Integer(ClientSession.ServerSessionID)));
aaWriteToLog('           ClientSession.Session.SessionID='+IntToStr(Integer(ClientSession.Session.SessionID)));
aaWriteToLog('           Header.ServerSessionID='+IntToStr(Integer(Header.SessionID)));
{$ENDIF}
{
     if (ClientSession.ConnectionID = Header.ConnectionID)
     and (
               (ClientSession.ServerSessionID = Header.SessionID)
            or (Header.ControlCode = ACRConnected)
            or (Header.ControlCode = ACRConnected+ACRLastPacket)
          )
}
     if (ClientSession.ConnectionID = Header.ConnectionID)
     and (
          (ClientSession.ServerSessionID = Header.SessionID)
{
          or
          (
            (ClientSession.ServerSessionID = INVALID_SESSION_ID)
            and
            (
             (Header.ControlCode = ACRConnected)
             or
             (Header.ControlCode = ACRConnected+ACRLastPacket)
            ) // and
          ) // or
}
         ) // and
     then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FManager.FSessions.UnlockList;
  end;
{$IFDEF LOG_CLIENT_THREADS}
if SessionFound then
aaWriteToLog('           Session Found')
else
aaWriteToLog('           Session Not Found')
;
{$ENDIF}
  if not SessionFound then
    goto KillPacket;

  // is listener stoped?
  if (Header.ControlCode = ACRMessage)
  or (Header.ControlCode = ACRMessage + ACRLastPacket)
  then
    if ClientSession.MsgControlCode <> ACRExecute then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Messages Listeninig blocked');
{$ENDIF}
      goto KillPacket;
     end
  else
    if ClientSession.ControlCode <> ACRExecute then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Commands Listeninig blocked');
{$ENDIF}
      goto KillPacket;
     end;
{
  EnterCriticalSection(FManager.FCSect);
  inc(ClientSession.Status);
  LeaveCriticalSection(FManager.FCSect);
}

// verify Sender
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sender');
{$ENDIF}
  if ClientSession.Session.ConnectParams.ServerID <> Header.Sender then
   begin
{$IFDEF MsgCommunicator}
    if (ClientSession.Session.ConnectParams.ServerID = Integer(MSG_INVALID_USER_ID)) // new direct session
    and ((Header.ControlCode = ACRConnected)                // just connected
      or (Header.ControlCode = ACRConnected+ACRLastPacket))
    then
     begin
      ClientSession.Session.FConnectParams.ServerID := Header.Sender;
     end
    else
{$ENDIF}
      goto KillPacket;
   end;
// check CurrentRequestID
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
{$ENDIF}
  if (Header.ControlCode = ACRMessageReceived) then
   begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ClientMessageID = '+IntToStr(ClientSession.ClientMessageID));
{$ENDIF}
    if ClientSession.ClientMessageID <> Header.CurrentRequestID then
      goto KillPacket;
   end
  else
  if (Header.ControlCode = ACRMessage)
  or (Header.ControlCode = (ACRMessage + ACRLastPacket))
  then
   begin
    if ClientSession.ServerMessageID <> Header.CurrentRequestID then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.ServerMessageID = '+IntToStr(ClientSession.ServerMessageID));
{$ENDIF}
      goto KillPacket;
     end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Message Listening Thread in '+IntToStr(Integer(ClientSession)));
{$ENDIF}
    if ClientSession.MsgListeningThreads = nil then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': MsgListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
      ClientSession.MsgListeningThreads.Add(self);
   end
  else
   begin
    if Header.ControlCode<>ACRDisconnect then
     if ClientSession.CurrentRequestID <> Header.CurrentRequestID then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientSession.CurrentRequestID = '+IntToStr(ClientSession.CurrentRequestID));
{$ENDIF}
       goto KillPacket;
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Listening Thread in '+IntToStr(Integer(ClientSession)));
{$ENDIF}
    if ClientSession.ListeningThreads = nil then
     begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
      ClientSession.ListeningThreads.Add(self);
   end;
// Process service packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Process service packet');
{$ENDIF}
  if (Header.ControlCode = ACRConnected)
  or (Header.ControlCode = (ACRConnected+ACRLastPacket))
    then ClientSession.ServerSessionID := Header.SessionID; // To allow disconnect immediately
  AllPacketsReceived := False;
  case Header.ControlCode of
   ACRDisconnect:
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRDisconnect');
{$ENDIF}
     if Header.SessionID = INVALID_SESSION_ID then
       goto KillPacket;
     inc(FManager.FLiveSessions);
     ClientSession.ControlCode := ACRTerminate;
     sleep(0);
     ClientSession.CurrentRequestID := Header.CurrentRequestID;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendACKN');
{$ENDIF}
     FManager.SendAcknowledgement(ClientSession);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session.OnDisconnect');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect);
     ClientSession.Session.OnDisconnect;
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TerminateAllSessionThreads');
{$ENDIF}
     FManager.TerminateAllSessionThreads(ClientSession);
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': FSessions.Remove...');
{$ENDIF}
     FManager.FSessions.Remove(ClientSession);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DeleteSession...');
{$ENDIF}
     FManager.DeleteSession(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT DISCONNECTED FROM SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
{$ENDIF}
     dec(FManager.FLiveSessions);
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
     goto KillPacket;
}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': End of Disconnect'); 
{$ENDIF} 
    end;
   ACRPacketResendRequest:
    begin 
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRPacketResendRequest');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect); 
     ClientSession.PacketIDsToResend.Append(Header.PacketID);
     LeaveCriticalSection(FManager.FCSect); 
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------'); 
aaWriteToLog('CLIENT received Command Packet Resend Request ID='+IntToStr(Header.PacketID));
{$ENDIF} 
     goto KillPacket;
    end; 
   ACRMessagePacketResendRequest: 
    begin
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRMessagePacketResendRequest'); 
{$ENDIF} 
     EnterCriticalSection(FManager.FCSect); 
     ClientSession.MsgPacketIDsToResend.Append(Header.PacketID); 
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received Message Packet Resend Request ID='+IntToStr(Header.PacketID));
{$ENDIF}
     goto KillPacket; 
    end;
   ACRAllPacketsReceived: 
    begin
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRAllPacketsReceived'); 
{$ENDIF}
     EnterCriticalSection(FManager.FCSect); 
     ClientSession.SendStatus := ACRSent;
     ClientSession.PacketIDsToResend.SetSize(0); 
     LeaveCriticalSection(FManager.FCSect); 
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received Command Acknowledgement'); 
{$ENDIF} 
     sleep(0); 
     goto KillPacket; 
    end; 
   ACRMessageReceived: 
    begin 
     EnterCriticalSection(FManager.FCSect);
     ClientSession.MsgSendStatus := ACRSent;
     ClientSession.MsgPacketIDsToResend.SetSize(0); 
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------'); 
aaWriteToLog('CLIENT received Message Acknowledgement');
{$ENDIF}
     goto KillPacket; 
    end; 
   ACRMessage, 
   (ACRMessage+ACRLastPacket):
    begin
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACR packet');
{$ENDIF}
//     EnterCriticalSection(FManager.FCSect); 
     if ClientSession.ServerMessageID = Header.CurrentRequestID then 
     if (ClientSession.MsgReceiveStatus = ACRFullAnswer) then // lost acknowledgement
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Send ACKN'); 
{$ENDIF} 
//       FManager.SendAcknowledgement(ClientSession, True); 
{$IFDEF DEBUG_LOG_NETWORK} 
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet while the answer received fully'); 
{$ENDIF} 
//       LeaveCriticalSection(FManager.FCSect); 
       goto KillPacket; // Do not replace correct packets with doubles
      end;
//     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Enter...');
{$ENDIF}
      EnterCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Entered!'); 
{$ENDIF}
// Is this packet already stored?
      if ClientSession.MsgPackets.List <> nil then // list exists
       if ((Header.PacketID+1) <= ClientSession.MsgPackets.Count) then // old packet
       if (ClientSession.MsgPackets.Items[Header.PacketID] <> nil) then // already received
        begin
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.MsgPackets.Items[Header.PacketID])));
{$ENDIF} 
{$IFDEF DEBUG_LOG_NETWORK} 
aaWriteToLog('--------------------------------------------------------------'); 
aaWriteToLog('CLIENT received existing packet'); 
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect); 
         goto KillPacket; // Do not replace correct packets with doubles 
        end;
// Add correct received data packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New packet');
if ((Header.PacketID+1) <= ClientSession.MsgPackets.Count) then // old packet 
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.MsgPackets.Items[Header.PacketID]))); 
{$ENDIF}
     if ClientSession.MsgPackets = nil then
       begin 
{$IFDEF LOG_CLIENT_THREADS} 
aaWriteToLog('           Packets = nil'); 
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
     if Terminated then
       begin 
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Terminated'); 
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
// Add packet
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
     if ClientSession.MsgPackets.Count < (Header.PacketID + 1) then // New Packet - Allocate
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Allocate');
aaWriteToLog('           MsgPackets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
{$ENDIF}
       ClientSession.MsgPackets.Count := (Header.PacketID + 1); // List fills hole by Nils
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           MsgPackets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add packet');
{$ENDIF}
     ClientSession.MsgPackets.Items[Header.PacketID] := Packet;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           MsgPackets.Count='+IntToStr(Integer(ClientSession.MsgPackets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': packet added');
{$ENDIF}
{
     if Terminated then
      begin
        Dispose(Packet);
        LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
      end;
}
// Are all packets received?
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.MsgPackets.Items[Header.PacketID])));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all packets received?');
{$ENDIF}
     if
//     (ClientSession.MsgPackets.Count > (Header.PacketID + 1)) or // repeated packet
     (Header.ControlCode >= ACRLastPacket)
     then // Last packet
      begin
       AllPacketsReceived := True;
       for i:=0 to ClientSession.MsgPackets.Count-1 do
        if ClientSession.MsgPackets.Items[i] = nil then
         begin
          AllPacketsReceived := False;
          break;
         end;
      end;
     if AllPacketsReceived then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
       ClientSession.MsgReceiveStatus := ACRFullAnswer  // Allow to extract buffer
      end
     else
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
       ClientSession.MsgReceiveStatus := ACRStartAnswer;
      end;
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Left!');
{$ENDIF}
// Is it the last packet?
     if Header.ControlCode >= ACRLastPacket then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       Header.ControlCode := Header.ControlCode - ACRLastPacket;
       if ClientSession.MsgReceiveStatus <> ACRFullAnswer then // Last packet received, but answer is not full
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Start resending thread');
{$ENDIF}
         if ClientSession.MsgResendRequestThread = nil then
           ClientSession.MsgResendRequestThread := TACRClientMsgResendRequestThread.Create(FManager, ClientSession, Buffer, ClientConnection.Network, FromHost, FromPort);
        end; // last packet - answer is not full
      end;
// Is Answer full?
       if ClientSession.MsgReceiveStatus = ACRFullAnswer then
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full answer');
{$ENDIF}
         if ClientSession.MsgResendRequestThread <> nil then
           ClientSession.MsgResendRequestThread.Terminate;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - start');
{$ENDIF}
         FManager.SendAcknowledgement(ClientSession, True);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - finish');
{$ENDIF}
         FManager.ReceiveMessage(ClientSession, Buf, BufSize);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DecompressAndDecryptBuffer');
{$ENDIF}
         FManager.DecompressAndDecryptBuffer(ClientSession.Session, Buf, BufSize);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT: Call ClientSession.ReceiveMessage');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
         TACRNetworkSession(ClientSession.Session).ReceiveMessage(Buf, BufSize);
        end; // answer is full
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End data packet process
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
}
     Exit;
    end; // message packet process
   ACRNewRequest,
   ACRConnected,
   ACRLastPacket..(ACRMessage+ACRLastPacket-1),
   (ACRMessage+ACRLastPacket+1)..127,
   ACRNoAction:
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Data packet');
{$ENDIF}
     if ClientSession.CurrentRequestID = Header.CurrentRequestID then
     if (ClientSession.AnswerStatus = ACRFullAnswer) then // lost acknowledgement
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Send ACKN');
{$ENDIF}
       FManager.SendAcknowledgement(ClientSession);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet while the answer received fully');
{$ENDIF}
       goto KillPacket; // Do not replace correct packets with doubles
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Enter...');
{$ENDIF}
      EnterCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Entered!');
{$ENDIF}
// Is this packet already stored?
      if ClientSession.Packets.List <> nil then // list exists
       if ((Header.PacketID+1) <= ClientSession.Packets.Count) then // old packet
       if (ClientSession.Packets.Items[Header.PacketID] <> nil) then // already received
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
{$ENDIF}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.Packets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT received existing packet');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
         goto KillPacket; // Do not replace correct packets with doubles
        end;
// Add correct received data packet
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New packet');
if ((Header.PacketID+1) <= ClientSession.Packets.Count) then // old packet
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.Packets.Items[Header.PacketID])));
{$ENDIF}
     if ClientSession.Packets = nil then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets = nil');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
     if Terminated then
       begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
      LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
       end;
// Add packet
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufferSize='+IntToStr(BufferSize));
{$ENDIF}
     if ClientSession.Packets.Count < (Header.PacketID + 1) then // New Packet - Allocate
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Allocate');
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
{$ENDIF}
       ClientSession.Packets.Count := (Header.PacketID + 1); // List fills hole by Nils
      end;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add packet');
{$ENDIF}
     ClientSession.Packets.Items[Header.PacketID] := Packet;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(ClientSession.Packets.Count)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': packet added');
{$ENDIF}
{
    if Terminated then
      begin
        Dispose(Packet);
        LeaveCriticalSection(FManager.FCSect);
        goto KillPacket;
      end;
}
// Are all packets received?
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog('           @Items='+IntToStr(Integer(ClientSession.Packets.Items[Header.PacketID])));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all packets received?');
{$ENDIF}
     if
//     (ClientSession.Packets.Count > (Header.PacketID + 1)) or // repeated packet
     (Header.ControlCode >= ACRLastPacket)
     then // Last packet
      begin
       AllPacketsReceived := True;
       for i:=0 to ClientSession.Packets.Count-1 do
        if ClientSession.Packets.Items[i] = nil then
         begin
          AllPacketsReceived := False;
          break;
         end;
      end;
     if AllPacketsReceived then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
       ClientSession.AnswerStatus := ACRFullAnswer  // Allow to extract buffer
      end
     else
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
       ClientSession.AnswerStatus := ACRStartAnswer;
      end;
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Left!');
{$ENDIF}
// Is it the last packet?
     if Header.ControlCode >= ACRLastPacket then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       Header.ControlCode := Header.ControlCode - ACRLastPacket;
       if ClientSession.AnswerStatus <> ACRFullAnswer then // Last packet received, but answer is not full
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Start resending thread');
{$ENDIF}
         if ClientSession.ResendRequestThread = nil then
           ClientSession.ResendRequestThread := TACRClientResendRequestThread.Create(FManager, ClientSession, Buffer, ClientConnection.Network, FromHost, FromPort);
        end; // last packet - answer is not full
      end;
// Is Answer full?
       if ClientSession.AnswerStatus = ACRFullAnswer then
        begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full answer');
{$ENDIF}
         if ClientSession.ResendRequestThread <> nil then
          ClientSession.ResendRequestThread.Terminate;
// To allow GetAnswer=Connected while Connect procedure
         if (Header.ControlCode = ACRConnected)
          then
           begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRConnected');
{$ENDIF}
            ClientSession.SendStatus := ACRSent;
            sleep(0); // TRY TO REMOVE!!!
{$IFDEF MsgCommunicator}
            if TACRClientSession(ClientSession.Session).Direct then
              ClientSession.Session.ConnectedUser(
                TACRClientSession(ClientSession.Session).RemoteUser.UserID,
                FromHost, FromPort                
                                                );
 {$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ConnectedUser!');
 {$ENDIF}
{$ENDIF}
           end
          else
           begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement');
{$ENDIF}
{
            if Terminated then
             begin
              Dispose(Packet);
//              LeaveCriticalSection(FManager.FCSect);
              goto KillPacket;
             end;
}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - start');
{$ENDIF}
            FManager.SendAcknowledgement(ClientSession);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendAcknowledgement - finish');
{$ENDIF}
           end;
         if (Header.ControlCode = ACREcho) then
          begin
           FManager.ReceiveBuffer(ClientSession.Session,Buf,BufSize);
           FManager.SendBuffer(ClientSession.Session,Buf,BufSize,ACRNoAction);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT SENT ECHO TO SERVER #'+IntToStr(ClientSession.Session.ConnectParams.ServerID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
           MemoryManager.FreeAndNilMem(Buf);
          end;
        end; // answer is full
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End data packet process
{
     EnterCriticalSection(FManager.FCSect);
     dec(ClientSession.Status);
     LeaveCriticalSection(FManager.FCSect);
}
     Exit;
    end // data packet process
   else
     raise EACRException.Create(40022, ErrorRUnknownControlCode, [Header.ControlCode]);
     goto KillPacket;
   end; // packet process
KillPacket:
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': KillPacket');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('CLIENT KILLED PACKET');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
  MemoryManager.FreeAndNilMem(Buffer);
{
  EnterCriticalSection(FManager.FCSect);
  dec(ClientSession.Status);
  LeaveCriticalSection(FManager.FCSect);
}
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientListenerThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
finally
//LeaveCriticalSection(FManager.FCSect);
end;
end;// Execute



////////////////////////////////////////////////////////////////////////////////
//
// TACRClientResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientResendRequestThread.Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                                            );
var
  Header:       PACRPacketHeader;
begin
  FManager := Manager;
  FManager.IncThreadCount;
  FClientSession := ClientSession;
  FBuffer := MemoryManager.AllocMem(SizeOf(TACRPacketHeader));
  Move(Buffer^,FBuffer^,SizeOf(TACRPacketHeader));
  Header := PACRPacketHeader(FBuffer);
  Header.Recepient := FClientSession.Session.ConnectParams.ServerID;
  FNetwork := Network;
  FHost := FromHost;
  FPort := FromPort;
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT ResendRequest THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientResendRequestThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT ResendRequest THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  MemoryManager.FreeAndNilMem(FBuffer);
  EnterCriticalSection(FManager.FCSect);
  if FClientSession.ResendRequestThread <> nil then
    FClientSession.ResendRequestThread := nil;
  LeaveCriticalSection(FManager.FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT ResendRequest THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRClientResendRequestThread.Execute;
var
  i, Count:           Integer;
  AllPacketsReceived: Boolean;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** CLIENT RESEND REQUEST START *****');
{$ENDIF}
  repeat
   AllPacketsReceived := True;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRClientResendRequestThread');
{$ENDIF}
   Count := FClientSession.Packets.Count;
   for i:=0 to Count-1 do
    begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF}
     if Terminated then
      begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF}
       Exit;
      end;
     Sleep(0);
     EnterCriticalSection(FManager.FCSect);
     if FClientSession.Packets <> nil then
     if FClientSession.Packets.Count > i then
     if FClientSession.Packets.Items[i] = nil then
      begin
       LeaveCriticalSection(FManager.FCSect);
       AllPacketsReceived := False;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketResendRequest, i='+IntToStr(i));
{$ENDIF}
       FManager.PacketResendRequest(FBuffer, FNetwork, FHost, FPort, i);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been requested');
{$ENDIF}
       Sleep(ACRRequestDelay);
       EnterCriticalSection(FManager.FCSect);
      end;
     LeaveCriticalSection(FManager.FCSect);
    end;
   if AllPacketsReceived then
    begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
     FClientSession.AnswerStatus := ACRFullAnswer;
     FManager.SendAcknowledgement(FClientSession);
    end;
  until FClientSession.AnswerStatus = ACRFullAnswer;
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientResendRequestThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
end;// Execute


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientMsgResendRequestThread.Create(
                       Manager:           TACRClientConnectionManager;
                       ClientSession:     PACRClntSession;
                       Buffer:            PChar;
                       Network:           TACRNetwork;
                       FromHost:          String;
                       FromPort:          Integer
                                            );
var
  Header:       PACRPacketHeader;
begin
  FManager := Manager;
  FManager.IncThreadCount;
  FClientSession := ClientSession;
  FBuffer := MemoryManager.AllocMem(SizeOf(TACRPacketHeader));
  Move(Buffer^,FBuffer^,SizeOf(TACRPacketHeader));
  Header := PACRPacketHeader(FBuffer);
  Header.Recepient := FClientSession.Session.ConnectParams.ServerID;
  FNetwork := Network;
  FHost := FromHost;
  FPort := FromPort;
  inherited Create(False);
  Priority := tpNormal; //tpLowest;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientMsgResendRequestThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  MemoryManager.FreeAndNilMem(FBuffer);
  EnterCriticalSection(FManager.FCSect);
  if FClientSession.MsgResendRequestThread <> nil then
    FClientSession.MsgResendRequestThread := nil;
  LeaveCriticalSection(FManager.FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_THREADS}
aaWriteToLog('CLIENT RESENDING THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRClientMsgResendRequestThread.Execute;
var
  i, Count:           Integer;
  AllPacketsReceived: Boolean;
begin
 try
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** CLIENT RESEND REQUEST START *****');
{$ENDIF}
  repeat
   AllPacketsReceived := True;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRClientMsgResendingThread');
{$ENDIF}
   Count := FClientSession.MsgPackets.Count;
   for i:=0 to Count-1 do
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF}
     if Terminated then
      begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Count='+IntToStr(Count));
{$ENDIF}
       Exit;
      end;
     Sleep(0);
     EnterCriticalSection(FManager.FCSect);
     if FClientSession.MsgPackets <> nil then
     if FClientSession.MsgPackets.Count > i then
     if FClientSession.MsgPackets.Items[i] = nil then
      begin
       LeaveCriticalSection(FManager.FCSect);
       AllPacketsReceived := False;
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketResendRequest, i='+IntToStr(i));
{$ENDIF}
       FManager.PacketResendRequest(FBuffer, FNetwork, FHost, FPort, i, True);
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been resent');
{$ENDIF}
       Sleep(ACRRequestDelay);
       EnterCriticalSection(FManager.FCSect);
      end;
     LeaveCriticalSection(FManager.FCSect);
    end;
   if AllPacketsReceived then
    begin
{$IFDEF LOG_CLIENT_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
     FClientSession.MsgReceiveStatus := ACRFullAnswer;
//     FManager.SendAcknowledgement(FClientSession, True);
    end;
  until FClientSession.MsgReceiveStatus = ACRFullAnswer;
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRClientMsgResendingThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
end;// Execute

{$ENDIF} // Client



{$IFDEF SERVER_VERSION}

////////////////////////////////////////////////////////////////////////////////
//
// TACRServerConnectionManager
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerConnectionManager.Create(Server: TComponent);
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Create> @Server='+IntToStr(Integer(FServer)));
{$ENDIF}
  InitializeCriticalSection(FCSect);
  inherited Create;
  FSessions := TThreadList.Create;
  FSessionID := -MAXINT;
  FServer := Server;
  FNetwork := TACRNetwork.Create(self);
  ThreadCount := 1;
  if Server = nil then
   begin
    FNetwork.FLocalClient := 1;
    FNetwork.LocalPort := ACRDefaultServerPort;
   end
  else
   begin
    FNetwork.FLocalClient := TACRServer(FServer).ServerID;
    FNetwork.LocalHost := TACRServer(FServer).LocalHost;
    FNetwork.LocalPort := TACRServer(FServer).LocalPort;
   end;
  FListenerStoped := False;
  ResendRequestThread := TACRServerResendRequestThread.Create(self);
  MsgResendRequestThread := TACRServerMsgResendRequestThread.Create(self);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Create> FINISHED');
{$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerConnectionManager.Destroy;
var
  Error:        String;
  StartTime:    DWord;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Destroy> @Server='+IntToStr(Integer(FServer)));
{$ENDIF}
  FListenerStoped := True;
  DisconnectAll(False);
// Terminate ResendRequestThread
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Destroy> Terminate ResendRequestThread');
{$ENDIF}
  if ResendRequestThread <> nil then
    ResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   sleep(0);
   if (ResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
       begin
        Error := 'command resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
                                          ['ServerConnectionManager',
                                                                  Error]);
       end;
     end;
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Destroy> command resend thread terminated');
{$ENDIF}
// Terminate MsgResendRequestThread
  if (MsgResendRequestThread <> nil) then
    MsgResendRequestThread.Terminate;
  StartTime := GetTickCount;
  repeat
   sleep(0);
   if (MsgResendRequestThread=nil)
    then break
    else
     begin
      if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
       begin
        Error := 'message resend requesting thread';
        raise EACRException.Create(40036, ErrorRThreadHangs,
                                          ['ServerConnectionManager',
                                                                  Error]);
       end;
     end;
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Destroy> message resend thread terminated');
{$ENDIF}
  sleep(0);
  EnterCriticalSection(FCSect);
  FSessions.Free;
  FSessions := nil;
  if FNetwork <> nil then
   begin
    FNetwork.Free;
    DecThreadCount;
   end;
  FNetwork := nil;
  LeaveCriticalSection(FCSect);
{ TODO -oAlex : Write ThreadCount to ErrorLog }
  DeleteCriticalSection(FCSect);
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Destroy - FINISHED');
{$ENDIF}
end;// Destoy


//------------------------------------------------------------------------------
// PacketResendRequest
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.PacketResendRequest(
                               Buffer:        PChar;
                               Network:       TACRNetwork;
                               RemoteHost:    String;
                               RemotePort:    Integer;
                               PacketID:      Integer = -1;
                               Msg:           Boolean = False
                                 );
var
  Header:          PACRPacketHeader;
begin
  Header := Pointer(Buffer);
  Header.Signature := ACRServerPacketSign;
  if Msg then
    Header.ControlCode := ACRMessagePacketResendRequest
  else
    Header.ControlCode := ACRPacketResendRequest;
  if (PacketID >= 0) then
    Header.PacketID := PacketID
  else
    Header.Recepient := Header.Sender;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('SERVER asks to resend packet # '+IntToStr(Header.PacketID));
{$ENDIF}
  EnterCriticalSection(FCSect);
  try
   Network.RemoteHost := RemoteHost;
   Network.RemotePort := RemotePort;
   Network.SendBuffer(Buffer, SizeOf(TACRPacketHeader));
  finally
   LeaveCriticalSection(FCSect);
  end;
end;// PacketResendRequest


//------------------------------------------------------------------------------
// OnDisconnect
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.OnDisconnect(
                               FNetwork:      TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
{$IFDEF MsgCommunicator}
  SSessions:            array of TACRServerSession;
{$ENDIF}
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.OnDisconnect> START');
aaWriteToLog('OnDisconnect> Broken client: '+FromHost+':'+IntToStr(FromPort));
{$ENDIF}
 try
 {$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> FSessions');
aaWriteToLog(IntToStr(Integer(FSessions)));
 {$ENDIF DEBUG_LOG_NETWORK_ERRORS}
{$IFDEF MsgCommunicator}
  SetLength(SSessions, 0);
{$ENDIF}
  Sessions:=FSessions.LockList;
 {$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> FSessions Locked!');
 {$ENDIF DEBUG_LOG_NETWORK_ERRORS}
   try
    for i:=Sessions.Count-1 downto 0 do
     begin
      if (i<0)
      or (i>=Sessions.Count)
      then
        break;
      ServerSession := Sessions.Items[i];
      if (ServerSession.RemoteHost = FromHost)
      and (ServerSession.RemotePort = FromPort)
      then
       begin
{$IFDEF MsgCommunicator}
        SetLength(SSessions, Length(SSessions)+1);
        SSessions[Length(SSessions)-1] := TACRServerSession(ServerSession.Session);
{$ENDIF}
        EnterCriticalSection(FCSect);
        if (ServerSession.ControlCode <> ACRTerminate)
        or (ServerSession.MsgControlCode <> ACRTerminate)
        then
         begin
          // Terminate sending threads
          ServerSession.ControlCode := ACRTerminate;
          ServerSession.MsgControlCode := ACRTerminate;
          LeaveCriticalSection(FCSect);
          // Block send message
{$IFDEF MsgCommunicator}
          EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
          FSessions.UnlockList;
          // Terminate listening threads
          TerminateAllSessionThreads(ServerSession);
(*
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> Terminate command listening threads...');
{$ENDIF}
          TerminateCommandThreads(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> Terminate message listening threads...');
{$ENDIF}
          TerminateMessageThreads(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> All sessions terminated');
{$ENDIF}
*)
          // delete session from list
          FSessions.Remove(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> Removed SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
{$IFDEF MsgCommunicator}
          LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> Session CS left');
{$ENDIF}
          // free memory and session
          DeleteSession(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> Session deleted');
{$ENDIF}
          Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> Sessions locked');
{$ENDIF}
         end
        else
          LeaveCriticalSection(FCSect);
       end;
     end;
   finally
    FSessions.UnlockList;
   end;

(*
  EnterCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> OLD: FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
{$ENDIF}
  if FNetwork <> nil then
   begin
    FNetwork.Free;
    DecThreadCount;
   end;
  FNetwork := nil;
  FNetwork := TACRNetwork.Create(self);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> NEW: FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
{$ENDIF}
  if FServer = nil then
   begin
    FNetwork.FLocalClient := 1;
    FNetwork.LocalPort := ACRDefaultServerPort;
   end
  else
   begin
    FNetwork.FLocalClient := TACRServer(FServer).ServerID;
    FNetwork.LocalHost := TACRServer(FServer).LocalHost;
    FNetwork.LocalPort := TACRServer(FServer).LocalPort;
   end;
  LeaveCriticalSection(FCSect);
*)

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog(IntToStr(GetTickCount)+':');
aaWriteToLog('OnDisconnect> FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
{$ENDIF}

{$IFDEF MsgCommunicator}
  if not FListenerStoped then
   // Send off-line events
   for i:=0 to Length(SSessions)-1 do
    begin
 {$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> DisconnectUser...');
 {$ENDIF}
     SSessions[i].DisconnectUser;
{
     ServerSession := Sessions.Items[0];
     TACRServerSession(ServerSession.Session).DisconnectUser(Users.Items[i]);
}
 {$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> DisconnectUser finished');
 {$ENDIF}
    end;
{$ENDIF}

{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('OnDisconnect> FINISH');
{$ENDIF DEBUG_LOG_NETWORK_ERRORS}
 except
  on E: Exception do
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerConnectionManager.OnDisconnect - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
    raise;
   end;
 end;
end; // OnDisconnect


//------------------------------------------------------------------------------
// NetworkListener
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.NetworkListener(
                               Buffer:        PChar;
                               BufferSize:    Integer;
                               Network:       TACRNetwork;
                               FromHost:      String;
                               FromPort:      Integer
                              );
begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.NetworkListener');
{$ENDIF}
  if FListenerStoped then
    MemoryManager.FreeAndNilMem(Buffer)
  else
   begin
    repeat
     if ThreadCount<ACRMaxThreadCount then
      break;
     sleep(0);
    until False;
    TACRServerListenerThread.Create(self, Buffer, BufferSize, FromHost, FromPort);
   end;
end;// NetworkListener


//------------------------------------------------------------------------------
// CommandReceived
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.CommandReceived(
                          ServerSession:  PACRSrvrSession;
                          ControlCode:    TACRControlCode
                                                      );
begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> STARTED');
{$ENDIF}
  if (ControlCode >= ACRLastPacket) then
    ControlCode := ControlCode - ACRLastPacket;
  EnterCriticalSection(FCSect);
  if ServerSession.ControlCode = ACRExecute then
    ServerSession.ControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);
(*
  if (ControlCode <> ACRConnect) then
   begin
    SendAcknowledgement(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: EXECUTE SESSION with ControlCode='+IntToStr(ControlCode));
aaWriteToLog('==============================================================');
{$ENDIF}
    if (ServerSession.Thread = nil)
    or (ControlCode = ACRDisconnect)
    then
      ServerSession.Thread := TACRServerSessionThread.Create(self,
                                                  ServerSession, ControlCode);
   end;
*)
  if (ControlCode <> ACRConnect) then
   begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> SendAcknowledgement');
{$ENDIF}
    SendAcknowledgement(ServerSession);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> SendAcknowledgement - SENT');
{$ENDIF}
    if (ControlCode = ACRDisconnect) then
     begin
      TACRServerSessionThread.Create(self, ServerSession, ControlCode);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> Disconnect thread CREATED');
{$ENDIF}
     end
    else
     begin
      // Wait for thread
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> Wait for thread');
{$ENDIF}
      repeat
       if ServerSession.Thread = nil then
         break;
       sleep(0);
      until False;
      if ServerSession.Thread = nil then
       begin
        ServerSession.Thread := TACRServerSessionThread.Create(self,
                                                 ServerSession, ControlCode);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> CREATED # '+IntToStr(Integer(ServerSession.Thread.ThreadID))+'/'+IntToStr(Integer(ServerSession.Thread.Handle)));
{$ENDIF}
       end
      else
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> Server Session already exists');
{$ENDIF}
       end;
     end;
  end;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.CommandReceived> FINISHED');
{$ENDIF}
end; // CommandReceived


//------------------------------------------------------------------------------
// MessageReceived
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.MessageReceived(
                                  ServerSession:  PACRSrvrSession
                                                      );
begin
  SendAcknowledgement(ServerSession, True);
  EnterCriticalSection(FCSect);
  if ServerSession.MsgControlCode = ACRExecute then
    ServerSession.MsgControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);
  // Wait for message thread
  repeat
    if ServerSession.MsgThread = nil then
      break;
    sleep(0);
  until False;
  // create new thread
  ServerSession.MsgThread := TACRServerSessionThread.Create(self,
                                                  ServerSession, ACRMessage);
end; // MessageReceived


//------------------------------------------------------------------------------
// ReceiveMessage
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.ReceiveMessage(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
var
  i:                    Integer;
  Packets:              TList;
  Packet:               PACRPacket;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER are receiving message from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
{$ENDIF}
  if ServerSession.MsgReceiveStatus<>ACRFullAnswer then
   begin
    while (ServerSession.MsgReceiveStatus<>ACRStartAnswer)
    and (ServerSession.MsgReceiveStatus<>ACRFullAnswer)
    do // Wait for starting answer receive
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRServerReceiveSleep);
     end;
    StartTime := GetTickCount;
    while (ServerSession.MsgReceiveStatus<>ACRFullAnswer)
    do // Wait for all packets to arrive
     begin
      if (GetTickCount - StartTime) > ACRServerReceiveTimeOut then
       raise EACRException.Create(40026, ErrorRTimeoutFullReceive,
                                [ServerSession.Session.SessionID, ACRServerReceiveTimeOut]);
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRServerReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received request from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
  EnterCriticalSection(FCSect);
  if ServerSession.MsgControlCode = ACRExecute then
    ServerSession.MsgControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);
  Packets := ServerSession.MsgPackets.LockList;
  try
   BufferSize := 0;
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
    end;
   Buffer := MemoryManager.GetMem(BufferSize);
   pBuf := Buffer;
  // Copy to buffer then free MsgPackets
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
        begin
         Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
         inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
         MemoryManager.FreeAndNilMem(Packet.Buffer);
        end;
       Dispose(Packet);
       Packets.Items[i] := nil;
      end;
    end;
   Packets.Count := 0;
  finally
   ServerSession.MsgPackets.UnlockList;
  end;
  inc(ServerSession.ClientMessageID);
  EnterCriticalSection(FCSect);
  if ServerSession.MsgControlCode = ACRSuspend then
    ServerSession.MsgControlCode := ACRExecute;
  LeaveCriticalSection(FCSect);
  ServerSession.MsgReceiveStatus := ACRNoAnswer;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER received full message from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('==============================================================');
{$ENDIF}
end; // ReceiveMessage


//------------------------------------------------------------------------------
// ReceiveBuffer
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.ReceiveBuffer(
                          ServerSession:  PACRSrvrSession;
                          var Buffer:     PChar;
                          var BufferSize: Integer
                          );
var
  i:                    Integer;
  Packet:               PACRPacket;
  Packets:              TList;
  pBuf:                 PChar;
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.ReceiveBuffer - START');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER are receiving request from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
{$ENDIF}
  if ServerSession.AnswerStatus <> ACRFullAnswer then
   begin
    while (ServerSession.AnswerStatus <> ACRStartAnswer)
    and (ServerSession.AnswerStatus <> ACRFullAnswer)
    do // Wait for starting answer receive
     begin
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRServerReceiveSleep);
     end;
    StartTime := GetTickCount;
    while (ServerSession.AnswerStatus <> ACRFullAnswer)
    do // Wait for all packets to arrive
     begin
      if (GetTickCount - StartTime) > ACRServerReceiveTimeOut then
       raise EACRException.Create(40025, ErrorRTimeoutFullReceive,
                                [ServerSession.Session.SessionID, ACRServerReceiveTimeOut]);
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
      Sleep(ACRServerReceiveSleep);
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received request from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.ReceiveBuffer> All packets received');
{$ENDIF}
  EnterCriticalSection(FCSect);
  if ServerSession.ControlCode = ACRExecute then
    ServerSession.ControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);
  BufferSize := 0;
  Packets := ServerSession.Packets.LockList;
  try
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     BufferSize := BufferSize + Packet.BufferSize - SizeOf(TACRPacketHeader);
    end;
   if BufferSize > 0 then
     Buffer := MemoryManager.GetMem(BufferSize)
   else
     Buffer := nil;
   pBuf := Buffer;
  // Copy to buffer then free MsgPackets
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
        begin
         if (Packet.BufferSize > SizeOf(TACRPacketHeader)) then
          begin
           Move(Pointer(Packet.Buffer+SizeOf(TACRPacketHeader))^, pBuf^, Packet.BufferSize-SizeOf(TACRPacketHeader));
           inc(pBuf, Packet.BufferSize-SizeOf(TACRPacketHeader));
          end;
         MemoryManager.FreeAndNilMem(Packet.Buffer);
        end;
       Dispose(Packet);
       Packets.Items[i] := nil;
      end;
    end;
   Packets.Count := 0;
  finally
   ServerSession.Packets.UnlockList;
  end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER received full buffer from CLIENT #'+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('==============================================================');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.ReceiveBuffer - FINISH');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('SERVER<<< '+IntToStr(ServerSession.CurrentRequestID)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}
end; // ReceiveBuffer


//------------------------------------------------------------------------------
// SendConnectAckn
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendConnectAckn(
                          ServerSession:        PACRSrvrSession;
                          CurrentRequestID:     Integer = -1
                                                      );
var
  Buffer,
  Buf:                            PChar;
  BufSize, BufferSize,
  SizeSID, SizeParams:            Integer;
  ConnectionParams:               TACRConnectionParams;
  RequestID:                      Integer;
begin
{
  ConnectionParams.LocalPort := ServerSession.RemotePort;
  ConnectionParams.RemoteHost := FNetwork.LocalHost;
  ConnectionParams.RemotePort := FNetwork.LocalPort;
  ConnectionParams.ServerID := FNetwork.FLocalClient;
}
  ConnectionParams.PacketSize := ServerSession.Session.ConnectParams.PacketSize;
  ConnectionParams.CompressionAlgorithm := ServerSession.Session.ConnectParams.CompressionAlgorithm;
  ConnectionParams.CompressionMode := ServerSession.Session.ConnectParams.CompressionMode;
  SizeParams := SizeOf(ConnectionParams);
  SizeSID := SizeOf(ServerSession.Session.SessionID);
  BufferSize := SizeParams + SizeSID;
  Buffer := MemoryManager.GetMem(BufferSize);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER: SessionID='+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('SizeSID='+IntToStr(SizeSID)+', SizeParams='+IntToStr(SizeParams)+', BufferSize='+IntToStr(BufferSize));
{$ENDIF}
  try
   Move(ConnectionParams, Buffer^, SizeParams);
   Move(ServerSession.Session.SessionID, (Buffer+SizeParams)^, SizeSID);

   if CurrentRequestID >= 0 then
    begin
     RequestID := ServerSession.CurrentRequestID;
     ServerSession.CurrentRequestID := CurrentRequestID;
    end;

   CompressAndEncryptBuffer(ServerSession.Session, Buffer, BufferSize, Buf, BufSize);
   try
    DoSendBuffer(ServerSession, Buf, BufSize, ACRConnected);
   finally
    if Buf<>Buffer then
      MemoryManager.FreeAndNilMem(Buf);
   end;

   if CurrentRequestID >= 0 then
     ServerSession.CurrentRequestID := RequestID;

   ServerSession.AnswerStatus := ACRNoAnswer;
  finally
   MemoryManager.FreeAndNilMem(Buffer);
  end;
end;// SendConnectAckn


//------------------------------------------------------------------------------
// SendDisconnectRequest
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendDisconnectRequest(
                                    ServerSession: PACRSrvrSession;
                                    WaitForAnswer: Boolean = True);
var
  Header:           PACRPacketHeader;
  Retry:            Integer;
  StartTime:        DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.SendDisconnectRequest> START');
{$ENDIF}
 try
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   Header.ControlCode := ACRDisconnect;
   Header.Signature := ACRServerPacketSign;
   Header.Recepient := ServerSession.RemoteClientID;
   EnterCriticalSection(FCSect);
   Header.Sender := FNetwork.FLocalClient;
   LeaveCriticalSection(FCSect);
   Header.ConnectionID := ServerSession.ConnectionID;
   Header.SessionID := ServerSession.Session.SessionID;
   Header.PacketID := 0;
   Header.CurrentRequestID := ServerSession.CurrentRequestID;
   Retry := 0;
   ServerSession.SendStatus := ACRNotSent;
   repeat
     EnterCriticalSection(FCSect);
     try
      FNetwork.RemoteHost := ServerSession.RemoteHost;
      FNetwork.RemotePort := ServerSession.RemotePort;
{
      if ServerSession.ControlCode = ACRTerminate then
        raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
}
      FNetwork.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
     finally
      LeaveCriticalSection(FCSect);
     end;
     if not WaitForAnswer then
       Exit;
     StartTime := GetTickCount;
     while ((GetTickCount - StartTime) < ACRDisconnectDelay) do // pause
      if ServerSession.SendStatus <> ACRSent then
       begin
        Sleep(ACRServerReceiveSleep);
{$IFDEF ProcessMessages}
        Application.ProcessMessages;
{$ENDIF ProcessMessages}
       end
      else
       Exit;
     inc(Retry);
   until  (Retry > ACRDisconnectRetryCount);
   if ServerSession.SendStatus <> ACRSent then
    raise EACRException.Create(40024, ErrorRCannotDisconnect,
                   ['client', ServerSession.ClientSessionID,
                    ACRDisconnectRetryCount, ACRDisconnectDelay]);
  finally
   MemoryManager.FreeAndNilMem(Header);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.SendDisconnectRequest> FINISH');
{$ENDIF}
  end;
 except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerSessionThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
     raise;
    end;
{$ELSE}
  raise;
{$ENDIF}
 end;
end;// SendDisconnectRequest


//------------------------------------------------------------------------------
// DeleteSession
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DeleteSession(ServerSession: PACRSrvrSession);
var
  i:                    Integer;
  Packet:               PACRPacket;
  Packets:              TList;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> START');
{$ENDIF}
{$IFDEF MsgCommunicator}
  EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
 // free listenning threads lists
  ServerSession.ListeningThreads.Free;
  ServerSession.MsgListeningThreads.Free;
  // block using ServerSession
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRTerminate;
  ServerSession.MsgControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
  // Wait for the end of server session thread if it is existing
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> WaitFor command Session thread...');
{$ENDIF}
  repeat
    if ServerSession.Thread = nil then
      break;
    sleep(0);
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free Session...');
{$ENDIF}
  // free PacketIDsToResend
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> wait for resending broken packets...');
{$ENDIF}
  repeat
    if ServerSession.PacketIDsToResend.ItemCount = 0 then // wait for resending broken packets
      break;
    sleep(0);
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free PacketIDsToResend...');
{$ENDIF}
  ServerSession.PacketIDsToResend.Free;
  // free packets
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free packets...');
{$ENDIF}
  Packets := ServerSession.Packets.LockList;
  try
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
         MemoryManager.FreeAndNilMem(Packet.Buffer);
       Dispose(Packet);
      end;
    end;
  finally
   ServerSession.Packets.UnlockList;
  end;
  // free packets list
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free packet list...');
{$ENDIF}
  ServerSession.Packets.Free;
  // Wait for message thread
  repeat
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> SessionID='+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.DeleteSession> WaitFor message Session thread...');
aaWriteToLog('TACRServerConnectionManager.DeleteSession> MsgThread='+IntToStr(Integer(ServerSession.MsgThread)));
{$ENDIF}
    if ServerSession.MsgThread = nil then
      break;
    sleep(0);
  until False;
  // free MsgPacketIDsToResend
{$IFDEF MsgCommunicator}
  EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> wait for resending broken message packets...');
{$ENDIF}
  repeat
    if ServerSession.MsgPacketIDsToResend.ItemCount = 0 then // wait for resending broken packets
      break;
    sleep(0);
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free MsgPacketIDsToResend...');
{$ENDIF}
  ServerSession.MsgPacketIDsToResend.Free;
{$IFDEF MsgCommunicator}
  LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
  // free message packets
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free message packets...');
{$ENDIF}
  Packets := ServerSession.MsgPackets.LockList;
  try
   for i := 0 to Packets.Count - 1 do
    begin
     Packet := Packets.Items[i];
     if Packet <> nil then
      begin
       if Packet.Buffer <> nil then
         MemoryManager.FreeAndNilMem(Packet.Buffer);
       Dispose(Packet);
      end;
    end;
  finally
   ServerSession.MsgPackets.UnlockList;
  end;
  // free message packets list
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> free message packet list...');
{$ENDIF}
  ServerSession.MsgPackets.Free;
  // wait for vacant ServerSession
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> wait for vacant ServerSession...');
{$ENDIF}
  repeat
    // block using ServerSession
    EnterCriticalSection(FCSect);
    ServerSession.ControlCode := ACRTerminate;
    ServerSession.MsgControlCode := ACRTerminate;
    LeaveCriticalSection(FCSect);
    sleep(0);
    if (ServerSession.Status = ACRVacant) then
      break;
  until False;
{$IFDEF MsgCommunicator}
  LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
  // free session
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
  ServerSession.Session.Free;
  ServerSession.Session := nil;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> Dispose(ServerSession)...');
{$ENDIF}
  Dispose(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DeleteSession> FINISH');
{$ENDIF}
end; // DeleteSession


//------------------------------------------------------------------------------
// DoDisconnect
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DoDisconnect(Session: TACRBaseSession);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> START');
{$ENDIF}
  Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> FSessions Locked');
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Sessions.Count='+IntToStr(Sessions.Count));
{$ENDIF}
  try
   for i:=Sessions.Count-1 downto 0 do
    begin
     if (i<0)
     or (i>=Sessions.Count)
     then
       break;
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
      begin
       EnterCriticalSection(FCSect);
       if (ServerSession.ControlCode <> ACRTerminate)
       or (ServerSession.MsgControlCode <> ACRTerminate)
       then
        begin
         // Terminate sending threads
         ServerSession.ControlCode := ACRTerminate;
         ServerSession.MsgControlCode := ACRTerminate;
         LeaveCriticalSection(FCSect);
         // block sendmessage
{$IFDEF MsgCommunicator}
         EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
         FSessions.UnlockList;
         // Terminate listening threads
         TerminateCommandThreads(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('DoDisconnect> Terminate message listening threads...');
{$ENDIF}
         TerminateMessageThreads(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('DoDisconnect> All sessions terminated');
{$ENDIF}
         // delete session from list
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRClientConnectionManager.DoDisconnect> Remove SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
         FSessions.Remove(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('DoDisconnect> Removed SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
{$IFDEF MsgCommunicator}
         LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('DoDisconnect> Session CS left');
{$ENDIF}
         // free memory and session
         DeleteSession(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('DoDisconnect> Session deleted');
{$ENDIF}
         Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('DoDisconnect> Sessions locked');
{$ENDIF}
        end
       else
         LeaveCriticalSection(FCSect);
(*
       // Terminate all threads
       FSessions.UnlockList;
       TerminateAllSessionThreads(ServerSession);
       // delete session from list
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Session remove session from list...');
{$ENDIF}
       FSessions.Remove(ServerSession);
       // free memory and session
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Delete SessionID='+IntToStr(ServerSession.Session.SessionID)+'...');
{$ENDIF}
       DeleteSession(ServerSession);
       Sessions:=FSessions.LockList;
*)
       break;
      end;
    end;
  finally
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoDisconnect> Sessions.Count='+IntToStr(Sessions.Count));
{$ENDIF}
   FSessions.UnlockList;
  end;
end; // DoDisconnect


//------------------------------------------------------------------------------
// Disconnect
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.Disconnect(Session: TACRBaseSession);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
  Found:                Boolean;
begin
  Found := False;
  Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.Disconnect> FSessions Locked');
{$ENDIF}
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
      begin
       Found := True;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not Found then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);

  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRSuspend;
  ServerSession.MsgControlCode := ACRSuspend;
  LeaveCriticalSection(FCSect);

  SendDisconnectRequest(ServerSession);

{
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRTerminate;
  ServerSession.MsgControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
}

  DoDisconnect(Session);
end; // Disconnect


//------------------------------------------------------------------------------
// DisconnectAll
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DisconnectAll(WaitForAllDisconected: Boolean = True);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> START');
{$ENDIF}
if FSessions<>nil then
begin
try
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> Lock FSessions...');
{$ENDIF}
 Sessions:=FSessions.LockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> FSessions Locked');
{$ENDIF}
 try
//  while Sessions.Count>=1 do
  for i:=Sessions.Count-1 downto 0 do
   begin
    if (i<0)
    or (i>=Sessions.Count)
    then
      break;
    ServerSession := Sessions.Items[i];
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> EnterCriticalSection...');
{$ENDIF}
    EnterCriticalSection(FCSect);
    if (ServerSession.ControlCode <> ACRTerminate)
    or (ServerSession.MsgControlCode <> ACRTerminate)
    then
     begin
      // Terminate sending threads
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> Terminate sending threads...');
{$ENDIF}
      ServerSession.ControlCode := ACRTerminate;
      ServerSession.MsgControlCode := ACRTerminate;
      LeaveCriticalSection(FCSect);
      // Block sending message
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> Block sending message...');
{$ENDIF}
{$IFDEF MsgCommunicator}
      EnterCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
      FSessions.UnlockList;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> SendDisconnectRequest');
{$ENDIF}
      // send
      try
       SendDisconnectRequest(ServerSession, False);
      except
       // ignore all exceptions while disconnecting 
      end;
      // Terminate all threads
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> TerminateAllSessionThreads...');
{$ENDIF}
      TerminateAllSessionThreads(ServerSession);
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> All sessions terminated');
{$ENDIF}
      // delete session from list
      FSessions.Remove(ServerSession);
      // free memory and session
{$IFDEF MsgCommunicator}
      LeaveCriticalSection(TACRServerSession(ServerSession.Session).FCSect);
{$ENDIF}
      DeleteSession(ServerSession);
      Sessions:=FSessions.LockList;
     end
    else
      LeaveCriticalSection(FCSect);
   end;
//  Sessions.Capacity :=0;
  // Wait for all sessions be disconnected
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> WaitForAllDisconected...');
{$ENDIF}
  if WaitForAllDisconected then
    repeat
      Sessions:=FSessions.LockList;
      try
       if Sessions.Count = 0 then
         break;
      finally
       FSessions.UnlockList;
      end;
      sleep(0);
    until False;
 finally
  FSessions.UnlockList;
 end;
except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
    end;
{$ENDIF}
end;
end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DisconnectAll> FINISH');
{$ENDIF}
end; // DisconnectAll


//------------------------------------------------------------------------------
// SessionsCount
//------------------------------------------------------------------------------
function TACRServerConnectionManager.SessionsCount: Integer;
var
  Sessions:           TList;
begin
  Sessions := FSessions.LockList;
  try
    Result := Sessions.Count;
  finally
    FSessions.UnlockList;
  end;
end;// SessionsCount


//------------------------------------------------------------------------------
// TerminateMessageThreads
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateMessageThreads(ServerSession: PACRSrvrSession);
var
  i:                    Integer;
  Count:                Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
  Error:                String;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> START');
{$ENDIF}
  EnterCriticalSection(FCSect);
  ServerSession.MsgControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> MsgListeningThreads.LockList...');
{$ENDIF}
   // Terminate message listening threads
   Threads := ServerSession.MsgListeningThreads.LockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> MsgListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
   try
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ServerSession.MsgListeningThreads.UnlockList;
   end;
   sleep(0);
   // Check
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> CHECKING...');
{$ENDIF}
   Threads := ServerSession.MsgListeningThreads.LockList;
   try
    Count := Threads.Count;
   finally
    ServerSession.MsgListeningThreads.UnlockList;
   end;
   if (Count = 0) then break
   else
    begin
     if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
      begin
       Error := IntToStr(Count)+' message listening threads';
       raise EACRException.Create(40036, ErrorRThreadHangs,
            ['Server session # '+IntToStr(ServerSession.Session.SessionID),
            Error]);
      end;
    end;
  until False;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.TerminateMessageThreads> FINISH');
{$ENDIF}
end; // TerminateMessageThreads


//------------------------------------------------------------------------------
// TerminateCommandThreads
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateCommandThreads(ServerSession: PACRSrvrSession);
var
  i:                    Integer;
  Count:                Integer;
  StartTime:            DWORD;
  Threads:              TList;
  Thread:               TThread;
  Error:                String;
begin
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ACRTerminate;
  LeaveCriticalSection(FCSect);
  // Wait for finishes of all the threads
  StartTime := GetTickCount;
  repeat
   // Terminate all command listening threads
   Threads := ServerSession.ListeningThreads.LockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
{$ENDIF}
   try
    for i := 0 to Threads.Count-1 do
     begin
      Thread := Threads.Items[i];
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> Terminate # '+IntToStr(Integer(Thread.ThreadID))+'/'+IntToStr(Integer(Thread.Handle)));
{$ENDIF}
      if (Thread <> nil) then
        Thread.Terminate;
      sleep(0);
     end;
   finally
    ServerSession.ListeningThreads.UnlockList;
   end;
   sleep(0);
   // Check
   Threads := ServerSession.ListeningThreads.LockList;
   try
    Count := Threads.Count;
   finally
    ServerSession.ListeningThreads.UnlockList;
   end;
   if (Count = 0) then
     break
   else
    begin
     if (GetTickCount - StartTime) > ACRThreadsTerminateDelay then
      begin
       Error := IntToStr(Count)+' command listening threads';
       raise EACRException.Create(40036, ErrorRThreadHangs,
             ['Server session # '+IntToStr(ServerSession.Session.SessionID),
              Error]);
      end;
    end;
  until False;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> ListeningThreads.Count='+IntToStr(Threads.Count));
aaWriteToLog('TACRServerConnectionManager.TerminateCommandThreads> FINISH');
{$ENDIF}
end; // TerminateCommandThreads


//------------------------------------------------------------------------------
// TerminateAllSessionThreads
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.TerminateAllSessionThreads(ServerSession: PACRSrvrSession);
var
  StartTime:            DWORD;
  Error:                String;
begin
  TerminateCommandThreads(ServerSession);
  TerminateMessageThreads(ServerSession);
end; // TerminateAllSessionThreads


//------------------------------------------------------------------------------
// DoSendBuffer
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.DoSendBuffer(
                          ServerSession:    PACRSrvrSession;
                          Buffer:           PChar;
                          BufferSize:       Integer;
                          Code:             Integer = ACRNoAction
                                                  );
var
  Header:               PACRPacketHeader;
  Packets:              TList;
  Packet:               PACRPacket;
  BytesSent, DataSize:  Integer;
  PacketID:             Integer;
  StartTime, Delay:     DWORD;
  StartSleepTime:       DWORD;
  i:                    Integer;

procedure WaitForCommandSent;
var
  PacketID,
  ControlCode,
  i:                    Integer;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
  j:                    Integer;
{$ENDIF}
  StartTime:            DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForCommandSent> - START');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  ControlCode := ServerSession.ControlCode;
  ServerSession.ControlCode := ACRExecute;
  LeaveCriticalSection(FCSect);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server waits for all packets');
{$ENDIF}
  if not ((Code = ACRConnect) or (Code = (ACRConnect + ACRLastPacket))) then
   begin
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRServerReceiveSleep); // To avoid delay if full answer is already received
    i := ServerSession.PacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ServerSession.SendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server has '+IntToStr(ServerSession.PacketIDsToResend.ItemCount)+' packets to resend:');
for j:=0 to ServerSession.PacketIDsToResend.ItemCount-1 do
aaWriteToLog(IntToStr(ServerSession.PacketIDsToResend.Items[j]));
{$ENDIF}
      if (GetTickCount - StartTime) > ACRServerSendTimeOut then
        raise EACRException.Create(40078, ErrorRServerTimeOutSending,
          ['message',
{$IFDEF MsgCommunicator}
          ServerSession.Session.UserID,
{$ELSE}
          ServerSession.Session.SessionID,
{$ENDIF}
          ACRServerSendTimeOut]);
      dec(i);
      if (i<0) then
        i := ServerSession.PacketIDsToResend.ItemCount - 1;
      try
       ServerSession.PacketIDsToResend.Lock;
       if (ServerSession.PacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ServerSession.PacketIDsToResend.Items[i];
           ServerSession.PacketIDsToResend.Unlock;
           if PacketID >= Packets.Count then // Erroneous situation which is must not exist
             Continue;
          end
         else // No packets to resend - Resend last packet
          begin
           ServerSession.PacketIDsToResend.Unlock;
           PacketID := Packets.Count - 1;
          end;
         Packet := Packets.Items[PacketID];
         // resend packet
         EnterCriticalSection(FCSect);
         try
          FNetwork.RemoteHost := ServerSession.RemoteHost;
          FNetwork.RemotePort := ServerSession.RemotePort;
          if ServerSession.ControlCode = ACRTerminate then
            raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
          FNetwork.SendBuffer(Packet.Buffer, Packet.BufferSize);
         finally
          LeaveCriticalSection(FCSect);
         end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server resent packet # '+IntToStr(PacketID));
{$ENDIF}
        end
      else
        ServerSession.PacketIDsToResend.Unlock;
      finally
       ServerSession.PacketIDsToResend.Lock;
       if  (ServerSession.PacketIDsToResend.ItemCount > 0)
       and (ServerSession.PacketIDsToResend.ItemCount > i)
       then
//         ServerSession.PacketIDsToResend.Delete(i);
         ServerSession.PacketIDsToResend.Remove(PacketID);
       ServerSession.PacketIDsToResend.Unlock;
      end; // finally
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRServerResendDelay);
      Sleep(0);
      if ServerSession.PacketIDsToResend.ItemCount = 0 then
       StartSleepTime := GetTickCount;
       while ((GetTickCount - StartSleepTime) < ACRServerResendDelay) do  // sleep after last packet send
        begin
         if ServerSession.SendStatus = ACRSent then
           break;
         Sleep(ACRServerReceiveSleep);
        end; // sleep
     end;
   end; // no Connect
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog('Server received all the packets');
{$ENDIF}
  EnterCriticalSection(FCSect);
  ServerSession.ControlCode := ControlCode;
  LeaveCriticalSection(FCSect);
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForCommandSent> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // WaitForCommandSent

procedure WaitForMessageSent;
var
  PacketID,
  ControlCode,
  i:                    Integer;
  StartTime:            DWORD;
  StartSleepTime:       DWORD;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> - START');
{$ENDIF}
 try
  EnterCriticalSection(FCSect);
  ControlCode := ServerSession.MsgControlCode;
  ServerSession.MsgControlCode := ACRExecute;
  LeaveCriticalSection(FCSect);
{$IFDEF ProcessMessages}
    Application.ProcessMessages;
{$ENDIF ProcessMessages}
    Sleep(ACRServerReceiveSleep); // To avoid delay if full answer is already received
    i := ServerSession.MsgPacketIDsToResend.ItemCount;
    StartTime := GetTickCount;
    while not (ServerSession.MsgSendStatus = ACRSent) do  // resend broken packets
     begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> New step');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> SessionID='+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.Session='+IntToStr(Integer(ServerSession.Session)));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> i='+IntToStr(i));
{$ENDIF}
      if (GetTickCount - StartTime) > ACRServerSendTimeOut then
        raise EACRException.Create(40078, ErrorRServerTimeOutSending,
          ['message',
{$IFDEF MsgCommunicator}
          ServerSession.Session.UserID,
{$ELSE}
          ServerSession.Session.SessionID,
{$ENDIF}
          ACRServerSendTimeOut]);
      dec(i);
      if i<0 then
        i := ServerSession.MsgPacketIDsToResend.ItemCount - 1;
      try
       ServerSession.MsgPacketIDsToResend.Lock;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> MsgPacketIDsToResend.ItemCount = '+IntToStr(ServerSession.MsgPacketIDsToResend.ItemCount));
{$ENDIF}
       if (ServerSession.MsgPacketIDsToResend.ItemCount > i) then
        begin
         if i>=0 then // Packet to resend exists
          begin
           PacketID := ServerSession.MsgPacketIDsToResend.Items[i];
           ServerSession.MsgPacketIDsToResend.Unlock;
           if PacketID >= Packets.Count then
             Continue;
          end
         else // No packets to resend - Resend last packet
          begin
           ServerSession.MsgPacketIDsToResend.Unlock;
           PacketID := Packets.Count - 1;
          end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> PacketID = '+IntToStr(PacketID));
{$ENDIF}
         Packet := Packets.Items[PacketID];
         // resend packet
         EnterCriticalSection(FCSect);
         try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(GetTickCount)+':');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> FNetwork=');
aaWriteToLog(IntToHex(Integer(FNetwork),6));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> RemoteHost=');
aaWriteToLog(ServerSession.RemoteHost);
{$ENDIF}
          FNetwork.RemoteHost := ServerSession.RemoteHost;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> RemotePort=');
aaWriteToLog(IntToStr(ServerSession.RemotePort));
{$ENDIF}
          FNetwork.RemotePort := ServerSession.RemotePort;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Sending...');
{$ENDIF}
          if ServerSession.MsgControlCode = ACRTerminate then
            raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
          FNetwork.SendBuffer(Packet.Buffer, Packet.BufferSize);
         finally
          LeaveCriticalSection(FCSect);
         end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Sent!');
{$ENDIF}
        end
      else
        ServerSession.MsgPacketIDsToResend.Unlock;
      finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Locking...');
{$ENDIF}
       ServerSession.MsgPacketIDsToResend.Lock;
       if  (ServerSession.MsgPacketIDsToResend.ItemCount > 0)
       and (ServerSession.MsgPacketIDsToResend.ItemCount > i)
       then
        begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Removing PacketID = '+IntToStr(PacketID));
{$ENDIF}
         ServerSession.MsgPacketIDsToResend.Remove(PacketID);
        end;
       ServerSession.MsgPacketIDsToResend.Unlock;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> UnlLocked');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.Session=');
aaWriteToLog(IntToStr(Integer(ServerSession.Session)));
{$ENDIF}
      end; // finally
{$IFDEF ProcessMessages}
      Application.ProcessMessages;
{$ENDIF ProcessMessages}
//      Sleep(ACRServerResendDelay);
      Sleep(0);
      if ServerSession.MsgPacketIDsToResend.ItemCount = 0 then
       StartSleepTime := GetTickCount;
       while ((GetTickCount - StartSleepTime) < ACRServerResendDelay) do  // sleep after last packet send
        begin
         if ServerSession.SendStatus = ACRSent then
           break;
         Sleep(ACRServerReceiveSleep);
        end; // sleep
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.Session=');
aaWriteToLog(IntToStr(Integer(ServerSession.Session)));
{$ENDIF}
     end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> Ending...');
{$ENDIF}
  EnterCriticalSection(FCSect);
  ServerSession.MsgControlCode := ControlCode;
  LeaveCriticalSection(FCSect);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> ServerSession.MsgSendStatus = '+IntToStr(ServerSession.MsgSendStatus));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> SessionID = '+IntToStr(ServerSession.Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> End!');
{$ENDIF}
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> WaitForMessageSent> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // WaitForMessageSent

begin // DoSendBuffer
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER are sending buffer...');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 if (ServerSession.Session.ConnectParams.PacketSize < SizeOf(TACRPacketHeader)) then
   raise EACRException.Create(40020, ErrorRPacketSizeTooSmall,
                              [ServerSession.Session.ConnectParams.PacketSize,
                                     SizeOf(TACRPacketHeader)]);

 Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
 try
  DataSize := ServerSession.Session.ConnectParams.PacketSize - SizeOf(TACRPacketHeader);
  Header.Signature := ACRServerPacketSign;
  Header.Recepient := ServerSession.RemoteClientID;
  EnterCriticalSection(FCSect);
  Header.Sender := FNetwork.FLocalClient;
  LeaveCriticalSection(FCSect);
  Header.ConnectionID := ServerSession.ConnectionID;
  if Code = ACRConnected then
    Header.SessionID := ServerSession.ClientSessionID
  else
    Header.SessionID := ServerSession.Session.SessionID;
  Header.PacketID := 0;
  Header.ControlCode := Code;
  if Code = ACRMessage then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER IS SENDING MESSAGE # '+IntToStr(ServerSession.ClientMessageID)+' TO CLIENT ID='+IntToStr(ServerSession.RemoteClientID));
aaWriteToLog('==============================================================');
{$ENDIF}
    Header.CurrentRequestID := ServerSession.ServerMessageID;
   end
  else
    Header.CurrentRequestID := ServerSession.CurrentRequestID;
  Packets := TList.Create;
   try
    BytesSent := 0;
    i := DataSize;
    if Code = ACRMessage then
     begin
      ServerSession.MsgSendStatus := ACRNotSent;
     end
    else
      ServerSession.SendStatus := ACRNotSent;
    repeat
//    while BytesSent < BufferSize do
//     begin // Create and send all packets
      New(Packet);
      Packets.Add(Packet);
      Packet.Buffer := MemoryManager.GetMem(ServerSession.Session.ConnectParams.PacketSize);
      if BytesSent + DataSize > BufferSize then
       DataSize := BufferSize - BytesSent;
      Packet.BufferSize := DataSize + SizeOf(TACRPacketHeader);
      if BytesSent + DataSize = BufferSize then
       Header.ControlCode := Header.ControlCode+ACRLastPacket;
      Move(Header^, Packet.Buffer^, SizeOf(TACRPacketHeader));
      Move(Pointer(Integer(Buffer)+Header.PacketID*i)^, Pointer(Integer(Packet.Buffer)+SizeOf(TACRPacketHeader))^, DataSize);
      inc(Header.PacketID);
      // send packet
      EnterCriticalSection(FCSect);
      try
       FNetwork.RemoteHost := ServerSession.RemoteHost;
       FNetwork.RemotePort := ServerSession.RemotePort;
       if Code = ACRMessage then
        begin
         if ServerSession.MsgControlCode = ACRTerminate then
           raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
        end
       else
         if ServerSession.ControlCode = ACRTerminate then
           raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
       FNetwork.SendBuffer(Packet.Buffer, Packet.BufferSize);
      finally
       LeaveCriticalSection(FCSect);
      end;
      // packet has been sent
      BytesSent := BytesSent + DataSize;
//     end;
    until BytesSent >= BufferSize;
    // Wait for the end of sending...
    if Code = ACRMessage then
     begin
       if ServerSession.MsgControlCode <> ACRTerminate then
         WaitForMessageSent;
     end
    else
     begin
      if Code <> ACRConnected then
       if ServerSession.ControlCode <> ACRTerminate then
         WaitForCommandSent;
     end;
   finally
    // Remove all sent packets and free memory
    for i:= 0 to Packets.Count - 1 do
     begin
      Packet := Packets.Items[i];
      MemoryManager.FreeAndNilMem(Packet.Buffer);
      Dispose(Packet);
     end;
    Packets.Free;
    // Allow to add new packets
    if Code = ACRMessage then
     begin
      if ServerSession.MsgControlCode = ACRTerminate then
        raise EACRException.Create(40070, ErrorRSessionTerminated,[Integer(ServerSession)]);
      inc(ServerSession.ServerMessageID);
      sleep(0);
      ServerSession.MsgPacketIDsToResend.Lock;
      if ServerSession.MsgControlCode <> ACRTerminate then
        ServerSession.MsgPacketIDsToResend.SetSize(0);
      ServerSession.MsgPacketIDsToResend.Unlock;
//      ServerSession.MsgReceiveStatus := ACRNoAnswer;
     end
    else
     begin
      if ServerSession.ControlCode = ACRTerminate then
        raise EACRException.Create(40079, ErrorRCommandSessionTerminated,[Integer(ServerSession)]);
      inc(ServerSession.CurrentRequestID);
      sleep(0);
      ServerSession.PacketIDsToResend.Lock;
      ServerSession.PacketIDsToResend.SetSize(0);
      ServerSession.PacketIDsToResend.Unlock;
      ServerSession.AnswerStatus := ACRNoAnswer;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('ServerSession.ControlCode = '+IntToStr(ServerSession.ControlCode));
{$ENDIF}
      EnterCriticalSection(FCSect);
      if ServerSession.ControlCode = ACRSuspend then
        ServerSession.ControlCode := ACRExecute;
      LeaveCriticalSection(FCSect);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('ServerSession.ControlCode = '+IntToStr(ServerSession.ControlCode));
{$ENDIF}
     end;
   end; // try
 finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('SendBuffer> FreeAndNilMem(Header)');
{$ENDIF}
  MemoryManager.FreeAndNilMem(Header);
 end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> ServerSession.MsgSendStatus = '+IntToStr(ServerSession.MsgSendStatus));
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> End!');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER sent buffer');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
 except
  on E: Exception do
    begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('**************************************************************');
aaWriteToLog('TACRServerConnectionManager.DoSendBuffer> - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('**************************************************************');
{$ENDIF}
     raise;
    end;
 end;
end; // DoSendBuffer


//------------------------------------------------------------------------------
// SendBuffer
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendBuffer(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer;
                          Code:       Integer = ACRNoAction
                                                  );
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  SessionFound:         Boolean;
  Packets:              TList;
  i:                    Integer;
  Buf:                  PChar;
  BufSize:              Integer;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.SendBuffer> - START');
{$ENDIF}
  SessionFound := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('SERVER>>> '+IntToStr(ServerSession.CurrentRequestID)+' :');
aaWriteBufferToLog(Buffer,BufferSize);
{$ENDIF}

  inc(ServerSession.Status);
  try
   CompressAndEncryptBuffer(ServerSession.Session, Buffer, BufferSize, Buf, BufSize);
   try
    DoSendBuffer(ServerSession, Buf, BufSize, Code);
   finally
    if Buf <> Buffer then
      MemoryManager.FreeAndNilMem(Buf);
   end;
   dec(ServerSession.Status);
  except
   dec(ServerSession.Status);
   raise;
  end;

  (*
  if Code <> ACRMessage then
   begin
    ServerSession.AnswerStatus := ACRNoAnswer;
    Packets := ServerSession.Packets.LockList;
    try
     for i:=0 to Packets.Count-1 do
       Packets.Items[i] := nil;
     Packets.Count := 0;
    finally
     ServerSession.Packets.UnlockList;
    end;
   end
  else
   begin
    ServerSession.MsgReceiveStatus := ACRNoAnswer;
    Packets := ServerSession.MsgPackets.LockList;
    try
     for i:=0 to Packets.Count-1 do
       Packets.Items[i] := nil;
     Packets.Count := 0;
    finally
     ServerSession.MsgPackets.UnlockList;
    end;
   end;

  DoSendBuffer(ServerSession, Buffer, BufferSize, Code);

  if Code <> ACRMessage then
   begin
//     if ServerSession.AnswerStatus = ACRNoAnswer then
//       inc(ServerSession.CurrentRequestID);
   end
  else
   begin
     inc(ServerSession.ServerMessageID);
   end;
*)
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('TACRServerConnectionManager.SendBuffer> - FINISH');
{$ENDIF}
end; // SendBuffer


//------------------------------------------------------------------------------
// SendMessage
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendMessage(
                          Session:    TACRBaseSession;
                          Buffer:     PChar;
                          BufferSize: Integer
                                                  );
var
  Buf:            PChar;
  SessionFound:   Boolean;
  Sessions:       TList;
  i:              Integer;
  ServerSession:  PACRSrvrSession;
  StartTime:      DWORD;
begin
  SessionFound := False;
  Sessions := FSessions.LockList;
  try
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FSessions.UnlockList;
  end;
  if not SessionFound then
    raise EACRException.Create(40021, ErrorRSessionNotConnected, [Integer(Session)]);
  // wait for Session be vacant
  StartTime := GetTickCount;
  repeat
   if Session = nil then
     raise EACRException.Create(40071, ErrorRSessionDeleted);
   if ServerSession.MsgControlCode = ACRTerminate then
     raise EACRException.Create(40072, ErrorRSendSessionTerminated);
   if (GetTickCount - StartTime) > ACRWaitForMessagesSend then
     raise EACRException.Create(40073, ErrorRTimeoutMessageSent,[
{$IFDEF MsgCommunicator}
          ServerSession.Session.UserID,
{$ELSE}
          ServerSession.Session.SessionID,
{$ENDIF}
          ACRWaitForMessagesSend]);
   if (ServerSession.Status = ACRVacant) then
      break;
{$IFDEF ProcessMessages}
   Application.ProcessMessages;
{$ENDIF ProcessMessages}
   Sleep(ACRServerResendDelay);
  until False;
  // block other threads to send at the same time
{$IFDEF MsgCommunicator}
  EnterCriticalSection(TACRServerSession(Session).FCSect);
{$ENDIF}
  try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - Start');
{$ENDIF}
(*
   if (TACRCompressionAlgorithm(Session.ConnectParams.CompressionAlgorithm) <> acaNone)
   or (Session.ConnectParams.CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None)
   then
    begin
     Buf := MemoryManager.GetMem(BufferSize);
     Move(Buffer^,Buf^,BufferSize);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - SendBuffer SessionID='+IntToStr(Session.SessionID));
{$ENDIF}
     SendBuffer(Session, Buf, BufferSize, ACRMessage);
     MemoryManager.FreeAndNilMem(Buf);
    end
   else
*)
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - SendBuffer SessionID='+IntToStr(Session.SessionID));
{$ENDIF}
     SendBuffer(Session, Buffer, BufferSize, ACRMessage);
    end;
  finally
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerConnectionManager.SendMessage> - SessionID='+IntToStr(Session.SessionID));
aaWriteToLog('TACRServerConnectionManager.SendMessage> - FINISH');
{$ENDIF}
{$IFDEF MsgCommunicator}
   LeaveCriticalSection(TACRServerSession(Session).FCSect);
{$ENDIF}
  end;
end; // SendMessage


//------------------------------------------------------------------------------
// SendAcknowledgement
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.SendAcknowledgement(
                          ServerSession:        PACRSrvrSession;
                          Msg:                  Boolean = False;
                          CurrentRequestID:     Integer = -1
                                                          );
var
  Header:          PACRPacketHeader;
begin
  Header := MemoryManager.GetMem(SizeOf(TACRPacketHeader));
  try
   if Msg then
     Header.ControlCode := ACRMessageReceived
   else
     Header.ControlCode := ACRAllPacketsReceived;
   Header.Signature := ACRServerPacketSign;
   Header.Recepient := ServerSession.RemoteClientID;
   Header.ConnectionID := ServerSession.ConnectionID;
   Header.SessionID := ServerSession.Session.SessionID;
   Header.PacketID := 0;
   if CurrentRequestID>=0 then
     Header.CurrentRequestID := CurrentRequestID
   else
    begin
     if Msg then
       Header.CurrentRequestID := ServerSession.ClientMessageID
     else
       Header.CurrentRequestID := ServerSession.CurrentRequestID;
    end;
   EnterCriticalSection(FCSect);
   try
    FNetwork.RemoteHost := ServerSession.RemoteHost;
    FNetwork.RemotePort := ServerSession.RemotePort;
    FNetwork.SendBuffer(PChar(Header), SizeOf(TACRPacketHeader));
   finally
    LeaveCriticalSection(FCSect);
   end;
  finally
   MemoryManager.FreeAndNilMem(Header);
  end;
end;// SendAcknowledgement


//------------------------------------------------------------------------------
// GetClientInfo
//------------------------------------------------------------------------------
function TACRServerConnectionManager.GetClientInfo(
                          Session:          TACRBaseSession;
                          var Host:         String;
                          var Port:         Integer;
                          var Application:  String
                                                    ): Boolean;
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
  Result := False;
  Sessions:=FSessions.LockList;
  try
   for i:=Sessions.Count-1 downto 0 do
    begin
     ServerSession := Sessions.Items[i];
     if ServerSession.Session = Session then
      begin
       Host := ServerSession.RemoteHost;
       Port := ServerSession.RemotePort;
       Application := ServerSession.Application;
       Result := True;
       break;
      end;
    end;
  finally
   FSessions.UnlockList;
  end;
{
  if not Found then
    raise EACRException.Create(40037, ErrorRSessionNotFound,
                                ['Server', Session.SessionID]);
}
end; // GetClientInfo


//------------------------------------------------------------------------------
// GetClientsList
//------------------------------------------------------------------------------
procedure TACRServerConnectionManager.GetClientsList(var Clients: TACRSessionsArray);
var
  Sessions:             TList;
  ServerSession:        PACRSrvrSession;
  i:                    Integer;
begin
  Sessions:=FSessions.LockList;
  try
   SetLength(Clients, Sessions.Count);
   for i:=0 to Sessions.Count-1 do
    begin
     ServerSession := Sessions.Items[i];
     Clients[i] := ServerSession.Session;
    end;
  finally
   FSessions.UnlockList;
  end;
end; // GetClientsList


//------------------------------------------------------------------------------
// if authorization buffer is valid (encrypted by the same crypto settings as in CryptoInfo)
// then return true else return false
//------------------------------------------------------------------------------
function TACRServerConnectionManager.IsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
begin
  Result := fnIsAuthorizationBufferValid(CryptoInfo, Buffer, BufferSize);
end; // IsAuthorizationBufferValid

// TACRServerConnectionManager




////////////////////////////////////////////////////////////////////////////////
//
// TACRServerSessionThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerSessionThread.Create(
                          Manager:          TACRServerConnectionManager;
                          ServerSession:    PACRSrvrSession;
                          Code:             Integer = ACRNoAction
                                            );
begin
  FManager := Manager;
  FManager.IncThreadCount;
  FServerSession := ServerSession;
  FCode := Code;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy;
//------------------------------------------------------------------------------
destructor TACRServerSessionThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER SESSION THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  inherited Destroy;
  if (FCode = ACRMessage) then
    FServerSession.MsgThread := nil
  else
  if (FCode <> ACRDisconnect)
  then
    FServerSession.Thread := nil;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER SESSION THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end;// Destroy;


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerSessionThread.Execute;
var
  Buf:                 PChar;
  BufSize, AuBufSize:  Integer;
  ConnectionParams:    PACRConnectionParams;
begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - START');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AnswerStatus = '+IntToStr(FServerSession.AnswerStatus));
{$ENDIF}
 try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Code='+IntToStr(FCode));
{$ENDIF}
  if (FCode <> ACRMessage)
  and (FCode <> ACRDisconnect)
  then
   begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ControlCode='+IntToStr(FServerSession.ControlCode));
{$ENDIF}
    FManager.ReceiveBuffer(FServerSession, Buf, BufSize);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer received');
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ControlCode='+IntToStr(FServerSession.ControlCode));
{$ENDIF}
   end;
  if (FCode = ACRConnect) then
   begin
(*
    if not FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('Decompress/Decrypt - Error!');
{$ENDIF}
      Exit;
     end;
*)
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': IsAuthorizationBufferValid');
{$ENDIF}
{$IFDEF MsgCommunicator}
    Move(Buf^, FServerSession.Session.FUserID, SizeOf(FServerSession.Session.FUserID));
    inc(Buf, SizeOf(FServerSession.Session.FUserID));
{$ENDIF}
    Move(PChar(Buf + SizeOf(TACRConnectionParams))^, AuBufSize, SizeOf(AuBufSize));
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('SERVER> AuthorizationBuffer:');
aaWriteBufferToLog(PChar(Buf + SizeOf(TACRConnectionParams) + SizeOf(AuBufSize)),AuBufSize);
{$ENDIF}
    if not FManager.IsAuthorizationBufferValid(
              FServerSession.Session.ConnectParams.CryptoInfo,
              PChar(Buf + SizeOf(TACRConnectionParams) + SizeOf(AuBufSize)),
              AuBufSize
                                               )
    then
     begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('IsAuthorizationBufferValid - Invalid!!!');
{$ENDIF}
      FManager.DoDisconnect(FServerSession.Session);
      Exit;
     end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('IsAuthorizationBufferValid - OK!');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ConnectionParams');
{$ENDIF}
    // Get ConnectParams
    ConnectionParams := PACRConnectionParams(Buf);
    FServerSession.Session.FConnectParams.PacketSize := ConnectionParams.PacketSize;
    FServerSession.Session.FConnectParams.CompressionAlgorithm := ConnectionParams.CompressionAlgorithm;
    FServerSession.Session.FConnectParams.CompressionMode := ConnectionParams.CompressionMode;
    // Get client Application name
    SetLength(FServerSession.Application, BufSize-SizeOf(TACRConnectionParams)-SizeOf(AuBufSize)-AuBufSize-1
{$IFDEF MsgCommunicator}
    -SizeOf(FServerSession.Session.FUserID)
{$ENDIF}
    );
{$IFDEF LOG_SERVER_THREADS}
aaWriteBufferToLog(Buf+SizeOf(TACRConnectionParams)+SizeOf(AuBufSize)+AuBufSize,BufSize-SizeOf(TACRConnectionParams)-SizeOf(AuBufSize)-AuBufSize-1);
{$ENDIF}
    StrCopy(PChar(FServerSession.Application), Buf+SizeOf(TACRConnectionParams)+SizeOf(AuBufSize)+AuBufSize);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('Application="'+FServerSession.Application+'"');
{$ENDIF}
    // Free
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Free Buf');
{$ENDIF}
{$IFDEF MsgCommunicator}
    dec(Buf, SizeOf(FServerSession.Session.FUserID));
{$ENDIF}
    MemoryManager.FreeAndNilMem(Buf);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SendConnectAckn');
{$ENDIF}
    FManager.SendConnectAckn(FServerSession);
{$IFDEF MsgCommunicator}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ConnectUser...');
{$ENDIF}
    if not TACRServerSession(FServerSession.Session).ConnectUser then
      FManager.Disconnect(FServerSession.Session);
{$ENDIF}
//    inc(FServerSession.CurrentRequestID);
//    ServerSession.Session.Connected := True;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER CONNECTED CLIENT #'+IntToStr(FServerSession.RemoteClientID));
aaWriteToLog('PacketSize='+IntToStr(FServerSession.Session.ConnectParams.PacketSize));
aaWriteToLog('CompressionAlgorithm='+IntToStr(FServerSession.Session.ConnectParams.CompressionAlgorithm));
aaWriteToLog('CompressionMode='+IntToStr(FServerSession.Session.ConnectParams.CompressionMode));
aaWriteToLog('CryptoAlgorithm='+IntToStr(FServerSession.Session.ConnectParams.CryptoInfo.CryptoAlgorithm));
aaWriteToLog('Application="'+FServerSession.Application+'"');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
   end
  else
  if (FCode = ACRDisconnect) then
   begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Disconnect...');
{$ENDIF}
    if FServerSession.Session <> nil then
     begin
{$IFDEF MsgCommunicator}
 {$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DisconnectUser starting...');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SessionID='+IntToStr(Integer(FServerSession.Session.SessionID)));
 {$ENDIF LOG_SERVER_THREADS}
      TACRServerSession(FServerSession.Session).DisconnectUser;
 {$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SessionID='+IntToStr(Integer(FServerSession.Session.SessionID)));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DisconnectUser finished!!!');
 {$ENDIF LOG_SERVER_THREADS}
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': DoDisconnect');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': SessionID='+IntToStr(Integer(FServerSession.Session.SessionID)));
{$ENDIF}
      FManager.DoDisconnect(FServerSession.Session);
     end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Disconnected');
{$ENDIF}
   end
  else
  if (FCode = ACREcho) then
   begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Echo...');
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID = '+IntToStr(FServerSession.CurrentRequestID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufSize = '+IntToStr(BufSize));
{$ENDIF}
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Decompressed BufSize = '+IntToStr(BufSize));
{$ENDIF}
      FManager.SendBuffer(FServerSession.Session, Buf, BufSize, ACRNoAction);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Compressed BufSize = '+IntToStr(BufSize));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID = '+IntToStr(FServerSession.CurrentRequestID));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Echo sent');
{$ENDIF}
{ TODO -oLeo :
You must free the buffer you received when it is not needed any more.
Please check the following line to be present in your code: }
    MemoryManager.FreeAndNilMem(Buf);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER SENT ECHO TO CLIENT #'+IntToStr(FServerSession.RemoteClientID));
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
   end
  else
  if (FCode = ACRMessage) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call FManager.ReceiveMessage');
{$ENDIF}
    FManager.ReceiveMessage(FServerSession, Buf, BufSize);
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
      FManager.TerminateMessageThreads(FServerSession);
      FServerSession.MsgControlCode := ACRExecute;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call ServerSession.ReceiveMessage');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
      TACRNetworkSession(FServerSession.Session).ReceiveMessage(Buf, BufSize);
     end;
   end
  else
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: Call ServerSession.ReceiveData');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
    FManager.TerminateCommandThreads(FServerSession);
    FServerSession.ControlCode := ACRSuspend;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': BufSize = '+IntToStr(BufSize));
{$ENDIF}
    if FManager.DecompressAndDecryptBuffer(FServerSession.Session, Buf, BufSize) then
     begin
{$IFDEF DEBUG_LOG_NETWORK_COMMUNICATION}
aaWriteToLog('Server started new request #'+IntToStr(FServerSession.CurrentRequestID));
aaWriteBufferToLog(Buf, BufSize);
{$ENDIF}
{$IFDEF MsgCommunicator}
      FServerSession.Session.ReceiveData(Buf, BufSize);
{$ELSE}
      FServerSession.Session.ReceiveData(Buf, BufSize);
{$ENDIF}
     end;
   end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: FINISHED EXECUTE SESSION *****************************');
aaWriteToLog('==============================================================');
{$ENDIF}
 except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerSessionThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
    end;
{$ELSE}
  raise;
{$ENDIF}
 end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - FINISH');
{$ENDIF}
end; // Execute

// TACRServerSessionThread




////////////////////////////////////////////////////////////////////////////////
//
// TACRServerListenerThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerListenerThread.Create(
                       Manager:       TACRServerConnectionManager;
                       Buffer:        PChar;
                       BufferSize:    Integer;
                       FromHost:      String;
                       FromPort:      Integer
                                            );
begin
  FManager := Manager;
  self.Buffer := Buffer;
  self.BufferSize := BufferSize;
  self.FromHost := FromHost;
  self.FromPort := FromPort;
  inherited Create(False);
  Priority := tpNormal;//tpHigher;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
  FManager.IncThreadCount;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerListenerThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if SessionFound then
   begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FOUND');
{$ENDIF}
    if ServerSession.ListeningThreads <> nil then
      ServerSession.ListeningThreads.Remove(self);
    if ServerSession.MsgListeningThreads <> nil then
      ServerSession.MsgListeningThreads.Remove(self);
   end;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - inherited Destroy');
{$ENDIF}
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerListenerThread.Execute;
label
  KillPacket;
var
  Header:               PACRPacketHeader;
  Packets:              TList;
  Packet:               PACRPacket;
  Sessions:             TList;
  i:                    Integer;

function AllPacketsReceived: Boolean;
var
  i:                    Integer;
begin
// Are all packets received?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Are all packets received?');
{$ENDIF}
 Result := True;
 Packets := ServerSession.Packets.LockList;
 try
  for i:=0 to Packets.Count-1 do
  if Packets.Items[i] = nil then
   begin
    Result := False;
   end;
  finally
  ServerSession.Packets.UnlockList;
  end;
end; // function AllPacketsReceived: Boolean;

begin // TACRServerListenerThread.Execute;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('TACRServerListenerThread.Execute');
{$ENDIF}
 try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: '+IntToStr(GetTickCount)+' > Listener started');
aaWriteToLog('==============================================================');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** SERVER LISTENER START *****');
{$ENDIF}
 try
  Header := Pointer(Buffer);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
aaWriteToLog('           Header.PacketID         = '+IntToStr(Header.PacketID));
aaWriteToLog('           Header.ControlCode      = '+IntToStr(Header.ControlCode));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connections');
{$ENDIF}
//EnterCriticalSection(FManager.FCSect);
{$IFDEF LOG_SERVER_THREADS}
//aaWriteToLog('ENTERED!');
{$ENDIF}
// Check for errors
  if CheckSum(Buffer, BufferSize) <> Header.CheckSum then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid CheckSum');
{$ENDIF}
    FManager.PacketResendRequest(Buffer, FManager.FNetwork, FromHost, FromPort);
    goto KillPacket;
   end;
// Check sign
  if (Header.Signature <> ACRClientPacketSign) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid Signature');
{$ENDIF}
    goto KillPacket;
   end;
// Check ServerID
  if (Header.Recepient <> TACRServer(FManager.FServer).ServerID) then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid ServerID');
{$ENDIF}
    goto KillPacket;
   end;
  if    // first packet in multi-packet connect request
  ((Header.ControlCode=ACRConnect) and (Header.PacketID=0))
  or    // single-packet connect request
  ((Header.ControlCode=(ACRConnect+ACRLastPacket)) and (Header.PacketID=0))
  then
   begin // Connect
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Connect');
{$ENDIF}
    SessionFound := False;
    Sessions := FManager.FSessions.LockList;
    try
     for i:=0 to Sessions.Count-1 do
      begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Check session '+IntToStr(i)+' of '+IntToStr(Sessions.Count));
{$ENDIF}
       ServerSession := Sessions.Items[i];
       if (ServerSession.ClientSessionID = Header.SessionID)
       and (ServerSession.ConnectionID = Header.ConnectionID)
       and (ServerSession.RemoteClientID = Header.Sender)
       then
        begin
         SessionFound := True;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ServerSession),6)+', Client SessionID='+IntToStr(ServerSession.ClientSessionID));
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Found: @session='+IntToHex(Integer(ServerSession),6)+', Server SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
         break;
        end;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
    if not SessionFound then
     begin // First connect
// new ServerSession
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New ServerSession');
{$ENDIF}
      New(ServerSession);
      ServerSession.Status := ACRVacant;
      ServerSession.AnswerStatus := ACRStartAnswer;
      ServerSession.RemoteHost := FromHost;
      ServerSession.RemotePort := FromPort;
      ServerSession.ConnectionID := Header.ConnectionID;
      ServerSession.ClientSessionID := Header.SessionID;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ServerSession),6)+', Client SessionID='+IntToStr(ServerSession.ClientSessionID));
{$ENDIF}
      ServerSession.RemoteClientID := Header.Sender;
      ServerSession.Session := TACRServerSession.Create(FManager.FServer);
      ServerSession.Session.FConnectParams.CryptoInfo := TACRServer(FManager.FServer).CryptoParams.GetCryptoParams;
      ServerSession.Thread := nil;
      ServerSession.MsgThread := nil;
      ServerSession.ListeningThreads := TThreadList.Create;
      ServerSession.ControlCode := ACRExecute;
      ServerSession.ListeningThreads.Add(self);
      ServerSession.MsgListeningThreads := TThreadList.Create;
      ServerSession.MsgControlCode := ACRExecute;
// Set SessionID
      EnterCriticalSection(FManager.FCSect);
      ServerSession.Session.SessionID := FManager.FSessionID;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+':   New: @session='+IntToHex(Integer(ServerSession),6)+', Server SessionID='+IntToStr(ServerSession.Session.SessionID));
{$ENDIF}
      inc(FManager.FSessionID);
      if FManager.FSessionID = INVALID_SESSION_ID then
       inc(FManager.FSessionID);
      LeaveCriticalSection(FManager.FCSect);
// Create PacketIDsToResend
      ServerSession.PacketIDsToResend := TACRThreadIntArray.Create;
      ServerSession.MsgPacketIDsToResend := TACRThreadIntArray.Create;
// Create new packets list
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packets list create');
{$ENDIF}
      ServerSession.Packets := TThreadList.Create;
      Packets := ServerSession.Packets.LockList;
      try
       Packets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
      finally
       ServerSession.Packets.UnlockList;
      end;
      ServerSession.MsgPackets := TThreadList.Create;
      Packets := ServerSession.MsgPackets.LockList;
      try
       Packets.Capacity := ACRDefaultPacketsInRequest; // Allocate some place in list
      finally
       ServerSession.MsgPackets.UnlockList;
      end;
// add new session to list
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': add new session to list');
{$ENDIF}
      FManager.FSessions.Add(ServerSession);
      SessionFound := True;
     end
    else // not first connect - reset and resend connection info
     begin
      if ServerSession.Thread <> nil then // Is buffer receiving now?
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Buffer is receiving now');
{$ENDIF}
        goto KillPacket;
       end;
      if (ServerSession.ControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Not Execute - Terminate ListeningThread, ServerSession.ControlCode='+IntToStr(ServerSession.ControlCode));
{$ENDIF}
        goto KillPacket;
       end;
      if (ServerSession.CurrentRequestID > Header.CurrentRequestID) then // packet from old connection request
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - SendAckn');
{$ENDIF}
        FManager.SendConnectAckn(ServerSession, Header.CurrentRequestID);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': OldPacket - Terminate ListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
      ServerSession.MsgThread := nil;
      ServerSession.ListeningThreads.Add(self);
// Set PacketIDsToResend
      ServerSession.PacketIDsToResend.Lock;
      ServerSession.PacketIDsToResend.SetSize(0);
      ServerSession.PacketIDsToResend.Unlock;
      ServerSession.MsgPacketIDsToResend.Lock;
      ServerSession.MsgPacketIDsToResend.SetSize(0);
      ServerSession.MsgPacketIDsToResend.Unlock;
(*
// Delete old packets from list
      Packets := ServerSession.Packets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ServerSession.Packets.UnlockList;
      end;
      Packets := ServerSession.MsgPackets.LockList;
      try
       for i:=0 to Packets.Count-1 do
         Packets.Items[i] := nil;
       Packets.Count := 0;
      finally
       ServerSession.Packets.UnlockList;
      end;
*)
     end;
// Put this first packet
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': New Packet');
{$ENDIF}
    New(Packet);
    Packet.Buffer := Buffer;
    Packet.BufferSize := BufferSize;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet add');
{$ENDIF}
    ServerSession.Packets.Add(Packet);
    ServerSession.CurrentRequestID := Header.CurrentRequestID;
    ServerSession.ClientMessageID := 0;
    ServerSession.ServerMessageID := 0;
// Get connection parameters
    if Header.ControlCode=ACRConnect+ACRLastPacket then
     ServerSession.AnswerStatus := ACRFullAnswer;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSessionThread.Create');
{$ENDIF}
    ServerSession.Thread := TACRServerSessionThread.Create(FManager,
                                            ServerSession, ACRConnect);
    Sleep(0);
    Exit;
   end; // Connect
// verify ConnectionID existing and SessionID existing in this connection
  SessionFound := False;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sessions');
{$ENDIF}
  Sessions := FManager.FSessions.LockList;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Sessions Locked');
{$ENDIF}
  try
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           @Sessions='+IntToStr(Integer(Sessions)));
aaWriteToLog('           Sessions.Count='+IntToStr(Integer(Sessions.Count)));
{$ENDIF}
   for i:=0 to Sessions.Count-1 do
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           i='+IntToStr(i));
aaWriteToLog('           Sessions.Count='+IntToStr(Integer(Sessions.Count)));
{$ENDIF}
     ServerSession := Sessions.Items[i];
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('           @ServerSession='+IntToStr(Integer(ServerSession)));
{$ENDIF}
     if (ServerSession.ConnectionID = Header.ConnectionID)
     and ( (ServerSession.Session.SessionID = Header.SessionID)
           or ( (ServerSession.ClientSessionID = Header.SessionID)
                and  ( (Header.ControlCode = ACRConnect)
                    or (Header.ControlCode = (ACRConnect + ACRLastPacket))
                     )
              )
         )
     then
       begin
        SessionFound := True;
        break;
       end;
    end;
  finally
   FManager.FSessions.UnlockList;
  end;
  if not SessionFound then
   begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session not found');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid SessionID or ConnectionID');
{$ENDIF}
    // client try to contimue work with new server without loggin first
    // instead of old server it was loggin into
    New(ServerSession);
    ServerSession.CurrentRequestID := Header.CurrentRequestID;
    ServerSession.RemoteClientID := Header.Sender;
    ServerSession.ConnectionID := Header.ConnectionID;
    ServerSession.Session := TACRServerSession.Create(FManager.FServer);
    ServerSession.Session.SessionID := Header.SessionID;
    ServerSession.RemoteHost := FromHost;
    ServerSession.RemotePort := FromPort;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Try to send ACKN to old client...');
{$ENDIF}
    if Header.ControlCode = ACRDisconnect then
      FManager.SendAcknowledgement(ServerSession)
    else
      FManager.SendDisconnectRequest(ServerSession, False);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session.Free...');
{$ENDIF}
    ServerSession.Session.Free;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Dispose...');
{$ENDIF}
    Dispose(ServerSession);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Old client is disconnected');
{$ENDIF}
    goto KillPacket;
   end;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Session found');
{$ENDIF}
// verify Sender
  if ServerSession.RemoteClientID <> Header.Sender then
   begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER: invalid Sender');
{$ENDIF}
    goto KillPacket;
   end;
// check CurrentRequestID
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Header.CurrentRequestID = '+IntToStr(Header.CurrentRequestID));
{$ENDIF}
  if (Header.ControlCode = ACRMessageReceived)
  or (Header.ControlCode = ACRMessagePacketResendRequest)
  then
   begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ServerSession.ServerMessageID = '+IntToStr(ServerSession.ServerMessageID));
{$ENDIF}
    if ServerSession.ServerMessageID <> Header.CurrentRequestID then
      goto KillPacket;
   end
  else
  if (Header.ControlCode = ACRMessage)
  or (Header.ControlCode = (ACRMessage + ACRLastPacket))
  then
   begin
    if ServerSession.ClientMessageID > Header.CurrentRequestID then // old message
     begin
      goto KillPacket;
     end;
    if ServerSession.ClientMessageID < Header.CurrentRequestID then // lost acknowledgement from client
      ServerSession.MsgSendStatus := ACRSent;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Listening Thread in '+IntToStr(Integer(ServerSession)));
{$ENDIF}
    if ServerSession.MsgListeningThreads = nil then
     begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': MsgListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
     begin
      if (ServerSession.MsgControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminate ACRListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
      ServerSession.MsgListeningThreads.Add(self);
     end;
   end
  else
   begin
    if Header.ControlCode <> ACRDisconnect then
     if ServerSession.CurrentRequestID > Header.CurrentRequestID then // old command
      goto KillPacket;
{
    if ServerSession.CurrentRequestID < Header.CurrentRequestID then // lost acknowledgement from client
      ServerSession.SendStatus := ACRSent;
}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Add Listening Thread in '+IntToStr(Integer(ServerSession)));
{$ENDIF}
    if ServerSession.ListeningThreads = nil then
     begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ListeningThreads = nil');
{$ENDIF}
      goto KillPacket;
     end
    else
     begin
      if Header.ControlCode <> ACRDisconnect then
      if (Header.ControlCode <> ACRAllPacketsReceived) then
      if (ServerSession.ControlCode <> ACRExecute) then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminate ListeningThread');
{$ENDIF}
        goto KillPacket;
       end;
      ServerSession.ListeningThreads.Add(self);
     end;
   end;
// Process service packet
  case Header.ControlCode of
   ACRDisconnect:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRDisconnect');
{$ENDIF}
     EnterCriticalSection(FManager.FCSect);
     ServerSession.ControlCode := ACRTerminate;
     ServerSession.MsgControlCode := ACRSuspend;
     LeaveCriticalSection(FManager.FCSect);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement...');
{$ENDIF}
     FManager.SendAcknowledgement(ServerSession, False, Header.CurrentRequestID);
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Acknowledgement has been sent');
{$ENDIF}
     TACRServerSessionThread.Create(FManager, ServerSession, Header.ControlCode);
     goto KillPacket;
    end;
   ACRPacketResendRequest:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRPacketResendRequest');
{$ENDIF}
     ServerSession.PacketIDsToResend.Lock;
     ServerSession.PacketIDsToResend.Append(Header.PacketID);
     ServerSession.PacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Resend Request');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     goto KillPacket;
    end;
   ACRMessagePacketResendRequest:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRMessagePacketResendRequest');
{$ENDIF}
     ServerSession.MsgPacketIDsToResend.Lock;
     ServerSession.MsgPacketIDsToResend.Append(Header.PacketID);
     ServerSession.MsgPacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Message Resend Request');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     goto KillPacket;
    end;
   ACRAllPacketsReceived:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRAllPacketsReceived');
{$ENDIF}
     ServerSession.SendStatus := ACRSent;
     ServerSession.PacketIDsToResend.Lock;
     ServerSession.PacketIDsToResend.SetSize(0);
     ServerSession.PacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Command Acknowledgement');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     sleep(0);
     goto KillPacket;
    end;
   ACRMessageReceived:
    begin
     ServerSession.MsgSendStatus := ACRSent;
     ServerSession.MsgPacketIDsToResend.Lock;
     ServerSession.MsgPacketIDsToResend.SetSize(0);
     ServerSession.MsgPacketIDsToResend.Unlock;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER received Message Acknowledgement');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
     goto KillPacket;
    end;
   ACRMessage,
   (ACRMessage+ACRLastPacket):
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Data packet');
{$ENDIF}
     if (ServerSession.MsgReceiveStatus = ACRFullAnswer) then // lost acknowledgement
      begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full Answer!');
{$ENDIF}
       goto KillPacket; // Do not replace correct packets with doubles
      end;
// new request?
// Is this packet already stored?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Is this packet already stored?');
{$ENDIF}
     Packets := ServerSession.MsgPackets.LockList;
     if Packets.List = nil then // list absent
      begin
       ServerSession.MsgPackets.UnlockList;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': list absent');
{$ENDIF}
       goto KillPacket;
      end
     else // list exists
      if ((Header.PacketID+1) <= Packets.Count) then // old packet
       if (Packets.Items[Header.PacketID] <> nil) then // already received
        begin
         ServerSession.MsgPackets.UnlockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(Packets.Count)));
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(Packets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('Server received existing packet');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('AllPacketsReceived?');
{$ENDIF}
        if AllPacketsReceived then
          begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('Yes');
{$ENDIF}
           ServerSession.MsgReceiveStatus := ACRFullAnswer;
           FManager.MessageReceived(ServerSession);
          end;
         goto KillPacket; // Do not replace correct packets with doubles
        end;
// First packet in new request?
     if Packets.Count = 0 then
       ServerSession.MsgReceiveStatus := ACRStartAnswer;
     ServerSession.MsgPackets.UnlockList;
// Add correct received data packet
     if Terminated then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
        goto KillPacket;
       end;
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
if ((Header.PacketID+1) <= Packets.Count) then // old packet
if (Packets.Items[Header.PacketID] = nil) then // not received yet
aaWriteToLog('SERVER received packet #'+IntToStr(Header.PacketID));
{$ENDIF}
     Packets := ServerSession.MsgPackets.LockList;
     try
      if Packets.Count < (Header.PacketID + 1) then
        Packets.Count := Header.PacketID + 1; // List fills hole by Nil
      if (ServerSession.MsgReceiveStatus <> ACRFullAnswer) then
        Packets.Items[Header.PacketID] := Packet
      else
       begin
        Dispose(Packet);
        Packet := nil;
       end;
     finally
      ServerSession.MsgPackets.UnlockList;
     end;
     if Packet = nil then
       goto KillPacket;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': server has added new packet');
{$ENDIF}
(*
     if Terminated then
      begin
       Dispose(Packet);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
       goto KillPacket;
      end;
*)
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER has added new packet');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// test last packet
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Last Packet');
{$ENDIF}
     if (Header.ControlCode >= ACRLastPacket) then
      begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       Header.ControlCode := Header.ControlCode - ACRLastPacket;
       if AllPacketsReceived then
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
         ServerSession.MsgReceiveStatus := ACRFullAnswer;  // Allow to extract buffer
        end
       else
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
         // start resend requesting
         ServerSession.MsgReceiveStatus := ACRAnswerNotFull;
        end;
      end // Last Packet
     else // Not last Packet
       ServerSession.MsgReceiveStatus := ACRStartAnswer;
// Full Answer
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ClientMessageID = '+IntToStr(ServerSession.ClientMessageID));
{$ENDIF}
     if ServerSession.MsgReceiveStatus = ACRFullAnswer then
       if Terminated then
         goto KillPacket
       else
         FManager.MessageReceived(ServerSession);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End messge packet process
     Exit;
    end; // Message packet process
   ACREcho,
   ACRConnect, // for packets with ID>0, not for first packet
   ACRNewRequest,
   ACRLastPacket..(ACRMessage+ACRLastPacket-1),
   (ACRMessage+ACRLastPacket+1)..127,
   ACRNoAction:
    begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Data packet');
{$ENDIF}
// Check for if the previous data received fully:
{
     if Header.ControlCode then
      begin
       discontinue the previous sending;
      end;
Suppose client session should not send new request if it did not receive the previous fully
}

//     if ServerSession.CurrentRequestID = Header.CurrentRequestID then
     if (ServerSession.AnswerStatus = ACRFullAnswer) then // lost acknowledgement
      begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Full Answer!');
{$ENDIF}
       if (Header.ControlCode <> ACRConnect)
       and (Header.ControlCode <> (ACRConnect+ACRLastPacket))
       then
         FManager.SendAcknowledgement(ServerSession);
       goto KillPacket; // Do not replace correct packets with doubles
      end;
// new request?
(*
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('new request?');
{$ENDIF}
     if Header.ControlCode >= ACRNewRequest then
      if ServerSession.CurrentRequestID < Header.CurrentRequestID then
       begin
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('Yes');
{$ENDIF}
        ServerSession.AnswerStatus := ACRStartAnswer;
        ServerSession.CurrentRequestID := Header.CurrentRequestID;
        Packets := ServerSession.Packets.LockList;
        try
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('Packets.Count='+IntToStr(Packets.Count));
{$ENDIF}
         for i:=0 to Packets.Count-1 do
           Packets.Items[i] := nil;
         Packets.Count := 0;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('Packets.Count='+IntToStr(Packets.Count));
{$ENDIF}
        finally
         ServerSession.Packets.UnlockList;
        end;
       end;
*)
// Is this packet already stored?
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Is this packet already stored?');
{$ENDIF}
     Packets := ServerSession.Packets.LockList;
     if Packets.List = nil then // list absent
      begin
       ServerSession.Packets.UnlockList;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': list absent');
{$ENDIF}
       goto KillPacket;
      end
     else // list exists
      if ((Header.PacketID+1) <= Packets.Count) then // old packet
       if (Packets.Items[Header.PacketID] <> nil) then // already received
        begin
         ServerSession.Packets.UnlockList;
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Packets.Count='+IntToStr(Integer(Packets.Count)));
{$ENDIF}
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Existing packet');
aaWriteToLog('           @Items='+IntToStr(Integer(Packets.Items[Header.PacketID])));
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('Server received existing packet');
{$ENDIF}
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('AllPacketsReceived?');
{$ENDIF}
        if AllPacketsReceived then
          begin
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('Yes');
{$ENDIF}
           ServerSession.AnswerStatus := ACRFullAnswer;
           FManager.CommandReceived(ServerSession, Header.ControlCode);
          end;
         goto KillPacket; // Do not replace correct packets with doubles
        end;
// First packet in new request?
     if Packets.Count = 0 then
       ServerSession.AnswerStatus := ACRStartAnswer;
     ServerSession.Packets.UnlockList;
// Add correct received data packet
     if Terminated then
       begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
        goto KillPacket;
       end;
     New(Packet);
     Packet.Buffer := Pointer(Buffer);
     Packet.BufferSize := BufferSize;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
if ((Header.PacketID+1) <= Packets.Count) then // old packet
if (Packets.Items[Header.PacketID] = nil) then // not received yet
aaWriteToLog('SERVER received packet #'+IntToStr(Header.PacketID));
{$ENDIF}
     Packets := ServerSession.Packets.LockList;
     try
      if Packets.Count < (Header.PacketID + 1) then
        Packets.Count := Header.PacketID + 1; // List fills hole by Nil
      if (ServerSession.AnswerStatus <> ACRFullAnswer) then
        Packets.Items[Header.PacketID] := Packet
      else
       begin
        Dispose(Packet);
        Packet := nil;
       end;
     finally
      ServerSession.Packets.UnlockList;
     end;
     if Packet = nil then
       goto KillPacket;
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': server has added new packet');
{$ENDIF}
      if Terminated
      then
       begin
        Dispose(Packet);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog('           Terminated');
{$ENDIF}
        goto KillPacket;
       end;
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('SERVER has added new packet');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// test last packet
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Last Packet');
{$ENDIF}
     if (Header.ControlCode >= ACRLastPacket) then
      begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ACRLastPacket');
{$ENDIF}
       Header.ControlCode := Header.ControlCode - ACRLastPacket;
       if AllPacketsReceived then
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Yes');
{$ENDIF}
         ServerSession.AnswerStatus := ACRFullAnswer;  // Allow to extract buffer
        end
       else
        begin
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': No');
{$ENDIF}
         // start resend requesting
         ServerSession.AnswerStatus := ACRAnswerNotFull;
        end;
      end // Last Packet
     else // Not last Packet
       ServerSession.AnswerStatus := ACRStartAnswer;
// Full Answer
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': CurrentRequestID = '+IntToStr(ServerSession.CurrentRequestID));
{$ENDIF}
     if ServerSession.AnswerStatus = ACRFullAnswer then
       if Terminated then
         goto KillPacket
       else
         FManager.CommandReceived(ServerSession, Header.ControlCode);
{$IFDEF LOG_Server_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** END OF DATA PACKET *****');
aaWriteToLog('--------------------------------------------------------------');
{$ENDIF}
// End data packet process
     Exit;
    end // data packet process
   else
     raise EACRException.Create(40022, ErrorRUnknownControlCode, [Header.ControlCode]);
   end; // packet process
KillPacket:
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('--------------------------------------------------------------');
aaWriteToLog('SERVER KILLED PACKET # '+ IntToStr(Header.PacketID)+ ' with control code ' +IntToStr(Header.ControlCode));
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': FreeAndNilMem(Buffer)');
{$ENDIF}
  MemoryManager.FreeAndNilMem(Buffer);
 finally
//LeaveCriticalSection(FManager.FCSect);
{$IFDEF DEBUG_LOG_NETWORK}
aaWriteToLog('==============================================================');
aaWriteToLog('SERVER: '+IntToStr(GetTickCount)+' > Listener finished');
aaWriteToLog('==============================================================');
{$ENDIF}
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog('SERVER LISTENER THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISHED');
{$ENDIF}
 end;
 except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerListenerThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
    end;
{$ELSE}
  raise;
{$ENDIF}
 end;
end; // Execute

// TACRServerListenerThread



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerResendRequestThread.Create(
                       Manager:           TACRServerConnectionManager
                                            );
begin
  FManager := Manager;
  FManager.IncThreadCount;
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('Server RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerResendRequestThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('Server RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if FManager.ResendRequestThread <> nil then
    FManager.ResendRequestThread := nil;
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('Server RESENDING THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerResendRequestThread.Execute;
var
  i, j, k,
  PacketsCount:       Integer;
  Packets:            TList;
  AllPacketsReceived: Boolean;
  Sessions:           TList;
  ServerSession:      PACRSrvrSession;
  Header:             TACRPacketHeader;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** SERVER RESEND REQUEST START *****');
{$ENDIF}
  j:=0;
  repeat
    if Terminated then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sessions loop');
{$ENDIF}
      Exit;
     end;
    sleep(1 + (ACRServerRequestDelay div 8));
    Sessions := FManager.FSessions.LockList;
    try
     if j >= Sessions.Count then
      begin
       j := 0;
       continue;
      end;
     ServerSession := Sessions.Items[j];
     inc(ServerSession.Status);
     if ServerSession.ControlCode = ACRTerminate then
      begin
       dec(ServerSession.Status);
       continue;
      end;
     if ServerSession.AnswerStatus <> ACRAnswerNotFull then
      begin
       dec(ServerSession.Status);
       continue;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerResendRequestThread');
{$ENDIF}
    // Make PacketHeader
    Header.Signature := ACRServerPacketSign;
    Header.Recepient := ServerSession.RemoteClientID;
    EnterCriticalSection(FManager.FCSect);
    Header.Sender := FManager.FNetwork.FLocalClient;
    LeaveCriticalSection(FManager.FCSect);
    Header.ConnectionID := ServerSession.ConnectionID;
    Header.SessionID := ServerSession.Session.SessionID;
    Header.PacketID := 0;
    Header.ControlCode := ACRPacketResendRequest;
    Header.CurrentRequestID := ServerSession.CurrentRequestID;
    // Search absent packet and send request to resend it
    Packets := ServerSession.Packets.LockList;
    AllPacketsReceived := True;
    try
     PacketsCount := Packets.Count;
    finally
     ServerSession.Packets.UnlockList;
    end;
    dec(ServerSession.Status);
    for i:=0 to PacketsCount-1 do
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketsCount='+IntToStr(PacketsCount));
{$ENDIF}
      if Terminated then
       begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in packets loop');
{$ENDIF}
        Exit;
       end;
      if ServerSession.ControlCode = ACRTerminate then
        break;
      inc(ServerSession.Status);
      if ServerSession.Packets <> nil then
       begin
        Packets := ServerSession.Packets.LockList;
        if Packets.List = nil then
         begin
          ServerSession.Packets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Count <= i then
         begin
          ServerSession.Packets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Items[i] <> nil then
         begin
          ServerSession.Packets.UnlockList;
          dec(ServerSession.Status);
         end
        else // send request
         begin
          ServerSession.Packets.UnlockList;
          AllPacketsReceived := False;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketResendRequest, i='+IntToStr(i));
{$ENDIF}
          FManager.PacketResendRequest(@Header, FManager.FNetwork, ServerSession.RemoteHost, ServerSession.RemotePort, i);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been resent');
{$ENDIF}
          dec(ServerSession.Status);
          for k:=0 to (ACRServerRequestDelay div 2) do
           begin // Pause
            if Terminated then
             begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sleep loop');
{$ENDIF}
              Exit;
             end;
            Sleep(2);
           end; // Pause
         end;// send request
       end
      else
       dec(ServerSession.Status);
     end; // next packet
    if AllPacketsReceived then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
      EnterCriticalSection(FManager.FCSect);
      if ServerSession.ControlCode = ACRExecute then
        ServerSession.ControlCode := ACRSuspend;
      LeaveCriticalSection(FManager.FCSect);
      ServerSession.AnswerStatus := ACRFullAnswer;
     end;
    inc(j);
   until False; // Get new Session
 except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerResendRequestThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
    end;
{$ELSE}
  raise;
{$ENDIF}
 end; // except
end;// Execute



////////////////////////////////////////////////////////////////////////////////
//
// TACRServerMsgResendRequestThread
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRServerMsgResendRequestThread.Create(
                       Manager:           TACRServerConnectionManager
                                            );
begin
  FManager := Manager;
  FManager.IncThreadCount;
  inherited Create(False);
  Priority := tpNormal;
  FreeOnTerminate := True;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('Server MESSAGE RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - START');
{$ENDIF}
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRServerMsgResendRequestThread.Destroy;
begin
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('Server MESSAGE RESENDING THREAD #'+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+' - FINISH');
{$ENDIF}
  if FManager.MsgResendRequestThread <> nil then
    FManager.MsgResendRequestThread := nil;
  inherited Destroy;
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
aaWriteToLog('Server MESSAGE RESENDING THREAD - FINISHED');
{$ENDIF}
  FManager.DecThreadCount;
end; // Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRServerMsgResendRequestThread.Execute;
var
  i, j, k,
  PacketsCount:       Integer;
  Packets:            TList;
  AllPacketsReceived: Boolean;
  Sessions:           TList;
  ServerSession:      PACRSrvrSession;
  Header:             TACRPacketHeader;
begin
 try
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': ***** SERVER MESSAGE RESEND REQUEST START *****');
{$ENDIF}
  j:=0;
  repeat
    if Terminated then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sessions loop');
{$ENDIF}
      Exit;
     end;
    sleep(1 + (ACRServerRequestDelay div 8));
    Sessions := FManager.FSessions.LockList;
    try
     if j >= Sessions.Count then
      begin
       j := 0;
       continue;
      end;
     ServerSession := Sessions.Items[j];
     inc(ServerSession.Status);
     if ServerSession.ControlCode = ACRTerminate then
      begin
       dec(ServerSession.Status);
       continue;
      end;
     if ServerSession.AnswerStatus <> ACRAnswerNotFull then
      begin
       dec(ServerSession.Status);
       continue;
      end;
    finally
     FManager.FSessions.UnlockList;
    end;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerMsgResendRequestThread');
{$ENDIF}
    // Make PacketHeader
    Header.Signature := ACRServerPacketSign;
    Header.Recepient := ServerSession.RemoteClientID;
    EnterCriticalSection(FManager.FCSect);
    Header.Sender := FManager.FNetwork.FLocalClient;
    LeaveCriticalSection(FManager.FCSect);
    Header.ConnectionID := ServerSession.ConnectionID;
    Header.SessionID := ServerSession.Session.SessionID;
    Header.PacketID := 0;
    Header.ControlCode := ACRMessagePacketResendRequest;
    Header.CurrentRequestID := ServerSession.CurrentRequestID;
    // Search absent packet and send request to resend it
    Packets := ServerSession.MsgPackets.LockList;
    AllPacketsReceived := True;
    try
     PacketsCount := Packets.Count;
    finally
     ServerSession.MsgPackets.UnlockList;
    end;
    dec(ServerSession.Status);
    for i:=0 to PacketsCount-1 do
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': i='+IntToStr(i));
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketsCount='+IntToStr(PacketsCount));
{$ENDIF}
      if Terminated then
       begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketsCount='+IntToStr(PacketsCount));
{$ENDIF}
        Exit;
       end;
      if ServerSession.MsgControlCode = ACRTerminate then
        break;
      inc(ServerSession.Status);
      if ServerSession.MsgPackets <> nil then
       begin
        Packets := ServerSession.MsgPackets.LockList;
        if Packets.List = nil then
         begin
          ServerSession.MsgPackets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Count <= i then
         begin
          ServerSession.MsgPackets.UnlockList;
          dec(ServerSession.Status);
         end
        else if Packets.Items[i] <> nil then
         begin
          ServerSession.MsgPackets.UnlockList;
          dec(ServerSession.Status);
         end
        else // send request
         begin
          ServerSession.MsgPackets.UnlockList;
          AllPacketsReceived := False;
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': PacketMsgResendRequest, i='+IntToStr(i));
{$ENDIF}
          FManager.PacketResendRequest(@Header, FManager.FNetwork, ServerSession.RemoteHost, ServerSession.RemotePort, i, True);
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Packet has been resent');
{$ENDIF}
          dec(ServerSession.Status);
          for k:=0 to (ACRServerRequestDelay div 2) do
           begin // Pause
            if Terminated then
             begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': Terminated in sleep loop');
{$ENDIF}
              Exit;
             end;
            Sleep(2);
           end; // Pause
         end;// send request
       end
      else
       dec(ServerSession.Status);
     end; // next packet
    if AllPacketsReceived then
     begin
{$IFDEF DEBUG_LOG_NETWORK_RESENDING}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': AllPacketsReceived');
{$ENDIF}
      EnterCriticalSection(FManager.FCSect);
      if ServerSession.MsgControlCode = ACRExecute then
        ServerSession.MsgControlCode := ACRSuspend;
      LeaveCriticalSection(FManager.FCSect);
      ServerSession.MsgReceiveStatus := ACRFullAnswer;
//      FManager.MessageReceived(ServerSession);
     end;
    inc(j);
   until False; // Get new Session
 except
{$IFDEF DEBUG_LOG_NETWORK_ERRORS}
  on E: Exception do
    begin
aaWriteToLog('**************************************************************');
aaWriteToLog('ACRConnection> TACRServerMsgResendRequestThread.Execute - Error:');
aaWriteToLog(E.Message);
aaWriteToLog('Thread # '+IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle)));
aaWriteToLog('**************************************************************');
    end;
{$ELSE}
  raise;
{$ENDIF}
 end; // except
{$IFDEF LOG_SERVER_THREADS}
aaWriteToLog(IntToStr(Integer(self.ThreadID))+'/'+IntToStr(Integer(self.Handle))+': TACRServerSessionThread.Execute - START');
{$ENDIF}
end;// Execute

{$ENDIF} // Server


{$IFNDEF MsgCommunicator}
{IFDEF SERVER_VERSION}
{$ENDIF MsgCommunicator}
//------------------------------------------------------------------------------
// if authorization buffer is valid (encrypted by the same crypto settings as in CryptoInfo)
// then return true else return false
//------------------------------------------------------------------------------
function fnIsAuthorizationBufferValid(
                      CryptoInfo: TACRCryptoInfo;
                      Buffer:     PChar;
                      BufferSize: Integer
                                    ): Boolean;
var ms:             TACRMemoryStream;
    crc32, crc32_1: Cardinal;
    size:           Integer;
begin
  Result := True;
  if BufferSize = 0 then
   begin
    if CryptoInfo.CryptoAlgorithm <> 0 then
      Result:= False;
    Exit;
   end;
  ms := TACRMemoryStream.Create(Buffer, BufferSize);
  try
   try
    LoadDataFromStream(size,SizeOf(crc32),ms,11322);
    LoadDataFromStream(crc32,SizeOf(crc32),ms,11323);
    if (ms.Size - ms.Position <> size) then
     raise EACRException.Create(11325,ErrorLInvalidStreamSize,[ms.Size,(ms.Position+size)]);
   except
     Result := False;
     Exit;
   end;
    ACRDecryptBuffer(CryptoInfo,PChar(Buffer+ms.Position),size);
    crc32_1 := ACRCountCRC(0,PChar(Buffer+ms.Position),size);
    Result := (crc32 = crc32_1);
  finally
    ms.Buffer := nil;
    ms.Free;
  end;
end; // fnIsAuthorizationBufferValid
{$IFNDEF MsgCommunicator}
{ENDIF SERVER_VERSION}
{$ENDIF MsgCommunicator}

initialization

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRConnection> initialized');
{$ENDIF}

end.

{ TODO -oLeo : Move these default values to published properties}
(*
  FParams.RemoteHost := ACRDefaultHost;
  FParams.LocalHost := ACRDefaultHost;
{$IFDEF SERVER_VERSION}
  FParams.RemotePort := ACRDefaultClientPort;
  FParams.LocalPort := ACRDefaultServerPort;
{$ENDIF}
{$IFDEF CLIENT_VERSION}
  FParams.RemotePort := ACRDefaultServerPort;
  FParams.LocalPort := ACRDefaultClientPort;
{$ENDIF}
  FParams.PacketSize := ACRDefaultPacketSize;
*)
{ default values has been set - End of Leo's TODO}

{ TODO -oLeo : You must disable possibilities to change ConnectionParams while session is connected }

{ TODO -oLeo : You must disable possibilities to change LocalPort and ServerID while there is active session on the server}



