unit ACRMain;

interface

{$I ACRVer.inc}

uses SysUtils, Classes, DB, IniFiles,
     Dialogs,
{$IFDEF MSWINDOWS}
     Windows, Forms, Controls,
{$ENDIF}


{$IFDEF D6H}
{$IFDEF MSWINDOWS}
     DBCommon,
{$ENDIF}
     Variants,
     SqlTimSt,
{$ENDIF}




////////////////////////////////////////////////////////////////////////////////
//
//  Accuracer units
//
////////////////////////////////////////////////////////////////////////////////




{$IFDEF LOCAL_VERSION}
     ACRLocalEngine,
{$ENDIF}

{$IFNDEF D5H}
     ACRD4Routines,
{$ENDIF}

{$IFDEF DEBUG_MEMCHECK}
     MemCheck,
{$ENDIF}

{$IFDEF CLIENT_VERSION}
     ACRClient,
{$ENDIF}
     ACRBase,
{$IFNDEF SQLMEMTABLE}
     ACRCrypto,
{$ENDIF}
     ACRCompression,
     ACRTypes,
     ACRVariant,
     //ACRConverts,
     ACRExcept,
     ACRSQLProcessor,
     ACRLexer,
     ACRConst,
 {$IFDEF DEBUG_LOG}
     ACRDebug,
 {$ENDIF}
     ACRMemory;       // UNIT ACRMemory MUST BE LAST !!!


const
  // TACRDataset flags
  dbfOpened     = 0;
  dbfPrepared   = 1;
  dbfExecSQL    = 2;
  dbfTable      = 3;
  dbfFieldList  = 4;
  dbfIndexList  = 5;
  dbfStoredProc = 6;
  dbfExecProc   = 7;
  dbfProcDesc   = 8;
  dbfDatabase   = 9;
  dbfProvider   = 10;

type



  // forward declarations
  TACRSession = class;
  TACRDatabase = class;
  TACRDataSet = class;
  TACRTable = class;
  TACRQuery = class;
  TACRAdvFieldDefs = class;
  TACRForeignKeyDefs = class;
  TACRForeignKeyDef = class;
  TACRArrayOfTACRVariant = array of TACRVariant;

  TACRBackupInfo = record
   Date:                TDateTime;
   FileSize:            Int64;
   Description:         String;
   Encrypted:           Boolean;
   EncryptedByPassword: Boolean;
   TableCount:          Integer;
  end;
  
  TCompressionAlgorithm = (caNone,caZLIB,caBZIP,caPPM);
  TACRDatabaseOperation = (dbopCompact,dbopRepair,dbopCopy,dbopChangeDatabaseSettings,dbopBackup,dbopRestore);
  TACRBatchMode = (batAppend, batUpdate, batAppendUpdate, batCopy, batDelete);
  TACRTableOperation = (tbopImport,tbopExport,tbopCopy,
                        tbopRestructure,tbopRepair,
                        tbopBatchAppend, tbopBatchUpdate, tbopBatchAppendUpdate,
                        tbopBatchCopy, tbopBatchDelete
                        );

  TACRDatabaseProgress = procedure (
                                      Sender:     TComponent;
                                      Progress:   Double;
                                      Operation:  TACRDatabaseOperation;
                                      var Abort:  Boolean
                                   ) of object;

  TACRTableProgress = procedure (
                                      Sender:     TComponent;
                                      Progress:   Double;
                                      Operation:  TACRTableOperation;
                                      var Abort:  Boolean
                                   ) of object;

  TACROnClientReceiveTextMessage = procedure (const Text: String)  of object;
  TACROnClientReceiveBinaryMessage = procedure (Buffer: PChar; Size: Integer) of object;
  TACROnClientReceiveStreamMessage = procedure (Stream: TStream) of object;

  TACROnError = procedure (
                       Sender:             TComponent;
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                     ) of object;

  TACRBeforeInsertRecord = procedure (
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant;
                       var Abort:          Boolean
                     ) of object;
  TACRAfterInsertRecord = procedure (
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant
                     ) of object;

  TACRBeforeUpdateRecord = procedure (
                       Sender:                TACRDataset;
                       const TableName:       String;
                       const OldFieldValues:  TACRArrayOfTACRVariant;
                       const NewFieldValues:  TACRArrayOfTACRVariant;
                       var Abort:             Boolean
                     ) of object;
  TACRAfterUpdateRecord = procedure (
                       Sender:                TACRDataset;
                       const TableName:       String;
                       const OldFieldValues:  TACRArrayOfTACRVariant;
                       const NewFieldValues:  TACRArrayOfTACRVariant
                     ) of object;

  TACRBeforeDeleteRecord = procedure (
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant;
                       var Abort:          Boolean
                     ) of object;
  TACRAfterDeleteRecord = procedure (
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant
                     ) of object;

  TACRBeforeExecuteSQL = procedure (
                       Sender:             TACRQuery;
                       var Abort:          Boolean
                     ) of object;
  TACRAfterExecuteSQL = procedure (
                       Sender:             TACRQuery
                     ) of object;


  TACRCryptoAlgorithm = (
                        craNone
{$IFDEF ENCRYPTION_ON}
                        ,
                        craRijndael_128,
                        craRijndael_256,
                        craBlowfish,
                        craTwofish_128,
                        craTwofish_256,
                        craSquare,
                        craDES_Single_8,
                        craDES_Double_8,
                        craDES_Double_16,
                        craDES_Triple_8,
                        craDES_Triple_16,
                        craDES_Triple_24
{$ENDIF}
                        );
  TACRCryptoMode = (acmCTS,acmCBC,acmCFB,acmOFB);


////////////////////////////////////////////////////////////////////////////////
//
// TACRBlobStream
//
////////////////////////////////////////////////////////////////////////////////


  TACRBlobStream = class (TACRStream)
  private
    FBlobStream: TACRStream; 
    FField:      TBlobField;
    FDataSet:    TACRDataSet;
    FWideMemo:   Boolean;
   protected
    // sets new size of the stream
    procedure InternalSetSize(const NewSize: Int64);
    // sets new size of the stream
    procedure SetSize(NewSize: Longint);
    {$IFDEF D6H}
      overload;
    {$ENDIF}
      override;
    {$IFDEF D6H}
    procedure SetSize(const NewSize: Int64); overload; override;
    {$ENDIF}
   public
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint;
    {$IFDEF D6H}
            overload;
    {$ENDIF}
      override;
    {$IFDEF D6H}
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; overload; override;
    {$ENDIF}
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    procedure Truncate;
  public
   // Blob stream interface
  end; // TACRBlobStream



////////////////////////////////////////////////////////////////////////////////
//
// TACRSessionList
//
////////////////////////////////////////////////////////////////////////////////

  // global list of sessions
  TACRSessionList = class(TObject)
   private
    FSessions: TThreadList;
    FSessionNumbers: TBits;

    // adds session to list
    procedure AddSession(ASession: TACRSession);
    // closes all sessions
    procedure CloseAll;
    // Gets sessions count
    function GetCount: Integer;
    // Gets session by No
    function GetSession(Index: Integer): TACRSession;
    // gets current session
    function GetCurrentSession: TACRSession;
    // Gets session by Name
    function GetSessionByName(const SessionName: string): TACRSession;
    // Sets current session
    procedure SetCurrentSession(Value: TACRSession);

   public
    constructor Create;
    destructor Destroy; override;
    // Finds session by name
    function FindSession(const SessionName: string): TACRSession;
    // Gets list of sessions names
    procedure GetSessionNames(List: TStrings);
    // Opens session by name
    function OpenSession(const SessionName: string): TACRSession;

    property Count: Integer read GetCount;
    property CurrentSession: TACRSession read GetCurrentSession write SetCurrentSession;
    property Sessions[Index: Integer]: TACRSession read GetSession; default;
    property List[const SessionName: string]: TACRSession read GetSessionByName;
  end;


////////////////////////////////////////////////////////////////////////////////
//
// TACRSession
//
////////////////////////////////////////////////////////////////////////////////

  TACRDatabaseEvent = (dbOpen, dbClose, dbAdd, dbRemove);
  TACRDatabaseNotifyEvent = procedure(DBEvent: TACRDatabaseEvent; const Param) of object;
  TACRPasswordEvent = procedure(Sender: TObject; var Continue: Boolean) of Object;

  // TSession replacement for thread-safe use
  TACRSession = class(TComponent)
  private
    FHandle: TACRSessionComponentManager;
    FDatabases: TList;
    FStreamedActive: Boolean;
    FKeepConnections: Boolean;
    FDefault: Boolean;
    FAutoSessionName: Boolean;
    FUpdatingAutoSessionName: Boolean;
    FSessionName: string;
    FSessionNumber: Integer;
    FLockCount: Integer;

    FOnDBNotify: TACRDatabaseNotifyEvent;
    FOnStartup: TNotifyEvent;

    // adds database
    procedure AddDatabase(Value: TACRDatabase);
    // raises exception if active
    procedure CheckInactive;
    // sends notification
    procedure DBNotification(DBEvent: TACRDatabaseEvent; const Param);
    // finds database with specified owner
    function DoFindDatabase(const DatabaseName: string; AOwner: TComponent): TACRDatabase;
    // opens database (thread-safe)
    function DoOpenDatabase(const DatabaseName: string;
                            AOwner: TComponent): TACRDatabase;
    // find DB manager by db name
    function FindDatabaseHandle(const DatabaseName: string): TACRBaseSession;
    // session is active?
    function GetActive: Boolean;
    // gets database by No
    function GetDatabase(Index: Integer): TACRDatabase;
    // gets count of connected databases
    function GetDatabaseCount: Integer;
    // gets handle
    function GetHandle: TACRSessionComponentManager;
    // not auto-session?
    function SessionNameStored: Boolean;
    // makes session current
    procedure MakeCurrent;
    // removes database from list
    procedure RemoveDatabase(Value: TACRDatabase);
    // opens session
    procedure SetActive(Value: Boolean);
    // sets auto-session name
    procedure SetAutoSessionName(Value: Boolean);
    // sets the name of session
    procedure SetSessionName(const Value: string);
    // sets session name to datasets and databases
    procedure SetSessionNames;
    // starts session
    procedure StartSession(Value: Boolean);
    // updates auto-session name
    procedure UpdateAutoSessionName;
    // auto-session name is valid?
    procedure ValidateAutoSession(AOwner: TComponent; AllSessions: Boolean);

  protected
    // loaded
    procedure Loaded; override;
    // send notification to datasets and databases
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property OnDBNotify: TACRDatabaseNotifyEvent read FOnDBNotify write FOnDBNotify;
    // set name of component
    procedure SetName(const NewName: TComponentName); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // locks session
    procedure LockSession;
    // unlocks session
    procedure UnlockSession;
    // closes session
    procedure Close;
    // closes database
    procedure CloseDatabase(Database: TACRDatabase);
    // drops all connections
    procedure DropConnections;
    // finds database by name
    function FindDatabase(const DatabaseName: string): TACRDatabase;
    // get list of database names
    procedure GetDatabaseNames(List: TStrings);
    // get list of database tables
    procedure GetTableNames(const DatabaseName: string; List: TStrings);
    // opens session
    procedure Open;
    // opens database
    function OpenDatabase(const DatabaseName: string): TACRDatabase;

    property DatabaseCount: Integer read GetDatabaseCount;
    property Databases[Index: Integer]: TACRDatabase read GetDatabase;
    property Handle: TACRSessionComponentManager read GetHandle;

  published
    property Active: Boolean read GetActive write SetActive default False;
    property AutoSessionName: Boolean read FAutoSessionName write SetAutoSessionName default False;
    property KeepConnections: Boolean read FKeepConnections write FKeepConnections default True;
    property SessionName: string read FSessionName write SetSessionName stored SessionNameStored;
    property OnStartup: TNotifyEvent read FOnStartup write FOnStartup;
  end;

////////////////////////////////////////////////////////////////////////////////
//
// TACRDataset
//
////////////////////////////////////////////////////////////////////////////////


 TACRDBFlags = set of 0..15;
 TACRDataset = class (TDataset)
  private
   FCurrentVersion:                 String;
   FHandle:                         TACRCursor;
   FSessionName:                    String;
   FFilterBuffer:                   TACRRecordBuffer; // filter record buffer
   FIndexFieldCount:                Integer;
   FIndexFieldMap:                  array of Word;
   FKeySize:                        Integer;
   FDBFlags:                        TACRDBFlags;
   FDatabase:                       TACRDatabase;
   FDatabaseName:                   String;
   FInMemory:                       Boolean;
   FReadOnly:                       Boolean;
   FStoreDefs:                      Boolean;  // for FFieldDefs
   FEditRecordBuffer:               TACRRecordBuffer; // for storing record on edit
   FACRConstraintDefs:              TACRConstraintDefs;  // Constraint definitions
   FExternalHandle:                 TACRCursor;
   FInsertOrEditComplete:           Boolean; // set to True in internal post
   FIsTable:                        Boolean; // True if TACRTable, False if TACRQuery
   FIsProjectionSet:                Boolean;
   FLogComponent:                   TACRDataset; // Table or Query for INSERT, UPDATE, DELETE record event in TACRDatabase
   FOldFieldValues:                 TACRArrayOfTACRVariant; // for update / delete
   FNewFieldValues:                 TACRArrayOfTACRVariant; // for update / insert
   FValuesChangedByEventHandler:    Boolean;
{$IFDEF D6H}
   FOnUpdateRecord:                 TUpdateRecordEvent;
{$ENDIF}
 protected
   FIndexDefs:                      TIndexDefs; // index definitions
   FACRFieldDefs:                   TACRFieldDefs; // fields definitions
   FAdvIndexDefs:                   TACRIndexDefs; // index definitions
   FAdvFieldDefs:                   TACRAdvFieldDefs; // USER fields definitions
   FForeignKeyDefs:                 TACRForeignKeyDefs; // foreign key definitions
   FRestructureIndexDefs:           TIndexDefs; // restructure index definitions
   FRestructureFieldDefs:           TACRAdvFieldDefs; // restructure field definitions
   FRestructureForeignKeyDefs:      TACRForeignKeyDefs; // restructure foreign key definitions
   FKeyBuffers:                     array[TACRKeyIndex] of TACRRecordBuffer;
   FKeyBuffer:                      TACRRecordBuffer;
   FDeleteRecordFlag:               Boolean;
{$IFDEF D6H}
  protected
   // IProviderSupport
   function PSGetUpdateException(E: Exception; Prev: EUpdateError): EUpdateError; override;
   function PSIsSQLSupported: Boolean; override;
   procedure PSReset; override;
   function PSUpdateRecord(UpdateKind: TUpdateKind; Delta: TDataSet): Boolean; override;
  protected
   // IProviderSupport
   procedure PSEndTransaction(Commit: Boolean); override;
   procedure FixParams(Query: TDataset);
   function FixNames(QueryText: String): String;
   function PSExecuteStatement(const ASQL: string; AParams: TParams;
      ResultSet: Pointer = nil): Integer; override;
   procedure PSGetAttributes(List: TList); override;
   function PSGetQuoteChar: string; override;
   function PSInTransaction: Boolean; override;
   function PSIsSQLBased: Boolean; override;
   procedure PSStartTransaction; override;
{$ENDIF}
  protected
   function InitKeyBuffer(Buffer: TACRRecordBuffer): TACRRecordBuffer;
   procedure AllocKeyBuffers;
   procedure FreeKeyBuffers;
   // field defs support
   function FieldDefsStored: Boolean;
   // index defs support
   function IndexDefsStored: Boolean;
   // set index definitions
   procedure SetIndexDefs(Value: TIndexDefs);
   // get active buffer
   function GetActiveRecordBuffer: PChar;
   procedure CheckDBSessionName;
   function GetDBHandle: TACRBaseSession;
   function GetDBSession: TACRSession;
   procedure CheckInMemoryDatabaseName;
   procedure SetDatabaseName(const Value: string);
  public
   procedure ChangeCurrentDatabaseName(const Value: string);
  protected
   procedure SetSessionName(const Value: string);
   procedure SetInMemory(const Value: Boolean);
   function GetCurrentVersion: String;
  protected
   procedure OpenCursor(InfoQuery: Boolean); override;
   procedure CloseCursor; override;
   procedure Disconnect; virtual;
   procedure SetDBFlag(Flag: Integer; Value: Boolean); virtual;
   function CreateHandle: TACRCursor; virtual;
   procedure DestroyHandle; virtual;

   procedure InternalDataConvert(Field: TField; Source, Dest: Pointer; ToNative: Boolean);

   function IsWideMemoField(Field: TField): Boolean;
   procedure DataConvert(Field: TField; Source, Dest: Pointer; ToNative: Boolean);
{$IFDEF D5H}
    override;
{$ENDIF}
  public
   procedure SetWideMemoField(Field: TField; Value: WideString);
   function GetWideMemoField(Field: TField): WideString;
  protected
   procedure SetActive(Value: Boolean); override;

   //---------------------------------------------------------------------------
   // indexes and ranges
   //---------------------------------------------------------------------------

   procedure SwitchToIndex(const IndexName: string);
   function GetIsIndexField(Field: TField): Boolean; override;
   procedure GetIndexInfo;
   function ResetCursorRange: Boolean;

   //---------------------------------------------------------------------------
   // navigation & bookmark methods
   //---------------------------------------------------------------------------

   // clear calculated fields
   procedure ClearCalcFields(Buffer: PChar); override;
   procedure InternalRefresh; override;
   function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
  public
   function GetCurrentRecord(Buffer: PChar): Boolean; override;
   function GetCurrentRecordID: TACRRecordID;
  protected
   // return record count
   function GetRecordCount: Integer; override;
   // go to record
   procedure SetRecNo(Value: Integer); override;
   // return current record number
   function GetRecNo: Integer; override;
   // go to first record
   procedure InternalFirst; override;
   // go to last record
   procedure InternalLast; override;
   // go to record in buffer
   procedure InternalSetToRecord(Buffer: PChar); override;
   // get bookmark flag
   function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
   // get bookmark data
   procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
   // go to bookmark
   procedure InternalGotoBookmark(Bookmark: Pointer); override;
   // set flag
   procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
   // set data
   procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
  public
   // compare bookmarks
   function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
   // checks if bookmark is valid
   function BookmarkValid(Bookmark: TBookmark): Boolean; override;
  protected

   //---------------------------------------------------------------------------
   // Filters and search
   //---------------------------------------------------------------------------

   // for OnFilterRecord Event
   function IsOnFilterRecordApplied: Boolean;
  public
   function InternalFilterRecord(Buffer: TACRRecordBuffer): Boolean;
   function FilterRecord(Buffer: TACRRecordBuffer; Dataset: Pointer): Boolean;
  protected
   procedure SetOnFilterRecord(const Value: TFilterRecordEvent); override;
   function IsIndexApplied: Boolean;
   procedure PrepareCursor; virtual;
   function GetCanModify: Boolean; override;
  public
   // set SQL Filter
   procedure SetSQLFilter(FilterExpr: TObject);
   // apply projection
   procedure ApplyProjection(FieldNamesList, AliasList: TStringList);
   // FindFirst, FindNext, Filters
   procedure ActivateFilters;
   procedure DeactivateFilters;
   procedure SetFilterData(const Text: string; Options: TFilterOptions);
   procedure SetFiltered(Value: Boolean); override;
   procedure SetFilterOptions(Value: TFilterOptions); override;
   procedure SetFilterText(const Value: string); override;
   function FindRecord(Restart, GoForward: Boolean): Boolean; override;
   function LocateRecord(
                         const KeyFields: String;
                         const KeyValues: Variant;
                         Options:         TLocateOptions
                        ): Boolean;
  public
   function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
   function Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant; override;

   //---------------------------------------------------------------------------
   // insert, edit, post, delete methods
   //---------------------------------------------------------------------------
  protected
   procedure InitRecord(Buffer: PChar); override;
   // appending table (Append flag - ignored, record will be inserted at first empty position)
   procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
   // insert record
   procedure InternalInsert; override;
   // edit record
   procedure InternalEdit; override;
   // cancels updates
   procedure InternalCancel; override;
   // update record
   procedure InternalPost; override;
   // delete record
   procedure InternalDelete; override;

   //---------------------------------------------------------------------------
   //  insert, update, delete record events of TACRDababase
   //---------------------------------------------------------------------------
   procedure CreateOldNewFieldValues;
   procedure FreeOldNewFieldValues;
   procedure SaveValues(var Values: TACRArrayOfTACRVariant);
   procedure ClearValues(var Values: TACRArrayOfTACRVariant);
  public
   function IsBeforeInsertRecordAssigned: Boolean;
   function IsAfterInsertRecordAssigned: Boolean;
   function IsBeforeUpdateRecordAssigned: Boolean;
   function IsAfterUpdateRecordAssigned: Boolean;
   function IsBeforeDeleteRecordAssigned: Boolean;
   function IsAfterDeleteRecordAssigned: Boolean;
   function IsBeforeExecuteSQLAssigned: Boolean;
   function IsAfterExecuteSQLAssigned: Boolean;
  protected 
   function DoBeforeInsertRecord: Boolean;
   procedure DoAfterInsertRecord;
   function DoBeforeUpdateRecord: Boolean;
   procedure DoAfterUpdateRecord;
   function DoBeforeDeleteRecord: Boolean;
   procedure DoAfterDeleteRecord;
  public
   //---------------------------------------------------------------------------
   //  methods for fast executing DELETE and UPDATE SQL statements
   //---------------------------------------------------------------------------
   procedure DeleteVisibleRecords;
   procedure UpdateVisibleRecords(
                                  FieldNames:   TStringList;
                                  values:       array of TACRVariant
                                 );
  protected
   //---------------------------------------------------------------------------
   // open, close methods
   //---------------------------------------------------------------------------

   procedure InternalHandleException; override;
   function IsCursorOpen: Boolean; override;
   procedure InternalOpen; override;
   procedure InternalClose; override;
   procedure InternalInitFieldDefs; override;


   //---------------------------------------------------------------------------
   // general methods
   //---------------------------------------------------------------------------

   // allocate record buffer
   function AllocRecordBuffer: PChar; override;
   // free record buffer
   procedure FreeRecordBuffer(var Buffer: PChar); override;
   // initialize record buffer
   procedure InternalInitRecord(Buffer: PChar); override;
   // return record size in bytes
   function GetRecordSize: Word; override;
   // return true if range is applied
   function IsRangeApplied: Boolean;
   // return true if distinct is applied
   function IsDistinctApplied: Boolean;

  protected
   property DBFlags: TACRDBFlags read FDBFlags;
  public
   constructor Create(AOwner: TComponent); override;
   destructor Destroy; override;
   function OpenDatabase: TACRDatabase;
   procedure CloseDatabase(Database: TACRDatabase);
  protected
   function GetCalcFieldNo(Field: TField): Integer;
  public
   // read field data to current record buffer
   function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
  protected
   // write field data from buffer to current record buffer
   procedure SetFieldData(Field: TField; Buffer: Pointer); override;
  public
   procedure GetFieldValue(Value: TACRVariant; FieldNo: Integer; DirectAccess: Boolean; CopyFlag: Boolean = True);
   procedure SetFieldValue(Value: TACRVariant; FieldNo: Integer; DirectAccess: Boolean);
   procedure LockTable;
   procedure UnlockTable;

   // create Blob stream
  private
   function InternalCreateBlobStream(
    					Field: TField;
              Mode: TBlobStreamMode
              ): TACRStream;
  public
   // create TACRBlobStream
   function CreateBlobStream(
    					Field: TField;
              Mode: TBlobStreamMode
              ): TStream; override;
    // close Blob stream, write Blob field value to Blob data file
    procedure CloseBlob(Field: TField); override;
  public
   // Table or Query for INSERT, UPDATE, DELETE record event in TACRDatabase
   property LogComponent: TACRDataset read FLogComponent write FLogComponent;
   // for INSERT, UPDATE events
   property ValuesChangedByEventHandler: Boolean read FValuesChangedByEventHandler write FValuesChangedByEventHandler;
   property IsTable: Boolean read FIsTable write FIsTable;
   property Handle: TACRCursor read FHandle;
   property Database: TACRDatabase read FDatabase;
//   property DBHandle: TACRBaseSession read GetDBHandle;
   property DBSession: TACRSession read GetDBSession;
   // index definitions, used by CreateTable;
   property IndexDefs: TIndexDefs read FIndexDefs write SetIndexDefs stored IndexDefsStored;
   // field definitions, used by CreateTable;
   property FieldDefs stored FieldDefsStored;
   // index definitions, used by CreateTable;
   property AdvIndexDefs: TACRIndexDefs read FAdvIndexDefs;
   // field definitions, used by CreateTable;
   property AdvFieldDefs: TACRAdvFieldDefs read FAdvFieldDefs;
   property ForeignKeyDefs: TACRForeignKeyDefs read FForeignKeyDefs;
   property KeySize: Integer read FKeySize;
  published
   property CurrentVersion: String read GetCurrentVersion write FCurrentVersion;
   // fielddefs support
   property StoreDefs: Boolean read FStoreDefs write FStoreDefs default False;
 {$IFDEF SQLMEMTABLE}
  public
 {$ENDIF}
   property DatabaseName: String read FDatabaseName write SetDatabaseName;
   property SessionName: String read FSessionName write SetSessionName;
   property InMemory: Boolean read FInMemory write SetInMemory;
 {$IFDEF SQLMEMTABLE}
  published
 {$ENDIF}
   property ReadOnly: Boolean read FReadOnly write FReadOnly;

   property Active;
   property AutoCalcFields;
   property Filter;
   property Filtered;
   property FilterOptions;
   property BeforeOpen;
   property AfterOpen;
   property BeforeClose;
   property AfterClose;
   property BeforeInsert;
   property AfterInsert;
   property BeforeEdit;
   property AfterEdit;
   property BeforePost;
   property AfterPost;
   property BeforeCancel;
   property AfterCancel;
   property BeforeDelete;
   property AfterDelete;
   property BeforeScroll;
   property AfterScroll;
{$IFDEF D5H}
   property BeforeRefresh;
   property AfterRefresh;
{$ENDIF}
   property OnCalcFields;
   property OnDeleteError;
   property OnEditError;
   property OnFilterRecord;
   property OnNewRecord;
   property OnPostError;
{$IFDEF D6H}
   property OnUpdateRecord: TUpdateRecordEvent read FOnUpdateRecord write FOnUpdateRecord;
{$ENDIF}
 end; // TACRDataset


////////////////////////////////////////////////////////////////////////////////
//
// TACRTable
//
////////////////////////////////////////////////////////////////////////////////


 TACRTable = class (TACRDataset)
  private
   FOnProgress:         TACRTableProgress;
   FTableName:          String;
   FExclusive:          Boolean;
   FTemporary:          Boolean;
   FIndexName:          String;
   FFieldsIndex:        Boolean;
   FMasterLink:         TMasterDataLink;
   FMemoryTableAllocBy: Integer;

{$IFDEF D6H}
  protected
    // IProviderSupport
    function PSGetDefaultOrder: TIndexDef; override;
    function PSGetKeyFields: string; override;
    function PSGetTableName: string; override;
    function PSGetIndexDefs(IndexTypes: TIndexOptions): TIndexDefs; override;
    procedure PSSetCommandText(const CommandText: string); override;
    procedure PSSetParams(AParams: TParams); override;
{$ENDIF}
  private
   procedure CheckBlankTableName;
   procedure SetTemporary(const Value: Boolean);
   function GetIndexFieldNames: string;
   function GetIndexName: string;
   procedure GetIndexParams(const IndexName: string; FieldsIndex: Boolean;
         var IndexedName: string);
   function IndexDefsStored: Boolean;
   procedure SetIndex(const Value: string; FieldsIndex: Boolean);
   procedure SetIndexFieldNames(const Value: string);
   procedure SetIndexName(const Value: string);
   function CreateCursor: TACRCursor;

 protected
   // progress event
   procedure DoOnProgress(
                          Progress:   Double;
                          Operation:  TACRTableOperation;
                          var Abort:  Boolean
                         );
  public
   procedure GetIndexNames(List: TStrings);
    // return index name of the index or '' if not found
   function FindIndex(FieldNamesList, AscDescList, CaseSensitivityList: TStringList): String;
    // return index name of the index
   function FindOrCreateIndex(FieldNamesList, AscDescList, CaseSensitivityList: TStringList): String;
   // set distinct
   procedure ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList);
   procedure GetTableNames(List: TStrings);
  private

   function GetTableExists: Boolean;
   //---------------------------- master-detail --------------------------------
   procedure CheckMasterRange;
   procedure UpdateRange;
   procedure MasterChanged(Sender: TObject);
   procedure MasterDisabled(Sender: TObject);
   procedure SetDataSource(Value: TDataSource);
   function GetMasterFields: string;
   procedure SetMasterFields(const Value: string);

  protected
   procedure PrepareCursor; override;
   function CreateHandle: TACRCursor; override;
   procedure DataEvent(Event: TDataEvent; Info: Longint); override;
   procedure DefChanged(Sender: TObject); override;
   procedure InitFieldDefs; override;
   procedure DestroyHandle; override;
   function GetHandle: TACRCursor;
   procedure UpdateIndexDefs; override;
   function GetIndexField(Index: Integer): TField;
   procedure SetIndexField(Index: Integer; Value: TField);
   procedure SetTableName(Value: String);

   //---------------------------- master-detail --------------------------------
   procedure SetLinkRanges(MasterFields: TList);
   function GetDataSource: TDataSource; override;
   procedure DoOnNewRecord; override;
   function GetIndexFieldCount: Integer;
  protected
   property MasterLink: TMasterDataLink read FMasterLink;
  public
   constructor Create(AOwner: TComponent); override;
   destructor Destroy; override;
   procedure CreateTable;
   procedure DeleteTable(Cascade: Boolean = False);
   procedure EmptyTable;
   procedure RenameTable(NewTableName: String);
   procedure LoadTableFromFile(
                        FileName: String
                       );
   procedure SaveTableToFile(
                        FileName: String;
                        CompressionAlgorithm: TCompressionAlgorithm = caNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
   procedure LoadTableFromStream(
                        Stream: TStream
                       );
   procedure SaveTableToStream(
                        Stream: TStream;
                        CompressionAlgorithm: TCompressionAlgorithm = caNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
   procedure LoadAllTablesFromFile(
                        FileName: String
                       );
   procedure SaveAllTablesToFile(
                        FileName: String;
                        CompressionAlgorithm: TCompressionAlgorithm = caNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
   procedure LoadAllTablesFromStream(
                        Stream: TStream
                       );
   procedure SaveAllTablesToStream(
                        Stream: TStream;
                        CompressionAlgorithm: TCompressionAlgorithm = caNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
   function ImportTable(
                          SourceTable: TDataset;
                          var Log:     String
                        ): Boolean; overload;
   function ImportTable(SourceTable: TDataset): Boolean; overload;
   function ExportTable(
                          DestinationTable:   TDataset;
                          CreateTablePointer: TProcedure;
                          var Log:            String
                       ): Boolean; overload;
   function ExportTable(
                          DestinationTable:   TDataset;
                          CreateTablePointer: TProcedure
                       ): Boolean; overload;
   function ExportTableToSQL(
                              ExportStructure:      Boolean = True;
                              AddDropTableCommand:  Boolean = True;
                              ExportIndexes:        Boolean = True;
                              AddDropIndexCommand:  Boolean = False;
                              ExportData:           Boolean = True;
                              ExportBLOBFields:     Boolean = True;
                              UseBracketsForNames:  Boolean = False
                            ): String;
   function BatchMove(ASource: TACRDataSet; AMode: TACRBatchMode): Int64;
   function RestructureTable(
                          var Log:            String
                       ): Boolean; overload;
   function RestructureTable: Boolean; overload;
   // repair table
  protected
   function InternalRepairOrRestructureTable(
                            Repair:       Boolean;
                            var Log:      String;
                            NewDatabase:  TACRDatabase = nil;
                            FKDefs:       TACRForeignKeyDefs = nil
                            ): Boolean;
  public
   function RepairTable(UseLowLevelTableAccess: Boolean = False): Boolean; overload;
   function RepairTable(var Log:                String;
                        UseLowLevelTableAccess: Boolean = False;
                        NewDatabase:            TACRDatabase = nil): Boolean; overload;

   // Rename Field by Name
   procedure RenameField(FieldName, NewFieldName: String);
   procedure AddForeignKey(ForeignKeyDef: TACRForeignKeyDef);
   procedure DeleteConstraint(Name: String; Cascade: Boolean = False);

   procedure AddIndex(
              const Name,
              Fields: String;
              Options: TIndexOptions;
              const DescFields: String = '';
              const CaseInsFields: String = ''
                     );
   procedure DeleteIndex(const Name: string);
   procedure DeleteAllIndexes;

   //---------------------------------------------------------------------------
   // key and range methods
   //---------------------------------------------------------------------------
  protected
    procedure CheckSetKeyMode;
    function GetKeyBuffer(KeyIndex: TACRKeyIndex): TACRRecordBuffer;
    function GetKeyExclusive: Boolean;
    function GetKeyFieldCount: Integer;
    procedure SetKeyExclusive(Value: Boolean);
    procedure SetKeyFieldCount(Value: Integer);
    procedure SetKeyBuffer(KeyIndex: TACRKeyIndex; Clear: Boolean);
    procedure SetKeyFields(KeyIndex: TACRKeyIndex; const Values: array of const);
    procedure PostKeyBuffer(Commit: Boolean);
  public
   function FindKey(const KeyValues: array of const): Boolean;
   procedure FindNearest(const KeyValues: array of const);
   function GotoKey: Boolean;
   procedure GotoNearest;
   procedure EditKey;
   procedure SetKey;

   function SetCursorRange: Boolean;
   procedure ApplyRange;
   procedure CancelRange;
   procedure EditRangeStart;
   procedure EditRangeEnd;
   procedure SetRange(const StartValues, EndValues: array of const);
   procedure SetRangeStart;
   procedure SetRangeEnd;

   procedure Post; override;

   // return LastAutoincValue for Field (FieldIndex started from 0)
   function LastAutoincValue(FieldIndex: Integer): Int64; overload;
   // return LastAutoincValue for Field
   function LastAutoincValue(FieldName: String): Int64; overload;
   // set LastAutoincValue for Field (FieldIndex started from 0)
   procedure SetLastAutoincValue(Value: Int64; FieldIndex: Integer); overload;
   // set LastAutoincValue for Field
   procedure SetLastAutoincValue(Value: Int64; FieldName: String); overload;

  public
   property IndexFieldCount: Integer read GetIndexFieldCount;
   property IndexFields[Index: Integer]: TField read GetIndexField write SetIndexField;
   property KeyExclusive: Boolean read GetKeyExclusive write SetKeyExclusive;
   property KeyFieldCount: Integer read GetKeyFieldCount write SetKeyFieldCount;
   property Temporary: Boolean Read FTemporary Write SetTemporary;
   // index definitions, used by RestructureTable;
   property RestructureIndexDefs: TIndexDefs read FRestructureIndexDefs;
   // field definitions, used by RestructureTable;
   property RestructureFieldDefs: TACRAdvFieldDefs read FRestructureFieldDefs;
   // foreign key definitions, used by RestructureTable;
   property RestructureForeignKeyDefs: TACRForeignKeyDefs read FRestructureForeignKeyDefs;
  published
   // fielddefs support
   property StoreDefs;
   // index definitions
   property IndexDefs: TIndexDefs read FIndexDefs write SetIndexDefs
              stored IndexDefsStored;
   property IndexFieldNames: String read GetIndexFieldNames write SetIndexFieldNames;
   property IndexName: String read GetIndexName write SetIndexName;
   // field definitions
   property FieldDefs stored FieldDefsStored;
   property TableName: string Read FTableName Write SetTableName;
{public}
   property Exclusive: Boolean read FExclusive write FExclusive;
{published}
   property Exists: Boolean read GetTableExists;
   property MasterFields: string read GetMasterFields write SetMasterFields;
   property MasterSource: TDataSource read GetDataSource write SetDataSource;
   property MemoryTableAllocBy: Integer read FMemoryTableAllocBy write FMemoryTableAllocBy;
   property OnProgress: TACRTableProgress read FOnProgress write FOnProgress;
 end; // TACRTable


TACRQuery = class (TACRDataset)
  private
    FStmtHandle:        TACRSQLProcessor;
    FSQL:               TStrings;
    FPrepared:          Boolean;
    FParams:            TParams;
    FText:              String;
    FDataLink:          TDataLink;
    FRowsAffected:      Integer;
    FRequestLive:       Boolean;
    FSQLBinary:         PChar;
    FParamCheck:        Boolean;
    FExecSQL:           Boolean;
    FCheckRowsAffected: Boolean;
{$IFDEF D6H}
   protected
    // IProviderSupport
    procedure PSExecute; override;
    function PSGetDefaultOrder: TIndexDef; override;
    function PSGetParams: TParams; override;
    function PSGetTableName: string; override;
    procedure PSSetCommandText(const CommandText: string); override;
    procedure PSSetParams(AParams: TParams); override;
{$ENDIF}    
   protected
    procedure GetStatementHandle(SQLText: PChar);
    procedure FreeStatement;
    function CreateCursor(GenHandle: Boolean): TACRCursor;
    function GetQueryCursor(GenHandle: Boolean): TACRCursor;
    function GetRowsAffected: Integer;
    procedure QueryChanged(Sender: TObject);
    function GetDataSource: TDataSource; override;
    procedure SetDataSource(Value: TDataSource);
    procedure SetQuery(Value: TStrings);

    function GetParamsCount: Word;
    procedure RefreshParams;
    procedure SetParamsList(Value: TParams);
    procedure SetParamsFromCursor;
  public
    function ParamByName(const Value: string): TParam;

    procedure Prepare;
    procedure UnPrepare;
  public
    procedure PrepareSQL(Value: PChar);
  protected
    procedure SetPrepared(Value: Boolean);
    procedure SetPrepare(Value: Boolean);
  private
    procedure ReadBinaryData(Stream: TStream);
    procedure WriteBinaryData(Stream: TStream);
    procedure ReadParamData(Reader: TReader);
    procedure WriteParamData(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    function CreateHandle: TACRCursor; override;
    procedure Disconnect; override;
    procedure SetDBFlag(Flag: Integer; Value: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ExecSQL;
    procedure GetDetailLinkFields(MasterFields, DetailFields: TList); override;
  public
    property Prepared: Boolean read FPrepared write SetPrepare;
    property ParamCount: Word read GetParamsCount;
    property StmtHandle: TACRSQLProcessor read FStmtHandle;
    property Text: string read FText;
    property RowsAffected: Integer read GetRowsAffected;
    property SQLBinary: PChar read FSQLBinary write FSQLBinary;
  published
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ParamCheck: Boolean read FParamCheck write FParamCheck default True;
    property RequestLive: Boolean read FRequestLive write FRequestLive default False;
    property SQL: TStrings read FSQL write SetQuery;
    property Params: TParams read FParams write SetParamsList stored False;
 end; // TACRQuery


////////////////////////////////////////////////////////////////////////////////
//
// TACRLockParamsEditor
//
////////////////////////////////////////////////////////////////////////////////

{$IFNDEF SQLMEMTABLE}
  TACRLockParamsEditor = class (TPersistent)
   private
    FRetryCount:                   Integer;
    FDelay:                        Integer;
   public
    constructor Create;
    function GetLockParams: TACRLockParams;
    procedure SetLockParams(Params: TACRLockParams);
    procedure SetRetryCount(Value: Integer);
    procedure SetDelay(Value: Integer);
    procedure Assign(Source: TPersistent); override;
   published
    property Delay: Integer read FDelay write SetDelay;
    property RetryCount: Integer read FRetryCount write SetRetryCount;
   end;// TACRLockParamsEditor
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACROptionsEditor
//
////////////////////////////////////////////////////////////////////////////////


{$IFNDEF SQLMEMTABLE}
  TACROptionsEditor = class (TPersistent)
   private
    FMaxSessionCount:           Integer;
    FPageSize:                  Integer;
    FExtentPageCount:           Integer;
    FRandomSearchRetryCount:    Integer;
   public
    constructor Create;
    function GetOptions: TACROptions;
    procedure SetOptions(NewOptions: TACROptions);
    procedure SetMaxSessionCount(Value: Integer);
    procedure SetPageSize(Value: Integer);
    procedure SetExtentPageCount(Value: Integer);
    procedure SetRandomSearchRetryCount(Value: Integer);
    procedure Assign(Source: TPersistent); override;
   published
    property MaxSessionCount: Integer read FMaxSessionCount write SetMaxSessionCount;
    property PageSize: Integer read FPageSize write SetPageSize;
    property ExtentPageCount: Integer read FExtentPageCount write SetExtentPageCount;
    property RandomSearchRetryCount: Integer read FRandomSearchRetryCount write SetRandomSearchRetryCount;
   end;// TACROptionsEditor
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACRCryptoParamsEditor
//
////////////////////////////////////////////////////////////////////////////////


{$IFNDEF SQLMEMTABLE}
  TACRCryptoParamsEditor = class (TPersistent)
   private
    FKeyInfo:         TACRCryptoKey;
    FInitVector:      array [0..ACR_MAX_VECTOR] of Byte;
    FPassword:        string; // ACRDefaultPassword by default
    FCryptoAlgorithm: TACRCryptoAlgorithm;  // acr_Cipher_None by Default
    FCryptoMode:      TACRCryptoMode;  // acr_CTS by Default
    FUseInitVector:   Boolean; // False by default
   public
    constructor Create;
    destructor Destroy; override;
    procedure SetCryptoParams(Params: TACRCryptoParams);
    function GetCryptoParams: TACRCryptoParams;

   protected
    function GetInitVectorValue(Index: Integer): Byte;
    procedure SetInitVectorValue(Index: Integer; Value: Byte);
    function GetVectorSize: Integer;

    function GetKeyValue(Index: Integer): Byte;
    procedure SetKeyValue(Index: Integer; Value: Byte);
    function GetKeySize: Integer;
    procedure SetKeySize(Value: Integer);
    function GetMaxKeySize: Integer;
   public
    procedure SetKey(Key: Pointer; KeySize: Integer);
    function GetKey: Pointer;
    procedure MakeRandomKey(KeySize: Integer);
    procedure MakeRandomInitVector;

    procedure SetInitVector(Vector: Pointer);
    function GetInitVector: Pointer;
    procedure Assign(Source: TPersistent); override;

   public
    property InitVector[Index: Integer]: Byte read GetInitVectorValue write SetInitVectorValue;
    property MaxInitVectorSize: Integer read GetVectorSize;

    property Key[Index: Integer]: Byte read GetKeyValue write SetKeyValue;
    property MaxKeySize: Integer read GetMaxKeySize;

   published
    property CryptoAlgorithm: TACRCryptoAlgorithm read FCryptoAlgorithm write FCryptoAlgorithm;
    property CryptoMode:TACRCryptoMode read FCryptoMode write FCryptoMode;
    property KeySize: Integer read GetKeySize write SetKeySize;
    property Password: string read FPassword write FPassword;
    property UseInitVector: Boolean read FUseInitVector write FUseInitVector;
   end;// TACRCryptoParamsEditor
{$ENDIF}


{$IFNDEF SQLMEMTABLE}


////////////////////////////////////////////////////////////////////////////////
//
// TACRNetworkSettingsEditor
//
////////////////////////////////////////////////////////////////////////////////


  TACRNetworkSettingsEditor = class (TPersistent)
   private
    FPacketSize:               Integer;
    FMaxThreadCount:           Integer;
    FConnectionParamsTunning:  Boolean;
    FTestPacketCount:          Integer;
    FDisconnectRetryCount:     Integer;
    FDisconnectDelay:          Integer;
   protected
    procedure SetPacketSize(Value: Integer);
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure CopySettingsToConnectParams(var ConnectParams: TACRConnectParams); virtual;
   published
    property PacketSize:               Integer read FPacketSize write SetPacketSize;
    property MaxThreadCount:           Integer read FMaxThreadCount write FMaxThreadCount;
    property ConnectionParamsTunning:  Boolean read FConnectionParamsTunning write FConnectionParamsTunning;
    property TestPacketCount:          Integer read FTestPacketCount write FTestPacketCount;
    property DisconnectRetryCount:     Integer read FDisconnectRetryCount write FDisconnectRetryCount;
    property DisconnectDelay:          Integer read FDisconnectDelay write FDisconnectDelay;
  end; // TACRNetworkSettingsEditor


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientNetworkSettingsEditor
//
////////////////////////////////////////////////////////////////////////////////


  TACRClientNetworkSettingsEditor = class (TACRNetworkSettingsEditor)
   private
    FConnectRetryCount:        Integer;
    FConnectDelay:             Integer;
    FStartReceiveTimeOut:      Integer;
    FReceiveTimeOut:           Integer;
    FReceiveSleep:             Integer;
    FMinSendTimeOut:           Integer;
    FSendTimeOut:              Integer;
    FWaitForSendSleep:         Integer;
    FResendDelay:              Integer;
    FRequestDelay:             Integer;
    FWaitForTimeOut:           Integer;
    FThreadsTerminateDelay:    Integer;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure CopySettingsToConnectParams(var ConnectParams: TACRConnectParams); override;
   published
    property ConnectRetryCount:        Integer read FConnectRetryCount write FConnectRetryCount;
    property ConnectDelay:             Integer read FConnectDelay write FConnectDelay;
    property StartReceiveTimeOut:      Integer read FStartReceiveTimeOut write FStartReceiveTimeOut;
    property ReceiveTimeOut:           Integer read FReceiveTimeOut write FReceiveTimeOut;
    property ReceiveSleep:             Integer read FReceiveSleep write FReceiveSleep;
    property MinSendTimeOut:           Integer read FMinSendTimeOut write FMinSendTimeOut;
    property SendTimeOut:              Integer read FSendTimeOut write FSendTimeOut;
    property WaitForSendSleep:         Integer read FWaitForSendSleep write FWaitForSendSleep;
    property ResendDelay:              Integer read FResendDelay write FResendDelay;
    property RequestDelay:             Integer read FRequestDelay write FRequestDelay;
    property WaitForTimeOut:           Integer read FWaitForTimeOut write FWaitForTimeOut;
    property ThreadsTerminateDelay:    Integer read FThreadsTerminateDelay write FThreadsTerminateDelay;
  end; // TACRClientNetworkSettingsEditor


////////////////////////////////////////////////////////////////////////////////
//
// TACRConnectionParamsEditor
//
////////////////////////////////////////////////////////////////////////////////


  TACRConnectionParamsEditor = class (TPersistent)
   private
    FRemoteHost:            String;
    FRemotePort:            Integer;
    FLocalPort:             Integer;
    FDatabaseName:          String;
    FCompressionAlgorithm:  TCompressionAlgorithm;
    FCompressionMode:       Byte;
    FServerID:              Integer;
    FCryptoParamsEditor:    TACRCryptoParamsEditor;
    FNetworkSettings:       TACRClientNetworkSettingsEditor;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function GetConnectParams: TACRConnectParams;
   published
    property RemoteHost: String read FRemoteHost write FRemoteHost;
    property RemotePort: Integer read FRemotePort write FRemotePort;
    property LocalPort: Integer read FLocalPort write FLocalPort;
    property DatabaseName: String read FDatabaseName write FDatabaseName;
    property CompressionAlgorithm: TCompressionAlgorithm
              read FCompressionAlgorithm write FCompressionAlgorithm;
    property CompressionMode: Byte read FCompressionMode write FCompressionMode;
    property CryptoParams: TACRCryptoParamsEditor read FCryptoParamsEditor write FCryptoParamsEditor;
    property ServerID: Integer read FServerID write FServerID;
    property NetworkSettings: TACRClientNetworkSettingsEditor read FNetworkSettings write FNetworkSettings;
  end;
{$ENDIF}


{$IFNDEF SQLMEMTABLE}
  TACRBackupParamsEditor = class (TPersistent)
   private
    FCompressionAlgorithm:  TCompressionAlgorithm;
    FCompressionMode:       Byte;
    FCryptoParamsEditor:    TACRCryptoParamsEditor;
    FDescription:           String;
    FBlockSize:             Integer;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
   published
    property CompressionAlgorithm: TCompressionAlgorithm
              read FCompressionAlgorithm write FCompressionAlgorithm;
    property CompressionMode: Byte read FCompressionMode write FCompressionMode;
    property CryptoParams: TACRCryptoParamsEditor read FCryptoParamsEditor write FCryptoParamsEditor;
    property Description: String read FDescription write FDescription;
    property BlockSize: Integer read FBlockSize write FBlockSize;
  end;
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACRDatabase
//
////////////////////////////////////////////////////////////////////////////////


 TACRDatabase = class (TComponent)
  private
    FOnError:               TACROnError;
    FAfterConnect:          TNotifyEvent;
    FAfterDisconnect:       TNotifyEvent;
    FBeforeConnect:         TNotifyEvent;
    FBeforeDisconnect:      TNotifyEvent;
    FAfterServerShutdown:   TNotifyEvent;
    FBeforeServerShutdown:  TNotifyEvent;
    FAfterConnectionLost:   TNotifyEvent;
    FBeforeConnectionLost:  TNotifyEvent;
    FBeforeInsertRecord:    TACRBeforeInsertRecord;
    FBeforeUpdateRecord:    TACRBeforeUpdateRecord;
    FBeforeDeleteRecord:    TACRBeforeDeleteRecord;
    FBeforeExecuteSQL:      TACRBeforeExecuteSQL;
    FAfterInsertRecord:     TACRAfterInsertRecord;
    FAfterUpdateRecord:     TACRAfterUpdateRecord;
    FAfterDeleteRecord:     TACRAfterDeleteRecord;
    FAfterExecuteSQL:       TACRAfterExecuteSQL;
    FOnProgress:            TACRDatabaseProgress;
    FOnTableProgress:       TACRTableProgress;
    FDataSets:              TList;
    FKeepConnection:        Boolean;
    FTemporary:             Boolean;
    FStreamedConnected:     Boolean;
    FAcquiredHandle:        Boolean;
    FHandleShared:          Boolean;
    FReadOnly:              Boolean;
    FRefCount:              Integer;
    FHandle:                TACRBaseSession;
    FSession:               TACRSession;
    FSessionName:           String;
    FDatabaseName:          String; // name of database
    FDatabaseFileName:      String; // database file name
    FExclusive:             Boolean;
    FLocalDatabase:         Boolean;
    FCSect:                 TRTLCriticalSection;
{$IFNDEF SQLMEMTABLE}
    FBackupParams:            TACRBackupParamsEditor;
    FConnectionParams:        TACRConnectionParamsEditor;
    FLockParamsEditor:        TACRLockParamsEditor;
    FOptionsEditor:           TACROptionsEditor;
    FCryptoParamsEditor:      TACRCryptoParamsEditor;
    FOnReceiveTextMessage:    TACROnClientReceiveTextMessage;
    FOnReceiveBinaryMessage:  TACROnClientReceiveBinaryMessage;
    FOnReceiveStreamMessage:  TACROnClientReceiveStreamMessage;
{$ENDIF}
    // raises exception if not active
    procedure CheckInactive;
    // raises exception if database name is not valid
    procedure CheckDatabaseName;
    // checks session name
    procedure CheckSessionName(Required: Boolean);
    procedure CheckConnected;
    procedure CheckDisconnected;
    // db connected?
    function GetConnected: Boolean;
    // connected dataset
    function GetDataSet(Index: Integer): TACRDataSet;
    // count of connected datasets
    function GetDataSetCount: Integer;
    // opens from existing DB
    function OpenFromExistingDB: Boolean;

    // sets specified file name
    procedure SetDatabaseFileName(Value: string);
    // sets specified database name
    procedure SetDatabaseName(Value: string);
    // sets handle
    procedure SetHandle(Value: TACRBaseSession);
    // keeps connection
    procedure SetKeepConnection(Value: Boolean);
    // sets read-only mode
    procedure SetReadOnly(Value: Boolean);
    // sets session name
    procedure SetSessionName(const Value: string);

    // connect / disconnect
    procedure SetConnected(value: boolean);
    // is database file exists
    function GetExists: boolean;
    // get database manager
    procedure CreateHandle;
    // release database manager
    procedure DestroyHandle;
    procedure DisconnectAllRemoteDatasets(ServerShutdown: Boolean);
    procedure DoOnError(
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       );

  protected
    // loaded
    procedure Loaded; override;
    // sends notification
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    // progress event
    procedure DoOnProgress(Progress: Double; Operation: TACRDatabaseOperation; var Abort: Boolean);
    // progress event
    procedure DoOnTableProgress(
                                      Sender:     TComponent;
                                      Progress:   Double;
                                      Operation:  TACRTableOperation;
                                      var Abort:  Boolean
                               );
    // before insert record
    procedure DoBeforeInsertRecord(
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant;
                       var Abort:          Boolean
                     );
    // after insert record
    procedure DoAfterInsertRecord(
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant
                     );
    procedure DoBeforeUpdateRecord(
                       Sender:                TACRDataset;
                       const TableName:       String;
                       const OldFieldValues:  TACRArrayOfTACRVariant;
                       const NewFieldValues:  TACRArrayOfTACRVariant;
                       var Abort:             Boolean
                     );
    procedure DoAfterUpdateRecord(
                       Sender:                TACRDataset;
                       const TableName:       String;
                       const OldFieldValues:  TACRArrayOfTACRVariant;
                       const NewFieldValues:  TACRArrayOfTACRVariant
                     );
    procedure DoBeforeDeleteRecord(
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant;
                       var Abort:          Boolean
                     );
    procedure DoAfterDeleteRecord(
                       Sender:             TACRDataset;
                       const TableName:    String;
                       const FieldValues:  TACRArrayOfTACRVariant
                     );
    procedure DoBeforeExecuteSQL(
                       Sender:             TACRQuery;
                       var Abort:          Boolean
                     );
    procedure DoAfterExecuteSQL(
                       Sender:             TACRQuery
                     );
  public
    // creates databases with specified directory
    constructor Create(AOwner: TComponent); override;
    // destructor
    destructor Destroy; override;
    // connected := true
    procedure Open;
    // connected := false
    procedure Close;
    // create database
    procedure CreateDatabase;
    // delete database
    procedure DeleteDatabase;
    // rename database
    procedure RenameDatabase(NewDatabaseFileName: String);
    // flush file buffers
    procedure FlushFileBuffers;
    // close all datasets
    procedure CloseDataSets;
    // validates name
    procedure ValidateName(const Name: string);
    // get list of tables in database file
    procedure GetTablesList(List: TStrings);
    // determine if table exists
    function TableExists(TableName: String): Boolean;
    procedure LoadAllTablesFromStream(
                        Stream: TStream
                       );
    procedure SaveAllTablesToStream(
                        Stream: TStream;
                        CompressionAlgorithm: TCompressionAlgorithm = caNone;
                        CompressionMode:      Byte = 0;
                        BlockSize:            Integer = 0
                      );
    function ExportDatabaseToSQL(
                              ExportStructure:      Boolean = True;
                              AddDropTableCommand:  Boolean = True;
                              ExportIndexes:        Boolean = True;
                              AddDropIndexCommand:  Boolean = False;
                              ExportData:           Boolean = True;
                              ExportBLOBFields:     Boolean = True;
                              UseBracketsForNames:  Boolean = False
                            ): String;

{$IFNDEF SQLMEMTABLE}
   protected
    // retrun true if database has active transaction
    function GetInTransaction: Boolean;
    // receive custom message from server and call defined event
    procedure ReceiveMessage(Buffer: PChar; Size: Integer);
   public
    // start a transaction
    procedure StartTransaction;
    // apply changes made by transaction and optionally flush file buffers
    procedure Commit(FlushFileBuffers: Boolean = True);
    // cancel changes made by transaction
    procedure Rollback;
    // remove all locks - called by disconnect of server session
    procedure RemoveAllLocks;

   protected
    function InternalRepairDatabase(IgnoreErrors: Boolean; var Log: String;
                                    UseLowLevelTableAccess: Boolean = False;
                                    Options: TACROptionsEditor = nil;
                                    CryptoParams: TACRCryptoParamsEditor = nil;
                                    NewDatabaseFileName: String = ''): Boolean;
   public
    // compact database
    procedure CompactDatabase;
    // repair database
    function RepairDatabase(UseLowLevelTableAccess: Boolean = False): Boolean; overload;
    function RepairDatabase(var Log: String; UseLowLevelTableAccess: Boolean = False): Boolean; overload;
    // change database settings
    function ChangeDatabaseSettings(Options: TACROptionsEditor;
                                    CryptoParams: TACRCryptoParamsEditor): Boolean;
   protected
    // return database format version
    function GetFormatVersion: Double;
    // empty proc for displaying property in Object Inspector
    procedure SetFormatVersion(Value: Double);
    // return total number of pages
    function GetTotalPageCount: Integer;
    // return number of free pages
    function GetFreePageCount: Integer;
    // return number of used pages
    function GetUsedPageCount: Integer;
    // return density = usedpagecoount / totalpagecount * 100.0 %
    function GetDensity: Double;

   public
    // return true if database is encrypted
    function IsDatabaseEncrypted: Boolean;
    // return true if database is encrypted by password or by key
    function IsDatabaseEncryptedByPassword: Boolean;
    // return true if CryptoParams are valid
    function IsCryptoParamsValid: Boolean;
    procedure SendMessage(const Text: String); overload;
    procedure SendMessage(Buffer: PChar; Size: Integer); overload;
    procedure SendMessage(Stream: TStream); overload;
    // makes Exe database from edb file
    procedure MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string);
    // removes database file from executable database file
    procedure RemoveDatabaseFromExe;
    // returns true if this file is an Accuracer database
    function IsAccuracerDatabaseFile: Boolean;
    // return true if file was copied
    function CopyDatabase(const NewDatabaseFileName: String): Boolean; overload;
    function CopyDatabase(var Log: String; const NewDatabaseFileName: String): Boolean; overload;

   protected
    // progress event
    procedure DoOnBackupProgress(Sender: TObject; Progress: Double; var Abort: Boolean);
    procedure DoOnRestoreProgress(Sender: TObject; Progress: Double; var Abort: Boolean);
    procedure LoadBackupHeader(
                         Stream:            TACRStream;
                         var Header:        TACRBackupHeader;
                         out Description:   String;
                         Tables:            TStrings = nil);
   public
    procedure Backup(const BackupFileName: String);
    procedure Restore(const BackupFileName: String);
    function GetBackupInfo(const BackupFileName: String; Tables: TStrings = nil): TACRBackupInfo;
    function IsAccuracerBackupFile(const BackupFileName: String): Boolean;
    function IsAccuracerBackupFileCryptoParamsValid(const BackupFileName: String): Boolean;
{$ENDIF}
   public
{$IFNDEF SQLMEMTABLE}
    property InTransaction: Boolean read GetInTransaction;
    property TotalPageCount: Integer read GetTotalPageCount;
    property FreePageCount: Integer read GetFreePageCount;
    property UsedPageCount: Integer read GetUsedPageCount;
    property Density: Double read GetDensity;
{$ENDIF}
    property DataSets[Index: Integer]: TACRDataSet read GetDataSet;
    property DataSetCount: Integer read GetDataSetCount;
    property Exists: Boolean read GetExists;
    property Handle: TACRBaseSession read FHandle write SetHandle;
    property Session: TACRSession read FSession;
    property Temporary: Boolean read FTemporary write FTemporary;

  published
{$IFNDEF SQLMEMTABLE}
    property FormatVersion: Double read GetFormatVersion write SetFormatVersion;
    property OnReceiveTextMessage: TACROnClientReceiveTextMessage read FOnReceiveTextMessage write FOnReceiveTextMessage;
    property OnReceiveBinaryMessage: TACROnClientReceiveBinaryMessage read FOnReceiveBinaryMessage write FOnReceiveBinaryMessage;
    property OnReceiveStreamMessage: TACROnClientReceiveStreamMessage read FOnReceiveStreamMessage write FOnReceiveStreamMessage;
{$ENDIF}
    property OnError: TACROnError read FOnError write FOnError;
    property BeforeInsertRecord: TACRBeforeInsertRecord read FBeforeInsertRecord write FBeforeInsertRecord;
    property BeforeUpdateRecord:    TACRBeforeUpdateRecord read FBeforeUpdateRecord write FBeforeUpdateRecord;
    property BeforeDeleteRecord:    TACRBeforeDeleteRecord  read FBeforeDeleteRecord write FBeforeDeleteRecord;
    property BeforeExecuteSQL:      TACRBeforeExecuteSQL read FBeforeExecuteSQL write FBeforeExecuteSQL;
    property AfterInsertRecord:     TACRAfterInsertRecord read FAfterInsertRecord write FAfterInsertRecord;
    property AfterUpdateRecord:     TACRAfterUpdateRecord read FAfterUpdateRecord write FAfterUpdateRecord;
    property AfterDeleteRecord:     TACRAfterDeleteRecord read FAfterDeleteRecord write FAfterDeleteRecord;
    property AfterExecuteSQL:       TACRAfterExecuteSQL read FAfterExecuteSQL write FAfterExecuteSQL;
    property LocalDatabase: Boolean read FLocalDatabase write FLocalDatabase default True;
    property Connected: Boolean read GetConnected write SetConnected default False;
    property DatabaseFileName: string read FDatabaseFileName write SetDatabaseFileName;
    property DatabaseName: string read FDatabaseName write SetDatabaseName;
    property Exclusive: Boolean read FExclusive write FExclusive;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly default False;
    property HandleShared: Boolean read FHandleShared write FHandleShared default False;
    property KeepConnection: Boolean read FKeepConnection write SetKeepConnection default True;
    property SessionName: string read FSessionName write SetSessionName;
{$IFNDEF SQLMEMTABLE}
    property BackupParams: TACRBackupParamsEditor read FBackupParams write FBackupParams;
    property ConnectionParams: TACRConnectionParamsEditor read FConnectionParams write FConnectionParams;
    property LockParams: TACRLockParamsEditor read FLockParamsEditor write FLockParamsEditor;
    property Options: TACROptionsEditor read FOptionsEditor write FOptionsEditor;
    property OnProgress: TACRDatabaseProgress read FOnProgress write FOnProgress;
    property OnTableProgress: TACRTableProgress read FOnTableProgress write FOnTableProgress;
    property CryptoParams: TACRCryptoParamsEditor read FCryptoParamsEditor write FCryptoParamsEditor;
    property AfterConnect: TNotifyEvent read FAfterConnect write FAfterConnect;
    property BeforeConnect: TNotifyEvent read FBeforeConnect write FBeforeConnect;
    property AfterDisconnect: TNotifyEvent read FAfterDisconnect write FAfterDisconnect;
    property BeforeDisconnect: TNotifyEvent read FBeforeDisconnect write FBeforeDisconnect;
    property AfterServerShutdown: TNotifyEvent read FAfterServerShutdown write FAfterServerShutdown;
    property BeforeServerShutdown: TNotifyEvent read FBeforeServerShutdown write FBeforeServerShutdown;
    property AfterConnectionLost: TNotifyEvent read FAfterConnectionLost write FAfterConnectionLost;
    property BeforeConnectionLost: TNotifyEvent read FBeforeConnectionLost write FBeforeConnectionLost;
{$ENDIF}
 end; // TACRDatabase



////////////////////////////////////////////////////////////////////////////////
//
//  TACRAdvFieldDef
//
////////////////////////////////////////////////////////////////////////////////


   TACRAdvFieldDef = class (TObject)
    private
     FName:                 String;
     FDataType:             TACRAdvancedFieldType;
     FRequired:             Boolean;
     FSize:                 Integer;

     FDefaultValue:         TACRVariant;
     FMinValue:             TACRVariant;
     FMaxValue:             TACRVariant;

     // Autoinc settings
     FAutoincIncrement:     Int64;
     FAutoincInitialValue:  Int64;
     FAutoincMinValue:      Int64;
     FAutoincMaxValue:      Int64;
     FAutoincCycled:        Boolean;

     FBlobCompressionAlgorithm: TCompressionAlgorithm;
     FBlobCompressionMode:      Byte;
     FBlobBlockSize:            Integer;
    public
     procedure Assign(Source: TACRAdvFieldDef);
     constructor Create;
     destructor Destroy; override;
    public
     property Name: String read FName write FName;
     property DataType: TACRAdvancedFieldType read FDataType write FDataType;
     property Required: Boolean read FRequired write FRequired;
     property Size: Integer read FSize write FSize;
     property DefaultValue: TACRVariant read FDefaultValue;
     property MinValue: TACRVariant read FMinValue;
     property MaxValue: TACRVariant read FMaxValue;
     property AutoincIncrement: Int64 read FAutoincIncrement write FAutoincIncrement;
     property AutoincInitialValue: Int64 read FAutoincInitialValue write FAutoincInitialValue;
     property AutoincMinValue:   Int64 read FAutoincMinValue write FAutoincMinValue;
     property AutoincMaxValue:  Int64 read FAutoincMaxValue write FAutoincMaxValue;
     property AutoincCycled:    Boolean read FAutoincCycled write FAutoincCycled;
     property BlobCompressionAlgorithm: TCompressionAlgorithm read FBlobCompressionAlgorithm write FBlobCompressionAlgorithm;
     property BlobCompressionMode: Byte read FBlobCompressionMode write FBlobCompressionMode;
     property BlobBlockSize: Integer read FBlobBlockSize write FBlobBlockSize;
   end;


////////////////////////////////////////////////////////////////////////////////
//
//  TACRAdvFieldDefs
//
////////////////////////////////////////////////////////////////////////////////

   TACRAdvFieldDefs = class (TObject)
    private
     FDefsList:      TList;
    private
     function GetCount: Integer;
     function GetDef(Index: Integer): TACRAdvFieldDef;
     procedure SetDef(Index: Integer; value: TACRAdvFieldDef);
    public
     procedure Assign(Source: TACRAdvFieldDefs);
     constructor Create;
     destructor Destroy; override;
     function AddFieldDef: TACRAdvFieldDef;
     procedure Add(const Name: String; DataType: TACRAdvancedFieldType; Size: Integer = 0; Required: Boolean = False);
     procedure DeleteFieldDef(const FieldName: string);
     function Find(const Name: string): TACRAdvFieldDef;
     function IsVarcharExists: Boolean;
     procedure Clear;
    public
     property Items[Index: Integer]: TACRAdvFieldDef read GetDef write SetDef; default;
     property Count: Integer read GetCount;
   end;


////////////////////////////////////////////////////////////////////////////////
//
//  TACRForeignKeyDef
//
////////////////////////////////////////////////////////////////////////////////
   TACRForeignKeyMatchType = (fkmtDefault,fkmtFull,fkmtPartial);
   TACRForeignKeyAction = (fkaDefault,fkaCascade,fkaSetNull,fkaSetDefault,fkaNoAction);

   TACRForeignKeyDef = class (TPersistent)
    private
     FName:                String;
     FMatchType:           TACRForeignKeyMatchType;
     FDeleteAction:        TACRForeignKeyAction;
     FUpdateAction:        TACRForeignKeyAction;
     FReferencedTableName: String;
     FColumns:             String;
    public
     procedure Assign(Source: TPersistent); override;
     constructor Create;
     destructor Destroy; override;
    public
     property Name: String read FName write FName;
     property ReferencedTableName: String read FReferencedTableName write FReferencedTableName;
     property Columns: String read FColumns write FColumns;
     property MatchType: TACRForeignKeyMatchType read FMatchType write FMatchType;
     property DeleteAction: TACRForeignKeyAction read FDeleteAction write FDeleteAction;
     property UpdateAction: TACRForeignKeyAction read FUpdateAction write FUpdateAction;
   end; // TACRForeignKeyDef


////////////////////////////////////////////////////////////////////////////////
//
//  TACRForeignKeyDefs
//
////////////////////////////////////////////////////////////////////////////////


   TACRForeignKeyDefs = class (TPersistent)
    private
     FDefsList:      TList;
    private
     function GetCount: Integer;
     function GetDef(Index: Integer): TACRForeignKeyDef;
     procedure SetDef(Index: Integer; value: TACRForeignKeyDef);
    public
     procedure Assign(Source: TPersistent); override;
     constructor Create;
     destructor Destroy; override;
     function AddForeignKeyDef: TACRForeignKeyDef;
     procedure Add(const Name: String;
                   const Columns: String;
                   const ReferencedTableName: String;
                   const MatchType: TACRForeignKeyMatchType = fkmtDefault;
                   const DeleteAction: TACRForeignKeyAction = fkaDefault;
                   const UpdateAction: TACRForeignKeyAction = fkaDefault
                   );
     procedure DeleteForeignKeyDef(const Name: string);
     procedure Delete(const Index: Integer);
     function Find(const Name: string): TACRForeignKeyDef;
     function IndexOf(const Name: string): Integer;
     procedure Clear;
    public
     property Items[Index: Integer]: TACRForeignKeyDef read GetDef write SetDef; default;
     property Count: Integer read GetCount;
   end; // TACRForeignKeyDefs


////////////////////////////////////////////////////////////////////////////////
//
//  TACRQueryDataLink
//
////////////////////////////////////////////////////////////////////////////////

  TACRQueryDataLink = class(TDetailDataLink)
  private
    FQuery: TACRQuery;
  protected
    procedure ActiveChanged; override;
    procedure RecordChanged(Field: TField); override;
    function GetDetailDataSet: TDataSet; override;
    procedure CheckBrowseMode; override;
  public
    constructor Create(AQuery: TACRQuery);
  end;


////////////////////////////////////////////////////////////////////////////////
//
//  TACRBatchMove
//
////////////////////////////////////////////////////////////////////////////////

  TACRBatchMove = class(TComponent)
  private
    FOnProgress:            TACRTableProgress;
    FDestination:           TACRTable;
    FSource:                TACRDataset;
    FProblemTable:          TACRTable;
    FChangedTable:          TACRTable;
    FKeyViolTable:          TACRTable;
    FMode:                  TACRBatchMode;
    FAbortOnKeyViol:        Boolean;
    FAbortOnProblem:        Boolean;
{$IFNDEF SQLMEMTABLE}
    FUseTransactions:       Boolean;
    FCommitCount:           Int64;
{$ENDIF}
    FRecordCount:           Int64;
    FMovedCount:            Int64;
    FKeyViolCount:          Int64;
    FProblemCount:          Int64;
    FChangedCount:          Int64;
    FMappings:              TStrings;
    FKeyViolTableName:      String;
    FProblemTableName:      String;
    FChangedTableName:      String;
  protected
    procedure SetMappings(Value: TStrings);
    procedure SetSource(Value: TACRDataset);
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    // progress event
    procedure DoOnProgress(
                            Progress:   Double;
                            var Abort:  Boolean
                           );
    procedure CopyRecords(BatchMode:    TACRBatchMode;
                          FieldCount:   Integer;
                          SourceFields: array of Integer;
                          DestFields:   array of Integer
                          );
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Execute;
  public
    property ChangedCount: Int64 read FChangedCount;
    property KeyViolCount: Int64 read FKeyViolCount;
    property MovedCount: Int64 read FMovedCount;
    property ProblemCount: Int64 read FProblemCount;
  published
    property AbortOnKeyViol: Boolean read FAbortOnKeyViol write FAbortOnKeyViol default True;
    property AbortOnProblem: Boolean read FAbortOnProblem write FAbortOnProblem default True;
{$IFNDEF SQLMEMTABLE}
    property CommitCount: Int64 read FCommitCount write FCommitCount default ACRDefaultCommitCount;
{$ENDIF}
    property ChangedTableName: String read FChangedTableName write FChangedTableName;
    property Destination: TACRTable read FDestination write FDestination;
    property KeyViolTableName: String read FKeyViolTableName write FKeyViolTableName;
    property Mappings: TStrings read FMappings write SetMappings;
    property Mode: TACRBatchMode read FMode write FMode default batAppend;
    property OnProgress: TACRTableProgress read FOnProgress write FOnProgress;
    property ProblemTableName: String read FProblemTableName write FProblemTableName;
    property RecordCount: Int64 read FRecordCount write FRecordCount default 0;
    property Source: TACRDataset read FSource write SetSource;
{$IFNDEF SQLMEMTABLE}
    property UseTransactions: Boolean read FUseTransactions write FUseTransactions default True;
{$ENDIF}
  end; // TACRBatchMove

 procedure ACRFixForeignKeysSelfReferences(TableName, NewTableName: String; ForeignKeyDefs: TACRForeignKeyDefs);
 procedure ACRConvertForeignKeyDefToConstraintDef(
            ForeignKeyDef: TACRForeignKeyDef;
            ConstraintDef: TACRConstraintDefForeignKey
                            );

 // convert foreign keys to constraint defs
 procedure ACRConvertForeignKeyDefsToConstraintDefs(
            ForeignKeyDefs: TACRForeignKeyDefs;
            ConstraintDefs: TACRConstraintDefs
            );

 // convert constraint defs to foreign keys
 procedure ConvertConstraintDefsToForeignKeyDefs(
            ConstraintDefs: TACRConstraintDefs;
            ForeignKeyDefs: TACRForeignKeyDefs
            );

 // convert TFieldDefs to ACRFieldDefs
 procedure ConvertFieldDefsToACRFieldDefs(
                FieldDefs:      TFieldDefs;
                ACRFieldDefs:   TACRFieldDefs
                                         );

 // convert AdvFieldDefs to ACRFieldDefs
 procedure ConvertAdvFieldDefsToACRFieldDefs(
                AdvFieldDefs:      TACRAdvFieldDefs;
                ACRFieldDefs:      TACRFieldDefs;
                IndexDefs:         TACRIndexDefs;
                ACRConstraintDefs: TACRConstraintDefs;
                Temporary:         Boolean
                                            );

 // Add Unic or Primary Key constraint
 function AddConstraintForIndex(
                                  IndexDef:           TACRIndexDef;
                                  ACRConstraintDefs:  TACRConstraintDefs
                                 ): TACRConstraintDef;

 // convert ACRFieldDefs to AdvFieldDefs
 procedure ConvertACRFieldDefsToAdvFieldDefs(
                VisibleFieldDefs:   TACRFieldDefs;
                ACRFieldDefs:   TACRFieldDefs;
                ACRConstraintDefs: TACRConstraintDefs;
                AdvFieldDefs:   TACRAdvFieldDefs
                                        );


 // convert ACRFieldDefs to TFieldDefs
 procedure ConvertACRFieldDefsToFieldDefs(
                ACRFieldDefs:   TACRFieldDefs;
                ACRConstraintDefs: TACRConstraintDefs;
                FieldDefs:      TFieldDefs
                                         );


 // convert AdvFieldDefs to FieldDefs
 procedure ConvertAdvFieldDefsToFieldDefs(AdvFieldDefs: TACRAdvFieldDefs; FieldDefs: TFieldDefs);

 // convert FieldDefs to AdvFieldDefs
 procedure ConvertFieldDefsToAdvFieldDefs(FieldDefs: TFieldDefs; AdvFieldDefs: TACRAdvFieldDefs);


 // get string list from string with names
 procedure GetNamesList(List: TStrings; const Names: string);

 // fill ACRIndexDef
 procedure FillACRIndexDef(
              ACRIndexDef:         TACRIndexDef;
              const Name,
              Fields: String;
              Options: TIndexOptions;
              const DescFields: String;
              const CaseInsFields: String;
              FieldDefs:           TFieldDefs;
              AdvFieldDefs:        TACRAdvFieldDefs
                           );

 // convert TIndexDef to TACRIndexDef
 procedure ConvertIndexDefToACRIndexDef(
                IndexDef:      TIndexDef;
                ACRIndexDef:   TACRIndexDef;
                FieldDefs:     TFieldDefs;
                AdvFieldDefs:  TACRAdvFieldDefs
                                         );

 // convert TIndexDefs to ACRIndexDefs
 procedure ConvertIndexDefsToACRIndexDefs(
                IndexDefs:      TIndexDefs;
                ACRIndexDefs:   TACRIndexDefs;
                FieldDefs:      TFieldDefs;
                AdvFieldDefs:   TACRAdvFieldDefs
                                         );

 // convert ACRIndexDefs to TIndexDefs
 procedure ConvertACRIndexDefsToIndexDefs(
                ACRIndexDefs:   TACRIndexDefs;
                IndexDefs:      TIndexDefs
                                         );

 // return true if field exists
 function FindFieldInFieldDefs(FieldDefs: TFieldDefs; FieldName : string): Boolean;

 // compression algorithm
function ConvertCompressionAlgorithmToACRCompressionAlgorithm(
            CompressionAlgorithm: TCompressionAlgorithm
          ): TACRCompressionAlgorithm;
 // compression algorithm
function ConvertACRCompressionAlgorithmToCompressionAlgorithm(
            CompressionAlgorithm: TACRCompressionAlgorithm
          ): TCompressionAlgorithm;

// copy records and return error log
function CopyDatasets(
            SourceDataset:                TDataset;
            DestinationDataset:           TDataset;
            UseSourceDatasetForProgress:  Boolean;
            Operation:                    TACRTableOperation
            ): String;



var
  Sessions:              TACRSessionList;
  ACRDatasetsList:       TList;
  ACRDatasets:           TThreadList;

implementation

uses
     {$IFDEF MEMORY_ENGINE}
     ACRMemEngine,
     {$ENDIF}
     {$IFDEF TEMPORARY_ENGINE}
     ACRTempEngine,
     {$ENDIF}
     {$IFDEF DISK_ENGINE}
     ACRDiskEngine,
     {$ENDIF}
     ACRBaseEngine, Math;
var
  FCSect:                TRTLCriticalSection;
  CurrentSessionManager: TACRSessionComponentManager;
  Session:               TACRSession;
  Initialized:           Boolean;
  IsDesignMode:          Boolean;


////////////////////////////////////////////////////////////////////////////////
//
//  TACRQueryDataLink
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRQueryDataLink.Create(AQuery: TACRQuery);
begin
  inherited Create;
  FQuery := AQuery;
end;// Create


//------------------------------------------------------------------------------
// ActiveChanged
//------------------------------------------------------------------------------
procedure TACRQueryDataLink.ActiveChanged;
begin
  if FQuery.Active then
    FQuery.RefreshParams;
end;// ActiveChanged


//------------------------------------------------------------------------------
// GetDetailDataSet
//------------------------------------------------------------------------------
function TACRQueryDataLink.GetDetailDataSet: TDataSet;
begin
  Result := FQuery;
end;// GetDetailDataSet


//------------------------------------------------------------------------------
// RecordChanged
//------------------------------------------------------------------------------
procedure TACRQueryDataLink.RecordChanged(Field: TField);
begin
  if (Field = nil) and FQuery.Active
    then FQuery.RefreshParams;
end;// RecordChanged


//------------------------------------------------------------------------------
// CheckBrowseMode
//------------------------------------------------------------------------------
procedure TACRQueryDataLink.CheckBrowseMode;
begin
  if FQuery.Active then
    FQuery.CheckBrowseMode;
end;// CheckBrowseMode


//------------------------------------------------------------------------------
// inits engine
//------------------------------------------------------------------------------
procedure InitializeACREngine;
begin
  if (not Initialized) then
   begin
     Initialized:=True;
     CurrentSessionManager:=TACRSessionComponentManager.Create;
   end;
end;


//------------------------------------------------------------------------------
// finalizes engine
//------------------------------------------------------------------------------
procedure FinalizeACREngine;
begin
  if (Initialized) then
   begin
     Initialized:=False;
     CurrentSessionManager.Free;
   end;
end;

//------------------------------------------------------------------------------
// gets default session
//------------------------------------------------------------------------------
function ACRDefaultSession: TACRSession;
begin
   Result := ACRMain.Session;
end;

//------------------------------------------------------------------------------
// gets current session manager
//------------------------------------------------------------------------------
function ACRGetCurrentSession: TACRSessionComponentManager;
begin
  if (not Initialized) then
   raise EACRException.Create(20007, ErrorAEngineNotInitialized);
  Result := CurrentSessionManager;
end;

//------------------------------------------------------------------------------
// sets current session manager
//------------------------------------------------------------------------------
procedure ACRSetCurrentSession(Value: TACRSessionComponentManager);
begin
  CurrentSessionManager := Value;
end;

//------------------------------------------------------------------------------
// creates session manager
//------------------------------------------------------------------------------
procedure ACRStartSession(var Value: TACRSessionComponentManager);
begin
  Value := TACRSessionComponentManager.Create;
end;

//------------------------------------------------------------------------------
// frees session manager
//------------------------------------------------------------------------------
procedure ACRCloseSession(Value: TACRSessionComponentManager);
begin
  if (Value <> nil) then
    Value.Free;
end;


////////////////////////////////////////////////////////////////////////////////
//
//  TACRSessionList
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRSessionList.Create;
begin
  inherited Create;
  FSessions:=TThreadList.Create;
  FSessionNumbers:=TBits.Create;
  InitializeCriticalSection(FCSect);
end;// Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRSessionList.Destroy;
begin
  CloseAll;
  DeleteCriticalSection(FCSect);
  FSessionNumbers.Free;
  FSessions.Free;
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// adds session to list
//------------------------------------------------------------------------------
procedure TACRSessionList.AddSession(ASession: TACRSession);
var
  List: TList;
begin
  List:=FSessions.LockList;
  try
   if (List.Count=0) then
     ASession.FDefault:=True;
   List.Add(ASession);
  finally
   FSessions.UnlockList;
  end;
end;// AddSession


//------------------------------------------------------------------------------
// closes all sessions
//------------------------------------------------------------------------------
procedure TACRSessionList.CloseAll;
var
  I: Integer;
  List: TList;
begin
  List:=FSessions.LockList;
  try
    for I:=List.Count-1 downto 0 do
      TACRSession(List[I]).Free;
  finally
    FSessions.UnlockList;
  end;
end;// CloseAll


//------------------------------------------------------------------------------
// Gets sessions count
//------------------------------------------------------------------------------
function TACRSessionList.GetCount: Integer;
var
  List: TList;
begin
  List:=FSessions.LockList;
  try
    Result:=List.Count;
  finally
    FSessions.UnlockList;
  end;
end;// GetCount


//------------------------------------------------------------------------------
// gets current session
//------------------------------------------------------------------------------
function TACRSessionList.GetCurrentSession: TACRSession;
var
  Handle: TACRSessionComponentManager;
  I: Integer;
  List: TList;
begin
  List:=FSessions.LockList;
  try
    Handle := CurrentSessionManager;
    for I:=0 to List.Count-1 do
      begin
       if (TACRSession(List[I]).FHandle=Handle) then
         begin
           Result:=TACRSession(List[I]);
           Exit;
         end;
      end;
    Result:=nil;
  finally
    FSessions.UnlockList;
  end;
end;// GetCurrentSession


//------------------------------------------------------------------------------
// Gets session by No
//------------------------------------------------------------------------------
function TACRSessionList.GetSession(Index: Integer): TACRSession;
var
  List: TList;
begin
  List:=FSessions.LockList;
  try
    Result:=TACRSession(List[Index]);
  finally
    FSessions.UnlockList;
  end;
end;// GetSession


//------------------------------------------------------------------------------
// Gets session by Name
//------------------------------------------------------------------------------
function TACRSessionList.GetSessionByName(const SessionName: string): TACRSession;
begin
  if (SessionName = '') then
    Result:=Session
  else
    Result := FindSession(SessionName);
  if (Result = nil) then
    raise EACRException.Create(20002, ErrorAInvalidSessionName, [SessionName]);
end;// GetSessionByName


//------------------------------------------------------------------------------
// Finds session by name
//------------------------------------------------------------------------------
function TACRSessionList.FindSession(const SessionName: string): TACRSession;
var
  I: Integer;
  List: TList;
begin
  if (SessionName='') then
    Result:=Session
  else
    begin
      List:=FSessions.LockList;
      try
        for I:=0 to List.Count-1 do
          begin
            Result:=List[I];
            if AnsiCompareText(Result.SessionName,SessionName)=0 then
               Exit;
          end;
        Result:=nil;
      finally
        FSessions.UnlockList;
      end;
    end;
end;// FindSession


//------------------------------------------------------------------------------
// Gets list of sessions names
//------------------------------------------------------------------------------
procedure TACRSessionList.GetSessionNames(List: TStrings);
var
  I: Integer;
  SList: TList;
begin
  List.BeginUpdate;
  try
    List.Clear;
    SList:=FSessions.LockList;
    try
      for I:=0 to SList.Count-1 do
        with TACRSession(SList[I]) do
          List.Add(SessionName);
    finally
      FSessions.UnlockList;
    end;
  finally
    List.EndUpdate;
  end;
end;// GetSessionNames


//------------------------------------------------------------------------------
// Opens session by name
//------------------------------------------------------------------------------
function TACRSessionList.OpenSession(const SessionName: string): TACRSession;
begin
  Result := FindSession(SessionName);
  if (Result = nil) then
    begin
      Result := TACRSession.Create(nil);
      Result.SessionName := SessionName;
    end;
  Result.SetActive(True);
end;// OpenSession


//------------------------------------------------------------------------------
// Sets current session
//------------------------------------------------------------------------------
procedure TACRSessionList.SetCurrentSession(Value: TACRSession);
begin
  ACRSetCurrentSession(Value.FHandle);
end;// SetCurrentSession


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//  TACRSession
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRSession.Create(AOwner: TComponent);
begin
  ValidateAutoSession(AOwner,False);
  inherited Create(AOwner);
  FDatabases:=TList.Create;
  Sessions.AddSession(Self);
  FKeepConnections:=True;
  FHandle := nil;
   
end;// Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRSession.Destroy;
begin
  SetActive(False);
  Sessions.FSessions.Remove(Self);
  FDatabases.Free;
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// adds database
//------------------------------------------------------------------------------
procedure TACRSession.AddDatabase(Value: TACRDatabase);
begin
  FDatabases.Add(Value);
  DBNotification(dbAdd,Value);
end;// AddDatabase


//------------------------------------------------------------------------------
// raises exception if active
//------------------------------------------------------------------------------
procedure TACRSession.CheckInactive;
begin
  if Active then
    DatabaseError(ErrorASessionActive, Self);
end;// CheckInactive


//------------------------------------------------------------------------------
// closes session
//------------------------------------------------------------------------------
procedure TACRSession.Close;
begin
  SetActive(False);
end;// Close


//------------------------------------------------------------------------------
// closes database
//------------------------------------------------------------------------------
procedure TACRSession.CloseDatabase(Database: TACRDatabase);
begin
  with Database do
  begin
    if FRefCount <> 0 then Dec(FRefCount);
    if (FRefCount = 0) and not KeepConnection then
      if not Temporary then Close else
         if not (csDestroying in ComponentState) then Free;
  end;
end;// CloseDatabase


//------------------------------------------------------------------------------
// sends notification
//------------------------------------------------------------------------------
procedure TACRSession.DBNotification(DBEvent: TACRDatabaseEvent; const Param);
begin
  if Assigned(FOnDBNotify) then FOnDBNotify(DBEvent, Param);
end;// DBNotification


//------------------------------------------------------------------------------
// drops all connections
//------------------------------------------------------------------------------
procedure TACRSession.DropConnections;
var
  I: Integer;
begin
  for I := FDatabases.Count - 1 downto 0 do
    with TACRDatabase(FDatabases[I]) do
      if Temporary and (FRefCount = 0) then
        Free;
end;// DropConnections


//------------------------------------------------------------------------------
// finds database by name
//------------------------------------------------------------------------------
function TACRSession.FindDatabase(const DatabaseName: string): TACRDatabase;
var
  I: Integer;
begin
  for I := 0 to FDatabases.Count - 1 do
  begin
    Result := FDatabases[I];
    if ((Result.DatabaseName <> '') or Result.Temporary) and
      (AnsiCompareText(Result.DatabaseName, DatabaseName) = 0) then
        Exit;
  end;
  Result := nil;
end;// FindDatabase


//------------------------------------------------------------------------------
// finds database with specified owner
//------------------------------------------------------------------------------
function TACRSession.DoFindDatabase(const DatabaseName: string; AOwner: TComponent): TACRDatabase;
var
  I: Integer;
begin
  if AOwner <> nil then
    for I := 0 to FDatabases.Count - 1 do
    begin
      Result := FDatabases[I];
      if (Result.Owner = AOwner) and (Result.HandleShared) and
        (AnsiCompareText(Result.DatabaseName, DatabaseName) = 0) then
         Exit;
    end;
  Result := FindDatabase(DatabaseName);
end;// DoFindDatabase


//------------------------------------------------------------------------------
// session is active?
//------------------------------------------------------------------------------
function TACRSession.GetActive: Boolean;
begin
  Result := FHandle <> nil;
end;// GetActive


//------------------------------------------------------------------------------
// gets database by No
//------------------------------------------------------------------------------
function TACRSession.GetDatabase(Index: Integer): TACRDatabase;
begin
  Result := FDatabases[Index];
end;// GetDatabase


//------------------------------------------------------------------------------
// gets count of connected databases
//------------------------------------------------------------------------------
function TACRSession.GetDatabaseCount: Integer;
begin
  Result := FDatabases.Count;
end;// GetDatabaseCount


//------------------------------------------------------------------------------
// get list of database names
//------------------------------------------------------------------------------
procedure TACRSession.GetDatabaseNames(List: TStrings);
var
  I: Integer;
  Names: TStringList;
begin
  Names := TStringList.Create;
  try
    Names.Sorted := True;
    for I := 0 to FDatabases.Count - 1 do
      with TACRDatabase(FDatabases[I]) do
         Names.Add(DatabaseName);
    List.Assign(Names);
  finally
    Names.Free;
  end;
end;// GetDatabaseNames


//------------------------------------------------------------------------------
// get list of database tables
//------------------------------------------------------------------------------
procedure TACRSession.GetTableNames(const DatabaseName: string; List: TStrings);
var
   Database:        TACRDatabase;
begin
  List.BeginUpdate;
  try
    List.Clear;
    try
      Database := OpenDatabase(DatabaseName);
      if (Database <> nil) then
       try
        if (Database <> nil) and (Database.Handle <> nil) then
          Database.Handle.GetTablesList(List);
       finally
        CloseDatabase(Database);
       end;
    except
     if (not (csDesigning in ComponentState)) then
      raise;
    end;
  finally
     List.EndUpdate;
  end;
end;// GetTableNames


//------------------------------------------------------------------------------
// gets handle
//------------------------------------------------------------------------------
function TACRSession.GetHandle: TACRSessionComponentManager;
begin
  if (FHandle <> nil) then
    ACRSetCurrentSession(FHandle)
  else
    SetActive(True);
  Result:=FHandle;
end;// GetHandle


//------------------------------------------------------------------------------
// loaded
//------------------------------------------------------------------------------
procedure TACRSession.Loaded;
begin
  inherited Loaded;
  if AutoSessionName then
     SetSessionNames;
  if FStreamedActive then
     SetActive(True);
end;// Loaded


//------------------------------------------------------------------------------
// locks session
//------------------------------------------------------------------------------
procedure TACRSession.LockSession;
begin
  if (FLockCount=0) then
    begin
      EnterCriticalSection(FCSect);
      Inc(FLockCount);
      try
        MakeCurrent;
      except
        UnlockSession;
        raise;
      end;
   end
 else
   Inc(FLockCount);
end;


//------------------------------------------------------------------------------
// unlocks session
//------------------------------------------------------------------------------
procedure TACRSession.UnlockSession;
begin
  Dec(FLockCount);
  if (FLockCount=0) then
    LeaveCriticalSection(FCSect);
end;// UnlockSession


//------------------------------------------------------------------------------
// makes session current
//------------------------------------------------------------------------------
procedure TACRSession.MakeCurrent;
begin
  if (FHandle <> nil) then
    ACRSetCurrentSession(FHandle)
  else
    SetActive(True);
end;// MakeCurrent


//------------------------------------------------------------------------------
// send notification to datasets and databases
//------------------------------------------------------------------------------
procedure TACRSession.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent,Operation);
  if AutoSessionName and (Operation=opInsert) then
    begin
      if (AComponent is TACRDataSet) then
        TACRDataSet(AComponent).FSessionName:=Self.SessionName
      else
       if (AComponent is TACRDatabase) then
        TACRDatabase(AComponent).FSession:=Self;
      end;
end;// Notification


//------------------------------------------------------------------------------
// opens session
//------------------------------------------------------------------------------
procedure TACRSession.Open;
begin
  SetActive(True);
end;// Open


//------------------------------------------------------------------------------
// opens database (thread-safe)
//------------------------------------------------------------------------------
function TACRSession.DoOpenDatabase(const DatabaseName: string;
                                     AOwner: TComponent): TACRDatabase;
var
  TempDatabase: TACRDatabase;
begin
  LockSession;
  try
    TempDatabase := nil;
    try
      Result := DoFindDatabase(DatabaseName, AOwner);
      if Result = nil then
      begin
        TempDatabase := TACRDatabase.Create(Self);
        TempDatabase.DatabaseName := DatabaseName;
        TempDatabase.KeepConnection := FKeepConnections;
        TempDatabase.Temporary := True;
        Result := TempDatabase;
      end;
      Result.Open;
      Inc(Result.FRefCount);
    except
      if (TempDatabase <> nil) then
        TempDatabase.Free;
      Result := nil;
    end;
  finally
    UnlockSession;
  end;
end;// DoOpenDatabase


//------------------------------------------------------------------------------
// find DB manager by db name
//------------------------------------------------------------------------------
function TACRSession.FindDatabaseHandle(const DatabaseName: string): TACRBaseSession;
var
  I: Integer;
  DB: TACRDatabase;
begin
  for I := 0 to FDatabases.Count - 1 do
  begin
    DB := FDatabases[I];
    if (DB.Handle <> nil) and
       (AnsiCompareText(DB.DatabaseName, DatabaseName) = 0) and
       DB.HandleShared then
    begin
      Result := DB.Handle;
      Exit;
    end;
  end;
  Result := nil;
end;// FindDatabaseHandle


//------------------------------------------------------------------------------
// opens database
//------------------------------------------------------------------------------
function TACRSession.OpenDatabase(const DatabaseName: string): TACRDatabase;
begin
  Result := DoOpenDatabase(DatabaseName,nil);
end;// OpenDatabase


//------------------------------------------------------------------------------
// not auto-session?
//------------------------------------------------------------------------------
function TACRSession.SessionNameStored: Boolean;
begin
  Result := not FAutoSessionName;
end;// SessionNameStored


//------------------------------------------------------------------------------
// removes database from list
//------------------------------------------------------------------------------
procedure TACRSession.RemoveDatabase(Value: TACRDatabase);
begin
  FDatabases.Remove(Value);
  DBNotification(dbRemove, Value);
end;// RemoveDatabase


//------------------------------------------------------------------------------
// opens session
//------------------------------------------------------------------------------
procedure TACRSession.SetActive(Value: Boolean);
begin
  if csReading in ComponentState then
    FStreamedActive := Value
  else
    if Active <> Value then
      StartSession(Value);
end;// SetActive


//------------------------------------------------------------------------------
// sets auto-session name
//------------------------------------------------------------------------------
procedure TACRSession.SetAutoSessionName(Value: Boolean);
begin
  if Value <> FAutoSessionName then
  begin
    if Value then
    begin
      CheckInActive;
      ValidateAutoSession(Owner, True);
      FSessionNumber := -1;
      EnterCriticalSection(FCSect);
      try
        with Sessions do
        begin
          FSessionNumber := FSessionNumbers.OpenBit;
          FSessionNumbers[FSessionNumber] := True;
        end;
      finally
        LeaveCriticalSection(FCSect);
      end;
      UpdateAutoSessionName;
    end
    else
    begin
      if FSessionNumber > -1 then
      begin
        EnterCriticalSection(FCSect);
        try
          Sessions.FSessionNumbers[FSessionNumber] := False;
        finally
          LeaveCriticalSection(FCSect);
        end;
      end;
    end;
    FAutoSessionName := Value;
  end;
end;// SetAutoSessionName


//------------------------------------------------------------------------------
// set name of component
//------------------------------------------------------------------------------
procedure TACRSession.SetName(const NewName: TComponentName);
begin
  inherited SetName(NewName);
  if FAutoSessionName then
    UpdateAutoSessionName;
end;// SetName


//------------------------------------------------------------------------------
// sets the name of session
//------------------------------------------------------------------------------
procedure TACRSession.SetSessionName(const Value: string);
var
  Ses: TACRSession;
begin
  if (FAutoSessionName and (not FUpdatingAutoSessionName)) then
   DatabaseError(ErrorAAutoSessionActive,Self);
  CheckInActive;
  if Value <> '' then
   begin
    Ses := Sessions.FindSession(Value);
    if (not ((Ses = nil) or (Ses = Self))) then
      DatabaseErrorFmt(ErrorADuplicateSessionName, [Value], Self);
   end;
  FSessionName := Value
end;// SetSessionName


//------------------------------------------------------------------------------
// sets session name to datasets and databases
//------------------------------------------------------------------------------
procedure TACRSession.SetSessionNames;
var
  I: Integer;
  Component: TComponent;
begin
  if Owner <> nil then
    for I := 0 to Owner.ComponentCount - 1 do
    begin
      Component := Owner.Components[I];
      if (Component is TACRDataSet) and
        (AnsiCompareText(TACRDataSet(Component).SessionName, Self.SessionName) <> 0) then
        TACRDataSet(Component).SessionName := Self.SessionName
      else if (Component is TACRDatabase) and
        (AnsiCompareText(TACRDatabase(Component).SessionName, Self.SessionName) <> 0) then
        TACRDatabase(Component).SessionName := Self.SessionName
    end;
end;// SetSessionNames


//------------------------------------------------------------------------------
// starts session
//------------------------------------------------------------------------------
procedure TACRSession.StartSession(Value: Boolean);
var
  I: Integer;
begin
  EnterCriticalSection(FCSect);
  try
    if Value then
      begin
        if Assigned(FOnStartup) then
          FOnStartup(Self);
        // session name missing?
        if (FSessionName = '') then
         DatabaseError(ErrorASessionNameMissing, Self);
        // activate default session
        if (ACRDefaultSession <> Self) then
            ACRDefaultSession.Active:=True;
        // default session?
        if FDefault then
          begin
            InitializeACREngine;
            FHandle := ACRGetCurrentSession;
          end
        else
          ACRStartSession(FHandle);
      end
    else
     begin
       ACRSetCurrentSession(FHandle);
       for I:=FDatabases.Count-1 downto 0 do
         begin
           with TACRDatabase(FDatabases[I]) do
             begin
               if Temporary then
                 Free
               else
                 Close;
             end;
         end;
       if FDefault then
         FinalizeACREngine
       else
         begin
           ACRCloseSession(FHandle);
           ACRSetCurrentSession(Session.FHandle);
         end;
       FHandle:=nil;
     end;
  finally
    LeaveCriticalSection(FCSect);
  end;
end;// StartSession


//------------------------------------------------------------------------------
// updates auto-session name
//------------------------------------------------------------------------------
procedure TACRSession.UpdateAutoSessionName;
begin
  FUpdatingAutoSessionName := True;
  try
    SessionName := Format('%s_%d', [Name, FSessionNumber + 1]);
  finally
    FUpdatingAutoSessionName := False;
  end;
  SetSessionNames;
end;// UpdateAutoSessionName


//------------------------------------------------------------------------------
// auto-session name is valid?
//------------------------------------------------------------------------------
procedure TACRSession.ValidateAutoSession(AOwner: TComponent; AllSessions: Boolean);
var
  I: Integer;
  Component: TComponent;
begin
  if AOwner <> nil then
    for I := 0 to AOwner.ComponentCount - 1 do
    begin
      Component := AOwner.Components[I];
      if (Component <> Self) and (Component is TACRSession) then
        if AllSessions then
         DatabaseError(ErrorAAutoSessionExclusive, Self)
        else
        if TACRSession(Component).AutoSessionName then
         DatabaseErrorFmt(ErrorAAutoSessionExists, [Component.Name], Self);
    end;
end;// ValidateAutoSession


////////////////////////////////////////////////////////////////////////////////
//
// TACRDataset
//
////////////////////////////////////////////////////////////////////////////////


{$IFDEF D6H}


////////////////////////////////////////////////////////////////////////////////
//
// IProviderSupport
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// GetUpdateException
//------------------------------------------------------------------------------
function TACRDataset.PSGetUpdateException(E: Exception; Prev: EUpdateError): EUpdateError;
begin
  Result := inherited PSGetUpdateException(E, Prev);
end; // PSGetUpdateException


//------------------------------------------------------------------------------
// IsSQLSupported
//------------------------------------------------------------------------------
function TACRDataset.PSIsSQLSupported: Boolean;
begin
  Result := True;
end; // PSIsSQLSupported


//------------------------------------------------------------------------------
// Reset
//------------------------------------------------------------------------------
procedure TACRDataset.PSReset;
begin
  inherited PSReset;
end; // PSReset


//------------------------------------------------------------------------------
// UpdateRecord
//------------------------------------------------------------------------------
function TACRDataset.PSUpdateRecord(UpdateKind: TUpdateKind; Delta: TDataSet): Boolean;
var
  UpdateAction: TUpdateAction;
begin
  Result := False;
  if Assigned(OnUpdateRecord) then
  begin
    UpdateAction := uaFail;
    if Assigned(FOnUpdateRecord) then
    begin
      FOnUpdateRecord(Delta, UpdateKind, UpdateAction);
      Result := UpdateAction = uaApplied;
    end;
  end;
end; // PSUpdateRecord


//------------------------------------------------------------------------------
// EndTransaction
//------------------------------------------------------------------------------
procedure TACRDataset.PSEndTransaction(Commit: Boolean);
begin
  try
{$IFNDEF SQLMEMTABLE}
    if (Commit) then
      Database.Commit
    else
      Database.Rollback;
{$ENDIF}
  finally
    SetDBFlag(dbfProvider, False);
  end;
end; // PSEndTransaction


//------------------------------------------------------------------------------
// fix params
//------------------------------------------------------------------------------
procedure TACRDataset.FixParams(Query: TDataset);
var Lexer:  TACRLexer;
    i:      Integer;
    s:      String;
    token:  TToken;
    Params: TParams;
begin
  Params := TACRQuery(Query).Params;
  s := '';
  i := 0;
  Lexer := TACRLexer.Create(TACRQuery(Query).SQL.Text,nil);
  try
    while (Lexer.GetNextCommand) do
     begin
       if (Lexer.GetCurrentToken(token)) then
        repeat
          if (token.TokenType = tktString) and (token.Text = '?') then
           begin
            if (i < Params.Count) then
             Params.Items[i].Name := 'P'+IntToStr(i);
            token.Text := ':'+Params.Items[i].Name;
            Inc(i);
           end;
          case token.TokenType of
           tktNone: s := s+' ';
           tktLeftParenthesis,tktRightParenthesis,
           tktComma,tktDot,
           tktInt,tktFloat,
           tktReservedWord,tktString:
              s := s + token.Text+' ';
           tktQuotedString: s := s + AnsiQuotedStr(token.Text,'"')+' ';
           tktBracketedString: s := s + '['+token.Text+'] ';
           tktParameter: s := s + ':'+token.Text+' ' ;
          end;
        until (not Lexer.GetNextToken(token));
       s := s +';'+#13#10;
     end; // command
    TACRQuery(Query).SQL.Text := s;
  finally
    Lexer.Free;
  end;
end; // FixParams


//------------------------------------------------------------------------------
// fixes buggy quoted field and table names
//------------------------------------------------------------------------------
function TACRDataset.FixNames(QueryText: String): String;
var Lexer:  TACRLexer;
    token:  TToken;

  procedure AddToken(SkipSpace: Boolean = False);
  var Space: String;
  begin
    if (SkipSpace) then
     Space := ''
    else
     Space := ' ';
    case token.TokenType of
     tktNone: Result := Result+Space;
     tktLeftParenthesis,tktRightParenthesis,
     tktComma,tktDot,
     tktInt,tktFloat,
     tktReservedWord,tktString:
        Result := Result + token.Text+Space;
     tktQuotedString: Result := Result + AnsiQuotedStr(token.Text,'"')+Space;
     tktBracketedString: Result := Result + '['+token.Text+']'+Space;
     tktParameter: Result := Result + ':'+token.Text+Space;
    end;
  end;


begin
  Result := '';
  Lexer := TACRLexer.Create(QueryText,nil);
  try
    while (Lexer.GetNextCommand) do
     begin
       if (Lexer.GetCurrentToken(token)) then
        repeat
          if (token.TokenType = tktString) and (token.Text = '@') then
           begin
            Result := Result + '[';
            if (Lexer.GetNextToken(token)) then
              AddToken(True);
            if (not Lexer.GetNextToken(token)) then
             break;
            Result := Result+'] ';
            if (not Lexer.GetNextToken(token)) then
             break;
           end;
         AddToken;
        until (not Lexer.GetNextToken(token));
       Result := Result +';'+#13#10;
     end; // command
  finally
    Lexer.Free;
  end;
end; // FixNames


//------------------------------------------------------------------------------
// ExecuteStatemnt
//------------------------------------------------------------------------------
function TACRDataset.PSExecuteStatement(const ASQL: string; AParams: TParams;
  ResultSet: Pointer = nil): Integer;
var
  InProvider: Boolean;
begin
  SetDBFlag(dbfProvider, True);
  InProvider := dbfProvider in DBFlags;
  try
    ResultSet := TACRQuery.Create(nil);
    TACRQuery(ResultSet).InMemory := FInMemory;
    TACRQuery(ResultSet).DatabaseName := FDatabaseName;
    TACRQuery(ResultSet).SessionName := FSessionName;
    TACRQuery(ResultSet).SQL.Text := FixNames(ASQL);
    TACRQuery(ResultSet).FParams.Assign(AParams);
    if (AParams <> nil) then
     if (AParams.Count > 0) then
      FixParams(ResultSet);
    TACRQuery(ResultSet).ExecSQL;
    Result := TACRQuery(ResultSet).RowsAffected;
  finally
    SetDBFlag(dbfProvider, InProvider);
  end;
end; // PSExecuteStatement


//------------------------------------------------------------------------------
// GetAttributes
//------------------------------------------------------------------------------
procedure TACRDataset.PSGetAttributes(List: TList);
begin
  inherited PSGetAttributes(List);
end; // PSGetAttributes


//------------------------------------------------------------------------------
// GetQuoteChar
//------------------------------------------------------------------------------
function TACRDataset.PSGetQuoteChar: string;
begin
  Result := ACRPSQuoteChar;
end; // PSGetQuoteChar


//------------------------------------------------------------------------------
// InTransaction
//------------------------------------------------------------------------------
function TACRDataset.PSInTransaction: Boolean;
var db: TACRDatabase;
begin
  Result := False;
{$IFNDEF SQLMEMTABLE}
  if (FHandle = nil) then
   begin
    if (FDatabase <> nil) then
     Result := FDatabase.InTransaction
    else
     begin
      db := Session.DoFindDatabase(FDatabaseName,nil);
      if (db <> nil) then
       Result := db.InTransaction;
     end;
   end
  else
   Result := FHandle.Session.InTransaction;
{$ENDIF}   
end; // PSInTransaction


//------------------------------------------------------------------------------
// IsSQLBased
//------------------------------------------------------------------------------
function TACRDataset.PSIsSQLBased: Boolean;
begin
  Result := True;
end; // PSIsSQLBased


//------------------------------------------------------------------------------
// StartTransaction
//------------------------------------------------------------------------------
procedure TACRDataset.PSStartTransaction;
begin
  SetDBFlag(dbfProvider, True);
  try
{$IFNDEF SQLMEMTABLE}
    Database.StartTransaction;
{$ENDIF}
  except
    SetDBFlag(dbfProvider, False);
    raise;
  end
end; // PSStartTransaction


{$ENDIF}


//------------------------------------------------------------------------------
// init key buffer
//------------------------------------------------------------------------------
function TACRDataset.InitKeyBuffer(Buffer: TACRRecordBuffer): TACRRecordBuffer;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10280,ErrorLNilPointer);
  Result := Buffer;
  FHandle.InternalInitKeyBuffer(Result);
end; // InitKeyBuffer


//------------------------------------------------------------------------------
// allocate key buffers
//------------------------------------------------------------------------------
procedure TACRDataset.AllocKeyBuffers;
var
  KeyIndex: TACRKeyIndex;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10281,ErrorLNilPointer);
  try
    for KeyIndex := Low(TACRKeyIndex) to High(TACRKeyIndex) do
      FKeyBuffers[KeyIndex] := InitKeyBuffer(FHandle.AllocateKeyRecordBuffer);
  except
    FreeKeyBuffers;
    raise;
  end;
end; // AllocKeyBuffers


//------------------------------------------------------------------------------
// free key buffers
//------------------------------------------------------------------------------
procedure TACRDataset.FreeKeyBuffers;
var
  KeyIndex: TACRKeyIndex;
begin
  for KeyIndex := Low(TACRKeyIndex) to High(TACRKeyIndex) do
   if (FKeyBuffers[KeyIndex] <> nil) then
    begin
     FreeRecordBuffer(FKeyBuffers[KeyIndex]);
     FKeyBuffers[KeyIndex] := nil;
    end;
end; // FreeKeyBuffers


//------------------------------------------------------------------------------
// field defs support
//------------------------------------------------------------------------------
function TACRDataset.FieldDefsStored: Boolean;
begin
  Result := StoreDefs and (FieldDefs.Count > 0);
end; // FieldDefsStored


//------------------------------------------------------------------------------
// index defs support
//------------------------------------------------------------------------------
function TACRDataset.IndexDefsStored: Boolean;
begin
  Result := StoreDefs and (IndexDefs.Count > 0);
end; // IndexDefsStored


//------------------------------------------------------------------------------
// set index definitions
//------------------------------------------------------------------------------
procedure TACRDataset.SetIndexDefs(Value: TIndexDefs);
begin
  IndexDefs.Assign(Value);
end; // SetIndexDefs


//------------------------------------------------------------------------------
// get active buffer
//------------------------------------------------------------------------------
function TACRDataSet.GetActiveRecordBuffer: PChar;
begin
 case State of
      dsBrowse:
        if IsEmpty then Result := nil else Result := ActiveBuffer;
      dsCalcFields:
        Result := CalcBuffer;
      dsFilter:
        Result := FFilterBuffer;
      dsEdit,dsInsert:
        Result:=ActiveBuffer;
      dsSetKey:
        Result := PChar(FKeyBuffer);
        else Result:=nil;
 end;
end; // GetActiveRecordBuffer


//------------------------------------------------------------------------------
// check session name
//------------------------------------------------------------------------------
procedure TACRDataset.CheckDBSessionName;
var
  S: TACRSession;
  Database: TACRDatabase;
begin
  if (SessionName <> '') and (DatabaseName <> '') then
  begin
    S := Sessions.FindSession(SessionName);
    if (Assigned(S) and not Assigned(S.DoFindDatabase(DatabaseName, Self))) then
    begin
      Database := ACRDefaultSession.DoFindDatabase(DatabaseName, Self);
      if Assigned(Database) then
       Database.CheckSessionName(True);
    end;
  end;
end;// CheckDBSessionName


//------------------------------------------------------------------------------
// get base session
//------------------------------------------------------------------------------
function TACRDataset.GetDBHandle: TACRBaseSession;
begin
  if FDatabase <> nil then
    Result := FDatabase.Handle
  else
    Result := nil;
end;// GetDBHandle


//------------------------------------------------------------------------------
// get ACRSession
//------------------------------------------------------------------------------
function TACRDataset.GetDBSession: TACRSession;
begin
  if (FDatabase <> nil) then
    Result := FDatabase.Session
  else
    Result := Sessions.FindSession(SessionName);
  if Result = nil then
   Result := ACRDefaultSession;
end;// GetDBSession


//------------------------------------------------------------------------------
// set InMemory property if memory database
//------------------------------------------------------------------------------
procedure TACRDataset.CheckInMemoryDatabaseName;
begin
  if (AnsiCompareText(FDatabaseName, ACRMemoryDatabaseName) = 0) then
    FInMemory := True
  else
    FInMemory := False;
end;// CheckInMemoryDatabaseName


//------------------------------------------------------------------------------
// set specified database name
//------------------------------------------------------------------------------
procedure TACRDataset.SetDatabaseName(const Value: string);
begin
  if csReading in ComponentState then
   begin
    FDatabaseName := Value;
    CheckInMemoryDatabaseName;
   end
  else if FDatabaseName <> Value then
  begin
    CheckInactive;
    if FDatabase <> nil then
     DatabaseError(ErrorADatabaseOpen, Self);
    FDatabaseName := Value;
    CheckInMemoryDatabaseName;
    DataEvent(dePropertyChange, 0);
  end;
end;// SetDatabaseName


//------------------------------------------------------------------------------
// change current database name
//------------------------------------------------------------------------------
procedure TACRDataset.ChangeCurrentDatabaseName(const Value: string);
begin
  FDatabaseName := Value;
end; // ChangeCurrentDatabaseName


//------------------------------------------------------------------------------
// set specified session name
//------------------------------------------------------------------------------
procedure TACRDataset.SetSessionName(const Value: string);
begin
  CheckInactive;
  FSessionName := Value;
  DataEvent(dePropertyChange, 0);
end;// SetSessionName


//------------------------------------------------------------------------------
// set in-memory
//------------------------------------------------------------------------------
procedure TACRDataset.SetInMemory(const Value: Boolean);
begin
  CheckInactive;
  FInMemory := Value;
  if (Value) then
   FDatabaseName := ACRMemoryDatabaseName;
end;// SetInMemory


//------------------------------------------------------------------------------
// return current version
//------------------------------------------------------------------------------
function TACRDataSet.GetCurrentVersion: String;
var c : char;
begin
 c := DecimalSeparator;
 DecimalSeparator := '.';
 Result := FloatToStrF(ACRVersion,ffFixed,3,2) + ' ' + ACRVersionText;
 DecimalSeparator := c;
end; // GetCurrentVersion


//------------------------------------------------------------------------------
// open cursor
//------------------------------------------------------------------------------
procedure TACRDataSet.OpenCursor(InfoQuery: Boolean);
begin

 FDeleteRecordFlag := False;
 SetDBFlag(dbfOpened, True);
 if (FDatabase <> nil) then
  if (FDatabase.ReadOnly) then
   ReadOnly := True;
 if (FHandle = nil) then
    FHandle := CreateHandle;
 if (FHandle = nil) then
    raise EACRException.Create(20001, ErrorAHandleError);
 inherited OpenCursor(InfoQuery);

end;// OpenCursor


//------------------------------------------------------------------------------
// close cursor
//------------------------------------------------------------------------------
procedure TACRDataSet.CloseCursor;
begin

 FDeleteRecordFlag := False;
 inherited CloseCursor;
 SetDBFlag(dbfOpened, False);
 if (FHandle <> nil) then
      DestroyHandle;
 FAdvFieldDefs.Clear;
 FAdvIndexDefs.Clear;

end;// CloseCursor


//------------------------------------------------------------------------------
// disconnect
//------------------------------------------------------------------------------
procedure TACRDataset.Disconnect;
begin
  Close;
end;// Disconnect


//------------------------------------------------------------------------------
// set DBFlag
//------------------------------------------------------------------------------
procedure TACRDataset.SetDBFlag(Flag: Integer; Value: Boolean);
begin
  if Value then
    begin
      if (not (Flag in FDBFlags)) then
        begin
         if (FDBFlags=[]) then
           begin
            CheckDBSessionName;
            FDatabase := OpenDatabase;
            if (FDatabase = nil) then
             raise EACRException.Create(11245,ErrorLCannotOpenDatabase,[FDatabaseName]);
            FDatabase.FDataSets.Add(Self);
           end;
         Include(FDBFlags,Flag);
       end;
    end
  else
    begin
      if (Flag in FDBFlags) then
        begin
         Exclude(FDBFlags,Flag);
         if (FDBFlags=[]) then
           begin
            FDatabase.FDataSets.Remove(Self);
            FDatabase.Session.CloseDatabase(FDatabase);
            FDatabase := nil;
           end;
        end;
      end;
end;// SetDBFlag


//------------------------------------------------------------------------------
// create handle
//------------------------------------------------------------------------------
function TACRDataSet.CreateHandle: TACRCursor;
begin
 Result := nil;
end;// CreateHandle


//------------------------------------------------------------------------------
// destroy handle
//------------------------------------------------------------------------------
procedure TACRDataSet.DestroyHandle;
begin
 if (FExternalHandle = nil) then
  if (FHandle <> nil) then
    begin
      FHandle.FSettingProjection := False;
      FHandle.Free;
    end;
 FHandle := nil;
end;// DestroyHandle


//------------------------------------------------------------------------------
// internal data convert - date, time and datetime
//------------------------------------------------------------------------------
procedure TACRDataset.InternalDataConvert(Field: TField; Source, Dest: Pointer; ToNative: Boolean);
begin
  case Field.DataType of
   ftDate:
     begin
      if ToNative then
        TACRDate(Dest^) := DateToACRDate(TDate(Source^))
      else
        TDate(Dest^) := ACRDateToDate(TACRDate(Source^));
     end;
   ftDateTime:
     begin
      if ToNative then
        TACRDateTime(Dest^) := DateTimeToACRDateTime(TDateTime(Source^))
      else
        TDateTime(Dest^) := ACRDateTimeToDateTime(TACRDateTime(Source^));
     end;
   ftTime:
     begin
      if ToNative then
        TACRTime(Dest^) := TimeToACRTime(TTime(Source^))
      else
        TTime(Dest^) := ACRTimeToTime(TACRTime(Source^));
     end;
{$IFDEF D6H}
   ftTimeStamp:
     begin
      if ToNative then
        begin
          TACRDateTime(Dest^) := DateTimeToACRDateTime(SQLTimeStampToDateTime(TSQLTimeStamp(Source^)));
        end
      else
        begin
          TSQLTimeStamp(Dest^) := DateTimeToSQLTimeStamp(ACRDateTimeToDateTime(TACRDateTime(Source^)));
        end;
     end;
{$ENDIF}
  end;
end; // InternalDataConvert


//------------------------------------------------------------------------------
// return true if field is wide memo
//------------------------------------------------------------------------------
function TACRDataset.IsWideMemoField(Field: TField): Boolean;
var k: Integer;
begin
  k := Field.FieldNo - 1;
  if (FHandle = nil) then
    raise EACRException.Create(11233,ErrorLNilPointer);
  if ((k < 0) or (k >= FHandle.VisibleFieldDefs.Count)) then
    raise EACRException.Create(11234,ErrorLInvalidFieldNumber,[k,FHandle.VisibleFieldDefs.Count]);
  Result := (FHandle.VisibleFieldDefs.Items[k].AdvancedFieldType = aftWideMemo);
end; // IsWideMemoField


//------------------------------------------------------------------------------
// data convert
//------------------------------------------------------------------------------
procedure TACRDataset.DataConvert(Field: TField; Source, Dest: Pointer;
  ToNative: Boolean);
const x: Word = $0000;

  procedure ProcessWideStringOrWideMemo(WideMemo: Boolean);
  begin
    if ToNative then
     Move(Source^,Dest^,Sizeof(PWideChar))
    else
     begin
      WideString(Dest^) := WideString(PWideChar(Source^));
     end;
  end; // ProcessWideStringOrWideMemo

begin
  case Field.DataType of
  {$IFDEF D6H}
   ftTimeStamp: InternalDataConvert(Field,Source,Dest,ToNative);
  {$ENDIF}
   ftWideString:
     begin
      ProcessWideStringOrWideMemo(False);
     end
   else
    begin
     {$IFDEF D5H}
     inherited DataConvert(Field, Source, Dest, ToNative);
     {$ENDIF}
    end
  end;//case
end;//DataConvert


//------------------------------------------------------------------------------
// set wide memo field
//------------------------------------------------------------------------------
procedure TACRDataset.SetWideMemoField(Field: TField; Value: WideString);
var bs: TStream;
    l:  Integer;
begin
  if (not IsWideMemoField(Field)) then
   DatabaseError(Format(ErrorLNotWideMemoField,[Field.FieldName]),nil);
  l := Length(Value) * 2;
  if (l <= 0) then
   Field.Clear
  else
   begin
     bs := Self.CreateBlobStream(Field,bmWrite);
     bs.WriteBuffer(PWideChar(@Value[1])^,l+2);
     bs.free;
   end;
end; // SetWideMemoField


//------------------------------------------------------------------------------
// get wide memo field
//------------------------------------------------------------------------------
function TACRDataset.GetWideMemoField(Field: TField): WideString;
var bs: TStream;
    l:  Integer;
begin
  if (not IsWideMemoField(Field)) then
   DatabaseError(Format(ErrorLNotWideMemoField,[Field.FieldName]),nil);
  bs := Self.CreateBlobStream(Field,bmRead);
  try
   l := bs.Size;
   if (l <= 0) then
    Result := ''
   else
    begin
     SetLength(Result,l);
     bs.ReadBuffer(PWideChar(@Result[1])^,l);
    end;
  finally
    bs.Free;
  end;
end; // GetWideMemoField


//------------------------------------------------------------------------------
// SetActive
//------------------------------------------------------------------------------
procedure TACRDataset.SetActive(Value: Boolean);
begin
 if (State in [dsEdit,dsInsert]) and (Value = False) then
   InternalCancel; 
 inherited;
end;//SetActive


//------------------------------------------------------------------------------
// switch to index
//------------------------------------------------------------------------------
procedure TACRDataset.SwitchToIndex(const IndexName: string);
begin
   ResetCursorRange;
   UpdateCursorPos;
   if (FHandle <> nil) then
    begin
      if (FHandle.RecordBitmap <> nil) then
       TACRRecordBitmap(FHandle.RecordBitmap).Active := False;
      FHandle.IndexName := IndexName;
    end;
   FKeySize := 0;
   FIndexFieldCount := 0;
   GetIndexInfo;
end;// SwitchToIndex


//------------------------------------------------------------------------------
// check whether field is in index
//------------------------------------------------------------------------------
function TACRDataset.GetIsIndexField(Field: TField): Boolean;
var
  I: Integer;
begin
  Result:=False;
  with Field do
   if FieldNo > 0 then
     for I := 0 to FIndexFieldCount - 1 do
      if FIndexFieldMap[I] = FieldNo then
        begin
          Result := True;
          Exit;
        end;
end;// GetIsIndexField


//------------------------------------------------------------------------------
// get info of acive index
//------------------------------------------------------------------------------
procedure TACRDataSet.GetIndexInfo;
var
  ACRIndexDef: TACRIndexDef;
  i:           Integer;
begin
  if (FHandle <> nil) and (FHandle.IndexName <> '') then
   begin
     ACRIndexDef := FHandle.IndexDefs.GetIndexDefByName(FHandle.IndexName);
     if (ACRIndexDef = nil) then
       raise EACRException.Create(10325,ErrorLCannotFindIndex,[FHandle.IndexName]);
     FIndexFieldCount := ACRIndexDef.ColumnCount;
     SetLength(FIndexFieldMap,FIndexFieldCount);
     FKeySize := 0;
     for i := 0 to FIndexFieldCount - 1 do
      begin
        FIndexFieldMap[i] := FHandle.FieldDefs.GetDefNumberByName(
                              ACRIndexDef.Columns[i].FieldName) + 1;
        Inc(FKeySize,FHandle.FieldDefs[FIndexFieldMap[i]-1].MemoryDataSize);
      end;
   end
  else
   begin
     FIndexFieldMap := nil;
     FIndexFieldCount := 0;
     FKeySize := 0;
   end;
end;// GetIndexInfo


//------------------------------------------------------------------------------
// ResetCursorRange
//------------------------------------------------------------------------------
function TACRDataSet.ResetCursorRange: Boolean;
begin
  if (FHandle = nil) then
   raise EACRException.Create(10287,ErrorLNilPointer);
  Result := False;
  if (PACRKeyBuffer(FKeyBuffers[kiCurRangeStart] + FHandle.KeyOffset)^.Modified or
    PACRKeyBuffer(FKeyBuffers[kiCurRangeEnd] + FHandle.KeyOffset)^.Modified) then
   begin
    InitKeyBuffer(FKeyBuffers[kiCurRangeStart]);
    InitKeyBuffer(FKeyBuffers[kiCurRangeEnd]);
    Result := True;
   end;
  FHandle.ResetRange;
end;// ResetCursorRange


//------------------------------------------------------------------------------
// clear calculated fields
//------------------------------------------------------------------------------
procedure TACRDataSet.ClearCalcFields(Buffer: PChar);
begin
 if (Buffer = nil) then
  raise EACRException.Create(10039,ErrorLNilPointer);
 if (FHandle = nil) then
  raise EACRException.Create(10040,ErrorLNilPointer);
 if (CalcFieldsSize > 0) then
  FillChar(Buffer[FHandle.CalculatedFieldsOffset],CalcFieldsSize,0);
end; // ClearCalcFields


//------------------------------------------------------------------------------
// refresh
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalRefresh;
begin
 FDeleteRecordFlag := False;
 FInsertOrEditComplete := False;
 if (Active) then
  begin
   DataEvent(deDataSetChange, 0);
   DeactivateFilters;
   if (Filtered) then
    ActivateFilters;
   if (FHandle.CurrentRecordBuffer = nil) then
    InternalFirst
   else
    if (GetRecord(FHandle.CurrentRecordBuffer,gmCurrent,False) <> grOK) then
     InternalFirst;
  end;
end; // InternalRefresh


//------------------------------------------------------------------------------
// get record
//------------------------------------------------------------------------------
function TACRDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var ACRGetMode:         TACRGetRecordMode;
    ACRGetResult:       TACRGetRecordResult;
    Bookmark:           PACRBookmarkInfo;
    bSkipMoveRecordID:  Boolean;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10041,ErrorLNilPointer);
 case (GetMode) of
  gmCurrent: ACRGetMode := grmCurrent;
  gmNext: ACRGetMode := grmNext;
  gmPrior: ACRGetMode := grmPrior
 else
  ACRGetMode := grmCurrent;
 end;
 bSkipMoveRecordID := False;
 if ((FInsertOrEditComplete) and (GetMode = gmCurrent)) then
  begin
   Move(FHandle.CurrentRecordBuffer^,Buffer^,FHandle.RecordSize);
   bSkipMoveRecordID := True;
  end;
 FHandle.CurrentRecordBuffer := Buffer;

 // bug fix with navigating dataset connected to DBGrid with multiple rows on the screen
 if ((not FDeleteRecordFlag) and (not bSkipMoveRecordID)) then
   if ((not FHandle.FirstPosition) and (not FHandle.LastPosition)) then
    if ((ActiveBuffer <> nil) and (Active)) then
     if (BufferCount > 1) then
      if ((CurrentRecord < BufferCount) and (CurrentRecord >= 0)) then
        Move(PChar(Buffers[CurrentRecord] + FHandle.BookmarkOffset)^,
          FHandle.CurrentRecordID, SizeOf(TACRRecordID));

 // avoid loading record from database file immediately after posting it
 if ((not FInsertOrEditComplete) or (GetMode <> gmCurrent)) then
   ACRGetResult := FHandle.GetRecordBuffer(ACRGetMode)
 else
   ACRGetResult := grrOK;

 FInsertOrEditComplete := False;
 FDeleteRecordFlag := False;

 case (ACRGetResult) of
  grrOK:
   begin
    Result := grOK;
    ClearCalcFields(Buffer);
    GetCalcFields(Buffer);
    // write bookmark info to record buffer
    Bookmark := PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset);
    Bookmark^.BookmarkData := FHandle.CurrentRecordID;
    Bookmark^.BookmarkFlag := abfCurrent;
   end;
  grrBOF: Result := grBOF;
  grrEOF: Result := grEOF
 else
  Result := grError;
 end;
 if (Result = grError) and DoCheck then
  raise EACRException.Create(10026,ErrorLGetRecordFailed);
end; // GetRecord


//------------------------------------------------------------------------------
// GetCurrentRecord
//------------------------------------------------------------------------------
function TACRDataset.GetCurrentRecord(Buffer: PChar): Boolean;
var
  OldBuffer: TACRRecordBuffer;
begin
  if (FHandle <> nil) then
    begin
      if ((not IsEmpty) and (GetBookmarkFlag(ActiveBuffer) = bfCurrent)) then
         begin
           UpdateCursorPos;
           OldBuffer := FHandle.CurrentRecordBuffer;
           FHandle.CurrentRecordBuffer := Buffer;
           try
             Result := (FHandle.GetRecordBuffer(grmCurrent) = grrOK);
           finally
             FHandle.CurrentRecordBuffer := OldBuffer;
           end;
         end
      else
        Result:=False;
    end
  else
    Result:=False;
end;// GetCurrentRecord


//------------------------------------------------------------------------------
// returns RecordID of the current record
//------------------------------------------------------------------------------
function TACRDataset.GetCurrentRecordID: TACRRecordID;
begin
 if (FHandle = nil) then
  raise EACRException.Create(11352,ErrorLNilPointer);
 Result := FHandle.CurrentRecordID;
end; // GetCurrentRecordID


//------------------------------------------------------------------------------
// return record count
//------------------------------------------------------------------------------
function TACRDataset.GetRecordCount: Integer;
begin
 if (not Active) then
  Result := 0
 else
  begin
   if (FHandle = nil) then
    raise EACRException.Create(10056,ErrorLNilPointer);
   Result := FHandle.RecordCount;
  end;
end; // GetRecordCount


//------------------------------------------------------------------------------
// go to record
//------------------------------------------------------------------------------
procedure TACRDataSet.SetRecNo(Value: Integer);
begin
 if (FHandle = nil) then
  raise EACRException.Create(10042,ErrorLNilPointer);
 DoBeforeScroll;
 FHandle.SetRecNo(Value);
 FDeleteRecordFlag := True;
 Resync([rmCenter]);
 DoAfterScroll;
end; // SetRecNo


//------------------------------------------------------------------------------
// return current record number
//------------------------------------------------------------------------------
function TACRDataSet.GetRecNo: Integer;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10043,ErrorLNilPointer);
 if (State = dsInsert) then
  Result := -1
 else
  begin
   FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
   if (FHandle.CurrentRecordBuffer = nil) then
    Result := -1
   else
    begin
      FHandle.FirstPosition := False;
      FHandle.LastPosition := False;
      FHandle.CurrentRecordID :=
        PACRBookmarkInfo(FHandle.CurrentRecordBuffer + FHandle.BookmarkOffset)^.BookmarkData;
      Result := FHandle.GetRecNo;
    end;
  end;
end; // GetRecNo


//------------------------------------------------------------------------------
// go to first record (before first record to BOF)
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalFirst;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10044,ErrorLNilPointer);
 FHandle.InternalFirst;
end; // InternalFirst


//------------------------------------------------------------------------------
// go to last record (after last record to EOF)
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalLast;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10045,ErrorLNilPointer);
 FHandle.InternalLast;
end; // InternalLast


//------------------------------------------------------------------------------
// go to record in buffer
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalSetToRecord(Buffer: PChar);
begin
 if (FHandle = nil) then
  raise EACRException.Create(10046,ErrorLNilPointer);
 if (Buffer = nil) then
  raise EACRException.Create(10047,ErrorLNilPointer);
 InternalGotoBookmark(Buffer + FHandle.BookmarkOffset); 
end; // InternalSetToRecord


//------------------------------------------------------------------------------
// get bookmark flag
//------------------------------------------------------------------------------
function TACRDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
 if (Buffer = nil) then
  raise EACRException.Create(10059,ErrorLNilPointer);
 if (FHandle = nil) then
  raise EACRException.Create(10060,ErrorLNilPointer);
 Result := bfCurrent; 
 case PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkFlag  of
  abfCurrent:
     Result := bfCurrent;
  abfBOF:
     Result := bfBOF;
  abfEOF:
     Result := bfEOF;
  abfInserted:
     Result := bfInserted;
 end;
end; // GetBookmarkFlag


//------------------------------------------------------------------------------
// get bookmark data
//------------------------------------------------------------------------------
procedure TACRDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
 if (Buffer = nil) then
  raise EACRException.Create(10034,ErrorLNilPointer);
 if (Data = nil) then
  raise EACRException.Create(10035,ErrorLNilPointer);
 if (FHandle = nil) then
  raise EACRException.Create(10036,ErrorLNilPointer);
 // copy bookmark
 Move(PChar(Buffer + FHandle.BookmarkOffset)^,PChar(Data)^,Sizeof(TACRRecordID));
end; // GetBookmarkData


//------------------------------------------------------------------------------
// go to bookmark
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalGotoBookmark(Bookmark: Pointer);
begin
 if (Bookmark = nil) then
  raise EACRException.Create(10037,ErrorLNilPointer);
 if (FHandle = nil) then
  raise EACRException.Create(10038,ErrorLNilPointer);
 FHandle.CurrentRecordID :=
   PACRBookmarkInfo(Bookmark)^.BookmarkData;
 FHandle.FirstPosition := False;
 FHandle.LastPosition := False;
end; // InternalGotoBookmark


//------------------------------------------------------------------------------
// set flag
//------------------------------------------------------------------------------
procedure TACRDataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
 if (FHandle = nil) then
  raise EACRException.Create(10057,ErrorLNilPointer);
 case Value of
  bfCurrent:
    PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkFlag :=
      abfCurrent;
  bfBOF:
    PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkFlag :=
      abfBOF;
  bfEOF:
    PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkFlag :=
      abfEOF;
  bfInserted:
    PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkFlag :=
      abfInserted;
 end;
end; // SetBookmarkFlag


//------------------------------------------------------------------------------
// set data
//------------------------------------------------------------------------------
procedure TACRDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
 if (FHandle = nil) then
  raise EACRException.Create(10058,ErrorLNilPointer);
 if (Buffer = nil) then
  raise EACRException.Create(10061,ErrorLNilPointer);
 if (Data = nil) then Exit;
 // copy bookmark
 Move(PChar(Data)^,PChar(Buffer + FHandle.BookmarkOffset)^,Sizeof(TACRRecordID));
end; // SetBookmarkData


//------------------------------------------------------------------------------
// compare bookmarks
//------------------------------------------------------------------------------
function TACRDataSet.CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer;
const
  RetCodes: array[Boolean, Boolean] of ShortInt = ((2,-1),(1,0));
var Pos1, Pos2:           TACRRecordNo;
    RecordID1, RecordID2: TACRRecordID;
    OldPos:               Pointer;
    OldBuffer:            PChar;
    TempBuffer:           PChar;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10063,ErrorLNilPointer);
 Result := RetCodes[Bookmark1 = nil, Bookmark2 = nil];
 if Result = 2 then
  begin
    OldPos := FHandle.SavePosition;
    OldBuffer := FHandle.CurrentRecordBuffer;
    TempBuffer := FHandle.AllocateRecordBuffer;
    FHandle.CurrentRecordBuffer := TempBuffer;
    try
      FHandle.FirstPosition := False;
      FHandle.LastPosition := False;
      Move(PChar(Bookmark1)^,RecordID1,Sizeof(TACRRecordID));
      Move(PChar(Bookmark2)^,RecordID2,Sizeof(TACRRecordID));

      FHandle.CurrentRecordID := RecordID1;
//      if (FHandle.GetRecordBuffer(grmCurrent) <> grrOK) then
      if (not BookmarkValid(Bookmark1)) then
       Pos1 := -1
      else
       Pos1 := FHandle.GetRecNo;

      FHandle.CurrentRecordID := RecordID2;
      FHandle.FirstPosition := False;
      FHandle.LastPosition := False;
      FHandle.CurrentRecordBuffer := TempBuffer;
//      if (FHandle.GetRecordBuffer(grmCurrent) <> grrOK) then
      if (not BookmarkValid(Bookmark2)) then
       Pos2 := -1
      else
       Pos2 := FHandle.GetRecNo;

      if (Pos1 = Pos2) then
       Result := 0
      else
      if (Pos1 > Pos2) then
       Result := 1
      else
       Result := -1;
    finally
      FHandle.RestorePosition(OldPos);
      FHandle.FreePosition(OldPos);
      FHandle.FreeRecordBuffer(TempBuffer);
      FHandle.CurrentRecordBuffer := OldBuffer;
    end;
  end; // Result = 2
end; // CompareBookmarks


//------------------------------------------------------------------------------
// checks if bookmark is valid
//------------------------------------------------------------------------------
function TACRDataSet.BookmarkValid(Bookmark: TBookmark): Boolean;
var
    OldPos:               Pointer;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10064,ErrorLNilPointer);
 Result := False;
 if (Bookmark <> nil) then
  begin
   OldPos := FHandle.SavePosition;
   try
     Move(PChar(Bookmark)^,FHandle.CurrentRecordID,sizeof(TACRRecordID));
     Result := FHandle.IsRecordExists;
   finally
     FHandle.RestorePosition(OldPos);
     FHandle.FreePosition(OldPos);
   end;

{
   OldBuffer := FHandle.CurrentRecordBuffer;
   TempBuffer := FHandle.AllocateRecordBuffer;
   FHandle.CurrentRecordBuffer := TempBuffer;
   try
     Move(PChar(Bookmark)^,FHandle.CurrentRecordID,sizeof(TACRRecordID));
     Result := (FHandle.GetRecordBuffer(grmCurrent) = grrOK);
   finally
      FHandle.RestorePosition(OldPos);
      FHandle.FreePosition(OldPos);
      FHandle.FreeRecordBuffer(TempBuffer);
      FHandle.CurrentRecordBuffer := OldBuffer;
   end;
}
  end;
end; // BookmarkValid


//------------------------------------------------------------------------------
// return True if OnFilterRecord is applied and Fitlered is True
//------------------------------------------------------------------------------
function TACRDataSet.IsOnFilterRecordApplied: Boolean;
begin
  Result := Filtered and (Assigned(OnFilterRecord));
end; // IsOnFilterRecordApplied


//------------------------------------------------------------------------------
// return True if OnFilterRecord is applied and Fitlered is True
//------------------------------------------------------------------------------
function TACRDataSet.InternalFilterRecord(Buffer: TACRRecordBuffer): Boolean;
var
 SaveState: TDataSetState;
begin
 if (IsOnFilterRecordApplied) then
  begin
 SaveState := SetTempState(dsFilter);
 FFilterBuffer := Buffer;
 try
   Result := True;
   OnFilterRecord(Self,Result);
 except
     {$IFDEF D6H}
   ApplicationHandleException(Self);
     {$ELSE}
     Application.HandleException(Self)
     {$ENDIF}
 end;
 RestoreState(SaveState);
  end
 else
  Result := True;
end;


//------------------------------------------------------------------------------
// return Accept if OnFilterRecord accepts current record
//------------------------------------------------------------------------------
function TACRDataSet.FilterRecord(Buffer: TACRRecordBuffer; Dataset: Pointer): Boolean;
begin
  Result := TACRDataSet(Dataset).InternalFilterRecord(Buffer);
end; // FilterRecord


//------------------------------------------------------------------------------
// applies OnFilterRecord
//------------------------------------------------------------------------------
procedure TACRDataSet.SetOnFilterRecord(const Value: TFilterRecordEvent);
begin
  if Active then
   begin
    if (FHandle = nil) then
     raise EACRException.Create(10275,ErrorLNilPointer);
    CheckBrowseMode;
    if (Assigned(OnFilterRecord) <> Assigned(Value)) then
     begin
      if (Filtered) then
       ActivateFilters
      else
       DeactivateFilters;
    end;
    inherited SetOnFilterRecord(Value);
    if (Filtered) then First;
   end
  else
   inherited SetOnFilterRecord(Value);
end;


//------------------------------------------------------------------------------
// return true if index applied
//------------------------------------------------------------------------------
function TACRDataSet.IsIndexApplied: Boolean;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10381,ErrorLNilPointer);
 Result := (FHandle.FIndexID <> INVALID_OBJECT_ID); 
end; // IsIndexApplied


//------------------------------------------------------------------------------
// prepare cursor
//------------------------------------------------------------------------------
procedure TACRDataSet.PrepareCursor;
begin
end; // PrepareCursor


//------------------------------------------------------------------------------
// return true if dataset can be modified (not read only)
//------------------------------------------------------------------------------
function TACRDataSet.GetCanModify: Boolean;
begin
  Result := not Self.FReadOnly;
end; // GetCanModify


//------------------------------------------------------------------------------
// set SQL Filter
//------------------------------------------------------------------------------
procedure TACRDataSet.SetSQLFilter(FilterExpr: TObject);
begin
  if (FHandle = nil) then
    raise EACRException.Create(10300,ErrorLNilPointer);
  FHandle.SetSQLFilter(FilterExpr);
end; // SetSQLFilter


//------------------------------------------------------------------------------
// apply projection
//------------------------------------------------------------------------------
procedure TACRDataSet.ApplyProjection(FieldNamesList, AliasList: TStringList);
begin
  if (FHandle = nil) then
    raise EACRException.Create(10320,ErrorLNilPointer);
  FHandle.ApplyProjection(FieldNamesList,AliasList);
  FHandle.FSettingProjection := True;
  FExternalHandle := FHandle;
  try
    Close;
    Open;
    FHandle.FSettingProjection := False;
  finally
    FExternalHandle := nil;
  end;
  FIsProjectionSet := True;
end; // ApplyProjection


//------------------------------------------------------------------------------
// activate filters
//------------------------------------------------------------------------------
procedure TACRDataSet.ActivateFilters;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10269,ErrorLNilPointer);
 // Filter property based filters
 if (Filter <> '') then
  FHandle.ActivateFilters(Filter,(foCaseInsensitive in FilterOptions),
                          (not (foNoPartialCompare in FilterOptions)))
 else
  FHandle.DeactivateFilters;
 if (Assigned(OnFilterRecord)) then
   FHandle.FilterRecord := @TACRDataset.FilterRecord
 else
   FHandle.FilterRecord := nil;
end; // ActivateFilters


//------------------------------------------------------------------------------
// deactivate filters
//------------------------------------------------------------------------------
procedure TACRDataSet.DeactivateFilters;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10270,ErrorLNilPointer);
 FHandle.DeactivateFilters; 
end; // DeactivateFilters


//------------------------------------------------------------------------------
// set filter
//------------------------------------------------------------------------------
procedure TACRDataSet.SetFilterData(const Text: string; Options: TFilterOptions);
var bResetFilters: Boolean;
begin
  bResetFilters := False;
  if Active then
   begin
    CheckBrowseMode;
    if (FHandle = nil) then
     raise EACRException.Create(10266,ErrorLNilPointer);
    if (Filter <> Text) or (FilterOptions <> Options) then
      bResetFilters := True;
   end;
  inherited SetFilterText(Text);
  inherited SetFilterOptions(Options);
  if (Active and Filtered) then
   begin
    if (bResetFilters) then
     ActivateFilters;
    First;
   end;
end; // SetFilterData


//------------------------------------------------------------------------------
// set filtered
//------------------------------------------------------------------------------
procedure TACRDataSet.SetFiltered(Value: Boolean);
begin
  if (Active) then
   begin
    CheckBrowseMode;
    if (FHandle = nil) then
     raise EACRException.Create(10267,ErrorLNilPointer);
    if (Filtered <> Value) then
     begin
      // filtered changed
      if (Value) then
       ActivateFilters
      else
       DeactivateFilters;
      inherited SetFiltered(Value);
    end;
    First;
   end
  else
   inherited SetFiltered(Value);
end; // SetFiltered


//------------------------------------------------------------------------------
// set filter options
//------------------------------------------------------------------------------
procedure TACRDataSet.SetFilterOptions(Value: TFilterOptions);
begin
  SetFilterData(Filter, Value);
end; // SetFilterOptions


//------------------------------------------------------------------------------
// set filter text
//------------------------------------------------------------------------------
procedure TACRDataSet.SetFilterText(const Value: string);
begin
  SetFilterData(Value, FilterOptions);
end; // SetFilterText


//------------------------------------------------------------------------------
// FindFirst, FindNext, Filters
//------------------------------------------------------------------------------
function TACRDataSet.FindRecord(Restart, GoForward: Boolean): Boolean;
var Buffer:     TACRRecordBuffer;
    GetResult:  TGetResult;
begin
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  if (FHandle = nil) then
   raise EACRException.Create(10268,ErrorLNilPointer);
  UpdateCursorPos;
  CursorPosChanged;
  if (not Filtered) then
   ActivateFilters;
  Buffer := AllocRecordBuffer;
  try
   if GoForward then
    begin
     if (Restart) then
      InternalFirst;
     GetResult := GetRecord(Buffer,gmNext,False);
    end
   else
    begin
     if (Restart) then
      InternalLast;
     GetResult := GetRecord(Buffer,gmPrior,False);
    end;
  finally
   FreeRecordBuffer(Buffer);
   if (not Filtered) then
    DeactivateFilters;
  end;
 if (GetResult = grOK) then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end;
  Result := Found;
  if (Result) then
    DoAfterScroll;
end; // FindRecord


//------------------------------------------------------------------------------
// Locate record
//------------------------------------------------------------------------------
function TACRDataSet.LocateRecord(
                         const KeyFields: String;
                         const KeyValues: Variant;
                         Options:         TLocateOptions
                        ): Boolean;
var Buffer: PChar;
begin
  if (FHandle = nil) then
   raise EACRException.Create(10271,ErrorLNilPointer);
  CheckBrowseMode;
  CursorPosChanged;
  SetTempState(dsFilter);
  Buffer := TempBuffer;
  FFilterBuffer := Buffer;
  try
    FHandle.CurrentRecordBuffer := Buffer;
    Result := FHandle.Locate(
                             KeyFields,KeyValues,
                             (loCaseInsensitive in Options),
                             (loPartialKey in Options)
                            );
  finally
    RestoreState(dsBrowse);
  end;
end; // LocateRecord


//------------------------------------------------------------------------------
// Locate
//------------------------------------------------------------------------------
function TACRDataSet.Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean;
begin
  DoBeforeScroll;
  Result := LocateRecord(KeyFields, KeyValues, Options);
  if (Result) then
   begin
    Resync([rmExact, rmCenter]);
    DoAfterScroll;
   end;
end; // Locate


//------------------------------------------------------------------------------
// Lookup
//------------------------------------------------------------------------------
function TACRDataSet.Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant;
begin
  Result := Null;
  if (LocateRecord(KeyFields, KeyValues, [])) then
  begin
    SetTempState(dsCalcFields);
    try
      CalculateFields(TempBuffer);
      Result := FieldValues[ResultFields];
    finally
      RestoreState(dsBrowse);
    end;
  end;
end; // Lookup


//------------------------------------------------------------------------------
// InitRecord
//------------------------------------------------------------------------------
procedure TACRDataSet.InitRecord(Buffer: PChar);
begin
  inherited InitRecord(Buffer);
  PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkFlag := abfInserted;
  PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkData.PageNo := INVALID_PAGE_NO;
  PACRBookmarkInfo(Buffer + FHandle.BookmarkOffset)^.BookmarkData.PageItemNo := $FFFF;
end;// InitRecord


//------------------------------------------------------------------------------
// appending table (Append flag - ignored, record will be inserted at first empty position)
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
 if (not Active) then
  raise EACRException.Create(10075,ErrorLDatasetIsNotOpened);
 if (State <> dsInsert) then
  raise EACRException.Create(10076,ErrorLDatasetIsNotInInsertOrEditMode);
 if (ReadOnly) then
  raise EACRException.Create(10077,ErrorLDatasetIsInReadOnlyMode);
 if (FHandle = nil) then
  raise EACRException.Create(10078,ErrorLNilPointer);

 FHandle.CurrentRecordBuffer := Buffer;

 if (IsBeforeInsertRecordAssigned or IsAfterInsertRecordAssigned) then
   SaveValues(FNewFieldValues);
 try
   if (IsBeforeInsertRecordAssigned) then
    if (not DoBeforeInsertRecord) then
     begin
      DatabaseError(ErrorLInsertRecordBlockedByDatabaseEvent,Self);
      Exit;
     end;
   FHandle.InternalPost(True);

   if (FHandle.ErrorCode <> ACR_ERR_OK) then
    begin
     if (FHandle.ErrorCode = ACR_ERR_INSERT_RECORD) then
      DatabaseError(ErrorLAddingRecord);
    end // not OK
   else
    begin
     FInsertOrEditComplete := True;
     if (IsAfterInsertRecordAssigned) then
      begin
       if (FValuesChangedByEventHandler) then
        SaveValues(FNewFieldValues);
       DoAfterInsertRecord;
      end;
    end;
 finally
  if (IsBeforeInsertRecordAssigned or IsAfterInsertRecordAssigned) then
   ClearValues(FNewFieldValues);
 end;
end; // InternalAddRecord


//------------------------------------------------------------------------------
// insert record
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalInsert;
begin
 if (not ReadOnly) then
   FHandle.InternalInsert;
end; // InternalInsert


//------------------------------------------------------------------------------
// edit record
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalEdit;
begin
 if (not ReadOnly) then
  begin
   if (FHandle = nil) then
    raise EACRException.Create(10065,ErrorLNilPointer);
   if (ActiveBuffer = nil) then
    raise EACRException.Create(10067,ErrorLNilPointer);
   FHandle.InternalEdit;
   if (FHandle.ErrorCode = ACR_ERR_OK) then
    begin
     FEditRecordBuffer := AllocRecordBuffer;
     Move(PChar(ActiveBuffer)^,FEditRecordBuffer^,FHandle.RecordSize);
     if (IsBeforeUpdateRecordAssigned or IsAfterUpdateRecordAssigned) then
      SaveValues(FOldFieldValues);
    end;
   if (FHandle.ErrorCode = ACR_ERR_UPDATE_RECORD_PROHIBITED) then
    DatabaseError(ErrorLUpdateRecordProhibited);
  end;
end; // InternalEdit


//------------------------------------------------------------------------------
// cancels updates
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalCancel;
begin
 if (State = dsEdit) then
  if (IsBeforeUpdateRecordAssigned or IsAfterUpdateRecordAssigned) then
   ClearValues(FOldFieldValues);
 if (FHandle = nil) then
  raise EACRException.Create(10066,ErrorLNilPointer);
 if (FEditRecordBuffer <> nil) then
   FreeRecordBuffer(FEditRecordBuffer);
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
 FHandle.InternalCancel(State = dsInsert);
end; // InternalCancel


//------------------------------------------------------------------------------
// update record
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalPost;
var bInsert: Boolean;
begin
 if (not Active) then
  raise EACRException.Create(10029,ErrorLDatasetIsNotOpened);
 if (State <> dsInsert) and (State <> dsEdit) then
  raise EACRException.Create(10030,ErrorLDatasetIsNotInInsertOrEditMode);
 if (ReadOnly) then
  raise EACRException.Create(10032,ErrorLDatasetIsInReadOnlyMode);
 if (FHandle = nil) then
  raise EACRException.Create(10048,ErrorLNilPointer);
 bInsert := (State = dsInsert);
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
 if (not bInsert) then
  FHandle.EditRecordBuffer := FEditRecordBuffer;
 if (bInsert) then
  if (IsBeforeInsertRecordAssigned or IsAfterInsertRecordAssigned) then
   SaveValues(FNewFieldValues);
 if (not bInsert) then
  if (IsBeforeUpdateRecordAssigned or IsAfterUpdateRecordAssigned) then
   SaveValues(FNewFieldValues);
 try
   if (bInsert) then
    if (IsBeforeInsertRecordAssigned) then
     if (not DoBeforeInsertRecord) then
      begin
       DatabaseError(ErrorLInsertRecordBlockedByDatabaseEvent,Self);
       Exit;
      end;
   if (not bInsert) then
    if (IsBeforeUpdateRecordAssigned) then
     if (not DoBeforeUpdateRecord) then
      begin
       DatabaseError(ErrorLUpdateRecordBlockedByDatabaseEvent,Self);
       Exit;
      end;

   FHandle.InternalPost(State = dsInsert);

   if (not bInsert) then
    if (FHandle.ErrorCode = ACR_ERR_OK) then
    FreeRecordBuffer(FEditRecordBuffer);
   if (FHandle.ErrorCode <> ACR_ERR_OK) then
     DatabaseError(FHandle.ErrorMessage)
   else
    begin
     FInsertOrEditComplete := True;
     if (bInsert) then
      if (IsAfterInsertRecordAssigned) then
        begin
         if (FValuesChangedByEventHandler) then
          SaveValues(FNewFieldValues);
         DoAfterInsertRecord;
        end;
     if (not bInsert) then
      if (IsAfterUpdateRecordAssigned) then
        begin
         if (FValuesChangedByEventHandler) then
          SaveValues(FNewFieldValues);
         DoAfterUpdateRecord;
        end;
    end;
 finally
   if (bInsert) then
    if (IsBeforeInsertRecordAssigned or IsAfterInsertRecordAssigned) then
     ClearValues(FNewFieldValues);
   if (not bInsert) then
    if (IsBeforeUpdateRecordAssigned or IsAfterUpdateRecordAssigned) then
     begin
       ClearValues(FOldFieldValues);
       ClearValues(FNewFieldValues);
     end;
 end;
end; // InternalPost


//------------------------------------------------------------------------------
// delete record
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalDelete;
begin
 if (not Active) then
  raise EACRException.Create(10068,ErrorLDatasetIsNotOpened);
 if (ReadOnly) then
  raise EACRException.Create(10069,ErrorLDatasetIsInReadOnlyMode);
 if (FHandle = nil) then
  raise EACRException.Create(10070,ErrorLNilPointer);
 if (ActiveBuffer = nil) then
  raise EACRException.Create(10071,ErrorLNilPointer);
 if (State = dsEdit) then
  Cancel;
 if (IsBeforeDeleteRecordAssigned or IsAfterDeleteRecordAssigned) then
  SaveValues(FOldFieldValues);
 try
   if (IsBeforeDeleteRecordAssigned) then
    if (not DoBeforeDeleteRecord) then
     begin
      DatabaseError(ErrorLDeleteRecordBlockedByDatabaseEvent,Self);
      Exit;
     end;

   FHandle.CurrentRecordBuffer := ActiveBuffer;
   FHandle.InternalDelete;
   if (FHandle.ErrorCode <> ACR_ERR_OK) then
    begin
     if (FHandle.ErrorCode = ACR_ERR_DELETE_RECORD_PROHIBITED) then
      DatabaseError(ErrorLDeleteRecordProhibited)
     else
     if (FHandle.ErrorCode = ACR_ERR_DELETE_RECORD) then
      DatabaseError(ErrorLDeletingRecord)
     else
     if (FHandle.ErrorCode = ACR_ERR_DELETE_RECORD_MODIFIED) then
      DatabaseError(ErrorLDeleteRecordModified)
     else
     if (FHandle.ErrorCode = ACR_ERR_DELETE_RECORD_DELETED) then
      DatabaseError(ErrorLDeleteRecordDeleted);
    end
   else
    begin
     FDeleteRecordFlag := True;
     if (IsAfterDeleteRecordAssigned) then
      DoAfterDeleteRecord;
    end;
 finally
   if (IsBeforeDeleteRecordAssigned or IsAfterDeleteRecordAssigned) then
     ClearValues(FOldFieldValues);
 end;
end; // InternalDelete


//------------------------------------------------------------------------------
// create old and new field values
//------------------------------------------------------------------------------
procedure TACRDataSet.CreateOldNewFieldValues;
var i: Integer;
begin
  FreeOldNewFieldValues;
  SetLength(FOldFieldValues,FAdvFieldDefs.Count);
  SetLength(FNewFieldValues,FAdvFieldDefs.Count);
  for i := 0 to FAdvFieldDefs.Count-1 do
   begin
    FOldFieldValues[i] := TACRVariant.Create;
    FNewFieldValues[i] := TACRVariant.Create;
   end;
end; // CreateOldNewFieldValues


//------------------------------------------------------------------------------
// free old and new field values
//------------------------------------------------------------------------------
procedure TACRDataSet.FreeOldNewFieldValues;
var i: Integer;
begin
  if (FOldFieldValues <> nil) then
   try
    for i := 0 to High(FOldFieldValues) do
     FOldFieldValues[i].Free;
   finally
     FOldFieldValues := nil;
   end;
  if (FNewFieldValues <> nil) then
   try
    for i := 0 to High(FNewFieldValues) do
     FNewFieldValues[i].Free;
   finally
     FNewFieldValues := nil;
   end;
end; // FreeOldNewFieldValues


//------------------------------------------------------------------------------
// save values
//------------------------------------------------------------------------------
procedure TACRDataSet.SaveValues(var Values: TACRArrayOfTACRVariant);
var i: Integer;
begin
  ValuesChangedByEventHandler := False;
  for i := 0 to FAdvFieldDefs.Count-1 do
   GetFieldValue(Values[i],i,False,True);
end; // SaveValues


//------------------------------------------------------------------------------
// clear values
//------------------------------------------------------------------------------
procedure TACRDataSet.ClearValues(var Values: TACRArrayOfTACRVariant);
var i: Integer;
begin
  for i := Low(Values) to High(Values) do
    Values[i].Clear;
end; // ClearValues


//------------------------------------------------------------------------------
// return true if database component for this table has event handler BeforeInsertRecord
//------------------------------------------------------------------------------
function TACRDataSet.IsBeforeInsertRecordAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.BeforeInsertRecord);
end; // IsBeforeInsertRecordAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler AfterInsertRecord
//------------------------------------------------------------------------------
function TACRDataSet.IsAfterInsertRecordAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.AfterInsertRecord);
end; // IsAfterInsertRecordAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler BeforeUpdateRecord
//------------------------------------------------------------------------------
function TACRDataSet.IsBeforeUpdateRecordAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.BeforeUpdateRecord);
end; // IsBeforeUpdateRecordAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler AfterUpdateRecord
//------------------------------------------------------------------------------
function TACRDataSet.IsAfterUpdateRecordAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.AfterUpdateRecord);
end; // IsAfterUpdateRecordAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler BeforeDeleteRecord
//------------------------------------------------------------------------------
function TACRDataSet.IsBeforeDeleteRecordAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.BeforeDeleteRecord);
end; // IsBeforeDeleteRecordAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler AfterDeleteRecord
//------------------------------------------------------------------------------
function TACRDataSet.IsAfterDeleteRecordAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.AfterDeleteRecord);
end; // IsAfterDeleteRecordAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler BeforeExecuteSQL
//------------------------------------------------------------------------------
function TACRDataSet.IsBeforeExecuteSQLAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.BeforeExecuteSQL);
end; // IsBeforeExecuteSQLAssigned


//------------------------------------------------------------------------------
// return true if database component for this table has event handler AfterExecuteSQL
//------------------------------------------------------------------------------
function TACRDataSet.IsAfterExecuteSQLAssigned: Boolean;
begin
  Result := False;
  if (FDatabase <> nil) then
   Result := Assigned(FDatabase.AfterExecuteSQL);
end; // IsAfterExecuteSQLAssigned


//------------------------------------------------------------------------------
// before insert record
//------------------------------------------------------------------------------
function TACRDataSet.DoBeforeInsertRecord: Boolean;
var TableName: String;
begin
  TableName := ErrorLUnknownTable;
  if (FHandle <> nil) then
   TableName := FHandle.TableName;
  FDatabase.DoBeforeInsertRecord(FLogComponent,TableName,FNewFieldValues,Result);
  Result := not Result;
end; // DoBeforeInsertRecord


//------------------------------------------------------------------------------
// after insert record
//------------------------------------------------------------------------------
procedure TACRDataSet.DoAfterInsertRecord;
var TableName: String;
begin
  TableName := ErrorLUnknownTable;
  if (FHandle <> nil) then
   TableName := FHandle.TableName;
  FDatabase.DoAfterInsertRecord(FLogComponent,TableName,FNewFieldValues);
end; // DoAfterInsertRecord


//------------------------------------------------------------------------------
// before update record
//------------------------------------------------------------------------------
function TACRDataSet.DoBeforeUpdateRecord: Boolean;
var TableName: String;
begin
  TableName := ErrorLUnknownTable;
  if (FHandle <> nil) then
   TableName := FHandle.TableName;
  FDatabase.DoBeforeUpdateRecord(FLogComponent,TableName,FOldFieldValues,FNewFieldValues,Result);
  Result := not Result;
end; // DoBeforeUpdateRecord


//------------------------------------------------------------------------------
// after update record
//------------------------------------------------------------------------------
procedure TACRDataSet.DoAfterUpdateRecord;
var TableName: String;
begin
  TableName := ErrorLUnknownTable;
  if (FHandle <> nil) then
   TableName := FHandle.TableName;
  FDatabase.DoAfterUpdateRecord(FLogComponent,TableName,FOldFieldValues,FNewFieldValues);
end; // DoAfterUpdateRecord


//------------------------------------------------------------------------------
// before delete record
//------------------------------------------------------------------------------
function TACRDataSet.DoBeforeDeleteRecord: Boolean;
var TableName: String;
begin
  TableName := ErrorLUnknownTable;
  if (FHandle <> nil) then
   TableName := FHandle.TableName;
  FDatabase.DoBeforeDeleteRecord(FLogComponent,TableName,FOldFieldValues,Result);
  Result := not Result;
end; // DoBeforeDeleteRecord


//------------------------------------------------------------------------------
// after delete record
//------------------------------------------------------------------------------
procedure TACRDataSet.DoAfterDeleteRecord;
var TableName: String;
begin
  TableName := ErrorLUnknownTable;
  if (FHandle <> nil) then
   TableName := FHandle.TableName;
  FDatabase.DoAfterDeleteRecord(FLogComponent,TableName,FOldFieldValues);
end; // DoAfterDeleteRecord


//------------------------------------------------------------------------------
// delete visible records
//------------------------------------------------------------------------------
procedure TACRDataSet.DeleteVisibleRecords;
begin
 if (not Active) then
  raise EACRException.Create(11304,ErrorLDatasetIsNotOpened);
 if (ReadOnly) then
  raise EACRException.Create(11305,ErrorLDatasetIsInReadOnlyMode);
 if (FHandle = nil) then
  raise EACRException.Create(11306,ErrorLNilPointer);
 First;
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
 FHandle.DeleteVisibleRecords;
end; // DeleteVisibleRecords


//------------------------------------------------------------------------------
// update visible records
//------------------------------------------------------------------------------
procedure TACRDataSet.UpdateVisibleRecords(
                              FieldNames:   TStringList;
                              values:       array of TACRVariant
                             );
begin
 if (not Active) then
  raise EACRException.Create(11463,ErrorLDatasetIsNotOpened);
 if (ReadOnly) then
  raise EACRException.Create(11464,ErrorLDatasetIsInReadOnlyMode);
 if (FHandle = nil) then
  raise EACRException.Create(11465,ErrorLNilPointer);
 First;
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
 FHandle.UpdateVisibleRecords(FieldNames,values,False);
end; // UpdateVisibleRecords


//------------------------------------------------------------------------------
// internal handle exception
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalHandleException;
begin
 Application.HandleException(Self);
end; // InternalHandleException


//------------------------------------------------------------------------------
// is cusor open
//------------------------------------------------------------------------------
function TACRDataSet.IsCursorOpen: Boolean;
begin
  Result := (FHandle <> nil);

end;// IsCursorOpen


//------------------------------------------------------------------------------
// internal open
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalOpen;
begin

  if (FHandle <> nil) then
    begin
      BookmarkSize := sizeOf(TACRBookmarkInfo);
      try
        FieldDefs.Updated := False;
        FieldDefs.Update;
        GetIndexInfo;

        if (FEditRecordBuffer <> nil) then
         FreeRecordBuffer(FEditRecordBuffer);
        if DefaultFields then
          CreateFields;
        BindFields(true);
      except
        if (FHandle <> nil) then
         DestroyHandle;
        raise;
      end;
      // for OnFilterRecord
      FHandle.Dataset := TACRDataset(Self);
      if (FHandle.RecordBufferSize + CalcFieldsSize > FHandle.RecordBufferSize) then
       FHandle.RecordBufferSize := FHandle.RecordBufferSize + CalcFieldsSize;
      if (Filtered) then
       ActivateFilters
      else
       DeactivateFilters;
      AllocKeyBuffers;
      ResetCursorRange;
      PrepareCursor;
      FInsertOrEditComplete := False;
    end;

end;// InternalOpen


//------------------------------------------------------------------------------
// internal close
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalClose;
begin

 FreeOldNewFieldValues;
 if (FHandle <> nil) then
  begin
   if (FEditRecordBuffer <> nil) then
    FreeRecordBuffer(FEditRecordBuffer);
   FreeKeyBuffers;
  end;

 BindFields(False);
 if DefaultFields then
  DestroyFields;
 FIndexFieldCount := 0;
 FIndexFieldMap := nil;
 FKeySize := 0;
 FIsProjectionSet := False;

end; // InternalClose


//------------------------------------------------------------------------------
// init field defs
//------------------------------------------------------------------------------
procedure TACRDataSet.InternalInitFieldDefs;
begin

   FieldDefs.Clear;

   if (FHandle <> nil) then
    begin
     FHandle.IsDesignMode := IsDesignMode;
     FHandle.InternalInitFieldDefs;
     // Fill ACRFieldDefs
     FACRFieldDefs.Assign(FHandle.VisibleFieldDefs);
     // Fill FieldDefs
     ConvertACRFieldDefsToFieldDefs(FHandle.VisibleFieldDefs, FHandle.FConstraintDefs, FieldDefs);
     //Fill AdvFieldDefs
     ConvertACRFieldDefsToAdvFieldDefs(FHandle.VisibleFieldDefs,FHandle.FieldDefs, FHandle.FConstraintDefs, AdvFieldDefs);
     // fill foreign key defs
     ConvertConstraintDefsToForeignKeyDefs(FHandle.FConstraintDefs,FForeignKeyDefs);
     FACRConstraintDefs.Assign(FHandle.ConstraintDefs);
     FRestructureFieldDefs.Assign(FAdvFieldDefs);
     FRestructureForeignKeyDefs.Assign(FForeignKeyDefs);
     CreateOldNewFieldValues;
    end;


end;// InternalInitFieldDefs


//------------------------------------------------------------------------------
// get calc field no
//------------------------------------------------------------------------------
function TACRDataset.GetCalcFieldNo(Field: TField): Integer;
var i: Integer;
begin
 Result := 0;
 for i := 0 to Field.FieldNo-2 do
  if (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup) then
   Inc(Result);
end; // GetCalcFieldNo


//------------------------------------------------------------------------------
// read field data to current record buffer
//------------------------------------------------------------------------------
function TACRDataset.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var RecordBuffer: PChar;
    aft:          TACRAdvancedFieldType;
    t:            TDateTime;
    xs:           Single;
    xext:         Extended;
begin
 Result := False;
 RecordBuffer := GetActiveRecordBuffer;
 if (not Active) or (RecordBuffer = nil) then Exit;

 if (FHandle = nil) then
  raise EACRException.Create(10050,ErrorLNilPointer);
 // this is a calculated or lookup field
 if (Field.FieldNo < 0) or (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup) then
  begin
   Inc(RecordBuffer,FHandle.CalculatedFieldsOffset + Field.Offset);
   Result := Boolean(RecordBuffer[0]);
   if ((Result) and (Buffer <> nil)) then
    begin
     Move(RecordBuffer[1],Buffer^,Field.DataSize);
    end;
  end
 else
  begin
   aft := FAdvFieldDefs[Field.FieldNo-1].FDataType;
   if (aft = aftSingle) then
     Result := FHandle.GetFieldData(Field.FieldNo-1,@xs,RecordBuffer)
   else
   if (aft = aftExtended) then
     Result := FHandle.GetFieldData(Field.FieldNo-1,@xext,RecordBuffer)
   else
     Result := FHandle.GetFieldData(Field.FieldNo-1,Buffer,RecordBuffer);


   // Convert
   //  Single --> Double
   //  Extended --> Double
   if (Result and (Buffer <> nil)) then
    begin
     aft := FAdvFieldDefs[Field.FieldNo-1].FDataType;
     case aft of
       aftSingle:     pDouble(Buffer)^ := xs;
       aftExtended:   pDouble(Buffer)^ := xext;
     end;
    end;

   // call data convert
   if (Result and (Buffer <> nil)) then
    if (Field.DataType in [ftDate,ftTime,ftDateTime]) then
     begin
        InternalDataConvert(Field,Buffer,@t,False);
        case (Field.DataType) of
         ftDate: TDateTimeRec(Buffer^).Date := Integer(Trunc(t)) + DateDelta;
         ftTime: TDateTimeRec(Buffer^).Time := Integer(Round(Frac(t) * MSecsPerDay));
        else
         TDateTimeRec(Buffer^).DateTime := (t + DateDelta) * MSecsPerDay;
        end;
     end;
  end;

end; // GetFieldData


//------------------------------------------------------------------------------
// write field data from buffer to current record buffer
//------------------------------------------------------------------------------
procedure TACRDataset.SetFieldData(Field: TField; Buffer: Pointer);

 procedure Finalize;
 begin
  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
   DataEvent(deFieldChange, Longint(Field));
 end; // Finalize

var
    RecordBuffer: PChar;
    t: TDateTime;
    aft:  TACRAdvancedFieldType;
begin


 if not (State in dsWriteModes) then
  raise EACRException.Create(10007, ErrorLDatasetIsNotInEditOrInsertMode);

 if ((State = dsSetKey) and
     ((Field.FieldNo < 0) or
      (FIndexFieldCount > 0) and not Field.IsIndexField)) then
  raise EACRException.Create(20019, ErrorANotIndexField, [Field.DisplayName]);

 RecordBuffer := GetActiveRecordBuffer;

 if (FHandle = nil) then
  raise EACRException.Create(10051,ErrorLNilPointer);
 ValuesChangedByEventHandler := True;
 // this is a calculated or lookup field
 if (Field.FieldNo < 0) or (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup) then
  begin
    Inc(RecordBuffer,FHandle.CalculatedFieldsOffset + Field.Offset);
    Boolean(RecordBuffer[0]) := (Buffer <> nil);
    if Boolean(RecordBuffer[0]) then
     Move(Buffer^,PChar(@RecordBuffer[1])^,Field.DataSize);
    Finalize;
  end
 else
  begin
   Field.Validate(Buffer);
   // call data convert
   if (Buffer <> nil) then
    if (Field.DataType in [ftDate,ftTime,ftDateTime]) then
     begin
        case (Field.DataType) of
         ftDate: t := TDateTimeRec(Buffer^).Date - DateDelta;
         ftTime: t := TDateTimeRec(Buffer^).Time / MSecsPerDay;
          else    t := (TDateTimeRec(Buffer^).DateTime / MSecsPerDay) - DateDelta;
        end;
        InternalDataConvert(Field,@t,Buffer,True);
     end;

   // Convert
   //  Double --> Single
   //  Extended --> Double
   if (Buffer <> nil) then
    begin
     aft := FAdvFieldDefs[Field.FieldNo-1].FDataType;
     case aft of
       aftSingle:
            begin
              //pSingle(Buffer)^ := Single(pDouble(Buffer)^);
              pSingle(Buffer)^ := StrToFloat(FloatToStr(pDouble(Buffer)^));
            end;
       aftExtended:
            begin
              pExtended(Buffer)^ := pDouble(Buffer)^;
            end;
     end;
    end;
   FHandle.SetFieldData(Field.FieldNo-1,Buffer,RecordBuffer);
   Finalize;
  end;

end; // SetFieldData


//------------------------------------------------------------------------------
// get field value
//------------------------------------------------------------------------------
procedure TACRDataset.GetFieldValue(Value: TACRVariant; FieldNo: Integer; DirectAccess: Boolean; CopyFlag: Boolean = True);
var
  Buffer:   PChar;
  bs:       TACRBlobStream;
  BaseType: TACRBaseFieldType;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10315,ErrorLNilPointer);
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
 BaseType := AdvancedFieldTypeToBaseFieldType(AdvFieldDefs[FieldNo].FDataType);
 if (IsBlobFieldType(AdvFieldDefs[FieldNo].FDataType)) then
   begin
     Value.SetNull(BaseType);
     bs := TACRBlobStream.Create(TBlobField(Fields[FieldNo]), bmRead);
     try
      if (bs.Size > 0) then
       begin
         Buffer := MemoryManager.GetMem(bs.Size);
         try
           bs.ReadBuffer(Buffer^, bs.Size);
           Value.SetData(Buffer, bs.Size, BaseType);
         finally
           MemoryManager.FreeAndNilMem(Buffer);
         end;
       end;//if
     finally
       bs.Free;
     end;
   end
 else
   begin
     FHandle.GetFieldValue(Value,FieldNo,DirectAccess,CopyFlag);
   end;
end; // GetFieldValue


//------------------------------------------------------------------------------
// set field value
//------------------------------------------------------------------------------
procedure TACRDataset.SetFieldValue(Value: TACRVariant; FieldNo: Integer; DirectAccess: Boolean);
var
  bs: TACRBlobStream;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10316,ErrorLNilPointer);
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;

 if (IsBlobFieldType(AdvFieldDefs[FieldNo].FDataType)) then
  begin
   if (not Value.IsNull) then
    begin
     bs := TACRBlobStream.Create(TBlobField(Fields[FieldNo]), bmWrite);
     try
       bs.WriteBuffer(Value.pData^, MemoryManager.GetMemoryBufferSize(Value.pData));
     finally
       bs.Free;
     end
    end
   else
    FHandle.SetFieldValue(Value,FieldNo,DirectAccess);
  end
 else
   begin
    FHandle.SetFieldValue(Value,FieldNo,DirectAccess);
   end;
end; // SetFieldValue


//------------------------------------------------------------------------------
// lock table in S mode and load internal data for fast navigation
//------------------------------------------------------------------------------
procedure TACRDataset.LockTable;
begin
  CheckActive;
  FHandle.LockTableForRead;
end; // LockTable


//------------------------------------------------------------------------------
// unlock table in S mode and free internal data
//------------------------------------------------------------------------------
procedure TACRDataset.UnlockTable;
begin
  CheckActive;
  FHandle.UnlockTableForRead;
end; // LockTable


//------------------------------------------------------------------------------
// create Blob stream
//------------------------------------------------------------------------------
function TACRDataset.InternalCreateBlobStream(
    					Field:  TField;
              Mode:   TBlobStreamMode
              ): TACRStream;
var OpenMode: TACRBlobOpenMode;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10108,ErrorLNilPointer);
 if (
      ((Mode = bmReadWrite) or (Mode = bmWrite)) and
      (State <> dsInsert) and (State <> dsEdit)
    ) then
  raise EACRException.Create(10111,ErrorLDatasetIsNotInEditOrInsertMode);

 OpenMode := bomRead;
 case Mode of
  bmRead:
    OpenMode := bomRead;
  bmReadWrite:
    OpenMode := bomReadWrite;
  bmWrite:
    OpenMode := bomWrite;
 end;
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer;
 if (FHandle.CurrentRecordBuffer = nil) then
  begin
   Result := FHandle.InternalCreateBlobStream(True,
    Field.FieldNo-1,OpenMode);
  end
 else
  begin
   Move(PChar(FHandle.CurrentRecordBuffer + FHandle.BookmarkOffset)^,
    FHandle.CurrentRecordID,sizeof(TACRRecordID));
   Result := FHandle.InternalCreateBlobStream((State = dsInsert),
    Field.FieldNo-1,OpenMode);
  end;
end; // InternalCreateBlobStream


//------------------------------------------------------------------------------
// create TACRBlobStream
//------------------------------------------------------------------------------
function TACRDataset.CreateBlobStream(
    					Field:  TField;
              Mode:   TBlobStreamMode
              ): TStream;
begin
 if ((Mode <> bmRead) and (TACRDataset(Field.DataSet).ReadOnly)) then
  raise EACRException.Create(10107,ErrorLDatasetIsInReadOnlyMode);
 Result := TACRBlobStream.Create(TBlobField(Field),Mode);
 if (Mode <> bmRead) then
  ValuesChangedByEventHandler := True;
end; // CreateBlobStream


//------------------------------------------------------------------------------
// close Blob stream, write Blob field value to Blob data file
//------------------------------------------------------------------------------
procedure TACRDataset.CloseBlob(Field: TField);
begin
 if (FHandle = nil) then
  raise EACRException.Create(10108,ErrorLNilPointer);
 FHandle.CurrentRecordBuffer := GetActiveRecordBuffer; 
 FHandle.InternalCloseBlob(Field.FieldNo-1);
end; // CloseBlob


//------------------------------------------------------------------------------
// allocate record buffer
//------------------------------------------------------------------------------
function TACRDataset.AllocRecordBuffer: PChar;
begin

 if (FHandle = nil) then
  raise EACRException.Create(10052,ErrorLNilPointer);
 Result := FHandle.AllocateRecordBuffer;
end; // AllocRecordBuffer


//------------------------------------------------------------------------------
// free record buffer
//------------------------------------------------------------------------------
procedure TACRDataset.FreeRecordBuffer(var Buffer: PChar);
begin

 if (FHandle = nil) then
  raise EACRException.Create(10053,ErrorLNilPointer);
 FHandle.FreeRecordBuffer(Buffer);
end; // FreeRecordBuffer


//------------------------------------------------------------------------------
// initialize record buffer
//------------------------------------------------------------------------------
procedure TACRDataset.InternalInitRecord(Buffer: PChar);
begin
 if (FHandle = nil) then
  raise EACRException.Create(10054,ErrorLNilPointer);
 FHandle.InternalInitRecord(Buffer,True);
end;


//------------------------------------------------------------------------------
// return record size in bytes
//------------------------------------------------------------------------------
function TACRDataset.GetRecordSize: Word;
begin
 if (FHandle = nil) then
  raise EACRException.Create(10055,ErrorLNilPointer);
 Result := FHandle.RecordSize;
end; // Get Record Size


//------------------------------------------------------------------------------
// return true if range is applied
//------------------------------------------------------------------------------
function TACRDataset.IsRangeApplied: Boolean;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10298,ErrorLNilPointer);
  Result := FHandle.IsRangeApplied;
end; // IsRangeApplied


//------------------------------------------------------------------------------
// return true if distinct is applied
//------------------------------------------------------------------------------
function TACRDataset.IsDistinctApplied: Boolean;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10326,ErrorLNilPointer);
  Result := (FHandle.DistinctFieldCount > 0);
end; // IsDistinctApplied


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRDataset.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFDEF SQLMEMTABLE}
  FDatabaseName := ACRMemoryDatabaseName; FSessionName := ''; FInMemory := True;
  {$ENDIF}
  FOldFieldValues := nil;
  FNewFieldValues := nil;
  FLogComponent := Self;
  ACRDatasetsList := ACRDatasets.LockList;
  ACRDatasetsList.Add(Self);
  ACRDatasets.UnlockList;
  FDatabase := nil;
  FExternalHandle := nil;
  FIsProjectionSet := False;
  FHandle := nil;
  FEditRecordBuffer := nil;
  FIndexDefs := TIndexDefs.Create(Self);
  FACRFieldDefs := TACRFieldDefs.Create;
  FAdvIndexDefs := TACRIndexDefs.Create;
  FAdvFieldDefs := TACRAdvFieldDefs.Create;
  FForeignKeyDefs := TACRForeignKeyDefs.Create;
  FRestructureIndexDefs := TIndexDefs.Create(Self);
  FRestructureFieldDefs := TACRAdvFieldDefs.Create;
  FRestructureForeignKeyDefs := TACRForeignKeyDefs.Create;
  FACRConstraintDefs := TACRConstraintDefs.Create;
  FInsertOrEditComplete := False;
  if (not IsDesignMode) then
   if (AOwner <> nil) then
    if (csDesigning in AOwner.ComponentState) then
     begin
      IsDesignMode := true;
     end;
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRDataset.Destroy;
begin
 if (FEditRecordBuffer <> nil) then
  FreeRecordBuffer(FEditRecordBuffer);
 Active := false;
 ACRDatasetsList := ACRDatasets.LockList;
 ACRDatasetsList.Remove(Self);
 ACRDatasets.UnlockList;
 inherited Destroy;
 FIndexDefs.Free;
 FACRFieldDefs.Free;
 FAdvIndexDefs.Free;
 FAdvFieldDefs.Free;
 FForeignKeyDefs.Free;
 FRestructureIndexDefs.Free;
 FRestructureFieldDefs.Free;
 FRestructureForeignKeyDefs.Free;
 FACRConstraintDefs.Free;
end; // Destroy


//------------------------------------------------------------------------------
// open database
//------------------------------------------------------------------------------
function TACRDataset.OpenDatabase: TACRDatabase;
begin
  with Sessions.List[FSessionName] do
    Result := DoOpenDatabase(FDatabaseName, Self.Owner);
end;// OpenDatabase


//------------------------------------------------------------------------------
// close database
//------------------------------------------------------------------------------
procedure TACRDataset.CloseDatabase(Database: TACRDatabase);
begin
  if Assigned(Database) then
    Database.Session.CloseDatabase(Database);
end;// CloseDatabase



////////////////////////////////////////////////////////////////////////////////
//
//  TACRTable
//
////////////////////////////////////////////////////////////////////////////////


{$IFDEF D6H}
//------------------------------------------------------------------------------
// get default order
//------------------------------------------------------------------------------
function TACRTable.PSGetDefaultOrder: TIndexDef;

  function GetIdx(IdxType: TIndexOption): TIndexDef;
  var
    i: Integer;
  begin
    Result := nil;
    for i := 0 to IndexDefs.Count - 1 do
      if (IdxType in IndexDefs[i].Options) then
      try
        Result := IndexDefs[i];
        GetFieldList(nil, Result.Fields);
        break;
      except
        Result := nil;
      end;
  end;

var
  DefIdx: TIndexDef;
begin
  DefIdx := nil;
  IndexDefs.Update;
  try
    if IndexName <> '' then
      DefIdx := IndexDefs.Find(IndexName)
    else
     if IndexFieldNames <> '' then
      DefIdx := IndexDefs.FindIndexForFields(IndexFieldNames);
    if Assigned(DefIdx) then
      GetFieldList(nil, DefIdx.Fields);
  except
    DefIdx := nil;
  end;
  if not Assigned(DefIdx) then
    DefIdx := GetIdx(ixPrimary);
  if not Assigned(DefIdx) then
    DefIdx := GetIdx(ixUnique);
  if Assigned(DefIdx) then
   begin
    Result := TIndexDef.Create(nil);
    Result.Assign(DefIdx);
   end
  else
   Result := nil;
end; // PSGetDefaultOrder


//------------------------------------------------------------------------------
// get key fields
//------------------------------------------------------------------------------
function TACRTable.PSGetKeyFields: string;
var
  i, Pos: Integer;
  IndexFound: Boolean;
begin
  Result := inherited PSGetKeyFields;
  if Result = '' then
   begin
    if not Exists then
     Exit;
    IndexFound := False;
    IndexDefs.Update;
    for i := 0 to IndexDefs.Count - 1 do
      if ixUnique in IndexDefs[I].Options then
       begin
        Result := IndexDefs[I].Fields;
        IndexFound := (FieldCount = 0);
        if not IndexFound then
         begin
          Pos := 1;
          while Pos <= Length(Result) do
           begin
            IndexFound := FindField(ExtractFieldName(Result, Pos)) <> nil;
            if not IndexFound then
             Break;
           end;
         end;
        if IndexFound then
         Break;
       end;
    if not IndexFound then
     Result := '';
   end;
end; // PSGetKeyFields


//------------------------------------------------------------------------------
// get table name
//------------------------------------------------------------------------------
function TACRTable.PSGetTableName: string;
begin
  Result := TableName;
end; // PSGetTableName


//------------------------------------------------------------------------------
// get index defs
//------------------------------------------------------------------------------
function TACRTable.PSGetIndexDefs(IndexTypes: TIndexOptions): TIndexDefs;
begin
  Result := GetIndexDefs(IndexDefs, IndexTypes);
end; // PSGetIndexDefs


//------------------------------------------------------------------------------
// set command text
//------------------------------------------------------------------------------
procedure TACRTable.PSSetCommandText(const CommandText: string);
begin
  if CommandText <> '' then
    TableName := CommandText;
end; // PSSetCommandText


//------------------------------------------------------------------------------
// set params
//------------------------------------------------------------------------------
procedure TACRTable.PSSetParams(AParams: TParams);

  procedure AssignFields;
  var
    I: Integer;
  begin
    for I := 0 to AParams.Count - 1 do
      if AParams[I].Name <> '' then
        FieldByName(AParams[I].Name).Value := AParams[I].Value
      else
        IndexFields[I].Value := AParams[I].Value;
  end;

begin
  if AParams.Count > 0 then
   begin
    Open;
    SetRangeStart;
    AssignFields;
    SetRangeEnd;
    AssignFields;
    ApplyRange;
   end
  else
   if Active then
    CancelRange;
  PSReset;
end; // PSSetParams
{$ENDIF}


//------------------------------------------------------------------------------
// check table name is blank?
//------------------------------------------------------------------------------
procedure TACRTable.CheckBlankTableName;
begin
  if (FTableName = '') then
    DatabaseError(ErrorANoTableName, Self);
end;// CheckBlankTableName


//------------------------------------------------------------------------------
// set temporary
//------------------------------------------------------------------------------
procedure TACRTable.SetTemporary(const Value: Boolean);
begin
  CheckInactive;
  FTemporary := Value;
  if (Value) then
   FDatabaseName := ACRTemporaryDatabaseName;
end;// SetTemporary


//------------------------------------------------------------------------------
// GetIndexFieldNames
//------------------------------------------------------------------------------
function TACRTable.GetIndexFieldNames: string;
begin
  if FFieldsIndex then
    Result := FIndexName
  else
    Result := '';
end;// GetIndexFieldNames


//------------------------------------------------------------------------------
// GetIndexName
//------------------------------------------------------------------------------
function TACRTable.GetIndexName: string;
begin
  if FFieldsIndex then
    Result := ''
  else
    Result := FIndexName;
end;// GetIndexName


//------------------------------------------------------------------------------
// GetIndexParams
//------------------------------------------------------------------------------
procedure TACRTable.GetIndexParams(const IndexName: string; FieldsIndex: Boolean;
         var IndexedName: string);
var
  IndexStr: string;
begin
  IndexStr := '';
  if (IndexName <> '') then
   begin
     FIndexDefs.Updated := False;
     FIndexDefs.Update;
     IndexStr := IndexName;
     if FieldsIndex then
       IndexStr := FIndexDefs.FindIndexForFields(IndexName).Name;
   end;
  IndexedName := IndexStr;
end;// GetIndexParams


//------------------------------------------------------------------------------
// IndexDefsStored
//------------------------------------------------------------------------------
function TACRTable.IndexDefsStored: Boolean;
begin
  Result := (FStoreDefs and (IndexDefs.Count > 0));
end;// IndexDefsStored


//------------------------------------------------------------------------------
// SetIndex
//------------------------------------------------------------------------------
procedure TACRTable.SetIndex(const Value: string; FieldsIndex: Boolean);
var
  IndexName: string;
begin
  if Active then
    CheckBrowseMode;
  if (FIndexName <> Value) or (FFieldsIndex <> FieldsIndex) then
    begin
      if Active then
        begin
         GetIndexParams(Value, FieldsIndex,IndexName);
         SwitchToIndex(IndexName);
         CheckMasterRange;
        end;
      FIndexName := Value;
      FFieldsIndex := FieldsIndex;
      if Active then
       Resync([]);
    end;
end;// SetIndex


//------------------------------------------------------------------------------
// SetIndexFieldNames
//------------------------------------------------------------------------------
procedure TACRTable.SetIndexFieldNames(const Value: string);
begin
  SetIndex(Value, Value <> '');
end;// SetIndexFieldNames


//------------------------------------------------------------------------------
// SetIndexName
//------------------------------------------------------------------------------
procedure TACRTable.SetIndexName(const Value: string);
begin
  SetIndex(Value, False);
end;// SetIndexName


//------------------------------------------------------------------------------
// create cursor: local or client
//------------------------------------------------------------------------------
function TACRTable.CreateCursor: TACRCursor;
begin
  Result := nil;
  if (FDatabase = nil) then
    raise EACRException.Create(10915,ErrorLNilPointer);
  {$IFDEF LOCAL_VERSION}
  if (FDatabase.LocalDatabase) then
    Result := TACRLocalCursor.Create;
  {$ENDIF}
  {$IFDEF CLIENT_VERSION}
  if (not FDatabase.LocalDatabase) then
    Result := TACRClientCursor.Create;
  {$ENDIF}
end; // CreateCursor


//------------------------------------------------------------------------------
// progress event
//------------------------------------------------------------------------------
procedure TACRTable.DoOnProgress(
                      Progress:   Double;
                      Operation:  TACRTableOperation;
                      var Abort:  Boolean
                     );
begin
 if (Assigned(FOnProgress)) then
  FOnProgress(Self,Progress,Operation,Abort);
end; // DoOnProgress


//------------------------------------------------------------------------------
// get index names
//------------------------------------------------------------------------------
procedure TACRTable.GetIndexNames(List: TStrings);
begin
  IndexDefs.Update;
  IndexDefs.GetItemNames(List);
end; // GetIndexNames


//------------------------------------------------------------------------------
// return index name of the index or '' if not found
//------------------------------------------------------------------------------
function TACRTable.FindIndex(FieldNamesList, AscDescList, CaseSensitivityList: TStringList): String;
begin
  if (FHandle = nil) then
    raise EACRException.Create(11382,ErrorLNilPointer);
  Result := FHandle.FindIndex(FieldNamesList, AscDescList, CaseSensitivityList);
end; // FindIndex



//------------------------------------------------------------------------------
// return index name
//------------------------------------------------------------------------------
function TACRTable.FindOrCreateIndex(FieldNamesList, AscDescList, CaseSensitivityList: TStringList): String;
begin
  raise EACRException.Create(11391,ErrorLOperationIsNotSupported);
  if (FHandle = nil) then
    raise EACRException.Create(10311,ErrorLNilPointer);
  Result := FHandle.FindOrCreateIndex(FieldNamesList, AscDescList, CaseSensitivityList);
end; // FindOrCreateIndex



//------------------------------------------------------------------------------
// apply distinct
//------------------------------------------------------------------------------
procedure TACRTable.ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList: TStringList);
begin
  raise EACRException.Create(11392,ErrorLOperationIsNotSupported);
  if (FHandle = nil) then
    raise EACRException.Create(10319,ErrorLNilPointer);
  FHandle.ApplyDistinct(FieldNamesList, AscDescList, CaseSensitivityList);
  IndexName := FHandle.IndexName;
end; // ApplyDistinct


//------------------------------------------------------------------------------
// get table names
//------------------------------------------------------------------------------
procedure TACRTable.GetTableNames(List: TStrings);
begin
 if (DBSession <> nil) then
  DBSession.GetTableNames(FDatabaseName,List);
end; // GetTableNames


//------------------------------------------------------------------------------
// return true if table exists
//------------------------------------------------------------------------------
function TACRTable.GetTableExists: Boolean;
var DatabaseClosed: Boolean;
begin
  Result := False;
  if (FDatabase = nil) then
    DatabaseClosed := True
  else
    DatabaseClosed := False;
  if (DatabaseClosed) then
    FDatabase := OpenDatabase;
  if (FDatabase <> nil) then
   begin
    Result := FDatabase.TableExists(TableName);
    if (DatabaseClosed) then
      begin
        CloseDatabase(FDatabase);
        FDatabase := nil;
      end;
   end;
end; // GetTableExists


//------------------------------------------------------------------------------
// CheckMasterRange
//------------------------------------------------------------------------------
procedure TACRTable.CheckMasterRange;
begin
  if (FMasterLink.Active and (FMasterLink.Fields.Count > 0)) then
  begin
    SetLinkRanges(FMasterLink.Fields);
    SetCursorRange;
  end;
end; // CheckMasterRange


//------------------------------------------------------------------------------
// update range
//------------------------------------------------------------------------------
procedure TACRTable.UpdateRange;
begin
  SetLinkRanges(FMasterLink.Fields);
end; // UpdateRange


//------------------------------------------------------------------------------
// Master changed
//------------------------------------------------------------------------------
procedure TACRTable.MasterChanged(Sender: TObject);
begin
  CheckBrowseMode;
  UpdateRange;
  ApplyRange;
end; // MasterChanged


//------------------------------------------------------------------------------
// Master disabled
//------------------------------------------------------------------------------
procedure TACRTable.MasterDisabled(Sender: TObject);
begin
  CancelRange;
end; // MasterDisabled


//------------------------------------------------------------------------------
// Set data source
//------------------------------------------------------------------------------
procedure TACRTable.SetDataSource(Value: TDataSource);
begin
  if (IsLinkedTo(Value)) then
    DatabaseError(ErrorLCircularDataLink, Self);
  FMasterLink.DataSource := Value;
end; // SetDataSource


//------------------------------------------------------------------------------
// get master fields
//------------------------------------------------------------------------------
function TACRTable.GetMasterFields: string;
begin
  Result := FMasterLink.FieldNames;
end; // GetMasterFields


//------------------------------------------------------------------------------
// set master fields
//------------------------------------------------------------------------------
procedure TACRTable.SetMasterFields(const Value: string);
begin
  FMasterLink.FieldNames := Value;
end; // SetMasterFields


//------------------------------------------------------------------------------
// prepare cursor
//------------------------------------------------------------------------------
procedure TACRTable.PrepareCursor;
begin
  CheckMasterRange;
end;


//------------------------------------------------------------------------------
// create and open cursor
//------------------------------------------------------------------------------
function TACRTable.CreateHandle: TACRCursor;
var
   IndexName: string;
begin

  CheckBlankTableName;
  FIndexDefs.Updated:=False;
  if (FExternalHandle = nil) then
    Result := GetHandle
  else
    Result := FExternalHandle;
  FHandle := Result;
  FIndexDefs.Update;
  GetIndexParams(FIndexName, FFieldsIndex, IndexName);
  if (IndexName <> '') then
    Result.IndexName := IndexName;

end;// CreateHandle


//------------------------------------------------------------------------------
// Data Event
//------------------------------------------------------------------------------
procedure TACRTable.DataEvent(Event: TDataEvent; Info: Longint);
begin
  if ((Event = dePropertyChange) and Assigned(IndexDefs)) then
   IndexDefs.Updated := False;
  inherited DataEvent(Event, Info);
end;// DataEvent


//------------------------------------------------------------------------------
// DefChanged
//------------------------------------------------------------------------------
procedure TACRTable.DefChanged(Sender: TObject);
begin
  StoreDefs := True;
end;// DefChanged


//------------------------------------------------------------------------------
// InitFieldDefs
//------------------------------------------------------------------------------
procedure TACRTable.InitFieldDefs;
var
  TmpCursor: TACRCursor;
begin
  if (FHandle <> nil) then
    InternalInitFieldDefs
  else
    begin
     SetDBFlag(dbfFieldList,True);
     try
       CheckBlankTableName;
       TmpCursor := GetHandle;
       try
         TmpCursor.InternalInitFieldDefs;
         // Fill ACRFieldDefs
         FACRFieldDefs.Assign(TmpCursor.VisibleFieldDefs);
         // Fill FieldDefs
         ConvertACRFieldDefsToFieldDefs(TmpCursor.VisibleFieldDefs, TmpCursor.FConstraintDefs, FieldDefs);
         // Fill AdvFieldDefs
         ConvertACRFieldDefsToAdvFieldDefs(TmpCursor.VisibleFieldDefs,TmpCursor.FieldDefs, TmpCursor.FConstraintDefs, AdvFieldDefs);
         FRestructureFieldDefs.Assign(FAdvFieldDefs);
         FRestructureForeignKeyDefs.Assign(FForeignKeyDefs);
       finally
        TmpCursor.Free;
       end;
     finally
       SetDBFlag(dbfFieldList,False);
     end;
    end;
end;// InitFieldDefs


//------------------------------------------------------------------------------
// destroy cursor
//------------------------------------------------------------------------------
procedure TACRTable.DestroyHandle;
begin

  inherited DestroyHandle;

end;// DestroyHandle


//------------------------------------------------------------------------------
// get cursor
//------------------------------------------------------------------------------
function TACRTable.GetHandle: TACRCursor;
var i: Integer;
begin

  Result := CreateCursor;
  if (Result = nil) then
    raise EACRException.Create(10916,ErrorLCannotCreateCursor);
  try
     Result.Session := FDatabase.Handle;
     Result.TableName := FTableName;
     Result.Exclusive := FExclusive;
     Result.ReadOnly := FReadOnly;
     Result.InMemory := FInMemory;
     Result.MemoryTableAllocBy := FMemoryTableAllocBy;
     Result.Temporary := FTemporary;
     Result.IsDesignMode := IsDesignMode;
     // adding fields, specified by editor
     for i := 0 to Fields.Count - 1 do
      if (Fields[i].FieldKind in [fkData]) then
       if (not FindFieldInFieldDefs(FieldDefs,Fields[i].FieldName)) then
        FieldDefs.Add(Fields[i].FieldName,Fields[i].DataType,Fields[i].Size,Fields[i].Required);
       // convert field and index defs
     if (FieldDefs.Count > 0) then
      begin
        FACRFieldDefs.Clear;
        ConvertFieldDefsToACRFieldDefs(FieldDefs,FACRFieldDefs);
      end;
     if (IndexDefs.Count > 0) then
      begin
       FAdvIndexDefs.Clear;
       ConvertIndexDefsToACRIndexDefs(IndexDefs,FAdvIndexDefs,FieldDefs,AdvFieldDefs);
       FRestructureIndexDefs.Assign(IndexDefs);
      end;
     Result.OpenTableByFieldDefs(FACRFieldDefs,FAdvIndexDefs, FACRConstraintDefs);
     FReadOnly := Result.ReadOnly;
   except
     Result.Free;
     raise;
   end;

end;// GetHandle


//------------------------------------------------------------------------------
// UpdateIndexDefs
//------------------------------------------------------------------------------
procedure TACRTable.UpdateIndexDefs;
var
  TmpCursor: TACRCursor;
  bIsOpen:   Boolean;
begin
 if (not FIndexDefs.Updated) then
  begin
   SetDBFlag(dbfIndexList, True);
   try
     if (FHandle = nil) then
       TmpCursor := GetHandle
     else
       TmpCursor := FHandle;
     try
      bIsOpen := TmpCursor.IsOpen;
      if (not FIsProjectionSet) then
       begin
        if (not bIsOpen) then
         TmpCursor.OpenTableByFieldDefs(nil,nil,nil);
        try
          TmpCursor.UpdateTableDefinitions;
        finally
          if (not bIsOpen) then
           TmpCursor.CloseTable;
        end;
       end;
      FieldDefs.Update;
      ConvertACRIndexDefsToIndexDefs(TmpCursor.IndexDefs, FIndexDefs);
      FAdvIndexDefs.Assign(TmpCursor.IndexDefs);
      FRestructureIndexDefs.Assign(FIndexDefs);
      FIndexDefs.Updated:=True;
     finally
      if (FHandle = nil) then
       TmpCursor.Free;
     end;
   finally
     SetDBFlag(dbfIndexList, False);
   end;
  end;
end;// UpdateIndexDefs


//------------------------------------------------------------------------------
// return field object for the specified field in the current index
//------------------------------------------------------------------------------
function TACRTable.GetIndexField(Index: Integer): TField;
var
  FieldNo: Integer;
begin
  if (Index < 0) or (Index >= FIndexFieldCount) then
    DatabaseError(ErrorLFieldIndexError, Self);
  FieldNo := FIndexFieldMap[Index];
  Result := FieldByNumber(FieldNo);
  if Result = nil then
    DatabaseErrorFmt(ErrorLIndexFieldMissing, [FieldDefs[FieldNo - 1].Name], Self);
end; // GetIndexField


//------------------------------------------------------------------------------
// set index field
//------------------------------------------------------------------------------
procedure TACRTable.SetIndexField(Index: Integer; Value: TField);
begin
  GetIndexField(Index).Assign(Value);
end; // SetIndexField


//------------------------------------------------------------------------------
// Set table name
//------------------------------------------------------------------------------
procedure TACRTable.SetTableName(Value: String);
begin
  if (csReading in ComponentState) or (csDesigning in ComponentState) then
    FTableName := Value
  else
  if (FTableName <> Value) then
   begin
    CheckInactive;
    FTableName := Value;
    DataEvent(dePropertyChange, 0);
   end;
end; // SetTableName


//------------------------------------------------------------------------------
// set link ranges
//------------------------------------------------------------------------------
procedure TACRTable.SetLinkRanges(MasterFields: TList);
var
  i: Integer;
  SaveState: TDataSetState;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10299,ErrorLNilPointer);
  SaveState := SetTempState(dsSetKey);
  try
    FKeyBuffer := InitKeyBuffer(FKeyBuffers[kiRangeStart]);
    PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.Modified := True;
    for i := 0 to MasterFields.Count - 1 do
     if (TField(MasterFields[i]).DataType = ftLargeInt) then
      begin
        if (TField(MasterFields[i]).IsNull) then
         GetIndexField(i).Clear
        else
         TLargeIntField(GetIndexField(i)).AsLargeInt := TLargeIntField(MasterFields[i]).AsLargeInt;
      end
     else
      GetIndexField(i).Assign(TField(MasterFields[i]));
    PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.FieldCount := MasterFields.Count;
  finally
    RestoreState(SaveState);
  end;
  Move(FKeyBuffers[kiRangeStart]^, FKeyBuffers[kiRangeEnd]^, FHandle.KeyBufferSize);
end; // SetLinkRanges


//------------------------------------------------------------------------------
// get data source
//------------------------------------------------------------------------------
function TACRTable.GetDataSource: TDataSource;
begin
  Result := FMasterLink.DataSource;
end; // GetDataSource


//------------------------------------------------------------------------------
// on new record
//------------------------------------------------------------------------------
procedure TACRTable.DoOnNewRecord;
var
  I: Integer;
begin
  if (FMasterLink.Active and (FMasterLink.Fields.Count > 0)) then
    for I := 0 to FMasterLink.Fields.Count - 1 do
      IndexFields[I] := TField(FMasterLink.Fields[I]);
  inherited DoOnNewRecord;
end; // DoOnNewRecord


//------------------------------------------------------------------------------
// get index field count
//------------------------------------------------------------------------------
function TACRTable.GetIndexFieldCount: Integer;
begin
  Result := FIndexFieldCount;
end; // GetIndexFieldCount


//------------------------------------------------------------------------------
// create
//------------------------------------------------------------------------------
constructor TACRTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIsTable := True;
  FTemporary := False;
  FMasterLink := TMasterDataLink.Create(Self);
  FMasterLink.OnMasterChange := MasterChanged;
  FMasterLink.OnMasterDisable := MasterDisabled;
  FMemoryTableAllocBy := ACRDefaultMemoryTableAllocBy;
  {$IFDEF SQLMEMTABLE}
  FTableName := GetTemporaryName('Table');
  {$ENDIF}
end;// Create


//------------------------------------------------------------------------------
// destory
//------------------------------------------------------------------------------
destructor TACRTable.Destroy;
begin
  inherited Destroy;
  FMasterLink.Free;
end; // Destroy


//------------------------------------------------------------------------------
// create table
//------------------------------------------------------------------------------
procedure TACRTable.CreateTable;
var i:          Integer;
    TempCursor: TACRCursor;
begin
  CheckInactive;
  CheckBlankTableName;
  SetDBFlag(dbfTable, True);
  try
   if (FDatabase <> nil) then
    if (FDatabase.ReadOnly) then
     DatabaseError(ErrorLDatabaseFileIsInReadOnlyMode);
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(10919,ErrorLCannotCreateCursor);
   try
    TempCursor.FCreateTableStarted := True;
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.MemoryTableAllocBy := Self.MemoryTableAllocBy;
    // adding fields, specified by editor
    for i := 0 to Fields.Count - 1 do
     if (Fields[i].FieldKind in [fkData]) then
      if (not FindFieldInFieldDefs(FieldDefs,Fields[i].FieldName)) then
        FieldDefs.Add(Fields[i].FieldName,Fields[i].DataType,Fields[i].Size,Fields[i].Required);

    // Synchronize FieldDefs and AdvFieldDefs
    if (AdvFieldDefs.Count > 0) then
      begin
        // Copy AdvFieldDefs to FieldDefs
        ConvertAdvFieldDefsToFieldDefs(AdvFieldDefs, FieldDefs);
      end
    else
     begin
        // Copy FieldDefs to AdvFieldDefs or
        ConvertFieldDefsToAdvFieldDefs(FieldDefs, AdvFieldDefs);
     end;

    if (IndexDefs.Count > 0) then
     begin
      FAdvIndexDefs.Clear;
      ConvertIndexDefsToACRIndexDefs(IndexDefs,FAdvIndexDefs, FieldDefs, AdvFieldDefs);
     end;

    // convert field and index defs
    ConvertAdvFieldDefsToACRFieldDefs(AdvFieldDefs, FACRFieldDefs,
                                      FAdvIndexDefs, FACRConstraintDefs, FTemporary);
    ACRConvertForeignKeyDefsToConstraintDefs(FForeignKeyDefs, FACRConstraintDefs);
    TempCursor.CreateTable(FACRFieldDefs,FAdvIndexDefs,FACRConstraintDefs);
   finally
    TempCursor.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end;// CreateTable


//------------------------------------------------------------------------------
// delete table
//------------------------------------------------------------------------------
procedure TACRTable.DeleteTable(Cascade: Boolean);
var
  TempCursor:   TACRCursor;
begin
  CheckInactive;
  CheckBlankTableName;
  SetDBFlag(dbfTable, True);
  try
    if (FDatabase <> nil) then
     if (FDatabase.ReadOnly) then
      DatabaseError(ErrorLDatabaseFileIsInReadOnlyMode);
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(10920,ErrorLCannotCreateCursor);
   try
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.DeleteTable(Cascade);
   finally
    TempCursor.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end; // DeleteTable


//------------------------------------------------------------------------------
// empty table
//------------------------------------------------------------------------------
procedure TACRTable.EmptyTable;
var
  TempCursor:   TACRCursor;
begin
  CheckInactive;
  SetDBFlag(dbfTable, True);
  try
    if (FDatabase <> nil) then
     if (FDatabase.ReadOnly) then
      DatabaseError(ErrorLDatabaseFileIsInReadOnlyMode);
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(10921,ErrorLCannotCreateCursor);
   try
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.EmptyTable;
   finally
    TempCursor.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end; // EmptyTable


//------------------------------------------------------------------------------
// rename table
//------------------------------------------------------------------------------
procedure TACRTable.RenameTable(NewTableName: String);
var
  TempCursor:   TACRCursor;
begin
  CheckInactive;
  if (FDatabase <> nil) then
   if (FDatabase.ReadOnly) then
    DatabaseError(ErrorLDatabaseFileIsInReadOnlyMode);
  SetDBFlag(dbfTable, True);
  try
   if (FDatabase <> nil) then
    if (FDatabase.ReadOnly) then
     DatabaseError(ErrorLDatabaseFileIsInReadOnlyMode);
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(10922,ErrorLCannotCreateCursor);
   try
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.RenameTable(NewTableName);
    FTableName := NewTableName;
   finally
    TempCursor.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end; // RenameTable


//------------------------------------------------------------------------------
// LoadTable
//------------------------------------------------------------------------------
procedure TACRTable.LoadTableFromFile(
                        FileName:             String
                       );
var
  Stream: TACRStream;
begin
  Stream := TACRFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite); 
  try 
    LoadTableFromStream(Stream);
  finally
    Stream.Free; 
  end; 
end; // LoadTableFromFile
	
 
//------------------------------------------------------------------------------
// SaveTable 
//------------------------------------------------------------------------------
procedure TACRTable.SaveTableToFile( 
                        FileName:             String;
                        CompressionAlgorithm: TCompressionAlgorithm; 
                        CompressionMode:      Byte;
                        BlockSize:            Integer 
                      ); 
var
  Stream: TACRStream; 
begin 
  Stream := TACRFileStream.Create(FileName, fmCreate); 
  try 
    SaveTableToStream(Stream,CompressionAlgorithm,CompressionMode,BlockSize); 
  finally
    Stream.Free;
  end;
end; // SaveTableToFile
	
 
//------------------------------------------------------------------------------
// LoadTable 
//------------------------------------------------------------------------------
procedure TACRTable.LoadTableFromStream( 
                        Stream:               TStream 
                       );
var 
  TempCursor:   TACRCursor;
begin 
  CheckInactive; 
  SetDBFlag(dbfTable, True);
  try
   TempCursor := CreateCursor; 
   if (TempCursor = nil) then 
    raise EACRException.Create(10923,ErrorLCannotCreateCursor); 
   try 
    TempCursor.Session := FDatabase.Handle; 
    TempCursor.TableName := FTableName; 
    TempCursor.Temporary := FTemporary; 
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True; 
    TempCursor.LoadTableFromStream(Stream);
    FTableName := TempCursor.FTableName;
   finally 
    TempCursor.Free;
   end;
  finally 
   SetDBFlag(dbfTable, False); 
  end; 
end; // LoadTableFromStream
	
 
//------------------------------------------------------------------------------
// SaveTable
//------------------------------------------------------------------------------ 
procedure TACRTable.SaveTableToStream( 
                        Stream: TStream;
                        CompressionAlgorithm: TCompressionAlgorithm;
                        CompressionMode:      Byte;
                        BlockSize:            Integer 
                      ); 
var 
  TempCursor:   TACRCursor; 
  CompAlg:      TACRCompressionAlgorithm;
begin 
  CheckInactive; 
  SetDBFlag(dbfTable, True); 
  try
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(10924,ErrorLCannotCreateCursor); 
   try
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary; 
    TempCursor.InMemory := FInMemory; 
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    CompAlg := ConvertCompressionAlgorithmToACRCompressionAlgorithm(CompressionAlgorithm);
    TempCursor.SaveTableToStream(Stream,CompAlg,CompressionMode,BlockSize);
   finally
    TempCursor.Free;
   end; 
  finally
   SetDBFlag(dbfTable, False);
  end; 
end; // SaveTableToStream
	
 
//------------------------------------------------------------------------------ 
// LoadTable 
//------------------------------------------------------------------------------ 
procedure TACRTable.LoadAllTablesFromFile(
                        FileName:             String 
                       ); 
var
  Stream: TACRStream;
begin
  Stream := TACRFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try 
    LoadAllTablesFromStream(Stream); 
  finally
    Stream.Free;
  end; 
end; // LoadTableFromFile 
 
	
//------------------------------------------------------------------------------
// SaveTable
//------------------------------------------------------------------------------
procedure TACRTable.SaveAllTablesToFile(
                        FileName:             String; 
                        CompressionAlgorithm: TCompressionAlgorithm;
                        CompressionMode:      Byte; 
                        BlockSize:            Integer
                      );
var
  Stream: TACRStream;
begin
  Stream := TACRFileStream.Create(FileName, fmCreate);
  try
    SaveAllTablesToStream(Stream,CompressionAlgorithm,CompressionMode,BlockSize);
  finally
    Stream.Free;
  end;
end; // SaveTableToFile


//------------------------------------------------------------------------------
// LoadTable
//------------------------------------------------------------------------------
procedure TACRTable.LoadAllTablesFromStream(
                        Stream:               TStream
                       );
begin
 CheckInactive;
 FDatabase := OpenDatabase;
 if (FDatabase = nil) then
  raise EACRException.Create(11341,ErrorLNilPointer);
 try
  FDatabase.LoadAllTablesFromStream(Stream);
 finally
  CloseDatabase(FDatabase);
  FDatabase := nil;
 end;
end; // LoadTableFromStream


//------------------------------------------------------------------------------
// SaveTable
//------------------------------------------------------------------------------
procedure TACRTable.SaveAllTablesToStream(
                        Stream: TStream;
                        CompressionAlgorithm: TCompressionAlgorithm;
                        CompressionMode:      Byte;
                        BlockSize:            Integer
                      );
begin
 CheckInactive;
 FDatabase := OpenDatabase;
 if (FDatabase = nil) then
  raise EACRException.Create(11340,ErrorLNilPointer);
 try
  FDatabase.SaveAllTablesToStream(Stream,CompressionAlgorithm,CompressionMode,BlockSize);
 finally
  CloseDatabase(FDatabase);
  FDatabase := nil;
 end;
end; // SaveTableToStream


//------------------------------------------------------------------------------
// import table
//------------------------------------------------------------------------------
function TACRTable.ImportTable(
                      SourceTable: TDataset;
                      var Log:     String
                      ): Boolean;
var
    i:               Integer;
    FixSize:         Integer; //fix
    s,SourceLog:     String;
begin
  Result := False;
  SourceLog := Log;
  CheckInactive;
  IndexDefs.Clear;
  AdvIndexDefs.Clear;
  FieldDefs.Clear;
  AdvFieldDefs.Clear;
  if (SourceTable is TACRDataset) then
   begin
    AdvFieldDefs.Assign(TACRDataset(SourceTable).AdvFieldDefs);
    IndexDefs.Assign(TACRDataSet(SourceTable).IndexDefs);
    ForeignKeyDefs.Assign(TACRDataSet(SourceTable).ForeignKeyDefs);
   end
  else
   FieldDefs.Assign(SourceTable.FieldDefs);
  for i := 0 to FieldDefs.Count - 1 do
   begin
    {$IFNDEF SQLMEMTABLE}
     if (FieldDefs[i].DataType = ftString) then
      begin
       FixSize := FieldDefs[i].Size; //fix
      FieldDefs[i].DataType := ftFixedChar;
       FieldDefs[i].Size := FixSize; //fix
      end;
    {$ENDIF}
    if (FieldDefs[i].DataType = ftWideString) then
     FieldDefs[i].Size := FieldDefs[i].Size * 2;
   end;
  try
    Self.CreateTable;
  except
   on e: Exception do
    begin
     Log := Log + Format(ErrorLImportTableCannotCreateTable,[e.Message]);
     Exit;
    end;
  end;
  Self.Open;
  if (FHandle = nil) then
    raise EACRException.Create(10259,ErrorLNilPointer);
  s := CopyDatasets(SourceTable,Self,False,tbopImport);
  if (s <> '') then
    Log := Log + Format(ErrorLImportTableCannotCopyData,[s]);

  Result := (Log = SourceLog);
  if (FTemporary) then
   FHandle.CreateTableStarted := True;
  Self.Close;
end; // ImportTable


//------------------------------------------------------------------------------
// import table
//------------------------------------------------------------------------------
function TACRTable.ImportTable(SourceTable: TDataset): Boolean;
var s: string;
begin
 Result := ImportTable(SourceTable,s);
end; // ImportTable


//------------------------------------------------------------------------------
// export table
//------------------------------------------------------------------------------
function TACRTable.ExportTable(
                      DestinationTable:   TDataset;
                      CreateTablePointer: TProcedure;
                      var Log:            String
                   ): Boolean;
var
    s,SourceLog:     String;
    OldActive:       Boolean;
    Bookmark:        TBookmark;
begin
  Result := False;
  SourceLog := Log;
  OldActive := Self.Active;
  DestinationTable.Close;
  if (OldActive) then
   Bookmark := Self.GetBookmark
  else
   Bookmark := nil;
  Self.Close;
  Self.Open;
  if (FHandle = nil) then
    raise EACRException.Create(10262,ErrorLNilPointer);
  if (DestinationTable is TACRDataset) then
   TACRDataset(DestinationTable).AdvFieldDefs.Assign(Self.AdvFieldDefs)
  else
   DestinationTable.FieldDefs.Assign(Self.FieldDefs);
  try
    CreateTablePointer;
  except
   on e: Exception do
    begin
     Log := Log + Format(ErrorLExportTableCannotCreateTable,[e.Message]);
     Exit;
    end;
  end;
  try
    DestinationTable.Open;
  except
   on e: Exception do
    begin
     Log := Log + Format(ErrorLExportTableCannotOpenTable,[e.Message]);
     Exit;
    end;
  end;

  s := CopyDatasets(Self,DestinationTable,True,tbopExport);
  if (s <> '') then
    Log := Log + Format(ErrorLExportTableCannotCopyData,[s]);
  Result := (Log = SourceLog);
  // restore table position
  if (OldActive) then
   begin
    try
      Self.GotoBookmark(Bookmark);
    finally
      Self.FreeBookmark(Bookmark);
    end;
   end
  else
   Self.Close;
end; // ExportTable


//------------------------------------------------------------------------------
// export table
//------------------------------------------------------------------------------
function TACRTable.ExportTable(
                      DestinationTable:   TDataset;
                      CreateTablePointer: TProcedure
                   ): Boolean;
var
  s: String;
begin
 Result := ExportTable(DestinationTable,CreateTablePointer, s);
end; // ExportTable


//------------------------------------------------------------------------------
// export table to SQL
//------------------------------------------------------------------------------
function TACRTable.ExportTableToSQL(
                            ExportStructure:      Boolean;
                            AddDropTableCommand:  Boolean;
                            ExportIndexes:        Boolean;
                            AddDropIndexCommand:  Boolean;
                            ExportData:           Boolean;
                            ExportBLOBFields:     Boolean;
                            UseBracketsForNames:  Boolean
                         ): String;
var
    OldActive:       Boolean;
begin
  OldActive := Self.Active;
  if (not OldActive) then
   Open
  else
   begin
    InitFieldDefs;
    FieldDefs.Update;
    IndexDefs.Update;
   end;
  try
    Result := FHandle.ExportTableToSQL(ExportStructure,AddDropTableCommand,
                ExportIndexes,AddDropIndexCommand,
                ExportData,ExportBLOBFields,
                UseBracketsForNames
                );
  finally
    if (not OldActive) then
     Close;
  end;
end; // ExportTableToSQL


//------------------------------------------------------------------------------
// executes batch move to this table and return number of records moved
//------------------------------------------------------------------------------
function TACRTable.BatchMove(ASource: TACRDataSet; AMode: TACRBatchMode): Int64;
begin
  with TACRBatchMove.Create(nil) do
  try
    Destination := Self;
    Source := ASource;
    Mode := AMode;
    OnProgress := Self.OnProgress;
    Execute;
    Result := MovedCount;
  finally
    Free;
  end;
end; // BatchMove


//------------------------------------------------------------------------------
// restructure table
//------------------------------------------------------------------------------
function TACRTable.RestructureTable(
                      var Log:            String
                   ): Boolean;
begin
 Result := InternalRepairOrRestructureTable(False,Log,nil);
end; // RestructureTable


//------------------------------------------------------------------------------
// restructure table
//------------------------------------------------------------------------------
function TACRTable.RestructureTable: Boolean;
var s: String;
begin
 Result := RestructureTable(s);
end; // RestructureTable


//------------------------------------------------------------------------------
// repair (high level) or restructure table
//------------------------------------------------------------------------------
function TACRTable.InternalRepairOrRestructureTable(
                            Repair:       Boolean;
                            var Log:      String;
                            NewDatabase:  TACRDatabase = nil;
                            FKDefs:       TACRForeignKeyDefs = nil
                            ): Boolean;
var TempTable:           TACRTable;
    SourceLog,s:         String;
    b:                   Boolean;
    OperationName:       String;

 procedure DeleteTempTable;
 begin
  try
    if (TempTable.Active) then
      TempTable.Close;
  except
  end;
  try
    TempTable.DeleteTable(True);
  except
  end;
 end;

 procedure DoClose;
 begin
  try
    if (Active) then
     Close;
  except
  end;
 end;

 procedure RecreateForeignKeyActions;
 var i:        Integer;
     table:    TACRTable;
     FKAction: TACRConstraintDefForeignKeyAction;
     FK,FK1:   TACRForeignKeyDef;
 begin
   table := TACRTable.Create(nil);
   FK := TACRForeignKeyDef.Create;
   try
     table.DatabaseName := TempTable.DatabaseName;
     table.InMemory := Temptable.InMemory;
     table.FExclusive := True;
     for i := 0 to FACRConstraintDefs.Count-1 do
      if (FACRConstraintDefs.Items[i].ConstraintType = ctFKAction) then
       begin
        FKAction := TACRConstraintDefForeignKeyAction(FACRConstraintDefs.Items[i]);
        FK.Name := FKAction.ReferencedFKName;
        table.TableName := FKAction.ReferencedTableName;
        table.Open;
        FK1 := table.ForeignKeyDefs.Find(FK.Name);
        if (FK1 = nil) then
         raise EACRException.Create(11581,ErrorLForeignKeyDefNotFound,[FK.Name]);
        FK.Columns := FK1.Columns;
        table.Close;
        FK.ReferencedTableName := TempTable.TableName;
        FK.MatchType := TACRForeignKeyMatchType(Integer(FKAction.MatchType));
        FK.DeleteAction := TACRForeignKeyAction(Integer(FKAction.DeleteAction));
        FK.UpdateAction := TACRForeignKeyAction(Integer(FKAction.UpdateAction));
        FK.Name := ACRConstraintFKTemporaryNamePrefix+FKAction.ReferencedFKName;
        table.AddForeignKey(FK);
       end;
   finally
     FK.Free;
     table.Free;
   end;
 end;

 procedure CreateStructureForRepair;
 begin
  OperationName := 'repair';
  Open;
  TempTable.AdvFieldDefs.Assign(FAdvFieldDefs);
  TempTable.IndexDefs.Assign(FIndexDefs);
  if (NewDatabase = nil) then
   begin
    TempTable.ForeignKeyDefs.Assign(FForeignKeyDefs);
    ACRFixForeignKeysSelfReferences(FTableName,TempTable.TableName,TempTable.ForeignKeyDefs);
   end
  else
   begin
    if (FKDefs <> nil) then
     FKDefs.Assign(FForeignKeyDefs);
    TempTable.ForeignKeyDefs.Clear;
   end;
  Close;
 end;

 procedure CreateStructureForRestructure;
 begin
  OperationName := 'restructure';
  TempTable.AdvFieldDefs.Assign(FRestructureFieldDefs);
  TempTable.IndexDefs.Assign(FRestructureIndexDefs);
  TempTable.ForeignKeyDefs.Assign(FRestructureForeignKeyDefs);
  ACRFixForeignKeysSelfReferences(FTableName,TempTable.TableName,TempTable.ForeignKeyDefs);
 end;

begin
  Result := False;
  CheckInactive;
  b := FExclusive;
  if (not b) then
   FExclusive := True;
  SourceLog := Log;
  TempTable := TACRTable.Create(nil);
  try
    TempTable.FInMemory := Self.FInMemory;
    TempTable.FTemporary := Self.FTemporary;
    if (NewDatabase = nil) then
     TempTable.DatabaseName := Self.FDatabaseName
    else
     TempTable.DatabaseName := NewDatabase.DatabaseName;
    if (NewDatabase = nil) then
     repeat
      TempTable.FTableName := GetTemporaryName('TEMP_TABLE_'+FTableName+'_');
     until (not TempTable.Exists)
    else
     TempTable.FTableName := FTableName;
    TempTable.FieldDefs.Clear;
    TempTable.AdvIndexDefs.Clear;
    try
      if (Repair) then
       CreateStructureForRepair
      else
       CreateStructureForRestructure;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCreateStructure,[OperationName,FTableName,e.Message]);
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCreateStructure,[OperationName,FTableName,ErrorLUnknownError]);
        Exit;
      end;
    end;
    // create temporary table with new settings
    try
     TempTable.CreateTable;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCreateTable,[OperationName,FTableName,TempTable.TableName,e.Message]);
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCreateTable,[OperationName,FTableName,TempTable.TableName,ErrorLUnknownError]);
        Exit;
      end;
    end;
    try
     TempTable.Open;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotOpenTempTable,[OperationName,FTableName,TempTable.TableName,e.Message]);
        DeleteTempTable;
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotOpenTempTable,[OperationName,FTableName,TempTable.TableName,ErrorLUnknownError]);
        DeleteTempTable;
        Exit;
      end;
    end;
    // open current table
    try
     Open;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotOpenTable,[OperationName,FTableName,e.Message]);
        DoClose;
        DeleteTempTable;
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotOpenTable,[OperationName,FTableName,ErrorLUnknownError]);
        DoClose;
        DeleteTempTable;
        Exit;
      end;
    end;
    // copy records from source table to temporary table
    try
      if (Log = SourceLog) then
       begin
        if (Repair) then
         s := CopyDatasets(Self,TempTable,True,tbopRepair)
        else
         s := CopyDatasets(Self,TempTable,True,tbopRestructure);
        if (s <> '') then
         Log := Log + Format(ErrorLRestructureTableCannotCopyData,[OperationName,FTableName,s]);
       end;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCopyRecords,[OperationName,FTableName,e.Message]);
        DoClose;
        DeleteTempTable;
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCopyRecords,[OperationName,FTableName,ErrorLUnknownError]);
        DoClose;
        DeleteTempTable;
        Exit;
      end;
    end;
    if (Log <> SourceLog) then
     begin
      DoClose;
      DeleteTempTable;
      Exit;
     end;
    try
     if (NewDatabase = nil) then
      RecreateForeignKeyActions;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotRecreateForeignKeys,[OperationName,FTableName,e.Message]);
        DoClose;
        DeleteTempTable;
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotRecreateForeignKeys,[OperationName,FTableName,ErrorLUnknownError]);
        DoClose;
        DeleteTempTable;
        Exit;
      end;
    end;
    try
      TempTable.Close;
    except
     on e: Exception do
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCloseTempTable,[OperationName,FTableName,e.Message]);
        DoClose;
        DeleteTempTable;
        Exit;
      end
     else
      begin
        Log := Log + Format(ErrorLRestructureTableCannotCloseTempTable,[OperationName,FTableName,ErrorLUnknownError]);
        DoClose;
        DeleteTempTable;
        Exit;
      end;
    end;
    DoClose;
    if (Log = SourceLog) or (Repair) then
     begin
      try
       if (NewDatabase = nil) then
        DeleteTable(True);
      except
       on e: Exception do
        begin
          Log := Log + Format(ErrorLRestructureTableCannotDeleteTable,[OperationName,FTableName,e.Message]);
          DoClose;
          DeleteTempTable;
          Exit;
        end
       else
        begin
          Log := Log + Format(ErrorLRestructureTableCannotDeleteTable,[OperationName,FTableName,ErrorLUnknownError]);
          DoClose;
          DeleteTempTable;
          Exit;
        end;
      end;
      try
       if (NewDatabase = nil) then
        TempTable.RenameTable(Self.FTableName);
      except
       on e: Exception do
        begin
          Log := Log + Format(ErrorLRestructureTableCannotRenameTable,[OperationName,FTableName,TempTable.TableName,e.Message]);
          Exit;
        end
       else
        begin
          Log := Log + Format(ErrorLRestructureTableCannotRenameTable,[OperationName,FTableName,TempTable.TableName,ErrorLUnknownError]);
          Exit;
        end;
      end;
     end
    else
     DeleteTempTable;
    if (Repair) then
     Result := True
    else
     Result := (Log = SourceLog);
  finally
    TempTable.Free;
    FExclusive := b;
    DoClose;
  end;
end; // InternalRepairOrRestructureTable


//------------------------------------------------------------------------------
// repair table
//------------------------------------------------------------------------------
function TACRTable.RepairTable(UseLowLevelTableAccess: Boolean = False): Boolean;
var s: String;
begin
  Result := RepairTable(s,UseLowLevelTableAccess);
end; // RepairTable


//------------------------------------------------------------------------------
// repair table
//------------------------------------------------------------------------------
function TACRTable.RepairTable(var Log:                String;
                               UseLowLevelTableAccess: Boolean = False;
                               NewDatabase:            TACRDatabase = nil): Boolean;
var
  TempCursor:   TACRCursor;
begin
  if (not UseLowLevelTableAccess) then
   Result := InternalRepairOrRestructureTable(True,Log,NewDatabase)
  else
   begin
    CheckInactive;
    CheckBlankTableName;
    SetDBFlag(dbfTable, True);
    try
     TempCursor := CreateCursor;
     if (TempCursor = nil) then
      raise EACRException.Create(11166,ErrorLCannotCreateCursor);
     try
      TempCursor.Session := FDatabase.Handle;
      TempCursor.TableName := FTableName;
      TempCursor.Temporary := FTemporary;
      TempCursor.InMemory := FInMemory;
      TempCursor.ReadOnly := False;
      TempCursor.Exclusive := True;
      if (NewDatabase <> nil) then
       Result := TempCursor.RepairTable(Log,NewDatabase.Handle)
      else
       Result := TempCursor.RepairTable(Log);
     finally
      TempCursor.Free;
     end;
    finally
     SetDBFlag(dbfTable, False);
    end;
   end;
end; // RepairTable


//------------------------------------------------------------------------------
// Rename Field by Name
//------------------------------------------------------------------------------
procedure TACRTable.RenameField(FieldName, NewFieldName: String);
var
  TempCursor:   TACRCursor;
begin
  CheckInactive;
  SetDBFlag(dbfTable, True);
  try
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(10959,ErrorLCannotCreateCursor);
   try
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.RenameField(FieldName,NewFieldName);
   finally
    TempCursor.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end;//RenameField


//------------------------------------------------------------------------------
// add foreign key to existing table
//------------------------------------------------------------------------------
procedure TACRTable.AddForeignKey(ForeignKeyDef: TACRForeignKeyDef);
var
  TempCursor:    TACRCursor;
  ConstraintDef: TACRConstraintDefForeignKey;
begin
  CheckInactive;
  SetDBFlag(dbfTable, True);
  try
   if (not Self.Exists) then
    DatabaseError(Format(ErrorLTableDoesNotExist,[FTableName]),Self);
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(11580,ErrorLCannotCreateCursor);
   ConstraintDef := TACRConstraintDefForeignKey.Create;
   try
    ACRConvertForeignKeyDefToConstraintDef(ForeignKeyDef,ConstraintDef);
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.OpenTableByFieldDefs(nil,nil,nil);
    TempCursor.InternalInitFieldDefs;
    TempCursor.AddForeignKey(ConstraintDef);
   finally
    TempCursor.Free;
    ConstraintDef.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end; // AddForeignKey


//------------------------------------------------------------------------------
// DeleteConstraint
//------------------------------------------------------------------------------
procedure TACRTable.DeleteConstraint(Name: String; Cascade: Boolean = False);
var
  TempCursor:   TACRCursor;
begin
  CheckInactive;
  SetDBFlag(dbfTable, True);
  try
   if (not Self.Exists) then
    DatabaseError(Format(ErrorLTableDoesNotExist,[FTableName]),Self);
   TempCursor := CreateCursor;
   if (TempCursor = nil) then
    raise EACRException.Create(11579,ErrorLCannotCreateCursor);
   try
    TempCursor.Session := FDatabase.Handle;
    TempCursor.TableName := FTableName;
    TempCursor.Temporary := FTemporary;
    TempCursor.InMemory := FInMemory;
    TempCursor.ReadOnly := False;
    TempCursor.Exclusive := True;
    TempCursor.OpenTableByFieldDefs(nil,nil,nil);
    TempCursor.DeleteConstraint(Name,Cascade,False);
   finally
    TempCursor.Free;
   end;
  finally
   SetDBFlag(dbfTable, False);
  end;
end; // DeleteConstraint


//------------------------------------------------------------------------------
// create new index
//------------------------------------------------------------------------------
procedure TACRTable.AddIndex(
              const Name,
              Fields: String;
              Options: TIndexOptions;
              const DescFields: String = '';
              const CaseInsFields: String = ''
                   );
var
  TmpCursor: TACRCursor;
  ACRIndexDef: TACRIndexDef;
begin
  FieldDefs.Update;
  SetDBFlag(dbfIndexList,True);
  try
    if (FHandle = nil) then
      TmpCursor := GetHandle
    else
      begin
       CheckBrowseMode;
       CursorPosChanged;
       TmpCursor := FHandle;
     end;
     ACRIndexDef := TACRIndexDef.Create;
    try
     if (Name = '') then
      raise EACRException.Create(20046, ErrorAInvalidIndexName, ['']);
     if (Fields = '') then
      raise EACRException.Create(20047, ErrorACannotFindIndexField, ['']);
     FillACRIndexDef(ACRIndexDef, Name, Fields, Options, DescFields, CaseInsFields, FieldDefs, AdvFieldDefs);
     TmpCursor.Exclusive := True;
     TmpCursor.AddIndex(ACRIndexDef);
    finally
       ACRIndexDef.Free;
       if (FHandle = nil) then
        TmpCursor.Free;
    end;
    FIndexDefs.Updated := False;
  finally
    SetDBFlag(dbfIndexList,False);
  end;
end;// AddIndex


//------------------------------------------------------------------------------
// DeleteIndex
//------------------------------------------------------------------------------
procedure TACRTable.DeleteIndex(const Name: string);
var
  TmpCursor: TACRCursor;
begin
  SetDBFlag(dbfIndexList, True);
  try
    if (FHandle = nil) then
      TmpCursor := GetHandle
    else
      begin
        CheckBrowseMode;
        TmpCursor := FHandle;
      end;
    try
      TmpCursor.DeleteIndex(Name);
    finally
      if (FHandle = nil) then
       TmpCursor.Free;
    end;
    FIndexDefs.Updated := False;
  finally
    SetDBFlag(dbfIndexList, False);
  end;
end;// DeleteIndex


//------------------------------------------------------------------------------
// DeleteAllIndexes
//------------------------------------------------------------------------------
procedure TACRTable.DeleteAllIndexes;
var
  TmpCursor: TACRCursor;
begin
  SetDBFlag(dbfIndexList, True);
  try
    if (FHandle = nil) then
      TmpCursor := GetHandle
    else
      begin
        CheckBrowseMode;
        TmpCursor := FHandle;
      end;
    try
      TmpCursor.DeleteAllIndexes;
    finally
      if (FHandle = nil) then
       TmpCursor.Free;
    end;
    FIndexDefs.Updated := False;
  finally
    SetDBFlag(dbfIndexList, False);
  end;
end;// DeleteAllIndexes


//------------------------------------------------------------------------------
// check set key mode
//------------------------------------------------------------------------------
procedure TACRTable.CheckSetKeyMode;
begin
  if (State <> dsSetKey) then
    DatabaseError(ErrorLDatasetIsNotInEditOrInsertMode, Self);
end; // CheckSetKeyMode


//------------------------------------------------------------------------------
// get key buffer
//------------------------------------------------------------------------------
function TACRTable.GetKeyBuffer(KeyIndex: TACRKeyIndex): TACRRecordBuffer;
begin
  Result := FKeyBuffers[KeyIndex];
end; // GetKeyBuffer


//------------------------------------------------------------------------------
// return true if key is exclusive
//------------------------------------------------------------------------------
function TACRTable.GetKeyExclusive: Boolean;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10282,ErrorLNilPointer);
  CheckSetKeyMode;
  Result := PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.Exclusive;
end; // GetKeyExclusive


//------------------------------------------------------------------------------
// return key field count
//------------------------------------------------------------------------------
function TACRTable.GetKeyFieldCount: Integer;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10284,ErrorLNilPointer);
  CheckSetKeyMode;
  Result := PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.FieldCount;
end; // GetKeyFieldCount


//------------------------------------------------------------------------------
// set key exclusive
//------------------------------------------------------------------------------
procedure TACRTable.SetKeyExclusive(Value: Boolean);
begin
  if (FHandle = nil) then
    raise EACRException.Create(10283,ErrorLNilPointer);
  CheckSetKeyMode;
  PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.Exclusive := Value;
end; // SetKeyExclusive


//------------------------------------------------------------------------------
// set key field count
//------------------------------------------------------------------------------
procedure TACRTable.SetKeyFieldCount(Value: Integer);
begin
  if (FHandle = nil) then
    raise EACRException.Create(10285,ErrorLNilPointer);
  CheckSetKeyMode;
  PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.FieldCount := Value;
end; // SetKeyFieldCount


//------------------------------------------------------------------------------
// set key buffer
//------------------------------------------------------------------------------
procedure TACRTable.SetKeyBuffer(KeyIndex: TACRKeyIndex; Clear: Boolean);
begin
  if (FHandle = nil) then
    raise EACRException.Create(10286,ErrorLNilPointer);
  CheckBrowseMode;
  FKeyBuffer := FKeyBuffers[KeyIndex];
  Move(FKeyBuffer^, FKeyBuffers[kiSave]^, FHandle.KeyBufferSize);
  if (Clear) then
    InitKeyBuffer(FKeyBuffer);
  SetState(dsSetKey);
  SetModified(PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.Modified);
  DataEvent(deDataSetChange, 0);
end; // SetKeyBuffer


//------------------------------------------------------------------------------
// set key fields
//------------------------------------------------------------------------------
procedure TACRTable.SetKeyFields(KeyIndex: TACRKeyIndex; const Values: array of const);
var
    i:          Integer;
    SaveState:  TDataSetState;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10289,ErrorLNilPointer);
  if (not IsIndexApplied) then
    DatabaseError(ErrorLNoFieldIndexes, Self);
  SaveState := SetTempState(dsSetKey);
  try
    FKeyBuffer := InitKeyBuffer(FKeyBuffers[KeyIndex]);
    for i := 0 to High(Values) do
      GetIndexField(i).AssignValue(Values[i]);
    PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.FieldCount := High(Values) + 1;
    PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.Modified := Modified;
  finally
    RestoreState(SaveState);
  end;
end; // SetKeyFields


//------------------------------------------------------------------------------
// post key buffer
//------------------------------------------------------------------------------
procedure TACRTable.PostKeyBuffer(Commit: Boolean);
begin
  if (FHandle = nil) then
    raise EACRException.Create(10288,ErrorLNilPointer);
  DataEvent(deCheckBrowseMode, 0);
  if (Commit) then
    PACRKeyBuffer(FKeyBuffer + FHandle.KeyOffset)^.Modified := Modified
  else
    Move(FKeyBuffers[kiSave]^, FKeyBuffer^, FHandle.KeyBufferSize);
  SetState(dsBrowse);
  DataEvent(deDataSetChange, 0);
end; // PostKeyBuffer


//------------------------------------------------------------------------------
// find key
//------------------------------------------------------------------------------
function TACRTable.FindKey(const KeyValues: array of const): Boolean;
begin
  CheckBrowseMode;
  SetKeyFields(kiLookup, KeyValues);
  Result := GotoKey;
end; // FindKey


//------------------------------------------------------------------------------
// find nearest
//------------------------------------------------------------------------------
procedure TACRTable.FindNearest(const KeyValues: array of const);
begin
  CheckBrowseMode;
  SetKeyFields(kiLookup, KeyValues);
  GotoNearest;
end; // FindNearest


//------------------------------------------------------------------------------
// goto key
//------------------------------------------------------------------------------
function TACRTable.GotoKey: Boolean;
var
  SearchCondition:    TACRSearchCondition;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10290,ErrorLNilPointer);
  CheckBrowseMode;
  DoBeforeScroll;
  CursorPosChanged;
  SearchCondition := scEqual;
  FHandle.KeyBuffer := GetKeyBuffer(kiLookup);
  FHandle.KeyFieldCount := PACRKeyBuffer(FHandle.KeyBuffer + FHandle.KeyOffset)^.FieldCount;
  if (FHandle.KeyFieldCount = 0) then
   FHandle.KeyFieldCount := FIndexFieldCount;
  Result := FHandle.FindKey(SearchCondition);
  if (Result) then
   begin
    Resync([rmExact, rmCenter]);
    DoAfterScroll;
   end;
end; // GotoKey


//------------------------------------------------------------------------------
// goto nearest
//------------------------------------------------------------------------------
procedure TACRTable.GotoNearest;
var
  SearchCondition: TACRSearchCondition;
  Result:          Boolean;
begin
  if (FHandle = nil) then
    raise EACRException.Create(10291,ErrorLNilPointer);
  CheckBrowseMode;
  CursorPosChanged;
  FHandle.KeyBuffer := GetKeyBuffer(kiLookup);
  if (PACRKeyBuffer(FHandle.KeyBuffer + FHandle.KeyOffset)^.Exclusive) then
    SearchCondition := scGreater
  else
    SearchCondition := scGreaterEqual;
  FHandle.KeyFieldCount := PACRKeyBuffer(FHandle.KeyBuffer + FHandle.KeyOffset)^.FieldCount;
  if (FHandle.KeyFieldCount = 0) then
    FHandle.KeyFieldCount := FIndexFieldCount;
  Result := FHandle.FindKey(SearchCondition);
  // if nearest record was not find go to last record
  if (not Result) then
   Last;
  Resync([rmCenter]);
end; // GotoNearest


//------------------------------------------------------------------------------
// edit key
//------------------------------------------------------------------------------
procedure TACRTable.EditKey;
begin
  SetKeyBuffer(kiLookup, False);
end; // EditKey


//------------------------------------------------------------------------------
// set key
//------------------------------------------------------------------------------
procedure TACRTable.SetKey;
begin
  SetKeyBuffer(kiLookup, True);
end; // SetKey


//------------------------------------------------------------------------------
// set range
//------------------------------------------------------------------------------
function TACRTable.SetCursorRange: Boolean;
var StartBuffer, EndBuffer:       TACRRecordBuffer;
    StartExclusive, EndExclusive: Boolean;
    StartKeyFieldCount,
    EndKeyFieldCount:             Integer;
begin
  Result := False;
  if (FHandle = nil) then
    raise EACRException.Create(10296,ErrorLNilPointer);
  if (not (
            BuffersEqual(FKeyBuffers[kiRangeStart], FKeyBuffers[kiCurRangeStart],
              FHandle.KeyBufferSize)
           and
            BuffersEqual(FKeyBuffers[kiRangeEnd], FKeyBuffers[kiCurRangeEnd],
              FHandle.KeyBufferSize))) then
  begin
    StartBuffer := FKeyBuffers[kiRangeStart];
    StartKeyFieldCount := PACRKeyBuffer(StartBuffer + FHandle.KeyOffset)^.FieldCount;
    if (StartKeyFieldCount = 0) then
     StartKeyFieldCount := FIndexFieldCount;
    StartExclusive := PACRKeyBuffer(StartBuffer + FHandle.KeyOffset)^.Exclusive;
    EndBuffer := FKeyBuffers[kiRangeEnd];
    EndKeyFieldCount := PACRKeyBuffer(EndBuffer + FHandle.KeyOffset)^.FieldCount;
    if (EndKeyFieldCount = 0) then
     EndKeyFieldCount := FIndexFieldCount;
    EndExclusive := PACRKeyBuffer(EndBuffer + FHandle.KeyOffset)^.Exclusive;
    FHandle.ApplyRange(StartBuffer,EndBuffer,
                       StartKeyFieldCount,EndKeyFieldCount,
                       StartExclusive, EndExclusive);
    Move(FKeyBuffers[kiRangeStart]^, FKeyBuffers[kiCurRangeStart]^,
      FHandle.KeyBufferSize);
    Move(FKeyBuffers[kiRangeEnd]^, FKeyBuffers[kiCurRangeEnd]^,
      FHandle.KeyBufferSize);
    Result := True;
  end;
end; // SetCursorRange


//------------------------------------------------------------------------------
// apply range
//------------------------------------------------------------------------------
procedure TACRTable.ApplyRange;
begin
  CheckBrowseMode;
  if (SetCursorRange) then
    First;
end; // ApplyRange


//------------------------------------------------------------------------------
// cancel range
//------------------------------------------------------------------------------
procedure TACRTable.CancelRange;
begin
  CheckBrowseMode;
  UpdateCursorPos;
  if (ResetCursorRange) then
    Resync([]);
end; // CancelRange


//------------------------------------------------------------------------------
// edit range start
//------------------------------------------------------------------------------
procedure TACRTable.EditRangeStart;
begin
  SetKeyBuffer(kiRangeStart, False);
end; // EditRangeStart


//------------------------------------------------------------------------------
// edit range end
//------------------------------------------------------------------------------
procedure TACRTable.EditRangeEnd;
begin
  SetKeyBuffer(kiRangeEnd, False);
end; // EditRangeEnd


//------------------------------------------------------------------------------
// set range
//------------------------------------------------------------------------------
procedure TACRTable.SetRange(const StartValues, EndValues: array of const);
begin
  CheckBrowseMode;
  SetKeyFields(kiRangeStart, StartValues);
  SetKeyFields(kiRangeEnd, EndValues);
  ApplyRange;
end; // SetRange


//------------------------------------------------------------------------------
// set range start
//------------------------------------------------------------------------------
procedure TACRTable.SetRangeStart;
begin
  SetKeyBuffer(kiRangeStart, True);
end; // SetRangeStart


//------------------------------------------------------------------------------
// set range end
//------------------------------------------------------------------------------
procedure TACRTable.SetRangeEnd;
begin
  SetKeyBuffer(kiRangeEnd, True);
end; // SetRangeEnd


//------------------------------------------------------------------------------
// post
//------------------------------------------------------------------------------
procedure TACRTable.Post;
begin
  inherited Post;
  if (State = dsSetKey) then
    PostKeyBuffer(True);
end; // Post


//------------------------------------------------------------------------------
// LastAutoincValue
//------------------------------------------------------------------------------
function TACRTable.LastAutoincValue(FieldIndex: Integer): Int64;
begin
  Result := FHandle.LastAutoincValue(FieldIndex);
end;//LastAutoincValue


//------------------------------------------------------------------------------
// LastAutoincValue
//------------------------------------------------------------------------------
function TACRTable.LastAutoincValue(FieldName: String): Int64;
begin
  Result := LastAutoincValue(Fields.IndexOf(Fields.FieldByName(FieldName)));
end;//LastAutoincValue


//------------------------------------------------------------------------------
// set LastAutoincValue for Field (FieldIndex started from 0)
//------------------------------------------------------------------------------
procedure TACRTable.SetLastAutoincValue(Value: Int64; FieldIndex: Integer);
begin
  FHandle.SetLastAutoincValue(Value,FieldIndex);
end; // SetLastAutoincValue


//------------------------------------------------------------------------------
// set LastAutoincValue for Field
//------------------------------------------------------------------------------
procedure TACRTable.SetLastAutoincValue(Value: Int64; FieldName: String);
begin
  SetLastAutoincValue(Value,Fields.IndexOf(Fields.FieldByName(FieldName)));
end; // SetLastAutoincValue


////////////////////////////////////////////////////////////////////////////////
//
// TACRQuery
//
////////////////////////////////////////////////////////////////////////////////

{$IFDEF D6H}
//------------------------------------------------------------------------------
// execute
//------------------------------------------------------------------------------
procedure TACRQuery.PSExecute;
begin
  ExecSQL;
end; // PSExecute


//------------------------------------------------------------------------------
// get default order
//------------------------------------------------------------------------------
function TACRQuery.PSGetDefaultOrder: TIndexDef;
begin
  Result := inherited PSGetDefaultOrder;
  if not Assigned(Result) then
    Result := GetIndexForOrderBy(SQL.Text, Self);
end; // PSGetDefaultOrder


//------------------------------------------------------------------------------
// get params
//------------------------------------------------------------------------------
function TACRQuery.PSGetParams: TParams;
begin
  Result := Params;
end; // PSGetParams


//------------------------------------------------------------------------------
// get table name
//------------------------------------------------------------------------------
function TACRQuery.PSGetTableName: string;
begin
  Result := GetTableNameFromSQL(SQL.Text);
end; // PSGetTableName


//------------------------------------------------------------------------------
// set command text
//------------------------------------------------------------------------------
procedure TACRQuery.PSSetCommandText(const CommandText: string);
begin
  if CommandText <> '' then
    SQL.Text := CommandText;
end; // PSSetCommandText


//------------------------------------------------------------------------------
// set params
//------------------------------------------------------------------------------
procedure TACRQuery.PSSetParams(AParams: TParams);
begin
  if AParams.Count <> 0 then
    Params.Assign(AParams);
  Close;
end; // PSSetParams


{$ENDIF}


//------------------------------------------------------------------------------
// GetStatementHandle
//------------------------------------------------------------------------------
procedure TACRQuery.GetStatementHandle(SQLText: PChar);
begin
  FStmtHandle := nil;
  if (FDatabase = nil) then
   begin
    {$IFDEF LOCAL_VERSION}
      FStmtHandle := TACRLocalSQLProcessor.Create(Self);
    {$ENDIF}
   end // local database
  else
  if (FDatabase.LocalDatabase or FInMemory) then
   begin
    {$IFDEF LOCAL_VERSION}
      FStmtHandle := TACRLocalSQLProcessor.Create(Self);
    {$ENDIF}
   end // local database
  else
   begin
    {$IFDEF CLIENT_VERSION}
      FStmtHandle := TACRClientSQLProcessor.Create(Self, FDatabase.Handle);
    {$ENDIF}
   end; // remote database
  if (FStmtHandle = nil) then
   DatabaseError(ErrorLSQLProcessorIsSwitchedOff);
  try
    FStmtHandle.RequestLive := FRequestLive;
    FStmtHandle.InMemory := FInMemory;
    FStmtHandle.PrepareStatement(SQLText);
  except
    FStmtHandle.Free;
    FStmtHandle := nil;
    raise;
  end;
end;// GetStatementHandle


//------------------------------------------------------------------------------
// FreeStatement
//------------------------------------------------------------------------------
procedure TACRQuery.FreeStatement;
begin
  if (FStmtHandle <> nil) then
    begin
      FStmtHandle.Free;
      FStmtHandle := nil;
    end;
end;// FreeStatement


//------------------------------------------------------------------------------
// CreateCursor
//------------------------------------------------------------------------------
function TACRQuery.CreateCursor(GenHandle: Boolean): TACRCursor;
var Abort: Boolean;
begin
  if (SQL.Count > 0) then
    begin
      if (IsBeforeExecuteSQLAssigned) then
       begin
        FDatabase.DoBeforeExecuteSQL(Self,Abort);
        if (Abort) then
         begin
          Result := nil;
          DatabaseError(ErrorLExecuteSQLBlockedByDatabaseEvent,Self);
          Exit;
         end;
       end;
      FExecSQL := not GenHandle;
      try
        SetPrepared(True);
      finally
        FExecSQL := False;
      end;
      if FDataLink.DataSource <> nil then
        SetParamsFromCursor;
      Result := GetQueryCursor(GenHandle);
      if (IsAfterExecuteSQLAssigned) then
       FDatabase.DoAfterExecuteSQL(Self);
    end
  else
    begin
      raise EACRException.Create(20034, ErrorAEmptySQLStatement);
      Result := nil;
    end;
  FCheckRowsAffected := (Result = nil);
end;// CreateCursor


//------------------------------------------------------------------------------
// GetQueryCursor
//------------------------------------------------------------------------------
function TACRQuery.GetQueryCursor(GenHandle: Boolean): TACRCursor;
var
  i:      Integer;
  Param:  TACRSQLParam;
begin
  Result := nil;
    // Set Params
  if FParams.Count > 0 then
   begin
    StmtHandle.SQLParams.Clear;
    for i:=0 to FParams.Count-1 do
     begin
      Param := StmtHandle.SQLParams.AddCreated;
      Param.Name := FParams[i].Name;
      if (FParams[i].IsNull) then
       Param.SetNull
      else
       begin
        Param.SetNull(AdvancedFieldTypeToBaseFieldType(
                                FieldTypeToACRAdvFieldType(FParams[i].DataType)));
        case FParams[i].DataType of
          ftBlob,ftGraphic:
            begin
              Param.DataType := bftBlob;
              Param.SetData(Pchar(Params[i].AsBlob), Length(Params[i].AsBlob), bftBlob);
            end;
         else
        Param.AsVariant := FParams[i].Value;
       end;
      end;
     end;
    end;

   if (GenHandle) then
    begin
      Result := StmtHandle.OpenQuery;
      // reopen cursor
      if Result <> nil then
        Result.FSettingProjection := True;
      FExternalHandle := Result;
      FReadOnly := StmtHandle.ReadOnly;
    end
   else
    StmtHandle.ExecuteQuery;
end;// GetQueryCursor


//------------------------------------------------------------------------------
// GetRowsAffected
//------------------------------------------------------------------------------
function TACRQuery.GetRowsAffected: Integer;
begin
  if Prepared then
    Result := StmtHandle.RowsAffected
  else
    Result := FRowsAffected;
end;//GetRowsAffected


//------------------------------------------------------------------------------
// QueryChanged
//------------------------------------------------------------------------------
procedure TACRQuery.QueryChanged(Sender: TObject);
var
  List: TParams;
begin
  if not (csReading in ComponentState) then
  begin
    Disconnect;
    StrDispose(SQLBinary);
    SQLBinary := nil;
    if ParamCheck or (csDesigning in ComponentState) then
    begin
      List := TParams.Create(Self);
      try
        FText := List.ParseSQL(SQL.Text, True);
        List.AssignValues(FParams);
        FParams.Clear;
        FParams.Assign(List);
      finally
        List.Free;
      end;
     end
    else
      FText := SQL.Text;
    DataEvent(dePropertyChange, 0);
   end
  else
    FText := FParams.ParseSQL(SQL.Text, False);
end;// QueryChanged


//------------------------------------------------------------------------------
// GetDataSource
//------------------------------------------------------------------------------
function TACRQuery.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;// GetDataSource


//------------------------------------------------------------------------------
// SetDataSource
//------------------------------------------------------------------------------
procedure TACRQuery.SetDataSource(Value: TDataSource);
begin
  if IsLinkedTo(Value) then
   DatabaseError(ErrorACircularDataLink, Self);
  FDataLink.DataSource := Value;
end;// SetDataSource


//------------------------------------------------------------------------------
// SetQuery
//------------------------------------------------------------------------------
procedure TACRQuery.SetQuery(Value: TStrings);
begin
  if SQL.Text <> Value.Text then
   begin
    Disconnect;
    SQL.BeginUpdate;
    try
      SQL.Assign(Value);
    finally
      SQL.EndUpdate;
    end;
   end;
end;// SetQuery


//------------------------------------------------------------------------------
// GetParamsCount
//------------------------------------------------------------------------------
function TACRQuery.GetParamsCount: Word;
begin
  Result := FParams.Count;
end;// GetParamsCount


//------------------------------------------------------------------------------
// RefreshParams
//------------------------------------------------------------------------------
procedure TACRQuery.RefreshParams;
var
  DataSet: TDataSet;
begin
  DisableControls;
  try
    if FDataLink.DataSource <> nil then
    begin
      DataSet := FDataLink.DataSource.DataSet;
      if DataSet <> nil then
        if DataSet.Active and (DataSet.State <> dsSetKey) then
        begin
          Close;
          Open;
        end;
    end;
  finally
    EnableControls;
  end;
end;// RefreshParams


//------------------------------------------------------------------------------
// SetParamsList
//------------------------------------------------------------------------------
procedure TACRQuery.SetParamsList(Value: TParams);
begin
  FParams.AssignValues(Value);
end;// SetParamsList


//------------------------------------------------------------------------------
// SetParamsFromCursor
//------------------------------------------------------------------------------
procedure TACRQuery.SetParamsFromCursor;
var
  I: Integer;
  DataSet: TDataSet;
begin
  if FDataLink.DataSource <> nil then
  begin
    DataSet := FDataLink.DataSource.DataSet;
    if DataSet <> nil then
    begin
      DataSet.FieldDefs.Update;
      for I := 0 to FParams.Count - 1 do
       if not FParams[I].Bound then
          begin
           FParams[I].AssignField(DataSet.FieldByName(FParams[I].Name));
           FParams[I].Bound := False;
          end;
    end;
  end;
end;// SetParamsFromCursor


//------------------------------------------------------------------------------
// ParamByName
//------------------------------------------------------------------------------
function TACRQuery.ParamByName(const Value: string): TParam;
begin
  Result:=FParams.ParamByName(Value);
end;// ParamByName


//------------------------------------------------------------------------------
// Prepare
//------------------------------------------------------------------------------
procedure TACRQuery.Prepare;
begin
  SetDBFlag(dbfPrepared, True);
  SetPrepared(True);
end;// Prepare


//------------------------------------------------------------------------------
// UnPrepare
//------------------------------------------------------------------------------
procedure TACRQuery.UnPrepare;
begin
  SetPrepared(False);
  SetDBFlag(dbfPrepared, False);
end;// UnPrepare


//------------------------------------------------------------------------------
// PrepareSQL
//------------------------------------------------------------------------------
procedure TACRQuery.PrepareSQL(Value: PChar);
begin
  GetStatementHandle(Value);
end;// PrepareSQL


//------------------------------------------------------------------------------
// SetPrepared
//------------------------------------------------------------------------------
procedure TACRQuery.SetPrepared(Value: Boolean);
begin
  if FHandle <> nil then
   if (not Value) then
     begin
       {$IFDEF CLIENT_VERSION}
       if (FHandle.IsClientCursor) then
        FHandle.Free;
       {$ENDIF}
       FHandle := nil;
       FExternalHandle := nil;
     end
   else
    raise EACRException.Create(30141, ErorrGDataSetOpen);

  if Value <> Prepared then
   begin
    if Value then
     begin
      FRowsAffected := -1;
      FCheckRowsAffected := True;
      if Length(SQL.Text) > 1 then
        PrepareSQL(PChar(SQL.Text))
      else
        raise EACRException.Create(30142, ErorrGEmptySQLStatement);
     end
    else
     begin
      if FCheckRowsAffected then
        FRowsAffected := RowsAffected;
      FreeStatement;
      FExternalHandle := nil;
     end;
    FPrepared := Value;
  end;
end;// SetPrepared


//------------------------------------------------------------------------------
// SetPrepare
//------------------------------------------------------------------------------
procedure TACRQuery.SetPrepare(Value: Boolean);
  begin
  if Value then
   Prepare
  else
   UnPrepare;
end;// SetPrepare


//------------------------------------------------------------------------------
// ReadBinaryData
//------------------------------------------------------------------------------
procedure TACRQuery.ReadBinaryData(Stream: TStream);
begin
  SQLBinary := StrAlloc(Stream.Size);
  Stream.ReadBuffer(SQLBinary^, Stream.Size);
end;// ReadBinaryData


//------------------------------------------------------------------------------
// WriteBinaryData
//------------------------------------------------------------------------------
procedure TACRQuery.WriteBinaryData(Stream: TStream);
begin
  Stream.WriteBuffer(SQLBinary^, StrBufSize(SQLBinary));
end;// WriteBinaryData


//------------------------------------------------------------------------------
// ReadParamData
//------------------------------------------------------------------------------
procedure TACRQuery.ReadParamData(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FParams);
end;// ReadParamData


//------------------------------------------------------------------------------
// WriteParamData
//------------------------------------------------------------------------------
procedure TACRQuery.WriteParamData(Writer: TWriter);
begin
  Writer.WriteCollection(Params);
end;// WriteParamData


//------------------------------------------------------------------------------
// define properties - for setting params
//------------------------------------------------------------------------------
procedure TACRQuery.DefineProperties(Filer: TFiler);

  function WriteData: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not FParams.IsEqual(TACRQuery(Filer.Ancestor).FParams)
    else
      Result := FParams.Count > 0;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadBinaryData, WriteBinaryData, SQLBinary <> nil);
  Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, WriteData);
end; // DefineProperties


//------------------------------------------------------------------------------
// CreateHandle
//------------------------------------------------------------------------------
function TACRQuery.CreateHandle: TACRCursor;
begin
  if (FExternalHandle = nil) then
    Result := CreateCursor(True)
  else
    Result := FExternalHandle;
end;// CreateHandle


//------------------------------------------------------------------------------
// Disconnect
//------------------------------------------------------------------------------
procedure TACRQuery.Disconnect;
begin
  Close;
  UnPrepare;
end;// Disconnect


//------------------------------------------------------------------------------
// SetDBFlag
//------------------------------------------------------------------------------
procedure TACRQuery.SetDBFlag(Flag: Integer; Value: Boolean);
begin
  if Value then
    inherited SetDBFlag(Flag,Value)
  else
    begin
      if ((DBFlags - [Flag]) = []) then
         SetPrepared(False);
      inherited SetDBFlag(Flag, Value);
    end;
end;// SetDBFlag


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRQuery.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIsTable := False;
  FSQL := TStringList.Create;
  TStringList(SQL).OnChange := QueryChanged;
  FParams := TParams.Create(Self);
  FDataLink := TACRQueryDataLink.Create(Self);
  RequestLive := False;
  ParamCheck := True;
  FRowsAffected := -1;
   
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRQuery.Destroy;
begin
  Destroying;
  Disconnect;
  SQL.Free;
  FParams.Free;
  FDataLink.Free;
  StrDispose(SQLBinary);
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// ExecSQL
//------------------------------------------------------------------------------
procedure TACRQuery.ExecSQL;
begin
  CheckInActive;
  SetDBFlag(dbfExecSQL, True);
  try
    CreateCursor(False);
  finally
    SetDBFlag(dbfExecSQL, False);
  end;
end;// ExecSQL


//------------------------------------------------------------------------------
// GetDetailLinkFields
//------------------------------------------------------------------------------
procedure TACRQuery.GetDetailLinkFields(MasterFields, DetailFields: TList);


  function AddFieldToList(
                          const FieldName:  String;
                          DataSet:          TDataSet;
                          List:             TList
                         ): Boolean;
  var
    Field: TField;
  begin
    Field := DataSet.FindField(FieldName);
    if (Field <> nil) then
      List.Add(Field);
    Result := (Field <> nil);
  end; // AddFieldToList


var
  i: Integer;
begin
  MasterFields.Clear;
  DetailFields.Clear;
  if (DataSource <> nil) and (DataSource.DataSet <> nil) then
    for i := 0 to Params.Count - 1 do
      if AddFieldToList(Params[i].Name, DataSource.DataSet, MasterFields) then
        AddFieldToList(Params[i].Name, Self, DetailFields);
end;// GetDetailLinkFields


////////////////////////////////////////////////////////////////////////////////
//
// TACRLockParamsEditor
//
////////////////////////////////////////////////////////////////////////////////


{$IFNDEF SQLMEMTABLE}
//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRLockParamsEditor.Create;
begin
  inherited;
  FRetryCount := ACRDefaultRetryCount;
  FDelay := ACRDefaultDelay;
end;//Create


//------------------------------------------------------------------------------
// GetLockParams
//------------------------------------------------------------------------------
function TACRLockParamsEditor.GetLockParams: TACRLockParams;
begin
  Result.RetryCount := FRetryCount;
  Result.Delay := FDelay;
end;// GetLockParams


//------------------------------------------------------------------------------
// etLockParams
//------------------------------------------------------------------------------
procedure TACRLockParamsEditor.SetLockParams(Params: TACRLockParams);
begin
  FRetryCount := Params.RetryCount;
  FDelay := Params.Delay;
end; // SetLockParams


//------------------------------------------------------------------------------
// SetRetryCount
//------------------------------------------------------------------------------
procedure TACRLockParamsEditor.SetRetryCount(Value: Integer);
begin
  if (Value >= ACRMinRetryCount) then
    FRetryCount := Value;
end; // SetRetryCount


//------------------------------------------------------------------------------
// SetDelay
//------------------------------------------------------------------------------
procedure TACRLockParamsEditor.SetDelay(Value: Integer);
begin
  if (Value >= ACRMinDelay) then
    FDelay := Value;
end; // SetDelay


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRLockParamsEditor.Assign(Source: TPersistent);
begin
 FDelay := TACRLockParamsEditor(Source).Delay;
 FRetryCount := TACRLockParamsEditor(Source).RetryCount;
end; // Assign
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACROptionsEditor
//
////////////////////////////////////////////////////////////////////////////////


{$IFNDEF SQLMEMTABLE}
//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACROptionsEditor.Create;
begin
  inherited;
  FMaxSessionCount := ACRDefaultSessionCount;
  if (FMaxSessionCount > ACRMaxSessionCount) then
   FMaxSessionCount := ACRMaxSessionCount;
  FPageSize := ACRDefaultPageSize;
  FExtentPageCount := ACRDefaultExtentPageCount;
  FRandomSearchRetryCount := ACRDefaultRandomSearchRetryCount;
end;//Create


//------------------------------------------------------------------------------
// GetOptions
//------------------------------------------------------------------------------
function TACROptionsEditor.GetOptions: TACROptions;
begin
  Result.MaxSessionCount := FMaxSessionCount;
  Result.PageSize := FPageSize;
  Result.ExtentPageCount := FExtentPageCount;
  Result.RandomSearchRetryCount := FRandomSearchRetryCount;
end;// GetOptions


//------------------------------------------------------------------------------
// SetOptions
//------------------------------------------------------------------------------
procedure TACROptionsEditor.SetOptions(NewOptions: TACROptions);
begin
 FMaxSessionCount := NewOptions.MaxSessionCount;
 FPageSize := NewOptions.PageSize;
 FExtentPageCount := NewOptions.ExtentPageCount;
 FRandomSearchRetryCount := NewOptions.RandomSearchRetryCount;
end; // SetOptions


//------------------------------------------------------------------------------
// SetMaxSessionCount
//------------------------------------------------------------------------------
procedure TACROptionsEditor.SetMaxSessionCount(Value: Integer);
begin
  if (Value <= 0) then
    FMaxSessionCount := 1
  else
  if (Value > ACRMaxSessionCount) then
    FMaxSessionCount := ACRMaxSessionCount
  else
    FMaxSessionCount := Value;
end; // SetMaxSessionCount


//------------------------------------------------------------------------------
// SetPageSize
//------------------------------------------------------------------------------
procedure TACROptionsEditor.SetPageSize(Value: Integer);
begin
  if (Value > ACRMaxPageSize) then
    FPageSize := ACRMaxPageSize
  else
  if (Value < ACRMinPageSize) then
    FPageSize := ACRMinPageSize
  else
    FPageSize := Value;
end; // SetPageSize


//------------------------------------------------------------------------------
// SetExtentPageCount
//------------------------------------------------------------------------------
procedure TACROptionsEditor.SetExtentPageCount(Value: Integer);
begin
  if (Value > ACRMaxExtentPageCount) then
    FExtentPageCount := ACRMaxExtentPageCount
  else
  if (Value < ACRMinExtentPageCount) then
    FExtentPageCount := ACRMinExtentPageCount
  else
    FExtentPageCount := Value;
end; // SetExtentPageCount


//------------------------------------------------------------------------------
// SetRandomSearchRetryCount
//------------------------------------------------------------------------------
procedure TACROptionsEditor.SetRandomSearchRetryCount(Value: Integer);
begin
  if (Value > ACRMaxRandomSearchRetryCount) then
    FRandomSearchRetryCount := ACRMaxRandomSearchRetryCount
  else
  if (Value < ACRMinRandomSearchRetryCount) then
    FRandomSearchRetryCount := ACRMinRandomSearchRetryCount
  else
    FRandomSearchRetryCount := Value;
end; // SetRandomSearchRetryCount


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACROptionsEditor.Assign(Source: TPersistent);
begin
 FMaxSessionCount := TACROptionsEditor(Source).MaxSessionCount;
 FExtentPageCount := TACROptionsEditor(Source).ExtentPageCount;
 FPageSize := TACROptionsEditor(Source).PageSize;
 FRandomSearchRetryCount := TACROptionsEditor(Source).RandomSearchRetryCount;
end; // Assign
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
// TACRCryptoParamsEditor
//
////////////////////////////////////////////////////////////////////////////////


{$IFNDEF SQLMEMTABLE}
//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRCryptoParamsEditor.Create;
begin
  inherited;
  FPassword := ACRDefaultPassword;
  FKeyInfo.KeySize := ACR_MAX_KEY+1;
  FillChar(FInitVector,MaxInitVectorSize,$00);
  FillChar(FKeyInfo.Key,MaxKeySize,$00);
  FCryptoAlgorithm := craNone;
  FCryptoMode := acmCTS;
  FUseInitVector := False;
end;//Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRCryptoParamsEditor.Destroy;
begin
  FillChar(FKeyInfo,SizeOf(FKeyInfo),$00);
  FillChar(FInitVector,SizeOf(FInitVector),$00);
  if (Length(FPassword) > 0) then
   FillChar(FPassword[1],Length(Password),$FF);
  FPassword := '';
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// set CryptoParams
//------------------------------------------------------------------------------
procedure TACRCryptoParamsEditor.SetCryptoParams(Params: TACRCryptoParams);
begin
  FUseInitVector := Params.UseInitVector;
  FKeyInfo := Params.KeyInfo;
  Move(Params.InitVector[0],FInitVector[0],MaxInitVectorSize);
  if (Length(FPassword) > 0) then
   FillChar(FPassword[1],Length(Password),$FF);
  FPassword := Params.Password;
  case Params.CryptoAlgorithm of
    ACR_Cipher_None:          FCryptoAlgorithm := craNone;
    ACR_Cipher_Rijndael_128:  FCryptoAlgorithm := craRijndael_128;
    ACR_Cipher_Rijndael_256:  FCryptoAlgorithm := craRijndael_256;
    ACR_Cipher_Blowfish:      FCryptoAlgorithm := craBlowfish;
    ACR_Cipher_Twofish_128:   FCryptoAlgorithm := craTwofish_128;
    ACR_Cipher_Twofish_256:   FCryptoAlgorithm := craTwofish_256;
    ACR_Cipher_Square:        FCryptoAlgorithm := craSquare;
    ACR_Cipher_Des_Single_8:  FCryptoAlgorithm := craDES_Single_8;
    ACR_Cipher_Des_Double_8:  FCryptoAlgorithm := craDES_Double_8;
    ACR_Cipher_Des_Double_16: FCryptoAlgorithm := craDES_Double_16;
    ACR_Cipher_Des_Triple_8:  FCryptoAlgorithm := craDES_Triple_8;
    ACR_Cipher_Des_Triple_16: FCryptoAlgorithm := craDES_Triple_16;
    ACR_Cipher_Des_Triple_24: FCryptoAlgorithm := craDES_Triple_24;
  end;
  case Params.CryptoMode of
    ACR_Cipher_Mode_CTS:    FCryptoMode := acmCTS;
    ACR_Cipher_Mode_CBC:    FCryptoMode := acmCBC;
    ACR_Cipher_Mode_CFB:    FCryptoMode := acmCFB;
    ACR_Cipher_Mode_OFB:    FCryptoMode := acmOFB;
  end;
end; // SetCryptoParams


//------------------------------------------------------------------------------
// GetCryptoParams
//------------------------------------------------------------------------------
function TACRCryptoParamsEditor.GetCryptoParams: TACRCryptoParams;
begin
  Result.UseInitVector := FUseInitVector;
  Result.KeyInfo := FKeyInfo;
  Move(FInitVector[0],Result.InitVector[0],MaxInitVectorSize);
  Result.Password := FPassword;
  case FCryptoAlgorithm of
    craNone:               Result.CryptoAlgorithm := ACR_Cipher_None;
    craRijndael_128:       Result.CryptoAlgorithm := ACR_Cipher_Rijndael_128;
    craRijndael_256:       Result.CryptoAlgorithm := ACR_Cipher_Rijndael_256;
    craBlowfish:           Result.CryptoAlgorithm := ACR_Cipher_Blowfish;
    craTwofish_128:        Result.CryptoAlgorithm := ACR_Cipher_Twofish_128;
    craTwofish_256:        Result.CryptoAlgorithm := ACR_Cipher_Twofish_256;
    craSquare:             Result.CryptoAlgorithm := ACR_Cipher_Square;
    craDES_Single_8:       Result.CryptoAlgorithm := ACR_Cipher_Des_Single_8;
    craDES_Double_8:       Result.CryptoAlgorithm := ACR_Cipher_Des_Double_8;
    craDES_Double_16:      Result.CryptoAlgorithm := ACR_Cipher_Des_Double_16;
    craDES_Triple_8:       Result.CryptoAlgorithm := ACR_Cipher_Des_Triple_8;
    craDES_Triple_16:      Result.CryptoAlgorithm := ACR_Cipher_Des_Triple_16;
    craDES_Triple_24:      Result.CryptoAlgorithm := ACR_Cipher_Des_Triple_24;
  end;
  case FCryptoMode of
    acmCTS:   Result.CryptoMode := ACR_Cipher_Mode_CTS;
    acmCBC:   Result.CryptoMode := ACR_Cipher_Mode_CBC;
    acmCFB:   Result.CryptoMode := ACR_Cipher_Mode_CFB;
    acmOFB:   Result.CryptoMode := ACR_Cipher_Mode_OFB;
  end;
end;// GetCryptoParams


function TACRCryptoParamsEditor.GetInitVectorValue(Index: Integer): Byte;
begin
 if (Index < 0) or (Index >= MaxInitVectorSize) then
  raise EACRException.Create(10717,ErrorLInvalidVectorIndex,[Index,MaxInitVectorSize]);
 Result := FInitVector[Index];
end; // FInitVector


procedure TACRCryptoParamsEditor.SetInitVectorValue(Index: Integer; Value: Byte);
begin
 if (Index < 0) or (Index >= MaxInitVectorSize) then
  raise EACRException.Create(10718,ErrorLInvalidVectorIndex,[Index,MaxInitVectorSize]);
 FInitVector[Index] := Value;
end; // SetInitVectorValue


function TACRCryptoParamsEditor.GetVectorSize: Integer;
begin
 Result := ACR_MAX_VECTOR+1;
end; // GetVectorSize


function TACRCryptoParamsEditor.GetKeySize: Integer;
begin
 Result := FKeyInfo.KeySize;
end; // GetVectorSize


procedure TACRCryptoParamsEditor.SetKeySize(Value: Integer);
begin
 if (Value < 0) or (Value > MaxKeySize) then
  raise EACRException.Create(10720,ErrorLInvalidKeySize,[Value,MaxKeySize]);
 FKeyInfo.KeySize := Value;
 if (Length(FPassword) > 0) then
  FillChar(FPassword[1],Length(Password),$FF);
 FPassword := '';
end; // GetVectorSize


function TACRCryptoParamsEditor.GetKeyValue(Index: Integer): Byte;
begin
 if (Index < 0) or (Index >= KeySize) then
  raise EACRException.Create(10721,ErrorLInvalidKeyIndex,[Index,KeySize]);
 Result := FKeyInfo.Key[Index];
end;


procedure TACRCryptoParamsEditor.SetKeyValue(Index: Integer; Value: Byte);
begin
 if (Index < 0) or (Index >= KeySize) then
  raise EACRException.Create(10722,ErrorLInvalidKeyIndex,[Index,KeySize]);
 FKeyInfo.Key[Index] := Value;
 if (Length(FPassword) > 0) then
  FillChar(FPassword[1],Length(Password),$FF);
 FPassword := '';
end; // SetKeyValue


function TACRCryptoParamsEditor.GetMaxKeySize: Integer;
begin
 Result := ACR_MAX_KEY+1;
end; // GetMaxKeySize


procedure TACRCryptoParamsEditor.SetKey(Key: Pointer; KeySize: Integer);
begin
 if (KeySize < 0) or (KeySize >= MaxKeySize) then
  raise EACRException.Create(10719,ErrorLInvalidKeySize,[KeySize,MaxKeySize]);
 Move(Key^,FKeyInfo.Key[0],KeySize);
 FKeyInfo.KeySize := KeySize;
 if (Length(FPassword) > 0) then
   FillChar(FPassword[1],Length(Password),$FF);
 FPassword := '';
end; //SetKey


function TACRCryptoParamsEditor.GetKey: Pointer;
begin
 Result := @FKeyInfo.Key;
end; // GetKey


procedure TACRCryptoParamsEditor.MakeRandomKey(KeySize: Integer);
begin
 if (KeySize < 0) or (KeySize >= MaxKeySize) then
  raise EACRException.Create(10723,ErrorLInvalidKeySize,[KeySize,MaxKeySize]);
 FKeyInfo.KeySize := KeySize;
 ACRGenerateRandomBuffer(@FKeyInfo.Key[0],KeySize);
 if (Length(FPassword) > 0) then
   FillChar(FPassword[1],Length(Password),$FF);
 FPassword := '';
end;

procedure TACRCryptoParamsEditor.MakeRandomInitVector;
begin
 ACRGenerateRandomBuffer(@FInitVector[0],MaxInitVectorSize);
 FUseInitVector := True;
end;


procedure TACRCryptoParamsEditor.SetInitVector(Vector: Pointer);
begin
  Move(Vector^, FInitVector[0], MaxInitVectorSize);
end; // SetInitVector


function TACRCryptoParamsEditor.GetInitVector: Pointer;
begin
  Result := @FInitVector;
end; // GetInitVector


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRCryptoParamsEditor.Assign(Source: TPersistent);
begin
 if (Source <> nil) then
  if (Source is TACRCryptoParamsEditor) then
   begin
     if (Length(FPassword) > 0) then
       FillChar(FPassword[1],Length(Password),$FF);
     FKeyInfo := TACRCryptoParamsEditor(Source).FKeyInfo;
     FInitVector := TACRCryptoParamsEditor(Source).FInitVector;
     FPassword := TACRCryptoParamsEditor(Source).FPassword;
     FCryptoAlgorithm := TACRCryptoParamsEditor(Source).CryptoAlgorithm;
     FCryptoMode := TACRCryptoParamsEditor(Source).CryptoMode;
     FUseInitVector := TACRCryptoParamsEditor(Source).FUseInitVector;
   end;
end; // Assign
{$ENDIF}


{$IFNDEF SQLMEMTABLE}
////////////////////////////////////////////////////////////////////////////////
//
// TACRNetworkSettingsEditor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// set packet size
//------------------------------------------------------------------------------
procedure TACRNetworkSettingsEditor.SetPacketSize(Value: Integer);
begin
  if (Value >= ACRMinPacketSize) and (Value <= ACRMaxPacketSize) then
   FPacketSize := Value;
end; // SetPacketSize


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRNetworkSettingsEditor.Create;
begin
  inherited Create;
  FPacketSize := ACRDefaultPacketSize;
  FMaxThreadCount := ACRMaxThreadCount;
  FConnectionParamsTunning := ACRConnectionParamsTunning;
  FTestPacketCount := ACRTestPacketCount;
  FDisconnectRetryCount := ACRDisconnectRetryCount;
  FDisconnectDelay := ACRDisconnectDelay;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRNetworkSettingsEditor.Destroy;
begin
  inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRNetworkSettingsEditor.Assign(Source: TPersistent);
begin
  if (Source <> nil) then
   if (Source is TACRNetworkSettingsEditor) then
    begin
      FPacketSize := TACRNetworkSettingsEditor(Source).PacketSize;
      FMaxThreadCount := TACRNetworkSettingsEditor(Source).MaxThreadCount;
      FConnectionParamsTunning := TACRNetworkSettingsEditor(Source).ConnectionParamsTunning;
      FTestPacketCount := TACRNetworkSettingsEditor(Source).TestPacketCount;
      FDisconnectRetryCount := TACRNetworkSettingsEditor(Source).DisconnectRetryCount;
      FDisconnectDelay := TACRNetworkSettingsEditor(Source).DisconnectDelay;
    end;
end; // Assign


//------------------------------------------------------------------------------
// Copy network settings to ConnectParams
//------------------------------------------------------------------------------
procedure TACRNetworkSettingsEditor.CopySettingsToConnectParams(var ConnectParams: TACRConnectParams);
begin
  ConnectParams.PacketSize := FPacketSize;
  ConnectParams.MaxThreadCount := FMaxThreadCount;
  ConnectParams.ConnectionParamsTunning := FConnectionParamsTunning;
  ConnectParams.TestPacketCount := FTestPacketCount;
  ConnectParams.DisconnectRetryCount := FDisconnectRetryCount;
  ConnectParams.DisconnectDelay := FDisconnectDelay;
end; // CopySettingsToConnectParams


////////////////////////////////////////////////////////////////////////////////
//
// TACRClientNetworkSettingsEditor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRClientNetworkSettingsEditor.Create;
begin
  inherited Create;
  FConnectRetryCount := ACRConnectRetryCount;
  FConnectDelay := ACRConnectDelay;
  FStartReceiveTimeOut := ACRStartReceiveTimeOut;
  FReceiveTimeOut := ACRReceiveTimeOut;
  FReceiveSleep := ACRReceiveSleep;
  FMinSendTimeOut := ACRMinSendTimeOut;
  FSendTimeOut := ACRSendTimeOut;
  FWaitForSendSleep := ACRWaitForSendSleep;
  FResendDelay := ACRResendDelay;
  FRequestDelay := ACRRequestDelay;
  FWaitForTimeOut := ACRWaitForTimeOut;
  FThreadsTerminateDelay := ACRThreadsTerminateDelay;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRClientNetworkSettingsEditor.Destroy;
begin
  inherited Destroy;
end; // Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRClientNetworkSettingsEditor.Assign(Source: TPersistent);
begin
  if (Source <> nil) then
   if (Source is TACRClientNetworkSettingsEditor) then
    begin
      inherited Assign(Source);
      FConnectRetryCount := TACRClientNetworkSettingsEditor(Source).ConnectRetryCount;
      FConnectDelay := TACRClientNetworkSettingsEditor(Source).ConnectDelay;
      FStartReceiveTimeOut := TACRClientNetworkSettingsEditor(Source).StartReceiveTimeOut;
      FReceiveTimeOut := TACRClientNetworkSettingsEditor(Source).ReceiveTimeOut;
      FReceiveSleep := TACRClientNetworkSettingsEditor(Source).ReceiveSleep;
      FMinSendTimeOut := TACRClientNetworkSettingsEditor(Source).MinSendTimeOut;
      FSendTimeOut := TACRClientNetworkSettingsEditor(Source).SendTimeOut;
      FWaitForSendSleep := TACRClientNetworkSettingsEditor(Source).WaitForSendSleep;
      FResendDelay := TACRClientNetworkSettingsEditor(Source).ResendDelay;
      FRequestDelay := TACRClientNetworkSettingsEditor(Source).RequestDelay;
      FWaitForTimeOut := TACRClientNetworkSettingsEditor(Source).WaitForTimeOut;
      FThreadsTerminateDelay := TACRClientNetworkSettingsEditor(Source).ThreadsTerminateDelay;
    end;
end; // Assign


//------------------------------------------------------------------------------
// Copy ClientNetwork settings to ConnectParams
//------------------------------------------------------------------------------
procedure TACRClientNetworkSettingsEditor.CopySettingsToConnectParams(var ConnectParams: TACRConnectParams);
begin
  inherited CopySettingsToConnectParams(ConnectParams);
  ConnectParams.ConnectRetryCount := FConnectRetryCount;
  ConnectParams.ConnectDelay := FConnectDelay;
  ConnectParams.StartReceiveTimeOut := FStartReceiveTimeOut;
  ConnectParams.ReceiveTimeOut := FReceiveTimeOut;
  ConnectParams.ReceiveSleep := FReceiveSleep;
  ConnectParams.MinSendTimeOut := FMinSendTimeOut;
  ConnectParams.SendTimeOut := FSendTimeOut;
  ConnectParams.WaitForSendSleep := FWaitForSendSleep;
  ConnectParams.ResendDelay := FResendDelay;
  ConnectParams.RequestDelay := FRequestDelay;
  ConnectParams.WaitForTimeOut := FWaitForTimeOut;
  ConnectParams.ThreadsTerminateDelay := FThreadsTerminateDelay;
end; // CopySettingsToConnectParams


////////////////////////////////////////////////////////////////////////////////
//
// TACRConnectionParamsEditor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRConnectionParamsEditor.Create;
begin
  inherited;
  FCryptoParamsEditor := TACRCryptoParamsEditor.Create;
  FServerID := ACRDefaultServerID;
  FRemoteHost := ACRDefaultHost;
  FRemotePort := ACRDefaultServerPort;
  FLocalPort := ACRDefaultClientPort;
  FDatabaseName := ACRDefaultDBName;
  FCompressionAlgorithm := caNone;
  FCompressionMode := 1;
  FNetworkSettings := TACRClientNetworkSettingsEditor.Create;
end;//Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRConnectionParamsEditor.Destroy;
begin
  FCryptoParamsEditor.Free;
  FNetworkSettings.Free;
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRConnectionParamsEditor.Assign(Source: TPersistent);
begin
  if (Source <> nil) then
   if (Source is TACRConnectionParamsEditor) then
    begin
      FServerID := TACRConnectionParamsEditor(Source).ServerID;
      FRemoteHost := TACRConnectionParamsEditor(Source).RemoteHost;
      FRemotePort := TACRConnectionParamsEditor(Source).RemotePort;
      FLocalPort := TACRConnectionParamsEditor(Source).LocalPort;
      FDatabaseName := TACRConnectionParamsEditor(Source).DatabaseName;
      FCompressionAlgorithm := TACRConnectionParamsEditor(Source).CompressionAlgorithm;
      FCompressionMode := TACRConnectionParamsEditor(Source).CompressionMode;
      FCryptoParamsEditor.Assign(TACRConnectionParamsEditor(Source).CryptoParams);
      FNetworkSettings.Assign(TACRConnectionParamsEditor(Source).NetworkSettings);
    end;
end; // Assign


//------------------------------------------------------------------------------
// return ConnectParams
//------------------------------------------------------------------------------
function TACRConnectionParamsEditor.GetConnectParams: TACRConnectParams;
begin
  Result.CompressionAlgorithm := Byte(ConvertCompressionAlgorithmToACRCompressionAlgorithm(FCompressionAlgorithm));
  Result.CompressionMode := FCompressionMode;
  Result.CryptoInfo := FCryptoParamsEditor.GetCryptoParams;
  Result.RemoteHost := FRemoteHost;
  Result.RemotePort := FRemotePort;
  Result.LocalPort := FLocalPort;
  Result.ServerID := ServerID;
  FNetworkSettings.CopySettingsToConnectParams(Result);
end; // GetConnectParams
{$ENDIF}



{$IFNDEF SQLMEMTABLE}
////////////////////////////////////////////////////////////////////////////////
//
// TACRBackupParamsEditor
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// constructor
//------------------------------------------------------------------------------
constructor TACRBackupParamsEditor.Create;
begin
  inherited;
  FCryptoParamsEditor := TACRCryptoParamsEditor.Create;
  FCompressionAlgorithm := caNone;
  FCompressionMode := 1;
  FDescription := '';
  FBlockSize := ACRDefaultBackupBlockSize;
end;//Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRBackupParamsEditor.Destroy;
begin
  FCryptoParamsEditor.Free;
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRBackupParamsEditor.Assign(Source: TPersistent);
begin
  FBlockSize := TACRBackupParamsEditor(Source).BlockSize;
  FDescription := TACRBackupParamsEditor(Source).Description;
  FCompressionAlgorithm := TACRBackupParamsEditor(Source).CompressionAlgorithm;
  FCompressionMode := TACRBackupParamsEditor(Source).CompressionMode;
  FCryptoParamsEditor.Assign(TACRBackupParamsEditor(Source).CryptoParams);
end; // Assign
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////
//
//  TACRDatabase
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// creates databases with specified directory
//------------------------------------------------------------------------------
constructor TACRDatabase.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  InitializeCriticalSection(FCSect);
  if (FSession=nil) then
    begin
      if (AOwner is TACRSession) then
        FSession:=TACRSession(AOwner)
      else
        FSession:=ACRDefaultSession;
    end;
  SessionName:=FSession.SessionName;
  FSession.AddDatabase(Self);
  FDataSets:=TList.Create;
  FKeepConnection:=True;
  if (not IsDesignMode) then
   if (Aowner <> nil) then
    if (csDesigning in AOwner.ComponentState) then
     IsDesignMode := true;
  FHandle := nil; // no manager
  {$IFDEF FILE_SERVER_VERSION}
  FExclusive := False;
  {$ELSE}
   {$IFNDEF SQLMEMTABLE}
  FExclusive := True;
   {$ENDIF}
  {$ENDIF}
  FLocalDatabase := True;
  FDatabaseName := GetTemporaryName('AccuracerDB_');
{$IFNDEF SQLMEMTABLE}
  FBackupParams := TACRBackupParamsEditor.Create;
  FLockParamsEditor := TACRLockParamsEditor.Create;
  FOptionsEditor := TACROptionsEditor.Create;
  FCryptoParamsEditor := TACRCryptoParamsEditor.Create;
  FConnectionParams := TACRConnectionParamsEditor.Create;
{$ENDIF}
end;//Create


//------------------------------------------------------------------------------
// destructor
//------------------------------------------------------------------------------
destructor TACRDatabase.Destroy;
begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Destroy starting ...'+#13#10+
              ', LocalDatabase = '+BoolToStr(FLocalDatabase,True)+#13#10+
              ', ConnectParamsDatabaseName = '+FConnectionParams.DatabaseName+#13#10+
              ', DatabaseName = '+FDatabaseName+#13#10+
              ', DatabaseFileName = '+FDatabaseFileName+#13#10+
              ', Self = '+IntToStr(Integer(Self))+#13#10+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
  Destroying;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Destroy closing ...'+#13#10+
              ', LocalDatabase = '+BoolToStr(FLocalDatabase,True)+#13#10+
              ', ConnectParamsDatabaseName = '+FConnectionParams.DatabaseName+#13#10+
              ', DatabaseName = '+FDatabaseName+#13#10+
              ', DatabaseFileName = '+FDatabaseFileName+#13#10+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
  Close;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Destroy closing ... OK'+#13#10+
              ', LocalDatabase = '+BoolToStr(FLocalDatabase,True)+#13#10+
              ', ConnectParamsDatabaseName = '+FConnectionParams.DatabaseName+#13#10+
              ', DatabaseName = '+FDatabaseName+#13#10+
              ', DatabaseFileName = '+FDatabaseFileName+#13#10+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
  DeleteCriticalSection(FCSect);
{$IFNDEF SQLMEMTABLE}
  FLockParamsEditor.Free;
  FOptionsEditor.Free;
  FCryptoParamsEditor.Free;
  FConnectionParams.Free;
  FBackupParams.Free;
{$ENDIF}
  FDataSets.Free;
  if (FSession <> nil) then
     FSession.RemoveDatabase(Self);
  inherited Destroy;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Destroy finished'+#13#10+#13#10);
 {$ENDIF}
end;//Destroy


//------------------------------------------------------------------------------
// connected := true
//------------------------------------------------------------------------------
procedure TACRDatabase.Open;
begin
  if (FHandle=nil) then
   begin
{$IFNDEF SQLMEMTABLE}
    if Assigned(BeforeConnect) then
     BeforeConnect(Self);
{$ENDIF}
     CheckDatabaseName;
     CheckSessionName(True);
     if not (HandleShared and OpenFromExistingDB) then
       begin
         FSession.LockSession;
         try
           try
             CreateHandle;
             if (FHandle = nil) then
              DatabaseError(ErrorADatabaseOpenError, Self);
             if (FLocalDatabase) then
              begin
               FHandle.ReadOnly := FReadOnly;
               FHandle.Exclusive := FExclusive;
              end;
             try
              FHandle.Connected := True;
              FReadOnly := FHandle.ReadOnly;
             except
              DestroyHandle;
              raise;
             end;
             if (not FLocalDatabase) then
               FExclusive := FHandle.Exclusive;
             {$IFNDEF SQLMEMTABLE}
             Options.SetOptions(FHandle.Options);
             CryptoParams.SetCryptoParams(FHandle.CryptoParams);
             {$ENDIF}
             // send notification
             Session.DBNotification(dbOpen,Self);
           except
             raise;
           end;
         finally
           FSession.UnlockSession;
         end;
       end
      else
        FAcquiredHandle:=False;
{$IFNDEF SQLMEMTABLE}
    if Assigned(AfterConnect) then
     AfterConnect(Self);
{$ENDIF}
   end; // FHandle = nil
end;// Open


//------------------------------------------------------------------------------
// connected := false
//------------------------------------------------------------------------------
procedure TACRDatabase.Close;
begin
  if (FHandle <> nil) then
    begin
{$IFNDEF SQLMEMTABLE}
      if Assigned(BeforeDisconnect) then
       BeforeDisconnect(Self);
{$ENDIF}
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Close session notification ...'+#13#10+
              ', LocalDatabase = '+BoolToStr(FLocalDatabase,True)+#13#10+
              ', ConnectParamsDatabaseName = '+FConnectionParams.DatabaseName+#13#10+
              ', DatabaseName = '+FDatabaseName+#13#10+
              ', DatabaseFileName = '+FDatabaseFileName+#13#10+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
      Session.DBNotification(dbClose,Self);
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Close closing datasets ...'+
              ', DatabaseName = '+FDatabaseName+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
      CloseDataSets;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Close closing datasets ... OK'+
              ', DatabaseName = '+FDatabaseName+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
      if (not FAcquiredHandle) then
        begin
         if (FHandle <> nil) then
          begin
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Close closing session ...'+
              ', DatabaseName = '+FDatabaseName+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
           DestroyHandle;
 {$IFDEF DEBUG_LOG_CLOSE_DISK_DATABASE}
 aaWriteToLog('TACRDatabase.Close session closed'+
              ', DatabaseName = '+FDatabaseName+
              ', FHandle = '+IntToStr(Integer(FHandle))+#13#10);
 {$ENDIF}
          end;
        end
      else
        FAcquiredHandle:=False;
      FHandle:=nil;
      FRefCount:=0;
{$IFNDEF SQLMEMTABLE}
      if Assigned(AfterDisconnect) then
       AfterDisconnect(Self);
{$ENDIF}
    end;
end;// Close


//------------------------------------------------------------------------------
// create database
//------------------------------------------------------------------------------
procedure TACRDatabase.CreateDatabase;
begin
  CheckDatabaseName;
  CheckSessionName(True);
  CheckConnected;

  FSession.LockSession;
  try
    CreateHandle;
    if (FHandle = nil) then
      DatabaseError(ErrorADatabaseCreate, Self);
    FHandle.CreateDatabase;
    DestroyHandle;
  finally
   FSession.UnlockSession;
  end;
end;// CreateDatabase


//------------------------------------------------------------------------------
// delete database
//------------------------------------------------------------------------------
procedure TACRDatabase.DeleteDatabase;
begin
  CheckDatabaseName;
  CheckSessionName(True);
  CheckConnected;

  DeleteFile(PChar(FDatabaseFileName));
end;// DeleteDatabase


//------------------------------------------------------------------------------
// rename database
//------------------------------------------------------------------------------
procedure TACRDatabase.RenameDatabase(NewDatabaseFileName: String);
begin
  CheckDatabaseName;
  CheckSessionName(True);
  CheckConnected;
  if (RenameFile(FDatabaseFileName, NewDatabaseFileName)) then
    FDatabaseFileName := NewDatabaseFileName;
end;// RenameDatabase


//------------------------------------------------------------------------------
// flush file buffers
//------------------------------------------------------------------------------
procedure TACRDatabase.FlushFileBuffers;
begin
  CheckDisconnected;
  FHandle.FlushFileBuffers;
end; // FlushFileBuffers


//------------------------------------------------------------------------------
// close all datasets
//------------------------------------------------------------------------------
procedure TACRDatabase.CloseDataSets;
begin
  while DataSetCount <> 0 do
   TACRDataset(DataSets[DataSetCount-1]).Disconnect;
end;// CloseDataSets


//------------------------------------------------------------------------------
// get list of tables in database file
//------------------------------------------------------------------------------
procedure TACRDatabase.GetTablesList(List: TStrings);
var bCon: Boolean;
begin
  if (not Assigned(List)) then
    raise EACRException.Create(10877,ErrorLNilPointer);
  bCon := Connected;
  if (not bCon) then
   Connected := True;
  FHandle.GetTablesList(List);
  Connected := bCon;
end;// GetTablesList


//------------------------------------------------------------------------------
// determine if table exists
//------------------------------------------------------------------------------
function TACRDatabase.TableExists(TableName: String) : Boolean;
var bCon: Boolean;
begin
  bCon := Connected;
  if (not bCon) then
   Connected := True;
  Result := FHandle.TableExists(TableName);
  Connected := bCon;
end;// TableExists


//------------------------------------------------------------------------------
// load all tables from file
//------------------------------------------------------------------------------
procedure TACRDatabase.LoadAllTablesFromStream(
                    Stream: TStream
                   );
var i,num: Integer;
    table: TACRTable;
begin
  LoadDataFromStream(num,SizeOf(num),Stream,11342);
  if (num > 0) then
   begin
    table := TACRTable.Create(Self);
    try
      for i := 0 to num - 1 do
       begin
        table.DatabaseName := Self.DatabaseName;
        table.TableName := '';
        table.LoadTableFromStream(Stream);
       end;
    finally
      table.Free;
    end;
   end;
end; // LoadAllTablesFromStream


//------------------------------------------------------------------------------
// save all tables to file
//------------------------------------------------------------------------------
procedure TACRDatabase.SaveAllTablesToStream(
                    Stream: TStream;
                    CompressionAlgorithm: TCompressionAlgorithm = caNone;
                    CompressionMode:      Byte = 0;
                    BlockSize:            Integer = 0
                  );
var num,i:  Integer;
    tables: TStringList;
    name:   string;
    table:  TACRTable;
begin
  tables := TStringList.Create;
  try
    GetTablesList(tables);
    num := tables.Count;
    SaveDataToStream(num,SizeOf(num),Stream,11339);
    if (num > 0) then
     begin
      table := TACRTable.Create(Self);
      try
        table.DatabaseName := Self.DatabaseName;
        for i := 0 to num - 1 do
         begin
          name := tables.Strings[i];
          table.TableName := name;
          table.SaveTableToStream(Stream,CompressionAlgorithm,CompressionMode,BlockSize);
         end;
      finally
        table.Free;
      end;
     end;
  finally
    tables.Free;
  end;
end; // SaveAllTablesToStream


//------------------------------------------------------------------------------
// export all tables in the database to SQL script
//------------------------------------------------------------------------------
function TACRDatabase.ExportDatabaseToSQL(
                          ExportStructure:      Boolean;
                          AddDropTableCommand:  Boolean;
                          ExportIndexes:        Boolean;
                          AddDropIndexCommand:  Boolean;
                          ExportData:           Boolean;
                          ExportBLOBFields:     Boolean;
                          UseBracketsForNames:  Boolean
                        ): String;
var bConnected: Boolean;
    i:          Integer;
    tables:     TStringList;
    table:      TACRTable;
    name:       String;
begin
 bConnected := Connected;
 Result := '';
 if (not bConnected) then
  Open;
 try
    tables := TStringList.Create;
    try
      GetTablesList(tables);
      table := TACRTable.Create(Self);
      try
        table.DatabaseName := Self.DatabaseName;
        for i := 0 to tables.Count - 1 do
         begin
          name := tables.Strings[i];
          table.TableName := name;
          if (i > 0) then
           Result := Result + CRLF;
          Result := Result + table.ExportTableToSQL(ExportStructure,
                    AddDropTableCommand,ExportIndexes,AddDropIndexCommand,
                    ExportData,ExportBLOBFields,UseBracketsForNames);
         end;
      finally
        table.Free;
      end;
    finally
      tables.Free;
    end;
 finally
  if (not bConnected) then
   Close;
 end;
end; // ExportDatabaseToSQL


{$IFNDEF SQLMEMTABLE}
//------------------------------------------------------------------------------
// retrun true if database has active transaction
//------------------------------------------------------------------------------
function TACRDatabase.GetInTransaction: Boolean;
begin
  CheckDisconnected;
  Result := FHandle.InTransaction;
end; // GetInTransaction


//------------------------------------------------------------------------------
// receive custom message from server and call defined event
//------------------------------------------------------------------------------
procedure TACRDatabase.ReceiveMessage(Buffer: PChar; Size: Integer);
{$IFDEF CLIENT_VERSION}
var MessageType: TACRMessageType;
    ms,ms1:      TACRMemoryStream;
    Text:        String;
    len:         Integer;
    Buf:         PChar;
    StreamSize:  Int64;
{$ENDIF}
begin
  CheckDisconnected;
  if (FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  {$IFDEF CLIENT_VERSION}
  ms := TACRMemoryStream.Create(Buffer);
  try
    LoadDataFromStream(MessageType,SizeOf(MessageType),ms,11369);
    case MessageType of
     aamtText:
      if (Assigned(FOnReceiveTextMessage)) then
       begin
         LoadDataFromStream(len,SizeOf(len),ms,11370);
         SetLength(Text,len);
         if (len > 0) then
          LoadDataFromStream(PChar(@Text[1])^,len,ms,11371);
         FOnReceiveTextMessage(Text);
       end;
     aamtBinary:
      if (Assigned(FOnReceiveBinaryMessage)) then
       begin
         LoadDataFromStream(len,SizeOf(len),ms,11372);
         if (len > 0) then
          begin
           Buf := MemoryManager.GetMem(len);
           try
             LoadDataFromStream(Buf^,len,ms,11373);
             FOnReceiveBinaryMessage(Buf,len);
           finally
             MemoryManager.FreeAndNilMem(Buf);
           end;
          end
         else
          FOnReceiveBinaryMessage(nil,len);
       end;
     aamtStream:
      if (Assigned(FOnReceiveStreamMessage)) then
       begin
        ms1 := TACRMemoryStream.Create;
        try
          LoadDataFromStream(StreamSize,SizeOf(StreamSize),ms,11374);
          if (StreamSize > 0) then
           ms1.LoadFromStreamWithPosition(ms,ms.Position,StreamSize);
          FOnReceiveStreamMessage(ms1);
        finally
          ms1.Free;
        end;
       end;
    end;
  finally
    ms.Free;
  end;
  {$ENDIF}
end; // ReceiveMessage


//------------------------------------------------------------------------------
// start a transaction
//------------------------------------------------------------------------------
procedure TACRDatabase.StartTransaction;
begin
  CheckDisconnected;
  FHandle.StartTransaction;
end; // StartTransaction


//------------------------------------------------------------------------------
// apply changes made by transaction
//------------------------------------------------------------------------------
procedure TACRDatabase.Commit(FlushFileBuffers: Boolean);
begin
  CheckDisconnected;
  FHandle.Commit(FlushFileBuffers);
end; // Commit


//------------------------------------------------------------------------------
// cancel changes made by transaction
//------------------------------------------------------------------------------
procedure TACRDatabase.Rollback;
begin
  CheckDisconnected;
  FHandle.Rollback;
end; // Rollback


//------------------------------------------------------------------------------
// remove all locks - called by disconnect of server session
//------------------------------------------------------------------------------
procedure TACRDatabase.RemoveAllLocks;
begin
  if (FHandle <> nil) then
   FHandle.RemoveAllLocks;
end; // RemoveAllLocks


//------------------------------------------------------------------------------
// repair database
//------------------------------------------------------------------------------
function TACRDatabase.InternalRepairDatabase(IgnoreErrors: Boolean; var Log: String ;
                                             UseLowLevelTableAccess: Boolean = False;
                                             Options: TACROptionsEditor = nil;
                                             CryptoParams: TACRCryptoParamsEditor = nil;
                                             NewDatabaseFileName: String = ''): Boolean;
var newDB:      TACRDatabase;
    i,j:        Integer;
    oldTable:   TACRTable;
    tableNames: TStringList;
    Abort,bOK:  Boolean;
    Operation:  TACRDatabaseOperation;
    Progress:   Double;
    bLowLevel:  Boolean;
    FKList:     TList;
    FKDefs:     TACRForeignKeyDefs;

 procedure DoAbort;
 begin
  Log := Log + #13#10 + 'Operation aborted';
  newDb.Close;
  newDB.DeleteDatabase;
  Self.Close;
 end; // DoAbort;

begin
 Log := '';
 Abort := False;
 if (IgnoreErrors) then
  Operation := dbopRepair
 else
  Operation := dbopCompact;
 if (Options<> nil) or (CryptoParams <> nil) then
  Operation := dbopChangeDatabaseSettings;
 if (NewDatabaseFileName <> '') then
  Operation := dbopCopy;
 if (Operation <> dbopRepair) then
   bLowLevel := False
 else
   bLowLevel := UseLowLevelTableAccess;
 CheckInactive;
 Result := False;
 if ((not FExclusive) and (Operation <> dbopCopy)) then
  raise EACRException.Create(10743,ErrorLDatabaseIsNotInExclusiveMode,[FDatabaseName]);
 tableNames := TStringList.Create;
 FKList := TList.Create;
 newDB := TACRDatabase.Create(nil);
 try
  // create new temporary database file
  newDB.DatabaseName := GetTemporaryName('TempDatabase');
  if (NewDatabaseFileName <> '') then
   newDB.DatabaseFileName := NewDatabaseFileName
  else
   newDB.DatabaseFileName := GetTempFileName;
  if Operation <> dbopChangeDatabaseSettings then
   begin
    newDB.Options.Assign(Self.Options);
    newDB.CryptoParams.Assign(Self.CryptoParams);
   end
  else
   begin
    newDB.Options.Assign(Options);
    newDB.CryptoParams.Assign(CryptoParams);
   end;
  newDB.LockParams.Assign(Self.LockParams);
  newDB.Exclusive := True;
  try
   newDB.CreateDatabase;
  except
  on e: Exception do
   begin
    Log := Log + Format(ErrorLRepairDatabaseCannotCreateDatabase,[newDB.DatabaseFileName,e.Message]);
    Exit;
   end;
  end;
 try
  Open;
 except
  on e: Exception do
   begin
    Log := Log + Format(ErrorLRepairDatabaseCannotOpenDatabase,[FDatabaseFileName,e.Message]);
    Exit;
   end;
 end;
 // get table names list
  try
   Self.GetTablesList(tableNames);
  except
  on e: Exception do
   begin
    Log := Log + Format(ErrorLRepairDatabaseCannotGetTableNamesList,[Self.DatabaseFileName,e.Message]);
    Exit;
   end;
  end;
  DoOnProgress(0,Operation,Abort);
  if (Abort) then
   begin
    DoAbort;
    Exit;
   end;
  try
   newDB.Open;
  except
  on e: Exception do
   begin
    Log := Log + Format(ErrorLRepairDatabaseCannotOpenDatabase,[newDB.DatabaseFileName,e.Message]);
    Exit;
   end;
  end;
  // copy all tables
  if (tableNames.Count > 0) then
   begin
    oldTable := TACRTable.Create(nil);
    try
      oldTable.FExclusive := True;
      oldTable.DatabaseName := Self.DatabaseName;
      for i := 0 to tableNames.Count - 1 do
       begin
        oldTable.OnProgress := FOnTableProgress;
        oldTable.Close;
        oldTable.TableName := tableNames[i];
        if (not bLowLevel) then
         begin
          FKDefs := TACRForeignKeyDefs.Create;
          FKList.Add(FKDefs);
         end
        else
         FKDefs := nil;
        if (not bLowLevel) then
         bOK := oldTable.InternalRepairOrRestructureTable(True,Log,newDB,FKDefs)
        else
         bOK := oldTable.RepairTable(Log,bLowLevel,newDB);
        if (not bOK) then
          if (not IgnoreErrors) then
           begin
            DoAbort;
            Exit;
           end;
        Progress := (i + 1) * 100;
        Progress := Progress / tableNames.Count;
        DoOnProgress(Progress,Operation,Abort);
        if (Abort) then
         begin
          DoAbort;
          Exit;
         end;
       end; // copy tables
      // recreate foreign keys 
      if (not bLowLevel) then
       for i := 0 to tableNames.Count-1 do
        begin
         oldTable.Close;
         oldTable.DatabaseName := newDB.DatabaseName;
         oldTable.TableName := tableNames[i];
         FKDefs := FKList.Items[i];
         if (FKDefs <> nil) then
          for j := 0 to FKDefs.Count-1 do
           try
            oldTable.AddForeignKey(FKDefs.Items[j]);
           except
            on e: Exception do
             begin
               Log := Log + Format(ErrorLCannotAddForeignKey,[FKDefs.Items[j].Name,oldTable.TableName,e.Message]);
             end
            else
             begin
               Log := Log + Format(ErrorLCannotAddForeignKey,[FKDefs.Items[j].Name,oldTable.TableName,'']);
             end;
           end;
        end;
    finally
     oldTable.Free;
    end;
   end; // there are some tables in the database
  try
   newDB.Close;
  except
  on e: Exception do
   begin
    Log := Log + Format(ErrorLRepairDatabaseCannotCloseDatabase,[newDB.DatabaseFileName,e.Message]);
    Exit;
   end;
  end;
  try
   Self.Close;
  except
  on e: Exception do
   begin
    Log := Log + Format(ErrorLRepairDatabaseCannotCloseDatabase,[Self.DatabaseFileName,e.Message]);
    Exit;
   end;
  end;
  if (Operation <> dbopCopy) then
   begin
    Self.DeleteDatabase;
    newDB.RenameDatabase(FDatabaseFileName);
   end;
  DoOnProgress(100.0,Operation,Abort);
  Result := True;
 finally
  newDB.Free;
  tableNames.Free;
  for i := 0 to FKList.Count-1 do
   TACRForeignKeyDefs(FKList.Items[i]).Free;
  FKList.Free; 
 end;
end; // InternalRepairDatabase


//------------------------------------------------------------------------------
// compact database
//------------------------------------------------------------------------------
procedure TACRDatabase.CompactDatabase;
var Log: String;
begin
 if ((DatabaseName = ACRMemoryDatabaseName) or (DatabaseName = ACRTemporaryDatabaseName)) then
  raise EACRException.Create(10740,ErrorLOperationCanBePerformedOnlyOnDiskDatabase);
 if (not InternalRepairDatabase(False,Log)) then
  raise EACRException.Create(10739,ErrorLCompactDatabaseFailed,[FDatabaseFileName,Log]);
end; // CompactDatabase


//------------------------------------------------------------------------------
// repair database
//------------------------------------------------------------------------------
function TACRDatabase.RepairDatabase(UseLowLevelTableAccess: Boolean = False): Boolean;
var Log: String;
begin
 Result := RepairDatabase(Log, UseLowLevelTableAccess);
end; // RepairDatabase


//------------------------------------------------------------------------------
// repair database
//------------------------------------------------------------------------------
function TACRDatabase.RepairDatabase(var Log: String; UseLowLevelTableAccess: Boolean): Boolean;
begin
 if ((DatabaseName = ACRMemoryDatabaseName) or (DatabaseName = ACRTemporaryDatabaseName)) then
  raise EACRException.Create(10741,ErrorLOperationCanBePerformedOnlyOnDiskDatabase);
 Result := InternalRepairDatabase(True,Log,UseLowLevelTableAccess);
end; // RepairDatabase


//------------------------------------------------------------------------------
// change database settings
//------------------------------------------------------------------------------
function TACRDatabase.ChangeDatabaseSettings(Options: TACROptionsEditor;
                                 CryptoParams: TACRCryptoParamsEditor): Boolean;
var Log: String;
begin
 if ((DatabaseName = ACRMemoryDatabaseName) or (DatabaseName = ACRTemporaryDatabaseName)) then
  raise EACRException.Create(50029,ErrorFOperationCanBePerformedOnlyOnDiskDatabase);
 Result := InternalRepairDatabase(False, Log, False, Options, CryptoParams);
end; // ChangeDatabaseSettings


//------------------------------------------------------------------------------
// get total count of pages
//------------------------------------------------------------------------------
function TACRDatabase.GetFormatVersion: Double;
begin
  Result := ACRVersion;
  if (Assigned(FHandle)) then
   Result := FHandle.GetFormatVersion;
end; // GetFormatVersion


//------------------------------------------------------------------------------
// empty proc for displaying property in Object Inspector
//------------------------------------------------------------------------------
procedure TACRDatabase.SetFormatVersion(Value: Double);
begin
;
end;


//------------------------------------------------------------------------------
// get total count of pages
//------------------------------------------------------------------------------
function TACRDatabase.GetTotalPageCount: Integer;
begin
  Result := 0;
  if (Assigned(FHandle)) then
   Result := FHandle.GetTotalPageCount;
end; // GetTotalPageCount


//------------------------------------------------------------------------------
// get count of free pages
//------------------------------------------------------------------------------
function TACRDatabase.GetFreePageCount: Integer;
begin
  Result := 0;
  if (Assigned(FHandle)) then
   Result := FHandle.GetFreePageCount;
end; // GetFreePageCount


//------------------------------------------------------------------------------
// return number of used pages
//------------------------------------------------------------------------------
function TACRDatabase.GetUsedPageCount: Integer;
begin
  Result := 0;
  if (Assigned(FHandle)) then
    Result := GetTotalPageCount - GetFreePageCount;
end; // GetUsedPageCount


//------------------------------------------------------------------------------
// return density = usedpagecoount / totalpagecount * 100.0 %
//------------------------------------------------------------------------------
function TACRDatabase.GetDensity: Double;
var d,d1: Double;
begin
  Result := 0;
  if (Assigned(FHandle)) then
   begin
    d := GetTotalPageCount;
    d1 := GetUsedPageCount;
    if (d = 0) then
     Result := 0
    else
     Result := d1 / d * 100.0;
   end;
end; // GetDensity


//------------------------------------------------------------------------------
// return true if database is encrypted
//------------------------------------------------------------------------------
function TACRDatabase.IsDatabaseEncrypted: Boolean;
var DatabaseClosed: Boolean;
begin
  if (LocalDatabase) then
   if (not Self.Exists) then
    raise EACRException.Create(10735,ErrorLDatabaseFileDoesNotExitst,[FDatabaseFileName]);
  CheckDatabaseName;
  CheckSessionName(True);

  FSession.LockSession;
  try
    DatabaseClosed := (FHandle = nil);
    if (DatabaseClosed) then
     CreateHandle;
    try
      if (FHandle = nil) then
        DatabaseError(ErrorADatabaseCreate, Self);
      Result := FHandle.IsDatabaseEncrypted;
    finally
      if (DatabaseClosed) then
        DestroyHandle;
    end;
  finally
   FSession.UnlockSession;
  end;
end; // IsDatabaseEncrypted


//------------------------------------------------------------------------------
// return true if database is encrypted by password, false if by key
//------------------------------------------------------------------------------
function TACRDatabase.IsDatabaseEncryptedByPassword: Boolean;
var DatabaseClosed: Boolean;
begin
  if (LocalDatabase) then
   if (not Self.Exists) then
    raise EACRException.Create(10736,ErrorLDatabaseFileDoesNotExitst,[FDatabaseFileName]);
  CheckDatabaseName;
  CheckSessionName(True);

  FSession.LockSession;
  try
    DatabaseClosed := (FHandle = nil);
    if (DatabaseClosed) then
     CreateHandle;
    try
      if (FHandle = nil) then
        DatabaseError(ErrorADatabaseCreate, Self);
      Result := FHandle.IsDatabaseEncryptedByPassword;
    finally
      if (DatabaseClosed) then
        DestroyHandle;
    end;
  finally
   FSession.UnlockSession;
  end;
end; // IsDatabaseEncryptedByPassword


//------------------------------------------------------------------------------
// return true if CryptoParams are valid
//------------------------------------------------------------------------------
function TACRDatabase.IsCryptoParamsValid: Boolean;
var DatabaseClosed: Boolean;
begin
  if (LocalDatabase) then
   if (not Self.Exists) then
    raise EACRException.Create(10737,ErrorLDatabaseFileDoesNotExitst,[FDatabaseFileName]);
  CheckDatabaseName;
  CheckSessionName(True);

  FSession.LockSession;
  try
    DatabaseClosed := (FHandle = nil);
    if (DatabaseClosed) then
     CreateHandle;
    try
      if (FHandle = nil) then
        DatabaseError(ErrorADatabaseCreate, Self);
      FHandle.CryptoParams := FCryptoParamsEditor.GetCryptoParams;
      Result := FHandle.IsCryptoParamsValid;
    finally
      if (DatabaseClosed) then
        DestroyHandle;
    end;
  finally
   FSession.UnlockSession;
  end;
end; // IsCryptoParamsValid


//------------------------------------------------------------------------------
// send message
//------------------------------------------------------------------------------
procedure TACRDatabase.SendMessage(const Text: String);
{$IFDEF CLIENT_VERSION}
var
  ClientSession:  TACRClientSession;
  ms:             TACRMemoryStream;
  MessageType:    TACRMessageType;
  len:            Integer;
{$ENDIF}
begin
  CheckDisconnected;
  if (FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  {$IFDEF CLIENT_VERSION}
  ClientSession := TACRClientSession(FHandle);
  MessageType := aamtText;
  len := Length(Text);
  ms := TACRMemoryStream.Create();
  try
   SaveDataToStream(MessageType,SizeOf(MessageType),ms,11273);
   SaveDataToStream(len,SizeOf(len),ms,11274);
   if (len > 0) then
     SaveDataToStream(PChar(@Text[1])^,len,ms,11275);
   ClientSession.SendMessage(ms.Buffer,ms.Size);
  finally
    ms.Free;
  end;
  {$ENDIF}
end; // SendMessage


//------------------------------------------------------------------------------
// send message
//------------------------------------------------------------------------------
procedure TACRDatabase.SendMessage(Buffer: PChar; Size: Integer);
{$IFDEF CLIENT_VERSION}
var
  ClientSession:  TACRClientSession;
  ms:             TACRMemoryStream;
  MessageType:    TACRMessageType;
{$ENDIF}
begin
  CheckDisconnected;
  if (FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  {$IFDEF CLIENT_VERSION}
  ClientSession := TACRClientSession(FHandle);
  if (ClientSession = nil) then
   raise EACRException.Create(11251,ErrorLClientIsNotConnected);
  MessageType := aamtBinary;
  ms := TACRMemoryStream.Create();
  try
   SaveDataToStream(MessageType,SizeOf(MessageType),ms,11276);
   SaveDataToStream(Size,SizeOf(Size),ms,11277);
   if (Size > 0) then
     SaveDataToStream(Buffer^,Size,ms,11278);
   ClientSession.SendMessage(ms.Buffer,ms.Size);
  finally
    ms.Free;
  end;
  {$ENDIF}
end; // SendMessage


//------------------------------------------------------------------------------
// send message
//------------------------------------------------------------------------------
procedure TACRDatabase.SendMessage(Stream: TStream);
{$IFDEF CLIENT_VERSION}
var
  ClientSession:  TACRClientSession;
  ms:             TACRMemoryStream;
  MessageType:    TACRMessageType;
  Size:           Int64;
{$ENDIF}
begin
  CheckDisconnected;
  if (FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  {$IFDEF CLIENT_VERSION}
  ClientSession := TACRClientSession(FHandle);
  if (ClientSession = nil) then
   raise EACRException.Create(11252,ErrorLClientIsNotConnected);
  MessageType := aamtStream;
  Size := Stream.Size;
  ms := TACRMemoryStream.Create();
  try
   SaveDataToStream(MessageType,SizeOf(MessageType),ms,11279);
   SaveDataToStream(Size,SizeOf(Size),ms,11280);
   if (Size > 0) then
    begin
     if (Stream is TACRStream) then
      TACRStream(Stream).SaveToStream(ms)
     else
      ms.CopyFrom(Stream,Size);
    end;
   ClientSession.SendMessage(ms.Buffer,ms.Size);
  finally
    ms.Free;
  end;
  {$ENDIF}
end; // SendMessage


//------------------------------------------------------------------------------
// makes Exe database from edb file
//------------------------------------------------------------------------------
procedure TACRDatabase.MakeExeDatabase(ExeFileName, ExeDatabaseFileName: string);
var DatabaseClosed: Boolean;
begin
  CheckDatabaseName;
  CheckSessionName(True);
  CheckConnected;
  if (not FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  if (LocalDatabase) then
   if (not Self.Exists) then
    raise EACRException.Create(11286,ErrorLDatabaseFileDoesNotExitst,[FDatabaseFileName]);
  if (not FExclusive) then
   raise EACRException.Create(11287,ErrorLDatabaseIsNotInExclusiveMode,[FDatabaseName]);
  FSession.LockSession;
  try
    DatabaseClosed := (FHandle = nil);
    if (DatabaseClosed) then
      CreateHandle;
    try
      if (FHandle = nil) then
        DatabaseError(ErrorADatabaseCreate, Self);
      FHandle.MakeExeDatabase(ExeFileName,ExeDatabaseFileName);
    finally
      if (DatabaseClosed) then
        DestroyHandle;
    end;
  finally
   FSession.UnlockSession;
  end;
end; // MakeExeDatabase


//------------------------------------------------------------------------------
// removes database file from executable database file
//------------------------------------------------------------------------------
procedure TACRDatabase.RemoveDatabaseFromExe;
var DatabaseClosed: Boolean;
begin
  CheckDatabaseName;
  CheckSessionName(True);
  CheckConnected;
  if (not FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  if (LocalDatabase) then
   if (not Self.Exists) then
    raise EACRException.Create(11297,ErrorLDatabaseFileDoesNotExitst,[FDatabaseFileName]);
  if (not FExclusive) then
   raise EACRException.Create(11299,ErrorLDatabaseIsNotInExclusiveMode,[FDatabaseName]);
  FSession.LockSession;
  try
    DatabaseClosed := (FHandle = nil);
    if (DatabaseClosed) then
      CreateHandle;
    try
      if (FHandle = nil) then
        DatabaseError(ErrorADatabaseCreate, Self);
      FHandle.RemoveDatabaseFromExe;
    finally
      if (DatabaseClosed) then
        DestroyHandle;
    end;
  finally
   FSession.UnlockSession;
  end;
end; // RemoveDatabaseFromExe


//------------------------------------------------------------------------------
// returns true if this file is an Accuracer database
//------------------------------------------------------------------------------
function TACRDatabase.IsAccuracerDatabaseFile: Boolean;
var DatabaseClosed: Boolean;
begin
  CheckDatabaseName;
  CheckSessionName(True);
  if (not FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  if (not Self.Exists) then
   begin
     Result := False;
     Exit;
   end;
  FSession.LockSession;
  try
    DatabaseClosed := (FHandle = nil);
    if (DatabaseClosed) then
      CreateHandle;
    try
      if (FHandle = nil) then
        DatabaseError(ErrorADatabaseCreate, Self);
      Result := FHandle.IsAccuracerDatabaseFile;
    finally
      if (DatabaseClosed) then
        DestroyHandle;
    end;
  finally
   FSession.UnlockSession;
  end;
end; // IsAccuracerDatabaseFile


//------------------------------------------------------------------------------
// return true if file was copied
//------------------------------------------------------------------------------
function TACRDatabase.CopyDatabase(const NewDatabaseFileName: String): Boolean;
var Log: String;
begin
  Result := CopyDatabase(Log,NewDatabaseFileName);
end; // CopyDatabase


//------------------------------------------------------------------------------
// return true if file was copied
//------------------------------------------------------------------------------
function TACRDatabase.CopyDatabase(var Log: String; const NewDatabaseFileName: String): Boolean; 
begin
  if (not FLocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  CheckConnected;
  Result := InternalRepairDatabase(False,Log,False,nil,nil,NewDatabaseFileName);
end; // CopyDatabase


//------------------------------------------------------------------------------
// on backup progress
//------------------------------------------------------------------------------
procedure TACRDatabase.DoOnBackupProgress(Sender: TObject; Progress: Double; var Abort: Boolean);
begin
  if (Assigned(FOnProgress)) then
   FOnProgress(Self,Progress,dbopBackup,Abort);
end; // DoOnBackupProgress


//------------------------------------------------------------------------------
// on restore progress
//------------------------------------------------------------------------------
procedure TACRDatabase.DoOnRestoreProgress(Sender: TObject; Progress: Double; var Abort: Boolean);
begin
  if (Assigned(FOnProgress)) then
   FOnProgress(Self,Progress,dbopRestore,Abort);
end; // DoOnRestoreProgress


//------------------------------------------------------------------------------
// load backup header
//------------------------------------------------------------------------------
procedure TACRDatabase.LoadBackupHeader(
                     Stream:            TACRStream;
                     var Header:        TACRBackupHeader;
                     out Description:   String;
                     Tables:            TStrings = nil);
var
    MS:             TACRMemoryStream;
    CryptoInfo:     TACRCryptoInfo;
    Len,I:          Integer;
    S:              String;
begin
  Stream.Position := 0;
  CryptoInfo := FBackupParams.CryptoParams.GetCryptoParams;
  LoadDataFromStream(Header,SizeOf(Header),Stream,11356);
  CryptoInfo.CryptoAlgorithm := Header.CryptoHeader.CryptoAlgorithm;
  CryptoInfo.CryptoMode := Header.CryptoHeader.CryptoMode;
  if (Header.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
   ACRDecryptBuffer(CryptoInfo,PCHar(@Header.DescHeader),SizeOf(Header.DescHeader));
  if (Header.DescHeader.Size > 0) then
   begin
    MS := TACRMemoryStream.Create;
    try
      MS.LoadFromStreamWithPosition(Stream,Stream.Position,Header.DescHeader.Size);
      if (Header.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
       ACRDecryptBuffer(CryptoInfo,MS.Buffer,MS.Size);
      MS.Position := 0;
      if (Header.DescHeader.DescLength > 0) then
       begin
        SetLength(Description,Header.DescHeader.DescLength);
        LoadDataFromStream(PChar(@Description[1])^,Header.DescHeader.DescLength,MS,11357);
       end;
      if (Tables <> nil) then
       begin
        Tables.Clear;
        for I := 0 to Header.DescHeader.TableCount - 1 do
         begin
          LoadDataFromStream(Len,SizeOf(Len),MS,11358);
          SetLength(S,Len);
          if (Len > 0) then
           LoadDataFromStream(PChar(@S[1])^,Len,MS,11359);
          Tables.Add(S);
         end;
       end;
    finally
      MS.Free;
    end;
   end;
end; // LoadBackupHeader


//------------------------------------------------------------------------------
// backup
//------------------------------------------------------------------------------
procedure TACRDatabase.Backup(const BackupFileName: String);
var 
    DBFile,
    BackupFile:     TACRFileStream;
    MS:             TACRMemoryStream;
    Header:         TACRBackupHeader;
    CryptoInfo:     TACRCryptoInfo;
    OnFileProgress: TACRProgressEvent;
    Len,I:          Integer;
    Tables:         TStringList;
begin
  if (not LocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  CheckInactive;
  OnFileProgress := nil;
  if (Assigned(FOnProgress)) then
   OnFileProgress := DoOnBackupProgress;
  CryptoInfo := FBackupParams.CryptoParams.GetCryptoParams;
  Len := Length(FBackupParams.Description);
  Tables := TStringList.Create;
  try
    GetTablesList(Tables);
    Header.Signature := ACRBackupSignature;
    Header.BlockSize := FBackupParams.BlockSize;
    Header.CompressionAlgorithm := Byte(FBackupParams.CompressionAlgorithm);
    Header.CompressionMode := FBackupParams.CompressionMode;
    Header.CryptoHeader := ACRCreateCryptoHeader(CryptoInfo);
    Header.DescHeader.TableCount := Tables.Count;
    Header.DescHeader.DescLength := Len;
    DBFile := TACRFileStream.Create(FDatabaseFileName,fmOpenRead and fmShareExclusive);
    try
     Header.DescHeader.Date := Now;
     Header.DescHeader.FileSize := DBFile.Size;
     BackupFile := TACRFileStream.Create(BackupFileName,fmCreate);
     try
       MS := TACRMemoryStream.Create;
       try
         if (Len > 0) then
          SaveDataToStream(PChar(@FBackupParams.Description[1])^,Len,MS,11353);
         if (Tables.Count > 0) then
          for I := 0 to Tables.Count - 1 do
           begin
            Len := Length(Tables.Strings[I]);
            SaveDataToStream(Len,SizeOf(Len),MS,11354);
            if (Len > 0) then
             SaveDataToStream(PChar(@(Tables.Strings[i][1]))^,Len,MS,11355);
           end;
         Header.DescHeader.Size := MS.Size;
         if (CryptoInfo.CryptoAlgorithm <> ACR_Cipher_None) then
          begin
           ACREncryptBuffer(CryptoInfo,PCHar(@Header.DescHeader),SizeOf(Header.DescHeader));
           ACREncryptBuffer(CryptoInfo,MS.Buffer,MS.Size);
          end;
         BackupFile.WriteBuffer(Header,SizeOf(Header));
         MS.SaveToStream(BackupFile);
       finally
         MS.Free;
       end;
       BackupFile.Position := BackupFile.Size;
       CompressAndEncryptStream(CryptoInfo,
                                Header.CompressionAlgorithm,
                                Header.CompressionMode,
                                Header.BlockSize,
                                DBFile,
                                BackupFile,
                                OnFileProgress);
     finally
       BackupFile.Free;
     end;
    finally
      DBFile.Free;
    end;
  finally
    Tables.Free;
  end;
end; // Backup


//------------------------------------------------------------------------------
// restore
//------------------------------------------------------------------------------
procedure TACRDatabase.Restore(const BackupFileName: String);
var BackupFile:         TACRFileStream;
    DBFile:             TACRFileStream;
    Header:             TACRBackupHeader;
    CryptoInfo:         TACRCryptoInfo;
    OnFileProgress:     TACRProgressEvent;
begin
  if (not LocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  CheckInactive;
  OnFileProgress := nil;
  if (Assigned(FOnProgress)) then
   OnFileProgress := DoOnRestoreProgress;
  if (not SysUtils.FileExists(BackupFileName)) then
   DatabaseError(ErrorLBackupFileDoesNotExists);
  CryptoInfo := FBackupParams.CryptoParams.GetCryptoParams;
  BackupFile := TACRFileStream.Create(BackupFileName,fmOpenRead or fmShareDenyWrite);
  try
    if (BackupFile.Size < SizeOf(Header)) then
     DatabaseError(ErrorLInvalidBackupFile)
    else
     begin
      BackupFile.ReadBuffer(Header,SizeOf(Header));
      CryptoInfo.CryptoAlgorithm := Header.CryptoHeader.CryptoAlgorithm;
      CryptoInfo.CryptoMode := Header.CryptoHeader.CryptoMode;
      if (Header.Signature <> ACRBackupSignature) then
       DatabaseError(ErrorLInvalidBackupFile);
      if (Header.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
       if (not ACRIsKeyValid(Header.CryptoHeader,FBackupParams.CryptoParams.GetCryptoParams)) then
        DatabaseError(ErrorLBackupFileInvalidCryptoParams)
       else
        ACRDecryptBuffer(CryptoInfo,PCHar(@Header.DescHeader),SizeOf(Header.DescHeader));
      BackupFile.Position := BackupFile.Position + Header.DescHeader.Size;
      DBFile := TACRFileStream.Create(FDatabaseFileName,fmCreate);
      try
        DecompressAndDecryptStream(CryptoInfo,
                                Header.CompressionAlgorithm,
                                Header.CompressionMode,
                                Header.BlockSize,
                                BackupFile,
                                DBFile,
                                OnFileProgress);
      finally
        DBFile.Free;
      end;
     end;
  finally
    BackupFile.Free;
  end;
end; // Restore


//------------------------------------------------------------------------------
// get backup info
//------------------------------------------------------------------------------
function TACRDatabase.GetBackupInfo(const BackupFileName: String; Tables: TStrings = nil): TACRBackupInfo;
var FS:         TACRFileStream;
    Header:     TACRBackupHeader;
    bShow:      Boolean;
    CryptoInfo: TACRCryptoInfo;
begin
  if (not LocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  if (not SysUtils.FileExists(BackupFileName)) then
   DatabaseError(ErrorLBackupFileDoesNotExists);
  FS := TACRFileStream.Create(BackupFileName,fmOpenRead or fmShareDenyNone);
  try
    if (FS.Size < SizeOf(Header)) then
     DatabaseError(ErrorLInvalidBackupFile)
    else
     begin
      FS.ReadBuffer(Header,SizeOf(Header));
      if (Header.Signature <> ACRBackupSignature) then
       DatabaseError(ErrorLInvalidBackupFile);
      Result.Encrypted := (Header.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None);
      Result.EncryptedByPassword := (Header.CryptoHeader.CryptoAskPassword <> 0);
      bShow := True;
      CryptoInfo := FBackupParams.CryptoParams.GetCryptoParams;
      CryptoInfo.CryptoAlgorithm := Header.CryptoHeader.CryptoAlgorithm;
      CryptoInfo.CryptoMode := Header.CryptoHeader.CryptoMode;
      if (Result.Encrypted) then
       if (not ACRIsKeyValid(Header.CryptoHeader,CryptoInfo)) then
         bShow := False;
      if (bShow) then
       begin
        LoadBackupHeader(FS,Header,Result.Description,Tables);
        Result.TableCount := Header.DescHeader.TableCount;
        Result.Date := Header.DescHeader.Date;
        Result.FileSize := Header.DescHeader.FileSize;
       end;
     end;
  finally
    FS.Free;
  end;
end; // GetBackupInfo


//------------------------------------------------------------------------------
// is Accuracer backup file
//------------------------------------------------------------------------------
function TACRDatabase.IsAccuracerBackupFile(const BackupFileName: String): Boolean;
var FS:       TACRFileStream;
    Header:   TACRBackupHeader;
begin
  if (not LocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  Result := SysUtils.FileExists(BackupFileName);
  if (Result) then
   begin
    FS := TACRFileStream.Create(BackupFileName,fmOpenRead or fmShareDenyNone);
    try
      if (FS.Size < SizeOf(Header)) then
       Result := False
      else
       begin
        FS.ReadBuffer(Header,SizeOf(Header));
        Result := (Header.Signature = ACRBackupSignature);
       end;
    finally
      FS.Free;
    end;
   end;
end; // IsAccuracerBackupFile


//------------------------------------------------------------------------------
// is Accuracer backup file can be restored using these crypto params
//------------------------------------------------------------------------------
function TACRDatabase.IsAccuracerBackupFileCryptoParamsValid(const BackupFileName: String): Boolean;
var FS:         TACRFileStream;
    Header:     TACRBackupHeader;
    CryptoInfo: TACRCryptoInfo;
begin
  if (not LocalDatabase) then
   DatabaseError(ErrorLOperationIsNotSupported);
  Result := SysUtils.FileExists(BackupFileName);
  if (Result) then
   begin
    FS := TACRFileStream.Create(BackupFileName,fmOpenRead or fmShareDenyNone);
    try
      if (FS.Size < SizeOf(Header)) then
       Result := False
      else
       begin
        FS.ReadBuffer(Header,SizeOf(Header));
        Result := (Header.Signature = ACRBackupSignature);
        if (Result) then
         if (Header.CryptoHeader.CryptoAlgorithm <> ACR_Cipher_None) then
          begin
            CryptoInfo := FBackupParams.CryptoParams.GetCryptoParams;
            CryptoInfo.CryptoAlgorithm := Header.CryptoHeader.CryptoAlgorithm;
            CryptoInfo.CryptoMode := Header.CryptoHeader.CryptoMode;
            Result := ACRIsKeyValid(Header.CryptoHeader,CryptoInfo);
          end;
       end;
    finally
      FS.Free;
    end;
   end;
end; // IsAccuracerBackupFileCryptoParamsValid
{$ENDIF}


//------------------------------------------------------------------------------
// connect / disconnect
//------------------------------------------------------------------------------
procedure TACRDatabase.SetConnected(value: boolean);
begin
  if (csReading in ComponentState) then
    FStreamedConnected := Value
  else
   begin
    if Value = GetConnected then
     Exit;
    try
      if Value then
       Open
      else
       Close;
    except
     on e: Exception do
      if (csDesigning in ComponentState) then
       MessageDlg(e.Message,mtError,[mbOK],0)
      else
       raise;
    end;
   end;
end;//SetConnected


//------------------------------------------------------------------------------
// set specified database name
//------------------------------------------------------------------------------
procedure TACRDatabase.SetDatabaseName(Value: string);
var i: Integer;
begin
  if FDatabaseName <> Value then
   begin
    if (not (csReading in ComponentState)) then
     begin
      CheckInactive;
      ValidateName(Value);
     end;
    ACRDatasetsList := ACRDatasets.LockList;
    try
     for i := 0 to ACRDatasetsList.Count - 1 do
      if (TACRDataset(ACRDatasetsList.Items[i]).DatabaseName = FDatabaseName) then
       TACRDataset(ACRDatasetsList.Items[i]).ChangeCurrentDatabaseName(Value);
    finally
      ACRDatasets.UnlockList;
    end;
    FDatabaseName := Value;
   end;
end;//SetDatabaseName


//------------------------------------------------------------------------------
// sets handle
//------------------------------------------------------------------------------
procedure TACRDatabase.SetHandle(Value: TACRBaseSession);
var
  DBSession: TACRSessionComponentManager;
begin
   if Connected then
      Close;
   if (Value <> nil) then
     begin
      DBSession := Value.SessionComponentManager;
      CheckDatabaseName;
      CheckSessionName(True);
      // database handle owned by another session
      if (FSession.FHandle <> DBSession) then
        DatabaseError(ErrorADatabaseHandleSet, Self);
      FHandle:=Value;
      Session.DBNotification(dbOpen,Self);
      FAcquiredHandle:=True;
     end;
end;// SetHandle


//------------------------------------------------------------------------------
// set specified file name
//------------------------------------------------------------------------------
procedure TACRDatabase.SetDatabaseFileName(Value: string);
begin
  if csReading in ComponentState then
    FDatabaseFileName := Value
  else
  if FDatabaseFileName <> Value then
   begin
    Connected := False;
    FDatabaseFileName := Value;
   end;
end;// SetDatabaseFileName


//------------------------------------------------------------------------------
// is database file exists
//------------------------------------------------------------------------------
function TACRDatabase.GetExists: boolean;
var bCon: Boolean;
begin
  bCon := Assigned(FHandle);
  if (not bCon) then
   CreateHandle;
  Result := FHandle.GetDatabaseExists;
  if (not bCon) then
   DestroyHandle;
end;// GetExists


//------------------------------------------------------------------------------
// get database manager
//------------------------------------------------------------------------------
procedure TACRDatabase.CreateHandle;
begin
  FHandle := nil;
  if (FLocalDatabase) then
   begin
    FHandle := TACRLocalSession.Create;
    FHandle.DatabaseName := FDatabaseName;
    FHandle.DatabaseFileName := FDatabaseFileName;
    {$IFNDEF SQLMEMTABLE}
    FHandle.LockParams := FLockParamsEditor.GetLockParams;
    FHandle.Options := FOptionsEditor.GetOptions;
    FHandle.CryptoParams := FCryptoParamsEditor.GetCryptoParams;
    {$ENDIF}
   end // LocalDatabase
  else
   begin
    {$IFDEF CLIENT_VERSION}
    {$IFNDEF SQLMEMTABLE}
      FHandle := TACRClientSession.Create;
      FHandle.DatabaseName := FConnectionParams.DatabaseName;
      FHandle.ConnectParams := FConnectionParams.GetConnectParams;
      FHandle.CryptoParams := FCryptoParamsEditor.GetCryptoParams;
      TACRClientSession(FHandle).OnReceiveMessage := Self.ReceiveMessage;
      TACRClientSession(FHandle).OnDisconnectRemoteDatasets := Self.DisconnectAllRemoteDatasets;
      TACRClientSession(FHandle).OnError := Self.DoOnError;
    {$ENDIF}
    {$ENDIF}
   end;
  FAcquiredHandle := False;
end;// CreateHandle


//------------------------------------------------------------------------------
// release database manager
//------------------------------------------------------------------------------
procedure TACRDatabase.DestroyHandle;
begin
 if (Assigned(FHandle)) then
  begin
   if (FHandle.Connected) then
    FHandle.Connected := False;
   FHandle.Free;
  end;
 FHandle := nil;
end;// DestroyHandle


//------------------------------------------------------------------------------
// disconnect all remote datasets
//------------------------------------------------------------------------------
procedure TACRDatabase.DisconnectAllRemoteDatasets(ServerShutdown: Boolean);
var i: Integer;
begin
  if ((not FLocalDatabase) and (Connected)) then
   begin
    EnterCriticalSection(FCSect);
    if (ServerShutdown) then
     begin
      if (Assigned(FBeforeServerShutdown)) then
        FBeforeServerShutdown(Self);
     end
    else
     begin
      if (Assigned(FBeforeConnectionLost)) then
        FBeforeConnectionLost(Self);
     end;
    try
      for i := 0 to FDataSets.Count - 1 do
       if (TDataset(FDatasets[i]).Active) then
        begin
          if (TACRDataset(FDataSets[i]).IsTable) then
           if (TACRTable(FDataSets[i]).Handle <> nil) then
            // skip sending CloseTable request to server
            TACRTable(FDataSets[i]).Handle.IsOpen := False;
          TDataset(FDatasets[i]).Close;
        end;
    finally
      LeaveCriticalSection(FCSect);
      Close;
      if (ServerShutdown) then
       begin
        if (Assigned(FAfterServerShutdown)) then
          FAfterServerShutdown(Self);
       end
      else
       begin
        if (Assigned(FAfterConnectionLost)) then
          FAfterConnectionLost(Self);
       end;
    end;
   end;
end; // DisconnectAllRemoteDatasets


//------------------------------------------------------------------------------
// on error
//------------------------------------------------------------------------------
procedure TACRDatabase.DoOnError(
                       const ErrorCode:    Integer;
                       const NativeError:  Integer;
                       const ErrorMessage: String
                       );
begin
  if (Assigned(FOnError)) then
   FOnError(Self,ErrorCode,NativeError,ErrorMessage);
end; // DoOnError


//------------------------------------------------------------------------------
// loaded
//------------------------------------------------------------------------------
procedure TACRDatabase.Loaded;
begin
  inherited Loaded;
  if FStreamedConnected then
    Open
  else
    CheckSessionName(False);
end;// Loaded


//------------------------------------------------------------------------------
// sends notification
//------------------------------------------------------------------------------
procedure TACRDatabase.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent,Operation);
  if (Operation = opRemove) and (AComponent = FSession) and
     (FSession <> ACRDefaultSession) then
    begin
      Close;
      SessionName := '';
    end;
end;// Notification


//------------------------------------------------------------------------------
// progress event
//------------------------------------------------------------------------------
procedure TACRDatabase.DoOnProgress(
            Progress:   Double;
            Operation:  TACRDatabaseOperation;
            var Abort:  Boolean);
begin
 if Assigned(FOnProgress) then
   FOnProgress(Self,Progress,Operation,Abort);
end; // DoOnProgress


//------------------------------------------------------------------------------
// progress event
//------------------------------------------------------------------------------
procedure TACRDatabase.DoOnTableProgress(
                                  Sender:     TComponent;
                                  Progress:   Double;
                                  Operation:  TACRTableOperation;
                                  var Abort:  Boolean
                           );
begin
 if (Assigned(FOnTableProgress)) then
  FOnTableProgress(Sender,Progress,Operation,Abort);
end; // DoOnTableProgress


//------------------------------------------------------------------------------
// before insert record
//------------------------------------------------------------------------------
procedure TACRDatabase.DoBeforeInsertRecord(
                   Sender:             TACRDataset;
                   const TableName:    String;
                   const FieldValues:  TACRArrayOfTACRVariant;
                   var Abort:          Boolean
                 );
begin
  Abort := False;
  if (Assigned(FBeforeInsertRecord)) then
   FBeforeInsertRecord(Sender,TableName,FieldValues,Abort);
end; // DoBeforeInsertRecord


//------------------------------------------------------------------------------
// after insert record
//------------------------------------------------------------------------------
procedure TACRDatabase.DoAfterInsertRecord(
                   Sender:             TACRDataset;
                   const TableName:    String;
                   const FieldValues:  TACRArrayOfTACRVariant
                 );
begin
  if (Assigned(FAfterInsertRecord)) then
   FAfterInsertRecord(Sender,TableName,FieldValues);
end; // DoAfterInsertRecord


//------------------------------------------------------------------------------
// before update record
//------------------------------------------------------------------------------
procedure TACRDatabase.DoBeforeUpdateRecord(
                   Sender:                TACRDataset;
                   const TableName:       String;
                   const OldFieldValues:  TACRArrayOfTACRVariant;
                   const NewFieldValues:  TACRArrayOfTACRVariant;
                   var Abort:             Boolean
                 );
begin
  if (Assigned(FBeforeUpdateRecord)) then
   FBeforeUpdateRecord(Sender,TableName,OldFieldValues,NewFieldValues,Abort);
end; // DoBeforeUpdateRecord


//------------------------------------------------------------------------------
// after update record
//------------------------------------------------------------------------------
procedure TACRDatabase.DoAfterUpdateRecord(
                   Sender:                TACRDataset;
                   const TableName:       String;
                   const OldFieldValues:  TACRArrayOfTACRVariant;
                   const NewFieldValues:  TACRArrayOfTACRVariant
                 );
begin
  if (Assigned(FAfterUpdateRecord)) then
   FAfterUpdateRecord(Sender,TableName,OldFieldValues,NewFieldValues);
end; // DoAfterUpdateRecord


//------------------------------------------------------------------------------
// before delete record
//------------------------------------------------------------------------------
procedure TACRDatabase.DoBeforeDeleteRecord(
                   Sender:             TACRDataset;
                   const TableName:    String;
                   const FieldValues:  TACRArrayOfTACRVariant;
                   var Abort:          Boolean
                 );
begin
  if (Assigned(FBeforeDeleteRecord)) then
   FBeforeDeleteRecord(Sender,TableName,FieldValues,Abort);
end; // DoBeforeDeleteRecord


//------------------------------------------------------------------------------
// after delete record
//------------------------------------------------------------------------------
procedure TACRDatabase.DoAfterDeleteRecord(
                   Sender:             TACRDataset;
                   const TableName:    String;
                   const FieldValues:  TACRArrayOfTACRVariant
                 );
begin
  if (Assigned(FAfterDeleteRecord)) then
   FAfterDeleteRecord(Sender,TableName,FieldValues);
end; // DoAfterDeleteRecord


//------------------------------------------------------------------------------
// before execute SQL script
//------------------------------------------------------------------------------
procedure TACRDatabase.DoBeforeExecuteSQL(
                   Sender:             TACRQuery;
                   var Abort:          Boolean
                 );
begin
  if (Assigned(FBeforeExecuteSQL)) then
   FBeforeExecuteSQL(Sender,Abort);
end; // DoBeforeExecuteSQL


//------------------------------------------------------------------------------
// after execute SQL script
//------------------------------------------------------------------------------
procedure TACRDatabase.DoAfterExecuteSQL(
                   Sender:             TACRQuery
                 );

begin
  if (Assigned(FAfterExecuteSQL)) then
   FAfterExecuteSQL(Sender);
end; // DoAfterExecuteSQL


//------------------------------------------------------------------------------
// keep connection
//------------------------------------------------------------------------------
procedure TACRDatabase.SetKeepConnection(Value: Boolean);
begin
  if FKeepConnection <> Value then
  begin
    FKeepConnection := Value;
    if not Value and (FRefCount = 0) then
     Close;
  end;
end;// SetKeepConnection


//------------------------------------------------------------------------------
// sets read-only mode
//------------------------------------------------------------------------------
procedure TACRDatabase.SetReadOnly(Value: Boolean);
begin
//  CheckInactive;
  FReadOnly := Value;
end;// SetReadOnly


//------------------------------------------------------------------------------
// set session name
//------------------------------------------------------------------------------
procedure TACRDatabase.SetSessionName(const Value: string);
begin
  if csReading in ComponentState then
    FSessionName := Value
  else
  begin
    CheckInactive;
    if FSessionName <> Value then
    begin
      FSessionName := Value;
      if not (csDestroying in ComponentState) then
        CheckSessionName(False);
    end;
  end;
end;// SetSessionName


//------------------------------------------------------------------------------
// checks session name
//------------------------------------------------------------------------------
procedure TACRDatabase.CheckSessionName(Required: Boolean);
var
  NewSession: TACRSession;
begin
  if Required then
    NewSession := Sessions.List[FSessionName]
  else
    NewSession := Sessions.FindSession(FSessionName);
  if (NewSession <> nil) and (NewSession <> FSession) then
  begin
    if (FSession <> nil) then FSession.RemoveDatabase(Self);
    FSession := NewSession;
    FSession.FreeNotification(Self);
    FSession.AddDatabase(Self);
    try
      ValidateName(FDatabaseName);
    except
      FDatabaseName := '';
      raise;
    end;
  end;
  if Required then FSession.Active := True;
end;// CheckSessionName


//------------------------------------------------------------------------------
// CheckConnected
//------------------------------------------------------------------------------
procedure TACRDatabase.CheckConnected;
begin
  if (Connected) then
    DatabaseError(ErrorADatabaseOpen, Self);
end;// CheckConnected


//------------------------------------------------------------------------------
// CheckDisconnected
//------------------------------------------------------------------------------
procedure TACRDatabase.CheckDisconnected;
begin
  if (not Connected) then
    DatabaseError(ErrorLDatabaseIsNotOpened, Self);
end;// CheckDisconnected


//------------------------------------------------------------------------------
// db connected?
//------------------------------------------------------------------------------
function TACRDatabase.GetConnected: Boolean;
begin
  Result := (FHandle <> nil);
end;// GetConnected


//------------------------------------------------------------------------------
// connected dataset
//------------------------------------------------------------------------------
function TACRDatabase.GetDataSet(Index: Integer): TACRDataSet;
begin
  Result := FDataSets[Index];
end;// GetDataSet


//------------------------------------------------------------------------------
// count of connected datasets
//------------------------------------------------------------------------------
function TACRDatabase.GetDataSetCount: Integer;
begin
  Result := FDataSets.Count;
end;// GetDataSetCount


//------------------------------------------------------------------------------
// opens from existing DB
//------------------------------------------------------------------------------
function TACRDatabase.OpenFromExistingDB: Boolean;
begin
  FHandle := FSession.FindDatabaseHandle(DatabaseName);
  Result := (FHandle <> nil);
  FAcquiredHandle := Result;
end;// OpenFromExistingDB


//------------------------------------------------------------------------------
// validates name
//------------------------------------------------------------------------------
procedure TACRDatabase.ValidateName(const Name: string);
var
  Database: TACRDatabase;
begin
  if (Name <> '') and (FSession <> nil) then
  begin
    Database := FSession.FindDatabase(Name);
    if ((Database <> nil) and (Database <> Self) and
          not (Database.HandleShared and HandleShared)) then
    begin
      if (not Database.Temporary or (Database.FRefCount <> 0)) then
       DatabaseErrorFmt(ErrorADatabaseOpen, [Name], Self);
      Database.Free;
    end;
  end;
end;// ValidateName


//------------------------------------------------------------------------------
// raises exception if not active
//------------------------------------------------------------------------------
procedure TACRDatabase.CheckInactive;
begin
  if FHandle <> nil then
    if (csDesigning in ComponentState) then
      Close
    else
      DatabaseError(ErrorADatabaseOpen, Self);
end;// CheckInactive


//------------------------------------------------------------------------------
// raises exception if database name is not valid
//------------------------------------------------------------------------------
procedure TACRDatabase.CheckDatabaseName;
begin
 if ((FDatabaseName = '') and (not Temporary)) then
    DatabaseError(ErrorADatabaseNameMissing, Self);
end;// CheckDatabaseName


////////////////////////////////////////////////////////////////////////////////
//
//  TACRAdvFieldDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRAdvFieldDef.Assign(Source: TACRAdvFieldDef);
begin
  FName := Source.Name;
  FDataType := Source.DataType;
  FRequired := Source.Required;
  FSize := Source.Size;

  //FDefaultValueType := Source.DefaultValueType;
  FDefaultValue.Assign(Source.DefaultValue,True);
  FMinValue.Assign(Source.FMinValue,True);
  FMaxValue.Assign(Source.MaxValue,True);

  // Blob data
  FBlobCompressionAlgorithm := Source.BlobCompressionAlgorithm;
  FBlobCompressionMode := Source.BlobCompressionMode;
  FBlobBlockSize := Source.BlobBlockSize;

  // Autoinc settings
  FAutoincIncrement := Source.FAutoincIncrement;
  FAutoincInitialValue := Source.FAutoincInitialValue;
  FAutoincMinValue  := Source.FAutoincMinValue;
  FAutoincMaxValue  := Source.FAutoincMaxValue;
  FAutoincCycled    := Source.FAutoincCycled;

end;


//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
constructor TACRAdvFieldDef.Create;
begin
  FName := '';
  FDataType := aftUnknown;
  FRequired := false;
  FSize := 0;

  FDefaultValue := TACRVariant.Create;
  FMinValue := TACRVariant.Create;
  FMaxValue := TACRVariant.Create;

  // Blob data
  FBlobCompressionAlgorithm := caNone;
  FBlobCompressionMode := 0;
  FBlobBlockSize := DefaultBlobBlockSize;

  // Autoinc settings
  FAutoincIncrement := 1;
  FAutoincInitialValue := 0;
  FAutoincMinValue  := 0;//Low(Int64);
  FAutoincMaxValue  := High(Int64);
  FAutoincCycled    := False;

end;//Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRAdvFieldDef.Destroy;
begin
  FDefaultValue.Free;
  FMinValue.Free;
  FMaxValue.Free;
  inherited;
end;//Destroy




////////////////////////////////////////////////////////////////////////////////
//
//  TACRAdvFieldDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRAdvFieldDefs.Create;
begin
  FDefsList := TList.Create;
end;//Create


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRAdvFieldDefs.Assign(Source: TACRAdvFieldDefs);
var i:  Integer;
begin
  Clear;
  for i := 0 to Source.Count - 1 do
    AddFieldDef.Assign(Source[i]);
end;


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRAdvFieldDefs.Destroy;
begin
  Clear;
  FDefsList.Free;
  inherited;
end;//Destroy



//------------------------------------------------------------------------------
// Count
//------------------------------------------------------------------------------
function TACRAdvFieldDefs.GetCount: Integer;
begin
  Result := FDefsList.Count;
end;//GetCount


//------------------------------------------------------------------------------
// GetDef
//------------------------------------------------------------------------------
function TACRAdvFieldDefs.GetDef(Index: Integer): TACRAdvFieldDef;
begin
  Result := TACRAdvFieldDef(FDefsList[Index]);
end;//GetDef


//------------------------------------------------------------------------------
// SetDef
//------------------------------------------------------------------------------
procedure TACRAdvFieldDefs.SetDef(Index: Integer; value: TACRAdvFieldDef);
begin
  if (FDefsList[Index] <> nil) then
    TACRAdvFieldDef(FDefsList[Index]).Free;
  FDefsList[Index] := Value;
end;//SetDef


//------------------------------------------------------------------------------
// AddFieldDef
//------------------------------------------------------------------------------
function TACRAdvFieldDefs.AddFieldDef: TACRAdvFieldDef;
begin
  Result := TACRAdvFieldDef.Create;
  FDefsList.Add(Result);
end;//AddFieldDef


//------------------------------------------------------------------------------
// Add
//------------------------------------------------------------------------------
procedure TACRAdvFieldDefs.Add(const Name: String;
  DataType: TACRAdvancedFieldType; Size: Integer; Required: Boolean);
var TmpAdvFieldDef: TACRAdvFieldDef;
begin
  TmpAdvFieldDef := AddFieldDef;
  TmpAdvFieldDef.Name := Name;
  TmpAdvFieldDef.DataType := DataType;
  TmpAdvFieldDef.Size := Size;
  TmpAdvFieldDef.Required := Required;
end;//Add


//------------------------------------------------------------------------------
// DeleteFieldDef
//------------------------------------------------------------------------------
procedure TACRAdvFieldDefs.DeleteFieldDef(const FieldName: string);
var i: Integer;
begin
  for i:=0 to Count-1 do
   if (AnsiStrIComp(PAnsiChar(TACRAdvFieldDef(Items[i]).Name), PAnsiChar(FieldName)) = 0) then
    begin
      TACRAdvFieldDef(FDefsList[i]).Free;
      FDefsList.Delete(i);
      Exit;
    end;
  raise EACRException.Create(30349, ErrorGFieldWithNameNotFound, [FieldName]);
end;//DeleteFieldDef


//------------------------------------------------------------------------------
// Find
//------------------------------------------------------------------------------
function TACRAdvFieldDefs.Find(const Name: string): TACRAdvFieldDef;
var i: Integer;
begin
  Result := nil;
  for i:=0 to Count-1 do
   if (AnsiStrIComp(PAnsiChar(TACRAdvFieldDef(Items[i]).Name), PAnsiChar(Name)) = 0) then
    begin
      Result := Items[i];
      break;
    end
end;//Find


//------------------------------------------------------------------------------
// varchar exists
//------------------------------------------------------------------------------
function TACRAdvFieldDefs.IsVarcharExists: Boolean;
var i: Integer;
begin
  Result := False;
  for i:=0 to FDefsList.Count-1 do
   if (IsVarcharFieldType(TACRAdvFieldDef(FDefsList[i]).FDataType)) then
    begin
     Result := True;
     break;
    end;
end; // IsVarcharExists


//------------------------------------------------------------------------------
// Clear
//------------------------------------------------------------------------------
procedure TACRAdvFieldDefs.Clear;
var i: Integer;
begin
  for i:=0 to FDefsList.Count-1 do
    TACRAdvFieldDef(FDefsList[i]).Free;
  FDefsList.Clear;
end;//Clear


////////////////////////////////////////////////////////////////////////////////
//
//  TACRForeignKeyDef
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRForeignKeyDef.Assign(Source: TPersistent);
begin
 FName := TACRForeignKeyDef(Source).Name;
 FDeleteAction := TACRForeignKeyDef(Source).DeleteAction;
 FUpdateAction := TACRForeignKeyDef(Source).UpdateAction;
 FMatchType := TACRForeignKeyDef(Source).MatchType;
 FReferencedTableName := TACRForeignKeyDef(Source).ReferencedTableName;
 FColumns := TACRForeignKeyDef(Source).Columns;
 FName := TACRForeignKeyDef(Source).FName;
end; // Assign


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRForeignKeyDef.Create;
begin
 FName := '';
 FDeleteAction := fkaDefault;
 FUpdateAction := fkaDefault;
 FMatchType := fkmtDefault;
 FColumns := '';
 FReferencedTableName := '';
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRForeignKeyDef.Destroy;
begin
  inherited;
end; // Destroy


////////////////////////////////////////////////////////////////////////////////
//
//  TACRForeignKeyDefs
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// GetCount
//------------------------------------------------------------------------------
function TACRForeignKeyDefs.GetCount: Integer;
begin
  Result := FDefsList.Count;
end; // GetCount


//------------------------------------------------------------------------------
// GetDef
//------------------------------------------------------------------------------
function TACRForeignKeyDefs.GetDef(Index: Integer): TACRForeignKeyDef;
begin
  if ((Index < 0) or (Index >= FDefsList.Count)) then
    Result := nil
  else
    Result := FDefsList[Index];
end; // GetDef


//------------------------------------------------------------------------------
// SetDef
//------------------------------------------------------------------------------
procedure TACRForeignKeyDefs.SetDef(Index: Integer; value: TACRForeignKeyDef);
begin
  if (FDefsList[Index] <> nil) then
    TACRForeignKeyDef(FDefsList[Index]).Free;
  FDefsList[Index] := Value;
end; // SetDef


//------------------------------------------------------------------------------
// Assign
//------------------------------------------------------------------------------
procedure TACRForeignKeyDefs.Assign(Source: TPersistent);
var i:  Integer;
begin
  Clear;
  for i := 0 to TACRForeignKeyDefs(Source).Count - 1 do
    AddForeignKeyDef.Assign(TACRForeignKeyDefs(Source)[i]);
end; // Assign


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRForeignKeyDefs.Create;
begin
  FDefsList := TList.Create;
end; // Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRForeignKeyDefs.Destroy;
begin
  Clear;
  FDefsList.Free;
  inherited;
end; // Destroy


//------------------------------------------------------------------------------
// AddForeignKeyDef
//------------------------------------------------------------------------------
function TACRForeignKeyDefs.AddForeignKeyDef: TACRForeignKeyDef;
begin
  Result := TACRForeignKeyDef.Create;
  FDefsList.Add(Result);
end; // AddForeignKeyDef


//------------------------------------------------------------------------------
// Clear
//------------------------------------------------------------------------------
procedure TACRForeignKeyDefs.Add(
             const Name: String;
             const Columns: String;
             const ReferencedTableName: String;
             const MatchType: TACRForeignKeyMatchType;
             const DeleteAction: TACRForeignKeyAction;
             const UpdateAction: TACRForeignKeyAction
             );
var Def: TACRForeignKeyDef;
begin
  Def := AddForeignKeyDef;
  Def.Name := Name;
  Def.Columns := Columns;
  Def.ReferencedTableName := ReferencedTableName;
  Def.MatchType := MatchType;
  Def.DeleteAction := DeleteAction;
  Def.UpdateAction := UpdateAction;
end; // Add


//------------------------------------------------------------------------------
// Delete ForeignKeyDef
//------------------------------------------------------------------------------
procedure TACRForeignKeyDefs.DeleteForeignKeyDef(const Name: string);
var i: Integer;
begin
  i := IndexOf(Name);
  if (i >= 0) then
   Delete(i)
  else
   raise EACRException.Create(11396, ErrorLForeignKeyDefNotFound, [Name]);
end; // DeleteForeignKeyDef


//------------------------------------------------------------------------------
// Delete
//------------------------------------------------------------------------------
procedure TACRForeignKeyDefs.Delete(const Index: Integer);
begin
  if (Index < 0) or (Index >= FDefsList.Count) then
   raise EACRException.Create(11397,ErrorLInvalidItemNumber,[Index,FDefsList.Count]);
  TACRForeignKeyDef(FDefsList[Index]).Free;
  FDefsList.Delete(Index);
end; // Delete


//------------------------------------------------------------------------------
// Find
//------------------------------------------------------------------------------
function TACRForeignKeyDefs.Find(const Name: string): TACRForeignKeyDef;
var i: Integer;
begin
  Result := nil;
  i := IndexOf(Name);
  if (i >= 0) then
   Result := FDefsList[i];
end; // Find


//------------------------------------------------------------------------------
// IndexOf
//------------------------------------------------------------------------------
function TACRForeignKeyDefs.IndexOf(const Name: string): Integer;
var i: Integer;
begin
  Result := -1;
  for i := 0 to FDefsList.Count - 1 do
   if (AnsiUpperCase(TACRForeignKeyDef(FDefsList[i]).Name) = AnsiUpperCase(Name)) then
    begin
     Result := i;
     break;
    end;
end; // IndexOf


//------------------------------------------------------------------------------
// Clear
//------------------------------------------------------------------------------
procedure TACRForeignKeyDefs.Clear;
var i: Integer;
begin
  for i:=0 to FDefsList.Count-1 do
    TACRForeignKeyDef(FDefsList[i]).Free;
  FDefsList.Clear;
end; // Clear


////////////////////////////////////////////////////////////////////////////////
//
// TACRBlobStream
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// sets new size of the stream
//------------------------------------------------------------------------------
procedure TACRBlobStream.InternalSetSize(const NewSize: Int64);
begin
 FBlobStream.Size := NewSize;
 if (FBlobStream.Size = NewSize) then
  FBlobStream.Modified := True;
end; // InternalSetSize


//------------------------------------------------------------------------------
// sets new size of the stream
//------------------------------------------------------------------------------
procedure TACRBlobStream.SetSize(NewSize: Longint);
begin
 InternalSetSize(NewSize);
end; // SetSize


{$IFDEF D6H}
//------------------------------------------------------------------------------
// sets new size of the stream
//------------------------------------------------------------------------------
procedure TACRBlobStream.SetSize(const NewSize: Int64);
begin
 InternalSetSize(NewSize);
end; // SetSize
{$ENDIF}


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRBlobStream.Read(var Buffer; Count: Longint): Longint;
begin
 Result := FBlobStream.Read(Buffer,Count);
end; // Read


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRBlobStream.Write(const Buffer; Count: Longint): Longint;
begin
 Result := FBlobStream.Write(Buffer,Count);
 if (Result > 0) then
   FBlobStream.Modified := True;
end; // Write


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRBlobStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
 Result := FBlobStream.Seek(Offset,Origin);
end; // Seek


{$IFDEF D6H}
//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
function TACRBlobStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
begin
 Result := FBlobStream.Seek(Offset,Origin);
end; // Seek
{$ENDIF}


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
constructor TACRBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
begin
 if ((Mode <> bmRead) and (TACRDataset(Field.DataSet).ReadOnly)) then
  raise EACRException.Create(10106,ErrorLDatasetIsInReadOnlyMode);
 FBlobStream := TACRDataset(Field.DataSet).InternalCreateBlobStream(
  Field.DataSet.FieldByName(Field.FieldName),Mode);
 BlockSize := FBlobStream.BlockSize;
 TACRLocalBlobStream(FBlobStream).UserBlobStream := Self;
 FField := Field;
 FBlobStream.Modified := False;
 FDataSet := FField.DataSet as TACRDataSet;
 if (Mode = bmWrite) then
  Truncate;
 FWideMemo := TACRDataset(Field.DataSet).IsWideMemoField(Field);
{
 if ((Self.Size > 0) and (FWideMemo)) then
  begin
   // convert from wide memo to string
   StrSize := Self.Size;
   SetLength(ws,StrSize div 2);
   OpenMode := TACRLocalBLOBStream(FBLOBStream).OpenMode;
   try
     TACRLocalBLOBStream(FBLOBStream).OpenMode := bomReadWrite;
     Self.Position := 0;
     Self.ReadBuffer(PWideChar(@ws[1])^,StrSize);
     FBLOBStream.Size := 0;
     SetLength(s,StrSize div 2);
     s := WideCharToString(PWideChar(@ws[1]));
//     for i := 1 to StrSize div 2 do
//      s[i] := WideCharToAnsiChar(ws[i]);
     FBlobStream.WriteBuffer(PChar(@s[1])^,StrSize div 2);
   finally
//     MemoryManager.FreeAndNilMem(buf);
     TACRLocalBLOBStream(FBLOBStream).OpenMode := OpenMode ;
   end;
  end; // WideMemo
}
end; // Create


//------------------------------------------------------------------------------
// set size of compressed stream
//------------------------------------------------------------------------------
destructor TACRBlobStream.Destroy;
begin
 Modified := FBlobStream.Modified;
 TACRLocalBlobStream(FBlobStream).UserBlobStream := nil;
{
 if (Modified) then
  if ((Self.Size > 0) and (FWideMemo)) then
   begin
     // convert from wide memo to string
     StrSize := Self.Size;
     buf := MemoryManager.GetMem(StrSize * 2);
     OpenMode := TACRLocalBLOBStream(FBLOBStream).OpenMode;
     try
       TACRLocalBLOBStream(FBLOBStream).OpenMode := bomReadWrite;
       Self.Position := 0;
       SetLength(s,StrSize);
       Self.ReadBuffer(PChar(@s[1])^,StrSize);
       FBLOBStream.Size := 0;
       StringToWideChar(s,buf,StrSize*2);
       FBlobStream.WriteBuffer(buf^,StrSize*2);
     finally
       MemoryManager.FreeAndNilMem(buf);
       TACRLocalBLOBStream(FBLOBStream).OpenMode := OpenMode ;
     end;
   end; // WideMemo
}   
 FDataset.CloseBlob(FField);

 if (Modified) then
  begin
   FField.Modified := True;
   try
     FDataSet.DataEvent(deFieldChange, Longint(FField));
   except
     {$IFDEF D6H}
     ApplicationHandleException(Self);
     {$ELSE}
     Application.HandleException(Self)
     {$ENDIF}
   end;
  end;
 inherited;
end; // Destroy


//------------------------------------------------------------------------------
// truncate
//------------------------------------------------------------------------------
procedure TACRBlobStream.Truncate;
begin
 Size := Position;
end; // Truncate


////////////////////////////////////////////////////////////////////////////////
//
// TACRBatchMove
//
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
// SetMappings
//------------------------------------------------------------------------------
procedure TACRBatchMove.SetMappings(Value: TStrings);
begin
  FMappings.Assign(Value);
end;// SetMappings


//------------------------------------------------------------------------------
// SetSource
//------------------------------------------------------------------------------
procedure TACRBatchMove.SetSource(Value: TACRDataset);
begin
  FSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;// SetSource


//------------------------------------------------------------------------------
// Notification
//------------------------------------------------------------------------------
procedure TACRBatchMove.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if Destination = AComponent then Destination := nil;
    if Source = AComponent then Source := nil;
  end;
end;// Notification


//------------------------------------------------------------------------------
// progress event
//------------------------------------------------------------------------------
procedure TACRBatchMove.DoOnProgress(
                        Progress:   Double;
                        var Abort:  Boolean
                       );
var Operation: TACRTableOperation;
begin
 case FMode of
  batAppend: Operation := tbopBatchAppend;
  batUpdate: Operation := tbopBatchUpdate;
  batAppendUpdate: Operation := tbopBatchAppendUpdate;
  batCopy: Operation := tbopBatchCopy;
  batDelete: Operation := tbopBatchDelete;
 end;
 if (Assigned(FOnProgress)) then
  FOnProgress(Self,Progress,Operation,Abort);
end; // DoOnProgress


//------------------------------------------------------------------------------
// Copy records
//------------------------------------------------------------------------------
procedure TACRBatchMove.CopyRecords(BatchMode:    TACRBatchMode;
                                    FieldCount:   Integer;
                                    SourceFields: array of Integer;
                                    DestFields:   array of Integer
                                    );
var
    CurRecNo: Int64;
    d,MaxRec: Double;
    Abort:    Boolean;

  procedure DoAbort;
  begin
{$IFNDEF SQLMEMTABLE}
   if (FUseTransactions) then
    if (FDestination.Database.InTransaction) then
     FDestination.Database.Rollback;
{$ENDIF}
  end; // DoAbort


  procedure WriteToProblemTable;
  var i: Integer;
  begin
   Inc(FProblemCount);
   if (FProblemTable <> nil) then
    begin
     FProblemTable.Insert;
     for i := 0 to FSource.FieldCount - 1 do
      if (FSource.Fields[i] is TLargeintField) then
       begin
        if (not FSource.Fields[i].IsNull) then
         TLargeintField(FProblemTable.Fields[i]).AsLargeInt :=
          TLargeintField(FSource.Fields[i]).AsLargeInt;
       end
      else
       FProblemTable.Fields[i].Assign(Source.Fields[i]);
     FProblemTable.Post;
    end;
  end; // WriteToProblemTable


  procedure WriteToChangedTable;
  var i: Integer;
  begin
   Inc(FChangedCount);
   if (FChangedTable <> nil) then
    begin
     FChangedTable.Insert;
     for i := 0 to FDestination.FieldCount - 1 do
      if (FDestination.Fields[i] is TLargeintField) then
       begin
        if (not FDestination.Fields[i].IsNull) then
         TLargeintField(FChangedTable.Fields[i]).AsLargeInt :=
          TLargeintField(FDestination.Fields[i]).AsLargeInt;
       end
      else
       FChangedTable.Fields[i].Assign(Destination.Fields[i]);
     FChangedTable.Post;
    end;
  end; // WriteToChangedTable


  procedure WriteToKeyViolTable;
  var i: Integer;
  begin
   Inc(FKeyViolCount);
   if (FKeyViolTable <> nil) then
    begin
     FKeyViolTable.Insert;
     for i := 0 to FSource.FieldCount - 1 do
      if (FSource.Fields[i] is TLargeintField) then
       begin
        if (not FSource.Fields[i].IsNull) then
         TLargeintField(FKeyViolTable.Fields[i]).AsLargeInt :=
          TLargeintField(FSource.Fields[i]).AsLargeInt;
       end
      else
       FKeyViolTable.Fields[i].Assign(Source.Fields[i]);
     FKeyViolTable.Post;
    end;
  end; // WriteToKeyViolTable


  function WriteToDestination: Boolean;
  var i: Integer;
  begin
   Result := True;
   try
     for i := 0 to FSource.FieldCount - 1 do
      if (FSource.Fields[i] is TLargeintField) then
       begin
        if (not FSource.Fields[i].IsNull) then
         begin
          TLargeintField(FDestination.Fields[i]).AsLargeInt :=
           TLargeintField(FSource.Fields[i]).AsLargeInt;
          if (TLargeintField(FDestination.Fields[i]).AsLargeInt <>
              TLargeintField(FSource.Fields[i]).AsLargeInt) then
           begin
            Result := False;
            Exit;
           end;
         end;
       end
      else
       begin
        FDestination.Fields[i].Assign(FSource.Fields[i]);
        if (FDestination.Fields[i].Value <> FSource.Fields[i].Value) then
         begin
          Result := False;
          Exit;
         end;
       end;
   except
     Result := False;
   end;
  end; // WriteToDestination


  procedure DoAppend;
  begin
   Destination.Insert;
   if (not WriteToDestination) then
    begin
     if (FAbortOnProblem) then
      DatabaseError(ErrorLBatchMoveAbortOnProblem,Self)
     else
      begin
       WriteToProblemTable;
       Destination.Cancel;
       Exit;
      end;
    end;
   try
     Destination.Post;
     Inc(FMovedCount);
   except
     on e: Exception do
      if (FAbortOnKeyViol) then
       DatabaseError(ErrorLBatchMoveAbortOnKeyViol+e.Message,Self)
      else
       begin
        Destination.Cancel;
        WriteToKeyViolTable;
       end;
   end;
  end; // DoAppend


  procedure DoAppendUpdateOrUpdate(DoAppendUpdate: Boolean);
  var i,j,SourceFieldNo,DestFieldNo:  Integer;
      DestField:                      TField;
      bUpdate:                        Boolean;
  begin
    Destination.SetKey;
    for i := 0 to Destination.IndexFieldCount-1 do
     begin
      DestField := Destination.GetIndexField(i);
      DestFieldNo := DestField.FieldNo - 1;
      SourceFieldNo := -1;
      for j := 0 to FieldCount-1 do
       if (DestFields[j] = DestFieldNo) then
        begin
         SourceFieldNo := SourceFields[j];
         break;
        end;
      if (SourceFieldNo > -1) then
       if (not FSource.Fields[SourceFieldNo].IsNull) then
         begin
          if (DestField is TLargeintField) then
           TLargeIntField(DestField).AsLargeInt :=
            TLargeIntField(FSource.Fields[SourceFieldNo]).AsLargeInt
          else
           DestField.Assign(FSource.Fields[SourceFieldNo]);
         end;
     end; // scan all index fields in Destination and set key values
    bUpdate := Destination.GotoKey;
    if (bUpdate) then
     begin
      WriteToChangedTable;
      Destination.Edit;
     end
    else
     begin
      if (DoAppendUpdate) then
       Destination.Insert
      else
       Exit; // do nothing
     end;
    if (not WriteToDestination) then
     begin
       if (FAbortOnProblem) then
        DatabaseError(ErrorLBatchMoveAbortOnProblem,Self)
       else
        begin
         WriteToProblemTable;
         Destination.Cancel;
         Exit;
        end;
     end;
    try
     Destination.Post;
     Inc(FMovedCount);
    except
     on e: Exception do
      if (FAbortOnKeyViol) then
       DatabaseError(ErrorLBatchMoveAbortOnKeyViol+e.Message,Self)
      else
       begin
        Destination.Cancel;
        WriteToKeyViolTable;
       end;
    end;
  end; // DoAppendUpdateOrUpdate


  procedure DoDelete;
  var i:                              Integer;
      SourceField:                    TField;
      KeyFields:                      String;
      KeyValues:                      variant;
  begin
    KeyValues := VarArrayCreate([0,(FieldCount-1)],varVariant);
    KeyFields := '';
    for i := 0 to FieldCount-1 do
     begin
      SourceField := FSource.Fields[SourceFields[i]];
      KeyValues[i] := SourceField.AsVariant;
      if (KeyFields = '') then
       KeyFields := FDestination.Fields[DestFields[i]].FieldName
      else
       KeyFields := KeyFields + ';'+FDestination.Fields[DestFields[i]].FieldName;
     end;
    if (Destination.LocateRecord(KeyFields,KeyValues,[])) then
     begin
      try
       WriteToChangedTable;
       Destination.Delete;
       Inc(FMovedCount);
      except
       on e: Exception do
        if (FAbortOnKeyViol) then
         DatabaseError(ErrorLBatchMoveAbortOnKeyViol+e.Message,Self)
        else
         WriteToKeyViolTable;
      end;
     end;
  end; // DoDelete


  function IsBatchmoveComplete: Boolean;
  begin
   Result := (Source.Eof) or ((CurRecNo >= FRecordCount) and (FRecordCount > 0));
  end;

begin
  CurRecNo := 0;
  Abort := False;
  Source.First;
  if (FRecordCount > 0) then
   MaxRec := FRecordCount
  else
   MaxRec := Source.RecordCount;
  FMovedCount := 0;
  FProblemCount := 0;
  FKeyViolCount := 0;
  FChangedCount := 0;
  while (not IsBatchmoveComplete) do
   begin
{$IFNDEF SQLMEMTABLE}
    if (FUseTransactions and (not FDestination.Database.InTransaction)) then
      FDestination.Database.StartTransaction;
{$ENDIF}
    case (BatchMode) of
     batAppend:        DoAppend;
     batAppendUpdate:  DoAppendUpdateOrUpdate(True);
     batUpdate:        DoAppendUpdateOrUpdate(False);
     batDelete:        DoDelete;
    end;
    Source.Next;
    Inc(CurRecNo);
    // commit if necessary
{$IFNDEF SQLMEMTABLE}
    if (FUseTransactions) then
     if (FCommitCount > 0) then
      if ((CurRecNo mod FCommitCount) = 0) then
        FDestination.Database.Commit(IsBatchmoveComplete);
{$ENDIF}
    // display progress
    d := CurRecNo / MaxRec;
    DoOnProgress(d,Abort);
    if (Abort) then
     begin
      DoAbort;
      Exit;
     end;
   end; // perform batch move
{$IFNDEF SQLMEMTABLE}
 if (FDestination.Database.InTransaction) then
  FDestination.Database.Commit(True);
{$ENDIF}
end; // CopyRecords


//------------------------------------------------------------------------------
// Create
//------------------------------------------------------------------------------
constructor TACRBatchMove.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAbortOnKeyViol := True;
  FAbortOnProblem := True;
  FRecordCount := 0;
{$IFNDEF SQLMEMTABLE}
  FUseTransactions := True;
  FCommitCount := ACRDefaultCommitCount;
{$ENDIF}
  FMappings := TStringList.Create;
  FProblemTableName := '';
  FKeyViolTableName := '';
  FChangedTableName := '';
end;// Create


//------------------------------------------------------------------------------
// Destroy
//------------------------------------------------------------------------------
destructor TACRBatchMove.Destroy;
begin
  FMappings.Free;
  inherited Destroy;
end;// Destroy


//------------------------------------------------------------------------------
// Execute
//------------------------------------------------------------------------------
procedure TACRBatchMove.Execute;
var
  SourceActive, DestinationActive: Boolean;
  BatchMode:                       TACRBatchMode;
  i, FieldCount:                   Integer;
  SourceFields,DestFields:         array of Integer;
  DestName, SourceName:            String;
  SourceCursorPos:                 Pointer;

  procedure GetMappingNames;
  var
    P:        Integer;
    Mapping:  String;
  begin
    Mapping := FMappings[I];
    P := Pos('=', Mapping);
    if P > 0 then
     begin
      DestName := Copy(Mapping, 1, P - 1);
      SourceName := Copy(Mapping, P + 1, 255);
     end
    else
     begin
      DestName := Mapping;
      SourceName := Mapping;
     end;
  end; // GetMappingNames


  procedure CreateProblemTable;
  begin
    FProblemTable := TACRTable.Create(Self);
    FProblemTable.DatabaseName := Destination.DatabaseName;
    FProblemTable.InMemory := Destination.InMemory;
    FProblemTable.TableName := FProblemTableName;
    FProblemTable.FExclusive := True;
    FProblemTable.AdvFieldDefs.Assign(Source.AdvFieldDefs);
    FProblemTable.CreateTable;
    FProblemTable.Open;
  end; // CreateProblemTable


  procedure CreateKeyViolTable;
  begin
    FKeyViolTable := TACRTable.Create(Self);
    FKeyViolTable.DatabaseName := Destination.DatabaseName;
    FKeyViolTable.InMemory := Destination.InMemory;
    FKeyViolTable.TableName := FKeyViolTableName;
    FKeyViolTable.FExclusive := True;
    FKeyViolTable.AdvFieldDefs.Assign(Source.AdvFieldDefs);
    FKeyViolTable.CreateTable;
    FKeyViolTable.Open;
  end; // CreateProblemTable


  procedure CreateChangedTable;
  begin
    FChangedTable := TACRTable.Create(Self);
    FChangedTable.DatabaseName := Destination.DatabaseName;
    FChangedTable.InMemory := Destination.InMemory;
    FChangedTable.TableName := FChangedTableName;
    FChangedTable.FExclusive := True;
    FChangedTable.AdvFieldDefs.Assign(Destination.AdvFieldDefs);
    FChangedTable.CreateTable;
    FChangedTable.Open;
  end; // CreateChangedTable


begin
  if ((Destination = nil) or (Source = nil) or (Destination = Source)) then
    DatabaseError(ErrorLInvalidBatchMove, Self);
{$IFNDEF SQLMEMTABLE}
  if (FUseTransactions and (CommitCount <= 0)) then
    DatabaseError(ErrorLInvalidBatchMoveCommitCount, Self);
{$ENDIF}
  SourceActive := Source.Active;
  DestinationActive := Destination.Active;
  SourceFields := nil;
  DestFields := nil;
  FProblemTable := nil;
  FChangedTable := nil;
  FKeyViolTable := nil;
  BatchMode := FMode;
  try
    Source.DisableControls;
    Destination.DisableControls;
    if (not SourceActive) then
      Source.Open;
    Source.CheckBrowseMode;
    Source.UpdateCursorPos;
    if (SourceActive) then
     SourceCursorPos := Source.Handle.SavePosition;
    if (BatchMode = batCopy) then
     begin
      Destination.Close;
      if (FMappings.Count = 0) then
        Destination.AdvFieldDefs.Assign(Source.AdvFieldDefs)
      else
       begin
        Destination.AdvFieldDefs.Clear;
        for I := 0 to FMappings.Count - 1 do
         begin
          GetMappingNames;
          with Source.AdvFieldDefs.Find(SourceName) do
            Destination.AdvFieldDefs.Add(DestName, DataType, Size, Required);
         end;
      end;
      Destination.IndexDefs.Clear;
      Destination.IndexDefs.Assign(Source.IndexDefs);
      Destination.FieldDefs.Clear;
      Destination.CreateTable;
      BatchMode := batAppend;
     end; // Copy
    Destination.Open;
    Destination.CheckBrowseMode;
    Destination.UpdateCursorPos;
    if ((FMode = batAppendUpdate) or (FMode = batUpdate)) then
     if (Destination.IndexFieldCount <= 0) then
      DatabaseError(ErrorLInvalidBatchMoveDestIndex,Self);
    FieldCount := Destination.AdvFieldDefs.Count;
    if (FProblemTableName <> '') then
     CreateProblemTable;
    if (FKeyViolTableName <> '') then
     CreateKeyViolTable;
    if (FChangedTableName <> '') then
     CreateChangedTable;
    SetLength(SourceFields, FieldCount);
    SetLength(DestFields, FieldCount);
    if (FMappings.Count <> 0) then
     begin
      for i := 0 to FMappings.Count - 1 do
       begin
        GetMappingNames;
        SourceFields[i] := Source.FieldDefs.Find(SourceName).FieldNo-1;
        DestFields[i] := Destination.FieldDefs.Find(DestName).FieldNo-1;
       end;
     end
    else
     begin
      for i := 0 to FieldCount-1 do
       begin
        SourceFields[i] := i;
        DestFields[i] := i;
       end;
     end;
{$IFNDEF SQLMEMTABLE}
    if (FUseTransactions) then
     Destination.Database.StartTransaction;
{$ENDIF}
    try
     try
      CopyRecords(BatchMode,FieldCount,SourceFields,DestFields);
     except
{$IFNDEF SQLMEMTABLE}
      if (FUseTransactions) then
       if (Destination.Database.InTransaction) then
        Destination.Database.Rollback;
{$ENDIF}
      raise;
     end;
    finally
      SourceFields := nil;
      DestFields := nil;
      if (DestinationActive) then
        Destination.First;
      if (SourceActive) then
       begin
        Source.Handle.RestorePosition(SourceCursorPos);
        Source.Handle.FreePosition(SourceCursorPos);
        Source.UpdateCursorPos;
       end;
    end;
  finally
    if (not DestinationActive) then
      Destination.Close;
    if (not SourceActive) then
      Source.Close;
    if (FProblemTable <> nil) then
     FProblemTable.Free;
    if (FChangedTable <> nil) then
     FChangedTable.Free;
    if (FKeyViolTable <> nil) then
     FKeyViolTable.Free;
    Destination.EnableControls;
    Source.EnableControls;
  end;
end;// Execute


////////////////////////////////////////////////////////////////////////////////
//
// Global functions
//
////////////////////////////////////////////////////////////////////////////////


procedure ACRFixForeignKeysSelfReferences(TableName, NewTableName: String; ForeignKeyDefs: TACRForeignKeyDefs);
var i:       Integer;
    s:       String;
begin
  s := AnsiUpperCase(TableName);
  for i := 0 to ForeignKeyDefs.Count - 1 do
   begin
    if (s = AnsiUpperCase(ForeignKeyDefs.Items[i].ReferencedTableName)) then
     ForeignKeyDefs.Items[i].ReferencedTableName := NewTableName;
   end;
end; // ACRFixForeignKeysSelfReferences


//------------------------------------------------------------------------------
// convert foreign keys to constraint defs
//------------------------------------------------------------------------------
procedure ACRConvertForeignKeyDefToConstraintDef(
            ForeignKeyDef: TACRForeignKeyDef;
            ConstraintDef: TACRConstraintDefForeignKey
                            );
var j:       Integer;
    sl:      TStringList;
begin
  sl := TStringList.Create;
  try
    ConstraintDef.Name := ForeignKeyDef.Name;
    ConstraintDef.ReferencedTableName := ForeignKeyDef.ReferencedTableName;
    ConstraintDef.ObjectID := INVALID_OBJECT_ID;
    ConstraintDef.MatchType := TACRConstraintForeignKeyMatchType(Byte(ForeignKeyDef.MatchType));
    ConstraintDef.DeleteAction := TACRConstraintForeignKeyAction(Byte(ForeignKeyDef.DeleteAction));
    ConstraintDef.UpdateAction := TACRConstraintForeignKeyAction(Byte(ForeignKeyDef.UpdateAction));
    sl.Clear;
    GetNamesList(sl,ForeignKeyDef.Columns);
    if (sl.Count <= 0) then
     raise EACRException.Create(11418,ErrorLNoColumnsInForeignKeyDefinition,[ForeignKeyDef.Name,ForeignKeyDef.ReferencedTableName]);
    SetLength(ConstraintDef.Columns,sl.Count);
    for j := 0 to sl.Count - 1 do
     begin
      ConstraintDef.Columns[j].ColumnName := sl.Strings[j];
      ConstraintDef.Columns[j].ColumnObjectID := INVALID_OBJECT_ID;
     end;
  finally
    sl.Free;
  end;
end; // ACRConvertForeignKeyDefToConstraintDef


//------------------------------------------------------------------------------
// convert foreign keys to constraint defs
//------------------------------------------------------------------------------
procedure ACRConvertForeignKeyDefsToConstraintDefs(
            ForeignKeyDefs: TACRForeignKeyDefs;
            ConstraintDefs: TACRConstraintDefs
            );
var i:      Integer;
    FKDef:  TACRForeignKeyDef;
    FK:     TACRConstraintDefForeignKey;
begin
    for i := 0 to ForeignKeyDefs.Count - 1 do
     begin
      FKDef := ForeignKeyDefs.Items[i];
      FK := ConstraintDefs.AddFK;
      ACRConvertForeignKeyDefToConstraintDef(FKDef,FK);
     end;
end; // ConvertForeignKeyDefsToConstraintDefs



//------------------------------------------------------------------------------
// convert constraint defs to foreign keys
//------------------------------------------------------------------------------
procedure ConvertConstraintDefsToForeignKeyDefs(
            ConstraintDefs: TACRConstraintDefs;
            ForeignKeyDefs: TACRForeignKeyDefs
            );
var i,j:    Integer;
    FKDef:  TACRForeignKeyDef;
    FK:     TACRConstraintDefForeignKey;
begin
  ForeignKeyDefs.Clear;
  for i := 0 to ConstraintDefs.Count - 1 do
   if (ConstraintDefs.Items[i].ConstraintType = ctFK) then
    begin
      FK := TACRConstraintDefForeignKey(ConstraintDefs.Items[i]);
      FKDef := ForeignKeyDefs.AddForeignKeyDef;
      FKDef.Name := FK.Name;
      FKDef.ReferencedTableName := FK.ReferencedTableName;
      FKDef.MatchType := TACRForeignKeyMatchType(Byte(FK.MatchType));
      FKDef.DeleteAction := TACRForeignKeyAction(Byte(FK.DeleteAction));
      FKDef.FUpdateAction := TACRForeignKeyAction(Byte(FK.UpdateAction));
      FKDef.Columns := '';
      for j := 0 to High(FK.Columns) do
       if (FKDef.Columns = '') then
        FKDef.Columns := FK.Columns[j].ColumnName
       else
        FKDef.Columns := FKDef.Columns + ';' + FK.Columns[j].ColumnName;
    end;
end; // ConvertForeignKeyDefsToConstraintDefs


//------------------------------------------------------------------------------
// convert TFieldDefs to ACRFieldDefs
//------------------------------------------------------------------------------
procedure ConvertFieldDefsToACRFieldDefs(
                FieldDefs:      TFieldDefs;
                ACRFieldDefs:   TACRFieldDefs
                                        );
var i:        Integer;
    FieldDef: TACRFieldDef;
begin
 ACRFieldDefs.Clear;
 for i := 0 to FieldDefs.Count-1 do
  begin
   FieldDef := ACRFieldDefs.AddCreated;
   FieldDef.Name := FieldDefs[i].Name;
   FieldDef.SetFieldDefDataType(
      FieldTypeToACRAdvFieldType(FieldDefs[i].DataType),
      FieldDefs[i].Size
                               );
  end;
end; // ConvertFieldDefsToACRFieldDefs



//------------------------------------------------------------------------------
// convert AdvFieldDefs to FieldDefs
//------------------------------------------------------------------------------
procedure ConvertAdvFieldDefsToFieldDefs(AdvFieldDefs: TACRAdvFieldDefs; FieldDefs: TFieldDefs);
var
  i: Integer;
  FieldDef: TFieldDef;
begin
  FieldDefs.Clear;
  for i:=0 to AdvFieldDefs.Count-1 do
   begin
    FieldDef := FieldDefs.AddFieldDef;
    FieldDef.Name := AdvFieldDefs[i].Name;
    FieldDef.DataType := ACRAdvFieldTypeToFieldType(AdvFieldDefs[i].DataType);
    FieldDef.Size := AdvFieldDefs[i].Size;
    //FieldDef.Precision := AdvFieldDefs[i].Precision;
    FieldDef.Required := AdvFieldDefs[i].Required;
    //FieldDef.FieldNo := i + 1;
   end;
end;

//------------------------------------------------------------------------------
// convert FieldDefs to AdvFieldDefs
//------------------------------------------------------------------------------
procedure ConvertFieldDefsToAdvFieldDefs(FieldDefs: TFieldDefs; AdvFieldDefs: TACRAdvFieldDefs);
var
  i: Integer;
  AdvFieldDef: TACRAdvFieldDef;
begin
  AdvFieldDefs.Clear;
  for i:=0 to FieldDefs.Count-1 do
   begin
    AdvFieldDef := AdvFieldDefs.AddFieldDef;
    AdvFieldDef.Name := FieldDefs[i].Name;
    AdvFieldDef.DataType := FieldTypeToACRAdvFieldType(FieldDefs[i].DataType);
    AdvFieldDef.Size := FieldDefs[i].Size;
    //AdvFieldDef.Precision := FieldDefs[i].Precision;
    AdvFieldDef.Required := FieldDefs[i].Required;
   end;
end;

//------------------------------------------------------------------------------
// convert AdvFieldDefs to ACRFieldDefs
//------------------------------------------------------------------------------
procedure ConvertAdvFieldDefsToACRFieldDefs(
                AdvFieldDefs:      TACRAdvFieldDefs;
                ACRFieldDefs:      TACRFieldDefs;
                IndexDefs:         TACRIndexDefs;
                ACRConstraintDefs: TACRConstraintDefs;
                Temporary:         Boolean
                                        );
var i: Integer;
    FieldDef: TACRFieldDef;
begin
 // Clear Lists
 ACRFieldDefs.Clear;
 ACRConstraintDefs.Clear;

 // Fill ACRFieldDefs
 for i := 0 to AdvFieldDefs.Count-1 do
  begin
   // Fill ACRFieldDefs
   FieldDef := ACRFieldDefs.AddCreated;
   FieldDef.Name := AdvFieldDefs[i].Name;
   FieldDef.SetFieldDefDataType(
      AdvFieldDefs[i].DataType,
      AdvFieldDefs[i].Size
                               );

   FieldDef.AutoincIncrement := AdvFieldDefs[i].AutoincIncrement;
   FieldDef.AutoincInitialValue := AdvFieldDefs[i].AutoincInitialValue;
   FieldDef.AutoincMinValue  := AdvFieldDefs[i].AutoincMinValue;
   FieldDef.AutoincMaxValue  := AdvFieldDefs[i].AutoincMaxValue;
   FieldDef.AutoincCycled    := AdvFieldDefs[i].AutoincCycled;

   FieldDef.DefaultValue.Assign(AdvFieldDefs[i].DefaultValue);

   FieldDef.BlobCompressionAlgorithm :=
     ConvertCompressionAlgorithmToACRCompressionAlgorithm(
      AdvFieldDefs[i].BlobCompressionAlgorithm);
   FieldDef.BlobCompressionMode := AdvFieldDefs[i].BlobCompressionMode;
   FieldDef.BlobBlockSize := AdvFieldDefs[i].BlobBlockSize;


   if (not Temporary) then
    begin
     if (AdvFieldDefs[i].Required) then
       ACRConstraintDefs.AddNotNull.ColumnName := AdvFieldDefs[i].Name;

     // Fill ACRConstraintDefs Check
     if ((not AdvFieldDefs[i].MinValue.IsNull) or (not AdvFieldDefs[i].MaxValue.IsNull)) then
      with ACRConstraintDefs.AddCheck do
      begin
        ColumnName := AdvFieldDefs[i].Name;
        MinValue.Assign(AdvFieldDefs[i].MinValue);
        MaxValue.Assign(AdvFieldDefs[i].MaxValue);
      end;
    end;//if

  end;//for

 if (not Temporary) then
   for i := 0 to IndexDefs.Count-1 do
     AddConstraintForIndex(IndexDefs[i], ACRConstraintDefs);

end;//ConvertAdvFieldDefsToACRFieldDefs


//------------------------------------------------------------------------------
// Add Unic or Primary Key constraint
//------------------------------------------------------------------------------
function AddConstraintForIndex(
                                  IndexDef:           TACRIndexDef;
                                  ACRConstraintDefs:  TACRConstraintDefs
                                 ): TACRConstraintDef;
var
  j: Integer;
  ConstraintDefPK: TACRConstraintDefPrimary;
  ConstraintDefUnique: TACRConstraintDefUnique;
begin
  Result := nil;
  if (IndexDef.Primary) then
    begin
     ConstraintDefPK := ACRConstraintDefs.AddPK;
     ConstraintDefPK.IndexName := IndexDef.Name;
     ConstraintDefPK.IndexObjectId := IndexDef.ObjectId;
     ConstraintDefPK.Name := IndexDef.Name;
//     ConstraintDefPK.Name := AutoNameConstraintPKPreffix;

     // Constraint Fields...
     SetLength(ConstraintDefPK.Columns, IndexDef.ColumnCount);
     for j:=0 to IndexDef.ColumnCount-1 do
      begin
       ConstraintDefPK.Columns[j].ColumnName := IndexDef.Columns[j].FieldName;
//       ConstraintDefPK.Name := ConstraintDefPK.Name + AutoNameSymbol +
//                               IndexDef.Columns[j].FieldName;

      end;
     Result := ConstraintDefPK;
    end
   else if (IndexDef.Unique) then
    begin
     ConstraintDefUnique := ACRConstraintDefs.AddUnique;
     ConstraintDefUnique.IndexName := IndexDef.Name;
     ConstraintDefUnique.IndexObjectId := IndexDef.ObjectId;
     ConstraintDefUnique.Name := IndexDef.Name;
//     ConstraintDefUnique.Name :=  AutoNameConstraintUniquePreffix;

     // Constraint Fields...
     SetLength(ConstraintDefUnique.Columns, IndexDef.ColumnCount);
     for j:=0 to IndexDef.ColumnCount-1 do
      begin
       ConstraintDefUnique.Columns[j].ColumnName := IndexDef.Columns[j].FieldName;
//       ConstraintDefUnique.Name := ConstraintDefUnique.Name + AutoNameSymbol +
//                               IndexDef.Columns[j].FieldName;

      end;
     Result := ConstraintDefUnique;
    end;
end;//AddCoustraintFromIndex


//------------------------------------------------------------------------------
// convert ACRFieldDefs to AdvFieldDefs
//------------------------------------------------------------------------------
procedure ConvertACRFieldDefsToAdvFieldDefs(
                VisibleFieldDefs:   TACRFieldDefs;
                ACRFieldDefs:   TACRFieldDefs;
                ACRConstraintDefs: TACRConstraintDefs;
                AdvFieldDefs:   TACRAdvFieldDefs
                                        );
var i,j,k,n:      Integer;
    AdvFieldDef: TACRAdvFieldDef;
begin
 // Clear Lists
 AdvFieldDefs.Clear;

 // Fill AdvFieldDefs
 for i := 0 to VisibleFieldDefs.Count-1 do
  begin
   AdvFieldDef := AdvFieldDefs.AddFieldDef;

   // Fill AdvFieldDefs
   AdvFieldDef.Name := VisibleFieldDefs[i].Name;
   AdvFieldDef.DataType := VisibleFieldDefs[i].AdvancedFieldType;
   AdvFieldDef.Size := VisibleFieldDefs[i].FieldSize;

   AdvFieldDef.AutoincIncrement := VisibleFieldDefs[i].AutoincIncrement;
   AdvFieldDef.AutoincInitialValue := VisibleFieldDefs[i].AutoincInitialValue;
   AdvFieldDef.AutoincMinValue   := VisibleFieldDefs[i].AutoincMinValue;
   AdvFieldDef.AutoincMaxValue  := VisibleFieldDefs[i].AutoincMaxValue;
   AdvFieldDef.AutoincCycled := VisibleFieldDefs[i].AutoincCycled;

   AdvFieldDef.BlobCompressionAlgorithm :=
    ConvertACRCompressionAlgorithmToCompressionAlgorithm(
       VisibleFieldDefs[i].BlobCompressionAlgorithm);
   AdvFieldDef.BlobCompressionMode := VisibleFieldDefs[i].BlobCompressionMode;
   AdvFieldDef.BlobBlockSize := VisibleFieldDefs[i].BlobBlockSize;

   AdvFieldDef.DefaultValue.Assign(VisibleFieldDefs[i].DefaultValue);
  end;

 // Fill Constraints Data
 for i := 0 to ACRConstraintDefs.Count-1 do
  begin
   case ACRConstraintDefs[i].ConstraintType of
     ctNotNull:
       begin
         n := ACRFieldDefs.GetDefNumberByObjectId(
                    TACRConstraintDefNotNull(ACRConstraintDefs[i]).ColumnObjectID);
         if n = -1 then
           raise EACRException.Create(30032, ErrorGFieldWithObjectIdNotFound,
                  [TACRConstraintDefNotNull(ACRConstraintDefs[i]).ColumnObjectID]);
         k := -1;
         for j := 0 to VisibleFieldDefs.Count - 1 do
           if (VisibleFieldDefs[j].FieldNoReference = n) then
            begin
             k := j;
             break;
            end;
         if (k > -1) then
          begin
           // Requared
           AdvFieldDefs[k].FRequired := True;
          end;
       end;
     ctCheck:
       begin
         n := ACRFieldDefs.GetDefNumberByObjectId(
                    TACRConstraintDefCheck(ACRConstraintDefs[i]).ColumnObjectID);
         if n = -1 then
           raise EACRException.Create(30031, ErrorGFieldWithObjectIdNotFound,
                  [TACRConstraintDefCheck(ACRConstraintDefs[i]).ColumnObjectID]);
         k := -1;
         for j := 0 to VisibleFieldDefs.Count - 1 do
           if (VisibleFieldDefs[j].FieldNoReference = n) then
            begin
             k := j;
             break;
            end;
         if (k > -1) then
          begin
           // Min
           AdvFieldDefs[k].MinValue.Assign(TACRConstraintDefCheck(ACRConstraintDefs[i]).MinValue);
           // Max
           AdvFieldDefs[k].MaxValue.Assign(TACRConstraintDefCheck(ACRConstraintDefs[i]).MaxValue);
          end;
       end;
     ctPK,
     ctFK,
     ctFKAction,
     ctUnique: ; // Nothing to do
     else
        raise EACRException.Create(30039, ErrorGNotImplementedYet);
   end;
  end;

end;//ConvertACRFieldDefsToAdvFieldDefs



//------------------------------------------------------------------------------
// convert ACRFieldDefs to TFieldDefs
//------------------------------------------------------------------------------
procedure ConvertACRFieldDefsToFieldDefs(
                ACRFieldDefs:   TACRFieldDefs;
                ACRConstraintDefs:  TACRConstraintDefs;
                FieldDefs:      TFieldDefs
                                        );
var i,j,n,Size: Integer;
    fd:    TFieldDef;
begin

 FieldDefs.Clear;
 for i := 0 to ACRFieldDefs.Count-1 do
  begin
   Size := ACRFieldDefs[i].FieldSize;
//   if (IsWideStringFieldType(ACRFieldDefs[i].AdvancedFieldType)) then
//    Size := Size * 2;
   FieldDefs.Add(
                  ACRFieldDefs[i].Name,
                  ACRAdvFieldTypeToFieldType(ACRFieldDefs[i].AdvancedFieldType),
                  Size,
                  False
                );
  end;

 for i := 0 to ACRConstraintDefs.Count-1 do
  if ACRConstraintDefs[i].ConstraintType = ctNotNull then
   begin
    n := TACRConstraintDefNotNull(ACRConstraintDefs[i]).ColumnObjectID;
    j := ACRFieldDefs.GetDefNumberByObjectId(n);
    if j <> -1 then
      begin
        fd := FieldDefs.Find(ACRFieldDefs[j].Name);
        if (fd <> nil) then
          fd.Required := true;
      end;
   end;


end; //


//------------------------------------------------------------------------------
// get string list from string with names
//------------------------------------------------------------------------------
procedure GetNamesList(List: TStrings; const Names: string);
var
  Pos: Integer;
  NewNames: String;
begin
  Pos := 1;
  NewNames := StringReplace(Names, ',', ';', [rfReplaceAll]);
  if Assigned(List) then
   begin
    List.Clear;
    while Pos <= Length(NewNames) do
     List.Add(ExtractFieldName(NewNames, Pos));
   end;
end;// GetNamesList


//------------------------------------------------------------------------------
// fill ACRIndexDef
//------------------------------------------------------------------------------
procedure FillACRIndexDef(
              ACRIndexDef:         TACRIndexDef;
              const Name,
              Fields:              String;
              Options:             TIndexOptions;
              const DescFields:    String;
              const CaseInsFields: String;
              FieldDefs:           TFieldDefs;
              AdvFieldDefs:        TACRAdvFieldDefs
                           );
var
  j:      Integer;
  FieldList, DescFieldList, CaseInsFieldList: TStringList;
begin
 FieldList := TStringList.Create;
 DescFieldList := TStringList.Create;
 CaseInsFieldList := TStringList.Create;
 try
     if (Name = '') then
      raise EACRException.Create(20048, ErrorAInvalidIndexName, ['']);
     if (Fields = '') then
      raise EACRException.Create(20049, ErrorACannotFindIndexField, ['']);

     ACRIndexDef.Name := Name;
     ACRIndexDef.IndexType := itBTree;
     ACRIndexDef.Unique := (ixUnique in Options);
     ACRIndexDef.Primary := (ixPrimary in Options);
     GetNamesList(FieldList, Fields);
     GetNamesList(DescFieldList, DescFields);
     GetNamesList(CaseInsFieldList, CaseInsFields);
     ACRIndexDef.ColumnCount := FieldList.Count;
     for j := 0 to FieldList.Count-1 do
      begin
       if (FieldDefs.IndexOf(FieldList.Strings[j]) = -1) then
        if (AdvFieldDefs.Find(FieldList.Strings[j]) = nil) then
         raise EACRException.Create(20050, ErrorACannotFindIndexField,
                                [FieldList.Strings[j]]);
       ACRIndexDef.Columns[j].FieldName := FieldList.Strings[j];
       ACRIndexDef.Columns[j].Descending :=
         (DescFieldList.IndexOf(FieldList.Strings[j]) >= 0) or
         (ixDescending in Options);
       ACRIndexDef.Columns[j].CaseInsensitive :=
         (CaseInsFieldList.IndexOf(FieldList.Strings[j]) >= 0) or
         (ixCaseInsensitive in Options);
      end;
 finally
   FieldList.Free;
   DescFieldList.Free;
   CaseInsFieldList.Free;
 end;
end;// FillACRIndexDef


//------------------------------------------------------------------------------
// convert TIndexDef to TACRIndexDef
//------------------------------------------------------------------------------
procedure ConvertIndexDefToACRIndexDef(
                IndexDef:      TIndexDef;
                ACRIndexDef:   TACRIndexDef;
                FieldDefs:     TFieldDefs;
                AdvFieldDefs:  TACRAdvFieldDefs
                                       );
begin
 FillACRIndexDef(ACRIndexDef,
                 IndexDef.Name, IndexDef.Fields, IndexDef.Options,
                 IndexDef.DescFields, IndexDef.CaseInsFields, FieldDefs, AdvFieldDefs);
end;// ConvertIndexDefToACRIndexDef


//------------------------------------------------------------------------------
// convert TIndexDefs to ACRIndexDefs
//------------------------------------------------------------------------------
procedure ConvertIndexDefsToACRIndexDefs(
                IndexDefs:      TIndexDefs;
                ACRIndexDefs:   TACRIndexDefs;
                FieldDefs:      TFieldDefs;
                AdvFieldDefs:   TACRAdvFieldDefs
                                         );
var
  i:      Integer;
begin
  ACRIndexDefs.Clear;
  for i := 0 to IndexDefs.Count-1 do
   ConvertIndexDefToACRIndexDef(IndexDefs[i], ACRIndexDefs.AddCreated, FieldDefs, AdvFieldDefs);
end;// ConvertIndexDefsToACRIndexDefs


//------------------------------------------------------------------------------
// convert ACRIndexDefs to TIndexDefs
//------------------------------------------------------------------------------
procedure ConvertACRIndexDefsToIndexDefs(
                ACRIndexDefs:   TACRIndexDefs;
                IndexDefs:      TIndexDefs
                                         );
var
  i,j: Integer;
  Options: TIndexOptions;
  Fields, DescFields, CaseInsFields: String;
begin
 IndexDefs.Clear;
 for i := 0 to ACRIndexDefs.Count-1 do
  begin
   Fields := '';
   DescFields := '';
   CaseInsFields := '';
   for j := 0 to ACRIndexDefs[i].ColumnCount-1 do
    begin
      if (Fields <> '') then
       Fields := Fields + ';' + ACRIndexDefs[i].Columns[j].FieldName
      else
       Fields := ACRIndexDefs[i].Columns[j].FieldName;

      if (ACRIndexDefs[i].Columns[j].Descending) then
       if (DescFields <> '') then
        DescFields := DescFields + ';' + ACRIndexDefs[i].Columns[j].FieldName
       else
        DescFields := ACRIndexDefs[i].Columns[j].FieldName;

      if (ACRIndexDefs[i].Columns[j].CaseInsensitive) then
       if (CaseInsFields <> '') then
        CaseInsFields := CaseInsFields + ';' + ACRIndexDefs[i].Columns[j].FieldName
       else
        CaseInsFields := ACRIndexDefs[i].Columns[j].FieldName;
    end;

   Options := [];
   if (ACRIndexDefs[i].Unique) then
    Options := Options + [ixUnique];
   if (ACRIndexDefs[i].Primary) then
    Options := Options + [ixPrimary];

   IndexDefs.Add(ACRIndexDefs[i].Name, Fields, Options);
   IndexDefs[IndexDefs.Count-1].DescFields := DescFields;
   IndexDefs[IndexDefs.Count-1].CaseInsFields := CaseInsFields;
  end;
end;// ConvertACRIndexDefsToIndexDefs


//------------------------------------------------------------------------------
// return true if field exists
//------------------------------------------------------------------------------
function FindFieldInFieldDefs(FieldDefs: TFieldDefs; FieldName : string): Boolean;
var i: integer;
    f: Boolean;
begin
 f := false;
 for i := 0 to FieldDefs.Count -1 do
   if (AnsiUpperCase(FieldDefs.Items[i].Name) = AnsiUpperCase(FieldName)) then
    begin
      f := true;
      break;
    end;
 Result := f;
end; // FindFieldInFieldDefs


//------------------------------------------------------------------------------
// compression algorithm
//------------------------------------------------------------------------------
function ConvertCompressionAlgorithmToACRCompressionAlgorithm(
    CompressionAlgorithm: TCompressionAlgorithm
          ): TACRCompressionAlgorithm;
begin
 Result := acaNone;
 case (CompressionAlgorithm) of
  caZLIB: Result := acaZLIB;
  caBZIP: Result := acaBZIP;
  caPPM: Result := acaPPM;
 end;
end; // ConvertCompressionAlgorithmToACRCompressionAlgorithm


//------------------------------------------------------------------------------
// compression algorithm
//------------------------------------------------------------------------------
function ConvertACRCompressionAlgorithmToCompressionAlgorithm(
            CompressionAlgorithm: TACRCompressionAlgorithm
          ): TCompressionAlgorithm;
begin
 Result := caNone;
 case (CompressionAlgorithm) of
  acaZLIB: Result := caZLIB;
  acaBZIP: Result := caBZIP;
  acaPPM: Result := caPPM;
 end;
end; // ConvertACRCompressionAlgorithmToCompressionAlgorithm



//------------------------------------------------------------------------------
// copy records and return error log
//------------------------------------------------------------------------------
function CopyDatasets(
            SourceDataset:                TDataset;
            DestinationDataset:           TDataset;
            UseSourceDatasetForProgress:  Boolean;
            Operation:                    TACRTableOperation
            ): String;
var
    CurrentValue:     Int64;
    CurrentRecordNo:  Int64;
    SourceRecCount:   Integer;
    i:                Integer;
    Bookmark:         TBookmark;
    Progress:         Double;
    bStart:           Boolean;
    bCalculateMinMax: Boolean;
    var min,max:      array of Int64;
    FieldNo:          Integer;
    ProgressShown:    Integer;

    function ShowProgress(Progress: Double): Boolean;
    begin
     Result := False;
     if (UseSourceDatasetForProgress) then
      TACRTable(SourceDataset).DoOnProgress(Progress,Operation,Result)
     else
      TACRTable(DestinationDataset).DoOnProgress(Progress,Operation,Result);
    end;

begin
 Result := '';
 CurrentRecordNo := 1;
 SourceDataset.CheckBrowseMode;
 SourceRecCount := SourceDataset.RecordCount;
 if (ShowProgress(0)) then
  begin
   Result := ErrorLOperationAborted;
   Exit;
  end;
 ProgressShown := 0;
 Bookmark := SourceDataset.GetBookmark;
 bStart := true;
 bCalculateMinMax := ((Operation = tbopRepair) or
                      (Operation = tbopImport) or
                      (Operation = tbopRestructure)) and
                     (DestinationDataset is TACRDataset);
 if (bCalculateMinMax) then
  begin
   SetLength(min,DestinationDataset.FieldCount);
   SetLength(max,DestinationDataset.FieldCount);
  end;
 try
   try
     SourceDataset.First;
   except
     Result := Result + ErrorLCannotRetrieveFirstRecord;
     Exit;
   end;
   while not SourceDataset.Eof do
    begin
      DestinationDataset.Insert;
      for i := 0 to SourceDataset.FieldCount - 1 do
       begin
        FieldNo := DestinationDataset.FieldDefs.IndexOf(SourceDataset.Fields[i].FieldName);
        if (FieldNo >= 0) then
         if (not SourceDataset.Fields[i].IsNull) then
          try
            if (DestinationDataset.Fields[FieldNo] is TLargeintField) then
             begin
              TLargeintField(DestinationDataset.Fields[FieldNo]).AsLargeInt :=
                TLargeintField(SourceDataset.Fields[i]).AsLargeInt
             end
            else
            if (DestinationDataset.Fields[FieldNo] is TBytesField) then
             begin
              DestinationDataset.Fields[FieldNo].AsString := SourceDataset.Fields[i].AsString;
             end
            else
             begin
               DestinationDataset.Fields[FieldNo].Assign(SourceDataset.Fields[i]);
             end;
            // calculate autoinc min or max values
            if (bCalculateMinMax) then
             if (IsAutoincFieldType(TACRDataset(DestinationDataset).AdvFieldDefs[FieldNo].DataType)) then
//             and
//                (not TACRDataset(DestinationDataset).AdvFieldDefs[FieldNo].AutoincCycled)) then
              begin
                if (DestinationDataset.Fields[FieldNo] is TLargeintField) then
                 CurrentValue := TLargeintField(DestinationDataset.Fields[FieldNo]).AsLargeInt
                else
                 CurrentValue := DestinationDataset.Fields[FieldNo].AsInteger;
                if (TACRDataset(DestinationDataset).AdvFieldDefs[FieldNo].FAutoincIncrement >= 0) then
                 begin
                  if (bStart) then
                   max[FieldNo] := CurrentValue
                  else
                   if (CurrentValue > max[FieldNo]) then
                     max[FieldNo] := CurrentValue;
                 end // increment >= 0
                else
                 begin
                  if (bStart) then
                   min[FieldNo] := CurrentValue
                  else
                   if (CurrentValue < min[FieldNo]) then
                     min[FieldNo] := CurrentValue;
                 end; // increment < 0
                bStart := False;
              end;
          except
           on e: Exception do
            begin
              Result := Result + Format(ErrorLCopyTableInvalidFieldValue,
                [CurrentRecordNo,SourceDataset.Fields[i].FieldName,e.Message]);
            end;
          end; // copy field value
       end; // copying fields
      try
       DestinationDataset.Post;
      except
       on e: Exception do
         Result := Result + Format(ErrorLCopyTablePostFailed,[CurrentRecordNo,e.Message]);
      end;
      Inc(CurrentRecordNo);
      try
       SourceDataset.Next;
      except
       Result := Result + #13#10 + Format(ErrorLCannotRetrieveRecord,[CurrentRecordNo,CurrentRecordNo-1]);
       Exit;
      end;
      Progress := CurrentRecordNo * 100.0 / SourceRecCount;
      if (Trunc(Progress) > ProgressShown) then
       begin
        if (ShowProgress(Progress)) then
         begin
          Result := Result + #13#10 + 'Operation aborted';
          break;
         end;
        ProgressShown := Trunc(Progress);
       end;
    end;
 finally
   if (bCalculateMinMax and (not bStart)) then
    begin
      for i := 0 to DestinationDataset.FieldCount - 1 do
       if (IsAutoincFieldType(TACRDataset(DestinationDataset).AdvFieldDefs[i].DataType)) then
//          (not TACRDataset(DestinationDataset).AdvFieldDefs[i].AutoincCycled)) then
        begin
         if (TACRDataset(DestinationDataset).AdvFieldDefs[i].AutoincIncrement >= 0) then
          CurrentValue := max[i]
         else
          CurrentValue := min[i];
         TACRTable(DestinationDataset).SetLastAutoincValue(CurrentValue,i);
        end;
      min := nil;
      max := nil;
    end;
   SourceDataset.GotoBookmark(Bookmark);
   SourceDataset.FreeBookmark(Bookmark);
   ShowProgress(100);
 end;
end; // CopyDatasets




var
  DBDatas: TList;

{$IFDEF MEMORY_ENGINE}
  MemDBData: TACRMemoryDatabaseData;
{$ENDIF}
{$IFDEF TEMPORARY_ENGINE}
  TempDBData: TACRTemporaryDatabaseData;
{$ENDIF}


initialization
{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRMain> initialization started');
{$ENDIF}

  {$IFDEF DEBUG_MEMCHECK}
  MemChk;
  {$ENDIF}


  ACRDatasets := TThreadList.Create;
  ACRDatasetsList := nil;
  DBDataList := TThreadList.Create;
  DBDatas := nil;
 {$IFDEF MEMORY_ENGINE}
  MemDBData := TACRMemoryDatabaseData.Create;
  DBDatas := DBDataList.LockList;
  DBDatas.Add(MemDBData);
  DBDataList.UnlockList;
 {$ENDIF}
 {$IFDEF TEMPORARY_ENGINE}
  TempDBData := TACRTemporaryDatabaseData.Create;
  DBDatas := DBDataList.LockList;
  DBDatas.Add(TempDBData);
  DBDataList.UnlockList;
 {$ENDIF}


  IsDesignMode := False;
  Sessions := TACRSessionList.Create;
  Session := TACRSession.Create(nil);
  Session.SessionName := 'Default';

{$IFDEF DEBUG_LOG_INIT}
aaWriteToLog('ACRMain> initialization finished');
{$ENDIF}

finalization
  Sessions.Free;

  DBDatas := DBDataList.LockList;
  while (DBDatas.Count > 0) do
    TACRDatabaseData(DBDatas.Items[0]).Free;
  DBDataList.UnlockList;
  DBDataList.Free;
  DBDataList := nil;

  ACRDatasets.Free;
  ACRDatasets := nil;
end.
